﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Linq.Expressions;

namespace Vita.Common {

  // Funclet is a wrapper around Func<> delegate. It is used for packing Func objects in dictionary. 
  // Funclets are used in data filters (in authorization). 
  // Note: We construct the Funclet from expression (not from delegate directly) just for the purpose 
  // of exc reporting or ToString() functionality - we want to have a nice representation of an expression
  public abstract class Funclet {
    public readonly Expression Expression;

    public Funclet(Expression expression) {
      Expression = expression; 
    }
    public abstract object Evaluate(params object[] args);

    public override string ToString() {
      return Expression == null ? this.GetType().Name : Expression.ToString();
    }
  }

  public class Funclet<T1, TResult> : Funclet {
    Func<T1, TResult> _compiledDelegate;

    public Funclet(Expression<Func<T1, TResult>> expression) : base(expression) {
      _compiledDelegate = (Expression == null) ? GetDefault : expression.Compile(); 
    }
    
    public override object Evaluate(params object[] args) {
      try {
        return _compiledDelegate((T1)args[0]);
      } catch (Exception ex) {
        ex.AddValue("Expression", this.ToString());
        throw; 
      }
    }

    public TResult Evaluate(T1 arg) {
      return _compiledDelegate(arg);
    }

    private TResult GetDefault(T1 arg) {
      return default(TResult); 
    }
  }//class

  // This Funclet is used in RecordFilter construction, with TResult = bool
  public class Funclet<T1, T2, TResult> : Funclet {
    Func<T1, T2, TResult> _compiledDelegate;

    public Funclet(Expression<Func<T1, T2, TResult>> expression) : base(expression) {
      _compiledDelegate = (expression == null) ? GetDefault : expression.Compile(); 
    }
    
    public override object Evaluate(params object[] args)  {
      return Evaluate((T1)args[0], (T2) args[1]);  
    }
    
    public TResult Evaluate(T1 arg1, T2 arg2) {
      try {
        return _compiledDelegate(arg1, arg2);
      } catch (Exception ex) {
        ex.AddValue("Expression", this.ToString());
        throw;
      }
    }

    private TResult GetDefault(T1 arg1, T2 arg2) {
      return default(TResult); 
    }
  }//class

}
