﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace Vita.Common {

  /// <summary>Provides simple interprocess event signaling. Used in cache invalidation.</summary>
  public class GlobalEventSignaler {
    public readonly string SignalName;
    public readonly TimeSpan Timeout;
    public event EventHandler<SignalEventArgs> Signaled;
    public int SignalDuration = 100; //ms
    EventWaitHandle _waitHandle;
    int _timeoutMs;
    bool _ignoreSelfSignal;
    long _activeSignalCount;
    bool _unloading;

    public GlobalEventSignaler(string signalName, TimeSpan timeout, bool ignoreSelfSignal = true) {
      SignalName = signalName; 
      Timeout = timeout;
      _ignoreSelfSignal = ignoreSelfSignal;
      _timeoutMs = (int) Timeout.TotalMilliseconds; 
      _waitHandle = new EventWaitHandle(false, EventResetMode.ManualReset, signalName);
      AppDomain.CurrentDomain.DomainUnload += CurrentDomain_DomainUnload;
      var task = new Task(Listen);
      task.Start(); 
    }

    void CurrentDomain_DomainUnload(object sender, EventArgs e) {
      //exit gracefully
      _unloading = true;
      _waitHandle.Set();
      Thread.Yield();
    }

    #region comments
    // Note: there's some annoyance in the way EventWaitHandle works, so here is an explanation of the workaround. 
    // Our goal is to propagate signal to all listening (waiting) threads, and to do it just once, if possible. 
    // With EventResetMode.AutoReset, calling Set() releases just one thread and resets the handle - so only one listener receives the signal. 
    // For AutoResetMode.ManualReset, it releases all waiting threads but does not reset the handle at all. 
    // The signaling thread must do it, but no earlier than all listeners got the signal. We try to achieve it using two delays. 
    // When signalling, we call Set(), then sleep for SignalDuration ms, then reset the wait handle - hoping that the delay is enough to wakeup all threads. 
    // When receiving the signal in Listen method, we add Sleep(SignalDuration * 2), to avoid multiple activations in a loop while the sending thread sleeps. 
    // Finally, to avoid a delay on the main signaling thread, we do the signaling on a secondary thread.  
    #endregion

    public void Signal() {
      var task = TaskHelper.NewTask(SignalAsync);
      task.Start(); 
      Thread.Yield(); 
    }

    private void SignalAsync() {
      Interlocked.Increment(ref _activeSignalCount); 
      _waitHandle.Set();
      Thread.Sleep(SignalDuration);
      _waitHandle.Reset();
      Interlocked.Decrement(ref _activeSignalCount);
    }

    private void Listen() {
      while (true) {
        var signaled = _waitHandle.WaitOne(_timeoutMs);
        if (_unloading)
          return; 
        bool ignore = false; 
        if (_ignoreSelfSignal) {
          var activeCount = Interlocked.Read(ref _activeSignalCount);
          ignore = (activeCount > 0);
        }
        if (!ignore)
          OnSignaled(timedOut: !signaled); // if WaitOne returns true, it indicates real signal event; if false, it is timeout
        Thread.Sleep(SignalDuration * 2); 
      }
    }

    private void OnSignaled(bool timedOut) {
      var temp = Signaled;
      if (temp != null)
        Signaled(this,  new SignalEventArgs(timedOut));
    }

  } //class

  public class SignalEventArgs : EventArgs {
    public readonly bool TimedOut;
    public SignalEventArgs(bool timedOut) {
      TimedOut = timedOut;
    }
  }
}//ns
