﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Xml;

namespace Vita.Common {

  //Helper class for creating HTML pages to view info in a browser. The main purpose is to create rough-looking but 
  // useful pages with content like meta-data, error log, etc. These pages are not for end-user, but for dev-only use. 
  // System WEB assemblies have good classes for HTML construction, but we want to avoid referencing these assemblies 
  // in VITA core and modules assemblies. 
  public class HtmlPageBuilder {

    public static string DefaultStyle =
      "table { border-collapse: collapse; border-spacing: 0px; font-family: Verdana; margin: 0em 0em 20px 20px;} " +
      "table th { border-right: 2px white solid; border-bottom: 1.5px white solid; font-weight: bold; background-color: #cecf9c;} " +
      "table td { border-right: 2px white solid; border-bottom: 1.5px white solid; background-color: #e5e5cc;font-family: Verdana;font-size: 12px;} " +
      "h1 { background-color: #003366; border-bottom: #336699 6px solid; color: #ffffff; font-family: Tahoma; font-size: 26px; font-weight: normal;margin: 0em 0em 10px -20px; padding-bottom: 8px; padding-left: 30px;padding-top: 16px;} " +
      "h2 { background-color: #003366; border-bottom: #336699 6px solid; color: #ffffff; font-family: Tahoma; font-size: 20px; font-weight: normal;margin: 0em 0em 10px -20px; padding-bottom: 8px; padding-left: 30px;padding-top: 16px;} " +
      "h3 { background-color: #ffffff; ; color: #000000; font-family: Tahoma; font-size: 20px; font-weight: bold; padding-bottom: 10px; padding-left: 10px;padding-top: 10px;} " +
      ".homelink { font-family: arial; color:gray; font-size: 10px; text-align:center; }" +
      ".footer { font-family: arial; color:gray; font-size: 10px; text-align:center; }" +
      ".prevnext { font-family: arial; color:gray; font-size: 12px; text-align:left; }";

    protected string Style;
    private MemoryStream _stream;
    private XmlWriter _writer;

    public HtmlPageBuilder(string style = null) {
      Style = style ?? DefaultStyle;
      //create stream and writer
      _stream = new MemoryStream(8192);
      var stt = new XmlWriterSettings();
      stt.Indent = true;
      stt.OmitXmlDeclaration = true;
      _writer = XmlWriter.Create(_stream, stt);
    }

    public XmlReader GetReader() {
      _stream.Position = 0;
      var reader = XmlReader.Create(_stream);
      return reader;
    }

    public string GetHtml() {
      _stream.Position = 0;
      var reader = new StreamReader(_stream);
      var html = reader.ReadToEnd();
      return html;
    }

    //Just in case derived class needs access to the writer
    protected XmlWriter GetWriter() {
      return _writer; 
    }

    #region Wrappers for common XmlWriter methods
    protected void WriteString(string text) {
      _writer.WriteString(text);
    }
    protected void WriteRaw(string data) {
      _writer.WriteRaw(data);
    }
    protected void WriteStartElement(string localName) {
      _writer.WriteStartElement(localName);
    }
    protected void WriteEndElement() {
      _writer.WriteEndElement(); 
    }
    protected void WriteElement(string localName, string value) {
      _writer.WriteElementString(localName, value);
    }
    protected void WriteAttribute(string localName, string value) {
      _writer.WriteAttributeString(localName, value);
    }
    #endregion

    protected void WriteAnchor(string text, string url, string target = null, string cssClass = null) {
      WriteStartElement("a");
      WriteAttribute("href", url);
      WriteAttribute("rel", "noreferrer");
      if (!string.IsNullOrEmpty(target))
        WriteAttribute("target", target);
      if (!string.IsNullOrEmpty(cssClass))
        WriteAttribute("class", cssClass);
      _writer.WriteString(text);
      _writer.WriteEndElement(); //a
    }

    protected void WriteTableHeader(params string[] captions) {
      WriteStartElement("table");
      WriteStartElement("tr");
      foreach (var caption in captions)
        WriteElement("th", caption);
      WriteEndElement();
    }

    protected void WriteMultilineText(string text) {
      if (string.IsNullOrWhiteSpace(text)) return; 
      WriteStartElement("p");
      var lines = text.Split(new string[] {Environment.NewLine}, StringSplitOptions.None);
      foreach(var line in lines) {
        WriteString(line);
        WriteRaw("<br/>");
      }
      WriteEndElement(); // /p
    }

    protected void BeginPage(string title, bool writeHeader = true) {
      WriteStartElement("html");
      WriteStartElement("head");
      WriteElement("title", title);
      WriteElement("style", Style);
      WriteEndElement();
      WriteStartElement("body");
      if (writeHeader)
        WriteElement("h1", title);
    }

    protected void EndPage() {
      WriteRaw("<br/>");
      WriteRaw("<hr/>");
      WriteFooter();
      WriteEndElement(); // body 
      WriteEndElement(); // html 
      _writer.Flush();
      //DEBUG only: dump all to debug output
      //if (System.Diagnostics.Debugger.IsAttached)
        // System.Diagnostics.Debug.WriteLine(GetHtml());
    }

    protected void WriteHomeLink(string text, string url) {
      WriteAnchor(text, url, null, "homelink");
    }

    protected void WriteFooter() {
      WriteStartElement("div");
      WriteAttribute("class", "footer");
      WriteString("Powered by ");
      WriteAnchor("VITA", "http://vita.codeplex.com", "_blank");
      WriteEndElement(); //div
    }
  }//class

}//ns
