﻿using System;
using System.Collections.Generic;
using System.Collections.Concurrent;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Threading;
using System.Diagnostics;

namespace Vita.Common {

  /// <summary>Used as a shared object to invalidate sets of entities in cache, without actually expelling them from cache immediately.</summary>
  public class SharedCacheStatus {
    public bool Invalidated;
  }

  public class ObjectCache<TKey, TValue> {
    #region CachedValueBox nested class
    class CachedValueBox {
      public readonly TKey Key; 
      public readonly TValue Value;
      public readonly long AddedOn;
      public long LastUsedOn;
      public readonly SharedCacheStatus SharedStatus;

      public CachedValueBox(TKey key, TValue value, long addedOn, SharedCacheStatus status) {
        Key = key; 
        Value = value;
        AddedOn = LastUsedOn = addedOn;
        SharedStatus = status; 
      }
      public override string ToString() {
        return Key + ":" + Value;
      }
    }
    #endregion

    public event EventHandler Cleanup;
    
    public readonly int Capacity;
    private int _expirationTimeMs; //ms
    private int _cleanupIntervalMs; // ms
    private bool _unloading; 

    ConcurrentDictionary<TKey, CachedValueBox> _data; 
    long _lastTickReading;
    SharedCacheStatus _dummySharedStatus; //used when client does not care about shared status
    ManualResetEventSlim _cleanupEvent;

    public ObjectCache(int capacity = 10000, TimeSpan? expirationTime = null, TimeSpan? cleanupInterval = null) {
      Capacity = Math.Max(10, capacity);
      // defaults are 60 and 30 seconds
      var expTimeMs = expirationTime == null ? 60 * 1000 : (int) expirationTime.Value.TotalMilliseconds;
      var cleanupMs = cleanupInterval == null ? 30 * 1000 : (int)cleanupInterval.Value.TotalMilliseconds; 
      _expirationTimeMs = Math.Max(100, expTimeMs); //no less than 100 ms
      _cleanupIntervalMs = Math.Max(100, cleanupMs);
      _cleanupIntervalMs = Math.Min(_cleanupIntervalMs, _expirationTimeMs); // cleanup interval is no less than expiration time
      _data = new ConcurrentDictionary<TKey, CachedValueBox>();
      _dummySharedStatus = new SharedCacheStatus();
      _cleanupEvent = new ManualResetEventSlim();
      _lastTickReading = GetCurrentMilliseconds();
      AppDomain.CurrentDomain.DomainUnload += CurrentDomain_DomainUnload;
      var cleanupTask = TaskHelper.NewTask(CleanupContinously);
      cleanupTask.Start(); 
    }

    void CurrentDomain_DomainUnload(object sender, EventArgs e) {
      _unloading = true;
      _cleanupEvent.Set(); //to exit gracefully
      Thread.Yield(); 
    }

    public override string ToString() {
      return string.Format("[Count={0}]", Count);
    }

    private long GetCurrentMilliseconds() {
      return Environment.TickCount;
    }

    public void Add(TKey key, TValue value, SharedCacheStatus sharedStatus = null) {
      sharedStatus = sharedStatus ?? _dummySharedStatus;
      _data[key] = new CachedValueBox(key, value, _lastTickReading, sharedStatus);
      if (_data.Count > Capacity)
        _cleanupEvent.Set(); //request immediate cleanup
    }

    public TValue Lookup(TKey key) {
      CachedValueBox vbox;
      if (!_data.TryGetValue(key, out vbox))
        return default(TValue);
      if (vbox.SharedStatus.Invalidated || vbox.AddedOn + _expirationTimeMs < _lastTickReading) {
        _data.TryRemove(key, out vbox);
        return default(TValue);
      }
      vbox.LastUsedOn = _lastTickReading;
      return vbox.Value; 
    }

    public bool TryRemove(TKey key) {
      CachedValueBox dummy;
      return _data.TryRemove(key, out dummy);
    }

    public int Count {
      get { return _data.Count; }
    }

    public void Clear() {
      _lastTickReading = GetCurrentMilliseconds();
      _data.Clear();
    }

    public void RunCleanup() {
      //wait until current cleanup is finished, and force one more
      while (_cleanupInProgress)
        Thread.Sleep(1);
      DoCleanup(); 
    }

    public void RefreshTime() {
      _lastTickReading = GetCurrentMilliseconds();
    }

    // Private stuff --------------------------------------------------------------------
    private void CleanupContinously() {
      while (true) {
        try {
          _cleanupEvent.Wait(_cleanupIntervalMs); // either wait for interval, or do immediate cleanup when signalled (on capacity overflow)
          if (_unloading)
            return; 
          DoCleanup();
          _cleanupEvent.Reset(); 
        } catch (Exception ex) {
          Trace.WriteLine("Object cache exception:\r\n" + ex.ToString());
        }
      }// while true
    }

    bool _cleanupInProgress; // a flag to prevent concurrent cleanup; 
    private void DoCleanup() {
      if (_cleanupInProgress)
        return;
      _cleanupInProgress = true;
      try {
        DoCleanupCycle(); //do at least once
        // If still over capacity (items were added while we were doing cleanup) - do it again and again
        while (_data.Count > Capacity) 
          DoCleanupCycle();
      } finally {
        _cleanupInProgress = false; 
      }
    }//method

    private void DoCleanupCycle() {
      _lastTickReading = GetCurrentMilliseconds();
      var vboxes = _data.Values.ToList();
      // Check if we need to limit count  
      var targetCount = Capacity;
      if (vboxes.Count > Capacity) {
        // sort with latest used first
        vboxes.Sort((x, y) => y.LastUsedOn.CompareTo(x.LastUsedOn));
        targetCount = (int)(Capacity * 0.8); //reduce to 80% of Capacity
      }
      CachedValueBox dummyCRef;
      int count = 0;
      foreach (var vbox in vboxes) {
        bool remove = count > targetCount || vbox.AddedOn + _expirationTimeMs < _lastTickReading;
        if (remove)
          _data.TryRemove(vbox.Key, out dummyCRef);
        else
          count++;
      }
      if (Cleanup != null)
        Cleanup(this, EventArgs.Empty);
    }//method

  }//class
}//ns
