﻿using System;
using System.Linq;
using System.Collections;
using System.Collections.Generic;
using System.Globalization;
using System.Xml;
using System.Linq.Expressions;

namespace Vita.Common {

  public static class Util {
    public static Exception Rethrow(Exception inner, string message, params object[] args) {
      var msg = SafeFormat(message, args);
      throw new Exception(msg, inner);
    }
    public static Exception Throw(string message, params object[] args) {
      var msg = SafeFormat(message, args);
      throw new Exception(msg);
    }
    public static Exception NotImplemented(string message = null, params object[] args) {
      message = message?? "Method not implemented.";
      message = SafeFormat(message, args); 
      throw new NotImplementedException(message); 
    }
    public static void Check(bool condition, string message, params object[] args) {
      if(!condition)
        Throw(message, args);
    }

    public static void CheckNotEmpty(string value, string message, params object[] args) {
      if (string.IsNullOrWhiteSpace(value))
        Throw(message, args);
    }

    public static void CheckAllNotNull(IEnumerable values, string message, params object[] args) {
      if (args == null)
        return; 
      var index = 0;
      var nullIndexes = new List<int>(); 
      foreach (var value in values) {
        if (value == null)
          nullIndexes.Add(index);
        index++;
      }
      if (nullIndexes.Count == 0) 
        return; 
      var msg = SafeFormat(message, args) + " (nulls at indexes " + string.Join(",", nullIndexes) + ")";
      Throw(message);
    }

    public static void Validate(bool condition, string message, params object[] args) {
      if (condition) return;
      var msg = SafeFormat(message, args);
      var ve = new Vita.Entities.ValidationException(msg);
      throw ve; 
    }

    public static void ValidateNotEmpty(string value, string message, params object[] args) {
      Validate(!string.IsNullOrWhiteSpace(value), message, args);
    }

    public static string Quote(this string value) {
      if (value == null) return "''";
      return "'" + value.Replace("'", "''") + "'";
    }

    public static string FirstCap(this string value) {
      if (value == null || value.Length == 0 || char.IsUpper(value[0])) return value;
      return char.ToUpperInvariant(value[0]) + value.Substring(1);
    }

    public static string TrimSuffix(this string value, string suffix) {
      if (value == null || suffix == null || value.Length <= suffix.Length) 
        return value;
      if (value.EndsWith(suffix, StringComparison.InvariantCultureIgnoreCase))
        return value.Substring(0, value.Length - suffix.Length);
      else 
        return value; 
    }

    public static string FormatUri(this string template, params object[] args) {
      if (args == null || args.Length == 0)
        return template;
      var sArgs = args.Select( a=> EscapeForUri(a)).ToArray(); //escape
      return string.Format(template, sArgs);
    }

    public static string EscapeForUri(object value) {
      return value == null ? string.Empty : Uri.EscapeDataString(value.ToString());
    }
    
    public static string[] EscapeManyForUri(object[] values) {
      if (values == null)
        return new string[] {};
      return values.Select(v => EscapeForUri(v)).ToArray();
    } 

    public static string SafeFormat(string message, params object[] args) {
      if (args == null || args.Length == 0)
        return message;
      try {
        return string.Format(CultureInfo.InvariantCulture, message, args);
      } catch(Exception ex) {
        return message + " (System error: failed to format message. " + ex.Message + ")";
      }
    }

    public static string[] SafeSplit(string value, params char[] separators) {
      if (string.IsNullOrWhiteSpace(value)) 
        return new string[]{};
      if (separators == null || separators.Length == 0)
        separators = new char[] { ',', ';' };
      return value.Split(separators);
    }




    public static string Pluralize(string name) {
      if (name.EndsWith("y"))
        return name.Substring(0, name.Length - 1) + "ies";
      if (name.EndsWith("s"))
        return name + "es";
      return name + "s";
    }

    public static string Unpluralize(string name) {
      if (name.EndsWith("ies")) // Categories
        return name.Substring(0, name.Length - 3) + "y";
      if (name.EndsWith("s")) //Orders
        return name.Substring(0, name.Length - 1);
      return name;
    }

    private static int _hashCount;
    public static int NewHash() {
      _hashCount++;
      var sHash = _hashCount + "_" + _hashCount;
      var code = sHash.GetHashCode();
      return code; 
    }

    public static DateTime CheckKind(this DateTime dateTime, DateTimeKind kindIfUndefined = DateTimeKind.Local) {
      if (dateTime.Kind != DateTimeKind.Unspecified)
        return dateTime;
      var result = new DateTime(dateTime.Ticks, kindIfUndefined);
      return result;
    }

    public static string CheckLength(string value, int maxLen) {
      if (string.IsNullOrEmpty(value)) return value;
      if (value.Length <= maxLen)
        return value;
      value = value.Substring(0, maxLen - 4) + " ...";
      return value; 
    }

    //.NET framework's GetHashCode() is not garanteed to be stable between .NET versions. 
    // If we want to keep hashes in database, we need a stable hash implementation 
    public static int StableHash(this string value) {
      int result = 0;
      var chars = value.ToCharArray();
      for (int i = 0; i < chars.Length; i++)
        result = result * 1013  + chars[i] * 7;
      return result; 
    }

    public static void Each<T>(this IEnumerable<T> list, Action<T> action) {
      foreach (var elem in list)
        action(elem); 
    }

    public static bool ContainsAll<T>(this IList<T> list, IList<T> subList) {
      foreach (var t in subList)
        if (!list.Contains(t)) return false;
      return true; 
    }

    public static T[] EmptyArray<T>() {
      return new T[] { };
    }



  }//class

}//namespace
