﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;

using Vita.Common;
using Vita.Entities;
using Vita.Entities.Model;
using Vita.Entities.Runtime;

namespace Vita.Data {

  [Flags]
  public enum DataStoreLinkOptions {
    None = 0x0,
    ReadPrimary = 0x1,
    ReadSecondary = 0x2,
    WritePrimary = 0x10,
    WriteSecondary = 0x20,
    LinqProvider = 0x100,

    Read = ReadPrimary | ReadSecondary,
    Write = WritePrimary | WriteSecondary,
  }

  //DataStoreLink represents an association between an EntityStore and a DataStore. One entity store can be linked to one or more Data stores through a list of DataStoreLink objects. 
  public class DataStoreLink {
    public string Name; 
    public IDataStore DataStore;
    public DataStoreLinkOptions Options;
    public StringSet Schemas = new StringSet(); 
    //Flags and last availability time
    public bool Enabled;
    public bool Available;
    public DateTime UnavailableSince;
    public DateTime LastAvailable;

    public DataStoreLink(string name, IDataStore dataStore, DataStoreLinkOptions options, bool enabled = true, bool available = true) {
      Name = name;
      DataStore = dataStore;
      Options = options;
      Enabled = enabled;
      Available = available; 
      Schemas.UnionWith(dataStore.GetSchemas()); 
    }
  }//class

  // MultiStore - a data store consisting of multiple "real" data stores underneath - databases/data sources, possibly identical or with similar structures; 
  // When entity store is connected to MultiStore, read operations are performed on one, PrimaryRead store (or SecondaryRead if primary is down);
  // write operations are performed against all underlying stores. Some underlying stores might be unavailable for some time, the system would continue to run
  // as long as there's at least one available. When underlying data store comes back to life, it might need "resync" first - this is not implemented yet.
  // This will be responsibility of a separate background-running synchronization service.
  // Some scenarios: 
  //   1. Client app with MultiStore consisting of connection to local database and REST connection to remote central database
  //   2. Server app with MultiStore with several identical databases on separate servers - for duplication and protection against hardware failure
  public class MultiStore : IDataStore {
    public List<DataStoreLink> Links = new List<DataStoreLink>();
    private int _opCount; //used for "occasional" checking of the connectivity - every 5th op or smth like this.
    private StringSet _schemas = new StringSet();
    //TODO: implement using TimeService
    //private ITimeService _timeService; 

    public MultiStore() {
    }

    public DataStoreLink AddStore(string name, IDataStore dataStore, DataStoreLinkOptions options) {
      _schemas.UnionWith(dataStore.GetSchemas());
      var link = new DataStoreLink(name, dataStore, options);
      Links.Add(link);
      return link; 
    }

    public DataStoreLink GetStore(string name) {
      return Links.FirstOrDefault(lnk => lnk.Name == name); 
    }

    #region IDataStore members
    public void Init(EntityStore entityStore) {
      foreach (var link in Links)
        link.DataStore.Init(entityStore); 
    }

    public IEnumerable<string> GetSchemas() {
      return _schemas; 
    }


    public IList<EntityRecord> ExecuteQuery(EntitySession session, IEntityCommand command, Paging paging, object[] args) {
      var link = FindAvailableForRead(command.Area);
      Util.Check(link != null, "Cannot execute query command {0} - no data store is available for read.", command.CommandName);
      List<Exception> exceptions = null; 
      var result = ExecuteSafe(link, () => link.DataStore.ExecuteQuery(session, command, paging, args), ref exceptions);
      if (exceptions != null && exceptions.Count > 0) {
        //TODO: try to find a different store
        throw exceptions[0]; 
      }
      return (IList<EntityRecord>)result; 
    }

    public object ExecuteNonQuery(EntitySession session, IEntityCommand command, object[] args) {
      if (command.Kind == EntityCommandKind.SelectAll) {
        var link = FindAvailable(DataStoreLinkOptions.Write, command.Area);
        Util.Check (link != null, "Cannot execute command {0} - no data store is available for read.", command.CommandName); 
        return link.DataStore.ExecuteNonQuery(session, command, args); 
      }
      RecheckConnectivity(); 
      int writeCount = 0;
      object result = null; 
      List<Exception> exceptions = null; 
      foreach (var lnk in Links) {
        if (lnk.Available && lnk.Options.IsSet(DataStoreLinkOptions.Write)) {
          result = ExecuteSafe(lnk, () => lnk.DataStore.ExecuteNonQuery(session, command, args), ref exceptions); 
          writeCount++; 
        }
      }
      Util.Check(writeCount == 0, "Failed to write to any of the writable data stores.");
      return result; 
    }

    public void SaveChanges(EntitySession session) {
      int writeCount = 0;
      foreach (var lnk in Links) {
        if (lnk.Available && lnk.Options.IsSet(DataStoreLinkOptions.Write)) {
          lnk.DataStore.SaveChanges(session); 
          writeCount++;
        }
      }
    }

    public bool CheckConnectivity() {
      CheckConnectivityImpl(throwErrors: true);
      return true; 
    }

    public object ExecuteDynamicQuery(EntitySession session, EntityQuery query) {
      foreach (var link in Links)
        if (link.Options.IsSet(DataStoreLinkOptions.LinqProvider))
          return link.DataStore.ExecuteDynamicQuery(session, query);
      Util.Throw("No child datastore supports direct LINQ query.");
      return null;
    }
    #endregion

    // pops pop count; if it is divisible by 5, call CheckConnectivity
    private void RecheckConnectivity() {
      _opCount++;
      if (_opCount % 5 != 0) return;
      _opCount = 0;
      CheckConnectivityImpl(throwErrors: false); 
    }

    // private implementation methods ==================================================================================
    private DataStoreLink FindAvailableForRead(EntityArea area) {
      RecheckConnectivity();
      ConnectionFailureException failureEx = null;
      //try primary
      var link = FindAvailable(DataStoreLinkOptions.ReadPrimary, area);
      if (link != null)
        return link; 
      //try secondary
      link = FindAvailable(DataStoreLinkOptions.ReadSecondary, area);
      if (link != null)
        return link; 
      //check all again and remember the exception
      failureEx = CheckConnectivityImpl(throwErrors: true);
      link = FindAvailable(DataStoreLinkOptions.Read, area);
      if (link != null)
        return link; 
      if (failureEx != null)
        throw failureEx;
      Util.Throw("System failure in MultiStore connection management. Could not locate available connection for read while failing to detect an error."); 
      return null;
    }

    private DataStoreLink FindAvailable(DataStoreLinkOptions options, EntityArea area) {
      return Links.FirstOrDefault(lnk => lnk.Available && lnk.Options.AllSet(options) && lnk.Schemas.Contains(area.SchemaName));
    }
    private DataStoreLink FindAny(DataStoreLinkOptions options) {
      return Links.FirstOrDefault(lnk => lnk.Options.IsSet(options));
    }

    private ConnectionFailureException CheckConnectivityImpl(bool throwErrors) {
      var linksToRetry = Links.FindAll(lnk => lnk.Enabled && !lnk.Available && (lnk.UnavailableSince.AddMilliseconds(500) < DateTime.Now));
      if (linksToRetry.Count == 0) return null;
      List<Exception> exceptions = null;
      foreach (var link in linksToRetry)
        ExecuteSafe(link, () => link.DataStore.CheckConnectivity(), ref exceptions);
      ConnectionFailureException ex = null;
      if (exceptions != null && exceptions.Count > 0) {
        var newList = new List<Exception>(); 
        ex = new ConnectionFailureException("Connection failed.", exceptions);
        if (throwErrors)
          throw ex; 
      }
      return ex;
    }

    private object ExecuteSafe<TResult>(DataStoreLink link, Func<TResult> func, ref List<Exception> exceptions) {
      try {
        var result = func(); 
        link.LastAvailable = DateTime.Now;
        return result;
      } catch (ConnectionFailureException ex) {
        link.Available = false;
        link.UnavailableSince = DateTime.Now;
        if (exceptions == null)
          exceptions = new List<Exception>();
        //We need to add only 'real' exceptions to the list.
        exceptions.AddRange(ex.Exceptions); 
        return null;
      } catch (Exception ex) {
        if (exceptions == null)
          exceptions = new List<Exception>();
        exceptions.Add(ex);
        return null; 
      }
    }

  }//class

  internal static class MultiStoreExtensions {
    public static bool IsSet(this DataStoreLinkOptions options, DataStoreLinkOptions option) {
      return (options & option) != 0;
    }
    public static bool AllSet(this DataStoreLinkOptions options, DataStoreLinkOptions checkOptions) {
      return (options & checkOptions) == checkOptions;
    }
  }//class


}//ns
