﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Vita.Entities.Model;
using Vita.Entities.Runtime;
using Vita.Entities.Authorization.Runtime;

namespace Vita.Entities.Authorization {

  /// <summary> A container for the runtime permission set for a user with a given role set. </summary>
  /// <remarks>Authority instances are created dynamically at runtime for authenticated users and are shared between users with the same roles.
  /// AuthorizationService holds a dynamic cache of authority instances. 
  /// </remarks>
  public class Authority {
    //A key to lookup the already created roles in authority cache
    public readonly string Key;
    public readonly List<Role> Roles; 
    internal HashSet<RuntimePermissionGrant> NonEntityPermissionGrants; //non-entity permission grants
    internal Dictionary<Type, RuntimeEntityPermissionSet> EntityPermissionTable; //compound permissions organized by entity type
    internal HashSet<DynamicActivityGrant> DynamicGrants;

    internal Authority(string key, IEnumerable<Role> roles) {
      Key = key;
      Roles = roles.ToList();
    }

    public override string ToString() {
      return "(Authority/" + Key + ")";
    }

    public override int GetHashCode() {
      return Key.GetHashCode();
    }

    public string GetPermissionsSummary(Type entityType) {
      var sep = Environment.NewLine + "  ";
      var sb = new StringBuilder();
      sb.AppendFormat("Permissions summary for Authority [{0}], Entity {1}", Key, entityType);//key contains list of user roles
      sb.AppendLine(); 
      RuntimeEntityPermissionSet permSet;
      if (entityType == null || !EntityPermissionTable.TryGetValue(entityType, out permSet)) {
        sb.AppendLine("(None)");
        return sb.ToString(); 
      }
      sb.AppendLine("-- Entity permisssion: ");
      sb.AppendLine(permSet.ToString());
      return sb.ToString();
    }

    public string GetPermissionsSummary() {
      var sep = Environment.NewLine + "  ";
      var sb = new StringBuilder();
      sb.AppendFormat("Permissions summary for Authority [{0}]", Key);//key contains list of user roles
      sb.AppendLine();
      if (this.NonEntityPermissionGrants.Count > 0) {
        sb.AppendLine("-- Non-entity permissions ----------------- ");
        sb.AppendLine(string.Join(sep, this.NonEntityPermissionGrants));
      }
      sb.AppendLine("-- Entity permissions --------------------- ");
      foreach (var de in EntityPermissionTable) 
        sb.Append(de.Value.ToString()); 
      return sb.ToString(); 
    }

    public string Summary {
      get {
        if (_summary == null)
          _summary = GetPermissionsSummary();
        return _summary; 
      }
    } string _summary;

    public static string BuildKey(IEnumerable<Role> roles) {
      var list = roles.ToList();
      list.Sort((x, y) => x.Id.CompareTo(y.Id));
      return string.Join(",", list.Select(r => r.Id));
    }
  }//class


}
