﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Vita.Common;
using Vita.Entities;
using Vita.Entities.Model;
using Vita.Entities.Runtime;

using Vita.Entities.Authorization.Runtime;
using Vita.Services;

namespace Vita.Entities.Authorization {

  public interface IAuthorizationService {
    ISecureSession OpenSecureSession(UserContext userContext);
    Authority GetAuthority(IEnumerable<Role> roles);
    Authority TryGetAuthority(string authorityKey);
  }

  public class AuthorityCacheSettings {
    public int Capacity;
    public TimeSpan Expiration;
    public TimeSpan CleanupInterval;

    public AuthorityCacheSettings(int capacity = 1000, TimeSpan? expiration = null, TimeSpan? cleanupInterval = null) {
      Capacity = capacity;
      Expiration = expiration == null ? TimeSpan.FromSeconds(60) : expiration.Value;
      CleanupInterval = cleanupInterval == null ? TimeSpan.FromSeconds(30) : cleanupInterval.Value;
    }
  }

  public class AuthorizationService: IAuthorizationService, IEntityServiceInit {
    EntityStore _entityStore;
    AuthorityCacheSettings _authorityCacheSettings;
    ObjectCache<string, Authority> _authorityCache;
    private static object _lock = new object(); 


    public AuthorizationService(EntityModel model, AuthorityCacheSettings authorityCacheSettings = null) {
      model.RegisterService(typeof(IAuthorizationService), this);
      _authorityCacheSettings = authorityCacheSettings ?? new AuthorityCacheSettings();
    }

    #region IEntityServiceInit Members

    public void Init(IEntityStore entityStore) {
      _entityStore = (EntityStore) entityStore;
      _authorityCache = new ObjectCache<string, Authority>(_authorityCacheSettings.Capacity, _authorityCacheSettings.Expiration, _authorityCacheSettings.CleanupInterval);
    }

    #endregion

    #region IAuthorizationService Members
    public ISecureSession OpenSecureSession(UserContext userContext) {
      Util.Check(userContext.Authority != null, "UserContext.Authority is null; must be set to user's authority to create secure session. User: {0}", userContext.UserName);
      return new SecureSession(_entityStore, userContext);
    }

    public Authority TryGetAuthority(string authorityKey) {
      return _authorityCache.Lookup(authorityKey); 
    }

    public Authority GetAuthority(IEnumerable<Role> userRoles) {
      var authorityKey = Authority.BuildKey(userRoles);
      Authority authority = TryGetAuthority(authorityKey);
      if (authority != null)
        return authority; 
      //build new authority. 
      lock (_lock) {
        // prepare roles
        foreach (var role in userRoles)
          if (role.RuntimeData == null)
            RuntimeAuthorityDataBuilder.BuildRuntimeData(_entityStore.Model.ModelInfo, role);
        authority = new Authority(authorityKey, userRoles);
        RuntimeAuthorityDataBuilder.BuildAuthorityPermissions(authority);
        _authorityCache.Add(authority.Key, authority);
      }
      return authority;
    }

    #endregion



  }//class
}
