﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Vita.Entities.Runtime;
using Vita.Entities.Model;

namespace Vita.Entities.Authorization {

  /// <summary>A flag set representing granular access rights (read, write, etc.) for an entity.</summary>
  [Flags]
  public enum Actions {
    /// <summary>Empty flag set.</summary>
    None,
    
    /// <summary>Restricted read access. Can read internally in code, but user may not see the value. </summary>
    Peek = 1, 

    /// <summary>Can read values and show it to the user in UI.</summary>
    ReadStrict = 1 << 1,

    /// <summary>Can create new entities.</summary>
    CreateStrict = 1 << 2,
    /// <summary>Can update an entity or entities.</summary>
    UpdateStrict = 1 << 3,
    /// <summary>Can delete an entity or entities.</summary>
    DeleteStrict = 1 << 4,

    // Usually the actions come combined: if the user can update the record, most likely he can view it as well. 
    /// <summary>Can read values and show it to the user in UI. Combination of ReadStrict and Peek access.</summary>
    Read = ReadStrict | Peek,
    /// <summary>Can create new and read existing entities.</summary>
    Create = CreateStrict | Read,
    /// <summary>Can view and update an entity or entities.</summary>
    Update = UpdateStrict | Read,
    /// <summary>Can view and delete an entity or entities.</summary>
    Delete = DeleteStrict | Read,
    /// <summary>Full CRUD access - a combination of Create, Read, Update, Delete actions.</summary>
    CRUD = Create | Update | Delete, 

    Custom01 = 1 << 8,
    Custom02 = 1 << 9,
    Custom03 = 1 << 10,
    Custom04 = 1 << 11,

  }

  public static class ActionsExtensions {
    public static bool IsSet(this Actions actions, Actions action) {
      return (actions & action) != 0;
    }

    public static Actions ToAction(this EntityStatus status) {
      switch (status) {
        case EntityStatus.New: return Actions.CreateStrict;
        case EntityStatus.Modified: return Actions.UpdateStrict;
        case EntityStatus.Deleting: return Actions.DeleteStrict;
        default: return Actions.Peek;
      }
    }

    public static Actions ToAction(this ReadAccessLevel mode) {
      return (mode == ReadAccessLevel.Read) ? Actions.ReadStrict : Actions.Peek;
    }

    public static Actions ToAuthorizationAction(this EntityCommandKind commandKind) {
      switch (commandKind) {
        case EntityCommandKind.SelectAll:
        case EntityCommandKind.SelectAllPaged:
        case EntityCommandKind.SelectByKey:
        case EntityCommandKind.SelectByKeyManyToMany:
        case EntityCommandKind.CustomSelect:
          return Actions.Peek;
        case EntityCommandKind.Update:
        case EntityCommandKind.CustomUpdate:
        case EntityCommandKind.PartialUpdate:
          return Actions.Update;
        case EntityCommandKind.Insert:
        case EntityCommandKind.CustomInsert:
          return Actions.Create;
        case EntityCommandKind.Delete:
        case EntityCommandKind.CustomDelete:
          return Actions.Delete;
        default:
          return Actions.None;
      }
    }//method

  }//class

}
