﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Vita.Common;
using Vita.Entities.Model;
using Vita.Entities.Runtime;

namespace Vita.Entities.Authorization {

  /// <summary> An activity grant enabled dynamically from business code for a certain period of time/execution path.</summary>
  /// <remarks>Dynamic activity grants allow enabling certain operations only in specific contexts (parts) of the application. </remarks>
  public class DynamicActivityGrant : ActivityGrant {
    public string DocumentKey;        // The key to save root Document's Id when activity starts
    public DataFilter DocumentFilter; // Filter to check the root entity at activity start

    public DynamicActivityGrant(Activity activity, DataFilter filter = null, 
           string documentKey = null, DataFilter documentFilter = null)
      : base(activity, filter) {
      DocumentKey = documentKey;
      DocumentFilter = documentFilter;
    }

    public void Begin(IUserContext context, object rootEntity = null, object rootId = null) {
      //Check rootEntity against startup filter 
      if (DocumentFilter != null) {
        Util.Check(rootEntity != null, "Root entity may not be null when activity is granted with Document filter.");
        var ok = DocumentFilter.MatchEntity(context, rootEntity);
        if (!ok) {
          var msg = Util.SafeFormat("Cannot start dynamic activity {0} - root entity does not pass startup record filter check. Entity: {1}",
            this.Activity.Name, rootEntity);
          var entType = EntityHelper.GetEntityType(rootEntity);
          var session = EntityHelper.GetSession(rootEntity) as SecureSession;
          throw new AuthorizationException(msg, entType, Actions.Peek, true, null, session);
        }
      }
      //Set root value
      if (DocumentKey != null)
        context.Values[DocumentKey] = rootId;
      var ctx = (UserContext)context;
      ctx.EnablementTable.Enable(this);
    }

    public void End(IUserContext context) {
      var ctx = (UserContext)context; 
      if (DocumentKey != null)
        ctx.RemoveValue(DocumentKey);
      ctx.EnablementTable.Disable(this);
    }

    // for use in 'using()' statements. ActivityGrant is enabled only for the duration of the block and is disabled at the end automatically. 
    // Ex:    
    //   using (editActivityGrant.Execute(context, payDoc, payDoc.Id )) { 
    //     doStuff(); 
    //   }
    public IDisposable Execute(IUserContext context, object root = null, object rootId = null) {
      var token = new DynamicGrantToken() { Context = (UserContext)context, Grant = this };
      try {
        Begin(context, root, rootId);
        return token; 
      } catch (Exception) {
        token.Dispose();
        throw; 
      }
    }

    public bool IsEnabled(IUserContext context) {
      var ctx = (UserContext)context; 
      return ctx.EnablementTable.IsEnabled(this);
    }

    //Used at runtime
    public override ActivityGrant CreateSimilarGrant(Activity activity) {
      return new DynamicActivityGrant(activity, this.Filter, this.DocumentKey, this.DocumentFilter); 
    }

    public override string ToString() {
      var result = Activity.Name;
      if (Filter != null)
        result += "/" + Filter;
      result += "(dynamic)";
      return result; 
    }

    #region DynamicActivityToken nested class
    // Used by DynamicActivityGrant.Execute method to enable activity inside a 'using' block
    internal class DynamicGrantToken : IDisposable {
      internal UserContext Context;
      internal DynamicActivityGrant Grant;
      bool _disposed;
      public void Dispose() {
        if (!_disposed)
          Grant.End(Context);
        _disposed = true; 
      }
    }
    #endregion

  }//class

}
