﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Vita.Common;
using Vita.Entities.Runtime;
using Vita.Entities.Model;

namespace Vita.Entities.Authorization.Runtime {

  //Static utility class
  internal static class RuntimeAuthorityDataBuilder {

    internal static void BuildAuthorityPermissions(Authority authority) {
      var entPermSets = authority.EntityPermissionTable = new Dictionary<Type, RuntimeEntityPermissionSet>();
      var nonEntPermGrants = authority.NonEntityPermissionGrants = new HashSet<RuntimePermissionGrant>();
      var dynGrants = authority.DynamicGrants = new HashSet<DynamicActivityGrant>();
      foreach (var role in authority.Roles) {
        foreach (var permGrant in role.RuntimeData.AllPermissionGrants) {
          var dynGrant = permGrant.Grant as DynamicActivityGrant;
          if (dynGrant != null)
            dynGrants.Add(dynGrant);
          var entPerm = permGrant.Permission as EntityGroupPermission;
          if (entPerm == null) {
            nonEntPermGrants.Add(permGrant);
            continue; 
          }
          //We have entity permission grant. Go through each entity
          foreach (var entGroupRes in entPerm.GroupResources) {
            foreach (var entRes in entGroupRes.Entities) {
              var entType = entRes.EntityType;
              //Find/create entity permission set for the entity type
              RuntimeEntityPermissionSet entPermSet;
              if (!entPermSets.TryGetValue(entType, out entPermSet)) {
                entPermSet = new RuntimeEntityPermissionSet(entType);
                entPermSets[entType] = entPermSet;
              }
              var msg = "  Source permission " + permGrant.Permission.Name + ":";
              // Go through each permission and try to merge
              bool merged = false;
              foreach (var oldPerm in entPermSet.Permissions)
                // Check if filters and grants are compatible, so permission can be merged
                if (CanMergeFromGrant(oldPerm, permGrant)) {
                  AddToRights(oldPerm.AccessRights, entPerm.Actions, entRes.MemberMask);
                  oldPerm.AllGrants.Add(permGrant.Grant); //add grant
                  merged = true;
                  msg += " - merged into " + oldPerm.Id;
                  break; //foreach rPerm
                }
              if (!merged) {
                //create new access rights and new runtime permission
                var rights = new EntityAccessRights(entPerm.Actions, false, entRes.MemberMask);
                var permId = "P" + entPermSet.Permissions.Count; //artificial Id
                var rPerm = new RuntimeEntityPermission(permId, entType, rights, permGrant.Grant);
                entPermSet.Permissions.Add(rPerm);
                msg += " - added as " + permId; 
              }
              entPermSet.LogBuilder.AppendLine(msg);
            } //foreach entResDet
          }//foreach entGroupRes
        }// foreach permGrant

      } //foreach role
      // Final processing: complete processing for each entPermSet
      foreach (var entPermSet in entPermSets.Values) 
        FinalizePermissionSet(entPermSet);
    }

    public static void BuildRuntimeData(EntityModelInfo entityModel, Role role) {
      if (role.RuntimeData != null) return;
      role.Init(entityModel); 
      role.RuntimeData = new RuntimeRoleData();
      //process all child activities
      foreach (var actGrant in role.ActivityGrants)
        CollectChildActivities(actGrant.Activity);
      // get refs to 'all' sets
      var allChildRoles = role.RuntimeData.AllChildRoles;
      var allChildGrants = role.RuntimeData.AllActivityGrants;
      allChildGrants.UnionWith(role.ActivityGrants); //initialize with direct grants
      //Process child roles; collect roles and grants
      foreach (var child in role.ChildRoles)
        if (allChildRoles.Add(child)) {
          BuildRuntimeData(entityModel, child);
          allChildRoles.UnionWith(child.RuntimeData.AllChildRoles); //merge child roles
          allChildGrants.UnionWith(child.RuntimeData.AllActivityGrants); //merge all child grants         
        }
      //check there's no circular parent/child relationships
      Util.Check(!allChildRoles.Contains(role),
        "Role {0} is an (indirect) child of itself.", role.Name);
      // Now we merged all activity grants from child roles into one flat set allChildRoles
      // All activity grants from this and all child roles are merged into  allChildGrants
      // What we need to add to allChildGrants is grand-child activities - child activities 
      // of direct grants for this role. If this role R has grant G on activity A, and activity
      // A has child activity B, then effectively role R is granted activity B through the same
      // grant G (or its clone). If grant G has conditions (filter, dynamic condition),
      // then activity B is granted to role R with the same conditions.
      // So we grant grand-child activities (B's) through clones of child grant G.
      foreach (var actGrant in role.ActivityGrants) {
        foreach (var grandChild in actGrant.Activity.RuntimeData.AllChildActivities) {
          var grandChildGrand = actGrant.CreateSimilarGrant(grandChild);
          allChildGrants.Add(grandChildGrand);
        }
      }
      //Consolidate all permissions in one plane set of (grant/permission) tuples
      foreach (var grant in role.RuntimeData.AllActivityGrants) {
        foreach (var perm in grant.Activity.Permissions) {
          role.RuntimeData.AllPermissionGrants.Add(new RuntimePermissionGrant(perm, grant));
        }
      }
    }

    private static void CollectChildActivities(Activity activity) {
      if (activity.RuntimeData != null) return;
      activity.RuntimeData = new RuntimeActivityData();
      activity.RuntimeData.AllChildActivities.UnionWith(activity.ChildActivities); //immediate children
      foreach (var child in activity.ChildActivities) {
        CollectChildActivities(child);
        activity.RuntimeData.AllChildActivities.UnionWith(child.RuntimeData.AllChildActivities);
      }
      Util.Check(!activity.RuntimeData.AllChildActivities.Contains(activity), "Activity {0} is an (indirect) child of itself.", activity.Name);
      //collect all permissions
      var allPerms = activity.RuntimeData.AllPermissions;
      allPerms.AddRange(activity.Permissions);
      foreach (var child in activity.RuntimeData.AllChildActivities)
        allPerms.AddRange(child.RuntimeData.AllPermissions);
    }

    private static void FinalizePermissionSet(RuntimeEntityPermissionSet permissionSet) {
      var permSet = permissionSet;
      permSet.HasRecordLevelPermissions = false;
      permSet.HasDynamicPermissions = false;
      permSet.FixedRights = new EntityAccessRights(Actions.None);
      permSet.StaticRightsForType = new EntityAccessRights(Actions.None);

      foreach (var perm in permissionSet.Permissions) {
        //complete perm
        var dynGrants = perm.AllGrants.Select(g => g as DynamicActivityGrant).Where(g => g != null).ToArray();
        perm.DynamicGrants = (dynGrants.Length == 0) ? null : dynGrants;
        permSet.HasRecordLevelPermissions |= (perm.RecordGroupIdReader != null);
        permSet.HasDynamicPermissions |= perm.DynamicGrants != null;
        if (!permSet.HasRecordLevelPermissions && !permSet.HasDynamicPermissions)
          permSet.FixedRights.Merge(perm.AccessRights);
        if (perm.DynamicGrants == null)
          permSet.StaticRightsForType.Merge(perm.AccessRights);
      }
      if (permSet.HasRecordLevelPermissions) {
        permSet.FixedRights = null; // FixedRights are there only if there are no record-level permissions
        permSet.StaticRightsForType.HasInstanceLevelRestrictions = true; // indicate that there are record-level restrictions
      }
      if (permSet.HasDynamicPermissions)
        permSet.StaticRightsForType = null;
      //finalize log
      permSet.ConstructionLog = permSet.LogBuilder.ToString();
      permSet.LogBuilder = null; 
    }

    // Extensions for merging rights
    public static void Merge(this EntityAccessRights toRights, EntityAccessRights other) {
      MergeActionsAndMasks(Actions.Peek, ref toRights.Actions, ref toRights.PeekMask, other.Actions, other.PeekMask);
      MergeActionsAndMasks(Actions.ReadStrict, ref toRights.Actions, ref toRights.ReadMask, other.Actions, other.ReadMask);
      MergeActionsAndMasks(Actions.UpdateStrict, ref toRights.Actions, ref toRights.UpdateMask, other.Actions, other.UpdateMask);
      toRights.Actions |= other.Actions;
    }

    public static void AddToRights(EntityAccessRights toRights, Actions otherActions, EntityMemberMask otherMask) {
      MergeActionsAndMasks(Actions.Peek, ref toRights.Actions, ref toRights.PeekMask, otherActions, otherMask);
      MergeActionsAndMasks(Actions.Read, ref toRights.Actions, ref toRights.ReadMask, otherActions, otherMask);
      MergeActionsAndMasks(Actions.Update, ref toRights.Actions, ref toRights.UpdateMask, otherActions, otherMask);
      toRights.Actions |= otherActions;
    }

    private static void MergeActionsAndMasks(Actions forAction,
             ref Actions intoActions, ref EntityMemberMask intoMask, Actions otherActions, EntityMemberMask otherMask) {
      //If none is set, nothing to do
      if (!intoActions.IsSet(forAction) && !otherActions.IsSet(forAction))
        return;
      //If both set, we need to merge both masks 
      if (intoActions.IsSet(forAction) && otherActions.IsSet(forAction)) {
        if (intoMask == null)
          return; //it is already 'allow all'
        if (otherMask == null) { //means allow al
          intoMask = null; //set it to allow-all
          return;
        }
        //both masks are not null: OR bits
        intoMask.Or(otherMask);
        return;
      }
      // Only one of the sides has Actions bit set. 
      if (intoActions.IsSet(forAction))
        return; // otherMask does not bring anything
      if (otherActions.IsSet(forAction)) {
        intoMask = otherMask;
      }
    }//method

    private static bool CanMergeFromGrant(RuntimeEntityPermission intoPerm, RuntimePermissionGrant fromGrant) {
      var actGrant = fromGrant.Grant;
      if (intoPerm.AllGrants.Contains(actGrant))
        return true; //Grant is already there
      if (intoPerm.DataFilter != actGrant.Filter)
        return false; // if filters do not match, then false
      //Check dynamic grant compatibility. Only dynamic grants matter
      if (actGrant is DynamicActivityGrant) {
        return intoPerm.AllGrants.Contains(actGrant); //match if dynamic grant is already in permission's grant list
      }
      return true;
    }


  }//class
}//ns
