﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Vita.Common;
using Vita.Entities.Model;

namespace Vita.Entities.Authorization.Runtime {

  internal class RuntimeActivityData {
    public HashSet<Activity> AllChildActivities = new HashSet<Activity>();
    public List<Permission> AllPermissions = new List<Permission>();
  }

  internal class RuntimeRoleData {
    public HashSet<Role> AllChildRoles = new HashSet<Role>();
    public HashSet<ActivityGrant> AllActivityGrants = new HashSet<ActivityGrant>();
    public HashSet<RuntimePermissionGrant> AllPermissionGrants = new HashSet<RuntimePermissionGrant>();
  }

  internal class RuntimePermissionGrant : HashedObject, IEquatable<RuntimePermissionGrant> {
    public Permission Permission;
    public ActivityGrant Grant;

    public RuntimePermissionGrant(Permission permission, ActivityGrant grant) {
      Permission = permission;
      Grant = grant;
    }
    //required if you override equals
    public override int GetHashCode() {
      return base.GetHashCode();
    }

    public override bool Equals(object obj) {
      return this.Equals(obj as RuntimePermissionGrant);
    }

    #region IEquatable<T> Members
    public bool Equals(RuntimePermissionGrant other) {
      if (other == null) 
        return false; 
      return Permission == other.Permission && Grant == other.Grant;
    }
    #endregion
  }//class

  internal class RuntimeEntityPermission {
    public string Id; // artificial name (P1, P2..) to refer to permission in creation log
    public Type EntityType;
    public EntityAccessRights AccessRights;
    public DataFilter DataFilter;
    internal HashSet<ActivityGrant> AllGrants = new HashSet<ActivityGrant>();

    public Funclet RecordGroupIdReader; //cached funclet for entity type
    //If there is at least one static grant, then this property is null
    public DynamicActivityGrant[] DynamicGrants;

    public RuntimeEntityPermission(string id, Type entityType, EntityAccessRights rights, ActivityGrant grant) {
      Id = id;
      EntityType = entityType;
      AccessRights = rights;
      AllGrants.Add(grant);
      DataFilter = grant.Filter;
      //Cache RecordGroupIdReader
      if (DataFilter != null)
        RecordGroupIdReader = DataFilter.GetFunclet(EntityType);
    }

    public override string ToString() {
      var result = Util.SafeFormat("{0}: {1}/{2}", Id, EntityType, AccessRights);
      if (DataFilter != null)
        result += "/Filter:" + DataFilter.Name;
      if (DynamicGrants != null)
        result += "/Dynamic, activated by: " + string.Join(",", DynamicGrants.Select(g => g.Activity.Name));
      return result;
    }

  }//class

  //Cumulative set of entity permissions for a given entity type. Computed in context of specific authority (role set)
  internal class RuntimeEntityPermissionSet {
    public Type EntityType;
    public bool HasRecordLevelPermissions;
    public bool HasDynamicPermissions;
    // FixedRights is not null only if no record-level permissions and no dynamic permissions
    // If not null, this field contains final access rights for all records
    public EntityAccessRights FixedRights;
    // Type-level access; not null if no dynamic permissions;  ignores record-level restrictions
    public EntityAccessRights StaticRightsForType;
    //Logging
    internal StringBuilder LogBuilder; // creation log, tracks all add/merge actions from original permissions
    internal string ConstructionLog;

    // Dynamic or record-level permissions
    internal List<RuntimeEntityPermission> Permissions = new List<RuntimeEntityPermission>();

    public RuntimeEntityPermissionSet(Type entityType) {
      EntityType = entityType;
      LogBuilder = new StringBuilder();
    }

    public override string ToString() {
      string perms = string.Join(Environment.NewLine, Permissions);
      var sDefaultRights = FixedRights == null ? "None" : FixedRights.ToString();
      var sDefaultStaticRightsForType = StaticRightsForType == null ? "None" : StaticRightsForType.ToString();
      var result = Util.SafeFormat(
@"  Entity:                       {0}
  DefaultRights:                {1}
  Static table-level Rights:    {2}    (ignores Data filter if any specified)
  HasRecordLevelPermissions:    {3}     
  HasDynamicPermissions:        {4}
............ Granted rights ...............................
{5}
............ Construction log: .................................
{6}
",
       EntityType, sDefaultRights, sDefaultStaticRightsForType, HasRecordLevelPermissions, HasDynamicPermissions, perms, ConstructionLog);
      return result;
    }

  }//class





}
