﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Vita.Common;
using Vita.Entities;
using Vita.Entities.Runtime;
using Vita.Entities.Model;

namespace Vita.Entities.Authorization {
  using Runtime;
  using Vita.Data;

  public class SecureSession : EntitySession, ISecureSession {
    #region ISecureSession members
    public ReadAccessLevel RequireReadAccess {
      get { return _requireReadAccess; }
      set {
        _requireReadAccess = value;
        _requireReadAction = _requireReadAccess.ToAction(); 
      } 
    } ReadAccessLevel _requireReadAccess; 

    public DenyReadActionType DenyReadAction { get; set; }
    #endregion

    public bool BypassReadCheck { get { return _elevationCount > 0; } }
    // Count of ElevateRead nested calls; if non-zero, reads are allowed without permission check. 
    private int _elevationCount;
    private Actions _requireReadAction; // Equiv of RequireReadAcces
    

    public bool IsActionAllowed<TEntity>(Actions actions) {
      var entInfo = GetEntityInfo(typeof(TEntity));
      var access = this.GetEntityTypeAccess(entInfo);
      //Check each action type separately
      if (actions.IsSet(Actions.CreateStrict) && ! access.Actions.IsSet(Actions.CreateStrict))
        return false;
      if (actions.IsSet(Actions.Peek) && !access.CanPeekAny()) 
        return false;
      if (actions.IsSet(Actions.ReadStrict) && !access.CanReadAny())
        return false;
      if (actions.IsSet(Actions.UpdateStrict) && !access.CanUpdateAny())
        return false;
      if (actions.IsSet(Actions.DeleteStrict) && !access.CanDelete())
        return false;
      return true; 
    }


    public SecureSession(IEntityStore entityStore, UserContext context) : base(entityStore, context) {
      RequireReadAccess = ReadAccessLevel.Peek;
      DenyReadAction = DenyReadActionType.Throw;
    }

    public override string ToString() {
      return "SecureSession(" + UserContext.UserName + ")";
    }
    //Note: we override mostly specific low-level methods we need to intercept
    protected override EntityRecord GetLoadedRecord(EntityKey primaryKey) {
      var rec = base.GetLoadedRecord(primaryKey);
      if (rec == null || BypassReadCheck)
        return rec;
      if (CheckRecordAccess(rec, _requireReadAction)) // throws if DenyReadMode is Throw
        return rec;
      return null; //if CheckRecordAccess did not throw  
    }

    public override EntityRecord NewRecord(EntityInfo entityInfo) {
      EntityAccessRights rights;
      //CheckEntityAccess throws only if DenyReadMode is Throw; but we throw anyway in NewRecord
      if (!CheckEntityAccess(entityInfo, Actions.CreateStrict, out rights))
        AccessDenied(Actions.Create, entityInfo, rights); 
      var rec = base.NewRecord(entityInfo);
      rec.AccessRights = rights;
      return rec; 
    }

    public override void DeleteRecord(EntityRecord record) {
      if (!CheckRecordAccess(record, Actions.DeleteStrict))
        return; //CheckRecordAccess throws if DenyReadMode is Throw
      base.DeleteRecord(record);
    }
    
    public override object ExecuteNonQuery(IEntityCommand command, params object[] args) {
      EntityAccessRights rights; 
      if (!CanExecute(command, out rights))
        AccessDenied(command, rights);
      return base.ExecuteNonQuery(command, args);
    }

    public override IQueryable<TEntity> EntitySet<TEntity>() {
      // We allow to execute LINQ queries in elevated mode
      if (BypassReadCheck)
        return base.EntitySet<TEntity>();
      var entInfo = GetEntityInfo(typeof(TEntity));
      EntityAccessRights rights;
      if (!CheckEntityAccess(entInfo, _requireReadAction, out rights))
        return (new List<TEntity>()).AsQueryable(); //return emtpy set
      
      // Note: we'll filter the returned results 
      return base.EntitySet<TEntity>();
    }

    public override IList<EntityRecord> ExecuteRecordQuery(IEntityCommand command, Paging paging, params object[] args) {
      if (BypassReadCheck)
        return base.ExecuteRecordQuery(command, paging, args);
      //Non-elevated - first check entity type-level access
      var entCommand = command as EntityCommand; 
      EntityAccessRights entityTypeAccess;
      if (!CheckEntityAccess(entCommand.TargetEntityInfo, entCommand.AuthorizationAction, out entityTypeAccess))
        return new List<EntityRecord>(); //return empty list, if exception is not thrown by CheckEntityAccess
      //execute query
      var records = base.ExecuteRecordQuery(command, paging, args);
      //If there are record-level permissions, filter records
      if (entityTypeAccess.HasInstanceLevelRestrictions)
        return FilterRecords(records);
      //Otherwise, set the rights on each record same as entity type access
      foreach (var rec in records)
        rec.AccessRights = entityTypeAccess;
      return records;
    }

    public override bool CanDeleteRecord(EntityRecord record, out Type[] blockingEntities) {
      // checking related records may require querying entities that may not be available to the current user.
      using (this.ElevateRead()) {
        return base.CanDeleteRecord(record, out blockingEntities);
      }
    }

    //TODO: Implement Attach to filter records in LINQ queries
    public override EntityRecord Attach(EntityRecord record) {
      record.AccessRights = null; //in case they're not null
      return base.Attach(record);
    }

    protected override void SubmitChanges() {
      // Access rights might change due to record field changes.
      // One example from books module: user creates BookReview record - he is granted the rights to do so. 
      // But then the code assigns other user as review owner - so user tries to create a review mascarading 
      // as another user. This situation must be caught here. 
      foreach (var rec in RecordsChanged) {
        rec.AccessRights = null; //reset rights and force new rights evaluation
        CheckRecordAccess(rec, rec.Status.ToAction());
        // verify write permissions for modified fields
        if (rec.AccessRights != null && rec.AccessRights.UpdateMask != null)
          VerifyModifiedValues(rec); 
      }
      base.SubmitChanges();
    }

    protected override void OnSaving() {
      // OnSave handlers may contain methods that travers internal record links. 
      // For example, ChangeLog might try to find GroupId for records, and needs to access props/entities that may not be available 
      // to the current user. So we need to use elevated mode here to disable access rights checks.
      using (this.ElevateRead()) {
        base.OnSaving();
      }
    }
    public override void ValidateChanges() {
      // Validation methods might access data that is not normally available to the current user. So we need to use elevated mode here
      // to disable access rights checks.
      using (this.ElevateRead()) {
        base.ValidateChanges();
      }
    }

    public override object RecordGetMemberValue(EntityRecord record, EntityMemberInfo member) {
      if (record.Status == EntityStatus.Loading || BypassReadCheck || !record.AccessCheckEnabled)
        return base.RecordGetMemberValue(record, member);
      //Not elevated
      if (!CheckRecordAccess(record, _requireReadAction))
        return member.DeniedValue;
      var rights = record.AccessRights;
      var canGet = _requireReadAccess == ReadAccessLevel.Peek ? rights.CanPeek(member) : rights.CanRead(member);
      if (canGet)
        return base.RecordGetMemberValue(record, member);
      //Access denied
      AccessDenied(_requireReadAction, record, member);
      // if AccessDenied did not throw, return default value
      return member.DeniedValue;
    }

    public override void RecordSetMemberValue(EntityRecord record, EntityMemberInfo member, object value) {
      //Check if get/set checks are enabled; they are disabled for new and loading records
      if (!record.AccessCheckEnabled) {
        base.RecordSetMemberValue(record, member, value);
        return; 
      }
      if (record.AccessRights == null)
         record.AccessRights = GetRecordAccess(record);
      if (record.AccessRights.CanUpdate(member))
        base.RecordSetMemberValue(record, member, value);
      else 
        AccessDenied(Actions.UpdateStrict, record, member);
    }


    public EntityAccessRights GetEntityTypeAccess(EntityInfo entity) {
      if (entity.Flags.IsSet(EntityFlags.IsSystem))
        return EntityAccessRights.AllowAll;
      RuntimeEntityPermissionSet permSet;
      if (!UserContext.Authority.EntityPermissionTable.TryGetValue(entity.EntityType, out permSet))
        return EntityAccessRights.AllowNone;
      return this.GetEntityTypeAccess(permSet, UserContext);
    }

    public EntityAccessRights GetRecordAccess(EntityRecord record) {
      var entity = record.EntityInfo;
      if (entity.Flags.IsSet(EntityFlags.IsSystem))
        return EntityAccessRights.AllowAll;
      RuntimeEntityPermissionSet permSet;
      if (!UserContext.Authority.EntityPermissionTable.TryGetValue(entity.EntityType, out permSet))
        return EntityAccessRights.AllowNone;
      return GetRecordAccess(permSet, record);
    }

    // Private utilities ========================================================================
    private EntityAccessRights GetEntityTypeAccess(RuntimeEntityPermissionSet permissionSet, UserContext context) {
      if (permissionSet.FixedRights != null)
        return permissionSet.FixedRights;
      if (permissionSet.StaticRightsForType != null)
        return permissionSet.StaticRightsForType;
      //We need to compute permissions
      var rights = new EntityAccessRights(Actions.None, hasInstanceLevelRestrictions: permissionSet.HasRecordLevelPermissions);
      //we ignore record-level filters when computing type-level access
      foreach (var perm in permissionSet.Permissions)
        if (PermissionIsActive(perm))
          rights.Merge(perm.AccessRights);
      return rights;
    }

    private EntityAccessRights GetRecordAccess(RuntimeEntityPermissionSet permissionSet, EntityRecord record) {
      if (permissionSet.FixedRights != null)
        return permissionSet.FixedRights;
      //We need to compute permissions
      var rights = new EntityAccessRights(Actions.None);
      foreach (var perm in permissionSet.Permissions)
        //Check enabled (for dynamic grants), and check record filter - might be false for dynamic actions
        if (PermissionIsActive(perm) && MatchRecordToPermissionFilter(perm, record))
          rights.Merge(perm.AccessRights);
      return rights;
    }

    private bool MatchRecordToPermissionFilter(RuntimeEntityPermission permission, EntityRecord record) {
      if (permission.DataFilter == null)
        return true;
      // Executing filter funclets involves accessing to properties and other referenced entities that might not be allowed for access by the user
      // We need to elevate the access for the period of evaluating filter expressions
      //if mode is already unrestricted, call directly
      if (this.BypassReadCheck)
        return permission.DataFilter.MatchEntity(UserContext, record.EntityInstance);
      //otherwise Elevate
      using (this.ElevateRead()) {
        return permission.DataFilter.MatchEntity(UserContext, record.EntityInstance);
      }
    }

    private bool PermissionIsActive(RuntimeEntityPermission permission) {
      var dynGrants = permission.DynamicGrants;
      if (dynGrants == null) return true;
      for (int i = 0; i < dynGrants.Length; i++)
        if (dynGrants[i].IsEnabled(UserContext)) return true;
      return false;
    }



    private bool CheckEntityAccess(EntityInfo entity, Actions action, out EntityAccessRights rights) {
      rights = GetEntityTypeAccess(entity);
      if (rights.Actions.IsSet(action)) return true;
      var isReadAction = action.IsSet(Actions.Read); 
      if (this.DenyReadAction == DenyReadActionType.Throw)
        AccessDenied(action, entity, rights);
      return false; 
    }

    private bool CheckRecordAccess(EntityRecord record, Actions action) {
      if (record.AccessRights == null)
        record.AccessRights = GetRecordAccess(record); 
      if (record.AccessRights.Actions.IsSet(action)) 
        return true;
      AccessDenied(action, record);
      return false; 
    }

    //this overload saves us some cycles when we know already entity type access and there's no record-level access
    private bool CheckRecordAccess(EntityRecord record, Actions action, EntityAccessRights typeAccessRights) {
      if (typeAccessRights.HasInstanceLevelRestrictions)
        return CheckRecordAccess(record, action);
      //Otherwise, assume record rights are the same as type access
      record.AccessRights = typeAccessRights;
      if (typeAccessRights.Actions.IsSet(action)) 
        return true;
      AccessDenied(action, record);
      return false;
    }

    private IList<EntityRecord> FilterRecords(IList<EntityRecord> records) {
      var outlist = new List<EntityRecord>();
      foreach (var rec in records) {
        if (CheckRecordAccess(rec, _requireReadAction))
          outlist.Add(rec);
      }
      return outlist;
    }

    protected virtual void VerifyModifiedValues(EntityRecord record) {
      var rights = record.AccessRights; 
      if (rights == null)
        return; 
      var members = record.EntityInfo.Members;
      var modValues = record.ValuesModified;
      for (int i = 0; i < members.Count; i++) {
        var member = members[i];
        if (modValues[member.ValueIndex] != null && !rights.CanUpdate(member)) //value was modified but update is not allowed for this member
          AccessDenied(Actions.UpdateStrict, record, member);
      }
    }


    //Throws or returns quietly depending on action
    // If action is Read, and CurrentDenyReadMode is DenyReadMode is filter, returns quietly.
    // Otherwise throws AuthorizationException
    private void AccessDenied(IEntityCommand command, EntityAccessRights grantedRights) {
      var entCommand = command as EntityCommand;
      AccessDenied(entCommand.AuthorizationAction, entCommand.TargetEntityInfo, grantedRights);
    }

    private void AccessDenied(Actions action, EntityInfo entity, EntityAccessRights grantedRights) {
      if (!MustThrowOnDenied(action)) return; 
      var msg = Util.SafeFormat("Actions(s) [{0}] denied for entity {1}, authority {2}.", 
        action, entity, UserContext.UserName);
      var authEx = new AuthorizationException(msg, entity.EntityType, action, false, grantedRights, this);
      throw authEx;
    }

    private void AccessDenied(Actions action, EntityRecord record, EntityMemberInfo member = null) {
      if (!MustThrowOnDenied(action)) return;
      var msg = Util.SafeFormat("Actions(s) [{0}] denied for record {1}, user {2}",
        action, record.ToString(), UserContext.UserName);
      if (member != null)
        msg += "(Property " + member.MemberName + ")";
      var authEx = new AuthorizationException(msg, record.EntityInfo.EntityType, action, true, record.AccessRights, this);
      throw authEx;
    }

    private bool MustThrowOnDenied(Actions action) {
      var isReadAction = (action & Actions.Read) == action; // Read = Peek | ReadStrict
      return !(isReadAction && this.DenyReadAction == DenyReadActionType.Filter);
    }

    public bool CanExecute(IEntityCommand command, out EntityAccessRights grantedRights) {
      var entCommand = command as EntityCommand;
      grantedRights = GetEntityTypeAccess(entCommand.TargetEntityInfo);
      var action = entCommand.AuthorizationAction;
      return (grantedRights.Actions & action) == action;
    }

    //Must be used in "using" statements, ex: using(context.ElevateRead()) {....}
    public IDisposable ElevateRead() {
      _elevationCount++;
      return new ElevationToken() { Session = this };
    }

    #region ElevationToken class

    internal class ElevationToken : IDisposable {
      internal SecureSession Session;

      public void Dispose() {
        if (Session._elevationCount > 0) //just in case we are over-elevated
          Session._elevationCount--;
      }
    }

    #endregion



  }//class
}
