﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Data;
using System.Linq;
using System.Text;
using System.Reflection;
using System.Reflection.Emit;

using Vita.Common;
using Vita.Entities.Model;
using Vita.Entities.Model.Construction;
using Vita.Entities.Runtime;
using Vita.Logging;
using Vita.Data;

namespace Vita.Entities {

  // Non-special model attributes contain their own code and system does not know any specifics about them. 
  // System calls their Apply methods, and attributes do all needed stuff

  // Model attributes are self-handlers
  public abstract class ModelAttribute : Attribute, IAttributeHandler {
    public virtual AttributeApplyOrder ApplyOrder { get; protected set;}
    public virtual void Apply(AttributeContext context, Attribute attribute, EntityInfo entity) { }
    public virtual void Apply(AttributeContext context, Attribute attribute, EntityMemberInfo member) { }
    public virtual CustomAttributeBuilder Clone(Attribute attribute) {
      return null; 
    }
    public ModelAttribute() {
      ApplyOrder = AttributeApplyOrder.Default;
    }
    protected string GetAttributeName() {
      const string suffix = "Attribute";
      var name = this.GetType().Name;
      if (name.EndsWith(suffix))
        name = name.Substring(0, name.Length - suffix.Length);
      return name; 

    }
  }

  public partial class PersistOrderInAttribute : ModelAttribute {
    public override void Apply(AttributeContext context, Attribute attribute, EntityMemberInfo member) {
      if (member.Kind != MemberKind.EntityList) {
        context.Log.Error("PersistOrderIn attribute may be specified only on list members. Entity {0}.", member.Entity.Name);
        return; 
      }
      EntityInfo orderedEntity = null; 
      //determine the entity that is ordered
      switch (member.ChildListInfo.RelationType) {
        case EntityRelationType.ManyToOne:   orderedEntity = member.ChildListInfo.TargetEntity; break;
        case EntityRelationType.ManyToMany: orderedEntity = member.ChildListInfo.LinkEntity; break; 
      }
      //check that there is a member 
      var orderMember = orderedEntity.GetMember(this.Property);
      if (orderMember == null) {
        context.Log.Error("Property '{0}' referenced in PersistOrderIn attribute on entity {1} not found in entity {2}.",
          this.Property, member.Entity.Name, orderedEntity.Name);
        return; 
      }
      //current limitation - index property must be int32 only
      if (orderMember.DataType != typeof(Int32)) {
        context.Log.Error("Invalid data type ({0}) for property '{1}' referenced in PersistOrderIn attribute on entity {2}: must be Int32.",
          orderMember.DataType.Name, this.Property, member.Entity.Name);
        return;
      }
      // Validation passed, assign order member
      member.ChildListInfo.PersistentOrderMember = orderMember;
    }
  }

  public partial class OrderByAttribute {
    public override void Apply(AttributeContext context, Attribute attribute, EntityInfo entity) {
      if (entity.DefaultOrderBy != null)
        context.Log.Error("More than one OrderBy attribute in entity {0}.", entity.Name);
      ConstructOrderBy(context, entity);
    }

    public override void Apply(AttributeContext context, Attribute attribute, EntityMemberInfo member) {
      var entity = member.Entity; 
      if (member.Kind == MemberKind.EntityList) {
        //This is a special case - OrderBy attribute specifies the order of entities in list property.
        member.ChildListInfo.OrderBy = OrderedPropertyRef.CreateFromNames(this.OrderByList);
      } else 
        context.Log.Error("OrderBy attribute may be used only on entities or list properties. Entity: {0}, property: {1}", entity.Name, member.MemberName);
    }

    private void ConstructOrderBy(AttributeContext context, EntityInfo entity) {
      var orderByList = new List<OrderByEntryInfo>();
      var propRefs = OrderedPropertyRef.CreateFromNames(OrderByList);
      foreach(var propRef in propRefs) {
        var member = entity.GetMember(propRef.Name);
        if (member == null) {
          context.Log.Error("Property {0} in OrderBy attribute not found in entity {1}.", propRef.Name, entity.FullName);
          continue; 
        }
        if (member.Kind != MemberKind.Column) {
          context.Log.Error("Invalid property {0} in OrderBy attribute in entity {1} - must be a plain property.", propRef.Name, entity.FullName);
          continue;
        }
        if (propRef.Order != "asc" && propRef.Order != "desc") {
          context.Log.Error("Property {0}.{1}: invalid order specification ({2}), should be 'asc' or 'desc'. Ex: 'LastName:desc'.", entity.FullName, propRef.Name, propRef.Order);
          continue; 
        }
        bool isDesc = propRef.Order == "desc";
        var entry = new OrderByEntryInfo(member, isDesc);
        orderByList.Add(entry);
      }
      if (orderByList.Count == 0)
        return;
      entity.DefaultOrderBy = orderByList.ToArray();
    }

  }// class

  public partial class OldNamesAttribute : ModelAttribute {
    public override void Apply(AttributeContext context, Attribute attribute, EntityInfo entity) {
      entity.OldNames = Util.SafeSplit(this.OldNames);
    }

    public override void Apply(AttributeContext context, Attribute attribute, EntityMemberInfo member) {
      member.OldNames = Util.SafeSplit(this.OldNames);
    }

  }// class

  public partial class PropertyGroupAttribute : ModelAttribute {
    public override void Apply(AttributeContext context, Attribute attribute, EntityInfo entity) {
      if (string.IsNullOrWhiteSpace(this.GroupName)) {
        context.Log.Error("Group name may not be empty. Entity: {0}.", entity.Name);
        return; 
      }
      var names =  Util.SafeSplit(this.MemberNames);
      foreach (var name in names) {
        var member = entity.GetMember(name);
        if (member == null) {
          context.Log.Error("PropertyGroup '{0}', entity {1}: member {2} not found.", this.GroupName, entity.Name, name);
          continue; 
        }
        var grp = entity.GetPropertyGroup(this.GroupName, create: true);
        if (!grp.Members.Contains(member))
          grp.Members.Add(member); 
      }//foreach
    }
  } //class

  public partial class GroupsAttribute : ModelAttribute {
    public override void Apply(AttributeContext context, Attribute attribute, EntityMemberInfo member) {
      if (string.IsNullOrWhiteSpace(this.GroupNames))
        return; 
      var names = Util.SafeSplit(this.GroupNames);
      foreach (var name in names) {
        if (string.IsNullOrWhiteSpace(name))
          continue; 
        var grp = member.Entity.GetPropertyGroup(name, create: true);
        if (!grp.Members.Contains(member))
          grp.Members.Add(member);
      }//foreach
    }
  } //class

  public partial class NoUpdateAttribute : ModelAttribute {
    public override void Apply(AttributeContext context, Attribute attribute, EntityMemberInfo member) {
      member.Flags |= EntityMemberFlags.ExcludeFromUpdate;
    }
  }// class

  public partial class DateOnlyAttribute {
    private SetValueHandler _defaultSetter;

    public override void Apply(AttributeContext context, Attribute attribute, EntityMemberInfo member) {
      var dataType = member.DataType;
      if (dataType.IsGenericType)
        dataType = dataType.GetGenericArguments()[0];
      bool isDateTime = dataType == typeof(DateTime) || dataType == typeof(DateTimeOffset);
      if (!isDateTime) {
        context.Log.Error("Property {0}.{1}: DateOnly attribute may be specified only on DataTime or DateTimeOffset properties. ",  member.Entity.Name, member.MemberName);
        return; 
      }
      //Inject interceptor
      _defaultSetter = member.SetValueRef;
      if (dataType == typeof(DateTime))
        member.SetValueRef = this.SetValueDateTime;
      else
        member.SetValueRef = this.SetValueDateTimeOffset;
    }

    //Interceptors for SetValue
    void SetValueDateTime(EntityRecord record, EntityMemberInfo member, object value) {
      if (value != null && value != DBNull.Value) {
        if (value.GetType() == typeof(DateTime?))
          value = (DateTime?)((DateTime?)value).Value.Date;
        else
          value = ((DateTime)value).Date;
      }// if value != null
      _defaultSetter(record, member, value);
    }

    void SetValueDateTimeOffset(EntityRecord record, EntityMemberInfo member, object value) {
      if (value != null && value != DBNull.Value) {
        if (value.GetType() == typeof(DateTimeOffset?)) {
          var date = ((DateTimeOffset?)value).Value.Date;
          value = (DateTimeOffset?) (new DateTimeOffset(date));
        }
        else
          value = ((DateTimeOffset)value).Date;
      }// if value != null
      _defaultSetter(record, member, value);
    }
  }// class

  public partial class UtcAttribute {
    private SetValueHandler _defaultSetter; //The previously assigned setter

    public override void Apply(AttributeContext context, Attribute attribute, EntityMemberInfo member) {
      var dataType = member.DataType;
      member.Flags |= EntityMemberFlags.Utc;
      if (member.DataType != typeof(DateTime) && member.DataType != typeof(DateTime?)) {
        context.Log.Error("Utc attribute may be specified only on DataTime properties. Entity.Member: {0}.{1}", member.Entity.Name, member.MemberName);
        return;
      }
      //Inject interceptor
      _defaultSetter = member.SetValueRef;
      member.SetValueRef = this.SetValueInterceptor;
    }
    //Interceptor for SetValue
    void SetValueInterceptor(EntityRecord record, EntityMemberInfo member, object value) {
      if (value == null || value == DBNull.Value) {
        _defaultSetter(record, member, value);
        return; 
      }
      DateTime dtValue;
      if (value.GetType() == typeof(DateTime?))
        dtValue = ((DateTime?)value).Value;
      else
        dtValue = (DateTime)value;
      // If value is coming from database or from serialization, its Kind shows Unspecified - set it to UTC explicitly; as we know that it is in fact UTC
      //if (record.Status == EntityStatus.Loading && dtValue.Kind == DateTimeKind.Unspecified)
      if (dtValue.Kind == DateTimeKind.Unspecified)
          dtValue = new DateTime(dtValue.Ticks, DateTimeKind.Utc);
      else
        dtValue = dtValue.ToUniversalTime();
      _defaultSetter(record, member, dtValue); 
    }

  }// class


  public partial class MemoAttribute {
    public MemoAttribute() : base() {  }
    public override void Apply(AttributeContext context, Attribute attribute, EntityMemberInfo member) {
      member.Size = -1;
    }

  }// class

  public partial class CurrencyAttribute {
    public override void Apply(AttributeContext context, Attribute attribute, EntityMemberInfo member) {
      member.ExplicitDbType = DbType.Currency;
    }
  }

  public partial class ComputedAttribute {
    private MethodInfo _method;

    public override void Apply(AttributeContext context, Attribute attribute, EntityMemberInfo member) {
      if (!this.Persist) {
        member.Kind = MemberKind.Transient;
      }
      member.Flags |= EntityMemberFlags.Computed;
      _method = this.MethodClass.GetMethod(this.MethodName);
      if (_method == null)
        context.Log.Error("Method {0} for computed column {1} not found in type {2}", 
          this.MethodName, member.MemberName, this.MethodClass);
      member.GetValueRef = GetComputedValue;
      member.SetValueRef = MemberValueGettersSetters.DummySetValue;
    }

    public object GetComputedValue(EntityRecord record, EntityMemberInfo member) {
      if (_method == null) return null;
      var value = _method.Invoke(null, new object[] { record.EntityInstance });
      if (Persist)
        record.SetValueDirect(member, value); 
      return value; 
    }

  }// class

  public partial class DependsOnAttribute {
    public override void Apply(AttributeContext context, Attribute attribute, EntityMemberInfo member) {
      var entity = member.Entity;
      var namesArr = Util.SafeSplit(this.MemberNames);
      foreach (var name in namesArr) {
        var targetMember = entity.GetMember(name);
        if (targetMember == null) {
          context.Log.Error("Member {0} referenced in DependsOn attribute on member {1}.{2} not found.", name, entity.Name, member.MemberName);
          continue; 
        }
        //add this member to DependentMembers array of targetMember
        if (targetMember.DependentMembers == null)
          targetMember.DependentMembers = new EntityMemberInfo[] { member };
        else {
          var mList = targetMember.DependentMembers.ToList(); 
          mList.Add(member);
          targetMember.DependentMembers = mList.ToArray(); 
        }
      }//foreach name
    } // method
  }//class

  public partial class ValidateAttribute {
    MethodInfo _method;

    public override void Apply(AttributeContext context, Attribute attribute, EntityInfo entity) {
      _method = MethodClass.GetMethod(MethodName);
      if (_method == null)
        context.Log.Error("Method {0} specified as Validation method for entity {1} not found in type {2}",
            MethodName, entity.EntityType, MethodClass);
      entity.Events.ValidatingChanges += Events_Validating;
    }

    void Events_Validating(EntityRecord record, EventArgs args) {
      _method.Invoke(null, new object[] { record.Session.Validator, record.EntityInstance });
    }
  }// class

  public partial class PagedAttribute : ModelAttribute {
    public override void Apply(AttributeContext context, Attribute attribute, EntityInfo entity) {
      entity.PagingMode = PagingMode.DataStore;
    }
  }// class

  public partial class AutoAttribute : ModelAttribute {
    public EntityMemberInfo Member;
    private SetValueHandler _defaultHandler; 
    // Used for generating temporary Identity values
    private static long _currentTemporaryIdentity;
   

    public override void Apply(AttributeContext context, Attribute attribute, EntityMemberInfo member) {
      Member = member;
      _defaultHandler = Member.SetValueRef; 
      var entity = member.Entity;
      member.Flags |= EntityMemberFlags.AutoValue;
      member.AutoValueType = this.Type;
      switch (this.Type) {
        case AutoType.Identity:
          member.Flags |=  EntityMemberFlags.ExcludeFromUpdate;
          if (this is IdentityAttribute) {
            member.Flags |= EntityMemberFlags.DbGenererated | EntityMemberFlags.ExcludeFromUpdate;
            if (!member.DataType.IsInt()) {
              context.Log.Error("Entity member {0}.{1}, type {2}: Identity attribute may be set only on member of integer types. ",
                Member.Entity, Member.MemberName, this.Type);
              return; 
            }
            entity.Events.New += EntityEvent_NewEntityHandleIdentity;
            entity.SaveEvents.SubmittedChanges += EntityEvent_IdentityEntitySubmitted;
          }// if is identity
          break;

        case AutoType.NewGuid:
          if (!CheckDataType(context, typeof(Guid)))
            return; 
          entity.Events.New += EntityEvent_HandleNewGuid;
          break;

        case AutoType.CreatedOn:
        case AutoType.UpdatedOn:
          if (!CheckDataType(context, typeof(DateTime), typeof(DateTimeOffset)))
            return; 
          if (this.Type == AutoType.CreatedOn)
            member.Flags |= EntityMemberFlags.ExcludeFromUpdate;
          if (member.DataType == typeof(DateTime) || member.DataType == typeof(DateTime?))
            entity.SaveEvents.SavingChanges += EntityEvent_HandleCreatedUpdatedOnDateTime;
          else
            entity.SaveEvents.SavingChanges += EntityEvent_HandleCreatedUpdatedOnDateTimeOffset;
          break;

        case AutoType.CreatedBy:
          if (!CheckDataType(context, typeof(string)))
            return;
          entity.Events.New += EntityEvent_HandleUpdatedCreatedBy;
          member.Flags |= EntityMemberFlags.ExcludeFromUpdate;
          break; 

        case AutoType.UpdatedBy:
          if (!CheckDataType(context, typeof(string)))
            return;
          entity.Events.New += EntityEvent_HandleUpdatedCreatedBy;
          entity.Events.Modified += EntityEvent_HandleUpdatedCreatedBy;
          break;
 
        case AutoType.CreatedById:
          entity.Events.New += EntityEvent_HandleUpdatedCreatedById;
          member.Flags |= EntityMemberFlags.ExcludeFromUpdate;
          break;
 
        case AutoType.UpdatedById:
          entity.Events.New += EntityEvent_HandleUpdatedCreatedById;
          entity.Events.Modified += EntityEvent_HandleUpdatedCreatedById;
          break; 

        case AutoType.RowVersion:
          member.Flags |= EntityMemberFlags.DbGenererated | EntityMemberFlags.RowVersion;
          member.Entity.Flags |= EntityFlags.HasRowVersion;
          member.ExplicitDbType = CustomDbTypes.RowVersion; 
          member.Size = 8;
          break; 
      }//swith AutoValueType
    }

    private bool CheckDataType(AttributeContext context, params Type[] types) {
      var dt = Member.DataType;
      if (dt.IsGenericType)
        dt = dt.GetGenericArguments()[0]; // check for DateTime?
      if (types.Contains(dt)) 
        return true;
      context.Log.Error("Entity member {0}.{1}: Auto({2})  attribute may be set only on member of type(s): {3}. ", 
        Member.Entity, Member.MemberName, this.Type, string.Join(", ", types.Select(t=>t.Name)));
      return false; 
    }

    void EntityEvent_NewEntityHandleIdentity(EntityRecord record, EventArgs args) {
      if (record.SuppressAutoValues) return; 
      var tempId = System.Threading.Interlocked.Decrement(ref _currentTemporaryIdentity);
      record.SetValueDirect(Member, tempId);
    }

    void EntityEvent_IdentityEntitySubmitted(EntityRecord record, EventArgs args) {
      if (record.Status != EntityStatus.New) return;
      var recs = record.Session.RecordsChanged;
      foreach (var refMember in record.EntityInfo.IncomingReferences) {
        var childRecs = recs.Where(r => r.EntityInfo == refMember.Entity);
        foreach (var childRec in childRecs) {
          var refBackValue = childRec.ValuesTransient[refMember.ValueIndex];
          if (refBackValue == record)
            refMember.SetValueRef(childRec, refMember, record.EntityInstance); //this will copy foreign key
        }//foreach refRec
      }
    }//method


    void EntityEvent_HandleNewGuid(EntityRecord record, EventArgs args) {
      if (record.SuppressAutoValues) return;
      var newGuid = Externals.NewSequentialGuid();
      record.SetValueDirect(Member, newGuid);
    }

    void EntityEvent_HandleUpdatedCreatedBy(EntityRecord record, EventArgs args) {
      if (record.SuppressAutoValues) return;
      var userName = record.Session.UserContext.UserName;
      record.SetValueDirect(Member, userName);
    }

    void EntityEvent_HandleUpdatedCreatedById(EntityRecord record, EventArgs args) {
      if (record.SuppressAutoValues) return;
      var userId = record.Session.UserContext.UserId;
      record.SetValueDirect(Member, userId);
    }

    void EntityEvent_HandleCreatedUpdatedOnDateTime(EntityRecord record, EventArgs args) {
      if (record.SuppressAutoValues) return;
      //Do CreatedOn only if it is new record
      if (this.Type == AutoType.CreatedOn && record.Status != EntityStatus.New) return;
      var dateTime = record.Session.LastTransactionDateTime.Value;
      if (Member.Flags.IsSet(EntityMemberFlags.Utc))
        dateTime = dateTime.ToUniversalTime();
      _defaultHandler(record, Member, dateTime); 
    }

    void EntityEvent_HandleCreatedUpdatedOnDateTimeOffset(EntityRecord record, EventArgs args) {
      if (record.SuppressAutoValues) return;
      //Do CreatedOn only if it is new record
      if (this.Type == AutoType.CreatedOn && record.Status != EntityStatus.New) return;
      var offset = new DateTimeOffset(record.Session.LastTransactionDateTime.Value);
      _defaultHandler(record, Member, offset);
    }

  }// class

  public partial class OwnerAttribute : ModelAttribute {
    public override void Apply(AttributeContext context, Attribute attribute, EntityMemberInfo member) {
      member.Flags |= EntityMemberFlags.IsOwner;
      if (member.Entity.OwnerMember == null)
        member.Entity.OwnerMember = member;
      else
        context.Log.Error("More than one Owner attribute is specified on entity {0}.", member.Entity.FullName);
    }
  }

  public partial class SecretAttribute : ModelAttribute {
    public override void Apply(AttributeContext context, Attribute attribute, EntityMemberInfo member) {
      member.Flags |= EntityMemberFlags.Secret | EntityMemberFlags.ExcludeFromUpdate; //updates are possible only thru custom update method
      member.GetValueRef = this.GetSecretValue;
    }

    private object GetSecretValue(EntityRecord record, EntityMemberInfo member) {
      //Always return value from Modified values - this value just set by the code; but no way to read value from database (which is in OriginalValues)
      var value = record.ValuesModified[member.ValueIndex];
      if (value == null)
        value = member.DeniedValue;
      if (value == DBNull.Value)
        return null;
      return value;
    }
  }

  public partial class IndexAttribute : ModelAttribute {
    public override void Apply(AttributeContext context, Attribute attribute, EntityInfo entity) {
      var mRefs = OrderedPropertyRef.CreateFromNames(this.MemberNames);
      var indName = this.IndexName;
      if (string.IsNullOrWhiteSpace(indName)) {
        var mNames = string.Join(string.Empty, mRefs.Select(mr => mr.Name));
        indName = GetAutoNamePrefix() + "_" + entity.Name + "_" + mNames;
        indName = entity.GetUniqueKeyName(indName);
      }
      var index = new EntityKeyInfo(indName, this.KeyType, entity); 
      foreach (var mRef in mRefs) {
        var member = entity.FindMemberOrColumn(mRef.Name);
        if (member == null) {
          context.Log.Error("Entity {0}: property {1} listed in {2} attribute not found.", entity.FullName, mRef.Name, this.GetAttributeName());
          return;
        }
        index.Members.Add(member);
        if (mRef.Order == "desc")
          index.DescendingMembers.Add(member); 
      } //foreach mRef      
    }//method

    public override void Apply(AttributeContext context, Attribute attribute, EntityMemberInfo member) {
      var entity = member.Entity;
      var indName = this.IndexName;
      if (string.IsNullOrWhiteSpace(indName)) {
        indName = GetAutoNamePrefix() + "_" + entity.Name + "_" + member.MemberName;
        indName = entity.GetUniqueKeyName(indName);
      }
      var index = new EntityKeyInfo(indName, this.KeyType, entity);
      index.Members.Add(member);
    }

    protected string GetAutoNamePrefix() {
      var result = "IX";
      if (this.KeyType.IsSet(KeyType.Clustered))
        result += "C";
      if (this.KeyType.IsSet(KeyType.Unique))
        result += "U";
      return result;
    }
  }//class

  public partial class IsSystemAttribute : ModelAttribute {
    public override void Apply(AttributeContext context, Attribute attribute, EntityInfo entity) {
      entity.Flags |= EntityFlags.IsSystem; 
    }
  }
  public partial class SmallTableAttribute : ModelAttribute {
    public override void Apply(AttributeContext context, Attribute attribute, EntityInfo entity) {
      entity.Flags |= EntityFlags.SmallTable;
    }
  }

  public partial class DynamicAttribute : ModelAttribute {
    public override void Apply(AttributeContext context, Attribute attribute, EntityInfo entity) {
      entity.Flags |= EntityFlags.Dynamic;
    }
  }

  public partial class DisplayAttribute : ModelAttribute {
    private MethodInfo _customMethodInfo;
    private string _adjustedFormat;
    private int[] _argIndexes;
    public override void Apply(AttributeContext context, Attribute attribute, EntityInfo entity) {
      if (this.MethodClass != null && this.MethodName != null) {
        _customMethodInfo = MethodClass.GetMethod(MethodName);
        if (_customMethodInfo == null) {
          context.Log.Error("Method {0} specified as Display method for entity {1} not found in type {2}", MethodName, entity.EntityType, MethodClass);
          return; 
        }
        entity.DisplayMethod = InvokeCustomDisplay;
        return; 
      }
      //Check if Format provided
      if (string.IsNullOrEmpty(Format)) {
        context.Log.Error("Invalid Display attribute on entity {0}. You must provide method reference or non-empty Format value.", entity.EntityType);
        return;
      }
      //Parse Format value, build argIndexes from referenced property names
      // Check if it is just single prop name
      if (Format.IndexOfAny(new char[] { '{', '}', ' ' }) < 0) {
        var member = entity.GetMember(Format);
        if (member != null) {
          _argIndexes = new int[] { member.Index };
          entity.DisplayMethod = GetDispalyStringSimple;
          return; 
        }
      }
      // Do real parsing
      string[] argNames;
      ParseTemplate(Format, out _adjustedFormat, out argNames); 
      //verify and build arg indexes
      _argIndexes = new int[argNames.Length];
      for (var i = 0; i < _argIndexes.Length; i++) {
        var argName = argNames[i];
        var member = entity.GetMember(argName);
        if (member == null) {
          context.Log.Error("Invalid Format expression in Display attribute on entity {0}. Property {1} not found.", entity.EntityType, argName);
          continue; 
        }
        _argIndexes[i] = member.Index;
      }
      entity.DisplayMethod = GetDisplayString;
    }

    private static void ParseTemplate(string template, out string adjustedTemplate, out string[] argNames) {
      // Double brace '{{' is 'escaped brace'. We start by replacing these doubles with \b symbol. At the end we'll replace it back with single brace
      template = template.Replace("{{", "\b");
      var argNameList = new StringList(); 
      //split by left brace {
      var segments = template.Split(new char[]{'{'}, StringSplitOptions.RemoveEmptyEntries);
      for (int i = 0; i < segments.Length; i++) {
        var segm = segments[i];
        if (string.IsNullOrEmpty(segm)) continue; 
        //find closing brace, extract prop name, replace it with index
        var rPos = segm.IndexOf('}');
        if (rPos < 0) rPos = segm.Length; // catch error by including all
        var argName = segm.Substring(0, rPos);
        var tail = segm.Substring(rPos + 1);
        argNameList.Add(argName);
        segments[i] = "{" + i + "}" + tail; // replace 'propName} abcd'   with '{0} abcd'
      }
      // Merge modified segments
      adjustedTemplate = string.Join(string.Empty, segments);
      adjustedTemplate = adjustedTemplate.Replace('\b', '{');
      argNames = argNameList.ToArray();      
    }//method

    private string InvokeCustomDisplay(EntityRecord record) {
      //TODO: implement compiled delegate for perf
      return (string) _customMethodInfo.Invoke(null, new object[] { record.EntityInstance });
    }

    private string GetDisplayString(EntityRecord record) {
      if (_argIndexes == null) return Format; 
      var args = new object[_argIndexes.Length];
      for (int i = 0; i < args.Length; i++)
        args[i] = record.GetValue(_argIndexes[i]);
      return Util.SafeFormat(_adjustedFormat, args);
    }

    private string GetDispalyStringSimple(EntityRecord record) {
      return string.Empty + record.GetValue(_argIndexes[0]);
      
    }

  }//class

  public partial class CascadeDeleteAttribute : ModelAttribute {
    public override void Apply(AttributeContext context, Attribute attribute, EntityMemberInfo member) {
      if (member.Kind != MemberKind.EntityRef) {
        context.Log.Error("CascadeDelete attribute may be used only on properties that are references to other entities. Property: {0}.{1}", 
            member.Entity.Name, member.MemberName);
        return;
      }
      member.Flags |= EntityMemberFlags.CascadeDelete;
    }
  }

  public partial class NoClusteredIndexAttribute : ModelAttribute {
    public override void Apply(AttributeContext context, Attribute attribute, EntityInfo entity) {
      entity.Flags |= EntityFlags.NoClusteredIndex;
    }
  }

}//ns
