﻿using System;
using System.Collections.Generic;
using System.Data;
using System.Linq;
using Vita.Common;

using Vita.Entities.Model;
using Vita.Entities.Model.Construction;

namespace Vita.Entities {
  // Special framework attributes, known to core framework and applied using specific internal logic

  public abstract class SpecialModelAttribute : ModelAttribute {
    public SpecialModelAttribute() {
      base.ApplyOrder = AttributeApplyOrder.System;
    }
  }

  public partial class EntityAttribute : SpecialModelAttribute {
    public override void Apply(AttributeContext context, Attribute attribute, EntityInfo entity) {
      if (!string.IsNullOrWhiteSpace(this.Name))
        entity.Name = this.Name;
      entity.TableName = this.TableName;
    }
  } //class

  public partial class ForEntityAttribute : SpecialModelAttribute { }

  public partial class ColumnAttribute : SpecialModelAttribute {

    public override void Apply(AttributeContext context, Attribute attribute, EntityMemberInfo member) {
      if (!string.IsNullOrWhiteSpace(this.ColumnName)) member.ColumnName = this.ColumnName;
      if (!string.IsNullOrWhiteSpace(this.Default)) member.ColumnDefault = this.Default;
      if (this.Scale > 0) member.Scale = this.Scale;
      if (this.Precision > 0) member.Precision = this.Precision;
      if (this.Size != 0)
        member.Size = this.Size;
      member.ExplicitDbType = this._dbTypeValue;
    }
  }// class

  public partial class NoColumnAttribute : SpecialModelAttribute {
    public override void Apply(AttributeContext context, Attribute attribute, EntityMemberInfo member) {
      member.Kind = MemberKind.Transient;
      member.GetValueRef = MemberValueGettersSetters.GetTransientValue;
      member.SetValueRef = MemberValueGettersSetters.SetTransientValue;
    }
  }// class

  public partial class SizeAttribute : SpecialModelAttribute {
    public override void Apply(AttributeContext context, Attribute attribute, EntityMemberInfo member) {
      member.Size = Size;
    }
  }// class

  public partial class NullableAttribute {

    public override void Apply(AttributeContext context, Attribute attribute, EntityMemberInfo member) {
      member.Flags |= EntityMemberFlags.Nullable;
      if (member.DataType.IsValueType) {
        member.Flags |= EntityMemberFlags.ReplaceDefaultWithNull;
        member.GetValueRef = MemberValueGettersSetters.GetValueTypeReplaceNullWithDefault;
        member.SetValueRef = MemberValueGettersSetters.SetValueTypeReplaceDefaultWithNull;
      }
    }
  }// class


  public partial class PrimaryKeyAttribute : SpecialModelAttribute {
    public override void Apply(AttributeContext context, Attribute attribute, EntityInfo entity) {
      if (string.IsNullOrWhiteSpace(this.MemberNames)) {
        context.Log.Error("Entity {0}: primary key must specify property name(s) when used on Entity interface.", entity.FullName);
        return;
      }
      if (!CreatePrimaryKey(context, entity)) return;
      var mRefs = OrderedPropertyRef.CreateFromNames(this.MemberNames);
      foreach (var mRef in mRefs) {
        var member = entity.FindMemberOrColumn(mRef.Name);
        if (member == null) {
          context.Log.Error("Entity {0}: property {1} listed in PrimaryKey attribute not found.", entity.FullName, mRef.Name);
          return;
        }
        member.Flags |= EntityMemberFlags.PrimaryKey;
        entity.PrimaryKey.Members.Add(member); 
      }
    }
    public override void Apply(AttributeContext context, Attribute attribute, EntityMemberInfo member) {
      var entity = member.Entity;
      if (!CreatePrimaryKey(context, entity)) return;
      entity.PrimaryKey.Members.Add(member);
      member.Flags |= EntityMemberFlags.PrimaryKey;
    }
    private bool CreatePrimaryKey(AttributeContext context, EntityInfo entity) {
      if (entity.PrimaryKey != null) {
        context.Log.Error("More than one primary key specified on entity {0}", entity.FullName);
        return false; 
      }
      entity.PrimaryKey = new EntityKeyInfo("PK_" + entity.Name, KeyType.PrimaryKey, entity);
      return true; 
    }
  } //class

  public partial class OneToManyAttribute : SpecialModelAttribute {

    public override void Apply(AttributeContext context, Attribute attribute, EntityMemberInfo member) {
      var listInfo = member.ChildListInfo = new ChildEntityListInfo();
      listInfo.RelationType = EntityRelationType.ManyToOne;
      var entType = member.Entity.EntityType;
      var targetType = member.DataType.GetGenericArguments()[0];
      listInfo.TargetEntity = context.RuntimeModel.GetEntityInfo(targetType, true);
      if (!string.IsNullOrWhiteSpace(this.ThisEntityRef)) {
        var fkMember = listInfo.TargetEntity.GetMember(this.ThisEntityRef);
        if (fkMember == null) {
          context.Log.Error("EntityList member {0}.{1}: could not find property {2} in target entity. ",
            entType, member.MemberName, this.ThisEntityRef);
          return;
        }
        this.ThisEntityRef = fkMember.MemberName;
        listInfo.ParentRefMember = fkMember;
      } else
        listInfo.ParentRefMember = listInfo.TargetEntity.FindEntityRefMember(this.ThisEntityRef, entType, member, context.Log);
      //Check that reference is found
      if (listInfo.ParentRefMember == null)
        context.Log.Error("EntityList member {0}.{1}: could not find reference property in target entity. ", entType, member.MemberName);
    }
  }// class

  public partial class ManyToManyAttribute : SpecialModelAttribute {
    public override void Apply(AttributeContext context, Attribute attribute, EntityMemberInfo member) {
      var linkEntityType = this.LinkEntity;
      var listInfo = member.ChildListInfo = new ChildEntityListInfo();
      listInfo.RelationType = EntityRelationType.ManyToMany;
      listInfo.LinkEntity = context.RuntimeModel.GetEntityInfo(linkEntityType, true);
      listInfo.ParentRefMember = listInfo.LinkEntity.FindEntityRefMember(this.ThisEntityRef, member.Entity.EntityType, member, context.Log);
      var targetEntType = member.DataType.GetGenericArguments()[0];
      listInfo.OtherEntityRefMember = listInfo.LinkEntity.FindEntityRefMember(this.OtherEntityRef, targetEntType, member, context.Log);
      if (listInfo.OtherEntityRefMember != null)
        listInfo.TargetEntity = context.RuntimeModel.GetEntityInfo(listInfo.OtherEntityRefMember.DataType, true);
    }//method
  }// class

  /// <summary>Marks an entity references, optional. Might be used to specify the names of foreign key columns. </summary>
  ///<remarks>The primary use is for reverse-engineering tool (database to c# code generator) to explicitly identify the foreign key columns in the database table.
  ///</remarks> 
  [AttributeUsage(AttributeTargets.Property)]
  public partial class EntityRefAttribute : SpecialModelAttribute {
    public string KeyColumns;

    public EntityRefAttribute(string keyColumns = null) {
      KeyColumns = keyColumns;
    }
    public override void Apply(AttributeContext context, Attribute attribute, EntityMemberInfo member) {
      if (member.Kind != MemberKind.EntityRef) {
        context.Log.Error("EntityRef attribute may be used only on properties that are references to other entities. Property: {0}.{1}",
            member.Entity.Name, member.MemberName);
        return;
      }
      var entity = member.Entity;
      // Create reference info
      var targetType = member.Property.PropertyType;
      var targetEntity = context.RuntimeModel.GetEntityInfo(targetType);
      Util.Check(targetEntity != null, "Target entity not found: {0}", targetType);
      //Create foreign key
      var fkName = "FK_" + entity.Name + "_" + member.MemberName;
      var fk = new EntityKeyInfo(fkName, KeyType.ForeignKey, entity, member);
      fk.Members.Add(member);
      member.ReferenceInfo = new EntityReferenceInfo(member, fk, targetEntity.PrimaryKey);
      member.ReferenceInfo.ForeignKeyColumns = this.KeyColumns;
    }
  }//class




}
