﻿using System;
using System.Collections.Generic;
using System.Data;
using System.Linq;
using System.Text;

using Vita.Common;
using Vita.Entities.Runtime;
using Vita.Logging;

namespace Vita.Entities.Model.Construction {
  public class EntityCommandBuilder {
    EntityModelInfo _modelInfo;
    Log _log; 

    public EntityCommandBuilder(EntityModelInfo modelInfo, Log log) {
      _modelInfo = modelInfo;
      _log = log;
    }

    public EntityCommand BuildCrudSelectAllCommand(EntityInfo entity) {
      var cmd = AddCommand(entity.Name + "SelectAll", "Selects all entities.", EntityCommandKind.SelectAll, entity);
      return cmd; 
    }

    public EntityCommand BuildCrudSelectAllPagedCommand(EntityInfo entity) {
      var cmd = AddCommand(entity.Name + "SelectAllPaged", "Selects all entities.", EntityCommandKind.SelectAllPaged, entity);
      cmd.Parameters.Add(new EntityCommandParameter("__skiprows", typeof(int), 0, 0));
      cmd.Parameters.Add(new EntityCommandParameter("__maxrows", typeof(int), 0, int.MaxValue));
      return cmd;
    }

    public EntityCommand BuildCrudSelectByKeyCommand(EntityInfo entity, EntityKeyInfo key, bool topOne = false) {
      var procName = entity.Name + "SelectBy" + key.GetMemberNames(string.Empty, removeUnderscore: true);
      var descr = "Selects entity(ies) by key.";
      if (topOne) {
        procName += "_TopOne";
        descr += " (top one).";
      }
      var cmd = AddCommand(procName, descr, EntityCommandKind.SelectByKey, entity, key);
      if (topOne)
        cmd.Flags |= EntityCommandFlags.SelectOne;
      AddParametersFromKey(cmd, key);
      return cmd;
    }

    public EntityCommand BuildCrudInsertCommand(EntityInfo entity) {
      var cmd = AddCommand(entity.Name + "Insert", "Inserts a new entity.", EntityCommandKind.Insert, entity);
      //Add PK first
      AddParametersFromKey(cmd, entity.PrimaryKey);
      // Add the rest
      foreach (var member in entity.Members) {
        if (member.Kind != MemberKind.Column | member.Flags.IsSet(EntityMemberFlags.PrimaryKey)) // PK is already added
          continue; 
        var prm = new EntityCommandParameter(member);
        cmd.Parameters.Add(prm);
      }
      // set param flags
      foreach (var prm in cmd.Parameters) {
        switch(prm.SourceMember.AutoValueType) {
          case AutoType.RowVersion: 
          case AutoType.Identity:   prm.Direction = ParameterDirection.Output; break;
          default: prm.Direction = ParameterDirection.Input; break; 
        }
      }
      return cmd;
    }

    public EntityCommand BuildCrudUpdateCommand(EntityInfo entity) {
      var cmd = AddCommand(entity.Name + "Update", "Updates an entity.", EntityCommandKind.Update, entity);
      AddParametersFromKey(cmd, entity.PrimaryKey);
      foreach (var member in entity.Members) {
        if (member.Kind != MemberKind.Column || member.Flags.IsSet(EntityMemberFlags.PrimaryKey | EntityMemberFlags.ExcludeFromUpdate))
          continue;
        cmd.Parameters.Add(new EntityCommandParameter(member));
      }
      // set param flags
      foreach (var prm in cmd.Parameters) {
        switch (prm.SourceMember.AutoValueType) {
          case AutoType.RowVersion: prm.Direction = ParameterDirection.InputOutput; break;
          default: prm.Direction = ParameterDirection.Input; break;
        }
      }
      return cmd;
    }

    // Note: for entities with RowVersion, we do NOT check row version when deleting by ID - just checking row count. 
    // this seems reasonable, no need to check that version matches when we are deleting, just check it is not deleted already.
    public EntityCommand BuildCrudDeleteCommand(EntityInfo entity) {
      var cmd = AddCommand(entity.Name + "Delete", "Deletes an entity.", EntityCommandKind.Delete, entity);
      AddParametersFromKey(cmd, entity.PrimaryKey);
      return cmd;
    }

    public EntityCommand BuildSelectByKeyManyToMany(ChildEntityListInfo listInfo) {
      //for many to many, create command selecting real entities
      var targetEnt = listInfo.TargetEntity;
      var parentEnt = listInfo.ParentRefMember.Entity;
      var parentRefKey = listInfo.ParentRefMember.ReferenceInfo.FromKey; //FK from link entity to parent entity
      var cmdName = targetEnt.Name + "SelectBy_" + listInfo.LinkEntity.Name + "_" + listInfo.ParentRefMember.MemberName;
      var cmd = AddCommand(cmdName, "Selects entities by many-to-many link.", EntityCommandKind.SelectByKeyManyToMany, targetEnt, parentRefKey);
      cmd.SelectKeySecondary = listInfo.OtherEntityRefMember.ReferenceInfo.FromKey;  
      AddParametersFromKey(cmd, cmd.SelectKey);
      return cmd; 
    }

        public void ProcessCustomCommand(EntityCommand command) {
          if (command.TargetEntityInfo == null && !AssignEntityInfo(command)) //it is an error - entity info not found
            return; 
          //Individual processing by command kind
          switch (command.Kind) {
            case EntityCommandKind.CustomSelect: case EntityCommandKind.CustomInsert:
            case EntityCommandKind.CustomUpdate: case EntityCommandKind.CustomDelete:
              //BuildCustomCommandParameters(command); 
              break; 
            case EntityCommandKind.PartialUpdate:
              SetupCustomPartialUpdateCommand(command); 
              break; 
          }//switch
        }
/*
        // Builds command.Parameters and command.ExtendedParameters lists.
        // Parameters is a list explicitly specified CommandParameters found in the base query, introduces by app code.
        // ExtendedParameters is a list of all parameters for DbCommand: LINQ converts all literal constants in queries into 
        // SQL parameters; so ExtendedParameters => Parameters + (LINQ constants)
        private void BuildCustomCommandParameters(EntityCommand command) {
          command.Parameters.Clear();
          command.ExtendedParameters.Clear();

          // TRACE and check indexes ( i+1?)
          var baseQueryInfo = command.CustomBaseQuery.QueryInfo;
          var evaluatables = baseQueryInfo.LocallyEvaluatable;
          for (int i = 0; i < evaluatables.Count; i++) {
            var expr = evaluatables[i];
            // Arguments of Take(n) and Skip(n) are not transformed into parameters by LINQ, so skip these
            if (baseQueryInfo.NonParameterConstants.Contains(expr))
              continue;
            var autoParamName = "p" + command.ExtendedParameters.Count;
            var constValue = baseQueryInfo.LocallyEvaluatedValues[i + 1];
            var queryParam = constValue as EntityCommandParameter;
            if (queryParam != null) {
              //it is parameter
              queryParam.Name = autoParamName;
              command.ExtendedParameters.Add(queryParam);
              // Some parameters are explicitly declared constants created using EntityModule.Const() method
              if (!queryParam.IsLinqConst)
                command.Parameters.Add(queryParam);
              baseQueryInfo.LocallyEvaluatedValues[i + 1] = queryParam.DefaultValue; //replace Parameter object with default value constant
            } else {
              var size = GetConstSize(constValue);
              var prm = new EntityCommandParameter(autoParamName, expr.Type, size, constValue);
              prm.IsLinqConst = true;
              command.ExtendedParameters.Add(prm);
            }
          }//for i
        }
*/
    private int GetConstSize(object constValue) {
      var str = constValue as string;
      if (str != null) {
        if (str.Length == 0) return 1; //to avoid nvarchar(0)
        return str.Length;
      }
      var bytes = constValue as byte[];
      if (bytes != null)
        return bytes.Length;
      return 0;
    }

    internal void BuildCommandDescriptiveTag(EntityCommand command) {
      command.DescriptiveTag = command.TargetEntityInfo.FullName + "/" + command.Kind;
      switch (command.Kind) {
        case EntityCommandKind.SelectAll:
        case EntityCommandKind.SelectAllPaged:
          break;
        case EntityCommandKind.SelectByKey:
        case EntityCommandKind.SelectByKeyManyToMany:
          command.DescriptiveTag += "/" + command.SelectKey.GetMemberNames();
          if (command.Flags.IsSet(EntityCommandFlags.SelectOne))
            command.DescriptiveTag += "/TopOne";
          break;
        case EntityCommandKind.CustomSelect:
        case EntityCommandKind.CustomInsert:
        case EntityCommandKind.CustomUpdate:
        case EntityCommandKind.CustomDelete:
          command.DescriptiveTag += "/" + command.CommandName;
          break;
        case EntityCommandKind.PartialUpdate:
          command.DescriptiveTag += "/" + string.Join(string.Empty, command.UpdateMemberNames);
          break;
      }
    } 

    
    // Private Utilities -----------------------------------------------------------------
    private void SetupCustomPartialUpdateCommand(EntityCommand command) {
      AddParametersFromKey(command, command.TargetEntityInfo.PrimaryKey);
      foreach (var memberName in command.UpdateMemberNames) {
        var member = command.TargetEntityInfo.GetMember(memberName);
        if (member == null) {
          _log.Error("Member {0} not found in entity {1}, partial update command {2}", memberName, command.TargetEntityInfo.FullName, command.CommandName);
          continue; 
        }
        //Skip PK members - forgive programmer if he accidentally includes them into update list
        if (member.Flags.IsSet(EntityMemberFlags.PrimaryKey)) continue; 
        command.Parameters.Add(new EntityCommandParameter(member));        
      }      
    }

    private bool AssignEntityInfo(EntityCommand command) {
      var entInfo = _modelInfo.GetEntityInfo(command.TargetEntityType);
      if (entInfo == null) {
        _log.Error("Failed to compile custom command {0}: type {1} is not part of entity model.", command, command.TargetEntityType);
        return false;
      }
      command.TargetEntityInfo = entInfo;
      command.Area = entInfo.Area;
      return true;
    }


    private EntityCommand AddCommand(string commandName, string description, EntityCommandKind kind, EntityInfo entity, EntityKeyInfo selectKey = null) {
      var command = new EntityCommand(commandName, description, kind, entity, selectKey);
      _modelInfo.AddCommand(command);
      return command;
    }

    private void AddParametersFromKey(EntityCommand command, EntityKeyInfo key) {
      foreach (var member in key.ExpandedMembers) 
        command.Parameters.Add(new EntityCommandParameter(member));
    }

  }//class

}//ns
