﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Linq.Expressions;

using Vita.Common;

namespace Vita.Entities.Model {

  /// <summary> Data filter detects if a given entity "belongs" to a data group identified by a value in UserContext. </summary>
  /// <remarks>For example, we can create a data filter that indicates if a certain record (document) is related to the department
  /// that user/employee works in. Based on this information, the authorization system might grant or deny access to the document.
  /// In this scenario, the Id of the department is placed into the user context. The context value reader is setup to return this value
  /// to the data filter. The data filter is configured to return true/false indicator for a given entity (record) instance, 
  /// using expressions that retrieve the associated department ID from an entity and match it with the ID in the user context.
  /// See an example of using the data filters in BooksModel sample assembly.
  /// </remarks>
  public abstract class DataFilter {
    public string Name;
    public IContextValueReader ContextValueReader;

    public DataFilter(string name, IContextValueReader contextValueReader) {
      Name = name;
      ContextValueReader = contextValueReader;
    }
    
    public override string ToString() {
      return Name; 
    }

    public override int GetHashCode() {
      return Name.GetHashCode(); 
    }

    public abstract Funclet GetFunclet(Type entityType);
    public abstract bool MatchEntity(IUserContext context, object entity);
  }//class

  /// <summary> Data filter detects if a given entity "belongs" to a data group identified by a value in UserContext. </summary>
  /// <remarks>For example, we can create a data filter that indicates if a certain record (document) is related to the department
  /// that user/employee works in. Based on this information, the authorization system might grant or deny access to the document.
  /// In this scenario, the Id of the department is placed into the user context. The context value reader is setup to return this value
  /// to the data filter. The data filter is configured to return true/false indicator for a given entity (record) instance, 
  /// using expressions that retrieve the associated department ID from an entity and match it with the ID in the user context.
  /// See an example of using the data filters in BooksModel sample assembly.
  /// </remarks>
  public class DataFilter<TEntityId> : DataFilter where TEntityId : IEquatable<TEntityId> {

    Dictionary<Type, Funclet> _funcs = new Dictionary<Type, Funclet>(); 

    public DataFilter(string name, IContextValueReader<TEntityId> contextValueReader) : base(name, contextValueReader) {
    }
    // Adds verification expression for an entity. (entity, contextValueId) -> true/false
    public void AddEntity<TEntity>(Expression<Func<TEntity, TEntityId, bool>> expression) {
      var func = new Funclet<TEntity, TEntityId, bool>(expression);
      _funcs.Add(typeof(TEntity), func);
    }

    // This method allows adding expressions in the form of target ID selector - expression that selects ID for an entity
    //    (entity) -> targetId
    // we transform it automatically into a verification expression, using extra == operator:
    //    (entity, targetId) -> true/false
    public void AddEntity<TEntity>(Expression<Func<TEntity, TEntityId>> documentIdSelector) {
      // Convert groupIdSelector into expression used by RecordFilter: (entity, groupId) => bool
      var prmEntity = Expression.Parameter(typeof(TEntity), "entity");
      var prmContextValueId = Expression.Parameter(typeof(TEntityId), "contextValueId");
      var invokeExpr = Expression.Invoke(documentIdSelector, prmEntity);
      var eqExpr = Expression.Equal(invokeExpr, prmContextValueId);
      var filterExpr = Expression.Lambda<Func<TEntity, TEntityId, bool>>(eqExpr, prmEntity, prmContextValueId);
      AddEntity<TEntity>(filterExpr); 
    }
    public override Funclet GetFunclet(Type entityType) {
      Funclet result;
      _funcs.TryGetValue(entityType, out result);
      return result; 
    }

    public override bool MatchEntity(IUserContext context, object entity) {
      Util.Check(entity != null, "Entity may not be null.");
      var type = EntityHelper.GetEntityType(entity); 
      Funclet func = GetFunclet(type); 
      if (func == null)
        return false;
      var docId = ContextValueReader.GetValue(context);
      return (bool)func.Evaluate(entity, docId);
    }

  }




}
