﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Reflection;
using System.Threading;
using System.Globalization;
using System.Linq.Expressions;
using System.Data;

using Vita.Common;
using Vita.Common.Graphs;
using Vita.Entities.Runtime;
using Vita.Entities.Model.Construction;

namespace Vita.Entities.Model {

  // Holds CRUD commands for a table
  public class EntityCrudCommands {
    public EntityCommand SelectAll;
    public EntityCommand SelectAllPaged;
    public EntityCommand SelectByPrimaryKey;
    public EntityCommand Insert;
    public EntityCommand Delete;
    public EntityCommand Update;
  }//class


  public delegate string RecordDisplayMethod(EntityRecord record);

  [Flags]
  public enum EntityFlags {
    None = 0,
    //============ Flags set by attributes ==========================
    //
    SmallTable = 1,
    // Marks entity as internal, accessed by system processes only, and access is granted by default
    IsSystem = 1 << 1,
    // Marks entities that are created on the fly when update is performed (ex: log records), and should be eliminated after transaction aborts or commits  
    Dynamic = 1 << 2,

    // Entities are cached
    Cached = 1 << 3,

    // =================== Flags computed at runtime (setup time) ========================
    //Self-referencing entities are not handled by SCC algorithm, so we track them with this flag
    SelfReferencing = 1 << 8, 
    // true if referenced by other entities
    IsReferenced = 1 << 9, 
    //if true, you can only select/insert/delete. Typical case is link table, with PK consisting of entity key references
    NoUpdate = 1 << 10,
    // Set by Scc algorithm, identifies that entity's SCC group has more than one entity
    TopologicalGroupNonTrivial = 1 << 11,
    // Indicates that there's no clustered index on the table. By default, system makes PK clustered - this flag suppresses this
    // Primary use - for code generation from database, to make entity Model exactly match the database model in cases when 
    // a table has no clustered index (bad practice anyway)
    NoClusteredIndex = 1 << 12, 
    // set if there's a RowVersion member
    HasRowVersion = 1 << 13,
    // Some 'strange' tables (like Production.Document in AdventureWorks database) have strange PKs (heirarchyid type). LINQ pukes on attempt to map such a table
    LinqDisabled = 1 << 14,

  }

  public enum CachingType {
    None,
    // User, //sparse cache attached to UserContext - for the future
    Sparse, // global sparse cache
    FullSet, // global full-set cache
  }


  public delegate EntityBase EntityCreatorMethod(EntityRecord record);

  //contains information about generated class
  public class EntityClassInfo {
    public Type Type;
    public EntityCreatorMethod CreateInstance;
  }

  // We use entity Model objects as keys in DbModel's tables(dictionaries) of all db objects - so they all derive from HashedObject.
  // For example, to find a table for an entity (find DbTableInfo for an EntityInfo). 
  // Note that there may be multiple DbModels for a single entity Model, so we may have a different DB object in each DbModel for an entity object.

  public class EntityInfo : HashedObject {
    public readonly EntityModule Module;
    public string Name;
    public string FullName { get; private set; }
    public string Description;
    public string TableName; // explicitly set table name in Entity attribute

    public readonly Type EntityType;
    public readonly HashSet<Type> ReplacesTypes = new HashSet<Type>(); //in case of entity replacement, the types that this type is replacing
    public EntityInfo ReplacedBy; 
    public readonly List<EntityMemberInfo> Members = new List<EntityMemberInfo>();
    public readonly List<EntityKeyInfo> Keys = new List<EntityKeyInfo>(); 
    public EntityKeyInfo PrimaryKey;
    public Dictionary<string, EntityMemberInfo> MembersByName = new Dictionary<string, EntityMemberInfo>(StringComparer.InvariantCultureIgnoreCase);
    public IList<Type> CompanionTypes = new List<Type>(); // Companion classes or interfaces - derived from Entity, 
                                                          // and specifying additional attributes
    public List<Attribute> Attributes = new List<Attribute>(); //attributes on entity itself and on companion types
    public List<EntityMemberInfo> RefMembers = new List<EntityMemberInfo>();
    public List<EntityMemberInfo> IncomingReferences = new List<EntityMemberInfo>(); // members that reference this entity
    public EntityMemberInfo OwnerMember; //A reference member marked with Owner attribute

    public int PersistentValuesCount; // Size of ValuesOriginal/ValuesModified arrays in EntityRecord
    public int TransientValuesCount;    // Size of ValuesTransient array in EntityRecord

    public IList<PropertyGroup> PropertyGroups = new List<PropertyGroup>();
    public RecordDisplayMethod DisplayMethod;

    public EntityFlags Flags;
    public CachingType CachingType;

    public EntityEvents Events = new EntityEvents();
    public EntitySaveEvents SaveEvents = new EntitySaveEvents();
    public object[] InitialColumnValues;


    //Topological sorting index, assigned by SCC (Strongly-Connected Components) algorithm
    public int TopologicalIndex;
    internal Vertex SccVertex; //used in SCC algorithm

    //Entity classes information
    public EntityClassInfo ClassInfo; 
    //Paging mode
    public PagingMode PagingMode;

    public OrderByEntryInfo[] DefaultOrderBy;

    public string[] OldNames; //From OldNames attribute 

    //CRUD
    public EntityCrudCommands CrudCommands {get; internal set;}

    public EntityArea Area { get { return Module.Area; } }

    public EntityInfo(EntityModule module, Type entityType) { 
      Module = module;
      EntityType = entityType; 
      Name = entityType.Name;
      if (Name.StartsWith("I"))
        Name = Name.Substring(1); 
      var ns = Area.EntityNamespace;
      FullName = string.IsNullOrWhiteSpace(ns) ? Name : ns + "." + Name;
      DisplayMethod = DefaultDisplayRecord;
    }

    private string DefaultDisplayRecord(EntityRecord record) {
      return record.ToString(); 
    }

    public override string ToString() {
      return EntityType.ToString();
    }
    public EntityMemberInfo GetMember(string name, bool throwIfNotFound = false) {
      EntityMemberInfo member;
      if(MembersByName.TryGetValue(name, out member)) 
        return member;
      if (throwIfNotFound)
        throw new Exception(string.Format("Member {0} not found in entity {1}.", name, EntityType.Name));
      return null; 
    }

    public EntityMemberInfo FindMemberOrColumn(string name) {
      foreach (var member in Members)
        if (member.MemberName == name || string.Equals(member.ColumnName, name, StringComparison.InvariantCultureIgnoreCase))
          return member; 
      return null;
    }

    public EntityMemberInfo GetMember<TEntity>(Expression<Func<TEntity, object>> propertySelector) {
      var propName = ReflectionHelper.GetSelectedProperty(propertySelector);
      return GetMember(propName, throwIfNotFound: true);
    }

    public T GetAttribute<T>() where T : Attribute {
      return (T)Attributes.Find(a => a is T);
    }
    public bool HasAttribute<AttributeType>() where AttributeType : Attribute {
      var attr = GetAttribute<AttributeType>();
      return attr != null;
    }

    public void AddMember(EntityMemberInfo member) {
      member.Index = this.Members.Count;
      Members.Add(member);
      MembersByName.Add(member.MemberName, member);
    }

    public PropertyGroup GetPropertyGroup(string name, bool create = false) {
      if (name.StartsWith("@"))
        name = name.Substring(1); //cutoff @
      var g = PropertyGroups.FirstOrDefault(gr => gr.Name == name);
      if (g == null && create) {
        g = new PropertyGroup(name);
        PropertyGroups.Add(g);
      }
      return g; 
    }

    // Implements comparison function using DefaultOrderBy segments; Used in sorting of cached entities in memory
    public int DefaultCompare(EntityRecord x, EntityRecord y) {
      return CompareStartingAt(0, x, y); 
    }
    public int CompareStartingAt(int orderByIndex, EntityRecord x, EntityRecord y) {
      var orderBy = DefaultOrderBy[orderByIndex];
      var result = orderBy.CompareValues(x, y);
      if (result != 0 || orderByIndex == DefaultOrderBy.Length - 1)
        return result; //if not equal and no more order by entries, return result
      // otherwise, look at next comparison
      return CompareStartingAt(orderByIndex + 1, x, y);
    }
  }//EntityInfo

  //Property groups are defined in PropertyGroup attribute
  public class PropertyGroup {
    public string Name;
    public List<EntityMemberInfo> Members = new List<EntityMemberInfo>();
    public PropertyGroup(string name) {
      Name = name; 
    }
  }

  [Flags]
  public enum EntityMemberFlags {
    None = 0, 

    Nullable = 0x01,
    AutoValue = 1 << 1,
    DbGenererated = 1 << 2,

    PrimaryKey = 1 << 4,
    ForeignKey = 1 << 5,
    ClusteredIndex = 1 << 6,
    NaturalKey = 1 << 7, //not used yet

    IsOwner = 1 << 8,
    ReplaceDefaultWithNull = 1 << 9, //for nullable value types (NOTE: not types like double?; but declared "double" and marked with Nullable attribute
    Computed = 1 << 10, //value is automatically computed
    CascadeDelete = 1 << 11,
    Secret = 1 << 12,  //Write only, no selects
    Utc = 1 << 13,
    IsMemo = 1 << 14, 

    ExcludeFromUpdate = 1 << 16, //If set, the field is not included in CRUD update statement
    RowVersion = 1 << 17,
  }
  
  public enum MemberKind {
    //Persistent members 
    Column,        // Regular value-type column
    // Non persistent (transient) members; values are stored in EntityRecord.TransientValues
    Transient,    // transient, non-persistent value; a 'convenience' extra property that is not persisted
    EntityRef,    //reference to other entity
    EntityList,   //List of child entities
  }


  //Returns true if objects are equal
  public delegate bool AreValuesEqualMethod(object x, object y);
  //Handlers for get/set value methods
  public delegate object GetValueHandler(EntityRecord record, EntityMemberInfo member);
  public delegate void SetValueHandler(EntityRecord record, EntityMemberInfo member, object value);
  public delegate string ValueToStringMethod(EntityMemberInfo member, object value);
  public delegate object ValueFromStringMethod(EntityMemberInfo member, string value);

  public class EntityMemberInfo :  HashedObject {
    public EntityMemberFlags Flags;
    public EntityInfo Entity;
    public string MemberName;
    public string Description; 
    public PropertyInfo Property;
    public PropertyInfo ClassProperty;
    public Type DataType;
    public int Size;
    public byte Precision;
    public byte Scale; 

    public MemberKind Kind; 
    public int Index; //index in Entity.Members list
    public int ValueIndex; // value index in either OriginalValues/ModifiedValues or MemoryOnlyValues
    public AutoType AutoValueType;
    public List<Attribute> Attributes = new List<Attribute>(); //attributes on property and on the same properties in buddy types

    //Get/set value handlers
    public GetValueHandler GetValueRef;
    public SetValueHandler SetValueRef;
    // Value converter for serialization/deserialization and for URLs
    public ValueToStringMethod ValueToStringRef;
    public ValueFromStringMethod ValueFromStringRef;
    // Comparison method
    public AreValuesEqualMethod AreValuesEqual; 

    //From ColumnAttribute
    public string ColumnName; //assigned by Column attribute, to specify explicit non-default name for table column
    public string ColumnDefault;
    // DbType specified explicitly in DbType attribute
    public DbType? ExplicitDbType = null; 

    //Specified by OldNames attribute, to handle schema change when renaming properties/columns
    public string[] OldNames; 

    //Meta-data for various kinds
    //For member Kind = EntityRef: 
    public EntityReferenceInfo ReferenceInfo;
    //For member Kind = ForeignKey 
    public EntityMemberInfo ForeignKeyOwner; //Entity ref member that owns this fk member
    //For member Kind = EntityList
    public ChildEntityListInfo ChildListInfo;

    // initial, default value
    public object DefaultValue;
    //Value returned when access denied
    public object DeniedValue;

    public EntityMemberInfo[] DependentMembers; //array of (computed) members that depend on this member


    public EntityMemberInfo(EntityInfo entity, MemberKind kind, PropertyInfo property) :
          this(entity, kind, property.Name, property.PropertyType){
      Property = property; 
    }

    public EntityMemberInfo(EntityInfo entity, MemberKind kind, string name, Type dataType) {
      Entity = entity;
      Kind = kind;
      MemberName = ColumnName = name;
      DataType = dataType;
      Entity.AddMember(this);
      // actions by kind
      switch (Kind) {
        case MemberKind.EntityRef: 
          Entity.RefMembers.Add(this); 
          break;
      }
      //Set to nullable if it is Nullable generic
      if (DataType.IsNullableValueType())
        Flags |= EntityMemberFlags.Nullable;
      if (DataType == typeof(decimal) || DataType == typeof(decimal?)) {
        this.Precision = 18; //defaults
        this.Scale = 4;
      }
      this.AreValuesEqual = MemberValueGettersSetters.AreObjectsEqual;
      //Assign default get/set handlers and to-from string converters, and comparer method
      MemberValueGettersSetters.AssignDefaultGetSetHandlers(this);
      StringConverters.AssignStringConverters(this);
    }

    public override string ToString() {
      return Util.SafeFormat("{0}.{1}({2})", Entity.Name, MemberName, DataType); 
    }

    public T GetAttribute<T>() where T : Attribute {
      return (T)Attributes.Find(a => a is T);
    }
    public bool HasAttribute<T>() where T : Attribute {
      var attr = GetAttribute<T>();
      return attr != null; 
    }

    internal void AddAttributes(object[] attributes) {
      //We can't use AddRange because we have object[] array
      foreach (Attribute attr in attributes)
        Attributes.Add(attr); 
    }

  }//class

  [Flags]
  public enum KeyType {
    PrimaryKey = 1,
    ForeignKey = 1 << 1,
    Index = 1 << 2,

    Unique = 1 << 4,
    Clustered = 1 << 5,
    Auto = 1 << 6, // auto-created by db server; ex: Index on auto-increment column created by MySQL 

    UniqueIndex = Unique | Index,
    ClusteredIndex = Clustered | Index,
    UniqueClusteredIndex = Unique | Clustered | Index,
    ClusteredPrimaryKey = Clustered | PrimaryKey,
  }


  public class EntityKeyInfo : HashedObject {
    public string Name; 
    public EntityInfo Entity;
    public KeyType KeyType;
    //Original members specified in attributes; may contain members that are entity references
    public List<EntityMemberInfo> Members = new List<EntityMemberInfo>();
    public HashSet<EntityMemberInfo> DescendingMembers = new HashSet<EntityMemberInfo>();    
    //Often the same as Members, except entity reference members had been "expanded" - replaced with value members that constitute the foreign key
    public List<EntityMemberInfo> ExpandedMembers = new List<EntityMemberInfo>();
    internal bool IsExpanded; //used in member expansion algorithm
    public bool HasIdentityMember;
    public EntityCommand SelectByKeyCommand;
    public EntityCommand SelectTopOneByKeyCommand;
    public EntityMemberInfo OwnerMember; //for FK, the ref member
    public Delegate CacheSelectMethod; // compiled method to select in cache

    public EntityKeyInfo(string keyName, KeyType keyType, EntityInfo entity, EntityMemberInfo ownerMember = null) {
      Name = keyName;
      KeyType = keyType;
      Entity = entity;
      OwnerMember = ownerMember;
      entity.Keys.Add(this);
    }
    
    public override string ToString() {
      return Util.SafeFormat("{0}:{1}({2})", Entity.FullName, KeyType, string.Join(",", Members.Select(m=>m.MemberName)));
      
    }

    public string GetMemberNames(string separator = "", bool removeUnderscore = false) {
      string result;
      if (ExpandedMembers.Count == 1)
        result = ExpandedMembers[0].MemberName; 
      else 
        result = string.Join(separator, ExpandedMembers.Select(m => m.MemberName));
      if (removeUnderscore)
        result = result.Replace("_", string.Empty);
      return result; 
    }

    public bool IsDescending(EntityMemberInfo member) {
      var owner = member.ForeignKeyOwner;
      if (owner != null)
        return DescendingMembers.Contains(owner);
      else
        return DescendingMembers.Contains(member); 
    }
  }//class

  public class EntityReferenceInfo :  HashedObject {
    public EntityMemberInfo FromMember;
    public EntityKeyInfo FromKey; 
    public EntityKeyInfo ToKey;
    public string ForeignKeyColumns;

    public EntityReferenceInfo(EntityMemberInfo fromMember, EntityKeyInfo fromKey, EntityKeyInfo toKey) {
      FromMember = fromMember;
      FromKey = fromKey;
      ToKey = toKey;
      ToKey.Entity.Flags |= EntityFlags.IsReferenced;
      var fromEnt = fromMember.Entity;
      fromEnt.RefMembers.Add(fromMember); 
      ToKey.Entity.IncomingReferences.Add(fromMember);
      if (fromEnt == ToKey.Entity)
        fromEnt.Flags |= EntityFlags.SelfReferencing;
    }
  }

  public enum EntityRelationType {
    ManyToOne,
    ManyToMany
  }

  public class ChildEntityListInfo :  HashedObject {
    public EntityRelationType RelationType;
    public EntityInfo TargetEntity; //entity in the list
    public EntityInfo LinkEntity; //many-to-many only, link entity
    //One-to-many: reference member in child entity that references back to "this" entity (that owns the entity list)
    //Many-to-many: reference member in link entity that references back to "this" entity
    public EntityMemberInfo ParentRefMember;
    //Many-to-many: a member of link entity that references the "other" entity 
    public EntityMemberInfo OtherEntityRefMember;
    public EntityCommand SelectManyToManyCommand; //for many2many - selects target entities
    public OrderedPropertyRef[] OrderBy;
    public EntityMemberInfo PersistentOrderMember; //set if owner property has [PersistOrderIn] attribute.

    //command selecting entities; for many2many, selects link entities
    public EntityCommand SelectCommand {
      get { return ParentRefMember.ReferenceInfo.FromKey.SelectByKeyCommand; }
    } 
  }

  //Used in entity replacement setup
  public class EntityReplacementInfo {
    public Type ReplacedType;
    public Type NewType;
  }

  // Helper container for parsing ordered column lists in attributes: Index("LastName:desc,FirstName:asc"); 
  public class OrderedPropertyRef {
    public readonly string Name;
    public readonly string Order = "asc";
    public OrderedPropertyRef(string spec) {
      var segms = Util.SafeSplit(spec, ':');
      Name = segms[0];
      if (segms.Length > 1)
        Order = (segms[1]??string.Empty).ToLowerInvariant();
    }

    public static OrderedPropertyRef[] CreateFromNames(string names) {
      return Util.SafeSplit(names).Select(n => new OrderedPropertyRef(n)).ToArray();
    }


  }


  public class OrderByEntryInfo {
    public static Comparer DefaultComparer = new Comparer(CultureInfo.InvariantCulture); 

    public EntityMemberInfo Member;
    public bool Descending;
    bool _isString;
    public OrderByEntryInfo(EntityMemberInfo member, bool descending) {
      Member = member;
      Descending = descending; 
      _isString = Member.DataType == typeof(string); 
    }

    // Note: this Compare method is used for sorting entities in cache. We need to match it with the way SQL sort works -
    // so that entities in cache appear in the same order as if they came from SQL SELECT.
    // By default SQL Server sorts strings ignoring case. That's why we use 'ignoreCase: true' for string comparison here.
    public int CompareValues(EntityRecord x, EntityRecord y) {
      var xv = x.GetValue(Member);
      var yv = y.GetValue(Member); 
      int result = _isString ? 
                       string.Compare((string)xv, (string)yv, ignoreCase: true) : 
                       DefaultComparer.Compare(xv, yv);
      if (Descending)
        result = - result; 
      return result; 
    }

    public override string ToString() {
      return Member.ColumnName + (Descending ? " DESC" : string.Empty);
    }

  }//class

  public class EntityRecordComparer : IComparer<EntityRecord> {
    public OrderByEntryInfo[] OrderBy;

    public EntityRecordComparer(OrderByEntryInfo[] orderBy) {
      OrderBy = orderBy;
    }

    #region IComparer<EntityRecord> Members
    public int Compare(EntityRecord x, EntityRecord y) {
      for(int i = 0; i< OrderBy.Length; i++) {
        var comp = OrderBy[i].CompareValues(x, y);
        if (comp != 0)
          return comp; 
      }
      return 0; 
    }
    #endregion
  }



  public static class EntityEnumExtensions {
    public static bool IsSet(this LoadFlags flags, LoadFlags flag) {
      return (flags & flag) != 0;
    }
    public static bool IsSet(this EntityFlags flags, EntityFlags flag) {
      return (flags & flag) != 0;
    }
    public static bool IsSet(this EntityMemberFlags flags, EntityMemberFlags flag) {
      return (flags & flag) != 0;
    }
    public static bool IsSet(this KeyType flags, KeyType flag) {
      return (flags & flag) != 0;
    }
    public static bool IsSet(this EntityCommandFlags flags, EntityCommandFlags flag) {
      return (flags & flag) != 0;
    }

  }//class


}//namespace
