﻿using System;
using System.Collections.Generic;
using System.Linq;

using Vita.Common;

namespace Vita.Entities.Model {

  public partial class EntityModelInfo  {
    public EntityModel Model;
    public Dictionary<string, Type> EnumTypes = new Dictionary<string, Type>(StringComparer.InvariantCultureIgnoreCase);

    internal string ClassesNamespace;

    private List<EntityCommand> _commands = new List<EntityCommand>();
    private IList<EntityInfo> _entities = new List<EntityInfo>();
    private Dictionary<Type, EntityInfo> _entitiesByType = new Dictionary<Type, EntityInfo>(); //interfaceType -> EntityInfo
    private Dictionary<string, EntityInfo> _entitiesByName = new Dictionary<string, EntityInfo>(StringComparer.InvariantCultureIgnoreCase); //name -> EntityInfo

    public ICollection<EntityInfo> Entities {
      get { return _entities; }
    }
    public IEnumerable<Type> GetAllEntityTypes() {
      return _entities.Select(e => e.EntityType);
    }

    public EntityModelInfo(EntityModel model) {
      Model = model;
    }


    // Remember that entities might be replaced when aggregating modules
    public void RegisterEntity(EntityInfo entityInfo) {
      //Only for final entities
      if (entityInfo.ReplacedBy == null) {
        //Register by full name and add to list of final entities
        _entitiesByName[entityInfo.FullName] = entityInfo;
        _entities.Add(entityInfo);
      }
      //For type->entity table, register final entity for original type 
      var finalEntityInfo = entityInfo.ReplacedBy ?? entityInfo;
      _entitiesByType[entityInfo.EntityType] = finalEntityInfo;
    }

    public EntityInfo GetEntityInfo(Type interfaceType, bool throwIfNotFound = false) {
      EntityInfo entityInfo;
      if(_entitiesByType.TryGetValue(interfaceType, out entityInfo)) return entityInfo;
      if (throwIfNotFound)
        Util.Throw("Type {0} is not registered as entity.", interfaceType);
      return null; 
    }

    public EntityInfo GetEntityInfo(string fullName, bool throwIfNotFound = false) {
      EntityInfo entityInfo;
      if (_entitiesByName.TryGetValue(fullName, out entityInfo)) return entityInfo;
      if (throwIfNotFound)
        Util.Throw("Type {0} is not registered as entity.", fullName);
      return null;
    }

    public Type GetEntityClass(Type interfaceType) {
      var entInfo = GetEntityInfo(interfaceType, true);
      return entInfo.ClassInfo.Type;
    }

    public void AddEnumType(Type type) {
      EnumTypes.Add(type.FullName.ToLowerInvariant(), type);
    }

    public Type FindEnumType(string fullName) {
      Type result;
      if (EnumTypes.TryGetValue(fullName, out result)) return result; 
      //try to find by short type name
      foreach (var de in EnumTypes)
        if (string.Compare(de.Value.Name, fullName, ignoreCase: true) == 0) return de.Value;
      //Otherwise return null
      return null;
    }

    #region Commands access
    public ICollection<EntityCommand> Commands {
      get {return _commands;}
    }
    
    public ICollection<EntityCommand> GetCrudCommands() {
      return _commands.Where(c => !c.Flags.IsSet(EntityCommandFlags.IsCustom)).ToList(); 
    }
    public ICollection<EntityCommand> GetCustomCommands() {
      return _commands.Where(c => c.Flags.IsSet(EntityCommandFlags.IsCustom)).ToList();
    }
    public void AddCommand(EntityCommand command) {
      _commands.Add(command); 
    }
    #endregion

  }//class

}//namespace
