﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Vita.Common; 
using Vita.Entities.Model;

namespace Vita.Entities.Runtime {


  public class EntityKey : IEquatable<EntityKey>  {
    public readonly EntityKeyInfo KeyInfo; 
    public readonly object[] Values;
    int _hashCode;

    public EntityKey(EntityKeyInfo keyInfo, EntityRecord record) {
      KeyInfo = keyInfo; 
      Values = new object[KeyInfo.ExpandedMembers.Count];
      for(int i = 0; i < Values.Length; i++)
        Values[i] = record.GetValueDirect(KeyInfo.ExpandedMembers[i]);
      ComputeHashCode(); 
    }

    public EntityKey(EntityKeyInfo key, params object[] values) {
      Util.Check(key.ExpandedMembers.Count == values.Length,
        "EntityKeyValue constructor: values count does not match key members count. Key={0}", key);
      KeyInfo = key; 
      Values = values;
      ComputeHashCode(); 
    }//constructor

    private void ComputeHashCode() {
      _hashCode = KeyInfo.GetHashCode() << 4;
      for (int i = 0; i < Values.Length; i++) {
        var v = Values[i];
        if (v != null)
          _hashCode = (_hashCode << 1) + v.GetHashCode(); 
      }
    }

    public override int GetHashCode() {
      return _hashCode;
    }

    public override bool Equals(object obj) {
      var keyObj = obj as EntityKey;
      if (keyObj == null)
        return false;
      return Equals(keyObj); 
    }

    public bool Equals(EntityKey other) {
      if (other.KeyInfo != this.KeyInfo)
        return false; 
      for (int i = 0; i < KeyInfo.ExpandedMembers.Count; i++) {
        var member = KeyInfo.ExpandedMembers[i];
        var eq = member.AreValuesEqual(Values[i], other.Values[i]);
        if (!eq)
          return false;
      }
      return true;
    }

    public override string ToString() {
      if (this.KeyInfo.KeyType.IsSet(KeyType.PrimaryKey))
        return Util.SafeFormat("{0}/{1}", this.KeyInfo.Entity.Name, ValuesToString()); 
      else if (this.KeyInfo.OwnerMember != null && this.KeyInfo.KeyType == KeyType.ForeignKey)
        return Util.SafeFormat("{0}.{1}/{2}/{3}", this.KeyInfo.Entity.Name, this.KeyInfo.OwnerMember.MemberName, this.KeyInfo.KeyType, ValuesToString());
      else 
        return Util.SafeFormat("{0}/{1}/{2}", this.KeyInfo.Entity.Name, this.KeyInfo.KeyType, ValuesToString());
    }

    public string ValuesToString() {
      //Fast path
      if (Values.Length == 1) {
        var m0 = KeyInfo.ExpandedMembers[0];
        return m0.ValueToStringRef(m0, Values[0]);
      }
      //full path
      var sv = new string[Values.Length];
      for (int i = 0; i < Values.Length; i++) {
        var m = KeyInfo.ExpandedMembers[i];
        sv[i] = m.ValueToStringRef(m, Values[i]);
      }
      return string.Join(";", sv);
    }

    public static EntityKey KeyFromString(EntityKeyInfo key, string valueString) {
      if (key.ExpandedMembers.Count == 1) { //shortcut
        var m0 = key.ExpandedMembers[0];
        var v = m0.ValueFromStringRef(m0, valueString);
        var kv = new EntityKey(key, new object[] { v });
        return kv;
      }
      //complex keys
      var sValues = valueString.Split(';');
      var values = new object[key.ExpandedMembers.Count];
      for (int i = 0; i < values.Length; i++) {
        var m = key.ExpandedMembers[i];
        values[i] = m.ValueFromStringRef(m, sValues[i]);
      }
      return new EntityKey(key, values);
    }

    public bool ValuesEmpty() {
      for (int i = 0; i < KeyInfo.ExpandedMembers.Count; i++) {
        var member = KeyInfo.ExpandedMembers[i];
        var v = Values[i];
        if (v != null && v != DBNull.Value)
          return false; 
      }
      return true; 
    }

  }//class

}//namespace
