﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.Concurrent;
using System.Linq;
using System.Linq.Expressions;
using System.Reflection;

using Vita.Common;
using Vita.Data;
using Vita.Entities.Model;
using Vita.Services;
using Vita.Logging;

namespace Vita.Entities.Runtime {


  /// <summary> Represents a live connection to the database with tracking of loaded and changed/added/deleted entities. </summary>
  /// <remarks>This class provides methods for data access: reading, updating, deleting entitites. </remarks>
  public partial class EntitySession : IEntitySession, IQueryProvider {
    public readonly EntityStore EntityStore;
    public readonly bool IsReadOnly;
    public readonly EntityModel Model;
    public readonly EntitySessionSettings Settings;
    public readonly UserContext UserContext; 

    public readonly EntityRecordWeakRefTable RecordsLoaded;
    public readonly IList<EntityRecord> RecordsChanged;
    public readonly HashSet<EntityInfo> EntityInfosForRecordsChanged;
    public readonly IList<IPropertyBoundList> ListsChanged;
    public readonly IList<ScheduledEntityCommand> ScheduledCommands = new List<ScheduledEntityCommand>();
    public EntityValidator Validator { get; private set; }
    public DateTime? LastTransactionDateTime { get; private set; }

    internal EntityQuery LastQuery; //last executed LINQ query; can be retrieved using EntityHelper.GetLastQuery(this session)
    private ITimeService _timeService;
    private IAppLogService _logService; 

    #region constructor
    public EntitySession(IEntityStore entityStore, UserContext userContext = null, EntitySessionSettings settings = null) {
      EntityStore = (EntityStore) entityStore;
      Model = EntityStore.Model;
      _logService = EntityStore.GetService<IAppLogService>(); 
      this.UserContext = userContext ?? new UserContext("Anonymous", null, null);
      this.Settings = settings ?? EntitySessionSettings.Default; 
      // Multithreaded sessions must be readonly
      var isConcurrent = Settings.Options.IsSet(EntitySessionOptions.ConcurrentReadonly);
      IsReadOnly = Settings.Options.IsSet(EntitySessionOptions.ReadOnly) || isConcurrent;
      _timeService = entityStore.GetService<ITimeService>();
      Util.Check(_timeService != null, "ITimeService is not available - make sure it is registered at startup.");
      RecordsLoaded = new EntityRecordWeakRefTable(isConcurrent);
      // These two lists are not used in Readonly sessions
      if (!IsReadOnly) {
        RecordsChanged = new List<EntityRecord>();
        ListsChanged = new List<IPropertyBoundList>(); 
      }
      EntityInfosForRecordsChanged = new HashSet<EntityInfo>();
      this.EntityStore.Events.OnNewSession(this); 
    }
    #endregion

    public override string ToString() {
      return "EntitySession(" + UserContext.UserName + ")";
    }

    #region IEntitySession Methods ===========================================================================================

    IUserContext IEntitySession.UserContext {
      get { return UserContext; } 
    }

    public virtual void LogMessage(string message, params object[] args) {
      var entry = new MessageLogEntry(LogEntryType.Information, message, args);
      //create local log if it does not exist
      var localLog = this.UserContext.Log;
      if (localLog == null)
        localLog = this.UserContext.Log = new Logging.Log(flushMode: LogFlushMode.None); //no flushing/writing anywhere, keep it in memory
      localLog.AddEntry(entry);
      var appLog = _logService.Log;
      if (appLog != null)
        appLog.AddEntry(entry);
    }

    public virtual TEntity NewEntity<TEntity>() where TEntity: class {
      CheckNotReadonly();
      var entityInfo = GetEntityInfo(typeof(TEntity));
      Util.Check(entityInfo != null, "Entity {0} is not part of Entity Model.", typeof(TEntity));
      var newRec = NewRecord(entityInfo);
      return (TEntity) (object) newRec.EntityInstance;
    }

    public virtual TEntity GetEntity<TEntity>(object primaryKeyValue, LoadFlags flags = LoadFlags.Default) where TEntity : class {
      Util.Check(primaryKeyValue != null, "Session.GetEntity<{0}>(): primary key may not be null.", typeof(TEntity));
      var entityInfo = GetEntityInfo(typeof(TEntity));
      //Check if it is an entity key object; if not, it is a "value" (or values) of the key
      var pkType = primaryKeyValue.GetType();
      EntityKey pk;
      if (pkType == typeof(EntityKey)) {
        pk = (EntityKey)primaryKeyValue;
      } else { 
        //try array of objects; if not, it is a single value, turn it into array
        object[] pkValues = pkType == typeof(object[]) ? (object[]) primaryKeyValue : new object[] { primaryKeyValue };
        pk = new EntityKey(entityInfo.PrimaryKey, pkValues);
      }
      var rec = GetRecord(pk, flags);
      if (rec != null)
        return (TEntity)(object) rec.EntityInstance;
      return default(TEntity);
    }

    public virtual void DeleteEntity<TEntity>(TEntity entity) where TEntity : class {
      CheckNotReadonly();
      var record = EntityHelper.GetRecord(entity);
      DeleteRecord(record);
    }

    public virtual bool CanDeleteEntity<TEntity>(TEntity entity, out Type[] blockingEntities) where TEntity : class {
      var record = EntityHelper.GetRecord(entity);
      Util.Check(record != null, "CanDeleteEntity: entity parameter is not an Entity.");
      return CanDeleteRecord(record, out blockingEntities);
    }

    public virtual IList<TEntity> GetEntities<TEntity>(int skip = 0, int take = int.MaxValue) where TEntity : class {
      var paging = new Paging(skip, take);
      var entityInfo = GetEntityInfo(typeof(TEntity));
      Util.Check(entityInfo != null, "Entity {0} not registered with the Model.", typeof(TEntity));
      //If there is no paging, check SmallTable attribute on entity
      if (paging.IsNoPaging()) {
        Util.Check(entityInfo.Flags.IsSet(EntityFlags.SmallTable),
          "Attempt to call Session.GetEntities<{0}>() method without paging parameters. Entity is not marked with [SmallTable] attribute.",
          entityInfo.EntityType.Name); 
      }
      var records = GetRecords(entityInfo, paging);
      return ToEntities<TEntity>(records); 
    }

    public virtual IList<TEntity> GetChildEntities<TParentEntity, TEntity>(TParentEntity parent, string parentRefProperty = null) where TEntity : class {
      Util.Check(parent != null, "GetChildEntities<{0},{1}>: parent parameter may not be null.", typeof(TParentEntity), typeof(TEntity));
      var parentRec = EntityHelper.GetRecord(parent);
      Util.Check(parentRec != null, "GetChildEntities<{0},{1}>: parent parameter ({2}) is not an entity.", typeof(TParentEntity), typeof(TEntity), parent);
      var parentEntity = parentRec.EntityInfo;
      var childEntity = GetEntityInfo(typeof(TEntity));
      Util.Check(childEntity != null, "Entity {0} not registered with the entity model.", typeof(TEntity));
      EntityMemberInfo member;
      if (string.IsNullOrEmpty(parentRefProperty)) {
        var members = childEntity.Members.FindAll(m => m.DataType == typeof(TParentEntity));
        Util.Check(members.Count > 0, "Failed to select child entities. Reference to entity {0} not found on entity {1}.", typeof(TParentEntity), typeof(TEntity));
        Util.Check(members.Count < 2, "Failed to select child entities. More than one reference to entity {0} found on entity {1}. Explicitly specify property to use.", 
          typeof(TParentEntity), typeof(TEntity));
        member = members[0];
      } else {
        member = childEntity.GetMember(parentRefProperty);
        Util.Check(member != null, "Failed to select child entities. Reference to entity {0} not found on entity {1}.", typeof(TParentEntity), typeof(TEntity));
        var refOk = member.Kind == MemberKind.EntityRef && member.ReferenceInfo.ToKey.Entity == parentEntity; 
        Util.Check(refOk, "Failed to select child entities. Property {0} on entity {1} does not reference entity {2}.", 
          parentRefProperty, typeof(TEntity), typeof(TParentEntity));
      }
      //actually retrieve entities
      var records = GetChildRecords(parentRec, member);
      return ToEntities<TEntity>(records);
    }

    public virtual void ValidateChanges() {
      Validator = new EntityValidator(this);
      Validator.Validate(this.RecordsChanged);
      Validator.ThrowValidation();
    }

    public void CancelChanges() {
      CheckNotReadonly(); 
      foreach (var rec in RecordsChanged) {
        rec.EntityInfo.Events.OnChangesCanceled(rec);
        rec.CancelChanges();
      }
      RecordsChanged.Clear();
      EntityInfosForRecordsChanged.Clear();
      ListsChanged.Clear(); 
    }

    public virtual void SaveChanges() {
      CheckNotReadonly();
      LastTransactionDateTime = _timeService.Now; 
      //We should validate first, before firing Saving event
      if (!_validationDisabled)
        ValidateChanges();
      OnSaving();
      try {
        SubmitChanges();
        OnSaved();
        RecordsChanged.Clear();
        EntityInfosForRecordsChanged.Clear(); 
        ListsChanged.Clear();
      } catch (Exception ex) {
        OnSaveAborted();
        EntityStore.Events.OnError(this, ex);
        throw;
      }
    }//method

    protected virtual void SubmitChanges() {
      this.EntityStore.DataStore.SaveChanges(this);
    }

    public virtual IList<TEntity> ExecuteQuery<TEntity>(IEntityCommand command, params object[] args) where TEntity: class {
      CheckCommandArgsCount(command, args); 
      var records = this.ExecuteRecordQuery(command, Paging.NoPaging, args);
      return ToEntities<TEntity>(records);
    }
    public virtual object ExecuteNonQuery(IEntityCommand command, params object[] args) {
      CheckNotReadonly();
      CheckCommandArgsCount(command, args);
      return EntityStore.ExecuteNonQuery(this, command, args);
    }

    public virtual IQueryable<TEntity> EntitySet<TEntity>() where TEntity : class {
      var entInfo = GetEntityInfo(typeof(TEntity));
      Util.Check(entInfo != null, "Type {0} is not an entity.", typeof(TEntity));
      return new EntitySet<TEntity>(this);
    }
    
    public void ExecuteOnSave(IEntityCommand command, ExecuteOptions options, params object[] args) {
      CheckNotReadonly();
      ScheduledCommands.Add(new ScheduledEntityCommand((EntityCommand)command, options, args));
    } 
    #endregion 

    #region Record manipulation methods
    public virtual EntityRecord GetRecord(EntityKey primaryKey, LoadFlags flags = LoadFlags.Default) {
      var record = GetLoadedRecord(primaryKey);
      if (record != null) {
        if (record.Status == EntityStatus.Stub && flags.IsSet(LoadFlags.Load)) record.Reload(); 
        return record; 
      }
      // TODO: review returning Record stub in the context of SecureSession. 
      // We might have security hole here if user is not authorized to access the record  
      if (flags.IsSet(LoadFlags.Stub))
        return CreateStub(primaryKey);
      if (!flags.IsSet(LoadFlags.Load))
        return null; 
      //Otherwise, load it
      var entityInfo = primaryKey.KeyInfo.Entity;
      var recs = this.ExecuteRecordQuery(entityInfo.CrudCommands.SelectByPrimaryKey, Paging.JustOne, primaryKey.Values);
      record = recs.Count > 0 ? recs[0] : null;
      return record; 
    }

    public virtual IList<EntityRecord> GetRecords(EntityInfo entityInfo, Paging paging = null) {
      paging = paging ?? Paging.NoPaging;
      var cmd = paging.IsNoPaging() ? entityInfo.CrudCommands.SelectAll : entityInfo.CrudCommands.SelectAllPaged;
      return ExecuteRecordQuery(cmd, paging, null);
    }

    public virtual IList<EntityRecord> GetChildRecords(EntityRecord parent, EntityMemberInfo refMember) {
      Util.Check(refMember.Kind == MemberKind.EntityRef, 
        "Invalid refMember parameter ({0}) - should be an Entity reference member(property).", refMember.MemberName);
      if (parent.Status != EntityStatus.New)
        return this.ExecuteRecordQuery(refMember.ReferenceInfo.FromKey.SelectByKeyCommand, Paging.NoPaging, parent.PrimaryKey.Values);
      //For new parent, look in the changed records      
      var childEntInfo = refMember.Entity;
      var list = new List<EntityRecord>(); 
      var refTarget = parent.EntityInstance;
      //look in new records in this session
      foreach(var rec in this.RecordsChanged) {
        if (rec.EntityInfo != childEntInfo) continue;
        var refValue = parent.GetValue(refMember);
        if (refValue == refTarget)
          list.Add(rec);
      }
      return list; 
    }

    public virtual EntityRecord NewRecord(EntityInfo entityInfo) {
      var record = new EntityRecord(entityInfo, EntityStatus.New);
      record = Attach(record);
      EntityStore.EntityEvents.OnNew(record);
      record.EntityInfo.Events.OnNew(record);
      return record;
    }

    public void DeleteRecord(EntityInfo entity, EntityKey primaryKeyValue) {
      var rec = GetRecord(primaryKeyValue, LoadFlags.Stub);
      if (rec == null) return;
      DeleteRecord(rec);
    }

    public virtual bool CanDeleteRecord(EntityRecord record, out Type[] blockingEntities) {
      Util.Check(record != null, "CanDeleteRecord: record parameter may not be null.");
      blockingEntities = null; 
      var entInfo = record.EntityInfo;
      var blockingTypeSet = new HashSet<Type>();
      //check all external ref members
      foreach (var refMember in entInfo.IncomingReferences) {
        // if it is cascading delete, this member is not a problem
        if (refMember.Flags.IsSet(EntityMemberFlags.CascadeDelete)) continue; 
        var selectOneCommand = refMember.ReferenceInfo.FromKey.SelectTopOneByKeyCommand;
        var childRecs = ExecuteRecordQuery(selectOneCommand, Paging.NoPaging, record.PrimaryKey.Values);
        if (childRecs.Count > 0)
          blockingTypeSet.Add(refMember.Entity.EntityType);
      }
      if (blockingTypeSet.Count == 0)
        return true;
      blockingEntities = blockingTypeSet.ToArray();
      return false; 
    }

    public virtual void DeleteRecord(EntityRecord record) {
      Util.Check(record != null, "DeleteRecord: record parameter may not be null.");
      record.MarkForDelete();
      record.EntityInfo.Events.OnDeleting(record);
      EntityStore.EntityEvents.OnDeleting(record);
    }

    public virtual EntityRecord Attach(EntityRecord record) {
      if (record == null) return null; 
      if (record.IsAttached && record.Session == this) 
        return record; //already attached
      //Check that entity Model is the same
      if (record.EntityInfo.Area.Model != Model) {
        var thisModelEntInfo = EntityStore.Model.ModelInfo.GetEntityInfo(record.EntityInfo.FullName);
        Util.Check(thisModelEntInfo != null, "Failed to attach record to session - entity info is not part of the Model.");
        record.EntityInfo = thisModelEntInfo;
      }
      //Actually attach
      record.Session = this;
      record.IsAttached = true; 
      var attachedRecord = record; //might switch to the record already loaded in session
      //Add record to appropriate collection
      switch (record.Status) {
        case EntityStatus.Loading:
        case EntityStatus.Stub:
        case EntityStatus.Loaded:
          record.Status = EntityStatus.Loaded;
          attachedRecord = RecordsLoaded.Add(record); //might return existing record
          if (attachedRecord == record) { // it is brand new record just loaded
            record.EntityInfo.Events.OnLoaded(record);
            EntityStore.EntityEvents.OnLoaded(record);
          }
          break; 
        case EntityStatus.New:
          RecordsChanged.Add(record);
          EntityInfosForRecordsChanged.Add(record.EntityInfo);
          break;
        case EntityStatus.Deleting:
        case EntityStatus.Modified:
          attachedRecord = RecordsLoaded.Add(record);
          RecordsChanged.Add(attachedRecord);
          EntityInfosForRecordsChanged.Add(record.EntityInfo);
          break; 
      }
      //Just in case, if 'record' is freshly loaded from database, and we have already the same 'attachedRecord' in session, 
      // refresh the column values
      if (attachedRecord != record) {
        attachedRecord.CopyOriginalValues(record);
        attachedRecord.ClearEntityRefValues(); 
      }
      return attachedRecord;
    }

    public void NotifyRecordStatusChanged(EntityRecord record, EntityStatus oldStatus) {
      var newStatus = record.Status; 
      if (newStatus == oldStatus)
        return; 
      //check new status
      switch (newStatus) {
        case EntityStatus.Loaded:
          break;
        case EntityStatus.New:
        case EntityStatus.Modified:
        case EntityStatus.Deleting:
          CheckNotReadonly(); 
          RecordsChanged.Add(record);
          EntityInfosForRecordsChanged.Add(record.EntityInfo);
          break; 
        case EntityStatus.Fantom:
          this.RecordsLoaded.TryRemove(record.PrimaryKey);
          break; 
      }
    }

    public virtual IList<EntityRecord> ExecuteRecordQuery(IEntityCommand command, Paging paging, params object[] args) {
      return EntityStore.ExecuteQuery(this, command, paging, args);
    }

    #endregion

    #region Dynamic query execution: ExecuteDynamicQuery
    internal object ExecuteDynamicQuery(EntityQuery query) {
      this.LastQuery = query; 
      return EntityStore.DataStore.ExecuteDynamicQuery(this, query);
    }
    #endregion

    #region Record Get/Set Value
    // Will be overridden SecureSession
    public virtual object RecordGetMemberValue(EntityRecord record, EntityMemberInfo member) {
      return member.GetValueRef(record, member);
    }
    public virtual void RecordSetMemberValue(EntityRecord record, EntityMemberInfo member, object value) {
      member.SetValueRef(record, member, value);
    }

    #endregion

    #region OnSave methods
    // We might have new entities added on the fly, as we fire events. For example, ChangeTracking hooks to EntityStore and it adds new IChangeTrackEntries.
    // We need to make sure that for all added records we fire the Entity's OnSaving event. (All Auto attributes are processed in this event). 
    // That's the reason for tracking 'done-Counts' and calling the invoke loop twice
    protected virtual void OnSaving() {
      int startListIndex = 0;
      int startRecordIndex = 0;
      InvokeOnSaving(ref startListIndex, ref startRecordIndex);
      EntityStore.Events.OnSavingChanges(this);
      InvokeOnSaving(ref startListIndex, ref startRecordIndex);
    }

    private void InvokeOnSaving(ref int startListIndex, ref int startRecordIndex) {
      //Notify all changed lists - these may add/delete records (many-to-many link records)
      for (int i = startListIndex; i < ListsChanged.Count; i++)
        ListsChanged[i].Notify(BoundListEventType.SavingChanges, this);
      startListIndex = ListsChanged.Count;
      //Saving event handlers may add/remove changed records, so we need to use 'for-i' loop here
      for (int i = startRecordIndex; i < this.RecordsChanged.Count; i++) {
        var rec = RecordsChanged[i];
        rec.EntityInfo.SaveEvents.OnSavingChanges(rec);
      }
      startRecordIndex = RecordsChanged.Count; 
    }

    protected void OnSaved() {
      //Notify all changed lists
      foreach (var list in ListsChanged)
        list.Notify(BoundListEventType.SavedChanges, this);
      //Process records
      for (int i = 0; i < this.RecordsChanged.Count; i++) {
        var rec = RecordsChanged[i];
        rec.CommitChanges();
        // Dynamic records are created on the fly when app submits changes (ex: trans logs). After commit, we discard them from session
        // RecordsChanged will be cleared anyway, so all we need to do is remove it from loaded records
        if (rec.EntityInfo.Flags.IsSet(EntityFlags.Dynamic)) 
          RecordsLoaded.TryRemove(rec.PrimaryKey);
        rec.EntityInfo.SaveEvents.OnSavedChanges(rec);
      }
      EntityStore.Events.OnSavedChanges(this);
      Validator = null; 
    }

    protected void OnSaveAborted() {
      // we need to iterate down, because we my eliminate records from the list
      for (int i = this.RecordsChanged.Count - 1; i >= 0; i--) {
        var rec = RecordsChanged[i];
        rec.EntityInfo.SaveEvents.OnSaveChangesAborted(rec);
        if (rec.EntityInfo.Flags.IsSet(EntityFlags.Dynamic)) {
          // Dynamic records are created on the fly when app submits changes (ex: trans logs). If trans aborted, they must be discarded
          RecordsChanged.RemoveAt(i);
          RecordsLoaded.TryRemove(rec.PrimaryKey); 
        }
      }
      EntityStore.Events.OnSaveChangesAborted(this);
      Validator = null;
    }
    #endregion

    #region Misc methods
    // Used for unit testing, to see how validation errors propagate to client. 
    // We disable validation on the client, while keep it enabled  on the server
    bool _validationDisabled;
    public void EnableValidation(bool enable) {
      _validationDisabled = !enable;
    }

    private void CheckNotReadonly() {
      Util.Check(!IsReadOnly, "Cannot modify records, session is readonly.");
    }

    private void CheckCommandArgsCount(IEntityCommand command, object[] args) {
      var entCommand = command as EntityCommand;
      var paramCount = entCommand.Parameters.Count;
      var argCount = args == null ? 0 : args.Length;
      Util.Check(paramCount == argCount, "Command {0}: Parameters count in command definition ({1}) and args count ({2}) do not match.",
        command.CommandName, paramCount, argCount);
    }
    #endregion

    #region private methods
    private EntityRecord CreateStub(EntityKey primaryKey) {
      var rec = new EntityRecord(primaryKey);
      rec.Session = this; 
      RecordsLoaded.Add(rec);
      return rec;
    }
    protected virtual EntityRecord GetLoadedRecord(EntityKey primaryKey) {
      if (primaryKey == null)
        return null;
      //1. Look in the cache
      //var entityInfo = primaryKey.Key.Entity;
      // var rec = Database.Cache.Lookup(entityInfo.EntityType, primaryKey);
      // if(rec != null) return rec;
      //2. Look in the group - if it was previously loaded
      var rec = RecordsLoaded.Find(primaryKey);
      // additionally check RecordsChanged; New records are not included in RecordsLoaded because they do not have PrimaryKes set (yet)
      if (rec == null && RecordsChanged != null && RecordsChanged.Count > 0)
        rec = FindNewRecord(primaryKey); 
      return rec;
    }

    private EntityRecord FindNewRecord(EntityKey primaryKey) {
      var ent = primaryKey.KeyInfo.Entity;
      for(int i=0; i < RecordsChanged.Count; i++) { //for-i loop is faster
        var rec = RecordsChanged[i]; 
        if (rec.Status != EntityStatus.New || rec.EntityInfo != ent) continue; 
        //we found new record of matching entity type. Check keys match. First check if key is loaded
        if (!rec.KeyIsLoaded(primaryKey.KeyInfo)) continue; 
        if (primaryKey.Equals(rec.PrimaryKey)) 
          return rec; 
      }
      return null; 
    }
    
    public virtual IList<TEntity> ToEntities<TEntity>(IEnumerable<EntityRecord> records) {
      var list = new List<EntityBase>();
      foreach (EntityRecord rec in records)
        list.Add(rec.EntityInstance);
      var entList = new ObservableEntityList<TEntity>(list); 
      return entList;
    }


    protected EntityInfo GetEntityInfo(Type entityType) {
      return EntityStore.Model.ModelInfo.GetEntityInfo(entityType, true);
    }

    private static bool IsSet(LoadFlags options, LoadFlags option) {
      return (options & option) != 0;
    }
    #endregion

    #region IQueryProvider Members

    IQueryable<TElement> IQueryProvider.CreateQuery<TElement>(Expression expression) {
      return CreateEntityQuery<TElement>(this, expression); 
    }

    IQueryable IQueryProvider.CreateQuery(Expression expression) {
      //Note: this method is almost never called, so we do not worry about efficiency here
      if (_createEntityQueryMethod == null)
        _createEntityQueryMethod = typeof(EntitySession).GetMethod("CreateEntityQuery", BindingFlags.Static | BindingFlags.NonPublic);
      var elementType = expression.Type.GetGenericArguments()[0];
      var genMethod = _createEntityQueryMethod.MakeGenericMethod(elementType);
      var query = (EntityQuery) genMethod.Invoke(null, new object[] { this, expression });
      return query;
    }

    TResult IQueryProvider.Execute<TResult>(Expression expression) {
      var qp = (IQueryProvider)this;
      return (TResult)qp.Execute(expression); 
    }

    object IQueryProvider.Execute(Expression expression) {
      var elemType = expression.Type.IsGenericType ? expression.Type.GenericTypeArguments[0] : typeof(object); 
      var query = new EntityQuery(this, elemType, expression);
      var result = ExecuteDynamicQuery(query);
      return result; 
    }

    //Helper method 
    static MethodInfo _createEntityQueryMethod;
    internal static EntityQuery<TEntity> CreateEntityQuery<TEntity>(EntitySession session, Expression expression) {
      return new EntityQuery<TEntity>(session, expression);

    }
    #endregion

    #region Cloning and attaching
    // forceFullCopy is true when we put (update) records into CacheSession
    public IList<EntityRecord> CloneAndAttach(IList<EntityRecord> records, bool forceFullCopy = false) {
      var newRecs = new List<EntityRecord>();
      for (int i = 0; i < records.Count; i++)
        newRecs.Add(CloneAndAttach(records[i], forceFullCopy));
      return newRecs;
    }

    public EntityRecord CloneAndAttach(EntityRecord record, bool forceFullCopy = false) {
      if (record.Session == this)
        return record; //do not clone if it is already in this session
      //If record with the same PK is already loaded in session, use it in result; otherwise, clone record from cache and attach
      var sessionRec = this.GetRecord(record.PrimaryKey, LoadFlags.None);
      if (sessionRec == null) {
        sessionRec = new EntityRecord(record); //make clone
        this.Attach(sessionRec);
      } else if (sessionRec.Status == EntityStatus.Stub || forceFullCopy) {
        sessionRec.CopyOriginalValues(record);
      }
      sessionRec.CachedRecord = record;
      return sessionRec;
    }


    #endregion

  }//class

}//namespace
