﻿using System;
using System.Collections.Generic;
using System.Collections.Concurrent;
using System.Linq;
using System.Text;

using Vita.Common;
using Vita.Data;
using Vita.Services;
using Vita.Entities.Model.Construction;
using Vita.Entities.Model;


namespace Vita.Entities.Runtime {

  public class EntityStore : IEntityStore {
    EntityModel _model;
    EntityModelInfo _modelInfo;

    public EntityEvents EntityEvents = new EntityEvents();
    public EntityStoreEvents Events = new EntityStoreEvents(); 
    public EntityModel Model { get { return _model; } }
    public IDataStore DataStore { get; private set; }

    #region IEntityStore Members

    public IEntitySession OpenSession(UserContext userContext = null) {
      return new EntitySession(this, userContext);
    }

    public TService GetService<TService>() where TService : class {
      return Model.GetService<TService>();
    }
    #endregion

    public EntityStore(EntityModel model, IDataStore dataStore, string appLogPath = null) {
      _model = model;
      DataStore = dataStore;
      //create default services
      Model.RegisterService(typeof(ITimeService), new TimeService());
      var logService =  new AppLogService();
      if (!string.IsNullOrEmpty(appLogPath))
        logService.SetLog(new Logging.Log(outFile: appLogPath));
      Model.RegisterService(typeof(IAppLogService), logService);
      // likely is replaced by another implementation writing to database - during modules initialization
      Model.RegisterService(typeof(IErrorLogService), new TraceErrorLogService());
      var appContext = new AppContext();
      Model.RegisterService(typeof(IAppContextProviderService), new AppContextProviderService(appContext));
      Model.RegisterService(typeof(IDataStoreListingService), new DataStoreListingService());
      //signup to TaskError event to log all async tasks exceptions
      TaskHelper.TaskError += TaskHelper_TaskError;
      //Notify modules that entity store is constructed
      foreach (var module in Model.Modules)
        module.Init(this);
      //init services
      var servList = Model.GetAllServices();
      for (int i = 0; i < servList.Count; i++) {
        var service = servList[i];
        var iServiceInit = service as IEntityServiceInit;
        if (iServiceInit != null)
          iServiceInit.Init(this);
      }
      //Init data store
      DataStore.Init(this);
    }

    public void Activate(AppActivationContext activationContext) {
      Events.OnInitializing(this, EntityStoreInitStep.Constructed, activationContext);
      Events.OnInitializing(this, EntityStoreInitStep.Activating, activationContext);
      // activate entity model - build entityModelInfo
      EntityModelInfoBuilder.Build(_model, activationContext.Log);
      _modelInfo = _model.ModelInfo;
      activationContext.CheckActivationErrors();
      this.Events.OnInitializing(this, EntityStoreInitStep.EntityModelConstructed, activationContext);
      //fire events
      this.Events.OnInitializing(this, EntityStoreInitStep.DbModelConstructing, activationContext);
      this.Events.OnInitializing(this, EntityStoreInitStep.DbModelConstructed, activationContext);
      // Activate modules
      foreach (var module in Model.Modules)
        module.Activate(activationContext);
      activationContext.CheckActivationErrors();
      this.Events.OnInitializing(this, EntityStoreInitStep.DbSchemaUpdating, activationContext); //database updates schema here
      this.Events.OnInitializing(this, EntityStoreInitStep.DbSchemaUpdated, activationContext);
      CheckAreas();
      this.Events.OnInitializing(this, EntityStoreInitStep.Activated, activationContext);
      if (!DataStore.CheckConnectivity())
        throw new ActivationException("Failed to connect to data store (database).", activationContext.Log.GetAllAsText());
      this.Events.OnInitializing(this, EntityStoreInitStep.Connected, activationContext);
    }

    //Check that each Entity Area is at least once represented in a datastore  
    private void CheckAreas() {
      //initialize areas
      var schemas = new StringSet(); 
      schemas.UnionWith(Model.Areas.Select(a => a.SchemaName));
      schemas.ExceptWith(DataStore.GetSchemas());
      if (schemas.Count == 0) 
        return; //everything is OK
      const string errorMsgTemplate = 
        "EntityAreas with schemas [{0}] in EntityModel areas are not associated with schema(s) of the data store. Register schemas using DbModelSetup methods.";
      var strMissingSchemas = string.Join(", ", schemas);
      Util.Throw(errorMsgTemplate, strMissingSchemas);
    }

    // Data access methods. 
    internal IList<EntityRecord> ExecuteQuery(EntitySession session, IEntityCommand command, Paging paging, object[] args) {
      // Switch to paged version if necessary
      try {
        command = CheckPaging(command, paging);
        return DataStore.ExecuteQuery(session, command, paging, args);
      } catch (NonFatalException nfx) {
        switch (nfx.ErrorCode) {
          case ExceptionStrings.CodeObjectNotFound: return new EntityRecord[] { };
          default: throw;
        }
      } catch (Exception ex) {
        ex.AddValue("command-name", command.CommandName);
        ex.AddValue("args", args);
        ex.AddValue("paging", paging);
        throw;
      }
    }

    internal object ExecuteNonQuery(EntitySession session, IEntityCommand command, object[] args) {
      try {
        return DataStore.ExecuteNonQuery(session, command, args);
      } catch (Exception ex) {
        ex.AddValue("command-name", command.CommandName);
        ex.AddValue("args", args);
        throw;
      }
    }
    private IEntityCommand CheckPaging(IEntityCommand command, Paging paging) {
      bool hasPaging = paging != null && paging != Paging.NoPaging;
      if (!hasPaging)
        return command; 
      //Check paging mode
      var entCommand = command as EntityCommand;
      var entInfo = entCommand.TargetEntityInfo;
      if (entCommand == entInfo.CrudCommands.SelectAll && entInfo.PagingMode == PagingMode.DataStore)
        command = entInfo.CrudCommands.SelectAllPaged; // replace with paged version; if Db driver does not support paging, it will flip it back.
      return command; 
    }

    #region ErrorLog, handling async errors
    void TaskHelper_TaskError(object sender, TaskErrorEventArgs e) {
      var errLog = this.ErrorLog;
      if (errLog != null)
        errLog.LogError(e.Task.Exception, null, null, "AsyncTask", Environment.MachineName);
    }

    public IErrorLogService ErrorLog {
      get {
        if (_errorLog == null)
          _errorLog = this.GetService<IErrorLogService>();
        return _errorLog;
      }
    } IErrorLogService _errorLog;
    #endregion


  }//class

}//ns
