﻿using System;
using System.Linq;
using System.Collections.Generic;

using Vita.Entities.Model;
using Vita.Entities.Runtime;

namespace Vita.Entities {

  /// <summary> A class working as a container for validation errors. It has utility methods for validating object properties.
  /// </summary>
  /// <remarks>Passed as a parameter to DataRecord.Validate method. </remarks>
  public class EntityValidator {
    public EntitySession Session;
    public IList<ValidationError> Errors = new List<ValidationError>();

    internal EntityValidator(EntitySession session) {
      Session = session; 
    }

    public ValidationError AddError(object invalidValue, string message, object[] messageArgs, string code, string targetProperty, object entity) {
      var err = new ValidationError(invalidValue,message, messageArgs, code, targetProperty, entity);
      Errors.Add(err);
      return err; 
    }
    public bool Check(bool condition, object invalidValue, string message, object[] messageArgs, string code, string targetProperty, object entity) {
      if (!condition) 
        AddError(invalidValue, message, messageArgs, code, targetProperty, entity);
      return condition;
    }

    public bool CheckNotNull(object value, string message, object[] messageArgs, string code, string targetProperty, object entity) {
      return Check(value != null && value != DBNull.Value, value, message, messageArgs, code, targetProperty, entity); 
    }

    public bool CheckNotEmpty(string value, string message, object[] messageArgs, string code, string targetProperty, object entity) {
      return Check(!string.IsNullOrWhiteSpace(value), value, message, messageArgs, code, targetProperty, entity);
    }

    public bool CheckLength(string value, int maxLength, string message, object[] messageArgs, string code, string targetProperty, object entity) {
      return Check(value == null || value.Length <= maxLength, value, message, messageArgs, code, targetProperty, entity);
    }

    public void ThrowValidation() {
      if (Errors.Count == 0)
        return;
      throw new ValidationException(Errors);
    }

    public bool Validate(IList<EntityRecord> records) {
      bool result = true;
      for (int i=0; i < records.Count; i++)
        result &= Validate(records[i]);
      return result;
    }

    public bool Validate(EntityRecord record) {
      record.IsValid = true;
      if (record.Status == EntityStatus.Deleting)
        return true;
      var initialCount = this.Errors.Count;
      record.EntityInfo.Events.OnValidatingChanges(record);
      foreach (var member in record.EntityInfo.Members) {
        switch (member.Kind) {
          case MemberKind.Column:
            // We do not validate columns that are not updated.
            // We do not validate foreign keys - instead, we validate corresponding EntityRef members.
            if (member.Flags.IsSet(EntityMemberFlags.ExcludeFromUpdate | EntityMemberFlags.ForeignKey))
              continue; 
            ValidateValueMember(record, member);
            continue;
          case MemberKind.EntityRef:
            ValidateEntityRefMember(record, member);
            continue;
          case MemberKind.EntityList:
          case MemberKind.Transient:
            continue; //do nothing
        }
      }//foreach member
      if (!record.IsValid) {
        var errs = Errors.Skip(initialCount).ToList();
        record.EntityInfo.Events.OnValidationFailed(record, errs);
      }
      return record.IsValid;
    }

    private void ValidateEntityRefMember(EntityRecord record, EntityMemberInfo member) {
      bool nullable = member.Flags.IsSet(EntityMemberFlags.Nullable);
      if (nullable) 
        return; //nothing to check
      var isNull = record.KeyIsNull(member.ReferenceInfo.FromKey);
      record.IsValid &= this.Check(!isNull, null,
          ExceptionStrings.MessageValueIsEmpty, new object[] { member.MemberName }, ExceptionStrings.CodeValueMissing, member.MemberName, record);
    }


    private void ValidateValueMember(EntityRecord record, EntityMemberInfo member) {
      bool nullable = member.Flags.IsSet(EntityMemberFlags.Nullable);
      var value = record.GetValue(member); 
      // Checking non-nullable fields; note that identity field for new records is initialized with temp value (in ValueBox)
      if (!nullable) {
        if (member.DataType == typeof(string))
          record.IsValid &= this.CheckNotEmpty(value as string,
              ExceptionStrings.MessageValueIsEmpty, new object[] { member.MemberName }, ExceptionStrings.CodeValueMissing, member.MemberName, record);
        else
          record.IsValid &= this.CheckNotNull(value,
              ExceptionStrings.MessageValueIsNull, new object[] { member.MemberName }, ExceptionStrings.CodeValueMissing, member.MemberName, record);
      }
      //Check string size
      if (member.DataType == typeof(string) && member.Size > 0)
        record.IsValid &= this.CheckLength(value as string, member.Size,
          ExceptionStrings.MessageValueIsTooLong, new object[] { member.MemberName, member.Size }, ExceptionStrings.CodeValueTooLong, member.MemberName, record);
    }

  }//class

}//ns
