﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml;
using System.Xml.Serialization;
using System.Diagnostics;

using Vita.Entities.Model;
using Vita.Common;
using Vita.Entities.Runtime;

namespace Vita.Entities.Xml {

  //This class will go away, eventually. Just need to implement more solid serialization infrastructure.
  public static class XmlSerializationExtensions {

    public static void WriteRecord(this EntityRecord record, XmlWriter writer, bool withType = false) {
      writer.WriteStartElement(record.EntityInfo.Name);
      WriteMembers(record, writer, withType);
      writer.WriteEndElement(); 
    }


    public static void WriteMembers(this EntityRecord record, XmlWriter writer, bool writeTypeInfo = false) {
      if (writeTypeInfo) {
        writer.WriteElementString("__type", record.EntityInfo.EntityType.FullName);
        writer.WriteElementString("__area", record.EntityInfo.Area.Name);
      }
      foreach (var member in record.EntityInfo.Members) {
        if (member.Kind != MemberKind.Column)
          continue;
        var value = record.GetValue(member);
        if (value == null) {
          writer.WriteElementString(member.MemberName, null);
        } else if (member.DataType == typeof(string) && member.Flags.IsSet(EntityMemberFlags.IsMemo)) {
          writer.WriteStartElement(member.MemberName);
          writer.WriteCData((string)value);
          writer.WriteEndElement();
        } else {
          var strValue = member.ValueToStringRef(member, value);
          writer.WriteElementString(member.MemberName, strValue);
        }
      }//foreach
    }//method

    public static void ReadRecord(this EntityRecord record, XmlReader reader, bool readEntityType = false) {
      //Read Record
      reader.ReadStartElement();
      //read member values
      ReadMembers(record, reader, readEntityType); 
      //Read /Record
      reader.ReadEndElement();
      record.Status = EntityStatus.Loaded;
    }

    public static void ReadMembers(this EntityRecord record, XmlReader reader, bool readEntityType = false) {
      reader.SkipWhitespace();
      if (readEntityType)
        ReadEntityTypeInfo(record, reader); 
      while (reader.NodeType == XmlNodeType.Element)
        ReadMember(record, reader);
    }

    private static void ReadEntityTypeInfo(this EntityRecord record, XmlReader reader) {
      //Read type element, recover EntityInfo
      string typeElement, typeName, areaElement, areaName;
      if (!reader.ReadSimpleElement(out typeElement, out typeName) || typeElement != "__type")
        Util.Throw("Xml deserialization error: expected __type element.");
      var type = ReflectionHelper.GetLoadedType(typeName);
      //Try reading area
      reader.SkipWhitespace();
      areaName = null;
      if (reader.NodeType == XmlNodeType.Element && reader.Name == "__area")
        reader.ReadSimpleElement(out areaElement, out areaName);
      var entInfo = EntityModelInfo.GetByInterface(type, areaName);
      Util.Check(entInfo != null, "Failed to retrieve entity info for type {0}.", type);
      record.EntityInfo = entInfo;
      record.InitValuesStorage();
      //Init values to defaults; this is useful when there are some schema mismatches: if some prop X was recently added to schema on the reciever,
      // but sender has no this change so prop X value is not sent in XML, then initialization of input record sets X to default value. This 
      // helps to avoid some nasty validation or get-value errors.
      Array.Copy(record.EntityInfo.InitialColumnValues, record.ValuesOriginal, record.ValuesOriginal.Length);
    }

    private static void ReadMember(this EntityRecord record, XmlReader reader) {
      string memberName = null;
      string memberValue = null;
      try {
        Util.Check(reader.ReadSimpleElement(out memberName, out memberValue), "Failed to read member element.");
        var member = record.EntityInfo.GetMember(memberName);
        if (member.Kind != MemberKind.Column)
          return;
        var value = member.ValueFromStringRef(member, memberValue);
        record.ValuesOriginal[member.ValueIndex] = value;
        reader.SkipWhitespace();
      } catch (Exception ex) {
        ex.AddValue("memberName", memberName);
        throw;
      }
    }


    public static void SkipWhitespace(this XmlReader reader) {
      while (reader.NodeType == XmlNodeType.Whitespace)
        reader.Read();
    }


    public static bool ReadSimpleElement(this XmlReader reader, out string element, out string value) {
      element = null;
      value = null;
      reader.SkipWhitespace();
      if (reader.NodeType != XmlNodeType.Element)
        return false;
      element = reader.Name;
      //Note: this is a messy case. If an element is empty
      if (reader.IsEmptyElement) {
        value = null;
        reader.Read();
      } else {
        reader.Read();
        value = reader.ReadString();
        reader.ReadEndElement();
      }
      reader.SkipWhitespace();
      return true; 
    }


    public static string GetValue(this XmlNode xmlNode, string path, string defaultValue = null) {
      var node = xmlNode.SelectSingleNode(path);
      if (node == null)
        return defaultValue;
      return node.InnerText.Trim();
    }

  } //class


}// ns
