﻿using System;
using System.Collections.Generic;
using System.Collections.Concurrent;
using System.Linq;
using System.Text;
using System.IO;
using System.Diagnostics;
using System.Threading.Tasks;

using Vita.Common;
using System.Timers;

namespace Vita.Logging {

  public enum LogFlushMode {
    None,
    Sync, 
    Async
  }

  public class Log {
    public int MaxEntries = 2000; //Prevent from overgrowing
    public bool HasErrors;
    public LogFlusher Flusher;

    // Signals that an entry was added; can be used for immediate flushing (writing to disk) of added entries, but this is not recommended.
    // Better use AsyncLogFlusher to write entries in bulk on secondary thread.
    public event EventHandler<EventArgs> EntryAdded;

    ConcurrentQueue<LogEntry> _entries = new ConcurrentQueue<LogEntry>();

    #region constructor and initialization
    public Log(int maxEntries = 2000, LogFlushMode flushMode = LogFlushMode.Async, int flushThreshold = 1000, string outFile = null) {
      MaxEntries = maxEntries;
      switch(flushMode) {
        case LogFlushMode.None: break;
        case LogFlushMode.Sync: Flusher = new LogFlusher(this, flushThreshold); break;
        case LogFlushMode.Async: Flusher = new AsyncLogFlusher(this, flushThreshold); break; 
      }
      if (!string.IsNullOrWhiteSpace(outFile)) {
        var logWriter = new FileLogWriter(outFile);
        AddWriter(logWriter); 
      }
    }
    #endregion

    public void AddFileWriter(string filePath) {
      if (Flusher == null)
        Flusher = new LogFlusher(this);
      var writer = new FileLogWriter(filePath); 
      Flusher.AddWriter(writer);
    }

    public void AddWriter(ILogWriter writer) {
      if (Flusher == null)
        Flusher = new LogFlusher(this);
      Flusher.AddWriter(writer); 
    }
    public void RemoveWriter(ILogWriter writer) {
      if (Flusher == null)
        return;
      Flusher.RemoveWriter(writer); 
    }

    public void AddEntry(LogEntry entry) {
      LogEntry dummy;
      while (_entries.Count > MaxEntries)
        _entries.TryDequeue(out dummy);
      if (entry.EntryType == LogEntryType.Error)
        HasErrors = true; 
      _entries.Enqueue(entry);
      var evtTemp = EntryAdded; 
      if (evtTemp != null) 
        evtTemp(this, EventArgs.Empty);
    }

    public void Write(LogEntryType entryType, string message, params object[] args) {
      AddEntry(new MessageLogEntry(entryType, message, args));
    }

    public void Error(string message, params object[] args) {
      AddEntry(new MessageLogEntry(LogEntryType.Error, message, args));
    }
    public void Info(string message, params object[] args) {
      AddEntry(new MessageLogEntry(LogEntryType.Information, message, args));
    }
    public void Warning(string message, params object[] args) {
      AddEntry(new MessageLogEntry(LogEntryType.Warning, message, args));
    }

    public bool IsEmpty {
      get { return _entries.IsEmpty; }
    }

    public int CurrentSize { 
      get { return _entries.Count; } 
    }
    public IList<LogEntry> GetAllEntries(bool clear = false) {
      if (clear) {
        HasErrors = false; 
        return DequeueAll();
      } else {
        return _entries.ToList();
      }
    }

    public void Flush() {
      if (IsEmpty || Flusher == null)
        return;
      Flusher.Flush(); 

    }

    public IList<LogEntry> DequeueAll() {
      var entries = new List<LogEntry>();
      LogEntry entry;
      while (_entries.TryDequeue(out entry)) {
        if (entry != null)
          entries.Add(entry);
      }
      return entries;
    }

    public string GetAllAsText(bool clear = false) {
      var entries = GetAllEntries(clear); 
      return string.Join(Environment.NewLine, entries);
    }

    public void Clear() {
      DequeueAll(); 
    }

  }// Log class

}
