﻿using System;
using System.Collections.Generic;
using System.Collections.Concurrent;
using System.Linq;
using System.Text;
using System.IO;
using System.Diagnostics;
using System.Threading.Tasks;
using System.Timers;

using Vita.Common;

namespace Vita.Logging {

  public interface ILogWriter {
    void Write(IList<LogEntry> entries);
  }

  //simple flusher, flashes immediately
  public class LogFlusher {
    public Log Log;
    public int Threshold;
    public IList<ILogWriter> Writers = new List<ILogWriter>(); 
    object _lock = new object();

    public LogFlusher(Log log, int threshold = 1) {
      Log = log;
      Threshold = threshold;
      Log.EntryAdded += Log_EntryAdded;
      AppDomain.CurrentDomain.DomainUnload += (o, e) => OnDomainUnloading();
    }

    void Log_EntryAdded(object sender, EventArgs e) {
      if (Threshold > 0 && Log.CurrentSize >= Threshold)
        Flush(); 
    }

    public void AddWriter(ILogWriter writer) {
      lock (_lock) {
        Writers.Add(writer); 
      }
    }
    public void RemoveWriter(ILogWriter writer) {
      lock (_lock) {
        Writers.Remove(writer);
      }
    }

    public void Flush() {
      // ensure single-entry
      lock (_lock) {
        FlushLocked();
      }
    }

    private void FlushLocked() {
      //repeat until all empty
      while (true) {
        var entries = Log.DequeueAll();
        if (entries.Count == 0)
          return;
        foreach (var writer in Writers)
          writer.Write(entries); 
      }//while true
    }

    protected virtual void OnDomainUnloading() {
      if (!Log.IsEmpty)
        Flush(); //this sometimes does not work - when domain is unloading, file writes OK, but the file itself does not appear on disk
    }
  
  }

  //Async log flusher; flushes log by timer, on background thread
  public class AsyncLogFlusher : LogFlusher, IDisposable {
    Timer _flushTimer;

    public AsyncLogFlusher(Log log, int threshold = 1000, int flushIntevalMs = 100) : base(log, threshold) {
      _flushTimer = new Timer(flushIntevalMs);
      _flushTimer.Elapsed += FlushTimer_Elapsed;
      _flushTimer.Start();
    }

    protected override void OnDomainUnloading() {
      _flushTimer.Stop();
      base.OnDomainUnloading();
    }

    #region Flushing
    void FlushTimer_Elapsed(object sender, ElapsedEventArgs e) {
      Flush();
    }
    
    #endregion

    #region IDisposable Members
    public void Dispose() {
      if (!Log.IsEmpty)
        Flush();
    }
    #endregion

  }// Log class

}
