﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using System.Text;

using Vita.Common;
using Vita.Entities.Model;
using Vita.Entities.Authorization;
using Vita.Entities.Runtime;
using Vita.Services;

namespace Vita.Entities {

  /// <summary>A static utility class for manipulating entities. </summary>
  public static class EntityHelper {

    /// <summary>Returns an entity session that entity is linked to.</summary>
    /// <param name="entity">The entity instance.</param>
    /// <returns>The entity session.</returns>
    public static IEntitySession GetSession(object entity) {
      Util.Check(entity != null, "GetSession: parameter 'entity' may not be null.");
      var rec = GetRecord(entity);
      return rec.Session;
    }

    /// <summary>Returns an entity session that the query is linked to.</summary>
    /// <typeparam name="TEntity">Entity type.</typeparam>
    /// <param name="dynamicQuery">An entity query.</param>
    /// <returns>The entity session.</returns>
    public static IEntitySession GetSession<TEntity>(IQueryable<TEntity> dynamicQuery) {
      var entityQuery = dynamicQuery as EntityQuery;
      Util.Check(entityQuery != null, "Invalid parameter: the object is not an entity query.");
      return entityQuery.Session; 
    }

    public static string GetUserLogContents(this IEntitySession session, bool clear = false) {
      if (session == null)
        return null;
      var ctx = (UserContext)session.UserContext;
      var log = ctx.Log;
      if (log == null)
        return null;
      var contents = log.GetAllAsText(clear);
      return contents; 
    }


    /// <summary>Returns an underlying <c>EntityRecord</c> object for an entity. </summary>
    /// <param name="entity">An entity instance.</param>
    /// <returns>The entity record.</returns>
    public static EntityRecord GetRecord(object entity) {
      if (entity == null || entity == DBNull.Value)
        return null;
      if (entity is EntityRecord)
        return (EntityRecord)entity;
      var entBase = entity as EntityBase;
      if (entBase == null)
        Util.Throw("Object {0} is not Entity, cannot retrieve Record field.", entity);
      return entBase.Record;
    }

    /// <summary>Returns the entity type (interface type) for an entity instance. </summary>
    /// <param name="entity">An entity instance.</param>
    /// <returns>The interface type for an entity.</returns>
    /// <remarks>Note that entity.GetType() returns an underlying class type (IL-generated at runtime), not entity interface type.</remarks>
    public static Type GetEntityType(object entity) {
      var rec = GetRecord(entity);
      return rec.EntityInfo.EntityType;
    }

    //Helper method that can be used to create composite primary keys
    /// <summary>Creates a primary key object for an entity identified by type. Use it for entity types with composite primary key, to create a key instance from 
    /// column values. The created key may be used in session.GetEntity() method as primary key parameter. </summary>
    /// <typeparam name="TEntity">Entity type.</typeparam>
    /// <param name="session">The entity session.</param>
    /// <param name="values">Value(s) of the primary key columns.</param>
    /// <returns>A primary key object.</returns>
    public static EntityKey CreatePrimaryKey<TEntity>(IEntitySession session, params object[] values) {
      var entSession = session as EntitySession;
      var modelInfo = entSession.EntityStore.Model.ModelInfo;
      var entInfo = modelInfo.GetEntityInfo(typeof(TEntity), true);
      var key = new EntityKey(entInfo.PrimaryKey, values);
      return key; 
    }

    /// <summary>Returns a display string for an entity. The display string is produced by a custom method specified in [Display] attribute.</summary>
    /// <param name="entity">The entity instance.</param>
    /// <returns>A display string.</returns>
    public static string GetDisplayString(object entity) {
      var rec = GetRecord(entity);
      if (rec == null) return string.Empty;
      return rec.EntityInfo.DisplayMethod(rec); 
    }

    /// <summary>Returns a display string for an entity record. The display string is produced by a custom method specified in [Display] attribute.</summary>
    /// <param name="record">The entity record.</param>
    /// <returns>A display string.</returns>
    public static string GetDisplayString(EntityRecord record) {
      if (record == null) return string.Empty;
      return record.EntityInfo.DisplayMethod(record);
    }

    /// <summary>Returns a value of the entity property.</summary>
    /// <param name="entity">The entity instance.</param>
    /// <param name="propertyName">The property name.</param>
    /// <returns>The value of the property.</returns>
    public static object GetProperty(object entity, string propertyName) {
      var rec = EntityHelper.GetRecord(entity);
      var value = rec[propertyName];
      return value;
    }

    /// <summary>Returns a strongly-typed value of the entity property.</summary>
    /// <typeparam name="T">The result type.</typeparam>
    /// <param name="entity">The entity instance.</param>
    /// <param name="propertyName">The property name.</param>
    /// <returns>The value of the property.</returns>
    public static T GetProperty<T>(object entity, string propertyName) {
      var rec = EntityHelper.GetRecord(entity);
      var value = rec[propertyName];
      return (T) value;
    }

    /// <summary>Sets a property value of an entity. </summary>
    /// <param name="entity">The entity instance.</param>
    /// <param name="propertyName">The property name.</param>
    /// <param name="value">The property value.</param>
    public static void SetProperty(object entity, string propertyName, object value) {
      var rec = EntityHelper.GetRecord(entity);
      rec[propertyName] = value;
    }

    /// <summary>Returns an instance of the <c>IEntityAccess</c> interface. This object may be used to retrieve the detailed information about authorization permissions 
    /// for the entity and the current user. </summary>
    /// <param name="entity">The entity instance.</param>
    /// <returns>The access info object.</returns>
    public static IEntityAccess GetEntityAccess<TEntity>(TEntity entity) {
      return GetRecord(entity); 
    }

    public static EntityQuery GetLastQuery(this IEntitySession session) {
      var entSession = (EntitySession)session;
      return entSession.LastQuery; 
    }

    /// <summary>Checks if specified properties of the entity are loaded from the database. Reloads the entity if necessary.</summary>
    /// <typeparam name="TEntity">The entity type.</typeparam>
    /// <param name="entity">The entity instance.</param>
    /// <param name="propertySelector">An expression identifying the properties to check.</param>
    /// <example>
    ///  EntityHelper.EnsureLoaded(book, b => b.Authors); 
    ///  EntityHelper.EnsureLoaded(book, b => new {b.Authors, b.Publisher}); 
    /// </example>
    public static void EnsureLoaded<TEntity>(TEntity entity, Expression<Func<TEntity, object>> propertySelector) {
      IEnumerable<string> names;
      var resultType = propertySelector.Body.Type;
      if (resultType.IsAnonymousType()) {
        names = resultType.GetProperties().Select(p=>p.Name); 
      } else {
        var memberExpr = propertySelector.Body as MemberExpression;
        Util.Check(memberExpr != null, "Invalid property selector: must be an anonymous type or single property selector.");
        names = new string[] {memberExpr.Member.Name};
      }
      var rec = GetRecord(entity);
      var entInfo = rec.EntityInfo; 
      var members = names.Select(n => entInfo.GetMember(n, throwIfNotFound: true));
      rec.EnsureLoaded(members);
    }

    /// <summary>Disables compiled query caching for a query. Use it for 'search' queries produced from 
    /// custom search forms, to prevent polluting the query cache with custom one-shot queries.</summary>
    /// <typeparam name="TEntity">Entity type.</typeparam>
    /// <param name="entitySet">A dynamic query instance; usually a source entity set.</param>
    /// <param name="disable">Set to true to disable caching of query definition.</param>
    /// <remarks><para>
    /// You can set this flag on any original EntitySet used in the query - in this case it will be 
    /// propagated to the final query. Example: 
    /// </para>
    /// <example>
    ///   var books = session.EntitySet&lt;IBook&gt;().DisableQueryCache();
    ///   var query = from b in books
    ///               where b.Title.StartsWith(userSearchTerm)
    ///               select b;
    /// </example>
    /// </remarks>
    public static IQueryable<TEntity> DisableQueryCache<TEntity>(this IQueryable<TEntity> entitySet) {
      var entQuery = entitySet as EntityQuery;
      if (entQuery != null)
        entQuery.DisableQueryCache = true; 
      return entitySet; 
    }

    /// <summary> Compares arrays of bytes. Primary use is for comparing row version properties (which are of type byte[]). </summary>
    /// <param name="rowVersion">Value to compare.</param>
    /// <param name="other">Value to compare with.</param>
    /// <returns>True if array lengths and byte values match. Otherwise, false.</returns>
    public static bool EqualValues(this byte[] rowVersion, byte[] other) {
      if (rowVersion == null && other == null) return true;
      if (rowVersion == null || other == null) return false;
      if (rowVersion.Length != other.Length)
        return false;
      for (int i = 0; i < rowVersion.Length; i++)
        if (rowVersion[i] != other[i])
          return false;
      return true;
    }

    public static void SetLog(this IAppLogService logService, Logging.Log log) {
      var srv = (AppLogService)logService;
      srv.Log = log; 
    }

  }//class
}//ns
