﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using Vita.Common;
using Vita.Data;
using Vita.Entities;
using Vita.Entities.Model;
using Vita.Entities.Model.Construction;
using Vita.Entities.Runtime;
using Vita.Services;
using Vita.Logging;

namespace Vita.Entities {

  /// <summary>Entity model. </summary>
  public partial class EntityModel {
    /// <summary>The model name. Defaults to type name. </summary>
    public string ModelName;
    /// <summary>A dictionary of custom attribute handlers.</summary>
    public readonly IDictionary<Type, CustomAttributeHandler> AttributeHandlers;

    public IEnumerable<EntityArea> Areas {
      get { return AreaList; }
    } internal IList<EntityArea> AreaList = new List<EntityArea>(); 

    public IEnumerable<EntityModule> Modules {
      get { return ModuleList; }
    } internal IList<EntityModule> ModuleList = new List<EntityModule>();

    public EntityModelInfo ModelInfo; //created at initialization time
    internal readonly List<EntityReplacementInfo> Replacements = new List<EntityReplacementInfo>();
    internal readonly HashSet<Type> CompanionTypes = new HashSet<Type>();

    public EntityModuleSettings DefaultModuleSettings;
    //services
    private IDictionary<Type, object> _servicesLookup = new Dictionary<Type, object>();
    

    /// <summary> Constructs a new EntityModelSetup object. </summary>
    public EntityModel(string modelName = null) {
      ModelName = modelName ?? this.GetType().Name; 
      AttributeHandlers = CustomAttributeHandler.GetDefaultHandlers();
      DefaultModuleSettings = new EntityModuleSettings();
    }

    public EntityArea AddArea(string name, string schemaName, string description = null) {
      return new EntityArea(this, name, schemaName, description);
    }

    /// <summary> Replaces one registered entity with another. </summary>
    /// <param name="replacedType">The entity type to be replaced.</param>
    /// <param name="replacementEntityType">The new replacing entity type.</param>
    /// <remarks><para>
    /// This method provides a way to extend entities defined in independently built modules. 
    /// The other use is to integrate the independently developed modules, so that the tables in database 
    /// coming from different modules can actually reference each other through foreign keys.
    /// </para><para>
    /// The replacement type might be registered with another module - in this case it remains with its module. 
    /// If it is not registered with any module, it is associated with the module of the entity being replaced.
    /// </para> 
    /// </remarks>
    public void ReplaceEntity(Type replacedType, Type replacementEntityType) {
      Util.Check(replacedType.IsEntity(), "Invalid type: {0}; expected Entity interface.", replacedType);
      Util.Check(replacementEntityType.IsEntity(), "Invalid type: {0}; expected Entity interface.", replacementEntityType);
      //Unless the type being replaced is an empty stub, then check that new type is compatible with the old type
      var oldTypeProps = replacedType.GetAllProperties();
      if (oldTypeProps.Count > 0) { //it is not an empty stub
        Util.Check(replacedType.IsAssignableFrom(replacementEntityType), "The replacing type ({0}) must be a subtype of the type being replaced ({1}). ", replacementEntityType, replacedType);
      }
      var replInfo = new EntityReplacementInfo() { ReplacedType = replacedType, NewType = replacementEntityType};
      Replacements.Add(replInfo); 
    }

    /// <summary> Registers companion types. </summary>
    /// <param name="companionTypes">Companion types for registered entities.</param>
    /// <remarks>Companion type is used as an alternative place to put attributes for an entity. 
    /// Any attribute you put on an entity (or its property) you can put on a companion type (or its property). 
    /// This facility might be useful for separation of concerns. For example, you can place all database
    /// index attribute on companion types that are located in a separate file that is maintaned by a 
    /// developer with appropriate knowledge. 
    /// </remarks>
    public void RegisterCompanionTypes(params Type[] companionTypes) {
      CompanionTypes.UnionWith(companionTypes);
    }

    public void RegisterService(Type serviceType, object service) {
      var objType = service.GetType();
      bool ok;
      if (serviceType.IsInterface)
        ok = objType.ImplementsInterface(serviceType); 
      else 
        ok = objType == serviceType || objType.IsSubclassOf(serviceType);
      Util.Check(ok,
          "Failed to register service {0} for service type {1} - the service instance must implement or inherit service type.", service, serviceType);
      _servicesLookup[serviceType] = service; 
    }

    public TService GetService<TService>()  where TService : class {
      object result;
      _servicesLookup.TryGetValue(typeof(TService), out result);
      return result as TService;
    }

    public object GetService(Type serviceType) {
      object result;
      _servicesLookup.TryGetValue(serviceType, out result);
      return result;
    }
    internal List<object> GetAllServices() {
      return _servicesLookup.Values.ToList(); 
    }

    public IEnumerable<Type> GetAllEntityTypes() {
      var result = new List<Type>();
      foreach (var module in this.Modules)
        result.AddRange(module.Entities);
      return result; 
    }

    public override string ToString() {
      return ModelName;
    }

  }//class

}//ns
