﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Linq.Expressions;
using System.ComponentModel;
using System.Reflection;

using Vita.Common; 
using Vita.Entities.Model;
using Vita.Entities.Runtime;

namespace Vita.Entities {

  public partial class EntityModule {
    public IEntityStore EntityStore;
    public EntityModel Model;  
    public string Name;
    public string Description;
    public EntityArea Area;

    public readonly HashSet<Type> Entities = new HashSet<Type>();
    internal readonly List<EntityCommand> CustomCommands = new List<EntityCommand>();

    /// <summary>Returns the settings object for an entity module. By default, the underlying field is null and the property returns 
    /// the shared object from model's DefaultModuleSettings property. </summary>
    public EntityModuleSettings Settings {
      get {
        if (_settings == null)
          return Model.DefaultModuleSettings;
        else
          return _settings; 
      }
      set { _settings = value; }
    } EntityModuleSettings _settings; 

    public EntityModule(EntityArea area, string name, string description = null) {
      Area = area; 
      Model = Area.Model;
      Name = name;
      Description = description;
      Model.ModuleList.Add(this);
    }

    /// <summary>Notifies that entity store is constructed. Initialization is not complete at this moment. </summary>
    /// <param name="entityStore">The entity store.</param>
    public virtual void Init(IEntityStore entityStore) {
      EntityStore = entityStore;
    }

    /// <summary>Notifies that entity store is activating. </summary>
    ///<param name="activationContext">Activation context.</param>
    public virtual void Activate(AppActivationContext activationContext) {
    }


    /// <summary>Registers entity type with the entity model. </summary>
    /// <param name="area">A schema (database schema, like "dbo") for tables associated with entities. </param>
    /// <param name="entities">A set of entity types.</param>
    /// <remarks>You do not need to register all entities explicitly. When you register an entity, the system scans its 
    /// properties for types that are entity types - and registers them as well. The process is recursive. </remarks>
    public void RegisterEntities(params Type[] entities) {
      foreach (var entType in entities) {
        Util.Check(entType.IsEntity(), "Invalid type: {0}; expected Entity interface.", entType);
        Entities.Add(entType);
      }//entType
    }

    /// <summary> Replaces one registered entity with another. </summary>
    /// <param name="entityToReplace">The entity type to be replaced.</param>
    /// <param name="newEntity">The new replacing entity type.</param>
    /// <remarks>Redirects the call to <c>EntityModelSetup.ReplaceEntity</c> method.</remarks>
    protected void ReplaceEntity(Type entityToReplace, Type newEntity) {
      Entities.Add(entityToReplace); //just in case, register the entity
      Model.ReplaceEntity(entityToReplace, newEntity);
    }

    /// <summary> Registers companion types. </summary>
    /// <param name="companionTypes">Companion types for registered entities.</param>
    /// <remarks>Companion type is used as an alternative place to put attributes for an entity. 
    /// Any attribute you put on an entity (or its property) you can alternatively put on a companion type (or its property). 
    /// This facility might be useful for separation of concerns. For example, you can place all database
    /// index attribute on companion types that are located in a separate file that is maintaned by a 
    /// developer with appropriate database management skills. 
    /// </remarks>
    protected void RegisterCompanionTypes(params Type[] companionTypes) {
      Model.RegisterCompanionTypes(companionTypes);
    }

/* Temporarily disabled
  
    #region Custom command construction
    /// <summary>Returns an queryable entity set representing a database table. Use it when constructing LINQ queries for custom commands. </summary>
    /// <typeparam name="TEntity">Entity type.</typeparam>
    /// <returns>Queryable entity set.</returns>
    protected IQueryable<TEntity> EntitySet<TEntity>() {
      return new EntitySet<TEntity>();
    }
    /// <summary>Creates a LINQ query parameter. Use it when constructing LINQ queries for custom commands.</summary>
    /// <typeparam name="T">Parameter type.</typeparam>
    /// <param name="size">Parameter size.</param>
    /// <returns>A query parameter instance.</returns>
    // NOTE: we cannot use default value for parameter here, LINQ rejects such functions, so we have two overloads
    protected static EntityQueryParameter<T> Param<T>(string name, int size = 0) {
      return new EntityQueryParameter<T>(name, size);
    }

    /// <summary>Creates a LINQ query parameter. Use it when constructing LINQ queries for custom commands.</summary>
    /// <typeparam name="T">Parameter type.</typeparam>
    /// <param name="size">Parameter size. Use for string or byte-array parameters to specify the exact size.</param>
    /// <returns>A query parameter instance.</returns>
    protected static EntityQueryParameter<T> Param<T>(int size) {
      return Param<T>(Unnamed, size);
    }

    /// <summary>Creates a LINQ query parameter. Use it when constructing LINQ queries for custom commands.</summary>
    /// <typeparam name="T">Parameter type.</typeparam>
    /// <returns>A query parameter instance.</returns>
    protected static EntityQueryParameter<T> Param<T>() {
      return Param<T>(Unnamed, 0);
    }

    /// <summary>Creates a string LINQ query parameter. Use it when constructing LINQ queries for custom commands.</summary>
    /// <returns>A query parameter instance.</returns>
    protected static EntityQueryParameter<string> StringParam(int size) {
      return Param<string>(Unnamed, size);
    }

    /// <summary>Creates a string Memo parameter. Use it when constructing LINQ queries for custom commands.</summary>
    /// <returns>A query parameter instance.</returns>
    protected static EntityQueryParameter<string> MemoParam() {
      return Param<string>(Unnamed, -1);
    }

    /// <summary>Creates a constant value to use in LINQ queries. </summary>
    /// <typeparam name="T">Constant type.</typeparam>
    /// <param name="value">Constant value.</param>
    /// <returns>Query parameter instance representing a constant.</returns>
    protected static EntityQueryParameter<T> Const<T>(T value) {
      var prm = new EntityQueryParameter<T>(Unnamed, 0, value);
      prm.IsLinqConst = true;
      return prm;
    }

    /// <summary>Creates a custom SELECT compiled command based on a custom query. The command is compiled into stored procedure 
    /// when they are supported in the database and enabled in the entity model. </summary>
    /// <typeparam name="TEntity">The result entity type.</typeparam>
    /// <param name="commandName">Command name. The name will be used as stored procedure name.</param>
    /// <param name="description">Command description.</param>
    /// <param name="query">A base LINQ query returning a list of entities.</param>
    /// <returns>A compiled command.</returns>
    /// <remarks>
    /// The LINQ query must be created using EntityModule.EntitySet&ltTEntity&gt; method as entity table references.
    /// Use 'Param' method for parameters in the query.
    /// </remarks>
    /// 
    protected IEntityCommand CreateSelect<TEntity>(string commandName, string description, IQueryable query) {
      return CreateCommandFromQuery<TEntity>(commandName, description, EntityCommandKind.CustomSelect, query);
    }

    /// <summary>Creates a custom INSERT compiled command based on a custom query. The command is compiled into stored procedure 
    /// when they are supported in the database and enabled in entity model. </summary>
    /// <typeparam name="TEntity">The type of the entity to insert.</typeparam>
    /// <param name="commandName">Command name. The name will be used as stored procedure name.</param>
    /// <param name="description">Command description.</param>
    /// <param name="query">A base LINQ query returning an anonymous type with properties matching the properties of the entity.</param>
    /// <returns>A compiled command.</returns>
    /// <remarks>The query should return an auto-type with properties matching the properties of the entity being inserted. 
    /// The LINQ query must be created using EntityModule.EntitySet&ltTEntity&gt; method as entity table references.
    /// Use 'Param' method for parameters in the query. </remarks>
    protected IEntityCommand CreateInsert<TEntity>(string commandName, string description, IQueryable query) {
      return CreateCommandFromQuery<TEntity>(commandName, description, EntityCommandKind.CustomInsert, query);
    }

    /// <summary>Creates a custom UPDATE compiled command based on a custom query. The command is compiled into stored procedure 
    /// when they are supported in the database and enabled in entity model. </summary>
    /// <typeparam name="TEntity">The type of the entity to update.</typeparam>
    /// <param name="commandName">Command name. The name will be used as stored procedure name.</param>
    /// <param name="description">Command description.</param>
    /// <param name="query">A base LINQ query returning an anonymous type with properties matching the properties of the entity.
    /// The properties must include primary key columns and properties to update.</param>
    /// <returns>A compiled command.</returns>
    /// <remarks>
    /// The LINQ query must be created using EntityModule.EntitySet&ltTEntity&gt; method as entity table references.
    /// Use 'Param' method for parameters in the query.
    /// </remarks>
    protected IEntityCommand CreateUpdate<TEntity>(string commandName, string description, IQueryable query) {
      return CreateCommandFromQuery<TEntity>(commandName, description, EntityCommandKind.CustomUpdate, query);
    }

    /// <summary>Creates a custom DELETE compiled command based on a custom query. The command is compiled into stored procedure 
    /// when they are supported in the database and enabled in entity model. </summary>
    /// <typeparam name="TEntity">The type of the entity to delete.</typeparam>
    /// <param name="commandName">Command name. The name will be used as stored procedure name.</param>
    /// <param name="description">Command description.</param>
    /// <param name="query">A base LINQ query returning an anonymous type with properties matching the primary key properties of the entity.</param>
    /// <returns>A compiled command.</returns>
    /// <remarks>
    /// The LINQ query must be created using EntityModule.EntitySet&ltTEntity&gt; method as entity table references.
    /// Use 'Param' method for parameters in the query.
    /// </remarks>
    protected IEntityCommand CreateDelete<TEntity>(string commandName, string description, IQueryable query) {
      return CreateCommandFromQuery<TEntity>(commandName, description, EntityCommandKind.CustomDelete, query);
    }
    #endregion
    */

    #region partial update commands
    /// <summary> Creates a custom update command. </summary>
    /// <typeparam name="TEntity">Entity type.</typeparam>
    /// <param name="commandName">Command name. The name will be used as stored procedure name.</param>
    /// <param name="description">Command description.</param>
    /// <param name="updateMemberSelector">A lambda expression returning anonymous type with properties to update.</param>
    /// <returns>A compiled command.</returns>
    /// <remarks>The method creates a custom command that updates selected values of a single row in the database table for an entity.
    /// The member selector expression should return an anonymous type with properties matching the entity properties.
    /// The primary key values of the entity are added automatically to the command parameters list. 
    /// Example: 
    /// <example> 
    ///   CreateUpdate&lt;IPerson&gt;("PersonUpdateLastName", "Updates last name", p=> new {p.LastName}); 
    /// </example>
    /// The list of arguments when invoking the method should start with primary key value followed by a list of the new values for 
    /// each property appearing in the anonymous type constructor.
    /// </remarks>
    public IEntityCommand CreatePartialUpdate<TEntity>(string commandName, string description, Expression<Func<TEntity, object>> updateMemberSelector) {
      var resultAutoType = updateMemberSelector.Body.Type;
      var autoProps = resultAutoType.GetProperties();
      var memberNames = autoProps.Select(p => p.Name);
      return CreatePartialUpdate<TEntity>(commandName, description, memberNames);
    }

    /// <summary> Creates a custom update command. </summary>
    /// <typeparam name="TEntity">Entity type.</typeparam>
    /// <param name="commandName">Command name. The name will be used as stored procedure name.</param>
    /// <param name="description">Command description.</param>
    /// <param name="updateMembers">A comma-delimited list of property names that should be updated.</param>
    /// <returns>A compiled command.</returns>
    /// <remarks>The method creates a custom command that updates selected values of a single row in the database table for an entity.
    /// The primary key values of the entity are added automatically to the command parameters list. 
    /// Example: 
    /// <example> 
    ///   CreateUpdate&lt;IPerson&gt;("PersonUpdateMarriedStatus", "Updates married status and last name", "MarriedStatus,LastName"); 
    /// </example>
    /// The list of arguments when invoking the method should start with primary key value followed by a list of new values for 
    /// each property appearing in the <c>updateMembers</c> list.
    /// </remarks>
    public IEntityCommand CreatePartialUpdate<TEntity>(string commandName, string description, string updateMembers) {
      return CreatePartialUpdate<TEntity>(commandName, description, Util.SafeSplit(updateMembers));
    }

    private IEntityCommand CreatePartialUpdate<TEntity>(string commandName, string description, IEnumerable<string> updateMembers) {
      var cmd = new EntityCommand(commandName, description, EntityCommandKind.PartialUpdate, typeof(TEntity), this.Area);
      cmd.UpdateMemberNames = updateMembers.ToList();
      CustomCommands.Add(cmd);
      return cmd;
    }

    #endregion

    #region private utilities

    private IEntityCommand CreateCommandFromQuery<TEntity>(string commandName, string description, EntityCommandKind kind, IQueryable query) {
/*
      Util.Check(query != null, "Query passed to CreateCommand is null. CommandName: {0}", commandName);
      var entType = typeof(TEntity);
      var entQuery = query as EntityQuery;
      Util.Check(entQuery != null, "Failed to create entity command {0}: 'query' parameter must be an entity query.", commandName);
      var preprocessor = new EntityQueryInfoBuilder();
      preprocessor.Build(entQuery);
      //Create command
      var cmd = new EntityCommand(commandName, description, kind, entType, this.Area, entQuery);
      CustomCommands.Add(cmd);
      return cmd;
 */
      return null; 
    }
    #endregion

  }//class

} // ns
