﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

using Vita.Entities;
using Vita.Entities.Runtime;

namespace Vita.Services {
  /// <summary> Time service interface definition.  </summary>
  /// <remarks> All code in VITA framework requests current time through a globally available time service, not from <c>DateTime.Now</c> directly.
  /// The goal is to allow easy unit testing of time-sensitive functionality. The test code can easily shift current time to test any functionality 
  /// involving current time. 
  /// Additionally time service provides a millisecond-resolution tick counter based on StopWatch; this counter may be used to 
  /// measure execution time of operations. Note that absolute value of EllapsedMilliseconds property is meaningless, only differences matter. 
  /// </remarks>
  public interface ITimeService {
    DateTime Now { get; }
    TimeSpan CurrentOffset { get; }
    long ElapsedMilliseconds { get; }
    void SetCurrentDateTime(DateTime current); //For testing
  }

  //implementation
  internal class TimeService : ITimeService, IEntityServiceInit, IDisposable {
    private Stopwatch _stopWatch;
    private long _millisecondOffset;

    public TimeService() {
      _stopWatch = new Stopwatch();
    }

    #region ITimeService Members

    public DateTime Now {
      get {
        var now = DateTime.Now;
        return _millisecondOffset == 0 ? now : now.AddMilliseconds(_millisecondOffset);
      }
    }

    public TimeSpan CurrentOffset {
      get {
        return TimeSpan.FromMilliseconds(_millisecondOffset);
      }
    }

    public long ElapsedMilliseconds {
      get { return _stopWatch.ElapsedMilliseconds + _millisecondOffset; }
    }

    public void SetCurrentDateTime(DateTime current) {
      var now = DateTime.Now;
      _millisecondOffset = (long)current.Subtract(now).TotalMilliseconds;
    }

    #endregion

    #region IDisposable Members

    public void Dispose() {
      if (_stopWatch != null)
        _stopWatch.Stop();
      _stopWatch = null;
    }

    #endregion

    #region IEntityServiceInit Members

    public void Init(IEntityStore entityStore) {
      _stopWatch.Start();
    }

    #endregion
  }//class

}
