﻿using System;
using System.Data;
using System.Linq;
using System.Text;
using System.Reflection;
using System.Collections.Generic;
using System.Threading;

using Vita.Common;
using Vita.Entities;
using Vita.Entities.Model;
using Vita.Data.Caching;
using Vita.Data.Model;
using Vita.Data.Driver;
using Vita.Services;
using Vita.Entities.Runtime;
using Vita.Entities.Model.Construction;
using Vita.Data;
using Vita.Data.Management;
using Vita.Data.Model.Construction;
using Vita.Logging;

namespace Vita.Activation {

  public static class ActivationHelper {

    /// <summary> Creates an entity store instance from an entity model using datastore, cache settings. </summary>
    public static IEntityStore CreateEntityStore(this EntityModel entityModel,
                                                      DbDriver driver,
                                                      string connectionString,
                                                      DbOptions dbOptions = DbOptions.Default,
                                                      bool updateSchema = true,
                                                      string logFilePath = null,
                                                      string activationLogFilePath = null,
                                                      CacheSettings cacheSettings = null) {
      var schemas = entityModel.Areas.Select(a => a.SchemaName);
      DbSchemaUpdateOptions schemaUpdateOptions = updateSchema ? DbSchemaUpdateOptions.DefaultDev : DbSchemaUpdateOptions.DefaultProduction;
      var dbSettings = new DbSettings(driver, dbOptions, connectionString, schemaUpdateOptions, schemas: schemas);
      return CreateEntityStore(entityModel, dbSettings, cacheSettings, activationLogFilePath, logFilePath);
    }

    /// <summary> Creates an entity store instance from an entity model using database settings and cache settings. </summary>
    public static IEntityStore CreateEntityStore(this EntityModel entityModel, DbSettings dbSettings, CacheSettings cacheSettings = null, string activationLogFilePath = null, string logFilePath = null) {
      var db = new Database(entityModel, dbSettings, entityModel.ModelName);
      //Create entity store
      IDataStore dataStore; 
      if (cacheSettings != null && cacheSettings.HasTypes()) 
        dataStore = new CacheDataStore(entityModel, db, cacheSettings);
       else 
        dataStore = db; 
      var entityStore = new EntityStore(entityModel, dataStore);
      //Activate entity store
      activationLogFilePath = activationLogFilePath ?? logFilePath;
      var actContext = new AppActivationContext(entityStore, activationLogFilePath);
      actContext.Log.Info("ActivationLog: Application started at " + DateTime.Now.ToString("s"));
      entityStore.Activate(actContext);
      actContext.CheckActivationErrors();
      return entityStore; 
    }

    public static void DeleteAllData(Database db, IEnumerable<Type> entityTypes) {
      // get existing tables; in unit tests it might happen that when we delete all for table that does not exist yet
      var driver = db.DbModel.Driver; 
      var sm = db.DbModel.Driver.CreateSchemaManager(db.ConnectionString);
      var tables = sm.GetTables();
      var tableNames = new StringSet();
      foreach (DataRow row in tables.Rows)
        tableNames.Add(row["TABLE_NAME"].ToString());

      //Figure out table list
      var modelInfo = db.DbModel.EntityModel.ModelInfo;
      var entList = new List<EntityInfo>();
      foreach (var entType in entityTypes) {
        var entityInfo = modelInfo.GetEntityInfo(entType);
        if (entityInfo == null) continue;
        entList.Add(entityInfo);
      }
      //sort in topological order
      entList.Sort((x, y) => x.TopologicalIndex.CompareTo(y.TopologicalIndex));
      // delete all one-by-one
      var conn = driver.CreateConnection(db.ConnectionString);
      try {
        conn.Open();
        foreach (var entityInfo in entList) {
          var table = db.DbModel.LookupDbObject<DbTableInfo>(entityInfo);
          if (table == null)
            continue;
          if (!tableNames.Contains(table.TableName))
            continue;
          var cmd = conn.CreateCommand();
          var fullName = driver.GetFullName(table.SchemaName, table.TableName);
          cmd.CommandText = Util.SafeFormat("DELETE FROM {0};", fullName);
          driver.ExecuteCommand(cmd, DbExecutionType.NonQuery);
        }
      } finally {
        conn.Close();
      }
    }

  }//class

}//namespace
