﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Linq.Expressions;
using System.Diagnostics;
using System.Reflection;
using System.Threading;

using Vita.Common;
using Vita.Entities;
using Vita.Entities.Model;
using Vita.Entities.Runtime;
using Vita.Logging;

using Vita.Data;
using Vita.Data.Linq;
using Vita.Data.Caching.Implementation;
using Vita.Services;
using Vita.Activation;

namespace Vita.Data.Caching {

  public class CacheDataStore : IDataStore {
    public readonly EntityModel Model; 
    public readonly IDataStore RealDataStore;
    public CacheSettings Settings;
    public readonly EntityCache Cache;
    EntityStore _entityStore;

    public CacheDataStore(EntityModel model, IDataStore realDataStore, CacheSettings settings) {
      Model = model; 
      RealDataStore = realDataStore;
      Settings = settings;
      Cache = new EntityCache(this);
      model.RegisterService(typeof(ICacheControlService), Cache);
    }

    #region IDataStore Members
    public void Init(EntityStore entityStore) {
      _entityStore = entityStore;
      _entityStore.Events.Initializing += Events_Initializing; 
      RealDataStore.Init(_entityStore);
    }

    void Events_Initializing(EntityStore entityStore, EntityStoreInitEventArgs args) {
      switch(args.InitStep) {
        case EntityStoreInitStep.Activated:
          Cache.Init(_entityStore);
          break; 
        case EntityStoreInitStep.Connected:
          Cache.Invalidate();
          break; 
      }
    }

    public IEnumerable<string> GetSchemas() {
      return RealDataStore.GetSchemas(); 
    }

    public IList<EntityRecord> ExecuteQuery(EntitySession session, IEntityCommand command, Paging paging, object[] args) {
      IList<EntityRecord> records;
      var entCommand = (EntityCommand)command;
      var targetEnt = entCommand.TargetEntityInfo; 
      var isCached = targetEnt.Flags.IsSet(EntityFlags.Cached);
      var entSession = (EntitySession)session; 
      if (isCached && Cache.TryExecuteQuery(entSession, command, paging, args, out records))  {
        records = entSession.CloneAndAttach(records); 
        return records;
      }
      //execute in real data store
      records = RealDataStore.ExecuteQuery(session, command, paging, args);
      // For sparse entities, add to cache
      if (targetEnt.CachingType == CachingType.Sparse)
        Cache.CacheDbQueryResults(entCommand, records);
      return records; 
    } 

    public object ExecuteNonQuery(EntitySession session, IEntityCommand command, object[] args) {
      return RealDataStore.ExecuteNonQuery(session, command, args);
    }

    public void SaveChanges(EntitySession session) {
      RealDataStore.SaveChanges(session);
    }

    public bool CheckConnectivity() {
      return RealDataStore.CheckConnectivity(); 
    }

    public object ExecuteDynamicQuery(EntitySession session, EntityQuery query) {
      object result;
      var entSession = (EntitySession)session;
      if (Cache.TryExecuteDynamicQuery(entSession, query, out result)) {
        //records/entities are already cloned by calls to CloneEntity inside query expression
        return result;
      }
      else 
        return RealDataStore.ExecuteDynamicQuery(session, query);
    }

    #endregion


  }//class
}
