﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Vita.Data.Caching.Implementation;

namespace Vita.Data.Caching {

  [Flags]
  public enum CacheOptions {
    None = 0,
    /// <summary>Update cached records asynchounously after updates are sent to database, so updating session does not wait for cache write lock and completion of update.
    /// <remarks>
    /// If the flag is not set, the cached records are updated immediately and updating thread waits for completion.</summary>
    /// Use this flag with caution: if you enable Async update, be careful when opening multiple sessions without shared UserContext
    ///  - these sessions may not see each other updates immediately.
    /// </remarks>
    AsyncUpdate = 1,
    UseSimpleInvalidator = 1 << 1,

    Default = UseSimpleInvalidator,
  }

  public class CacheSettings {
    public readonly HashSet<Type> FullyCachedTypes = new HashSet<Type>();
    public readonly HashSet<Type> SparselyCachedTypes = new HashSet<Type>();
    public int SparseCacheCapacity;
    public TimeSpan FullSetCacheExpiration;
    public TimeSpan SparseCacheExpiration;
    public CacheOptions Options;
    public ICacheInvalidator Invalidator;

    public CacheSettings(IEnumerable<Type> fullyCachedTypes = null, IEnumerable<Type> sparselyCachedTypes = null,
                TimeSpan? fullSetCacheExpiration = null, TimeSpan? sparseCacheExpiration = null, 
                int sparseCacheCapacity = 10000,  CacheOptions options = CacheOptions.Default, 
                ICacheInvalidator invalidator = null) {
      if (fullyCachedTypes != null)
        FullyCachedTypes.UnionWith(fullyCachedTypes);
      if (sparselyCachedTypes != null)
        SparselyCachedTypes.UnionWith(sparselyCachedTypes);
      FullSetCacheExpiration = fullSetCacheExpiration == null ? TimeSpan.FromMinutes(5) : fullSetCacheExpiration.Value;
      SparseCacheExpiration = sparseCacheExpiration == null ? TimeSpan.FromMinutes(1) : sparseCacheExpiration.Value;
      SparseCacheCapacity = sparseCacheCapacity;
      Options = options;
      Invalidator = invalidator;
      // Some sanity check - does not make sense to keep entity in sparse cache if it is in full cache
      SparselyCachedTypes.ExceptWith(FullyCachedTypes);
    }

    public bool HasTypes() {
      return FullyCachedTypes.Count > 0 || SparselyCachedTypes.Count > 0;
    }
  } //class

  public static class CacheSettingsExtensions {
    public static bool IsSet(this CacheOptions options, CacheOptions option) {
      return (options & option) != 0;
    }
  }

}
