﻿using System;
using System.Collections;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

using Vita.Common;
using Vita.Entities;
using Vita.Entities.Model;
using Vita.Entities.Runtime;
using Vita.Data.Linq;
using Vita.Logging;
using Vita.Services;

namespace Vita.Data.Caching.Implementation {

  public partial class EntityCache : ICacheControlService {
    EntityStore _entityStore; 
    public readonly CacheSettings Settings;
    public CacheDataStore CacheDataStore;
    public EntitySession CacheSession;
    IErrorLogService _errorLog;
    bool _enabled;

    public readonly ReaderWriterLockSlim Lock = new ReaderWriterLockSlim(LockRecursionPolicy.SupportsRecursion);
    public Dictionary<Type, FullyCachedSet> FullyCachedSets = new Dictionary<Type, FullyCachedSet>(); 
    ConcurrentDictionary<string, CompiledCacheQuery> _compiledQueryCache = new ConcurrentDictionary<string, CompiledCacheQuery>();
    ObjectCache<EntityKey, EntityRecord> _sparseCache;

    private static EntityRecord[] _empty = new EntityRecord[] { };
    ITimeService _timeService;
    IAppLogService _appLogService;

    public EntityCache(CacheDataStore cacheDataStore) {
      CacheDataStore = cacheDataStore;
      if (EntityCache.CloneEntityMethod == null)
        InitStaticFields(); 
      Settings = CacheDataStore.Settings;
    }

    public void Init(EntityStore entityStore) {
      _entityStore = entityStore; 
      var model = entityStore.Model;
      _timeService = _entityStore.GetService<ITimeService>();
      _errorLog = _entityStore.GetService<IErrorLogService>();
      _appLogService = entityStore.GetService<IAppLogService>();
      foreach (var entType in Settings.SparselyCachedTypes)
        RegisterCachedEntity(entType, CachingType.Sparse);
      foreach (var entType in Settings.FullyCachedTypes)
        RegisterCachedEntity(entType, CachingType.FullSet); 
      //sparse cache
      _sparseCache = new ObjectCache<EntityKey, EntityRecord>(Settings.SparseCacheCapacity, Settings.SparseCacheExpiration, Settings.SparseCacheExpiration);
      //Create entity session
      var ctx = new UserContext("Entity-Cache", null);
      var sessionSettings = new EntitySessionSettings("EntityCache", EntitySessionOptions.ConcurrentReadonly);
      CacheSession = new EntitySession(_entityStore, ctx, sessionSettings);
      //Hook to events
      entityStore.Events.SavedChanges += Events_SavedChanges;
      //Set invalidator
      var inv = Settings.Invalidator; 
      if (inv == null && Settings.Options.IsSet(CacheOptions.UseSimpleInvalidator))
        inv = new SimpleCacheInvalidator(Settings.FullSetCacheExpiration);
      SetInvalidator(inv);
      _enabled = true;
    }

    private void RegisterCachedEntity(Type entityType, CachingType cachingType) {
      var modelInfo = _entityStore.Model.ModelInfo;
      var entInfo = modelInfo.GetEntityInfo(entityType);
      Util.Check(entInfo != null, "Type {0} specified in cache settings is not an entity.", entityType);
      entInfo.Flags |= EntityFlags.Cached;
      entInfo.CachingType = cachingType;
      if (cachingType == CachingType.FullSet) 
        FullyCachedSets[entityType] = new FullyCachedSet(entInfo);
    }

    public void Reload(bool waitForComplete = false) {
      Status = CacheStatus.NonCurrent;
      ActivateAsyncWorker();
      if (!waitForComplete) 
        return;
      var endTime = _timeService.Now.AddSeconds(30); //default, safe 
      while (_timeService.Now < endTime)
        if (Status == CacheStatus.Current)
          return;
      Util.Throw("Failed to reload cache in 30 seconds.");
    }

    #region ICacheControlService Members

    public void Invalidate(bool reload = false, bool waitForReloadComplete = false) {
      Status = CacheStatus.NonCurrent;
      if (reload) 
        Reload(waitForReloadComplete);
    }

    public CacheStatus Status {
      get { return (CacheStatus)_status; }
      set { _status = (int)value; }
    } int _status; //used in Interlocked operations, which does not support enums - so we use int

    /// <summary>True if status is current and no items in update queue.</summary>
    public bool Ready {
      get {
        return Status == CacheStatus.Current && _updateQueue.Count == 0;
      } 
    }

    public bool CacheEnabled {
      get { return _enabled; }
      set {
        var old = _enabled;
        _enabled = value;
        if (!old && _enabled) 
          //if we are enabling, then invalidate
          Invalidate();
      }
    }  

    public DateTime LastReloadedOn { get; protected set; }

    public CacheSettings GetSettings() {
      return this.Settings;
    }
    #endregion

    #region Logging
    private long CurrentTickCount {
      get {return  _timeService.ElapsedMilliseconds;}
    }

    private void AddLogEntry(EntitySession session, LogEntry entry) {
      var appLog = _appLogService.Log;
      if (appLog != null)
        appLog.AddEntry(entry);
      var userLog = session.UserContext.Log;
      if (userLog != null)
        userLog.AddEntry(entry);
    }
    #endregion
  
  }
}
