﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;
using Vita.Entities;
using Vita.Entities.Runtime;

namespace Vita.Data.Caching.Implementation {

  public partial class EntityCache {
    public static MethodInfo CloneEntityMethod;
    public static MethodInfo CloneEntitiesMethod;
    public static MethodInfo EntitiesEqualMethod;
    public static MethodInfo EntitiesNotEqualMethod;


    private static void InitStaticFields() {
      var thisType = typeof(EntityCache); 
      var flags = BindingFlags.Static | BindingFlags.Public;
      CloneEntityMethod = thisType.GetMethod("CloneEntity",  flags);
      CloneEntitiesMethod = thisType.GetMethod("CloneEntities", flags);
      EntitiesEqualMethod = thisType.GetMethod("EntitiesEqual", flags);
      EntitiesNotEqualMethod = thisType.GetMethod("EntitiesNotEqual", flags);
    }


    //Static methods used in rewritten queries againts cache. Calls to these methods is added to query expressions
    // to automatically clone entities coming from cache.
    // Note: doing iterator here (with yield return) does not work here - errors out for queries returning lists of lists
    public static IList<TEntity> CloneEntities<TEntity>(EntitySession session, IEnumerable<TEntity> entities) {
      var result = new List<TEntity>();
      foreach (TEntity entity in entities) {
        var ent = CloneEntity(session, entity);
        if (ent != null)
          result.Add(ent);
      }
      return result;
    }

    public static TEntity CloneEntity<TEntity>(EntitySession session, TEntity entity) {
      if (entity == null)
        return default(TEntity);
      var rec = EntityHelper.GetRecord(entity);
      var clone = session.CloneAndAttach(rec);
      clone.CachedRecord = rec; 
      return (TEntity)(object)clone.EntityInstance;
    }

    // Used in comparison operations in dynamic LINQ queries executed in cache. Accounts for the fact that entities might be attached to different sessions.
    // The situation comes up when we use an entity instance as parameter in dynamic query: 'var msBooks = session.EntitySet<IBook>().Where(b => b.Publisher == msPub);'
    // In this query 'msPub' is a local variable, an entity attached to local session. During execution it is matched against entities attached to cache session,
    // so by default '==' operator would not work, as MS publisher entity is a different object in both session. To make it work, the equal and not-equal operators
    // in LINQ queries are replaced with a call to EntitiesEqual, which compares primary keys if entities belong to different sessions.
    public static bool EntitiesEqual(object x, object y) {
      if (x == y) return true;
      if (x == null) return y == null;
      if (y == null) return x == null; 
      // both are not nulls, and not the same object; there's a chance they are identical entities from different sessions
      var rx = EntityHelper.GetRecord(x);
      var ry = EntityHelper.GetRecord(y);
      if (rx.Session == ry.Session)
        return rx == ry; //compare as objects
      // Sessions are different - compare entity types and primary keys
      if (rx.EntityInfo != ry.EntityInfo) //should never happen, but just in case 
        return false; 
      var keysEqual = rx.PrimaryKey.Equals(ry.PrimaryKey);
      return keysEqual; 
    }

    public static bool EntitiesNotEqual(object x, object y) {
      return !EntitiesEqual(x, y);
    }

  }
}
