﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

using Vita.Common;
using Vita.Entities;
using Vita.Entities.Runtime;
using Vita.Services;

namespace Vita.Data.Caching.Implementation {


  // Loading, invalidation methods and properties
  public partial class EntityCache {
    ICacheInvalidator _invalidator;
    Exception _loadError; 

    public void SetInvalidator(ICacheInvalidator invalidator) {
      if (_invalidator != null)
        _invalidator.Invalidated -= CacheInvalidator_Invalidated;
      _invalidator = invalidator;
      if (_invalidator != null)
        _invalidator.Invalidated += CacheInvalidator_Invalidated;
    }

    void CacheInvalidator_Invalidated(object sender, CacheInvalidateEventArgs e) {
      if (e.IsTimeout) {
        var expires = LastReloadedOn.Add(Settings.FullSetCacheExpiration);
        if (expires > _timeService.Now)
          return; //if it was reloaded recently enough, ignore timout
      }
      Invalidate();
    }

    private CacheStatus SetStatusInterlocked(CacheStatus newValue) {
      return (CacheStatus)Interlocked.Exchange(ref _status, (int)newValue);
    }

    public void ReloadCache() {
      try {
        Lock.EnterWriteLock();
        // Repeat until we have clean run with status == Loading at the end. 
        // Some user session could invalidate cache while we load it and set status from Loading to NonCurrent.
        // Note that cache invalidation from update operations to cache entities is done without lock. Updating entity in cache enqueues cache update request 
        // and invalidates the cache (sets _status to NonCurrent) - all this without lock, to avoid 
        CacheStatus endStatus; 
        do {
          Status = CacheStatus.Loading;
          ReloadCacheLocked();
          // Try setting status to Current and check current (old value)
          // carefully read the current value with Interlock. If current value is not Loading, cache was invalidated again.
          // we are inside the lock here, but in other places we set status without lock (in Events_Saved method for ex)
          endStatus = SetStatusInterlocked(CacheStatus.Current); 
        } while (endStatus != CacheStatus.Loading);
      }catch(Exception ex) {
        _loadError = ex;
        var logContents = CacheSession.GetUserLogContents(); 
        _errorLog.LogError(ex, logContents);
        Status = CacheStatus.NonCurrent;
      } finally {
        this.LastReloadedOn = _timeService.Now;
        Lock.ExitWriteLock();
      }
    }

    private void ReloadCacheLocked() {
      var appLog = _appLogService.Log;
      if (appLog != null) 
        appLog.Info("-- Starting EntityCache reload...");
      // clear update request queue - we're reloading cache completely. We also clear sparse cache
      ClearUpdateQueue();
      _sparseCache.Clear();
      var realDataStore = this.CacheDataStore.RealDataStore;
      foreach (var entSet in FullyCachedSets.Values) {
        //if Status was set back to non-current while we were reloading, abandon this reload - we need to start from scratch
        if (Status == CacheStatus.NonCurrent)
          return; 
        var fullyCachedSet = (FullyCachedSet)entSet;
        var cmd = entSet.EntityInfo.CrudCommands.SelectAll;
        var records = this.CacheDataStore.RealDataStore.ExecuteQuery(CacheSession, cmd, Paging.NoPaging, null);
        fullyCachedSet.SetRecords(records);
      }
      if (appLog != null)
        appLog.Info("-- EntityCache reload completed.");
    }

  }//class

}
