﻿using System;
using System.Collections;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading;

using Vita.Common;
using Vita.Entities;
using Vita.Entities.Model;
using Vita.Entities.Runtime;
using Vita.Data.Linq;
using Vita.Logging;
using Vita.Services;

namespace Vita.Data.Caching.Implementation {

  // Cache querying
  public partial class EntityCache {
    public bool CanQuery(EntitySession session) {
      if (!_enabled) 
        return false;
      var varErr = _loadError;
      _loadError = null;
      if (varErr != null)
        throw new Exception("EntityCache load error: " + varErr.Message, varErr);
      //TODO: check required version
      if (Ready)
        return true; 
      //start reloading 
      if (Status == CacheStatus.NonCurrent)
        ActivateAsyncWorker(); 
      return false;
    }

    private bool CheckFullSetCacheExpired() {
      var status = this.Status;
      if (status == CacheStatus.Current) {
        //check expiration
        var lastReloaded = LastReloadedOn;
        if (lastReloaded.Add(Settings.FullSetCacheExpiration) < _timeService.Now)
          Status = CacheStatus.NonCurrent;
      }
      return Status == CacheStatus.Current;
    }

    public bool TryExecuteQuery(EntitySession session, IEntityCommand command, Paging paging, object[] args, out IList<EntityRecord> records) {
      records = null;
      //quick check, without lock
      if (!CanQuery(session)) 
        return false;
      bool result = false;
      var entCommand = (EntityCommand)command;
      if (!entCommand.TargetEntityInfo.Flags.IsSet(EntityFlags.Cached))
        return false;
      var start = CurrentTickCount;
      try {
        Lock.EnterReadLock();
        if (Status != CacheStatus.Current) return false; //check again under lock
        if (session.UserContext.RequireCacheVersion > _currentCacheVersion) //check version after getting lock
          return false; //most likely refresh is scheduled and about to execute
        result = TryExecuteQueryLocked(session, entCommand, paging, args, out records);
        if (result)
          session.UserContext.RequireCacheVersion = _currentCacheVersion;
      } finally {
        Lock.ExitReadLock();
      }
      //Log
      if (result) {
        var rowCount = records == null ? 0 : records.Count;
        var cachingType = entCommand.TargetEntityInfo.CachingType;
        AddLogEntry(session, new CacheCommandLogEntry(command.CommandName, args, paging, _timeService.Now, CurrentTickCount - start, rowCount, cachingType));
      }
      return result; 
    }


    private bool TryExecuteQueryLocked(EntitySession session, EntityCommand command, Paging paging, object[] args, out IList<EntityRecord> records) {
      records = _empty;
      var cachingType = command.TargetEntityInfo.CachingType;
      if (cachingType == CachingType.None)
        return false;
      if (cachingType == CachingType.FullSet && !CheckFullSetCacheExpired())
        return false;
      var entType = command.TargetEntityInfo.EntityType;
      FullyCachedSet fullSet;
      switch (command.Kind) {
        case EntityCommandKind.SelectAll:
        case EntityCommandKind.SelectAllPaged:
          if (cachingType != CachingType.FullSet)
            return false;
          if (!CheckFullSetCacheExpired())
            return false; 
          fullSet = GetFullyCachedSet(entType);
          var recs = fullSet.Records.ToList();
          if (paging == null || (paging.Skip == 0 && paging.Take >= fullSet.RecordsByPrimaryKey.Count))
            records = recs;
          else
            records = recs.Skip(paging.Skip).Take(paging.Take).ToList();
          return true; 

        case EntityCommandKind.SelectByKey:
          if (command.SelectKey.KeyType.IsSet(KeyType.PrimaryKey)) {
            var pk = new EntityKey(command.SelectKey, args);
            EntityRecord rec = null;
            switch(cachingType) {
              case CachingType.FullSet:
                fullSet = GetFullyCachedSet(entType);
                rec = fullSet.LookupByPrimaryKey(pk);
                break; 
              case CachingType.Sparse:
                rec = _sparseCache.Lookup(pk);
                break; 
            }//switch
            if (rec != null) {
              records = new EntityRecord[] { rec };
              return true;
            } else {
              records = _empty; 
              return false; // not found, records already set to empty array
            }
          }//if PrimaryKey 
          //foreign key
          if (cachingType != CachingType.FullSet)
            return false; 
          fullSet = GetFullyCachedSet(entType);
          records = fullSet.SelectByForeignKey(command.SelectKey, args);
          return true;

        case EntityCommandKind.SelectByKeyManyToMany:
          // TODO: SelectBykeyManyToMany should not be called for cache entities (but is called currently) - fix it in PropertyBoundListManyToMany
          return false; 
        default:
          return false; //never happens
      }//switch
    }

    public bool TryExecuteDynamicQuery(EntitySession session, EntityQuery query, out object result) {
      result = null;
      //quick check, without lock
      if (!CanQuery(session))
        return false;
      if (!CheckFullSetCacheExpired())
        return false; 
      var queryInfo = EntityQueryRuntimeInfo.GetQueryInfo(query); //returns already constructed object
      foreach (var ent in queryInfo.Entities)
        if (ent.CachingType != CachingType.FullSet)
          return false; 
      var start = CurrentTickCount;
      AssignCompiledQuery(queryInfo);
      queryInfo.ParamValues.Add(session);
      var prmValues = queryInfo.ParamValues.ToArray();
      try {
        // actually execute with read lock
        Lock.EnterReadLock();
        // check again under lock
        if (Status != CacheStatus.Current) return false; 
        if (session.UserContext.RequireCacheVersion > _currentCacheVersion) //check version
          return false; //most likely refresh is scheduled and about to execute
        result = queryInfo.EntityCacheQuery.CompiledDelegate.DynamicInvoke(prmValues);
        session.UserContext.RequireCacheVersion = _currentCacheVersion;
        //log; first clear session reference, so it does not hang out there
        prmValues[prmValues.Length - 1] = null; 
        AddLogEntry(session, new CacheQueryLogEntry(queryInfo.Lambda, prmValues, _timeService.Now, CurrentTickCount - start, GetRowCount(result)));
        return true;
      } finally {
        Lock.ExitReadLock(); 
      }
    }

    private int GetRowCount(object queryResult) {
      if (queryResult == null)
        return 0;
      var rList = queryResult as IList;
      return rList == null ? 1 : rList.Count;

    }

    public FullyCachedSet GetFullyCachedSet(Type entityType) {
      FullyCachedSet entSet;
      if (FullyCachedSets.TryGetValue(entityType, out entSet))
        return entSet;
      return null;
    }

    #region private methods
    private void AssignCompiledQuery(EntityQueryRuntimeInfo queryInfo) {
      if (_compiledQueryCache.TryGetValue(queryInfo.CacheKey, out queryInfo.EntityCacheQuery))
        return; 
      var rewriter = new CacheQueryRewriter();
      var rewritten = rewriter.Rewrite(queryInfo.Lambda, this);
      var compiledCacheLambda = rewritten.Compile();
      var cacheQuery = queryInfo.EntityCacheQuery = new CompiledCacheQuery(queryInfo.CacheKey, queryInfo.Lambda, compiledCacheLambda);
      _compiledQueryCache[cacheQuery.Key] = cacheQuery;
    }

    //  Adding to cache - we add only sparse cached records from queries, and ignore fully-cached entities
    internal void CacheDbQueryResults(EntityCommand command, IList<EntityRecord> records) {
      if (records == null || records.Count == 0)
        return;
      if (command.TargetEntityInfo.CachingType != CachingType.Sparse)
        return;
      switch (command.Kind) {
        case EntityCommandKind.SelectAll:
        case EntityCommandKind.SelectAllPaged:
        case EntityCommandKind.SelectByKey:
        case EntityCommandKind.SelectByKeyManyToMany:
          //synchronuous, no-lock adding to record list
          var cacheRecords = CacheSession.CloneAndAttach(records, forceFullCopy: true);
          foreach (var cacheRec in cacheRecords) {
            cacheRec.CachedRecord = null; //CloneAndAttach had set this prop to source record - clear it, cached records do not reference anything
            _sparseCache.Add(cacheRec.PrimaryKey, cacheRec);
          }
          break;
      }
    }
    #endregion

  }
}
