﻿using System;
using System.Collections.Generic;
using System.Collections.Concurrent;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using Vita.Common;
using Vita.Entities;
using Vita.Entities.Model;
using Vita.Entities.Runtime;

namespace Vita.Data.Caching.Implementation {

  /*
   Cache updates.
     Whenever an application updates an entity which is cached, the cache update process is started. 
     1. We start the update process when changes are committed and finalized in user session. Entity cache hooks to Saved event of the entity store 
        (which is fired after finalizing update), and then processes all updates. 
     2. We need to make sure that updates (if more than one record) become visible to other sessions in transactional manner - all at once.
        We need to use WriteLock for this; to avoid waiting times, we do actual update asynchonuously (unless AsyncUpdate option in CacheSettings is off).
        We pack all updates into an update pack, push them into the update request queue, and then activate background thread. The background thread processes all queued updates.
     3. Because of async nature of update, we must ensure that after cache update the app code will see the updated cache. We use cache version value for this.
        Whenever update happens, we increment next_version value (future version) and save it in UserContext. This future version becomes current when update is completed.
        Each time we try to access the cache we check that the current cache version is equal or greater than the required cache version in UserContext.
     4. For multi-domain server environments we use global signaler (WaitHandle) inside SimpleEntityCacheInvalidator to signal other domains that the cached entities 
        had been updated so they need to reload cache. 
    
   */ 

  class CachedRecordUpdate {
    public EntityStatus StatusBeforeSave;  
    public EntityKey PrimaryKey;
    public object[] NewValues;
  }
  
  class CacheUpdateRequest {
    public long NewCacheVersion;
    public readonly List<CachedRecordUpdate> Updates = new List<CachedRecordUpdate>();
  }


  public partial class EntityCache {
    ConcurrentQueue<CacheUpdateRequest> _updateQueue = new ConcurrentQueue<CacheUpdateRequest>();
    ManualResetEvent _asyncWorkSignal = new ManualResetEvent(false);
    // Two incrementing versions. Current identifies the 'current' version, _next is the version that it will become after completing all pending updates
    // We do updates async-ly; so after update we increment the next version and save it in user context - next time this user comes to cache, we'd require 
    // that current cache version is at least at this version. 
    long _currentCacheVersion;
    long _nextCacheVersion;


    void Events_SavedChanges(EntitySession session, EventArgs args) {
      if (!_enabled)
        return; 
      if (!AnyCached(session))
        return; 
      // we cannot update records immediately in the cache - other operations (queries, updates) might be running. 
      // We collect all updates and apply them on a background thread with WriteLock. 
      var updateRequest = new CacheUpdateRequest();
      foreach (var rec in session.RecordsChanged) {
        if (rec.EntityInfo.CachingType != CachingType.None)
          updateRequest.Updates.Add(new CachedRecordUpdate() { StatusBeforeSave = rec.StatusBeforeSave, PrimaryKey = rec.PrimaryKey, NewValues = rec.ValuesOriginal });
      }
      if (updateRequest.Updates.Count == 0)
        return;
      // Save incremented cache version in UserContext. We update cache asynchronously, so that original thread does not wait (waiting for write lock might take some time).
      // But the user on the next operation must see the updated cache. To ensure this, we save incremented cache version in user context, and check it when user tries 
      // to read the cache next time. 
      var nextVersion = Interlocked.Increment(ref _nextCacheVersion);
      session.UserContext.RequireCacheVersion = updateRequest.NewCacheVersion = nextVersion;
      _updateQueue.Enqueue(updateRequest);
      // fire signal to other domains that cache was updated
      if (_invalidator != null)
        _invalidator.SignalInvalidate();
      if (Settings.Options.IsSet(CacheOptions.AsyncUpdate)) {
        ActivateAsyncWorker();
      } else {
        //do it synchonously
        ProcessCacheUpdateQueue(); 
      }
    }//method

    private bool AnyCached(EntitySession session) {
      foreach(var ent in session.EntityInfosForRecordsChanged)
        if (ent.CachingType != CachingType.None)
          return true; 
      return false; 
    }

    private void ProcessCacheUpdateQueue() {
      if (_updateQueue.Count == 0)
        return; 
      try {
        Lock.EnterWriteLock();
        while (_updateQueue.Count > 0) {
          ProcessCacheUpdateQueueLocked(); 
        }
      } catch(Exception ex) {
        if (_errorLog != null)
          _errorLog.LogError(ex, null, Environment.UserName, "CacheService", Environment.MachineName);
        throw; 
      } finally {
        Lock.ExitWriteLock();
      }
    }

    private void ProcessCacheUpdateQueueLocked() {
      long newVersion = _currentCacheVersion;
      var updList = GetAllPendingUpdateRequests();
      var entTypes = new HashSet<Type>(); //set of all fully cached entity types involved
      foreach (var updateReq in updList) {
        newVersion = Math.Max(newVersion, updateReq.NewCacheVersion);
        //Process individual updates
        foreach (var update in updateReq.Updates) {
          var entInfo = update.PrimaryKey.KeyInfo.Entity;
          switch (entInfo.CachingType) {
            case CachingType.FullSet:
              ApplyFullyCachedEntityUpdate(update);
              entTypes.Add(entInfo.EntityType); //remember entity type to refresh lists later
              break;
            case CachingType.Sparse:
              ApplySparseCachedEntityUpdate(update);
              break;
          }//switch
        }//foreach update
      }//foreach updateReq
      //Refresh lists in all full sets
      foreach (var entType in entTypes) {
        var entSet = GetFullyCachedSet(entType);
        entSet.RefreshLists();
      }
      // set new version
      _currentCacheVersion = newVersion;
    }

    private void ApplyFullyCachedEntityUpdate(CachedRecordUpdate update) {
      var entSet = GetFullyCachedSet(update.PrimaryKey.KeyInfo.Entity.EntityType);
      EntityRecord cachedRecord = null;
      switch (update.StatusBeforeSave) {
        case EntityStatus.Deleting:
          if (entSet.RecordsByPrimaryKey.TryGetValue(update.PrimaryKey, out cachedRecord))
            entSet.RecordsByPrimaryKey.Remove(update.PrimaryKey);
          break;
        case EntityStatus.Modified:
          if (entSet.RecordsByPrimaryKey.TryGetValue(update.PrimaryKey, out cachedRecord)) 
            cachedRecord.ValuesOriginal = update.NewValues;
          break;
        case EntityStatus.New:
          cachedRecord = new EntityRecord(entSet.EntityInfo, EntityStatus.Loaded);
          cachedRecord.ValuesOriginal = update.NewValues;
          cachedRecord = CacheSession.Attach(cachedRecord);
          entSet.RecordsByPrimaryKey[cachedRecord.PrimaryKey] = cachedRecord;
          break;
      }//switch
      if (cachedRecord != null)
        ClearRelatedLists(cachedRecord); 
    }

    private void ApplySparseCachedEntityUpdate(CachedRecordUpdate update) {
      EntityRecord cachedRecord = null;
      switch (update.StatusBeforeSave) {
        case EntityStatus.Deleting:
          cachedRecord = this._sparseCache.Lookup(update.PrimaryKey);
          if (cachedRecord != null)
            _sparseCache.TryRemove(update.PrimaryKey);
          break;
        case EntityStatus.Modified:
          cachedRecord = this._sparseCache.Lookup(update.PrimaryKey);
          if (cachedRecord == null)
            return;
          cachedRecord.ValuesOriginal = update.NewValues; 
          break;
        case EntityStatus.New:
          var entInfo = update.PrimaryKey.KeyInfo.Entity;
          cachedRecord = new EntityRecord(entInfo, EntityStatus.Loaded);
          cachedRecord.ValuesOriginal = update.NewValues;
          cachedRecord = CacheSession.Attach(cachedRecord);
          _sparseCache.Add(cachedRecord.PrimaryKey, cachedRecord); 
          break;
      }//switch UpdateType
    }// method

    private void ClearUpdateQueue() {
      //just get all pending updates from the queue
      GetAllPendingUpdateRequests();
    }

    private IList<CacheUpdateRequest> GetAllPendingUpdateRequests() {
      var result = new List<CacheUpdateRequest>();
      while (true) {
        CacheUpdateRequest temp;
        if (!_updateQueue.TryDequeue(out temp)) return result;
        result.Add(temp);
      }
    }


    //Whenever we update/delete a record, it might change the related list. Ex: Deleting a book might change the Publisher.Books list. 
    // So we need to 'reset' this list by clearing out all transient values in book.Publisher entity.
    private void ClearRelatedLists(EntityRecord record) {
      foreach (var refMember in record.EntityInfo.RefMembers) {
        var targetEnt = record.GetValue(refMember);
        if (targetEnt == null) continue;
        var rec = EntityHelper.GetRecord(targetEnt);
        rec.ClearTransientValues(); 
      }
    }

  }//class

}
