﻿using System;
using System.Collections;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;
using Vita.Common;
using Vita.Entities.Model;
using Vita.Entities.Runtime;

namespace Vita.Data.Caching.Implementation {

  public class FullyCachedSet {
    // Used in LINQ queries against cache
    public static MethodInfo GetEntityListMethod; 

    public EntityInfo EntityInfo;
    public IDictionary<EntityKey, EntityRecord> RecordsByPrimaryKey;
    public List<EntityRecord> Records;
    public IList Entities; 

    public FullyCachedSet(EntityInfo entityInfo) {
      // take care of static field
      if (GetEntityListMethod == null)
        GetEntityListMethod = this.GetType().GetMethod("GetEntityList", BindingFlags.Instance | BindingFlags.Public);
      EntityInfo = entityInfo;
      Records = new List<EntityRecord>(); 
      RecordsByPrimaryKey = new Dictionary<EntityKey,EntityRecord>();
      var listType = typeof(List<>).MakeGenericType(EntityInfo.EntityType);
      Entities = Activator.CreateInstance(listType) as IList;
    }

    //Note: this method is referenced by GetEntityListMethod static property of this class, which is in turn used by LINQ queries against cache;
    // be careful changing signature of this method.
    public IList<T> GetEntityList<T>() {
      return (IList<T>)Entities;
    }

    public IList<EntityRecord> SelectByForeignKey(EntityKeyInfo keyInfo, object[] keyValues) {
      var result = this.Records.Where(rec => rec.KeyMatches(keyInfo, keyValues)).ToList();
      return result;
    }
    public EntityRecord LookupByPrimaryKey(EntityKey primaryKey) {
      EntityRecord rec;
      if (RecordsByPrimaryKey.TryGetValue(primaryKey, out rec))
        return rec;
      return null; 
    }

    // assumes records are sorted in proper order
    public void SetRecords(IList<EntityRecord> records) {
      RecordsByPrimaryKey.Clear();
      foreach (var rec in records) 
        RecordsByPrimaryKey[rec.PrimaryKey] = rec;
      RefreshLists(); 
    }

    // Refreshes Records and Entities lists from RecordsByPrimaryKey dictionary.
    // RecordsByPrimaryKey is a primary store of entities/records; cache updates first change this dictionary, then call RefreshLists to repopulate lists 
    // This is necessary, because lists must hold entities in default sort order. If a property in entity changes, it may change the entity's position in 
    // the cached list. 
    public void RefreshLists() {
      var recs = RecordsByPrimaryKey.Values.ToList();
      if (EntityInfo.DefaultOrderBy != null)
        recs.Sort(EntityInfo.DefaultCompare);
      Records.Clear();
      Entities.Clear();
      foreach (var rec in recs) {
        Records.Add(rec);
        Entities.Add(rec.EntityInstance);
      }
    }

  } //class

}
