﻿using System;
using System.Collections.Generic;
using System.Data;
using System.Data.Common;
using System.Linq.Expressions;
using System.Linq;
using System.Collections;
using System.Diagnostics;

using Vita.Common;
using Vita.Entities;
using Vita.Entities.Runtime;
using Vita.Entities.Model;
using Vita.Logging;

using Vita.Data.Model;
using Vita.Data.Driver;
using Vita.Data.Management;
using Vita.Data.Caching;
using Vita.Data.Linq.Translation;
using Vita.Data.Linq;
using Vita.Services;
using Vita.Activation;
using Vita.Data.Model.Construction;

namespace Vita.Data {

  // Implementation
  public partial class Database : IDataStore {
    public readonly string Alias; 
    /// <summary>Current database model. </summary>
    public DbModel DbModel { get; private set; }
    public readonly DbSettings Settings;

    /// <summary>Connection string.</summary>
    public string ConnectionString { get { return Settings.ConnectionString; } }

    EntityModel _entityModel;
    EntityStore _entityStore; 
    private DbDriver _driver;
    private ITimeService _timeService;
    private IAppLogService _appLogService;

    // LINQ
    private ObjectCache<string, SqlQuery> _queryCache;

    // constructors
    public Database(EntityModel entityModel, DbSettings settings, string alias = null) {
      _entityModel = entityModel; 
      Settings = settings;
      Alias = alias;
      if (string.IsNullOrEmpty(Alias) && _entityModel != null)
        Alias = _entityModel.ModelName;
      _driver = Settings.Driver;
    }
    public Database(DbModel dbModel, string alias = null) {
      DbModel = dbModel;
      _entityModel = DbModel.EntityModel;
      Settings = DbModel.Settings;
      Alias = alias;
      if (string.IsNullOrEmpty(Alias) && _entityModel != null)
        Alias = _entityModel.ModelName;
      _driver = Settings.Driver;
    }

    #region IDataStore Members
    public void Init(EntityStore entityStore) {
      _entityStore = entityStore; 
      _timeService = entityStore.GetService<ITimeService>();
      _appLogService = entityStore.GetService<IAppLogService>();
      _queryCache = new ObjectCache<string, SqlQuery>(capacity: 10000, expirationTime: TimeSpan.FromMinutes(10), cleanupInterval: TimeSpan.FromMinutes(1));
      var dataStoreListService = _entityStore.GetService<IDataStoreListingService>();
      if (dataStoreListService != null) //in fact, should always be there.
        dataStoreListService.RegisterDataStore(this);
      _entityStore.Events.Initializing += Events_Initializing;
    }

    void Events_Initializing(EntityStore entityStore, EntityStoreInitEventArgs args) {
      switch(args.InitStep) {
        case EntityStoreInitStep.DbModelConstructing:
          DbModel = DbModelBuilder.Build(_entityModel, Settings, args.ActivationContext.Log);
          break;
        case EntityStoreInitStep.DbSchemaUpdating:
          UpdateDbSchema(args.ActivationContext);
          break; 
      }
    }

    public void UpdateDbSchema(AppActivationContext activationContext) {
      var updateOptions = Settings.SchemaUpdateOptions;
      if (updateOptions.IsSet(DbSchemaUpdateOptions.UpdateSchema | DbSchemaUpdateOptions.ThrowOnSchemaMismatch)) {
        var log = activationContext.Log;
        var updater = new DbSchemaUpdater(this, log);
        var updates = updater.GetSchemaUpdates();
        if (updates.Actions.Count > 0) {
          //Log changes
          var strUpdates = updates.GetAllAsText();
          log.Info("Schema changes detected. ================================================================");
          log.Info(strUpdates);
          log.Info("End schema changes list. ================================================================");
          if (Settings.SchemaUpdateOptions.IsSet(DbSchemaUpdateOptions.UpdateSchema))
            updater.ApplySchemaUpdates(updates);
          else if (Settings.SchemaUpdateOptions.IsSet(DbSchemaUpdateOptions.ThrowOnSchemaMismatch)) {
            throw new ActivationException("Schema changes detected. Schema update settings set to throw exception on schema changes.", log.GetAllAsText());
          }
        }// if Count > 0
      }
      if (!DbModel.CustomCommandsCompiled)
        DbModel.CompileCustomCommands(this.ConnectionString);
    }

    public IEnumerable<string> GetSchemas() {
      return DbModel.Schemas;
    }

    public IList<EntityRecord> ExecuteQuery(EntitySession session, IEntityCommand command, Paging paging, object[] args){
      var entCommand = command as EntityCommand;
      Util.Check(entCommand != null, "Cannot execute command {0} - cast to EntityCommand failed.", command.CommandName);
      Util.Check(DbModel.Schemas.Contains(entCommand.Area.SchemaName), "Cannot execute query command {0} - command Area {1} is not part of DB model.", 
                                                   entCommand.CommandName, entCommand.Area.SchemaName);
      EntityInfo entInfo = entCommand.TargetEntityInfo;
      var dbCommandInfo = GetDbCommandInfo(entCommand);
      if (dbCommandInfo == null) {
        //check if it is request for SelectAllPaged and driver does not support paging - then substitute with SelecAll
        if (entCommand == entInfo.CrudCommands.SelectAllPaged)
          dbCommandInfo = GetDbCommandInfo(entInfo.CrudCommands.SelectAll);
      }
      if (dbCommandInfo == null)
        Util.Throw("Db command not found for entity command " + entCommand.CommandName);

      if (dbCommandInfo.Flags.IsSet(DbCommandFlags.WithPaging))
        args = AddPagingArguments(args, paging);
      List<EntityRecord> records; 
      using (var connection = OpenConnection(session)) {
        records = (List<EntityRecord>) ExecuteCommand(connection, dbCommandInfo, args, paging); 
      }

      for (int i = 0; i < records.Count; i++) 
        records[i] = session.Attach(records[i]);
      return records;
    }

    public object ExecuteNonQuery(EntitySession session, IEntityCommand command, object[] args) {
      var entCommand = (EntityCommand)command;
      if (!DbModel.Schemas.Contains(entCommand.Area.SchemaName))
        return null;
      using (var conn = OpenConnection(session)) {
        var dbCommand = GetDbCommandInfo(entCommand);
        return ExecuteCommand(conn, dbCommand, args);
      }
    }

    public void SaveChanges(EntitySession session) {
      //Create batch and invoke Validate, Saving events
      var batch = new EntityUpdateBatch(session, _timeService.Now);
      //check if we need to sequence records
      var useRefIntegrity = DbModel.Settings.Options.IsSet(DbOptions.UseRefIntegrity);
      var canDeferIntegrCheck = DbModel.Driver.Supports(DbCapabilities.DeferredConstraintCheck);
      var needSequencing = useRefIntegrity && !canDeferIntegrCheck;
      if (needSequencing && !batch.IsSequenced)
        if (!batch.IsSequenced) {
          batch.SequenceRecords();
        }
      //actually update
      var records = batch.Session.RecordsChanged;
      var commands = batch.Session.ScheduledCommands;
      //Start explicit transaction only if we have more than one command to execute.
      //TODO: see when and how to use TransactionMode.Context (with transaction spanning multiple connections)
      var transMode = records.Count + commands.Count > 1 ? TransactionMode.Connection : TransactionMode.None;
      if (transMode == TransactionMode.Connection)
        LogComment(batch.Session, "\r\nBEGIN TRANSACTION;");
      var conn = OpenConnection(batch.Session, transMode);
      var start = CurrentTickCount;
      try {
        //execute commands
        if (commands.Count > 0)
          ExecuteScheduledCommands(conn, commands, before: true);

        var recList = batch.IsolatedDeletes;
        for (int i = 0; i < recList.Count; i++)
          ApplyUpdate(conn, recList[i]);

        recList = batch.Inserts;
        for (int i = 0; i < recList.Count; i++)
          ApplyUpdate(conn, recList[i]);

        recList = batch.Updates;
        for (int i = 0; i < recList.Count; i++)
          ApplyUpdate(conn, recList[i]);

        recList = batch.Deletes;
        for (int i = 0; i < recList.Count; i++)
          ApplyUpdate(conn, recList[i]);

        recList = batch.IsolatedInserts;
        for (int i = 0; i < recList.Count; i++)
          ApplyUpdate(conn, recList[i]);

        if (commands.Count > 0)
          ExecuteScheduledCommands(conn, commands, before: false);
        if (transMode == TransactionMode.Connection)
          LogComment(batch.Session, "COMMIT TRANSACTION; --  time: {0}ms \r\n", CurrentTickCount - start);
        conn.Commit();
        conn.Close();
      } catch {
        conn.Abort();
        throw;
      } //catch
    }

    public bool CheckConnectivity() {
      IDbConnection conn = null;
      try {
        conn = this.DbModel.Driver.CreateConnection(this.ConnectionString);
        conn.Open(); 
        return true;
      } catch(Exception) {
        return false; 
      } finally {
        if (conn != null)
          conn.Close();
      }
    }

    #endregion IDataStore interface

    // Private methods
    private void ExecuteScheduledCommands(DatabaseConnection connection, IList<ScheduledEntityCommand> commands, bool before) {
      for (int i = 0; i < commands.Count; i++) {
        var scheduledCmd = commands[i];
        var cmdBefore = (scheduledCmd.Options & ExecuteOptions.Before) != 0;
        if (cmdBefore != before) continue;
        var dbCommand = GetDbCommandInfo(scheduledCmd.Command);
        ExecuteCommand(connection, dbCommand, scheduledCmd.Args);
      }
    }

    private DatabaseConnection OpenConnection(EntitySession session, TransactionMode mode = TransactionMode.None) {
      return new DatabaseConnection(session, this, mode);
    }


    private void ApplyUpdate(DatabaseConnection connection, EntityRecord record) {
      //Check record's area is in this database
      var area = record.EntityInfo.Area;
      if (!this.DbModel.Schemas.Contains(area.SchemaName))
        return;
      if (record.Status == EntityStatus.Modified && record.EntityInfo.Flags.IsSet(EntityFlags.NoUpdate))
        return; //if for whatever reason we have such a record, just ignore it
      var crud = record.EntityInfo.CrudCommands;
      EntityCommand cmd;
      switch (record.Status) {
        case EntityStatus.New: cmd = crud.Insert; break;
        case EntityStatus.Modified: cmd = crud.Update; break;
        case EntityStatus.Deleting: cmd = crud.Delete; break;
        default:
          return;
      }
      var cmdInfo = GetDbCommandInfo(cmd);
      Util.Check(cmdInfo != null, "No DB command defined for entity command {0}. Database: {1}", cmd, this.Alias);
      this.ExecuteCrudCommand(connection, cmdInfo, record);
      record.SubmitCount++;
      record.EntityInfo.SaveEvents.OnSubmittedChanges(record);
    }

    private object ExecuteCommand(DatabaseConnection connection, DbCommandInfo commandInfo, object[] args, Paging paging = null) {
      var cmd = CreateDbCommand(commandInfo, connection);
      SetCommandParameterValues(commandInfo, cmd, args);
      var result = ExecuteWithLog(connection, cmd, commandInfo, paging);
      if (commandInfo.OutputParameters.Count > 0)
        ReadOutputParameterValues(cmd, commandInfo, args);
      return result;
    }

    private object ExecuteCrudCommand(DatabaseConnection connection, DbCommandInfo commandInfo, EntityRecord record) {
      try {
        var cmd = CreateDbCommand(commandInfo, connection);
        SetCrudCommandParameterValues(commandInfo, cmd, record);
        var result = ExecuteWithLog(connection, cmd, commandInfo);
        if (commandInfo.OutputParameters.Count > 0)
          ReadCrudOutputParameterValues(cmd, commandInfo, record);
        return result;
      } catch (Exception ex) {
        ex.AddValue("entity-command-name", commandInfo.EntityCommand.CommandName);
        ex.AddValue("record", record);
        throw;
      }
    }

    private IDbCommand CreateDbCommand(DbCommandInfo commandInfo, DatabaseConnection connection) {
      var  cmd = _driver.CreateDbCommand(connection.DbConnection, commandInfo);
      cmd.Connection = connection.DbConnection;
      cmd.Transaction = connection.DbTransaction;
      return cmd; 
    }

    private object ExecuteWithLog(DatabaseConnection connection, IDbCommand command, DbCommandInfo commandInfo, Paging paging = null) {
      object result;
      try {
        var start = CurrentTickCount;
        int recordCount = -1; 
        result = _driver.ExecuteCommand(command, commandInfo.ExecutionType); 
        var end = CurrentTickCount;
        if (commandInfo.ExecutionType == DbExecutionType.Query) {
          connection.ActiveReader = (IDataReader)result;
          var records = ReadRecords(connection.ActiveReader, paging, commandInfo, command);
          recordCount = records.Count;
          connection.ActiveReader.Close(); 
          result = records;
        }
        LogCommand(connection.Session, command, end - start, recordCount);
        return result;
      } catch (Exception ex) {
        ex.AddValue("command", command);
        OnException(ex, commandInfo); 
        LogCommand(connection.Session, command, 0);
        LogException(connection.Session, ex);
        if (connection.ActiveReader != null)
          connection.ActiveReader.Close(); 
        throw;
      }
    }//method

    //Inspect specific exception types and add information to exception
    private void OnException(Exception exception, DbCommandInfo commandInfo) {
      var uex = exception as UniqueIndexViolationException;
      if (uex != null) {
        //We need to add entity name and member names to the exception
        var dbKey = DbModel.FindKey(uex.DbKeyName);
        if (dbKey != null) {
          var entKey = dbKey.EntityKey;
          uex.EntityName = entKey.Entity.Name;
          uex.KeyName = entKey.Name;
          uex.MemberNames = entKey.GetMemberNames(",");
        }
        return; 
      }
      var cex = exception as ConcurrentUpdateException;
      if (cex != null) {
        var table = DbModel.GetTable(cex.TableName);
        if (table != null)
          cex.EntityName = table.Entity.Name;
      }
      var dex = exception as DbException;
      if (dex != null)
        dex.EntityCommandName = commandInfo.EntityCommand.CommandName;
    }//method


    // Sets parameter values.
    // For most commands (except LINQ-based custom commands) args directly match command Parameters. 
    // For custom procedures based on LINQ queries, some actual Db parameters are created from literal constants 
    // in the base linq query - these parameters are not listed explicitly in args. When we create DbCommand from 
    // DbCommandInfo, we initialize all parameter values from defaults in DbCommandInfo.Parameters - so extra LINQ
    // parameters/constants get their correct values. Next we need to assign (all or some) values from input command arguments.
    // We use prmInfo.ArgIndex to match arguments to command parameters. 
    private void SetCommandParameterValues(DbCommandInfo commandInfo, IDbCommand command, object[] args) {
      for (int i = 0; i < commandInfo.Parameters.Count; i++) {
        var prmInfo = commandInfo.Parameters[i];
        if (prmInfo.ArgIndex >= 0) {
          var prm = (IDbDataParameter)command.Parameters[i];
          var v = args[prmInfo.ArgIndex];
          prm.Value = prmInfo.TypeInfo.WriteConverter(v);
        }
      } //for i

    }

    private IList<EntityRecord> ReadRecords(IDataReader reader, Paging paging, DbCommandInfo commandInfo, IDbCommand command) {
      if (commandInfo.QueryColumnMappings == null)
        BuildColumnMappings(commandInfo, reader);
      var records = new List<EntityRecord>();
      //check if we need to return anything; we do it AFTER executing command on purpose - to allow command test-call without returning records
      if (paging.Take == 0)
        return records;
      //Check if we need to do paging of the results in code
      bool doResultPaging = !paging.IsNoPaging() && !commandInfo.Flags.IsSet(DbCommandFlags.WithPaging);
      if (doResultPaging && paging.Skip > 0)
        for (int i = 0; i < paging.Skip; i++)
          if (!reader.Read())
            return records;
      //Actually read records
      var entInfo = commandInfo.EntityCommand.TargetEntityInfo;
      while (reader.Read()) {
        var record = new EntityRecord(entInfo, EntityStatus.Loading);
        //Read column values
        object dbValue, value;
        for (int i = 0; i < commandInfo.QueryColumnMappings.Length; i++) {
          var colInfo = commandInfo.QueryColumnMappings[i];
          value = null;
          dbValue = null;
          // try/catch to provide more info 
          try {
            dbValue = reader[i];
            value = colInfo.TypeInfo.ReadConverter(dbValue);
          } catch (Exception ex) {
            ex.AddValue("columnName", colInfo.ColumnName);
            ex.AddValue("reader", reader);
            throw;
          }
          record.ValuesOriginal[colInfo.Member.ValueIndex] = value;
        }
        record.Status = EntityStatus.Loaded;
        records.Add(record);
        if (records.Count == paging.Take)
          break;
      }
      return records;
    }

    #region private utilities

    private DbCommandInfo GetDbCommandInfo(EntityCommand entityCommand) {
      var cmd = DbModel.LookupDbObject<DbCommandInfo>(entityCommand);
      if (cmd == null) {
        Util.Throw("Command {0} not implemented by the database. Most likely driver does not support this type of command.", entityCommand.CommandName);
      }        
      return cmd;
    }


    private object[] AddPagingArguments(object[] args, Paging paging = null) {
      paging = paging ?? Paging.NoPaging; 
      if (args == null)
        args = new object[2];
      else
        Array.Resize(ref args, args.Length + 2);
      //Note that row numbers are 1-based
      args[args.Length - 2] = paging.Skip; // offset
      args[args.Length - 1] = paging.Take; // max rows
      return args; 
    }

    protected object ReadOutputParameterValues(IDbCommand command, DbCommandInfo commandInfo, object[] args) {
      object returnValue = null;
      for (int i = 0; i < commandInfo.OutputParameters.Count; i++) {
        var prmInfo = commandInfo.OutputParameters[i];
        var prm = command.Parameters[prmInfo.Name] as IDbDataParameter;
        args[prmInfo.ArgIndex] = prm.Value;
      }//for
      //check return value
      if (commandInfo.ReturnValueParameter != null) {
        var prm = command.Parameters[commandInfo.ReturnValueParameter.Name] as IDbDataParameter;
        returnValue = prm.Value;
      }
      return returnValue;
    }

    private void SetCrudCommandParameterValues(DbCommandInfo commandInfo, IDbCommand command, EntityRecord record) {
      if (record.Status == EntityStatus.Stub)
        record.Reload();
      for (int i = 0; i < commandInfo.Parameters.Count; i++) {
        var prm = (IDbDataParameter)command.Parameters[i];
        prm.Value = DBNull.Value;
        var prmInfo = commandInfo.Parameters[i];
        var isInput = prmInfo.Direction == ParameterDirection.Input || prmInfo.Direction == ParameterDirection.InputOutput;
        if (!isInput)  continue; 
        var col = prmInfo.SourceColumn;
        if (col == null || col.Member == null) continue; 
        var value = record.GetValueDirect(col.Member);
        value = prmInfo.TypeInfo.WriteConverter(value);
        if (value == null)
          value = DBNull.Value;
        prm.Value = value; 
      } //for i
    }

    protected void ReadCrudOutputParameterValues(IDbCommand command, DbCommandInfo commandInfo, EntityRecord record) {
      for (int i = 0; i < commandInfo.OutputParameters.Count; i++) {
        var prmInfo = commandInfo.OutputParameters[i];
        var col = prmInfo.SourceColumn;
        if (col == null) continue; 
        var prm = command.Parameters[prmInfo.Name] as IDbDataParameter;
        var v = prmInfo.TypeInfo.ReadConverter(prm.Value);
        record.ValuesModified[col.Member.ValueIndex] = v; 
      }//for
    }

    private void BuildColumnMappings(DbCommandInfo commandInfo, IDataReader reader) {
      lock (this.DbModel.LockObject) {
        if (commandInfo.QueryColumnMappings != null)
          return;
        var count = reader.FieldCount;
        var colList = new List<DbColumnInfo>();
        for (int i = 0; i < count; i++) {
          var colName = reader.GetName(i);
          var colInfo = commandInfo.Table.Columns.FindByName(colName);
          //There maybe some artificial columns (for paged queries) that do not match to members - just ignore them.
          if (colInfo == null)
            continue;
          colList.Add(colInfo);
        }
        commandInfo.QueryColumnMappings = colList.ToArray();
      }
    }
    #endregion

    #region Logging
    private long CurrentTickCount {
      get {
        return _timeService == null ? -1 : _timeService.ElapsedMilliseconds;
      }
    }

    private void LogCommand(EntitySession session, IDbCommand command, long executionTime, int rowCount = -1) {
      var now = _timeService.Now;
      var entry = new DbCommandLogEntry(command, DbModel.Driver.CommandCallLogFormat, now, executionTime, rowCount);
      AddLogEntry(session, entry); 
    }

    protected void LogComment(EntitySession session, string comment, params object[] args) {
      var entry = new MessageLogEntry(LogEntryType.Information, comment, args);
      AddLogEntry(session, entry); 
    }

    protected void LogException(EntitySession session, Exception ex) {
      var entry = new ExceptionLogEntry(_timeService.Now, ex);
      AddLogEntry(session, entry);
    }

    private void AddLogEntry(EntitySession session, LogEntry entry) {
      var appLog = _appLogService.Log;
      if (appLog != null)
        appLog.AddEntry(entry);
      var userLog = session.UserContext.Log;
      if (userLog != null)
        userLog.AddEntry(entry); 
    }
    #endregion

  }//class

}//ns
