﻿using System;
using System.Collections.Generic;
using System.Data;
using System.Linq;
using System.Text;

using Vita.Entities;
using Vita.Data.Driver;
using Vita.Common;
using Vita.Entities.Runtime;

namespace Vita.Data {

  public enum TransactionMode {
    None, // also identifies implicit transaction for a single command. 
    Connection,
    Context,
  }

  /// <summary>
  /// A wrapper class for IDbConnection, IDbTransaction objects and some utility methods.
  /// </summary>
  public class DatabaseConnection : IDisposable {
    public EntitySession Session;
    public DbDriver Driver; 
    public IDbConnection DbConnection;
    public IDbTransaction DbTransaction;
    public System.Transactions.TransactionScope Scope;
    public IDataReader ActiveReader; // holds an open reader that uses this connection. Allows application to automatically close the reader when connection is closed/reopened 

    public DatabaseConnection(EntitySession session, Database database, TransactionMode mode) {
      Session = session;
      Driver = database.DbModel.Driver;
      DbConnection = Driver.CreateConnection(database.ConnectionString);
      DbConnection.Open();
      switch (mode) {
        case TransactionMode.None:
        case TransactionMode.Context: 
          break;
        case TransactionMode.Connection:
          DbTransaction = DbConnection.BeginTransaction(); 
          break; 
      }//switch
    }

    public void Open() {
      try {
        if (ActiveReader != null) {
          ActiveReader.Close();
          ActiveReader = null; 
        }
        DbConnection.Open();
      } catch (Exception ex) {
        throw new ConnectionFailureException("Opening DatabaseConnection failed: " + ex.Message, ex);
      }
    }

    //Close connection if there's no active transaction
    public void Close() {
      Util.Check(DbTransaction == null, "Cannot close connection - it has active transaction associated with it.");
      if (ActiveReader != null)
        ActiveReader.Close();
      ActiveReader = null; 
      if (DbConnection != null && DbConnection.State != ConnectionState.Closed)
        DbConnection.Close();
    }

    public void BeginTransaction(IsolationLevel isolationLevel = IsolationLevel.Serializable) {
      if (DbConnection.State != ConnectionState.Open)
        Open();
      DbTransaction = DbConnection.BeginTransaction(isolationLevel);
    }

    public void Commit() {
      if (DbTransaction == null)
        return; 
      DbTransaction.Commit();
      DbTransaction = null;
    }

    public void Abort(bool closeConnection = true) {
      if (DbTransaction != null)
        DbTransaction.Rollback();
      DbTransaction = null;
      if (closeConnection)
        Close(); 
    }

    public void Dispose() {
      Abort();
      Close();
    }
  } //class
}
