﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Linq;
using System.Text;
//using System.Data.Linq;
using System.Linq.Expressions;
using System.Reflection;
using System.Collections;
using System.Diagnostics;
using System.Data;

using Vita.Common;
using Vita.Entities;
using Vita.Entities.Model;
using Vita.Entities.Runtime;
using Vita.Data.Model;

namespace Vita.Data {

  using Vita.Data.Caching;
  using Vita.Data.Linq;
  
  //Dynamic LINQ execution

  public partial class Database {

    public object ExecuteDynamicQuery(EntitySession session, EntityQuery query) {
      object result;
      var queryInfo = SetupSqlQuery(query);
      try {
        var isQueryable = typeof(IQueryable).IsAssignableFrom(query.Expression.Type);
        if (isQueryable) {
          var list = ExecuteSelect(session, queryInfo);
          result = ReflectionHelper.ConvertToTypedList(list, query.ElementType);
        } else {
          result = SelectScalar(session, queryInfo);
        }
        return result;
      } catch (Exception ex) {
        this._driver.CheckDatabaseException(ex);
        var dbExc = new DbException(ex);
        dbExc.AddValue("QueryExpression", query.Expression);
        if (queryInfo.DbCommand != null)
          dbExc.AddValue("DbCommand", queryInfo.DbCommand);
        LogException(session, dbExc);
        throw dbExc;
      } finally {
        if (queryInfo.DbCommand != null && queryInfo.DbCommand.Connection != null)
          queryInfo.DbCommand.Connection = null; //clear reference to connection object, but do not close it
      }
    }

    private EntityQueryRuntimeInfo SetupSqlQuery(EntityQuery query) {
      var queryInfo = EntityQueryRuntimeInfo.GetQueryInfo(query);
      // if cacheKey is not null, try looking up in cache
      var cacheKey = queryInfo.CacheKey;
      if (cacheKey != null) {
        queryInfo.SqlQuery = _queryCache.Lookup(cacheKey);
        if (queryInfo.SqlQuery != null)
          return queryInfo;
      }
      // build fresh
      var queryBuilder = new Vita.Data.Linq.Translation.QueryBuilder(this.DbModel);
      queryInfo.SqlQuery = queryBuilder.BuildDbQuery(queryInfo);
      // Add to cache if query is allowed to be cached
      if (cacheKey != null && queryInfo.SqlQuery.CanCache)
        _queryCache.Add(cacheKey, queryInfo.SqlQuery);
      return queryInfo;
    }

    private IEnumerable ExecuteSelect(EntitySession session, EntityQueryRuntimeInfo queryInfo) {
      IList results = new List<object>();
      var dbQuery = queryInfo.SqlQuery;
      // handle the special case where the query is empty, meaning we don't need the DB
      if (string.IsNullOrEmpty(dbQuery.Sql)) {
        results.Add(dbQuery.RowObjectCreator(null, null));
        return results;
      }
      //really execute it
      var start = _timeService.ElapsedMilliseconds; 
      CreateDbQueryCommand(queryInfo);
      var dbCommand = queryInfo.DbCommand;
      DatabaseConnection conn = null;
      try {
        conn = OpenConnection(session);
        dbCommand.Connection = conn.DbConnection; //it is opened by default
        var reader = conn.ActiveReader = dbCommand.ExecuteReader();
        while (reader.Read()) {
          var row = dbQuery.RowObjectCreator(reader, session);
          results.Add(row);
        }
        var time = _timeService.ElapsedMilliseconds - start;
        LogCommand(session, dbCommand, time, results.Count);
      } catch (Exception ex) {
        ex.AddValue("DataReader", conn.ActiveReader);
        throw;
      } finally {
        if (conn != null)
          conn.Close();
      }
      return results;
    }

    private object SelectScalar(EntitySession session, EntityQueryRuntimeInfo queryInfo) {
      switch (queryInfo.SqlQuery.ExecuteMethodName) {
        case null: // some calls, like Count() generate SQL and the resulting projection method name is null (never initialized)
          return SelectSingle(session, queryInfo, false); // Single() for safety, but First() should work
        case "First":
          return SelectFirst(session, queryInfo, false);
        case "FirstOrDefault":
          return SelectFirst(session, queryInfo, true);
        case "Single":
          return SelectSingle(session, queryInfo, false);
        case "SingleOrDefault":
          return SelectSingle(session, queryInfo, true);
        case "Last":
          return SelectLast(session, queryInfo, false);
        default:
          Util.Throw("Unhandled method '{0}'", queryInfo.SqlQuery.ExecuteMethodName);
          return null;
      }
    }

    /// <summary>
    /// Returns first item in query.
    /// If no row is found then if default allowed returns default(S), throws exception otherwise
    /// </summary>
    /// <param name="selectQuery"></param>
    /// <param name="allowDefault"></param>
    /// <returns></returns>
    protected virtual object SelectFirst(EntitySession session, EntityQueryRuntimeInfo queryInfo, bool allowDefault) {
      foreach (var row in ExecuteSelect(session, queryInfo))
        return row;
      if (allowDefault) 
        return null; 
      Util.Throw("Queryable.First(...): the source query returned no elements.");
      return null; 
    }

    /// <summary>
    /// Returns single item in query
    /// If more than one item is found, throws an exception
    /// If no row is found then if default allowed returns default(S), throws exception otherwise
    /// </summary>
    /// <typeparam name="S"></typeparam>
    /// <param name="selectQuery"></param>
    /// <param name="allowDefault"></param>
    /// <returns></returns>
    protected virtual object SelectSingle(EntitySession session, EntityQueryRuntimeInfo queryInfo, bool allowDefault) {
      object firstRow = null;
      int rowCount = 0;
      foreach (var row in ExecuteSelect(session, queryInfo)) {
        if (rowCount > 1)
          Util.Throw("Queryable.Single(...): the source query returned more than one element.");
        firstRow = row;
        rowCount++;
      }
      if (!allowDefault && rowCount == 0)
        throw new InvalidOperationException();
      return firstRow;
    }

    /// <summary>
    /// Returns last item in query
    /// </summary>
    /// <param name="allowDefault"></param>
    /// <returns></returns>
    protected virtual object SelectLast(EntitySession session, EntityQueryRuntimeInfo queryInfo, bool allowDefault) {
      object lastRow = null;
      int rowCount = 0;
      foreach (var row in ExecuteSelect(session, queryInfo)) {
        lastRow = row;
        rowCount++;
      }
      if (!allowDefault && rowCount == 0)
          Util.Throw("Queryable.Last(...): the source query returned no elements.");
      return lastRow;
    }

    private IDbCommand CreateDbQueryCommand(EntityQueryRuntimeInfo queryInfo) {
      var paramArray = queryInfo.ParamValues.ToArray(); 
      var cmd = DbModel.Driver.CreateDbCommand();
      queryInfo.DbCommand = cmd; 
      cmd.CommandType = CommandType.Text;
      cmd.CommandText = queryInfo.SqlQuery.Sql;
      foreach (var qParam in queryInfo.SqlQuery.Parameters) {
        var value = qParam.ReadValue(paramArray);
        var adoParam = cmd.CreateParameter();
        adoParam.ParameterName = DbModel.LinqSqlProvider.GetParameterName(qParam.Name);
        adoParam.Value = value;
        DbModel.LinqSqlProvider.CheckDbParameter(adoParam);
        cmd.Parameters.Add(adoParam);
      }
      return cmd;
    }

  }
}
