﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data;
using System.Transactions;
using System.Diagnostics;

using Vita.Common;
using Vita.Entities.Model;
using Vita.Data.Model;
using Vita.Activation;

namespace Vita.Data.Driver {

  public abstract class DbDriver {
    public DbCapabilities Capabilities { get; protected set; }
    // Parameter prefix. For MS SQL, it is '@' for both. For MySql, we need to use '@' for dynamic SQLs but no prefix or smth like 'prm' for stored procs
    // DbSqlBuilder chooses prefix depending on UseStoredProc flag in Settings.Options - which controls if app uses stored procs or dynamic SQL for CRUD 
    public string DynamicSqlParameterPrefix = string.Empty;
    public string StoredProcParameterPrefix = string.Empty;
    public const string GeneratedProcedureTag = "-- VITA:Generated;Tag";
    public string CommandCallLogFormat = "CALL {0} ({1});";


    public DbDriver() {
    }

    #region virtual and abstract methods - to be overridden to customize behavior for particular server
    public abstract IDbConnection CreateConnection(string connectionString);

    public abstract DbSchemaManager CreateSchemaManager(string connectionString);
    public virtual DbSqlBuilder CreateSqlBuilder(DbModel dbModel) {
      return new DbSqlBuilder(dbModel);
    }

    public virtual DbTypeMapper CreateTypeMapper(DbSettings settings) {
      return new DbTypeMapper(settings); 
    }

    public virtual Vita.Data.Driver.LinqSqlProvider GetLinqSqlProvider() {
      return null; 
    }
    #endregion

    public bool Supports(DbCapabilities capability) {
      return (Capabilities & capability) != 0;
    }

    public virtual string GetProviderVersion(string connectionString) {
      var conn =  CreateConnection(connectionString) as System.Data.Common.DbConnection;
      if (conn == null) 
        return string.Empty; 
      return conn.ServerVersion;
    }

    private string _fullNameFormat;
    public virtual string GetFullName(string schema, string name) {
      if (_fullNameFormat == null) {
        var supportsSchemas = Supports(DbCapabilities.Schemas);
        _fullNameFormat = supportsSchemas ? "[{0}].[{1}]" : "[{1}]";
      }
      return string.Format(_fullNameFormat, schema, name);
    }

    public virtual object ExecuteCommand(IDbCommand command, DbExecutionType executionType) {
      try {
        switch (executionType) {
          case DbExecutionType.Query:
            return command.ExecuteReader();
          case DbExecutionType.NonQuery:
            return command.ExecuteNonQuery();
          case DbExecutionType.Scalar:
            return command.ExecuteScalar();
        }
        return null; //never happens
      } catch (Exception ex) {
        CheckDatabaseException(ex, command); //throws if different
        throw new DbException(ex, command); //rethrow it as DbException
      }
    }

    public abstract IDbCommand CreateDbCommand();

    public virtual IDbCommand CreateDbCommand(IDbConnection connection, DbCommandInfo commandInfo) {
      var command = connection.CreateCommand();
      if (commandInfo.StoredProcDefined()) {
        command.CommandType = CommandType.StoredProcedure;
        command.CommandText = commandInfo.FullCommandName;
      } else {
        command.CommandType = CommandType.Text;
        command.CommandText = commandInfo.Sql;
      }
      //Create parameters collection
      for (int index = 0; index < commandInfo.Parameters.Count; index++) {
        var prm = CreateDbParameter(command, commandInfo.Parameters[index]);
        command.Parameters.Add(prm);
      }
      return command;
    }//method

    protected virtual IDbDataParameter CreateDbParameter(IDbCommand command, DbParamInfo prmInfo) {
      var prm = command.CreateParameter();
      var typeInfo = prmInfo.TypeInfo;
      prm.DbType = typeInfo.DbType;
      prm.Direction = prmInfo.Direction;
      prm.ParameterName = prmInfo.Name;
      if (typeInfo.Precision > 0) prm.Precision = typeInfo.Precision;
      if (typeInfo.Scale > 0) prm.Scale = typeInfo.Scale;
      if (typeInfo.Size > 0) prm.Size = typeInfo.Size;
      prm.Value = prmInfo.DefaultValue;
      return prm; 
    }

    protected virtual IDbDataParameter CloneDbParameter(IDbCommand command, IDbDataParameter parameter) {
      var prm = command.CreateParameter();
      prm.ParameterName = parameter.ParameterName;
      prm.DbType = parameter.DbType;
      prm.Direction = parameter.Direction;
      prm.Precision = parameter.Precision;
      prm.Scale = parameter.Scale;
      prm.Size = parameter.Size;
      prm.SourceColumn = parameter.SourceColumn;
      prm.SourceVersion = parameter.SourceVersion;
      return prm; 
    }

    // The method should check specific exception types and if necessary throw a different kind of exception. 
    // One exception to detect and rethrow is unique index violation. When database throws an exception on unique violation, most of the time it is user error -
    // he tries to create a duplicate entity. Therefore this error should be detected and rethrown as Validation exception.
    // Detecting this type of exception is provider-specific. MS SQL driver overrides this method to detect SqlException for unique index violation. 
    public virtual void CheckDatabaseException(Exception exception, IDbCommand command = null) {
    }

    // Stored procedures ==============================================================
    public virtual bool ConvertToStoredProc(DbCommandInfo command) {
      Util.Throw("Stored procedures are not supported by driver {0}", this.GetType().Name);
      return false;
    }

    public virtual void OnDbModelConstructed(DbModel dbModel) {

    }

    //By default we compare full text.  
    public virtual bool StoredProceduresMatch(DbCommandInfo x, DbCommandInfo y) {
      if (x.StoredProcText.Trim() == y.StoredProcText.Trim())
        return true;
      return false;
    }


  }//class

  public static class DbDriverEnumExtensions {
    public static bool IsSet(this DbColumnSpecOptions flags, DbColumnSpecOptions flag) {
      return (flags & flag) != 0;
    }
    public static bool IsSet(this DbCapabilities flags, DbCapabilities flag) {
      return (flags & flag) != 0;
    }
  }

}//namespace
