﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data;

using Vita.Common;
using Vita.Entities.Model;
using Vita.Data.Model;

namespace Vita.Data.Driver {

  [Flags]
  public enum DbColumnSpecOptions {
    None = 0,
    NewColumn = 0x01,
    // for new columns without defaults we create them as nullable 
    ForceNull = 0x02,
/*
    //Suppress NULL/NOT NULL altogether. SqlCe does not allow any null option in ntext columns.
    SuppressNull = 0x04,
 */ 
  }


  public abstract class DbSchemaManager : IDisposable {
    public readonly IDbConnection Connection;
    private DbDriver _driver; 

    protected DataTable EmptyTable = new DataTable(); 

    public DbSchemaManager(DbDriver driver, string connectionString) {
      _driver = driver; 
      Connection = _driver.CreateConnection(connectionString);
      Connection.Open(); 
    }

    private bool Supports(DbCapabilities capability) {
      var result = _driver.Supports(capability);
      return result;
    }

    public void Close() {
      if (Connection.State != ConnectionState.Closed)
        Connection.Close(); 
    }

    public void Dispose() {
      Close(); 
    }

    #region utility methods: SQL execution
    public virtual void ExecuteNonQuery(string sqlTemplate, params object[] args) {
      if (args != null && args.Length > 0)
        sqlTemplate = string.Format(sqlTemplate, args);
      var cmd = Connection.CreateCommand();
      cmd.CommandText = sqlTemplate;
      _driver.ExecuteCommand(cmd, DbExecutionType.NonQuery); 
    }

    public virtual DataTable ExecuteSelect(string sql) {
      var cmd = Connection.CreateCommand();
      cmd.CommandText = sql;
      var reader = (IDataReader) _driver.ExecuteCommand(cmd, DbExecutionType.Query);
      var table = new DataTable();
        //Have to put table into Dataset to disable constraints - see below about SQL CE
        var ds = new DataSet();
        ds.EnforceConstraints = false;
        ds.Tables.Add(table);
      table.Load(reader); // Smth stupid: SQL CE is giving "failed to enable constraints..." on table.Load(reader);  
      return table; 
    }
    #endregion

    #region Reading schema - using INFORMATION_SCHEMA views
    public virtual DataTable GetDatabases() {
      var sql = "SELECT DISTINCT CATALOG_NAME FROM INFORMATION_SCHEMA.SCHEMATA;";
      return ExecuteSelect(sql);
    }

    //Columns:CATALOG_NAME	SCHEMA_NAME	SCHEMA_OWNER	DEFAULT_CHARACTER_SET_CATALOG	DEFAULT_CHARACTER_SET_SCHEMA 
    //	DEFAULT_CHARACTER_SET_NAME  
    public virtual DataTable GetSchemas() {
      var sql = "SELECT * FROM INFORMATION_SCHEMA.SCHEMATA;";
      return ExecuteSelect(sql);
    }

    //Columns: TABLE_CATALOG	TABLE_SCHEMA	TABLE_NAME	TABLE_TYPE  (table_type: "BASE TABLE") 
    public virtual DataTable GetTables(string schema = null) {
      var sql = "SELECT * FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_TYPE='TABLE' ";
      if (!string.IsNullOrEmpty(schema))
        sql += string.Format(" AND TABLE_SCHEMA='{0}'", schema); 
      return ExecuteSelect(sql);
    }

    //Columns: TABLE_CATALOG	TABLE_SCHEMA	TABLE_NAME	COLUMN_NAME	ORDINAL_POSITION	
    //COLUMN_DEFAULT	IS_NULLABLE	DATA_TYPE	CHARACTER_MAXIMUM_LENGTH	CHARACTER_OCTET_LENGTH	
    //NUMERIC_PRECISION	NUMERIC_PRECISION_RADIX	NUMERIC_SCALE	DATETIME_PRECISION	
    //CHARACTER_SET_CATALOG	CHARACTER_SET_SCHEMA	CHARACTER_SET_NAME	COLLATION_CATALOG	
    //COLLATION_SCHEMA	COLLATION_NAME	DOMAIN_CATALOG	DOMAIN_SCHEMA	DOMAIN_NAME
    public virtual DataTable GetColumns(string schema = null) {
      var sql = "SELECT * FROM INFORMATION_SCHEMA.COLUMNS " ;
      if (!string.IsNullOrEmpty(schema))
        sql += string.Format(" WHERE TABLE_SCHEMA='{0}'", schema);
      sql += " ORDER BY TABLE_SCHEMA, TABLE_NAME, ORDINAL_POSITION;";
      return ExecuteSelect(sql);
    }

    //Columns: CONSTRAINT_CATALOG	CONSTRAINT_SCHEMA	CONSTRAINT_NAME	TABLE_CATALOG	
    //TABLE_SCHEMA	TABLE_NAME	CONSTRAINT_TYPE	IS_DEFERRABLE	INITIALLY_DEFERRED
    public virtual DataTable GetTableConstraints(string schema = null) {
      var sql = "SELECT * FROM INFORMATION_SCHEMA.TABLE_CONSTRAINTS";
      if (!string.IsNullOrEmpty(schema))
        sql += string.Format(" WHERE TABLE_SCHEMA='{0}'", schema);
      return ExecuteSelect(sql);
    }
    
    //Columns: TABLE_CATALOG	TABLE_SCHEMA	TABLE_NAME	COLUMN_NAME ORDINAL_POSITION	CONSTRAINT_CATALOG	
    //CONSTRAINT_SCHEMA	CONSTRAINT_NAME
    public virtual DataTable GetConstraintColumnUsage(string schema = null) {
      // NOTE: changed to KEY_COLUMN_USAGE because SQL CE does not support CONSTRAINT_COLUMN_USAGE. Seem to be equivalent
      //var sql = "SELECT * FROM INFORMATION_SCHEMA.CONSTRAINT_COLUMN_USAGE";
      var sql = "SELECT * FROM INFORMATION_SCHEMA.KEY_COLUMN_USAGE";
      if (!string.IsNullOrEmpty(schema))
        sql += string.Format(" WHERE TABLE_SCHEMA='{0}'", schema);
      sql += " ORDER BY TABLE_SCHEMA, TABLE_NAME, ORDINAL_POSITION;";
      return ExecuteSelect(sql);
    }

    //Columns: CONSTRAINT_CATALOG, CONSTRAINT_SCHEMA, CONSTRAINT_NAME, UNIQUE_CONSTRAINT_CATALOG, UNIQUE_CONSTRAINT_SCHEMA, 
    //  UNIQUE_CONSTRAINT_NAME, MATCH_OPTION, UPDATE_RULE, DELETE_RULE
    // Not used, we use ext version below
    public virtual DataTable GetReferentialConstraints(string schema = null) {
      var sql = "SELECT * FROM INFORMATION_SCHEMA.REFERENTIAL_CONSTRAINTS";
      if (!string.IsNullOrEmpty(schema))
        sql += string.Format(" WHERE CONSTRAINT_SCHEMA='{0}'", schema);
      return ExecuteSelect(sql);
    }

    //Columns: CONSTRAINT_CATALOG, CONSTRAINT_SCHEMA, CONSTRAINT_NAME, UNIQUE_CONSTRAINT_CATALOG, UNIQUE_CONSTRAINT_SCHEMA, 
    //  UNIQUE_CONSTRAINT_NAME, MATCH_OPTION, UPDATE_RULE, DELETE_RULE, C_TABLE, U_TABLE
    // Extended version with joins to bring Table names for both constraints
    public virtual DataTable GetReferentialConstraintsExt(string schema = null) {
      var sql = @"
        SELECT rc.*, tc1.TABLE_NAME as C_TABLE, tc2.TABLE_NAME AS U_TABLE 
        FROM INFORMATION_SCHEMA.REFERENTIAL_CONSTRAINTS rc
          INNER JOIN INFORMATION_SCHEMA.TABLE_CONSTRAINTS tc1 
            ON (tc1.CONSTRAINT_SCHEMA = rc.CONSTRAINT_SCHEMA AND tc1.CONSTRAINT_NAME = rc.CONSTRAINT_NAME)
          INNER JOIN INFORMATION_SCHEMA.TABLE_CONSTRAINTS tc2 
            ON (tc2.CONSTRAINT_SCHEMA = rc.UNIQUE_CONSTRAINT_SCHEMA AND tc2.CONSTRAINT_NAME = rc.UNIQUE_CONSTRAINT_NAME)";
      if (!string.IsNullOrEmpty(schema))
        sql += string.Format(" WHERE tc1.CONSTRAINT_SCHEMA='{0}'", schema);
      return ExecuteSelect(sql);
    }


    //All Columns: SPECIFIC_CATALOG	SPECIFIC_SCHEMA	SPECIFIC_NAME	ROUTINE_CATALOG	ROUTINE_SCHEMA 
    //	ROUTINE_NAME	ROUTINE_TYPE	MODULE_CATALOG	MODULE_SCHEMA	MODULE_NAME	UDT_CATALOG	
    //UDT_SCHEMA	UDT_NAME	DATA_TYPE	CHARACTER_MAXIMUM_LENGTH	CHARACTER_OCTET_LENGTH	
    //COLLATION_CATALOG	COLLATION_SCHEMA	COLLATION_NAME	CHARACTER_SET_CATALOG	
    //CHARACTER_SET_SCHEMA	CHARACTER_SET_NAME	NUMERIC_PRECISION	NUMERIC_PRECISION_RADIX	
    //NUMERIC_SCALE	DATETIME_PRECISION	INTERVAL_TYPE	INTERVAL_PRECISION	TYPE_UDT_CATALOG	
    //TYPE_UDT_SCHEMA	TYPE_UDT_NAME	SCOPE_CATALOG	SCOPE_SCHEMA	SCOPE_NAME	MAXIMUM_CARDINALITY	
    //DTD_IDENTIFIER	ROUTINE_BODY	ROUTINE_DEFINITION	EXTERNAL_NAME	EXTERNAL_LANGUAGE	
    //PARAMETER_STYLE	IS_DETERMINISTIC	SQL_DATA_ACCESS	IS_NULL_CALL	SQL_PATH	SCHEMA_LEVEL_ROUTINE	
    //MAX_DYNAMIC_RESULT_SETS	IS_USER_DEFINED_CAST	IS_IMPLICITLY_INVOCABLE	CREATED	LAST_ALTERED
    // Note: we select only schema, name, definition and type
    public virtual DataTable GetProcedures(string schema = null) {
      var sql = @"
  SELECT ROUTINE_SCHEMA, ROUTINE_NAME, ROUTINE_DEFINITION, ROUTINE_TYPE 
    FROM INFORMATION_SCHEMA.ROUTINES 
    WHERE ROUTINE_TYPE='PROCEDURE' ";
      if (!string.IsNullOrEmpty(schema))
        sql += string.Format(" AND ROUTINE_SCHEMA='{0}'", schema);
      return ExecuteSelect(sql);
    }

    //INFORMATION_SCHEMA does not have a view for indexes, column defaults and identity columns
    //Drivers should override this method with particular implementations
    public virtual DataTable GetIndexes(string schema = null) {
      return EmptyTable; 
    }
    public virtual DataTable GetIndexColumns(string schema = null) {
      return EmptyTable;
    }
    #endregion


    // A chance for a driver to add provider-specific information to the loaded column. 
    public virtual void OnColumnLoaded(DbModel model, DbColumnInfo column, DataRow columnRow) {

    }

    // A chance for a driver to add provider-specific information to the loaded model. 
    public virtual void OnModelLoaded(DbModel model) {

    }

    #region Modifying schema methods

    protected virtual string GetDefaultConstraintName(DbColumnInfo column) {
      var tbl = column.Table;
      return "DEFAULT_" + tbl.SchemaName + "_" + tbl.TableName + "_" + column.ColumnName;
    }


    public virtual void CreateDatabase(string name) {
      ExecuteNonQuery("CREATE DATABASE [{0}]", name);
    }

    public virtual void CreateSchema(string schema) {
      ExecuteNonQuery("CREATE SCHEMA [{0}];", schema);
    }//method

    public virtual void CreateTable(DbTableInfo table) {
      var specs = table.Columns.Select(c => GetColumnSpec(c, DbColumnSpecOptions.NewColumn));
      var columnSpecs = string.Join("," + Environment.NewLine, specs);
      //ExecuteNonQuery(" CREATE TABLE {0} ( \r\n {1} \r\n ) ", table.FullName, columnSpecs);
      var templ = @"CREATE TABLE {0} (" + "\r\n {1} \r\n); ";
      ExecuteNonQuery(templ, table.FullName, columnSpecs);
    }//method

    //ALTER TABLE employees ADD [employee_pwd] nvarchar(20) Null;
    public virtual void AddColumn(DbColumnInfo column) {
      var options = DbColumnSpecOptions.NewColumn;
      var nullable = column.Flags.IsSet(DbColumnFlags.Nullable);
      // Special case - identity columns do not allow nulls
      var forceNullable = !nullable && !column.Flags.IsSet(DbColumnFlags.Identity) && string.IsNullOrEmpty(column.DefaultExpression);
      if (forceNullable)
        options |= DbColumnSpecOptions.ForceNull;
      var colSpec = GetColumnSpec(column, options);
      ExecuteNonQuery("ALTER TABLE {0} ADD {1}", column.Table.FullName, colSpec);
    }
    //ALTER TABLE employees ALTER COLUMN [employee_name] nvarchar(100) Null;
    public virtual void ModifyColumn(DbColumnInfo column) {
      var colSpec = GetColumnSpec(column, DbColumnSpecOptions.None);
      ExecuteNonQuery("ALTER TABLE {0} ALTER COLUMN {1};", column.Table.FullName, colSpec);
    }

    // Syntax differs between database vendors; also not all drivers implement renaming directly. 
    // By default we are doing it through creating new column, copying data and dropping old one
    public virtual void RenameColumn(DbColumnInfo oldColumn, DbColumnInfo newColumn) {
      AddColumn(newColumn);
      ExecuteNonQuery("UPDATE {0} SET \"{1}\" = \"{2}\";", oldColumn.Table.FullName, newColumn.ColumnName, oldColumn.ColumnName);
      DropColumn(oldColumn); 
      // just in case, the new column was initially created nullable, but should not be, drop nullable
      if (!newColumn.Flags.IsSet(DbColumnFlags.Nullable))
        ModifyColumn(newColumn); // this will modify it to be non-nullable
    }
    public virtual void RenameTable(DbTableInfo oldTable, DbTableInfo newTable) {
      throw new NotImplementedException("Method RenameTable is not implemented by driver " + this.GetType().Name);
    }
    //ALTER TABLE employees DROP COLUMN [employee_pwd];
    public virtual void DropColumn(DbColumnInfo column) {
      ExecuteNonQuery("ALTER TABLE {0} DROP COLUMN \"{1}\"", column.Table.FullName, column.ColumnName);
    }

    public virtual void InitNullColumnValues(DbColumnInfo column) {
      const string sqlUpdateValues = "UPDATE {0} SET \"{1}\" = {2} WHERE \"{1}\" IS NULL;";
      var fullTableRef = column.Table.FullName;
      ExecuteNonQuery(sqlUpdateValues, fullTableRef, column.ColumnName, column.TypeInfo.InitExpression);
    }

    // Keys
    public virtual void CreateDbKey(DbKeyInfo key) {
      if (key.KeyType.IsSet(KeyType.PrimaryKey))
        CreatePrimaryKey(key);
      else if (key.KeyType.IsSet(KeyType.Index)) {
        CreateIndex(key);
      }
    }

    public virtual void CreateIndex(DbKeyInfo key) {
      var unique = key.KeyType.IsSet(KeyType.Unique) ? "UNIQUE" : string.Empty;
      string clustered = GetClusteredExpression(key); 
      string indexFields;
      if (_driver.Supports(DbCapabilities.OrderedColumnsInIndexes))
        indexFields = key.GetSqlNameListWithOrderSpec();
      else 
        indexFields = key.Columns.GetSqlNameList();
      ExecuteNonQuery("CREATE {0} {1} INDEX \"{2}\"  ON {3} ( {4} )",
                                        unique, clustered, key.Name, key.Table.FullName, indexFields);
    }
    public virtual void CreatePrimaryKey(DbKeyInfo key) {
      string clustered = GetClusteredExpression(key);
      var pkFields = key.Columns.GetSqlNameList();
      ExecuteNonQuery("ALTER TABLE {0} ADD CONSTRAINT \"{1}\" PRIMARY KEY {2} ({3});", key.Table.FullName, key.Name, clustered, pkFields);
    }

    protected string GetClusteredExpression(DbKeyInfo key) {
      string clustered = string.Empty;
      if (Supports(DbCapabilities.ClusteredIndexes))
        clustered = key.KeyType.IsSet(KeyType.Clustered) ? "CLUSTERED" : "NONCLUSTERED";
      return clustered;
    }

    public virtual void CreateReferentialConstraint(DbRefConstraintInfo refConstraint) {
      const string sqlTemplate = "ALTER TABLE {0} ADD CONSTRAINT \"{1}\" FOREIGN KEY ({2}) REFERENCES {3} ({4}) {5};";
      var srcTable = refConstraint.FromKey.Table;
      var targetTable = refConstraint.ToKey.Table;
      var fullSrcTableRef = srcTable.FullName;
      var fullTargetTableRef = targetTable.FullName;
      var srcCols = refConstraint.FromKey.Columns.GetSqlNameList();
      var targetCols = refConstraint.ToKey.Columns.GetSqlNameList();
      bool cascade = refConstraint.OwnerReference.FromMember.Flags.IsSet(EntityMemberFlags.CascadeDelete);
      var onDeleteClause = cascade ? "ON DELETE CASCADE" : string.Empty;
      ExecuteNonQuery(sqlTemplate, fullSrcTableRef, refConstraint.FromKey.Name, 
        srcCols, fullTargetTableRef, targetCols, onDeleteClause);
    }

    public virtual void AddColumnDefault(DbColumnInfo column) {
      var constrName = GetDefaultConstraintName(column);
      this.ExecuteNonQuery("ALTER TABLE {0} ADD CONSTRAINT \"{2}\" DEFAULT  ({3}) FOR \"{4}\";", 
          column.Table.FullName, constrName, column.DefaultExpression, column.ColumnName);
    }

    public virtual void DropColumnDefault(DbColumnInfo column) {
      var constrName = GetDefaultConstraintName(column);
      this.ExecuteNonQuery("ALTER TABLE {0} DROP CONSTRAINT \"{1}\";", column.Table.FullName, constrName);
    }

    // Deleting objects ##################################################################
    public virtual void DropDatabase(string name) {
      ExecuteNonQuery("DROP DATABASE \"{0}\";", name);
    }
    public virtual void DropSchema(string schema) {
      ExecuteNonQuery("DROP SCHEMA \"{0}\";", schema);
    }//method

    public virtual void DropTable(DbTableInfo table) {
      ExecuteNonQuery("DROP TABLE {0}", table.FullName);
    }//method

    public virtual void DropRoutine( string fullName) {
      ExecuteNonQuery("DROP PROCEDURE {0};", fullName);
    }//method

    public virtual void DropReferentialConstraint(DbRefConstraintInfo dbRefConstraint) {
      DropTableConstraint(dbRefConstraint.FromKey); 
    }

    public virtual void DropDbKey(DbKeyInfo dbKey) {
      switch (dbKey.KeyType) {
        case KeyType.ForeignKey: //FKs are deleted with Ref constraints
          break; 
        case KeyType.PrimaryKey:
        case KeyType.ClusteredPrimaryKey:
          DropTableConstraint(dbKey);
          break;
        default:
          DropIndex(dbKey);
          break;
      }
    }

    public virtual void DropIndex(DbKeyInfo key) {
      ExecuteNonQuery("DROP INDEX \"{0}\" ON {1};", key.Name, key.Table.FullName);
    }

    public virtual void DropTableConstraint(DbKeyInfo key) {
      ExecuteNonQuery("ALTER TABLE {0} DROP CONSTRAINT \"{1}\";", key.Table.FullName, key.Name);
    }

    public virtual void AddRoutine(DbCommandInfo command) {
      this.ExecuteNonQuery(command.StoredProcText);
    }

    protected virtual string GetColumnSpec(DbColumnInfo column, DbColumnSpecOptions options) {
      var typeStr = column.TypeInfo.SqlTypeSpec;
      var nullable = options.IsSet(DbColumnSpecOptions.ForceNull) || column.Flags.IsSet(DbColumnFlags.Nullable);
      var nullStr = nullable ? "NULL" : "NOT NULL";
      var idStr = string.Empty;
      if (options.IsSet(DbColumnSpecOptions.NewColumn) && column.Flags.IsSet(DbColumnFlags.Identity)) {
        idStr = "IDENTITY(1,1)";
      }
      string defaultStr = null;
      //Default constraint can be set only on new columns
      if (!string.IsNullOrWhiteSpace(column.DefaultExpression) && options.IsSet(DbColumnSpecOptions.NewColumn))
        defaultStr = "DEFAULT " + column.DefaultExpression;
      var spec = string.Format(" \"{0}\" {1} {2} {3} {4}", column.ColumnName, typeStr, idStr, defaultStr, nullStr);
      return spec;
    }
    #endregion

    public virtual string GetProcedureTag(DataRow row) {
      return null;
    }

    //Must be overridden in ancestorss
    public virtual void DropSchemaObjects(string schema) {
      Util.NotImplemented("DropSchema object is not implemented by driver {0}.", this.GetType()); 
    }

    public virtual void ValidateCommand(IDbCommand command) {
      command.Connection = Connection;
      command.Prepare();
      command.Connection = null; 
    }

  }//class

}
