﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data;
using System.Data.Common;
using System.Reflection;

using Vita.Common;
using Vita.Entities;
using Vita.Entities.Model;
using Vita.Logging;
using Vita.Data.Model;
using Vita.Activation;

namespace Vita.Data.Driver {
  
  //Builds CRUD and custom SQL commands/stored procedures
  public partial class DbSqlBuilder {

    protected DbSettings Settings;
    protected DbModel DbModel;
    protected DbTypeMapper TypeMapper;

    public DbSqlBuilder(DbModel dbModel) { 
      DbModel = dbModel; 
      Settings = DbModel.Settings;
      TypeMapper = DbModel.Driver.CreateTypeMapper(Settings); 
    }

    public virtual string GetFullName(string schema, string name) {
      return Settings.Driver.GetFullName(schema, name); 
    }

    public virtual DbCommandInfo BuildCommand(EntityCommand command) {
      switch (command.Kind) {
        case EntityCommandKind.SelectAll:
          return BuildSelectAllCommand(command);
        case EntityCommandKind.SelectAllPaged:
          return BuildSelectAllPagedCommand(command);
        case EntityCommandKind.SelectByKey:
          return BuildSelectByKeyCommand(command);
        case EntityCommandKind.SelectByKeyManyToMany:
          return BuildSelectManyToManyCommand(command);
        case EntityCommandKind.Insert:
          return BuildSqlInsertCommand(command); 
        case EntityCommandKind.Update:
        case EntityCommandKind.PartialUpdate:
          return BuildSqlUpdateCommand(command); 
        case EntityCommandKind.Delete:
          return BuildSqlDeleteCommand(command); 
        default:
          // it is custom command - should be compiled on the fly
          return null; 
      }
    }

    // Builds a SELECT command without paging. For now paging is provider-dependent.
    public virtual DbCommandInfo BuildSelectAllCommand(EntityCommand entityCommand) {
      var table = DbModel.LookupDbObject<DbTableInfo>(entityCommand.TargetEntityInfo, throwNotFound: true);
      const string SqlSelectAllTemplate = 
@"SELECT {0} {1} 
FROM {2} 
{3} {4}; ";
      //Build column list
      var strColumns = table.Columns.GetSelectable().GetSqlNameList();
      var strOrderBy = BuildOrderBy(table.DefaultOrderBy);
      var limit = GetSqlLimitClause(Settings.QueryLimit.ToString());
      var sql = string.Format(SqlSelectAllTemplate, limit.TopClause, strColumns, table.FullName, strOrderBy, limit.LimitClause);
      var cmdName = Settings.NamingPolicy.ConstructDbCommandName(entityCommand, table.TableName, "SelectAll");
      var cmdInfo = CreateDbCommandInfo(entityCommand, cmdName, table, DbExecutionType.Query, sql);
      return cmdInfo;
    }

    // Paging is provider-specific.  Drivers should overwrite this method to implement paging.
    public virtual DbCommandInfo BuildSelectAllPagedCommand(EntityCommand entityCommand) {
      return null; 
    }

    public virtual DbCommandInfo BuildSqlDeleteCommand(EntityCommand entityCommand) {
      const string SqlDeleteTemplate = "DELETE FROM {0} WHERE {1};";
      var table = DbModel.LookupDbObject<DbTableInfo>(entityCommand.TargetEntityInfo, throwNotFound: true);
      //Load by primary key
      var cmdName = Settings.NamingPolicy.ConstructDbCommandName(entityCommand, table.TableName, "Delete");
      var cmdInfo = CreateDbCommandInfo(entityCommand, cmdName, table, DbExecutionType.NonQuery, null);
      var strWhere = BuildWhereClause(cmdInfo.Parameters);
      cmdInfo.Sql = string.Format(SqlDeleteTemplate, table.FullName, strWhere);
      return cmdInfo; 
    }

    public virtual DbCommandInfo BuildSqlUpdateCommand(EntityCommand entityCommand) {
      const string SqlUpdateTemplate = 
@"UPDATE {0} 
  SET {1} 
WHERE {2};";
      var table = DbModel.LookupDbObject<DbTableInfo>(entityCommand.TargetEntityInfo, throwNotFound: true);
      if (entityCommand == null)
        return null; 
      var cmdName = Settings.NamingPolicy.ConstructDbCommandName(entityCommand, table.TableName, "Update");
      var cmdInfo = CreateDbCommandInfo(entityCommand, cmdName, table, DbExecutionType.NonQuery, null);
      var pkParams = cmdInfo.Parameters.Where(p => p.SourceColumn.Flags.IsSet(DbColumnFlags.PrimaryKey));
      // find members to update. For partial update, with explicitly specified members to update, we ignore NoUpdate flag
      var excludeFlags = DbColumnFlags.PrimaryKey;
      if (entityCommand.Kind == EntityCommandKind.Update)
        excludeFlags |= DbColumnFlags.NoUpdate;
      var updateParams = cmdInfo.Parameters.Where(p => !p.SourceColumn.Flags.IsSet(excludeFlags));
      // Some tables (like many-to-many link entities) might have no columns to update
      if (!updateParams.Any())
        return null;
      //Build Where expression
      var whereExpr = BuildWhereClause(pkParams.ToList());
      //Build Update clause
      var qt = "\"";
      var updList = new StringList();
      foreach (var prm in updateParams) 
        updList.Add(qt + prm.SourceColumn.ColumnName + qt + " = " + prm.Name);
      var strUpdates = string.Join(", ", updList);
      //Build SQL
      cmdInfo.Sql = string.Format(SqlUpdateTemplate, table.FullName, strUpdates, whereExpr);
      return cmdInfo; 
    }

    public virtual DbCommandInfo BuildSqlInsertCommand(EntityCommand entityCommand) {
      const string SqlInsertTemplate = "INSERT INTO {0} \r\n  ({1}) \r\n  VALUES \r\n    ({2}); \r\n";
      var table = DbModel.LookupDbObject<DbTableInfo>(entityCommand.TargetEntityInfo, throwNotFound: true);
      var idClause = string.Empty; 
      var listColumns = new List<DbColumnInfo>(); 
      var listValues = new StringList();
      var cmdName = Settings.NamingPolicy.ConstructDbCommandName(entityCommand, table.TableName, "Insert");
      var dbCmdInfo = CreateDbCommandInfo(entityCommand, cmdName, table, DbExecutionType.NonQuery, null);
      foreach (var prm in dbCmdInfo.Parameters) {
        var col = prm.SourceColumn;
        if (!col.Flags.IsSet(DbColumnFlags.NoInsert)) {
          listColumns.Add(col);
          listValues.Add(prm.Name);
        }
      }
      //build SQL
      var strColumns = listColumns.GetSqlNameList();
      var strValues = string.Join(", ", listValues);
      dbCmdInfo.Sql = string.Format(SqlInsertTemplate, table.FullName, strColumns, strValues) + idClause;
      return dbCmdInfo; 
    }

    public virtual DbCommandInfo BuildSelectByKeyCommand(EntityCommand entityCommand) {
      const string SqlSelectByFkTemplate = @"
SELECT {0} {1} 
  FROM {2} 
  WHERE {3} 
  {4} {5}; ";
      var topOne = entityCommand.Flags.IsSet(EntityCommandFlags.SelectOne); 
      string suffix = topOne? "SelectFirstBy" : "SelectBy";
      int topCount = topOne? 1 : Settings.QueryLimit;
      var table = DbModel.LookupDbObject<DbTableInfo>(entityCommand.TargetEntityInfo, throwNotFound: true);
      var dbKey = DbModel.LookupDbObject<DbKeyInfo>(entityCommand.SelectKey); 
      var keyCols = dbKey.Columns.GetNames(removeUnderscores: true);
      var cmdName = Settings.NamingPolicy.ConstructDbCommandName(entityCommand, table.TableName, suffix, keyCols);
      var cmdInfo = CreateDbCommandInfo(entityCommand, cmdName, table, DbExecutionType.Query, null);
      //Build column list
      var strColumns = table.Columns.GetSelectable().GetSqlNameList();
      //build WHERE clause
      //TODO: improve matching of DataMember.ForeignKey to DbConstraint
      var whereExpr = BuildWhereClause(cmdInfo.Parameters);
      string orderByExpr = null;
      if (dbKey.KeyType == KeyType.PrimaryKey)
        orderByExpr = null; 
      else 
        orderByExpr = BuildOrderBy(table.DefaultOrderBy);
      var limit = GetSqlLimitClause(topCount.ToString());
      cmdInfo.Sql = string.Format(SqlSelectByFkTemplate, limit.TopClause, strColumns, table.FullName, whereExpr, orderByExpr, limit.LimitClause);
      return cmdInfo;
    }

    public virtual DbCommandInfo BuildSelectManyToManyCommand(EntityCommand entityCommand) {
      const string SqlTemplate = @"
SELECT {0} {1} 
  FROM {2} L INNER JOIN {3} T ON {4}  
  WHERE {5} {6} {7}; ";
      var parentRefKey = entityCommand.SelectKey;
      var targetRefKey = entityCommand.SelectKeySecondary;
      var targetTable = DbModel.LookupDbObject<DbTableInfo>(entityCommand.TargetEntityInfo);
      var dbParentRefKey = DbModel.LookupDbObject<DbKeyInfo>(parentRefKey);
      var dbTargetRefKey = DbModel.LookupDbObject<DbKeyInfo>(targetRefKey);
      var linkTable = dbParentRefKey.Table;
      var parentRefColNames = dbParentRefKey.Columns.GetNames(removeUnderscores: true);
      var cmdName = Settings.NamingPolicy.ConstructDbCommandName(entityCommand, targetTable.TableName, "SelectBy", parentRefColNames);
      var cmdInfo = CreateDbCommandInfo(entityCommand, cmdName, targetTable, DbExecutionType.Query, null);
      var strOutColumns = targetTable.Columns.GetSelectable().GetSqlNameList("T");
      var whereExpr = BuildWhereClause(cmdInfo.Parameters, "L");
      var orderByExpr = BuildOrderBy(linkTable.DefaultOrderBy, "L");
      string joinOnClause = BuildJoinClause(dbTargetRefKey, targetTable.PrimaryKey, "L", "T");
      var limit = GetSqlLimitClause(Settings.QueryLimit.ToString()); 
      cmdInfo.Sql = string.Format(SqlTemplate, limit.TopClause, strOutColumns, linkTable.FullName, targetTable.FullName, 
        joinOnClause, whereExpr, orderByExpr, limit.LimitClause);
      return cmdInfo;
    }

    protected virtual string BuildOrderBy(IList<DbOrderByEntryInfo> orderByEntries, string tableAlias = null) {
      var cols = BuildOrderByColumns(orderByEntries, tableAlias);
      if (string.IsNullOrEmpty(cols))
        return cols;
      else 
        return "ORDER BY " + cols; 
    }

    protected virtual string BuildOrderByColumns(IList<DbOrderByEntryInfo> orderByEntries, string tableAlias = null) {
      if (orderByEntries == null || orderByEntries.Count == 0)
        return null; 
      if (string.IsNullOrEmpty(tableAlias))
        return string.Join(", ", orderByEntries);
      var delim = ", " + tableAlias + ".";
      return tableAlias + "." + string.Join(delim, orderByEntries);
    }

    protected virtual string BuildWhereClause(List<DbParamInfo> keyParameters, string tableAlias = null) {
      string colPrefix = string.IsNullOrEmpty(tableAlias) ? string.Empty : tableAlias + ".";
      var whereExprs = new StringList();
      foreach (var prm in keyParameters)
        whereExprs.Add(colPrefix + "\"" + prm.SourceColumn.ColumnName + "\"" + " = " + prm.Name);
      return string.Join(" AND ", whereExprs);
    }

    protected virtual string BuildJoinClause(DbKeyInfo key1, DbKeyInfo key2, string table1Alias, string table2Alias) {
      var prefix1 = string.IsNullOrEmpty(table1Alias) ? string.Empty : table1Alias + ".";
      var prefix2 = string.IsNullOrEmpty(table2Alias) ? string.Empty : table2Alias + ".";
      var count = key1.Columns.Count;
      var parts = new string[count];
      for (int i = 0; i < count; i++)
        parts[i] = string.Format("{0}\"{1}\" = {2}\"{3}\"", prefix1, key1.Columns[i].ColumnName, prefix2, key2.Columns[i].ColumnName);
      return string.Join(" AND ", parts);
    }

    public virtual string GetParameterPrefix() {
      var driver = this.Settings.Driver;
      var prefix = this.Settings.Options.IsSet(DbOptions.UseStoredProcs) ? driver.StoredProcParameterPrefix : driver.DynamicSqlParameterPrefix;
      return prefix; 
    }

    public virtual DbCommandInfo CreateDbCommandInfo(EntityCommand entityCommand, string name, DbTableInfo mainTable, 
          DbExecutionType executionType, string sql) {
      var policy = DbModel.Settings.NamingPolicy;
      var prmPrefix = GetParameterPrefix(); 
      var cmdInfo = new DbCommandInfo(entityCommand, name, mainTable, executionType, sql);
      //Create parameters from entity command parameters 
      for(int i=0; i< entityCommand.Parameters.Count; i++) {
        var entParam = entityCommand.Parameters[i];
        var paramName = prmPrefix + policy.ConstructDbParameterName(entParam.Name);
        DbParamInfo prmInfo;
        if (entParam.SourceMember != null) {
          var col = DbModel.LookupDbObject<DbColumnInfo>(entParam.SourceMember);
          prmInfo = cmdInfo.AddParameter(entParam, paramName, col, i);
        } else {
          var typeInfo = this.TypeMapper.GetDbTypeInfo(entParam.DataType, entParam.Size);
          prmInfo = cmdInfo.AddParameter(entParam, paramName, typeInfo, i);
        }
        // SQL CE does not support output parameters
        if (!this.DbModel.Driver.Supports(DbCapabilities.OutputParameters))
          prmInfo.Direction = ParameterDirection.Input;
      }//foreach entParam
      return cmdInfo;
    }

    protected class SqlLimitClause {
      public string TopClause; //Top clause, used by SQL Server 
      public string LimitClause; //Limit/Fetch clause, added at the end; used by MySql
    }
    protected virtual SqlLimitClause GetSqlLimitClause(string count) {
      return new SqlLimitClause() { TopClause = "TOP " + count };
    }


  }//class
}//namespace
