﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data;
using System.Data.Linq;

using Vita.Common;
using Vita.Entities;
using Vita.Entities.Model;
using Vita.Data.Model;
using Vita.Data.Management;
using Vita.Activation;

namespace Vita.Data {

  public class DbTypeMapper {
    protected DbSettings Settings;

    public DbTypeMapper(DbSettings settings) {
      Settings = settings; 
    }

    public virtual DbTypeInfo GetDbTypeInfo(EntityMemberInfo member) {
      var typeInfo = GetDbTypeInfo(member.DataType, member.Size, member.ExplicitDbType, member.Flags, member.Precision, member.Scale);
      if (typeInfo != null && typeInfo.ClrType == null)
        typeInfo.ClrType = member.DataType; 
      return typeInfo;
    }

    // Used by DbModelBuilder to derive DB column information from EntityMember
    public virtual DbTypeInfo GetDbTypeInfo(Type dataType, int size = 0, DbType? explicitDbType = null, 
                                            EntityMemberFlags flags = EntityMemberFlags.None, byte precision = 0, byte scale = 0) {
      var isNullable = flags.IsSet(EntityMemberFlags.Nullable);
      switch (explicitDbType) {
        case null: break;
        case DbType.Binary: return new DbTypeInfo(DbType.Binary, "binary({0})", isNullable, size, writeConverter: BinaryToByteArray, readConverter: ByteArrayToBinary);
        case CustomDbTypes.VarBinary: 
          return new DbTypeInfo(CustomDbTypes.VarBinary, "varbinary({0})", isNullable, size, writeConverter: BinaryToByteArray, readConverter: ByteArrayToBinary);
      }
      if (dataType.IsNullableValueType())
        dataType = dataType.GetValueTypeFromNullable();
      //Some special cases: Enum, Guid, byte arrray
      if (dataType.IsEnum)
        return new DbTypeInfo(DbType.Int32, "int", isNullable);
      if (dataType == typeof(Guid))
        return new DbTypeInfo(DbType.Guid, "uniqueidentifier", isNullable);
      if (dataType == typeof(byte[])) {
        if (size <= 0)
          size = 4000; //default
        // default for byte[] data type is VarBinary, not Binary; Binary may be specified explicitly
        return new DbTypeInfo(CustomDbTypes.VarBinary, "varbinary({0})", isNullable, size);
      }
      if (dataType == typeof(Binary)) {
        if (size <= 0)
          size = 4000; //default
        // default for byte[] data type is VarBinary, not Binary; Binary may be specified explicitly
        return new DbTypeInfo(CustomDbTypes.VarBinary, "varbinary({0})", isNullable, size, writeConverter: BinaryToByteArray, readConverter: ByteArrayToBinary);
      }
    
      //General case - go thru type code
      var typeCode = Type.GetTypeCode(dataType);
      switch (typeCode) {
        // Numeric  
        case TypeCode.Int16:
          return new DbTypeInfo(DbType.Int16, "smallint", isNullable);
        case TypeCode.Int32:
          return new DbTypeInfo(DbType.Int32, "int", isNullable);
        case TypeCode.Int64:
          return new DbTypeInfo(DbType.Int64, "bigint", isNullable);
        case TypeCode.Single:
          return new DbTypeInfo(DbType.Single, "real", isNullable);
        case TypeCode.Double:
          return new DbTypeInfo(DbType.Double, "float", isNullable);
        // Decimal
        case TypeCode.Decimal:
          if (precision == 0) precision = 18;
          if (scale == 0) scale = 4;
          return new DbTypeInfo(DbType.Decimal, "numeric({1},{2})", isNullable, size, null, precision, scale);
        
        case TypeCode.Char:
          switch (explicitDbType) {
            case DbType.AnsiStringFixedLength:
              return new DbTypeInfo(DbType.AnsiStringFixedLength, "char(1)", isNullable, 1, null, 0, 0, s => ((string)s)[0], c => c.ToString());
            case DbType.StringFixedLength:
            default:
              return new DbTypeInfo(DbType.StringFixedLength, "nchar(1)", isNullable, 1, null, 0, 0, s => ((string)s)[0], c => c.ToString());
          }

        case TypeCode.String:
          switch (explicitDbType) {
            case DbType.AnsiStringFixedLength: return new DbTypeInfo(DbType.AnsiStringFixedLength, "char({0})", isNullable, size);
            case DbType.StringFixedLength: return new DbTypeInfo(DbType.StringFixedLength, "nchar({0})", isNullable, size);
            case DbType.AnsiString: return new DbTypeInfo(DbType.AnsiString, "varchar({0})", isNullable, size);
            default:
              return new DbTypeInfo(DbType.String, "nvarchar({0})", isNullable, size);
          }

        case TypeCode.DateTime:
          DbValueConverter readTime = null;
          if (flags.IsSet(EntityMemberFlags.Utc))
            readTime = ReadUtcTime;
          switch (explicitDbType) {
            case DbType.Date: return new DbTypeInfo(DbType.Date, "date", isNullable, readConverter: readTime);
            case DbType.DateTime: return new DbTypeInfo(DbType.DateTime, "datetime", isNullable, readConverter: readTime);
            case DbType.Time: return new DbTypeInfo(DbType.Time, "time", isNullable, readConverter: readTime);
            default:
              return new DbTypeInfo(DbType.DateTime2, "datetime2", isNullable, readConverter: readTime);
          }//switch

        case TypeCode.Boolean:
          return new DbTypeInfo(DbType.Boolean, "bit", isNullable);
        // NOT supported by standard
        default:
          return null;
      }
    }

    private static Byte[] BinaryToByteArray(object value) {
      if (value == null || value == DBNull.Value) 
        return null;
      if (value is byte[]) 
        return (byte[]) value;
      var b = value as Binary;
      if (b != null)
        return b.ToArray();
      Util.Throw("Invalid binary value ({0}), expected byte[] or Binary.", value.GetType());
      return null; 
    }
    private static Binary ByteArrayToBinary(object value) {
      if (value == null || value == DBNull.Value) 
        return null;
      if (value is byte[])
        return new Binary((byte[])value);
      var b = value as Binary;
      if (b != null)
        return b.ToArray();
      Util.Throw("Invalid binary value ({0}), expected byte[] or Binary.", value.GetType());
      return null;
    }


    protected object ReadUtcTime(object value) {
      try {
        if (value == DBNull.Value) return value;
        var dt = (DateTime)value;
        //just in case it is already correct
        if (dt.Kind == DateTimeKind.Utc) 
          return dt;
        // Most likely not, it has kind = Unspecified
        var utc = new DateTime(dt.Ticks, DateTimeKind.Utc);
        return utc; 
      } catch (Exception ex) {
        ex.AddValue("value", value);
        throw;
      }
    }

    // Used by DbModelLoader to derive Db column information from INFORMATION_SCHEMA.COLUMNS row
    public virtual DbTypeInfo GetDbTypeInfo(DataRow schemaRow) {
      var dataTypeString = schemaRow.GetAsString("DATA_TYPE");
      var isNullable = (schemaRow.GetAsString("IS_NULLABLE") == "YES");
      int charSize = schemaRow.GetAsInt("CHARACTER_MAXIMUM_LENGTH");
      int byteSize = schemaRow.GetAsInt("CHARACTER_OCTET_LENGTH");
      string typeSpec = string.Empty; 
      dataTypeString = dataTypeString.ToLowerInvariant(); 
      switch (dataTypeString) {
        case "int":
          return new DbTypeInfo(DbType.Int32, dataTypeString, isNullable);
        case "tinyint":
          return new DbTypeInfo(DbType.Byte, dataTypeString, isNullable);
        case "smallint":
          return new DbTypeInfo(DbType.Int16, dataTypeString, isNullable);
        case "bigint":
          return new DbTypeInfo(DbType.Int64, dataTypeString, isNullable);
        case "real":
          return new DbTypeInfo(DbType.Single, dataTypeString, isNullable);
        case "float":
          return new DbTypeInfo(DbType.Double, dataTypeString, isNullable);
        case "numeric":
        case "decimal":
          var precision = schemaRow.GetAsByte("NUMERIC_PRECISION");
          var scale = schemaRow.GetAsByte("NUMERIC_SCALE");
          typeSpec = string.Format("numeric({0},{1})", precision, scale);  
          return new DbTypeInfo(DbType.Decimal, typeSpec, isNullable, 0, null, precision, scale);
        case "char":
          typeSpec = string.Format("char({0})", charSize);
          return new DbTypeInfo(DbType.AnsiStringFixedLength, typeSpec, isNullable, charSize);
        case "nchar":
          typeSpec = string.Format("nchar({0})", charSize);  
          return new DbTypeInfo(DbType.StringFixedLength, typeSpec, isNullable, charSize);
        case "varchar":
          if (charSize == -1) 
            typeSpec = "text";
          else
            typeSpec = string.Format("varchar({0})", charSize);
          return new DbTypeInfo(DbType.AnsiString, typeSpec, isNullable, charSize);
        case "nvarchar":
          if (charSize == -1) 
            typeSpec = "ntext";
          else
            typeSpec = string.Format("nvarchar({0})", charSize);
          return new DbTypeInfo(DbType.String, typeSpec, isNullable, charSize);

        case "ntext":
          return new DbTypeInfo(DbType.String, dataTypeString, isNullable, -1);
        case "xml":
          return new DbTypeInfo(DbType.Xml, "xml", isNullable, -1, clrType: typeof(string));

        case "date":
          return new DbTypeInfo(DbType.Date, dataTypeString, isNullable);
        case "time": 
          return new DbTypeInfo(DbType.Time, dataTypeString, isNullable);
        case "datetime":
          return new DbTypeInfo(DbType.DateTime, dataTypeString, isNullable);
        case "datetime2":
          return new DbTypeInfo(DbType.DateTime2, dataTypeString, isNullable);
        case "timestamp":
          return new DbTypeInfo(CustomDbTypes.RowVersion, dataTypeString, isNullable, 8);

        case "bit":
          return new DbTypeInfo(DbType.Boolean, dataTypeString, isNullable);
        case "uniqueidentifier":
          return new DbTypeInfo(DbType.Guid, dataTypeString, isNullable);
        case "binary":
          return new DbTypeInfo(DbType.Binary, "binary({0})", isNullable, byteSize);
        case "varbinary":
          typeSpec = string.Format("varbinary({0})", byteSize);
          return new DbTypeInfo(CustomDbTypes.VarBinary, typeSpec, isNullable, byteSize);
        default:
          return null;
      }
    }
/*
    // Used by command compiler to derive DbParamInfo from the parameter of the DB command generated by LINQ provider
    public virtual DbTypeInfo GetDbTypeInfo(IDbDataParameter fromParameter) {
      var prm = fromParameter;
      string typeSpec;
      switch (prm.DbType) {
        case DbType.Int16:
          return new DbTypeInfo(prm.DbType, "smallint", prm.IsNullable, prm.Size);
        case DbType.Int32:
          return new DbTypeInfo(prm.DbType, "int", prm.IsNullable, prm.Size);
        case DbType.Int64:
          return new DbTypeInfo(prm.DbType, "bigint", prm.IsNullable, prm.Size);
        case DbType.Single:
          return new DbTypeInfo(prm.DbType, "real", prm.IsNullable, prm.Size);
        case DbType.Double:
          return new DbTypeInfo(prm.DbType, "float", prm.IsNullable, prm.Size);
        case DbType.Currency:
        case DbType.Decimal:
          typeSpec = string.Format("numeric({0},{1})", prm.Precision, prm.Scale);  
          return new DbTypeInfo(prm.DbType, typeSpec, prm.IsNullable, prm.Size);
        case DbType.String:
          //Overriders - check if size==-1 - in this case it should be memo/text type
          typeSpec = string.Format("nvarchar({0})", prm.Size);
          return new DbTypeInfo(prm.DbType, typeSpec, prm.IsNullable, prm.Size);
        case DbType.StringFixedLength:
          typeSpec = string.Format("nchar({0})", prm.Size);
          return new DbTypeInfo(prm.DbType, typeSpec, prm.IsNullable, prm.Size);

        case DbType.Guid:
          return new DbTypeInfo(prm.DbType, "uniqueidentifier" , prm.IsNullable, prm.Size);
        case DbType.Boolean:
          return new DbTypeInfo(prm.DbType, "bit" , prm.IsNullable, prm.Size);

        case DbType.DateTime:
          return new DbTypeInfo(prm.DbType, "datetime" , prm.IsNullable, prm.Size);
        case DbType.DateTime2:
          return new DbTypeInfo(prm.DbType, "datetime2" , prm.IsNullable, prm.Size);

        case DbType.Binary:
          typeSpec = (prm.Size < 0) ? "binary" : "varbinary({0})";
          return new DbTypeInfo(prm.DbType, typeSpec, prm.IsNullable, prm.Size);
        default:
          return null; 
      }
    }
*/
  }//class
}//ns
