﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Text;
using System.Linq.Expressions;
using System.Diagnostics;
using System.Reflection;

using Vita.Common;
using Vita.Entities.Model;
using Vita.Entities.Runtime;
using Vita.Data.Linq;
using Vita.Entities;

namespace Vita.Data.Linq {

  // The main task is to build runtime query info - some basic information about submitted dynamic query, to allow it to be executed using 
  // previously compiled and cached version. 
  /*
  Tasks
      1. Transform expression to lambda
        - convert arguments of Take, Skip methods to parameters
        - convert references to external object instances to parameters
        - convert reads of static properties/fields to parameters; (by default they are MemberAccessExpression instances)
      2. Create cache key as expr.ToString() + (list of param types)
      3. Collect parameter values. Local variables are all represented by a single parameter - an instance of a 'DisplayClass' generated by LINQ.
            For entity cache, this parameter will be used as-is; for db query, we will create db query params (primitives)
             that will 'read' the values from this displayClass 
   */

  public class EntityQueryRuntimeInfoBuilder : ExpressionVisitor {
    EntityModelInfo _modelInfo; 
    List<ParameterExpression> _params;
    List<object> _paramValues;
    bool _disableQueryCache;
    HashSet<EntityInfo> _entityInfos;

    internal EntityQueryRuntimeInfoBuilder() { }

    public EntityQueryRuntimeInfo Build(EntityQuery query) {
      _modelInfo = query.Session.Model.ModelInfo;
      _params = new List<ParameterExpression>();
      _paramValues = new List<object>();
      _disableQueryCache = false;
      _entityInfos = new HashSet<EntityInfo>(); 
      var convertedExpr = Visit(query.Expression);

      var lambda = Expression.Lambda(convertedExpr, _params);
      string cacheKey = null;
      if (!_disableQueryCache) {
        // We need to add param types to cache key to uniqely identify queries that might have similar shape. Parameters are mostly instances of auto-generated 
        // "DisplayClass'1" instances holding local variables. Reading these variables for query evaluation requires exact match of the DisplayClass types.
        var paramTypes = string.Join(",", _params.Select(p => p.Type.Name));
        cacheKey = _disableQueryCache ? null : lambda.ToString() + "; ParamTypes: " + paramTypes;
      }
      var rtInfo = new EntityQueryRuntimeInfo(lambda, _entityInfos, _paramValues, cacheKey);
      query.RuntimeInfo = rtInfo;
      return rtInfo; 
    }

    protected override Expression VisitMethodCall(MethodCallExpression node) {
      var method = node.Method;
      // Take and Skip arguments are integer constants inside the query expression. We want to transform these into parameters,
      // so that requests for different pages of same query can reuse the common query definition
      if (method.DeclaringType == typeof(Queryable)) {
        switch(method.Name) {
          case "Take": case "Skip":
            var arg0 = Visit(node.Arguments[0]);
            var arg1 = node.Arguments[1] as ConstantExpression;
            var prmArg1 = GetOrCreateParameter(arg1.Value, typeof(int)); 
            return Expression.Call(null, node.Method, arg0, prmArg1);
        }
      }
      return base.VisitMethodCall(node);
    }

    protected override Expression VisitMember(MemberExpression node) {
      if (node == null)
        return null;
      var member = node.Member;
      // Static field/property - convert to parameter
      if (node.Expression == null) {
        var value = member.GetMemberValue(null);
        return GetOrCreateParameter(value, node.Type);
      } 
      // check if it is subquery stored in local variable; if yes, retrieve the value and convert to constant
      if (node.Expression.NodeType == ExpressionType.Constant && node.Type.IsGenericQueryable()) {
        var objExpr = (ConstantExpression) node.Expression;
        var subQuery = (IQueryable) member.GetMemberValue(objExpr.Value); 
        var newExpr = Visit(subQuery.Expression); // we need to visit it as well, to process sub-query nodes
        return newExpr; 
      } //if isqueryable
      
      return base.VisitMember(node);
    }

    protected override Expression VisitConstant(ConstantExpression node) {
      if (node == null || node.Value == null)
        return node;
      var value = node.Value;
      var type = node.Type;
      if (type.IsDbPrimitive())
        return node;
      // Check for EntityQuery/EntitySet
      if (typeof(EntityQuery).IsAssignableFrom(type)) {
        var entQuery = (EntityQuery)node.Value;
        _disableQueryCache |= entQuery.DisableQueryCache; //set DisableQueryCache flag is it is set in any of the sub-queries or entity sets
        // Note: EntitySet's Expression is a constant holding EntitySet itself (we have a loop), so don't visit its expression
        if (entQuery.IsEntitySet) {
          var entInfo = _modelInfo.GetEntityInfo(entQuery.ElementType);
          _entityInfos.Add(entInfo);
        }
        return node; 
      }
      // It is not primitive value, and it is not a subquery or entity set. It is some object - make it a parameter
      var prm = GetOrCreateParameter(value, type);
      return prm;
    }

    private ParameterExpression GetOrCreateParameter(object value, Type type) {
      // See if we had already seen this instance and registered it as a parameter
      var prmIndex = _paramValues.IndexOf(value);
      if (prmIndex >= 0)
        return _params[prmIndex];
      var prmName = "@P" + _params.Count;
      var prm = Expression.Parameter(type, prmName);
      _params.Add(prm);
      _paramValues.Add(value);
      return prm; 
    }


  }//class

}

