﻿
using System;
using System.Collections.Generic;
using System.Linq.Expressions;
using System.Reflection;

using Vita.Data.Linq.Translation.Expressions;
using Vita.Data.Model;

namespace Vita.Data.Linq.Translation {
  internal class BuilderContext {

    public DbModel DbModel;

    public Stack<MethodInfo> CallStack { get; private set; }

    /// <summary>Values coming from the code executing the query - parameters. Top-level parameters for the entire query.</summary>
    public readonly IList<ExternalValueExpression> ExternalValues;

    // Build context: values here are related to current context, and can change with it
    private int _currentScopeIndex;
    public SelectExpression CurrentSelect {
      get { return SelectExpressions[_currentScopeIndex]; }
      set { SelectExpressions[_currentScopeIndex] = value; }
    }
    public IList<SelectExpression> SelectExpressions { get; private set; }
    public IDictionary<Type, MetaTableExpression> MetaTables { get; private set; }
    public IDictionary<string, Expression> LambdaParameters { get; private set; }

    public bool ExpectMetaTableDefinition { get; set; }

    /// <summary>
    /// Helper to enumerate all registered tables
    /// </summary>
    /// <returns></returns>
    public IEnumerable<TableExpression> EnumerateAllTables() {
      foreach (var scopePiece in SelectExpressions) {
        foreach (var table in scopePiece.Tables)
          yield return table;
      }
    }

    /// <summary>
    /// Helper to enumerate all registered columns
    /// </summary>
    /// <returns></returns>
    public IEnumerable<TableExpression> EnumerateScopeTables() {
      for (SelectExpression currentSelect = CurrentSelect; currentSelect != null; currentSelect = currentSelect.Parent) {
        foreach (var table in currentSelect.Tables)
          yield return table;
      }
    }

    /// <summary>
    /// Helper to enumerate all registered columns
    /// </summary>
    /// <returns></returns>
    public IEnumerable<ColumnExpression> EnumerateScopeColumns() {
      for (SelectExpression currentSelect = CurrentSelect; currentSelect != null; currentSelect = currentSelect.Parent) {
        foreach (var column in currentSelect.Columns)
          yield return column;
      }
    }

    public BuilderContext(DbModel dbModel) {
      DbModel = dbModel;
      CallStack = new Stack<MethodInfo>();
      SelectExpressions = new List<SelectExpression>();
      _currentScopeIndex = SelectExpressions.Count;
      SelectExpressions.Add(new SelectExpression());
      ExternalValues = new List<ExternalValueExpression>();
      MetaTables = new Dictionary<Type, MetaTableExpression>();
      LambdaParameters = new Dictionary<string, Expression>();
    }

    public BuilderContext(BuilderContext source) {
      this.DbModel = source.DbModel;
      this.CallStack = source.CallStack;
      this.ExternalValues = source.ExternalValues;
      this.MetaTables = source.MetaTables;
      this.SelectExpressions = source.SelectExpressions;
      this.ExpectMetaTableDefinition = source.ExpectMetaTableDefinition;
      this.LambdaParameters = source.LambdaParameters;
      this._currentScopeIndex = source._currentScopeIndex;
    }


    /// <summary>Creates a new BuilderContext where parameters have a local scope.</summary>
    public BuilderContext NewQuote() {
      var builderContext = new BuilderContext(this);
      // scope dependent Parts
      builderContext.LambdaParameters = new Dictionary<string, Expression>(LambdaParameters);
      return builderContext;
    }

    /// <summary>Creates a new BuilderContext with a new query scope. </summary>
    public BuilderContext NewSelect() {
      var builderContext = new BuilderContext(this);
      // scope dependent Parts
      builderContext._currentScopeIndex = SelectExpressions.Count;
      SelectExpressions.Add(new SelectExpression(CurrentSelect));
      return builderContext;
    }

    /// <summary>Creates a new BuilderContext with a new query scope with the same parent of the CurrentSelect. </summary>
    public BuilderContext NewSisterSelect() {
      var builderContext = new BuilderContext(this);
      builderContext._currentScopeIndex = SelectExpressions.Count;
      SelectExpressions.Add(new SelectExpression(CurrentSelect.Parent));
      return builderContext;
    }

    /// <summary>
    /// Creates a new BuilderContext with a new query scope which is parent of the current one
    /// </summary>
    /// <returns></returns>
    public void NewParentSelect() {
      SelectExpression currentSelect = this.CurrentSelect;
      SelectExpression newParentSelect = new SelectExpression(currentSelect.Parent);
      while (currentSelect != null) {
        currentSelect.Parent = newParentSelect;
        currentSelect = currentSelect.NextSelectExpression;
      }
      this._currentScopeIndex = SelectExpressions.Count;
      SelectExpressions.Add(newParentSelect);
    }

    public bool IsExternalInExpressionChain { get; set; }
  }
}
