using System;
using System.Data;

namespace Vita.Data.Linq.Translation {
    public static class DataRecordExtensions
    {
        // please note that sometimes (depending on driver), GetValue() returns DBNull instead of null
        // so at this level, we handle both

        public static string GetAsString(this IDataRecord dataRecord, int index)
        {
            if (dataRecord.IsDBNull(index))
                return null;
            object o = dataRecord.GetValue(index);
            if (o == null) // this is not supposed to happen
                return null;
            return o.ToString();
        }

        public static bool GetAsBool(this IDataRecord dataRecord, int index)
        {
            object b = dataRecord.GetValue(index);
            return DataRecordHelper.ToBoolean(b);
        }

        public static bool? GetAsNullableBool(this IDataRecord dataRecord, int index)
        {
            if (dataRecord.IsDBNull(index))
                return null;
            return GetAsBool(dataRecord, index);
        }

        public static char GetAsChar(this IDataRecord dataRecord, int index)
        {
            object c = dataRecord.GetValue(index);
            return DataRecordHelper.ToChar(c);
        }

        public static char? GetAsNullableChar(this IDataRecord dataRecord, int index)
        {
            if (dataRecord.IsDBNull(index))
                return null;
            return GetAsChar(dataRecord, index);
        }

        public static U GetAsNumeric<U>(this IDataRecord dataRecord, int index)
        {
            if (dataRecord.IsDBNull(index))
                return default(U);
            return GetAsNumeric<U>(dataRecord.GetValue(index));
        }

        public static U? GetAsNullableNumeric<U>(this IDataRecord dataRecord, int index)
            where U : struct
        {
            if (dataRecord.IsDBNull(index))
                return null;
            return GetAsNumeric<U>(dataRecord.GetValue(index));
        }

        private static U GetAsNumeric<U>(object o)
        {
            if (o == null || o is DBNull)
                return default(U);
            return DataRecordHelper.ToNumber<U>(o);
        }

        public static int GetAsEnum(this IDataRecord dataRecord, Type enumType, int index)
        {
            int enumAsInt = dataRecord.GetAsNumeric<int>(index);
            return enumAsInt;
        }

        public static byte[] GetAsBytes(this IDataRecord dataRecord, int index)
        {
            if (dataRecord.IsDBNull(index))
                return new byte[0];
            object obj = dataRecord.GetValue(index);
            if (obj == null)
                return new byte[0]; //nullable blob?
            byte[] bytes = obj as byte[];
            if (bytes != null)
                return bytes; //works for BLOB field
            Console.WriteLine("GetBytes: received unexpected type:" + obj);
            //return _rdr.GetInt32(index);
            return new byte[0];
        }

        public static System.Data.Linq.Binary GetAsBinary(this IDataRecord dataRecord, int index)
        {
            byte[] bytes = GetAsBytes(dataRecord, index);
            if (bytes.Length == 0)
                return null;
            return new System.Data.Linq.Binary(bytes);
        }

        public static object GetAsObject(this IDataRecord dataRecord, int index)
        {
            if (dataRecord.IsDBNull(index))
                return null;
            object obj = dataRecord.GetValue(index);
            return obj;
        }

        public static DateTime GetAsDateTime(this IDataRecord dataRecord, int index)
        {
            // Convert an InvalidCastException (thrown for example by Npgsql when an
            // operation like "SELECT '2012-'::timestamp - NULL" that is perfectly
            // legal on PostgreSQL returns a null instead of a DateTime) into the
            // correct InvalidOperationException.

            if (dataRecord.IsDBNull(index))
                throw new InvalidOperationException("NULL found where DateTime expected");
            return dataRecord.GetDateTime(index);
        }

        public static DateTime? GetAsNullableDateTime(this IDataRecord dataRecord, int index)
        {
            if (dataRecord.IsDBNull(index))
                return null;
            return GetAsDateTime(dataRecord, index);
        }

        public static DateTimeOffset? GetAsNullableDateTimeOffset(this IDataRecord dataRecord, int index) {
          if (dataRecord.IsDBNull(index))
            return null;
          var value = dataRecord.GetValue(index);
          if (value is DateTimeOffset)
            return (DateTimeOffset)value; 
          //RI: a temp hack to make it work in SQL CE
          if (value is string) {
            return StringToDateTimeOffset((string)value);
          }
          return null; 
        }

        // The following two methods are used as read/write converters for SQL parameters
        //Note: default ToString() representation of DatetimeOffset is local time (!) followed by offset.
        // To make values comparable in database, we use UtcTime in string representation, so we do our custom parsing
        public static DateTimeOffset? StringToDateTimeOffset(string svalue) {
          if (string.IsNullOrWhiteSpace(svalue)) return null;
          var datePart = svalue.Substring(0, 27);
          var offsPart = svalue.Substring(28, 6);
          var utcDate = DateTime.ParseExact(datePart, "yyyy-MM-dd HH:mm:ss.fffffff", null);
          var offs = TimeSpan.Parse(offsPart);
          var localDate = utcDate.Add(offs);
          var result = new DateTimeOffset(localDate, offs); //expects local time
          return result;
        }



        public static Guid GetAsGuid(this IDataRecord dataRecord, int index)
        {
            return dataRecord.GetGuid(index);
        }
        public static Guid? GetAsNullableGuid(this IDataRecord dataRecord, int index)
        {
            if (dataRecord.IsDBNull(index))
                return null;
            return GetAsGuid(dataRecord, index);
        }
    }
}
