﻿using System;
using System.Data;
using System.Linq.Expressions;

using Vita.Common;
using Vita.Data.Model;
using Vita.Entities.Runtime;

namespace Vita.Data.Linq.Translation {

    internal class DataRecordReader 
    {

        /// <summary>
        /// Returns a Expression reading a property from a IDataRecord, at the specified index
        /// The lambda parameters are:
        /// - IDataRecord
        /// - MappingContext
        /// - int (field index)
        /// </summary>
        /// <param name="returnType">The expected return type (to be mapped to the property)</param>
        /// <returns>An expression returning the field value</returns>
        public virtual LambdaExpression GetPropertyReader(Type returnType)
        {
            // if we have a nullable, then use its inner type
            if (returnType.IsNullableValueType())
            {
                var nonNullableReturnType = returnType.GetUnderlyingType();
                return GetNullablePropertyReader(nonNullableReturnType);
            }
            // otherwise, it's simple
            return GetNullablePropertyReader(returnType);
        }

        protected virtual LambdaExpression GetNullablePropertyReader(Type simpleReturnType)
        {
            if (simpleReturnType == typeof(string))
            {
                return (Expression<Func<IDataRecord, EntitySession, int, string>>)((dataRecord, session, valueIndex)
                                                                                    => GetAsString(dataRecord, valueIndex, session));
            }
            if (simpleReturnType == typeof(bool))
            {
                return (Expression<Func<IDataRecord, EntitySession, int, bool?>>)((dataRecord, session, valueIndex)
                                                                                   => dataRecord.GetAsNullableBool(valueIndex));
            }
            if (simpleReturnType == typeof(char))
            {
                return (Expression<Func<IDataRecord, EntitySession, int, char?>>)((dataRecord, session, valueIndex)
                                                                                   => dataRecord.GetAsNullableChar(valueIndex));
            }
            if (simpleReturnType == typeof(byte))
            {
                return (Expression<Func<IDataRecord, EntitySession, int, byte?>>)((dataRecord, session, valueIndex)
                                                                                   => dataRecord.GetAsNullableNumeric<byte>(valueIndex));
            }
            if (simpleReturnType == typeof(sbyte))
            {
                return (Expression<Func<IDataRecord, EntitySession, int, sbyte?>>)((dataRecord, session, valueIndex)
                                                                                    => dataRecord.GetAsNullableNumeric<sbyte>(valueIndex));
            }
            if (simpleReturnType == typeof(short))
            {
                return (Expression<Func<IDataRecord, EntitySession, int, short?>>)((dataRecord, session, valueIndex)
                                                                                    => dataRecord.GetAsNullableNumeric<short>(valueIndex));
            }
            if (simpleReturnType == typeof(ushort))
            {
                return (Expression<Func<IDataRecord, EntitySession, int, ushort?>>)((dataRecord, session, valueIndex)
                                                                                     => dataRecord.GetAsNullableNumeric<ushort>(valueIndex));
            }
            if (simpleReturnType == typeof(int))
            {
                return (Expression<Func<IDataRecord, EntitySession, int, int?>>)((dataRecord, session, valueIndex)
                                                                                  => dataRecord.GetAsNullableNumeric<int>(valueIndex));
            }
            if (simpleReturnType == typeof(uint))
            {
                return (Expression<Func<IDataRecord, EntitySession, int, uint?>>)((dataRecord, session, valueIndex)
                                                                                   => dataRecord.GetAsNullableNumeric<uint>(valueIndex));
            }
            if (simpleReturnType == typeof(long))
            {
                return (Expression<Func<IDataRecord, EntitySession, int, long?>>)((dataRecord, session, valueIndex)
                                                                                   => dataRecord.GetAsNullableNumeric<long>(valueIndex));
            }
            if (simpleReturnType == typeof(ulong))
            {
                return (Expression<Func<IDataRecord, EntitySession, int, ulong?>>)((dataRecord, session, valueIndex)
                                                                                    => dataRecord.GetAsNullableNumeric<ulong>(valueIndex));
            }
            if (simpleReturnType == typeof(float))
            {
                return (Expression<Func<IDataRecord, EntitySession, int, float?>>)((dataRecord, session, valueIndex)
                                                                                    => dataRecord.GetAsNullableNumeric<float>(valueIndex));
            }
            if (simpleReturnType == typeof(double))
            {
                return (Expression<Func<IDataRecord, EntitySession, int, double?>>)((dataRecord, session, valueIndex)
                                                                                     => dataRecord.GetAsNullableNumeric<double>(valueIndex));
            }
            if (simpleReturnType == typeof(decimal))
            {
                return (Expression<Func<IDataRecord, EntitySession, int, decimal?>>)((dataRecord, session, valueIndex)
                                                                                      => dataRecord.GetAsNullableNumeric<decimal>(valueIndex));
            }
            if (simpleReturnType == typeof(DateTime))
            {
                return (Expression<Func<IDataRecord, EntitySession, int, DateTime?>>)((dataRecord, session, valueIndex)
                                                                                       => dataRecord.GetAsNullableDateTime(valueIndex));
            }
            if (simpleReturnType == typeof(DateTimeOffset)) {
              return (Expression<Func<IDataRecord, EntitySession, int, DateTimeOffset?>>)((dataRecord, session, valueIndex)
                                                                                     => dataRecord.GetAsNullableDateTimeOffset(valueIndex));
            }
            if (simpleReturnType == typeof(Guid))
            {
                return (Expression<Func<IDataRecord, EntitySession, int, Guid?>>)((dataRecord, session, valueIndex)
                                                                                       => dataRecord.GetAsNullableGuid(valueIndex));
            }
            if (simpleReturnType == typeof(byte[]))
            {
                return (Expression<Func<IDataRecord, EntitySession, int, byte[]>>)((dataRecord, session, valueIndex)
                                                                                    => dataRecord.GetAsBytes(valueIndex));
            }
            if (simpleReturnType.IsEnum)
            {
                return (Expression<Func<IDataRecord, EntitySession, int, int?>>)((dataRecord, session, valueIndex)
                                                                                  => dataRecord.GetAsNullableNumeric<int>(valueIndex));
            }
            if (simpleReturnType == typeof(System.Data.Linq.Binary))
            {
                return (Expression<Func<IDataRecord, EntitySession, int, System.Data.Linq.Binary>>)((dataRecord, session, valueIndex)
                                                                                    => dataRecord.GetAsBinary(valueIndex));
            }
            // for polymorphic types especially for ExecuteQuery<>()
            if (simpleReturnType == typeof(object))
            {
                return (Expression<Func<IDataRecord, EntitySession, int, object>>)((dataRecord, session, valueIndex)
                                                                                    => GetAsObject(dataRecord, valueIndex, session));
            }
            //s_rdr.GetUInt32();
            //s_rdr.GetFloat();
            string msg = "RowEnum TODO L381: add support for type " + simpleReturnType;
            System.Diagnostics.Debugger.Break();
            System.Diagnostics.Debug.WriteLine(msg);
            //                propertyReader = null;
            //              throw new ApplicationException(msg);
            // TODO: 
            return (Expression<Func<IDataRecord, EntitySession, int, object>>)((dataRecord, session, valueIndex)
                                                                                => GetAsObject(dataRecord, valueIndex, session));
        }

        /// <summary>
        /// Wrapper to call the MappingContext
        /// </summary>
        /// <param name="dataRecord"></param>
        /// <param name="columnIndex"></param>
        /// <param name="session"></param>
        /// <returns></returns>
        protected virtual string GetAsString(IDataRecord dataRecord, int columnIndex, EntitySession session)
        {
            var value = dataRecord.GetAsString(columnIndex);
            return value;
        }

        protected virtual object GetAsObject(IDataRecord dataRecord, int columnIndex, EntitySession session)
        {
            var value = dataRecord.GetAsObject(columnIndex);
            return value;
        }
    }
}