﻿
using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using System.Reflection;

using Vita.Common;
using Vita.Entities.Runtime;
using Vita.Data.Linq.Translation.Expressions;
using Vita.Data.Linq;
using Vita.Entities.Model; 


namespace Vita.Data.Linq.Translation {

    partial class ExpressionDispatcher
    {
        public Expression Analyze(ExpressionChain expressionChain, Expression parameter, BuilderContext builderContext)
        {
            Expression tableExpression = parameter;

            Expression last = expressionChain.Last();
            ExpressionLanguageParser languageParser = new ExpressionLanguageParser(); 
            foreach (Expression e in expressionChain)
            {
                if (e == last)
                    builderContext.IsExternalInExpressionChain = true;

                // write full debug

                // Convert linq Expressions to QueryOperationExpressions and QueryConstantExpressions 
                // Query expressions language identification
                var currentExpression = languageParser.Parse(e, builderContext);
                // Query expressions query identification 
                currentExpression = this.Analyze(currentExpression, tableExpression, builderContext);

                tableExpression = currentExpression;
            }

            return tableExpression;
        }

        /// <summary>
        /// Entry point for Analyzis
        /// </summary>
        /// <param name="expression"></param>
        /// <param name="parameter"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        protected virtual Expression Analyze(Expression expression, Expression parameter, BuilderContext builderContext)
        {
            return Analyze(expression, new[] { parameter }, builderContext);
        }

        protected virtual Expression Analyze(Expression expression, BuilderContext builderContext)
        {
            return Analyze(expression, new Expression[0], builderContext);
        }

        protected virtual Expression Analyze(Expression expression, IList<Expression> parameters, BuilderContext builderContext)
        {
          switch (expression.NodeType)
            {
                case ExpressionType.Call:
                    return AnalyzeCall((MethodCallExpression)expression, parameters, builderContext);
                case ExpressionType.Lambda:
                    return AnalyzeLambda(expression, parameters, builderContext);
                case ExpressionType.Parameter:
                case (ExpressionType) CustomExpressionType.InputParameter:
                    return AnalyzeParameter(expression, builderContext);
                case ExpressionType.Quote:
                    return AnalyzeQuote(expression, parameters, builderContext);
                case ExpressionType.MemberAccess:
                    return AnalyzeMember(expression, builderContext);
                //RI: added this special case
                case ExpressionType.Equal:
                case ExpressionType.NotEqual:
                    var expr = CheckEqualOperator((BinaryExpression)expression, builderContext);
                    return AnalyzeExpression(expr, builderContext);
                case ExpressionType.Add:
                case ExpressionType.AddChecked:
                case ExpressionType.Divide:
                case ExpressionType.Modulo:
                case ExpressionType.Multiply:
                case ExpressionType.MultiplyChecked:
                case ExpressionType.Power:
                case ExpressionType.Subtract:
                case ExpressionType.SubtractChecked:
                case ExpressionType.And:
                case ExpressionType.Or:
                case ExpressionType.ExclusiveOr:
                case ExpressionType.LeftShift:
                case ExpressionType.RightShift:
                case ExpressionType.AndAlso:
                case ExpressionType.OrElse:
                case ExpressionType.GreaterThanOrEqual:
                case ExpressionType.GreaterThan:
                case ExpressionType.LessThan:
                case ExpressionType.LessThanOrEqual:
                case ExpressionType.Coalesce:
                    return AnalyzeBinaryOperator((BinaryExpression)expression, builderContext);
                //case ExpressionType.ArrayIndex
                //case ExpressionType.ArrayLength
                case ExpressionType.Convert:
                case ExpressionType.ConvertChecked:
                case ExpressionType.Negate:
                case ExpressionType.NegateChecked:
                case ExpressionType.Not:
                //case ExpressionType.TypeAs
                case ExpressionType.UnaryPlus:
                    return AnalyzeUnaryOperator((UnaryExpression)expression, builderContext);

                case ExpressionType.New:
                    return AnalyzeNewOperator(expression, builderContext);
                case ExpressionType.Constant:
                    return AnalyzeConstant(expression, builderContext);
                case ExpressionType.Invoke:
                    return AnalyzeInvoke(expression, parameters, builderContext);
                case ExpressionType.MemberInit:
                case ExpressionType.Conditional: //RI: added this
                    return AnalyzeExpression(expression, builderContext);
          }
          return expression;
        }

        protected virtual Expression AnalyzeCall(MethodCallExpression expression, IList<Expression> parameters, BuilderContext builderContext)
        {
            var operands = expression.GetOperands().ToList();
            var operarandsToSkip = expression.Method.IsStatic ? 1 : 0;
            var originalParameters = operands.Skip(parameters.Count + operarandsToSkip);
            var newParameters = parameters.Union(originalParameters).ToList();

            return AnalyzeQueryableCall(expression.Method, newParameters, builderContext) ??
                AnalyzeStringCall(expression.Method, newParameters, builderContext) ??
                AnalyzeMathCall(expression.Method, newParameters, builderContext) ??
                AnalyzeUnknownCall(expression, newParameters, builderContext);
        }

        private Expression AnalyzeQueryableCall(MethodInfo method, IList<Expression> parameters, BuilderContext builderContext)
        {
            if (!(method.DeclaringType == typeof(Queryable) || method.DeclaringType == typeof(Enumerable)))
                return null;
            var popCallStack = PushCallStack(method, builderContext);
            // all methods to handle are listed here:
            // ms-help://MS.VSCC.v90/MS.MSDNQTR.v90.en/fxref_system.core/html/2a54ce9d-76f2-81e2-95bb-59740c85386b.htm
            string methodName = method.Name;
            switch (methodName)
            {
                case "All":
                    return popCallStack(AnalyzeAll(parameters, builderContext));
                case "Any":
                    return popCallStack(AnalyzeAny(parameters, builderContext));
                case "Average":
                    return popCallStack(AnalyzeProjectionQuery(SpecialExpressionType.Average, parameters, builderContext, canHaveFilter: false));
                case "Concat":
                    return popCallStack(AnalyzeSelectOperation(SelectOperatorType.UnionAll, parameters, builderContext));
                case "Contains":
                    return popCallStack(AnalyzeContains(parameters, builderContext));
                case "Count":
                    return popCallStack(AnalyzeProjectionQuery(SpecialExpressionType.Count, parameters, builderContext));
                case "DefaultIfEmpty":
                    return popCallStack(AnalyzeOuterJoin(parameters, builderContext));
                case "Distinct":
                    return popCallStack(AnalyzeDistinct(parameters, builderContext));
                case "Except":
                    return popCallStack(AnalyzeSelectOperation(SelectOperatorType.Exception, parameters, builderContext));
                case "First":
                case "FirstOrDefault":
                    return popCallStack(AnalyzeScalar(methodName, 1, parameters, builderContext));
                case "GroupBy":
                    return popCallStack(AnalyzeGroupBy(parameters, builderContext));
                case "GroupJoin":
                    return popCallStack(AnalyzeGroupJoin(parameters, builderContext));
                case "Intersect":
                    return popCallStack(AnalyzeSelectOperation(SelectOperatorType.Intersection, parameters, builderContext));
                case "Join":
                    return popCallStack(AnalyzeJoin(parameters, builderContext));
                case "Last":
                    return popCallStack(AnalyzeScalar(methodName, null, parameters, builderContext));
                case "Max":
                    return popCallStack(AnalyzeProjectionQuery(SpecialExpressionType.Max, parameters, builderContext, canHaveFilter: false));
                case "Min":
                    return popCallStack(AnalyzeProjectionQuery(SpecialExpressionType.Min, parameters, builderContext, canHaveFilter: false));
                case "OrderBy":
                case "ThenBy":
                    return popCallStack(AnalyzeOrderBy(parameters, false, builderContext));
                case "OrderByDescending":
                case "ThenByDescending":
                    return popCallStack(AnalyzeOrderBy(parameters, true, builderContext));
                case "Select":
                    return popCallStack(AnalyzeSelect(parameters, builderContext));
                case "SelectMany":
                    return popCallStack(AnalyzeSelectMany(parameters, builderContext));
                case "Single":
                case "SingleOrDefault":
                    return popCallStack(AnalyzeScalar(methodName, 2, parameters, builderContext));
                case "Skip":
                    return popCallStack(AnalyzeSkip(parameters, builderContext));
                case "Sum":
                    return popCallStack(AnalyzeProjectionQuery(SpecialExpressionType.Sum, parameters, builderContext, canHaveFilter: false));
                case "Take":
                    return popCallStack(AnalyzeTake(parameters, builderContext));
                case "Union":
                    return popCallStack(AnalyzeSelectOperation(SelectOperatorType.Union, parameters, builderContext));
                case "Where":
                    return popCallStack(AnalyzeWhere(parameters, builderContext));
                default:
                    if (method.DeclaringType == typeof(Queryable))
                        Util.Throw("S0133: Implement QueryMethod Queryable.{0}.", methodName);
                    return popCallStack(null);
            }
        }

        Func<Expression, Expression> PushCallStack(MethodInfo method, BuilderContext builderContext)
        {

            builderContext.CallStack.Push(method);
            Func<Expression, Expression> popCallStack = r =>
            {
                builderContext.CallStack.Pop();
                return r;
            };
            return popCallStack;
        }

        private Expression AnalyzeStringCall(MethodInfo method, IList<Expression> parameters, BuilderContext builderContext)
        {
            if (method.DeclaringType != typeof(string))
                return null;
            var popCallStack = PushCallStack(method, builderContext);
            switch (method.Name)
            {
                case "Contains":
                    return popCallStack(AnalyzeLike(parameters, builderContext));
                case "EndsWith":
                    return popCallStack(AnalyzeLikeEnd(parameters, builderContext));
                case "IndexOf":
                    return popCallStack(AnalyzeGenericSpecialExpressionType(SpecialExpressionType.IndexOf, parameters, builderContext));
                case "Remove":
                    return popCallStack(AnalyzeGenericSpecialExpressionType(SpecialExpressionType.Remove, parameters, builderContext));
                case "Replace":
                    return popCallStack(AnalyzeGenericSpecialExpressionType(SpecialExpressionType.Replace, parameters, builderContext));
                case "StartsWith":
                    return popCallStack(AnalyzeLikeStart(parameters, builderContext));
                case "Substring":
                    return popCallStack(AnalyzeSubString(parameters, builderContext));
                case "ToLower":
                    return popCallStack(AnalyzeToLower(parameters, builderContext));
                case "ToString":
                    return popCallStack(AnalyzeToString(method, parameters, builderContext));
                case "ToUpper":
                    return popCallStack(AnalyzeToUpper(parameters, builderContext));
                case "Trim":
                    return popCallStack(AnalyzeGenericSpecialExpressionType(SpecialExpressionType.Trim, parameters, builderContext));
                case "TrimEnd":
                    return popCallStack(AnalyzeGenericSpecialExpressionType(SpecialExpressionType.RTrim, parameters, builderContext));
                case "TrimStart":
                    return popCallStack(AnalyzeGenericSpecialExpressionType(SpecialExpressionType.LTrim, parameters, builderContext));
                default:
                    Util.Throw("Method not supported in LINQ: String.{0}.", method.Name);
                    return null; 
            }
        }

        private Expression AnalyzeMathCall(MethodInfo method, IList<Expression> parameters, BuilderContext builderContext)
        {
            if (method.DeclaringType != typeof(System.Math))
                return null;
            var popCallStack = PushCallStack(method, builderContext);
            switch (method.Name)
            {
                case "Abs":
                case "Exp":
                case "Floor":
                case "Pow":
                case "Round":
                case "Sign":
                case "Sqrt":
                    return popCallStack(AnalyzeGenericSpecialExpressionType((SpecialExpressionType)Enum.Parse(typeof(SpecialExpressionType), method.Name), parameters, builderContext));
                case "Log":
                    return popCallStack(AnalyzeLog(parameters, builderContext));
                case "Log10":
                    return popCallStack(AnalyzeGenericSpecialExpressionType(SpecialExpressionType.Log, parameters, builderContext));
                default:
                    Util.Throw("S0133: Implement QueryMethod Math.{0}.", method.Name);
                    return null; 
            }
        }

        private Expression AnalyzeUnknownCall(MethodCallExpression expression, IList<Expression> parameters, BuilderContext builderContext)
        {
            var method = expression.Method;
            switch (method.Name)
            {
                case "Parse":
                    if (method.IsStatic && parameters.Count == 1)
                        return AnalyzeParse(method, parameters, builderContext);
                    break;
                case "ToString": // Can we sanity check this type?
                    return AnalyzeToString(method, parameters, builderContext);
                case "Contains":
                  // handle List.Contains
                    if (typeof(IList).IsAssignableFrom(method.DeclaringType) || method.DeclaringType == typeof(Queryable))
                      // AnalyzeContains handles static Queryable.Contains which expects 2 parameters
                      return AnalyzeContains(new Expression[] { expression.Object, expression.Arguments[0] }, builderContext);
                    else
                      Util.Throw("Unsupported version of Contains method.");
                  break; 

                default:
                  //TODO: add support for custom functions through Linq engine extensions
                  Util.Throw("Function {0} not supported in queries", method.Name);
                  break; 

            }

            var args = new List<Expression>();
            foreach (var arg in expression.Arguments)
            {
                Expression newArg = arg;
                var pe = arg as ParameterExpression;
                if (pe != null)
                {
                    if (!builderContext.LambdaParameters.TryGetValue(pe.Name, out newArg))
                        throw new NotSupportedException("Do not currently support expression: " + expression);
                }
                else
                    newArg = Analyze(arg, builderContext);
                args.Add(newArg);
            }
            return Expression.Call(expression.Object, expression.Method, args);
        }

        protected virtual Expression AnalyzeLog(IList<Expression> parameters, BuilderContext builderContext)
        {
            if (parameters.Count == 1)
                return new SpecialExpression(SpecialExpressionType.Ln, parameters.Select(p => Analyze(p, builderContext)).ToList());
            else if (parameters.Count == 2)
                return new SpecialExpression(SpecialExpressionType.Log, parameters.Select(p => Analyze(p, builderContext)).ToList());
            else
                throw new NotSupportedException();
        }

        protected virtual Expression AnalyzeGenericSpecialExpressionType(SpecialExpressionType specialType, IList<Expression> parameters, BuilderContext builderContext)
        {
            return new SpecialExpression(specialType, parameters.Select(p => Analyze(p, builderContext)).ToList());
        }

      
        protected virtual Expression AnalyzeParse(MethodInfo method, IList<Expression> parameters, BuilderContext builderContext)
        {
          throw new Exception("Parse method not supported.");
        }

        protected virtual Expression AnalyzeToString(MethodInfo method, IList<Expression> parameters, BuilderContext builderContext)
        {
            if (parameters.Count != 1)
                throw new ArgumentException();

            Expression parameter = parameters.First();
            Expression parameterToHandle;

            if(parameter.Type.IsNullableValueType())
                parameter = Analyze(Expression.Convert(parameter, parameter.Type.GetUnderlyingType()), builderContext);

            parameterToHandle = Analyze(parameter, builderContext);
            if (parameterToHandle is ExternalValueExpression)
              Util.Throw("ToString() method is not supported for query parameter.");
            bool insideWhere = builderContext.CallStack.Count > 0 && builderContext.CallStack.Peek().Name == "Where";
            if (insideWhere)
              Util.Throw("ToString() method is not supported inside WHERE clause.");
            bool canHandle = parameter.Type.IsDbPrimitive() || parameterToHandle.Type == typeof(string);

            //if (!parameter.Type.IsPrimitive && parameterToHandle.Type != typeof(string))
            if (!canHandle)
              Util.Throw("ToString() can only be translated for primitive types. Type {0} is not primitive.", parameter.Type);
            // RI: new method, based on object.ToString 
            var  convertMethod = ExpressionHelper.GetToStringConverter(parameterToHandle.Type);
            return Expression.Convert(parameterToHandle, typeof(string), convertMethod); // 
           
        }

        static int? FirstIndexOf(Stack<MethodInfo> callStack, string methodName)
        {
            int? index = null;
            callStack.Where((m, i) =>
            {
                if (m.Name == methodName)
                {
                    index = i;
                    return true;
                }
                return false;
            }).FirstOrDefault();
            return index;
        }

        /// <summary>
        /// Limits selection count
        /// </summary>
        /// <param name="parameters"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        protected virtual Expression AnalyzeTake(IList<Expression> parameters, BuilderContext builderContext)
        {
            AddLimit(Analyze(parameters[1], builderContext), builderContext);
            return Analyze(parameters[0], builderContext);
        }

        protected virtual void AddLimit(Expression limit, BuilderContext builderContext)
        {
            var previousLimit = builderContext.CurrentSelect.Limit;
            if (previousLimit != null)
                builderContext.CurrentSelect.Limit = MergeLimits(previousLimit, limit);
            else
                builderContext.CurrentSelect.Limit = limit;
        }

        //FirstOrDefault sets Limit=1; if there's already a limit expression, 
        private Expression MergeLimits(Expression limit1, Expression limit2) {
          int valueLimit;
          if (limit1.IsConst(out valueLimit) && valueLimit <= 1)
            return limit1;
          if (limit2.IsConst(out valueLimit) && valueLimit <= 1)
            return limit2; 
          //othersie throw
          Util.Throw("Multiple limit(Take) clauses not supported.");
          return null; //never happens
        }
        
        /// <summary>
        /// Skip selection items
        /// </summary>
        /// <param name="parameters"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        protected virtual Expression AnalyzeSkip(IList<Expression> parameters, BuilderContext builderContext)
        {
            AddOffset(Analyze(parameters[1], builderContext), builderContext);
            return Analyze(parameters[0], builderContext);
        }

        protected virtual void AddOffset(Expression offset, BuilderContext builderContext)
        {
            var previousOffset = builderContext.CurrentSelect.Offset;
            if (previousOffset != null)
                builderContext.CurrentSelect.Offset = Expression.Add(offset, previousOffset);
            else
                builderContext.CurrentSelect.Offset = offset;
        }

        /// <summary>
        /// Registers a scalar method call for result
        /// </summary>
        /// <param name="methodName"></param>
        /// <param name="limit"></param>
        /// <param name="parameters"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        protected virtual Expression AnalyzeScalar(string methodName, int? limit, IList<Expression> parameters, BuilderContext builderContext)
        {
            builderContext.CurrentSelect.ExecuteMethodName = methodName;
            if (limit.HasValue)
                AddLimit(Expression.Constant(limit.Value), builderContext);
            var table = Analyze(parameters[0], builderContext);
            CheckWhere(table, parameters, 1, builderContext);
            return table;
        }

        /// <summary>
        /// Some methods, like Single(), Count(), etc. can get an extra parameter, specifying a restriction.
        /// This method checks if the parameter is specified, and adds it to the WHERE clauses
        /// </summary>
        /// <param name="table"></param>
        /// <param name="parameters"></param>
        /// <param name="extraParameterIndex"></param>
        /// <param name="builderContext"></param>
        private void CheckWhere(Expression table, IList<Expression> parameters, int extraParameterIndex, BuilderContext builderContext)
        {
            if (extraParameterIndex >= 0 && extraParameterIndex < parameters.Count) // a lambda can be specified here, this is a restriction
                RegisterWhere(Analyze(parameters[extraParameterIndex], table, builderContext), builderContext);
        }

        /// <summary>
        /// Returns a projection method call
        /// </summary>
        /// <param name="specialExpressionType"></param>
        /// <param name="parameters"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        protected virtual Expression AnalyzeProjectionQuery(SpecialExpressionType specialExpressionType, IList<Expression> parameters,
                                                            BuilderContext builderContext, bool canHaveFilter = true)
        {

            if (builderContext.IsExternalInExpressionChain)
            {
                var operand0 = Analyze(parameters[0], builderContext);
                Expression functionOperand = null; 
                Expression projectionOperand;

                if (    builderContext.CurrentSelect.NextSelectExpression != null 
                    ||  builderContext.CurrentSelect.Operands.Count() > 0
                    ||  builderContext.CurrentSelect.Group.Count > 0
                   )
                {
                    // No TableInfo in projection
                    operand0 = new SubSelectExpression(builderContext.CurrentSelect, operand0.Type, "source", null);
                    builderContext.NewParentSelect();

                    // In the new scope we should not have MaximumDatabaseLoad
                    //builderContext.QueryContext.MaximumDatabaseLoad = false;

                    builderContext.CurrentSelect.Tables.Add(operand0 as TableExpression);
                }

                // basically, we have three options for projection methods:
                // - projection on grouped table (1 operand, a GroupExpression)
                // - projection on grouped column (2 operands, GroupExpression and ColumnExpression)
                // - projection on table/column, with optional restriction
                var groupOperand0 = operand0 as GroupExpression;
                if (groupOperand0 != null)
                {
                    if (parameters.Count > 1)
                        projectionOperand = Analyze(parameters[1], groupOperand0.GroupedExpression, builderContext);
                    else
                        projectionOperand = Analyze(groupOperand0.GroupedExpression, builderContext);
                }
                else
                {
                    projectionOperand = operand0;
                    if (parameters.Count > 1)
                      functionOperand = Analyze(parameters[1], operand0, builderContext);
                    int filterIndex = canHaveFilter ? 1 : -1; //special case for Average - its second parameter is NOT filter
                    CheckWhere(projectionOperand, parameters, filterIndex, builderContext);
                }

                if (projectionOperand is TableExpression)
                    projectionOperand = RegisterTable((TableExpression)projectionOperand, builderContext);

                if (groupOperand0 != null) {
                  var childColumns = GetChildColumns(projectionOperand, builderContext);
                  projectionOperand = new GroupExpression(projectionOperand, groupOperand0.KeyExpression, childColumns);
                }

                var opList = new List<Expression>();
                opList.Add(projectionOperand);
                if (functionOperand != null)
                  opList.Add(functionOperand); 
                return new SpecialExpression(specialExpressionType, opList);
            }
            else
            {
                var projectionQueryBuilderContext = builderContext.NewSelect();

                var tableExpression = Analyze(parameters[0], projectionQueryBuilderContext);

                //RI: new stuff - handling grouping with aggregates
                if (tableExpression is GroupExpression && IsAggregate(specialExpressionType)) {
                  var grpExpr = tableExpression as GroupExpression;
                  var srcTable = grpExpr.GroupedExpression as TableExpression;
                  SpecialExpression specialExpr;
                  if (parameters.Count > 1) {
                    var predicate = Analyze(parameters[1], srcTable, projectionQueryBuilderContext);
                    specialExpr = new SpecialExpression(specialExpressionType, tableExpression, predicate);
                  } else
                    specialExpr = new SpecialExpression(specialExpressionType, tableExpression);
                  return specialExpr; 
                }
                //RI: end my special code


                // from here we build a custom clause:
                // <anyClause> ==> "(select count(*) from <table> where <anyClause>)>0"
                // TODO (later...): see if some vendors support native Any operator and avoid this substitution
                if (parameters.Count > 1)
                {
                    var predicate = Analyze(parameters[1], tableExpression, projectionQueryBuilderContext);
                    RegisterWhere(predicate, projectionQueryBuilderContext);
                }
                var specExpr = new SpecialExpression(specialExpressionType, tableExpression);
                var currSelect = projectionQueryBuilderContext.CurrentSelect;
                projectionQueryBuilderContext.CurrentSelect = currSelect.ChangeOperands(new Expression[] {specExpr}, currSelect.Operands);
                // we now switch back to current context, and compare the result with 0
                return projectionQueryBuilderContext.CurrentSelect;
            }
        }

        private bool IsAggregate(SpecialExpressionType type) {
          switch(type) {
            case SpecialExpressionType.Min: case SpecialExpressionType.Max: case SpecialExpressionType.Average: 
            case SpecialExpressionType.Count: case SpecialExpressionType.Sum: 
              return true;
            default:
              return false; 
          }
        }
        /// <summary>
        /// Entry point for a Select()
        /// static Select(this Expression table, λ(table))
        /// </summary>
        /// <param name="parameters"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        protected virtual Expression AnalyzeSelect(IList<Expression> parameters, BuilderContext builderContext)
        {
            // just call back the underlying lambda (or quote, whatever)
            Expression ex = Analyze(parameters[1], parameters[0], builderContext);

/* RI: removing this, the problem mentioned in the article (cache pollution) is not relevant for us here
            // http://social.msdn.microsoft.com/Forums/en-US/linqprojectgeneral/thread/1ce25da3-44c6-407d-8395-4c146930004b
            if (ex.NodeType == ExpressionType.MemberInit &&
                    builderContext.QueryContext.DbModel.Mapping.GetMetaType(ex.Type) != null)
                throw new NotSupportedException(
                    string.Format("Explicit construction of entity type '{0}' in query is not allowed.",
                        ex.Type.FullName));
 */ 
            // RI: changed the following to explicitly check if table included (instead checking for Count==0)
            //TableExpression table = parameters[0] as TableExpression;
            // if (table != null && builderContext.CurrentSelect.Tables.Count == 0)
               // RegisterTable(table, builderContext);
            // RI: Special case - for queries like 'books.Select(b => 1)', no book columns are in output, so books table is not registered - this results in invalid query.
            // to compensate for this, we check and register table explicitly. We need to check for both TableExpression and GroupExpression
            var p0 = parameters[0];
            TableExpression table = null;
            if (p0 is GroupExpression) {
              var gex = (GroupExpression)p0;
               table = gex.GroupedExpression as TableExpression;
            } else 
              table = p0 as TableExpression;
            if (table != null && !builderContext.CurrentSelect.Tables.Contains(table))
              table = RegisterTable(table, builderContext); 
            return ex;
        }

        /// <summary>
        /// Entry point for a Where()
        /// static Where(this Expression table, λ(table))
        /// </summary>
        /// <param name="parameters"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        protected virtual Expression AnalyzeWhere(IList<Expression> parameters, BuilderContext builderContext)
        {
            var tablePiece = parameters[0];
            var where = Analyze(parameters[1], tablePiece, builderContext);
           //RI: bit/bool
            where = ExpressionHelper.CheckNeedConvert(where, typeof(bool));
            RegisterWhere(where, builderContext);
            return tablePiece;
        }

        /// <summary>
        /// Handling a lambda consists in:
        /// - filling its input parameters with what's on the stack
        /// - using the body (parameters are registered in the context)
        /// </summary>
        /// <param name="expression"></param>
        /// <param name="parameters"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        protected virtual Expression AnalyzeLambda(Expression expression, IList<Expression> parameters, BuilderContext builderContext)
        {
            var lambdaExpression = expression as LambdaExpression;
            if (lambdaExpression == null)
                Util.Throw("S0227: Unknown type for AnalyzeLambda() ({0})", expression.GetType());
            // for a lambda, first parameter is body, others are input parameters
            // so we create a parameters stack
            for (int parameterIndex = 0; parameterIndex < lambdaExpression.Parameters.Count; parameterIndex++)
            {
                var parameterExpression = lambdaExpression.Parameters[parameterIndex];
                builderContext.LambdaParameters[parameterExpression.Name] = Analyze(parameters[parameterIndex], builderContext);
            }
            // we keep only the body, the header is now useless
            // and once the parameters have been substituted, we don't pass one anymore
            return Analyze(lambdaExpression.Body, builderContext);
        }

        /// <summary>
        /// When a parameter is used, we replace it with its original value
        /// </summary>
        /// <param name="expression"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        protected virtual Expression AnalyzeParameter(Expression expression, BuilderContext builderContext)
        {
          if (expression is ExternalValueExpression) {
            var inpPrm = (ExternalValueExpression)expression;
            inpPrm.FreeUseCount++;
            return inpPrm; 
          }
            // first check input QUERY parameters (top-level lambda)
            var inputPrm = builderContext.ExternalValues.FirstOrDefault(ip => ip.SourceExpression == expression);
            if (inputPrm != null) {
              inputPrm.FreeUseCount++; 
              return inputPrm;
            }

            var prmExpr = (ParameterExpression)expression;
            Expression unaliasedExpression;
            var parameterName = GetParameterName(expression);
            builderContext.LambdaParameters.TryGetValue(parameterName, out unaliasedExpression);
            if (unaliasedExpression == null)
                Util.Throw("S0257: can not find parameter '{0}'", parameterName);

            #region set alias helper

            // for table...
            var unaliasedTableExpression = unaliasedExpression as TableExpression;
            if (unaliasedTableExpression != null && unaliasedTableExpression.Alias == null)
                unaliasedTableExpression.Alias = parameterName;
            // .. or column
            var unaliasedColumnExpression = unaliasedExpression as ColumnExpression;
            if (unaliasedColumnExpression != null && unaliasedColumnExpression.Alias == null)
                unaliasedColumnExpression.Alias = parameterName;

            #endregion

            //var groupByExpression = unaliasedExpression as GroupByExpression;
            //if (groupByExpression != null)
            //    unaliasedExpression = groupByExpression.ColumnExpression.Table;

            return unaliasedExpression;
        }

        /// <summary>
        /// Analyzes a member access.
        /// This analyzis is down to top: the highest identifier is at bottom
        /// </summary>
        /// <param name="expression"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        protected virtual Expression AnalyzeMember(Expression expression, BuilderContext builderContext)
        {
            var memberExpression = (MemberExpression)expression;

            // RI: added this optimization
            if (memberExpression.Expression != null && memberExpression.Expression.NodeType == ExpressionType.MemberAccess) {
              // possible optimization: convert 'book.Publisher.Id' to 'book.Publisher_id', to avoid unnecessary joins
              var optExpression = OptimizeChainedMemberAccess(memberExpression, builderContext);
              if (optExpression != null)
                return optExpression;
            }

            Expression objectExpression = null;
            //maybe is a static member access like DateTime.Now
            bool isStaticMemberAccess = memberExpression.Member.IsStaticMember();

            var memberInfo = memberExpression.Member;
            var memberType = memberInfo.GetMemberType(); 
            if (!isStaticMemberAccess && memberInfo.Name == "Count")
                return AnalyzeProjectionQuery(SpecialExpressionType.Count, new[] { memberExpression.Expression }, builderContext);

            if (!isStaticMemberAccess)
                objectExpression = Analyze(memberExpression.Expression, builderContext);

            //RI: check if objectExpression is a subquery (EntityQuery) wrapped into a constant. We get this in join queries that use sub-queres
            // saved in local variables
            if (objectExpression.NodeType == ExpressionType.Constant && typeof(EntityQuery).IsAssignableFrom(objectExpression.Type)) {
              var constExpr = (ConstantExpression) objectExpression;
              var entQuery = constExpr.Value as EntityQuery;
              objectExpression = Analyze(entQuery.Expression, builderContext);
            }
            // then see what we can do, depending on object type
            // - MetaTable --> then the result is a table
            // - Table --> the result may be a column or a join
            // - Object --> external parameter or table (can this happen here? probably not... to be checked)


            if (objectExpression is MetaTableExpression)
            {
                var metaTableExpression = (MetaTableExpression)objectExpression;
                var tableExpression = metaTableExpression.GetTableExpression(memberInfo);
                if (tableExpression == null)
                    Util.Throw("S0270: MemberInfo '{0}' not found in MetaTable", memberInfo.Name);
                return tableExpression;
            }

            if (objectExpression is GroupExpression)
            {
                if (memberInfo.Name == "Key")
                    return ((GroupExpression)objectExpression).KeyExpression;
            }

            // if object is a table, then we need a column, or an association
            if (objectExpression is TableExpression) {
                var tableExpression = (TableExpression)objectExpression;
                if (!memberType.IsDbPrimitive()) {
                  var queryAssociationExpression = RegisterAssociation(tableExpression, memberInfo, memberType,
                                                                       builderContext);
                  if (queryAssociationExpression != null) {
                    return queryAssociationExpression;
                  }

                }
                // try to map to column; will throw if member cannot be mapped.
                var queryColumnExpression = RegisterColumn(tableExpression, memberInfo, builderContext);
                return queryColumnExpression;
            }

            // if object is parameter, then we have parameter-derived value
            if (objectExpression is ExternalValueExpression) 
              return DeriveInputParameter((ExternalValueExpression)objectExpression, memberInfo, builderContext);

            // we have here a special cases for nullables
            if (!isStaticMemberAccess && objectExpression.Type != null && objectExpression.Type.IsNullableValueType())
            {
                // Value means we convert the nullable to a value --> use Convert instead (works both on CLR and SQL, too)
                if (memberInfo.Name == "Value")
                    return Expression.Convert(objectExpression, memberType);
                // HasValue means not null (works both on CLR and SQL, too)
                if (memberInfo.Name == "HasValue")
                    return new SpecialExpression(SpecialExpressionType.IsNotNull, objectExpression);
            }


            if (memberInfo.DeclaringType == typeof(DateTime))
                return AnalyzeDateTimeMemberAccess(objectExpression, memberInfo, isStaticMemberAccess);

            // TODO: make this expresion safe (objectExpression can be null here)
            if (objectExpression.Type == typeof(TimeSpan))
                return AnalyzeTimeSpanMemberAccess(objectExpression, memberInfo);


            if (objectExpression is MemberInitExpression)
            {
                var foundExpression = AnalyzeMemberInit((MemberInitExpression)objectExpression, memberInfo, builderContext);
                if (foundExpression != null)
                    return foundExpression;
            }

            return AnalyzeCommonMember(objectExpression, memberInfo, builderContext);
        }

        private ExternalValueExpression DeriveInputParameter(ExternalValueExpression prm, MemberInfo memberInfo, BuilderContext builderContext) {
          prm.FreeUseCount--;
          var newSource = Expression.MakeMemberAccess(prm.SourceExpression, memberInfo);
          var newPrm = new ExternalValueExpression(newSource);
          newPrm.FreeUseCount++;
          builderContext.ExternalValues.Add(newPrm);
          return newPrm; 
        }

        // RI: new  stuff 
        // Converts expressions like 'book.Publisher.Id' to 'book.Publisher_id', to avoid unnecessary joins
        // Note that we cannot perform this optimization on original expression tree; the Publisher_id property is not available 
        //  on IBook interface, so we must first convert IBook to book table, and then create member access expr 'BookTable.Publisher_id'
        private Expression OptimizeChainedMemberAccess(MemberExpression memberExpression, BuilderContext builderContext) {
          var bookPubIdExpr = memberExpression; //just rename it for clarity
          var bookPubExpr = bookPubIdExpr.Expression as MemberExpression;
          var bookExpr = bookPubExpr.Expression;
          if (bookPubExpr.Member.IsStaticMember())
            return null;
          // check that child expr src object (book) is a Table
          var bookTable = builderContext.DbModel.GetTable(bookExpr.Type, throwIfNotFound: false);
          if (bookTable == null)
            return null;
          // check that child expr result (book.Publisher) has Table result type
          var pubTable = builderContext.DbModel.GetTable(bookPubExpr.Type, throwIfNotFound: false);
          if (pubTable == null)
            return null; 
          //check that expr is trying read PK member of the target entity (Id of Publisher)
          var idMemberName = bookPubIdExpr.Member.Name;
          var idMember = pubTable.PrimaryKey.EntityKey.ExpandedMembers.FirstOrDefault(m => m.MemberName == idMemberName);
          if (idMember == null)
            return null;
          //OK, it might be a case for optimization. Analyze bottom parameter (Book reference) 
          var analyzedBookExpr = Analyze(bookExpr, builderContext);
          // Don't optimize it if it is based on input parameter; just return derived parameter expression.
          // note: we cannot return null here, it is too late - bookExpr had been already analyzed, it will mess it all
          if (analyzedBookExpr is ExternalValueExpression) {
            var bookPubPrm = DeriveInputParameter((ExternalValueExpression)analyzedBookExpr, bookPubExpr.Member, builderContext);
            var bookPubIdPrm = DeriveInputParameter(bookPubPrm, bookPubIdExpr.Member, builderContext);
            return bookPubIdPrm;
          }
          // Find FK member on src table (Publisher_id on book entity)
          var idMemberIndex = pubTable.PrimaryKey.EntityKey.ExpandedMembers.IndexOf(idMember); 
          var pubRefMember = bookPubExpr.Member;
          var pubRefMemberInfo = bookTable.Entity.Members.First(m => m.Property == pubRefMember);
          var bookPubIdMemberInfo = pubRefMemberInfo.ReferenceInfo.FromKey.ExpandedMembers[idMemberIndex];
          ColumnExpression queryColumnExpression = RegisterColumn((TableExpression)analyzedBookExpr, bookPubIdMemberInfo.ClassProperty, builderContext);
          return queryColumnExpression; 
        }

        // TODO: (RI:) should be removed most likely
        protected Expression AnalyzeTimeSpanMemberAccess(Expression objectExpression, MemberInfo memberInfo) {
            //A timespan expression can be only generated in a c# query as a DateTime difference, as a function call return or as a paramter
            //this case is for the DateTime difference operation

            if (!(objectExpression is BinaryExpression))
                throw new NotSupportedException();

            var operands = objectExpression.GetOperands();

            bool absoluteSpam = memberInfo.Name.StartsWith("Total");
            string operationKey = absoluteSpam ? memberInfo.Name.Substring(5) : memberInfo.Name;

            Expression currentExpression;
            switch (operationKey)
            {
                case "Milliseconds":
                    currentExpression = Expression.Convert(new SpecialExpression(SpecialExpressionType.DateDiffInMilliseconds, operands.First(), operands.ElementAt(1)), typeof(double));
                    break;
                case "Seconds":
                    currentExpression = Expression.Divide(
                        Expression.Convert(new SpecialExpression(SpecialExpressionType.DateDiffInMilliseconds, operands.First(), operands.ElementAt(1)), typeof(double)),
                        Expression.Constant(1000.0));
                    break;
                case "Minutes":
                    currentExpression = Expression.Divide(
                            Expression.Convert(new SpecialExpression(SpecialExpressionType.DateDiffInMilliseconds, operands.First(), operands.ElementAt(1)), typeof(double)),
                            Expression.Constant(60000.0));
                    break;
                case "Hours":
                    currentExpression = Expression.Divide(
                            Expression.Convert(new SpecialExpression(SpecialExpressionType.DateDiffInMilliseconds, operands.First(), operands.ElementAt(1)), typeof(double)),
                            Expression.Constant(3600000.0));
                    break;
                case "Days":
                    currentExpression = Expression.Divide(
                            Expression.Convert(new SpecialExpression(SpecialExpressionType.DateDiffInMilliseconds, operands.First(), operands.ElementAt(1)), typeof(double)),
                            Expression.Constant(86400000.0));
                    break;
                default:
                    throw new NotSupportedException(string.Format("The operation {0} over the TimeSpan isn't currently supported", memberInfo.Name));
            }

            if (!absoluteSpam)
            {
                switch (memberInfo.Name)
                {
                    case "Milliseconds":
                        currentExpression = Expression.Convert(Expression.Modulo(Expression.Convert(currentExpression, typeof(long)), Expression.Constant(1000L)), typeof(int));
                        break;
                    case "Seconds":
                        currentExpression = Expression.Convert(Expression.Modulo(Expression.Convert(currentExpression, typeof(long)),
                                                              Expression.Constant(60L)), typeof(int));
                        break;
                    case "Minutes":
                        currentExpression = Expression.Convert(Expression.Modulo(Expression.Convert(currentExpression, typeof(long)),
                                                                Expression.Constant(60L)), typeof(int));
                        break;
                    case "Hours":
                        currentExpression = Expression.Convert(Expression.Modulo(Expression.Convert(
                                                                                        currentExpression, typeof(long)),
                                                                Expression.Constant(24L)), typeof(int));
                        break;
                    case "Days":
                        currentExpression = Expression.Convert(currentExpression, typeof(int));
                        break;
                }

            }
            return currentExpression;
        }

        protected Expression AnalyzeDateTimeMemberAccess(Expression objectExpression, MemberInfo memberInfo, bool isStaticMemberAccess)
        {
            if (isStaticMemberAccess)
            {
                if (memberInfo.Name == "Now")
                    return new SpecialExpression(SpecialExpressionType.Now);
                else
                    throw new NotSupportedException(string.Format("DateTime Member access {0} not supported", memberInfo.Name));
            }
            else
            {
                switch (memberInfo.Name)
                {
                    case "Year":
                        return new SpecialExpression(SpecialExpressionType.Year, objectExpression);
                    case "Month":
                        return new SpecialExpression(SpecialExpressionType.Month, objectExpression);
                    case "Day":
                        return new SpecialExpression(SpecialExpressionType.Day, objectExpression);
                    case "Hour":
                        return new SpecialExpression(SpecialExpressionType.Hour, objectExpression);
                    case "Minute":
                        return new SpecialExpression(SpecialExpressionType.Minute, objectExpression);
                    case "Second":
                        return new SpecialExpression(SpecialExpressionType.Second, objectExpression);
                    case "Millisecond":
                        return new SpecialExpression(SpecialExpressionType.Millisecond, objectExpression);
                    case "Date":
                        return new SpecialExpression(SpecialExpressionType.Date, objectExpression);
                    default:
                        throw new NotSupportedException(string.Format("DateTime Member access {0} not supported", memberInfo.Name));
                }
            }
        }

        /// <summary>
        /// This method analyzes the case of a new followed by a member access
        /// for example new "A(M = value).M", where the Expression can be reduced to "value"
        /// Caution: it may return null if no result is found
        /// </summary>
        /// <param name="expression"></param>
        /// <param name="memberInfo"></param>
        /// <param name="builderContext"></param>
        /// <returns>A member initializer or null</returns>
        protected virtual Expression AnalyzeMemberInit(MemberInitExpression expression, MemberInfo memberInfo,
                                                       BuilderContext builderContext)
        {
            // TODO: a method for NewExpression that we will use directly from AnalyzeMember and indirectly from here
            foreach (var binding in expression.Bindings)
            {
                var memberAssignment = binding as MemberAssignment;
                if (memberAssignment != null)
                {
                    if (memberAssignment.Member == memberInfo)
                        return memberAssignment.Expression;
                }
            }
            return null;
        }

        protected virtual Expression AnalyzeCommonMember(Expression objectExpression, MemberInfo memberInfo, BuilderContext builderContext)
        {
            if (typeof(string).IsAssignableFrom(objectExpression.Type))
            {
                switch (memberInfo.Name)
                {
                    case "Length":
                        return new SpecialExpression(SpecialExpressionType.StringLength, objectExpression);
                }
            }
            //Util.Throw("S0324: Don't know how to handle Piece");
            return Expression.MakeMemberAccess(objectExpression, memberInfo);
        }

        /// <summary>
        /// A Quote creates a new local context, outside which created parameters disappear
        /// This is why we clone the BuilderContext
        /// </summary>
        /// <param name="piece"></param>
        /// <param name="parameters"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        protected virtual Expression AnalyzeQuote(Expression piece, IList<Expression> parameters, BuilderContext builderContext)
        {
            var builderContextClone = builderContext.NewQuote();
            var firstExpression = piece.GetOperands().First();
            return Analyze(firstExpression, parameters, builderContextClone);
        }

        /// <summary>
        /// </summary>
        /// <param name="expression"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        protected virtual Expression AnalyzeOperatorSubstract(Expression expression, BuilderContext builderContext)
        {
            return AnalyzeExpression(expression, builderContext);
        }

        //Checks operands of Join expression: if it returns Entity, replaces it with PK of entity
        protected virtual Expression CheckJoinKeySelector(Expression selector, BuilderContext builderContext) {
          var tmp = selector;
          if (tmp.NodeType == ExpressionType.Quote) 
            tmp = ((UnaryExpression) tmp).Operand;
          // It should be lambda
          var lambda = (LambdaExpression) tmp;
          var retType = lambda.Body.Type;
          if (retType.IsDbPrimitive())
            return selector;
          var newBody = GetPrimaryKeyAccessor(lambda.Body, builderContext);
          if (newBody == null)
            return selector; 
          //Build new quoted lambda
          var newSelector = Expression.Quote(Expression.Lambda(newBody, lambda.Parameters));
          return newSelector;           
        }

        //RI: new stuff to handle 3 special cases: 1. Entity comparison 2. Comparison with NULL 3. Equals of bools (rejected)
        // Comparison with null used to be in ExpressionOptimizer, moved here - it is not optimization, it is essential for correct SQL
        protected virtual Expression CheckEqualOperator(BinaryExpression expression, BuilderContext builderContext) {
          var left = expression.Left;
          var right = expression.Right;
          // Check bool operands - reject
          // TODO: check the case when left & right are bit columns
          // if (left.Type == typeof(bool))
            // Util.Throw("Equal operator is not supported for bool operands.");
          // Try getting PK expressions - not null if we have entity references; replace expression with comparison of PKs
          var leftKey = GetPrimaryKeyAccessor(left, builderContext);
          var rightKey = GetPrimaryKeyAccessor(right, builderContext);
          if (leftKey != null && rightKey != null)
             return Expression.MakeBinary(expression.NodeType, leftKey, rightKey);
          // check if it is null comparison
          var specialExprType = expression.NodeType == ExpressionType.Equal ? SpecialExpressionType.IsNull : SpecialExpressionType.IsNotNull;
          if (leftKey != null && right.IsNullConstant())
            return new SpecialExpression(specialExprType, leftKey);
          if (rightKey != null && left.IsNullConstant())
            return new SpecialExpression(specialExprType, rightKey);
          if (left.IsNullConstant())
            return new SpecialExpression(specialExprType, right);
          if (right.IsNullConstant())
            return new SpecialExpression(specialExprType, left);
          return expression;
        }

        private MemberExpression GetPrimaryKeyAccessor(Expression expression, BuilderContext builderContext) {
          if (expression.Type.IsDbPrimitive())
            return null; 
          var table = builderContext.DbModel.GetTable(expression.Type, throwIfNotFound: false);
          if (table == null)
            return null;
          var pkMembers = table.PrimaryKey.EntityKey.Members;
          if (pkMembers.Count > 1)
            Util.Throw("Equal operator for entities with composite key is not supported. Expression: {0}.", expression);
          var pkProp = pkMembers[0].Property; // expression.Type.FindProperty(pkMembers[0].MemberName);
          var newExpr = Expression.MakeMemberAccess(expression, pkProp);
          return newExpr;
        }

        // generic, non specific node analysis
        protected virtual Expression AnalyzeExpression(Expression expression, BuilderContext builderContext) {
            var operands = expression.GetOperands().ToList();
            var newOperands = new List<Expression>(); 
            for (int operandIndex = 0; operandIndex < operands.Count; operandIndex++)
            {
                var operand = operands[operandIndex];
                newOperands.Add(Analyze(operand, builderContext));
            }

            return expression.ChangeOperands(newOperands, operands);
        }

        protected virtual Expression AnalyzeUnaryOperator(UnaryExpression expression, BuilderContext builderContext) {
          string parameterName;
          if (expression.NodeType == ExpressionType.Convert &&
                  expression.Method == null &&
                  (parameterName = GetParameterName(expression.Operand)) != null) {
            Expression unaliasedExpression;
            builderContext.LambdaParameters.TryGetValue(parameterName, out unaliasedExpression);
            var unaliasedTableExpression = unaliasedExpression as TableExpression;
            if (unaliasedTableExpression != null)
              return unaliasedTableExpression;
          }
          var newOp = Analyze(expression.Operand, builderContext);
          if (expression.Type == typeof(bool))
            newOp = CheckBoolBitExpression(newOp); //convert it if it is bit field
          return expression.ChangeOperands(new Expression[] { newOp }, new Expression[] { expression.Operand });
        }

        protected virtual Expression AnalyzeBinaryOperator(BinaryExpression expression, BuilderContext builderContext) {
          // Detect if it is a logical expression - if yes, check for bit column argument
          var isLogicOp = false;
          if (expression.Type == typeof(bool)) {
            switch (expression.NodeType) {
              case ExpressionType.And:  case ExpressionType.AndAlso:  case ExpressionType.Or: case ExpressionType.OrElse: isLogicOp = true; break;
            }
          }//if
          // process arguments
          var ops = expression.GetOperands();
          var newOps = new List<Expression>(); 
          foreach(var op in ops) {
            var newOp = Analyze(op, builderContext);
            if (isLogicOp)
              newOp = CheckBoolBitExpression(newOp); 
            newOps.Add(newOp); 
          }
          var newExpr = expression.ChangeOperands(newOps, ops);
          // Check bitwise op
          var result = CheckBitwiseOp(newExpr);
          return result; 
        }

        private Expression CheckBitwiseOp(BinaryExpression expression) {
          if (!expression.Type.IsInt())
            return expression; 
          switch(expression.NodeType) {
            case ExpressionType.And:
            case ExpressionType.Or:
            case ExpressionType.ExclusiveOr:
              return new SpecialExpression(expression.NodeType.ToBitwise(), expression.Left, expression.Right);
          }//switch
          return expression; 
        }

        // Translates bit column reference into an expression comparing with 1:
        //  bitCol -->   Convert(bitCol, typeof(int)) == 1
        // The Convert call is artificial, it will be eliminated during SQL translation, and resulting SQL will be smth like ' WHERE BitCol = 1; '
        public static Expression CheckBoolBitExpression(Expression expression) {
          var colExpr = expression as ColumnExpression;
          if (colExpr != null) {
            var bitExpr = Expression.Convert(colExpr, typeof(int));
            return Expression.Equal(bitExpr, Expression.Constant(1));
          }
          var constExpr = expression as ConstantExpression;
          if (constExpr != null && constExpr.Type == typeof(bool)) {
            var value = (bool)constExpr.Value;
            var bitValue = value ? 1 : 0;
            return Expression.Constant(bitValue);
          }
          //otherwise return expression
          return expression;
        }


        protected virtual Expression AnalyzeNewOperator(Expression expression, BuilderContext builderContext)  {
            if (builderContext.ExpectMetaTableDefinition)
            {
                // first, check if we have a MetaTable definition
                Type metaType;
                var typeInitializers = GetTypeInitializers((NewExpression)expression, true, out metaType);
                var aliases = new Dictionary<MemberInfo, MutableExpression>();
                foreach (var memberInfo in typeInitializers.Keys)
                {
                    var e = Analyze(typeInitializers[memberInfo], builderContext);
                    var tableExpression = e as TableExpression;
                    if (tableExpression != null)
                    {
                        aliases[memberInfo] = tableExpression;
                    }
                    else
                    {
                        aliases[memberInfo] = Analyze(typeInitializers[memberInfo], builderContext) as MetaTableExpression;
                    }
                }
                if (IsMetaTableDefinition(aliases))
                    return RegisterMetaTable(metaType, aliases, builderContext);
            }
            return AnalyzeExpression(expression, builderContext);
        }

        protected virtual bool IsMetaTableDefinition(IDictionary<MemberInfo, MutableExpression> aliases)
        {
            if (aliases.Count != 2)
                return false;
            foreach (var tableExpression in aliases.Values)
            {
                if (tableExpression == null)
                    return false;
            }
            return true;
        }

        /// <summary>
        /// SelectMany() joins tables
        /// </summary>
        /// <param name="parameters"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        protected virtual Expression AnalyzeSelectMany(IList<Expression> parameters, BuilderContext builderContext)
        {
            if (parameters.Count == 3)
            {
                // ms-help://MS.VSCC.v90/MS.MSDNQTR.v90.en/fxref_system.core/html/3371348f-7811-b0bc-8c0a-2a595e08e086.htm
                var tableExpression = parameters[0];
                var projectionExpression = Analyze(parameters[1], new[] { tableExpression }, builderContext);
                //var manyPiece = Analyze(parameters[2], new[] { tableExpression, projectionExpression }, builderContext);
                // from here, our manyPiece is a MetaTable definition
                //var newExpression = manyPiece as NewExpression;
                //if (newExpression == null)
                //    Util.Throw("S0377: Expected a NewExpression as SelectMany() return value");
                //Type metaTableType;
                //var associations = GetTypeInitializers<TableExpression>(newExpression, true, out metaTableType);
                //return RegisterMetaTable(metaTableType, associations, builderContext);
                var metaTableDefinitionBuilderContext = new BuilderContext(builderContext);
                metaTableDefinitionBuilderContext.ExpectMetaTableDefinition = true;
                var expression = Analyze(parameters[2], new[] { tableExpression, projectionExpression },
                                         metaTableDefinitionBuilderContext);
                return expression;
            }
            Util.Throw("S0358: Don't know how to handle this SelectMany() overload ({0} parameters)", parameters.Count);
            return null; 
        }

        protected virtual IDictionary<MemberInfo, Expression> GetTypeInitializers(NewExpression newExpression,
                                                                            bool checkCast, out Type metaType)
        {
            var associations = new Dictionary<MemberInfo, Expression>();
            metaType = null;
            for (int ctorParameterIndex = 0; ctorParameterIndex < newExpression.Arguments.Count; ctorParameterIndex++)
            {
                var aliasedExpression = newExpression.Arguments[ctorParameterIndex];
                if (aliasedExpression == null && checkCast)
                    Util.Throw("S0541: Expected an specific Expression type for GetTypeInitializers()");
                var memberInfo = newExpression.Members[ctorParameterIndex];
                metaType = memberInfo.ReflectedType;
                // the property info is the reflecting property for the memberInfo, if memberInfo is a get_*
                // otherwise we keep the memberInfo as is, since it is a field
                var propertyInfo = GetExposingProperty(memberInfo) ?? memberInfo;
                associations[propertyInfo] = aliasedExpression;
            }
            if (metaType == null && checkCast)
                Util.Throw("S0550: Empty NewExpression found"); // this should never happen, otherwise we may simply ignore it or take the type from elsewhere
            return associations;
        }

        /// <summary> If memberInfo is a method related to a property, returns the PropertyInfo. </summary>
        private PropertyInfo GetExposingProperty(MemberInfo memberInfo) {
          var prop = memberInfo.ReflectedType.GetProperties().FirstOrDefault(p => p.GetMethod == memberInfo || p.SetMethod == memberInfo);
          return prop; 
        }

        //protected virtual IDictionary<MemberInfo, E> GetTypeInitializers<E>(NewExpression newExpression)
        //    where E : Expression
        //{
        //    Type metaType;
        //    return GetTypeInitializers<E>(newExpression, out metaType);
        //}

        /// <summary>
        /// Analyzes a Join statement (explicit join)
        /// </summary>
        /// <param name="parameters"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        protected virtual Expression AnalyzeJoin(IList<Expression> parameters, BuilderContext builderContext)
        {
            return AnalyzeJoin(parameters, TableJoinType.Inner, builderContext);
        }

        /// <summary>
        /// Analyzes a Join statement (explicit join)
        /// </summary>
        /// <param name="parameters"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        protected virtual Expression AnalyzeGroupJoin(IList<Expression> parameters, BuilderContext builderContext)
        {
            return AnalyzeJoin(parameters, TableJoinType.Inner, builderContext);
        }

        protected virtual Expression AnalyzeOuterJoin(IList<Expression> parameters, BuilderContext builderContext)
        {
            var expression = Analyze(parameters[0], builderContext);
            var tableExpression = expression as TableExpression;
            if (tableExpression != null)
            {
                tableExpression.SetOuterJoin();
            }
            return expression;
        }

        private Expression AnalyzeJoin(IList<Expression> parameters, TableJoinType joinType, BuilderContext builderContext)
        {
            if (parameters.Count == 5)
            {
                var outerExpr = Analyze(parameters[0], builderContext);
                var innerExpr = Analyze(parameters[1], builderContext);
                var innerTable = innerExpr as TableExpression;
                // TODO: fix this. When joined table is a subquery, we have this exception
                if (innerTable == null)
                   Util.Throw("Join with sub-query is not supported. Sub-query: {0}.", innerExpr);
                // RI: check if key selectors return Entity: if yes, change to PK
                var outerSel = CheckJoinKeySelector(parameters[2], builderContext);
                var innerSel = CheckJoinKeySelector(parameters[3], builderContext);
                var outerKeySelector = Analyze(outerSel, outerExpr, builderContext);
                var innerKeySelector = Analyze(innerSel, innerTable, builderContext);
                // from here, we have two options to join:
                // 1. left and right are tables, we can use generic expressions (most common)
                // 2. left is something else (a meta table)
                var outerTable = outerExpr as TableExpression;
                if (outerTable == null)
                {
                    var outerKeyColumn = outerKeySelector as ColumnExpression;
                    if (outerKeyColumn == null)
                        Util.Throw("S0701: No way to find left table for Join");
                    outerTable = outerKeyColumn.Table;
                }
                var joinExpr = ExpressionHelper.MakeBinary(ExpressionType.Equal, outerKeySelector, innerKeySelector);
                innerTable.Join(joinType, outerTable, joinExpr,
                                string.Format("join{0}", builderContext.EnumerateAllTables().Count()));
                // last part is lambda, with two tables as parameters
                var metaTableDefinitionBuilderContext = new BuilderContext(builderContext);
                metaTableDefinitionBuilderContext.ExpectMetaTableDefinition = true;
                var expression = Analyze(parameters[4], new[] { outerExpr, innerTable }, metaTableDefinitionBuilderContext);
                return expression;
            }
            Util.Throw("S0530: Don't know how to handle GroupJoin() with {0} parameters", parameters.Count);
            return null; 
        }

        /// <summary>
        /// "Distinct" means select X group by X
        /// </summary>
        /// <param name="parameters"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        protected virtual Expression AnalyzeDistinct(IList<Expression> parameters, BuilderContext builderContext)
        {
            var expression = Analyze(parameters[0], builderContext);
            // we select and group by the same criterion
            // RI: adding explicit list of columns, to catch situation of duplicate column names in joins
            var childColumns = GetChildColumns(expression, builderContext); 
            // some providers (ex SQL CE) do not allow Text columns in distinct output
            CheckDistinctClauseColumns(childColumns, builderContext); 
            
            var group = new GroupExpression(expression, expression, childColumns);
            if (builderContext.CurrentSelect.NextSelectExpression != null)
            {
                var tableInfo = builderContext.DbModel.GetTable(expression.Type);
                expression = new SubSelectExpression(builderContext.CurrentSelect, expression.Type, "source", tableInfo);
                builderContext.NewParentSelect();

                // In the new scope we should not have MaximumDatabaseLoad
                //builderContext.QueryContext.MaximumDatabaseLoad = false;

                builderContext.CurrentSelect.Tables.Add(expression as TableExpression);
            }
            builderContext.CurrentSelect.Group.Add(group);
            
            //RI: added this special check
            // var table  
            // var cols = RegisterAllColumns()

            // "Distinct" method is equivalent to a GroupBy
            // but for some obscure reasons, Linq expects a IQueryable instead of an IGrouping
            // so we return the column, not the group
            return expression;
        }

        //RI: special case. Some providers (SQL CE) do not allow Memo columns in DISTINCT clause
        private void CheckDistinctClauseColumns(IList<ColumnExpression> columns, BuilderContext builderContext) {
          var driver = builderContext.DbModel.Driver;
          if (driver.Supports(Vita.Data.Driver.DbCapabilities.NoMemoInDistinct)) {
            var memos = columns.Where(c => c.ColumnInfo.Member.Flags.IsSet(EntityMemberFlags.IsMemo)).Select(c => c.Name).ToList();
            if (memos.Count > 0)
              Util.Throw("Memo columns are not allowed in DISTINCT clause: {0}", string.Join(", ", memos));
          }
        }

        //RI: adding this to get all child columns in GroupExpression
        private IList<ColumnExpression> GetChildColumns(Expression expression, BuilderContext builderContext) {
          var columns = new List<ColumnExpression>();
          // extract columns (for SQL build)
          expression.Recurse(
              delegate(Expression e) {
                if (e is ColumnExpression)
                  columns.Add((ColumnExpression)e);
                else if (e is TableExpression) {
                  var t = (TableExpression)e;
                  columns.AddRange(RegisterAllColumns(t, builderContext));
                }
                return e;
              }//
          );
          return columns; 
        }

        /// <summary>
        /// Creates a group by clause
        /// </summary>
        /// <param name="parameters"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        protected virtual Expression AnalyzeGroupBy(IList<Expression> parameters, BuilderContext builderContext)
        {
            var table = Analyze(parameters[0], builderContext);
            var keyExpression = Analyze(parameters[1], table, builderContext);

            Expression result = null;
            if (parameters.Count == 2)
                result = table; // we return the whole table
            else if (parameters.Count == 3) {
              Expression[] prms = null;
              // there are 2 overloads
              var selector = ExpressionHelper.Unqoute(parameters[2]) as LambdaExpression;
              if (selector != null) {
                switch(selector.Parameters.Count) {
                  case 2: prms = new Expression[] { keyExpression, table }; break;
                  case 3: prms = new Expression[] { table }; break; 
                }
                if (prms != null)
                  result = Analyze(parameters[2], prms, builderContext); // 3 parameters for a projection expression
              }
            } 
            if (result == null)
              throw Util.Throw("This overload of GroupBy is not supported.", parameters.Count);

            var childColumns = GetChildColumns(keyExpression, builderContext);
            var group = new GroupExpression(result, keyExpression, childColumns);
            builderContext.CurrentSelect.Group.Add(group);
            return group;
        }

        /// <summary>
        /// All() returns true if the given condition satisfies all provided elements
        /// </summary>
        /// <param name="parameters"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        protected virtual Expression AnalyzeAll(IList<Expression> parameters, BuilderContext builderContext)
        {
            var allBuilderContext = builderContext.NewSelect();
            var tableExpression = Analyze(parameters[0], allBuilderContext);
            var allClause = Analyze(parameters[1], tableExpression, allBuilderContext);
            // from here we build a custom clause:
            // <allClause> ==> "(select count(*) from <table> where not <allClause>)==0"
            // TODO (later...): see if some vendors support native All operator and avoid this substitution
            var whereExpression = Expression.Not(allClause);
            RegisterWhere(whereExpression, allBuilderContext);
            var countExpr = new SpecialExpression(SpecialExpressionType.Count, tableExpression);
            var currSelect = allBuilderContext.CurrentSelect;
            allBuilderContext.CurrentSelect = currSelect.ChangeOperands(new Expression[] {countExpr}, currSelect.Operands);
            // TODO: see if we need to register the tablePiece here (we probably don't)

            // we now switch back to current context, and compare the result with 0
            var allExpression = Expression.Equal(allBuilderContext.CurrentSelect, Expression.Constant(0));
            return allExpression;
        }

        /// <summary>
        /// Any() returns true if the given condition satisfies at least one of provided elements
        /// </summary>
        /// <param name="parameters"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        protected virtual Expression AnalyzeAny(IList<Expression> parameters, BuilderContext builderContext)
        {
            if (builderContext.IsExternalInExpressionChain)
            {
                var tableExpression = Analyze(parameters[0], builderContext);
                Expression projectionOperand;

                if (builderContext.CurrentSelect.NextSelectExpression != null)
                {
                    TableExpression currentTableExpression = tableExpression as TableExpression;
                    tableExpression = new SubSelectExpression(builderContext.CurrentSelect, currentTableExpression.Type, "source", currentTableExpression.TableInfo);
                    builderContext.NewParentSelect();

                    // In the new scope we should not have MaximumDatabaseLoad
                    //builderContext.QueryContext.MaximumDatabaseLoad = false;

                    builderContext.CurrentSelect.Tables.Add(tableExpression as TableExpression);
                }

                // basically, we have three options for projection methods:
                // - projection on grouped table (1 operand, a GroupExpression)
                // - projection on grouped column (2 operands, GroupExpression and ColumnExpression)
                // - projection on table/column, with optional restriction
                var groupOperand0 = tableExpression as GroupExpression;
                if (groupOperand0 != null)
                {
                    if (parameters.Count > 1)
                    {
                        projectionOperand = Analyze(parameters[1], groupOperand0.GroupedExpression,
                                                    builderContext);
                    }
                    else
                        projectionOperand = Analyze(groupOperand0.GroupedExpression, builderContext);
                }
                else
                {
                    projectionOperand = tableExpression;
                    CheckWhere(projectionOperand, parameters, 1, builderContext);
                }

                if (projectionOperand is TableExpression)
                    projectionOperand = RegisterTable((TableExpression)projectionOperand, builderContext);

                if (groupOperand0 != null) {
                  var childColumns = GetChildColumns(groupOperand0.KeyExpression, builderContext);
                  projectionOperand = new GroupExpression(projectionOperand, groupOperand0.KeyExpression, childColumns);
                }

                return Expression.GreaterThan(new SpecialExpression(SpecialExpressionType.Count, projectionOperand), Expression.Constant(0));
            }
            else
            {
                var anyBuilderContext = builderContext.NewSelect();
                var tableExpression = Analyze(parameters[0], anyBuilderContext);

                if (!(tableExpression is TableExpression))
                    tableExpression = Analyze(tableExpression, anyBuilderContext);

                // from here we build a custom clause:
                // <anyClause> ==> "(select count(*) from <table> where <anyClause>)>0"
                // TODO (later...): see if some vendors support native Any operator and avoid this substitution
                if (parameters.Count > 1)
                {
                    var anyClause = Analyze(parameters[1], tableExpression, anyBuilderContext);
                    RegisterWhere(anyClause, anyBuilderContext);
                }
                var countExpr = new SpecialExpression(SpecialExpressionType.Count, tableExpression);
                var currSelect = anyBuilderContext.CurrentSelect;
                anyBuilderContext.CurrentSelect = currSelect.ChangeOperands(new Expression[] {countExpr}, currSelect.Operands);
                // TODO: see if we need to register the tablePiece here (we probably don't)

                // we now switch back to current context, and compare the result with 0
                var anyExpression = Expression.GreaterThan(anyBuilderContext.CurrentSelect, Expression.Constant(0));
                return anyExpression;
            }
        }

        protected virtual Expression AnalyzeLikeStart(IList<Expression> parameters, BuilderContext builderContext)
        {
            return AnalyzeLike(parameters[0], null, parameters[1], "%", builderContext);
        }

        protected virtual Expression AnalyzeLikeEnd(IList<Expression> parameters, BuilderContext builderContext)
        {
            return AnalyzeLike(parameters[0], "%", parameters[1], null, builderContext);
        }

        protected virtual Expression AnalyzeLike(IList<Expression> parameters, BuilderContext builderContext)
        {
            return AnalyzeLike(parameters[0], "%", parameters[1], "%", builderContext);
        }

        protected virtual Expression AnalyzeLike(Expression value, string before, Expression operand, string after, BuilderContext builderContext)
        {
            operand = Analyze(operand, builderContext);
            if (before != null)
                operand = new SpecialExpression(SpecialExpressionType.Concat, Expression.Constant(before), operand);
            if (after != null)
                operand = new SpecialExpression(SpecialExpressionType.Concat, operand, Expression.Constant(after));
            return new SpecialExpression(SpecialExpressionType.Like, Analyze(value, builderContext), operand);
        }

        protected virtual Expression AnalyzeSubString(IList<Expression> parameters, BuilderContext builderContext)
        {
            var stringExpression = Analyze(parameters[0], builderContext);
            var startExpression = new StartIndexOffsetExpression(builderContext.DbModel.LinqSqlProvider.StringIndexStartsAtOne,
                                                        Analyze(parameters[1], builderContext));
            if (parameters.Count > 2)
            {
                var lengthExpression = parameters[2];
                return new SpecialExpression(SpecialExpressionType.Substring, stringExpression, startExpression, lengthExpression);
            }
            return new SpecialExpression(SpecialExpressionType.Substring, stringExpression, startExpression);
        }

        private Expression ConvertArrayToConstant(ExternalValueExpression parameter, BuilderContext builderContext) {
          object value = null;
          try {
            //value = parameter.GetValue(null);
          } catch (Exception ex) {
            var msg = "Failed to translate 'Contains' call: supported only for arrays or lists that can be converted to constant list. Inner error: " + ex.Message;
            throw new Exception(msg, ex);
          }
          // do not unregister - it will break parameter sequence
          //UnregisterParameter(parameter, builderContext);
          return Expression.Constant(value);
        }

        protected virtual Expression AnalyzeContains(IList<Expression> parameters, BuilderContext builderContext)
        {
            var t0 = parameters[0].Type;
            if (t0.IsArray || typeof(IList).IsAssignableFrom(t0))
            {
              Expression array = Analyze(parameters[0], builderContext);
              var expression = Analyze(parameters[1], builderContext);
              return new SpecialExpression(SpecialExpressionType.In, expression, array);
            }
            else
            {
                if (typeof(IQueryable).IsAssignableFrom(parameters[0].Type))
                {
                  System.Diagnostics.Debugger.Break(); 
                }
            }
            throw Util.Throw("S0548: Can't analyze Contains() method");
        }

        protected virtual Expression AnalyzeToUpper(IList<Expression> parameters, BuilderContext builderContext)
        {
            return new SpecialExpression(SpecialExpressionType.ToUpper, Analyze(parameters[0], builderContext));
        }

        protected virtual Expression AnalyzeToLower(IList<Expression> parameters, BuilderContext builderContext)
        {
            return new SpecialExpression(SpecialExpressionType.ToLower, Analyze(parameters[0], builderContext));
        }

        /// <summary>
        /// Registers ordering request
        /// </summary>
        /// <param name="parameters"></param>
        /// <param name="descending"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        protected virtual Expression AnalyzeOrderBy(IList<Expression> parameters, bool descending, BuilderContext builderContext)
        {
            var table = Analyze(parameters[0], builderContext);
            // the column is related to table
            var column = Analyze(parameters[1], table, builderContext);
            builderContext.CurrentSelect.OrderBy.Add(new OrderByExpression(descending, column));
            return table;
        }

        /// <summary>
        /// Analyzes constant expression value, and eventually extracts a table
        /// </summary>
        /// <param name="expression"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        protected virtual Expression AnalyzeConstant(Expression expression, BuilderContext builderContext)
        {
            var constantExpression = expression as ConstantExpression;
            if (constantExpression != null)
            {
/*
                //RI: new stuff
                var iEntSet = constantExpression.Value as Vita.Entities.Linq.IEntitySet;
                if (iEntSet != null)
                  return this.CreateTable(iEntSet.ElementType, builderContext); 
*/
                var queriedType = GetQueriedType(expression);
                if (queriedType != null)
                {
                  // check if it is EntityQuery 
                  var entQuery = constantExpression.Value as EntityQuery;
                  if (entQuery != null) {
                    // check if it is entity set
                    if (entQuery.IsEntitySet) {
                      var tableType = constantExpression.Type.GenericTypeArguments[0];
                      return CreateTable(tableType, builderContext);
                    } else 
                      return AnalyzeSubQuery(entQuery.Expression, builderContext);  
                  }//if entQuery != null
                }
            }
            return expression;
        }

        // RI: my new stuff
        protected virtual Expression AnalyzeSubQuery(Expression expression, BuilderContext builderContext) {
          // Build a new Context for the query
          ExpressionChain exprChain = ExpressionChain.Build(expression);
          Expression tableExpression = CreateTableExpression(exprChain[0], builderContext);

          return this.Analyze(exprChain, tableExpression, builderContext);
        }

        protected virtual Expression AnalyzeSelectOperation(SelectOperatorType operatorType, IList<Expression> parameters, BuilderContext builderContext)
        {
            // a special case: if we have several SELECT expressions linked together,
            // we maximize the load to the database, since the result must use the same parameters
            // types and count.
            //builderContext.QueryContext.MaximumDatabaseLoad = true; // all select expression goes to SQL tier
            var subQuery = parameters[1];
            if (subQuery != null) {
              // Handle second select first
              BuilderContext newContext = builderContext.NewSisterSelect();
              Expression tableExpression = AnalyzeSubQuery(subQuery, newContext);
              BuildSelect(tableExpression, newContext);

              // add the second select select to the chain
              if (newContext.CurrentSelect.NextSelectExpression != null) {
                var typedTable = tableExpression as TableExpression; 
                var dbTable = typedTable == null ? null : typedTable.TableInfo;
                var operand0 = new SubSelectExpression(newContext.CurrentSelect, tableExpression.Type, "source", dbTable);
                newContext.NewParentSelect();
                newContext.CurrentSelect.Tables.Add(operand0);
              }
              SelectExpression selectToModify = builderContext.CurrentSelect;
              while (selectToModify.NextSelectExpression != null)
                selectToModify = selectToModify.NextSelectExpression;

              selectToModify.NextSelectExpression = newContext.CurrentSelect;
              selectToModify.NextSelectExpressionOperator = operatorType;

              Expression firstSelection = Analyze(parameters[0], builderContext);
              BuildSelect(firstSelection, builderContext);

              return firstSelection;
            }
            return Analyze(parameters[0], builderContext);
        }


        /// <summary>
        /// Analyses InvokeExpression
        /// </summary>
        /// <param name="expression"></param>
        /// <param name="parameters"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        protected virtual Expression AnalyzeInvoke(Expression expression, IList<Expression> parameters,
                                                   BuilderContext builderContext)
        {
            var invocationExpression = (InvocationExpression)expression;
            var lambda = invocationExpression.Expression as LambdaExpression;
            if (lambda != null)
            {
                var localBuilderContext = builderContext.NewQuote();
                //for (int parameterIndex = 0; parameterIndex < lambda.Parameters.Count; parameterIndex++)
                //{
                //    var parameter = lambda.Parameters[parameterIndex];
                //    localBuilderContext.Parameters[parameter.Name] = Analyze(invocationExpression.Arguments[parameterIndex], builderContext);
                //}
                //return Analyze(lambda, localBuilderContext);
                return Analyze(lambda, invocationExpression.Arguments, localBuilderContext);
            }
            // TODO: see what we must do here
            return expression;
        }
    }
}
