﻿
using System;
using System.Collections.Generic;
using System.Data;
using System.Linq;
using System.Linq.Expressions;
using System.Reflection;

using Vita.Common;
using Vita.Data.Model;
using Vita.Entities.Model;
using Vita.Data.Linq.Translation.Expressions;


namespace Vita.Data.Linq.Translation {
    
    internal partial class ExpressionDispatcher
    {

        /// <summary>
        /// Returns a registered column, or null if not found
        /// This method requires the table to be already registered
        /// </summary>
        /// <param name="table"></param>
        /// <param name="name"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        protected virtual ColumnExpression GetRegisteredColumn(TableExpression table, string name,
                                                               BuilderContext builderContext)
        {
            return
                (from queryColumn in builderContext.EnumerateScopeColumns()
                 where queryColumn.Table.IsEqualTo(table) && queryColumn.Name == name
                 select queryColumn).SingleOrDefault();
        }

        /// <summary>
        /// Returns an existing table or registers the current one
        /// </summary>
        /// <param name="tableExpression"></param>
        /// <param name="builderContext"></param>
        /// <returns>A registered table or the current newly registered one</returns>
        public virtual TableExpression RegisterTable(TableExpression tableExpression, BuilderContext builderContext)
        {
            // 1. Find the table in current scope
            var foundTableExpression = (from t in builderContext.EnumerateScopeTables()
                                        where t.IsEqualTo(tableExpression)
                                        select t).SingleOrDefault();
            if (foundTableExpression != null)
                return foundTableExpression;
            // 2. Find it in all scopes, and promote it to current scope.
            foundTableExpression = PromoteTable(tableExpression, builderContext);
            if (foundTableExpression != null)
                return foundTableExpression;
            // 3. Add it
            builderContext.CurrentSelect.Tables.Add(tableExpression);
            return tableExpression;
        }

        /// <summary>
        /// Promotes a table to a common parent between its current scope and our current scope
        /// </summary>
        /// <param name="tableExpression"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        protected virtual TableExpression PromoteTable(TableExpression tableExpression, BuilderContext builderContext)
        {
            int currentIndex = 0;
            SelectExpression oldSelect = null;
            SelectExpression commonScope = null;
            TableExpression foundTable = null;
            do
            {
                // take a select
                oldSelect = builderContext.SelectExpressions[currentIndex];

                // look for a common scope
                if (oldSelect != builderContext.CurrentSelect)
                {
                    commonScope = FindCommonScope(oldSelect, builderContext.CurrentSelect);
                    if (commonScope != null)
                        // if a common scope exists, look for an equivalent table in that select
                        for (int tableIndex = 0; tableIndex < oldSelect.Tables.Count && foundTable == null; tableIndex++)
                        {
                            if (oldSelect.Tables[tableIndex].IsEqualTo(tableExpression))
                            {
                                // found a matching table!
                                foundTable = oldSelect.Tables[tableIndex];
                            }
                        }
                }
                ++currentIndex;
            }
            while (currentIndex < builderContext.SelectExpressions.Count && foundTable == null);

            if (foundTable != null)
            {
                oldSelect.Tables.Remove(foundTable);
                commonScope.Tables.Add(foundTable);
            }
            return foundTable;
        }

        /// <summary>
        /// Find the common ancestor between two ScopeExpressions
        /// </summary>
        /// <param name="a"></param>
        /// <param name="b"></param>
        /// <returns></returns>
        protected virtual SelectExpression FindCommonScope(SelectExpression a, SelectExpression b)
        {
            for (var aScope = a; aScope != null; aScope = aScope.Parent)
            {
                for (var bScope = b; bScope != null; bScope = bScope.Parent)
                {
                    if (aScope == bScope)
                        return aScope;
                }
            }
            return null;
        }

        /// <summary>
        /// Registers a column
        /// This method requires the table to be already registered
        /// </summary>
        /// <param name="table"></param>
        /// <param name="memberInfo"></param>
        /// <param name="name"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        public ColumnExpression RegisterColumn(TableExpression table, string name, BuilderContext builderContext)
        {
            var queryColumn = GetRegisteredColumn(table, name, builderContext);
            if (queryColumn == null)
            {
                table = RegisterTable(table, builderContext);
                queryColumn = CreateColumn(table, name, builderContext);
                builderContext.CurrentSelect.Columns.Add(queryColumn);
            }
            return queryColumn;
        }

        /// <summary>
        /// Registers a column with only a table and a MemberInfo (this is the preferred method overload)
        /// </summary>
        /// <param name="tableExpression"></param>
        /// <param name="memberInfo"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        public ColumnExpression RegisterColumn(TableExpression tableExpression, MemberInfo memberInfo, BuilderContext builderContext) {
            return RegisterColumn(tableExpression, memberInfo.Name, builderContext);
        }

        public ColumnExpression CreateColumn(TableExpression table, string memberName, BuilderContext builderContext) {
            var col = table.TableInfo.GetColumnByMemberName(memberName);
            Util.Check(col != null, "Column for member [{0}] not found in table {1}. Member type is not supported by LINQ.", memberName, table.Name);
            return new ColumnExpression(table, col);
        }

        /// <summary>
        /// Creates a default TableExpression
        /// </summary>
        /// <param name="tableType"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        public virtual TableExpression CreateTable(Type tableType, BuilderContext builderContext)
        {
          var tableInfo = builderContext.DbModel.GetTable(tableType);
          return new TableExpression(tableType, tableInfo.FullName, tableInfo); // DataMapper.GetTableName(tableType, builderContext.QueryContext.session), tableInfo);
        }

        /// <summary>
        /// Registers an association
        /// </summary>
        /// <param name="tableExpression">The table holding the member, to become the joinedTable</param>
        /// <param name="tableMemberInfo"></param>
        /// <param name="otherType"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        public virtual TableExpression RegisterAssociation(TableExpression tableExpression, MemberInfo tableMemberInfo,
                                                           Type otherType, BuilderContext builderContext)
        {
            IList<MemberInfo> otherKeys;
            TableJoinType joinType;
            string joinID;
            var theseKeys = GetAssociationMembers(tableExpression, tableMemberInfo, otherType, out otherKeys,
                                                      out joinType, out joinID, builderContext.DbModel);
            // if the memberInfo has no corresponding association, we get a null, that we propagate
            if (theseKeys == null)
                return null;

            // the current table has the foreign key, the other table the referenced (usually primary) key
            if (theseKeys.Count != otherKeys.Count)
                Util.Throw("S0128: Association arguments (FK and ref'd PK) don't match");

            // we first create the table, with the JoinID, and we MUST complete the table later, with the Join() method
            var otherTableInfo = builderContext.DbModel.GetTable(otherType); 
            var otherTableExpression = new TableExpression(otherType, otherTableInfo.FullName, otherTableInfo, joinID);

            Expression joinExpression = null;

            var createdColumns = new List<ColumnExpression>();
            for (int keyIndex = 0; keyIndex < theseKeys.Count; keyIndex++)
            {
                // joinedKey is registered, even if unused by final select (required columns will be filtered anyway)
                Expression otherKey = RegisterColumn(otherTableExpression, otherKeys[keyIndex], builderContext);
                // foreign is created, we will store it later if this assocation is registered too
                Expression thisKey = CreateColumn(tableExpression, theseKeys[keyIndex].Name, builderContext);
                createdColumns.Add((ColumnExpression)thisKey);

                // if the key is nullable, then convert it
                // TODO: this will probably need to be changed
                if (otherKey.Type.IsNullableValueType())
                    otherKey = Expression.Convert(otherKey, otherKey.Type.GetUnderlyingType());
                if (thisKey.Type.IsNullableValueType())
                    thisKey = Expression.Convert(thisKey, thisKey.Type.GetUnderlyingType());
                // the other key is set as left operand, this must be this way
                // since some vendors (SQL Server) don't support the opposite
                var referenceExpression = Expression.Equal(otherKey, thisKey);

                // if we already have a join expression, then we have a double condition here, so "AND" it
                if (joinExpression != null)
                    joinExpression = Expression.And(joinExpression, referenceExpression);
                else
                    joinExpression = referenceExpression;
            }
            // we complete the table here, now that we have all join information
            otherTableExpression.Join(joinType, tableExpression, joinExpression);

            // our table is created, with the expressions
            // now check if we didn't register exactly the same
            var existingTable = (from t in builderContext.EnumerateScopeTables() where t.IsEqualTo(otherTableExpression) select t).SingleOrDefault();
            if (existingTable != null)
                return existingTable;
 
            builderContext.CurrentSelect.Tables.Add(otherTableExpression);
            foreach (var createdColumn in createdColumns)
                builderContext.CurrentSelect.Columns.Add(createdColumn);
            return otherTableExpression;
        }

        /// <summary>
        /// Returns association definition, if any
        /// </summary>
        /// <param name="thisTableExpression">The table referenced by the assocation (the type holding the member)</param>
        /// <param name="memberInfo">The memberInfo related to association</param>
        /// <param name="otherType"></param>
        /// <param name="otherKey">The keys in the associated table</param>
        /// <param name="joinType"></param>
        /// <param name="joinID"></param>
        /// <param name="session"></param>
        /// <returns>ThisKey</returns>
        public virtual IList<MemberInfo> GetAssociationMembers(TableExpression thisTableExpression, MemberInfo memberInfo, Type otherType,
                             out IList<MemberInfo> otherKey, out TableJoinType joinType, out string joinID, DbModel dbModel) {
          var refMemberInfo = thisTableExpression.TableInfo.Entity.GetMember(memberInfo.Name);
          if (refMemberInfo != null && refMemberInfo.Kind == Vita.Entities.Model.MemberKind.EntityRef) {
            // by default, join is inner
            joinType = TableJoinType.Inner;
            joinID = refMemberInfo.MemberName;
            otherType = refMemberInfo.ReferenceInfo.ToKey.Entity.ClassInfo.Type;
            otherKey = refMemberInfo.ReferenceInfo.ToKey.ExpandedMembers.Select(mi => (MemberInfo)mi.ClassProperty).ToList();
            var thisKey = refMemberInfo.ReferenceInfo.FromKey.ExpandedMembers.Select(mi => (MemberInfo)mi.ClassProperty).ToList();
            if (refMemberInfo.Flags.IsSet(EntityMemberFlags.Nullable))
              joinType |= TableJoinType.LeftOuter;
            return thisKey;
          }
          otherKey = null;
          joinType = TableJoinType.Default;
          joinID = null;
          return null;
        }



        /// <summary>
        /// Registers a MetaTable
        /// </summary>
        /// <param name="metaTableType"></param>
        /// <param name="aliases"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        public virtual MetaTableExpression RegisterMetaTable(Type metaTableType, IDictionary<MemberInfo, MutableExpression> aliases,
                                                             BuilderContext builderContext)
        {
            MetaTableExpression metaTableExpression;
            if (!builderContext.MetaTables.TryGetValue(metaTableType, out metaTableExpression))
            {
                metaTableExpression = new MetaTableExpression(aliases, metaTableType);
                builderContext.MetaTables[metaTableType] = metaTableExpression;
            }
            return metaTableExpression;
        }

        /// <summary>
        /// Registers a where clause in the current context scope
        /// </summary>
        /// <param name="whereExpression"></param>
        /// <param name="builderContext"></param>
        public virtual void RegisterWhere(Expression whereExpression, BuilderContext builderContext)
        {
            var where = CheckBoolBitExpression(whereExpression);
            builderContext.CurrentSelect.Where.Add(where);
        }

        /// <summary>
        /// Registers all columns of a table.
        /// </summary>
        /// <param name="tableExpression"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        public virtual IList<ColumnExpression> RegisterAllColumns(TableExpression tableExpression, BuilderContext builderContext)
        {
          var result = new List<ColumnExpression>(); 
          foreach (var colInfo in tableExpression.TableInfo.Columns)
          {
            if (colInfo.Member == null)
              continue; 
            var colExpr = RegisterColumn(tableExpression, colInfo.Member.MemberName, builderContext);
            result.Add(colExpr); 
          }
          return result; 
        }

        /// <summary>
        /// Registers an expression to be returned by main request.
        /// The strategy is to try to find it in the already registered parameters, and if not found, add it
        /// </summary>
        /// <param name="expression">The expression to be registered</param>
        /// <param name="builderContext"></param>
        /// <returns>Expression index</returns>
        public virtual int RegisterOutputParameter(Expression expression, BuilderContext builderContext)
        {
            var scope = builderContext.CurrentSelect;
            var operands = scope.Operands.ToList();
            for (int index = 0; index < operands.Count; index++)
            {
                if (ExpressionEquals(operands[index], expression))
                    return index;
            }
            operands.Add(expression);
            builderContext.CurrentSelect = (SelectExpression)scope.Mutate(operands);
            return operands.Count - 1;
        }

        protected virtual bool ExpressionEquals(Expression a, Expression b)
        {
            // TODO: something smarter, to compare contents and not only references (works fine only for columns)
            return a == b;
        }

        // RI: new stuff, optimized entity reader
        protected virtual Expression GetOutputTableReader(TableExpression tableExpression,
                                                          ParameterExpression dataRecordParameter, ParameterExpression sessionParameter,
                                                          BuilderContext builderContext) {
          var entMatzer = new EntityMaterializer(tableExpression.TableInfo);
          var allColExprs = RegisterAllColumns(tableExpression, builderContext);
          foreach (var col in allColExprs) {
            var colIndex = RegisterOutputParameter(col, builderContext);
            entMatzer.AddColumn(col.ColumnInfo, colIndex);
          }
          var entMatzerConst = Expression.Constant(entMatzer);
          var callReadEntity = Expression.Call(entMatzerConst, EntityMaterializer.ReadMethodInfo, dataRecordParameter, sessionParameter);
          var convExpr = Expression.Convert(callReadEntity, tableExpression.Type);
          return convExpr; 
        }


        /// <summary>
        /// Registers the expression as returned by the SQL request.
        /// </summary>
        /// <param name="expression"></param>
        /// <param name="dataRecordParameter"></param>
        /// <param name="mappingContextParameter"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        protected virtual Expression GetOutputValueReader(Expression expression,
                                                          ParameterExpression dataRecordParameter, ParameterExpression mappingContextParameter,
                                                          BuilderContext builderContext)
        {
            int valueIndex = RegisterOutputParameter(expression, builderContext);
            return GetOutputValueReader(expression.Type, valueIndex, dataRecordParameter, mappingContextParameter);
        }

        /// <summary>
        /// Registers the ColumnExpression as returned by the SQL request.
        /// </summary>
        /// <param name="expression"></param>
        /// <param name="dataRecordParameter"></param>
        /// <param name="mappingContextParameter"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        protected virtual Expression GetOutputValueReader(ColumnExpression expression, Type memberType,
                                                          ParameterExpression dataRecordParameter, ParameterExpression mappingContextParameter,
                                                          BuilderContext builderContext)
        {
            int valueIndex = RegisterOutputParameter(expression, builderContext);
            return GetOutputValueReader(memberType, valueIndex, dataRecordParameter, mappingContextParameter);
        }


        /// <summary>
        /// Registers the expression as returned column
        /// </summary>
        /// <param name="outputType"></param>
        /// <param name="valueIndex"></param>
        /// <param name="dataRecordParameter"></param>
        /// <param name="mappingContextParameter"></param>
        /// <returns></returns>
        protected virtual Expression GetOutputValueReader(Type outputType, int valueIndex, ParameterExpression dataRecordParameter,
                                                          ParameterExpression mappingContextParameter)
        {
            var propertyReaderLambda = DataRecordReader.GetPropertyReader(outputType);
            Expression invoke = new ParameterBinder().BindParams(propertyReaderLambda,
                dataRecordParameter, mappingContextParameter, Expression.Constant(valueIndex));
            if (!outputType.IsNullableValueType())
                invoke = Expression.Convert(invoke, outputType);
            return invoke;
        }
    }

    class ParameterBinder
    {
        Dictionary<Expression, Expression> map;

        public Expression BindParams(LambdaExpression expr, params Expression[] args)
        {
            map = new Dictionary<Expression, Expression>();

            if (expr.Parameters.Count != args.Length)
                throw new NotImplementedException();
            for (int i = 0; i < expr.Parameters.Count; ++i)
                map[expr.Parameters[i]] = args[i];
            return Visit(expr.Body);
        }

        Expression Visit(Expression expr)
        {
            switch (expr.NodeType)
            {
                case ExpressionType.Call:
                    MethodCallExpression call = expr as MethodCallExpression;
                    Expression[] new_args = new Expression[call.Arguments.Count];
                    for (int i = 0; i < new_args.Length; ++i)
                        new_args[i] = Visit(call.Arguments[i]);
                    return Expression.Call(call.Object, call.Method, new_args);
                case ExpressionType.Parameter:
                    Expression new_expr;
                    if (map.TryGetValue(expr, out new_expr))
                        return new_expr;
                    break;
                default:
                    throw new Exception("Can't handle " + expr.NodeType);
            }

            return expr;
        }
    }
}