﻿
using System;
using System.Collections.Generic;
using System.Data;
using System.Linq;
using System.Linq.Expressions;
using System.Reflection;

using Vita.Common;
using Vita.Entities.Runtime;
using Vita.Data.Linq.Translation.Expressions;

namespace Vita.Data.Linq.Translation {

    internal partial class ExpressionDispatcher 
    {
        public ExpressionQualifier ExpressionQualifier { get; set; }
        public DataRecordReader DataRecordReader { get; set; }

        public ExpressionDispatcher()
        { 
          ExpressionQualifier = new ExpressionQualifier(); 
          DataRecordReader = new DataRecordReader(); 
        }

        /// <summary>
        /// Registers the first table. Extracts the table type and registeres the piece
        /// </summary>
        /// <param name="requestingExpression"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        public virtual Expression CreateTableExpression(Expression requestingExpression, BuilderContext builderContext)
        {
          Type reqType = null; 
          switch(requestingExpression.NodeType) {
            case ExpressionType.Call:
              var callExpression = (MethodCallExpression)requestingExpression;
              reqType = callExpression.Arguments[0].Type;
              break; 
            case ExpressionType.Constant:
              var constExpr = (ConstantExpression)requestingExpression;
              reqType = constExpr.Value.GetType();
              break; 
            default:
              Util.Throw("LINQ engine error: unhandled node type '{0}' for CreateTableExpression", requestingExpression.NodeType);
              return null; //never happens
          }
          return CreateTable(GetQueriedType(reqType), builderContext);
        }

        /// <summary>
        /// Registers the first table. Extracts the table type and registeres the piece
        /// </summary>
        /// <param name="requestingExpression"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        public virtual Expression GetTable(Expression requestingExpression, BuilderContext builderContext)
        {
            var callExpression = (MethodCallExpression)requestingExpression;
            var requestingType = callExpression.Arguments[0].Type;
            return CreateTable(GetQueriedType(requestingType), builderContext);
        }

        /// <summary>
        /// Builds the upper select clause
        /// </summary>
        /// <param name="selectExpression"></param>
        /// <param name="builderContext"></param>
        public virtual void BuildSelect(Expression selectExpression, BuilderContext builderContext)
        {
            // collect columns, split Expression in
            // - things we will do in CLR
            // - things we will do in SQL
            LambdaExpression lambdaSelectExpression;
            // if we have a GroupByExpression, the result type is not the same:
            // - we need to read what is going to be the Key expression
            // - the final row generator builds a IGrouping<K,T> instead of T
            var selectGroupExpression = selectExpression as GroupExpression;
            if (selectGroupExpression != null)
            {
                lambdaSelectExpression = CutOutOperands(selectGroupExpression.GroupedExpression, builderContext);
                var lambdaSelectKeyExpression = CutOutOperands(selectGroupExpression.KeyExpression, builderContext);
                lambdaSelectExpression = BuildSelectGroup(lambdaSelectExpression, lambdaSelectKeyExpression,
                                                          builderContext);
            }
            else
                lambdaSelectExpression = CutOutOperands(selectExpression, builderContext);
            // look for tables and use columns instead
            // (this is done after cut, because the part that went to SQL must not be converted)
            //selectExpression = selectExpression.Recurse(e => CheckTableExpression(e, builderContext));
            // the last return value becomes the select, with CurrentScope
            builderContext.CurrentSelect.Reader = lambdaSelectExpression;
        }

        /// <summary>
        /// Builds the lambda as:
        /// (dr, mc) => new LineGrouping<K,T>(selectKey(dr,mc),select(dr,mc))
        /// </summary>
        /// <param name="select"></param>
        /// <param name="selectKey"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        protected virtual LambdaExpression BuildSelectGroup(LambdaExpression select, LambdaExpression selectKey,
                                                            BuilderContext builderContext)
        {
            var dataRecordParameter = Expression.Parameter(typeof(IDataRecord), "dataRecord");
            var sessionParameter = Expression.Parameter(typeof(EntitySession), "session");
            var kType = selectKey.Body.Type;
            var lType = select.Body.Type;
            var groupingType = typeof(LineGrouping<,>).MakeGenericType(kType, lType);
            var groupingCtor = groupingType.GetConstructor(new[] { kType, lType });
            var invokeSelectKey = Expression.Invoke(selectKey, dataRecordParameter, sessionParameter);
            var invokeSelect = Expression.Invoke(select, dataRecordParameter, sessionParameter);
            var newLineGrouping = Expression.New(groupingCtor, invokeSelectKey, invokeSelect);
            var iGroupingType = typeof(IGrouping<,>).MakeGenericType(kType, lType);
            var newIGrouping = Expression.Convert(newLineGrouping, iGroupingType);
            var lambda = Expression.Lambda(newIGrouping, dataRecordParameter, sessionParameter);
            return lambda;
        }

        /// <summary>
        /// Cuts Expressions between CLR and SQL:
        /// - Replaces Expressions moved to SQL by calls to DataRecord values reader
        /// - SQL expressions are placed into Operands
        /// - Return value creator is the returned Expression
        /// </summary>
        /// <param name="selectExpression"></param>
        /// <param name="builderContext"></param>
        protected virtual LambdaExpression CutOutOperands(Expression selectExpression, BuilderContext builderContext)
        {
            var dataRecordParameter = Expression.Parameter(typeof(IDataRecord), "dataRecord");
            var sessionParameter = Expression.Parameter(typeof(EntitySession), "session");
            var expression = CutOutOperands(selectExpression, dataRecordParameter, sessionParameter, builderContext);
            return Expression.Lambda(expression, dataRecordParameter, sessionParameter);
        }

        /// <summary>
        /// Cuts tiers in CLR / SQL.
        /// The search for cut is top-down
        /// </summary>
        /// <param name="expression"></param>
        /// <param name="dataRecordParameter"></param>
        /// <param name="sessionParameter"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        protected virtual Expression CutOutOperands(Expression expression,
                                                    ParameterExpression dataRecordParameter, ParameterExpression sessionParameter,
                                                    BuilderContext builderContext)
        {
            // two options: we cut and return
            if (GetCutOutOperand(expression, builderContext))
            {
                // "cutting out" means we replace the current expression by a SQL result reader
                // before cutting out, we check that we're not cutting a table in this case, we convert it into its declared columns
                if (expression is TableExpression)
                  // RI: entity reader comes here
                  return GetOutputTableReader((TableExpression)expression, dataRecordParameter,
                                                sessionParameter, builderContext);
                // RI: single-value result goes here
                return GetOutputValueReader(expression, dataRecordParameter, sessionParameter, builderContext);
            }
            // RI: Anon types, custom types go here
            var newOperands = new List<Expression>();
            var operands = expression.GetOperands();
            foreach (var operand in operands)
            {
                newOperands.Add(operand == null 
                    ? null
                    : CutOutOperands(operand, dataRecordParameter, sessionParameter, builderContext));
            }
            Expression newExpr;
            if (expression is MetaTableExpression)
              newExpr = ((MetaTableExpression)expression).ConvertToNew(newOperands);
            else 
              newExpr =  expression.ChangeOperands(newOperands, operands);
            return newExpr; 
        }

        /// <summary>
        /// Returns true if we must cut out the given Expression
        /// </summary>
        /// <param name="operand"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        private bool GetCutOutOperand(Expression operand, BuilderContext builderContext)
        {
          //RI: changed this (simplified?), see DbLinq code for old version
            var tier = ExpressionQualifier.GetTier(operand);
            if ((tier & ExpressionTier.Clr) != 0)
              return false;
            if ((tier & ExpressionTier.Sql) != 0) // we can cut out only if the following expressiong can go to SQL
              return true;
            return false; 
        }

        /// <summary>
        /// Checks any expression for a TableExpression, and eventually replaces it with the convenient columns selection
        /// </summary>
        /// <param name="expression"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        protected virtual Expression CheckTableExpression(Expression expression, BuilderContext builderContext)
        {
            if (expression is TableExpression)
                return GetSelectTableExpression((TableExpression)expression, builderContext);
            return expression;
        }

        /// <summary>
        /// Replaces a table selection by a selection of all mapped columns (ColumnExpressions).
        /// ColumnExpressions will be replaced at a later time by the tier splitter
        /// </summary>
        /// <param name="tableExpression"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        protected virtual Expression GetSelectTableExpression(TableExpression tableExpression, BuilderContext builderContext)
        {
            var bindings = new List<MemberBinding>();
            foreach (var columnExpression in RegisterAllColumns(tableExpression, builderContext))
            {
                var binding = Expression.Bind((MethodInfo) columnExpression.MemberInfo, columnExpression);
                bindings.Add(binding);
            }
            var newExpression = Expression.New(tableExpression.Type);
            return Expression.MemberInit(newExpression, bindings);
        }

        /// <summary>
        /// Returns a queried type from a given expression, or null if no type can be found
        /// </summary>
        /// <param name="expression"></param>
        /// <returns></returns>
        public virtual Type GetQueriedType(Expression expression)
        {
            return GetQueriedType(expression.Type);
        }

        /// <summary>
        /// Extracts the type from the potentially generic type
        /// </summary>
        /// <param name="type"></param>
        /// <returns></returns>
        public virtual Type GetQueriedType(Type type)
        {
            if (typeof(IQueryable).IsAssignableFrom(type))
            {
                if (type.IsGenericType)
                    return type.GetGenericArguments()[0];
            }
            return null;
        }

        /// <summary>
        /// Returns the parameter name, if the Expression is a ParameterExpression, null otherwise
        /// </summary>
        /// <param name="expression"></param>
        /// <returns></returns>
        public virtual string GetParameterName(Expression expression)
        {
            if (expression is ParameterExpression)
                return ((ParameterExpression)expression).Name;
            return null;
        }

        /// <summary>
        /// Merges a parameter and a parameter list
        /// </summary>
        /// <param name="p1"></param>
        /// <param name="p2"></param>
        /// <returns></returns>
        public virtual IList<Expression> MergeParameters(Expression p1, IEnumerable<Expression> p2)
        {
            var p = new List<Expression>();
            p.Add(p1);
            p.AddRange(p2);
            return p;
        }
    }
}
