﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;

using Vita.Data.Linq.Translation.Expressions;
using Vita.Common;

namespace Vita.Data.Linq.Translation {

  public static class ExpressionHelper {

    public static BinaryExpression MakeBinary(ExpressionType nodeType, Expression left, Expression right) {
      // After replacing arguments with column expressions there might be a problem with nullable columns.
      // The following code compensates for this
      if (left.Type == right.Type || left.IsConstNull() || right.IsConstNull())
        return Expression.MakeBinary(nodeType, left, right);
      if (left.Type.IsAssignableFrom(right.Type))
        return Expression.MakeBinary(nodeType, left, Expression.Convert(right, left.Type));
      if (right.Type.IsAssignableFrom(left.Type))
        return Expression.MakeBinary(nodeType, Expression.Convert(left, right.Type), right);
      Util.Throw("Cannot create Binary expression for arguments of types {0} and {1}", left.Type, right.Type);
      return null; //never happens
    }


    public static SpecialExpressionType ToBitwise(this ExpressionType type) {
      switch (type) {
        case ExpressionType.And: return SpecialExpressionType.AndBitwise;
        case ExpressionType.Or: return SpecialExpressionType.OrBitwise;
        case ExpressionType.ExclusiveOr: return SpecialExpressionType.XorBitwise;
        default:
          Util.Throw("No bitwise form of expression type {0}.", type);
          return default(SpecialExpressionType);
      }
    }

    public static bool IsConstNull(this Expression expression) {
      if (expression.NodeType != ExpressionType.Constant) return false;
      var c = (ConstantExpression)expression;
      return c.Value == null; 
    }

    public static bool IsConst<T>(this Expression expression, out T value) {
      value = default(T);
      if (expression.NodeType != ExpressionType.Constant || expression.Type != typeof(T)) return false;
      var c = (ConstantExpression)expression;
      value = (T)c.Value;
      return true; 
    }

    public static Expression CheckNeedConvert(Expression expression, Type toType) {
      if (expression.Type == toType)
        return expression;
      if (toType == typeof(bool)) {
        //One special case - used for bit columns
        if (expression.Type.IsInt())
          return Expression.Equal(expression, Expression.Constant(1));
      }
      if (toType.IsValueType && expression.Type.IsNullableValueType()) {
        return Expression.Convert(expression, toType);
      }
      //don't know what to do, let's hope SQL swallows it            
      return expression; 
    }

    public static bool IsNullConstant(this Expression expression) {
      if (expression.NodeType == ExpressionType.Constant && ((ConstantExpression)expression).Value == null)
        return true;
      if (expression.NodeType == ExpressionType.Convert) {
        var conv = (UnaryExpression)expression;
        if (IsNullConstant(conv.Operand))
          return true; 
      }
      return false; 
    }


    private static MethodInfo _convertToStringMethod;
    public static MethodInfo GetToStringConverter(Type type) {
      if (_convertToStringMethod == null)
        _convertToStringMethod = typeof(ExpressionHelper).GetMethod("ConvertToString"); //, BindingFlags.Static | BindingFlags.Public);
      return _convertToStringMethod.MakeGenericMethod(type); 
    }

    public static string ConvertToString<T>(T value) {
      if (value == null)
        return string.Empty;
      return value.ToString(); 
    }

    public static Expression Unqoute(Expression node) {
      if (node.NodeType != ExpressionType.Quote) return node;
      var unExpr = node as UnaryExpression;
      return unExpr.Operand;
    }

  }
}
