﻿
using System;
using System.Collections;
using System.Collections.Generic;
using System.Data;
using System.Diagnostics;
using System.Linq;
using System.Linq.Expressions;
using System.Text.RegularExpressions;
using Vita.Data.Linq.Translation.Expressions;

using Vita.Common;
using Vita.Data.Linq;
using Vita.Data.Model;
using Vita.Entities.Runtime;
using Vita.Data.Linq.Translation.SqlGen;

namespace Vita.Data.Linq.Translation {

    /// <summary>
    /// Full query builder, with cache management
    /// 1. Parses Linq Expression
    /// 2. Generates SQL
    /// </summary>
    internal partial class QueryBuilder  {
        private DbModel _dbModel; 

        public ExpressionLanguageParser ExpressionLanguageParser { get; set; }
        public ExpressionDispatcher ExpressionDispatcher { get; set; }
        public PrequelAnalyzer PrequelAnalyzer { get; set; }
        public ExpressionOptimizer ExpressionOptimizer { get; set; }
        public SpecialExpressionTranslator SpecialExpressionTranslator { get; set; }
        public SqlBuilder SqlBuilder { get; set; }

        public QueryBuilder(DbModel dbModel)  {
          _dbModel = dbModel; 
          ExpressionLanguageParser = new ExpressionLanguageParser(); 
          ExpressionDispatcher = new ExpressionDispatcher(); 
          PrequelAnalyzer = new PrequelAnalyzer(); 
          ExpressionOptimizer = new ExpressionOptimizer(); 
          SpecialExpressionTranslator = new SpecialExpressionTranslator(); 
          SqlBuilder = new SqlBuilder(_dbModel); 
        }

        public SqlQuery BuildDbQuery(EntityQueryRuntimeInfo queryInfo) {
          try {
            return BuildDbQueryImpl(queryInfo); 
          } catch (Exception ex) {
            var note = "Linq2Sql translation failed. Possibly facilities you are trying to use are not supported. " +
                        "Try to reformulate/simplify the query. Hint: do not use c# functions/methods inside query directly. ";
            ex.AddValue("Note", note);
//          ex.AddValue("Query", query.Expression);
            ex.AddValue("Query-Lambda", queryInfo.Lambda);
            throw;
          }
        }

        private SqlQuery BuildDbQueryImpl(EntityQueryRuntimeInfo queryInfo) {
          // RI: extra actions
          // prepare parameters
          var builderContext = new BuilderContext(_dbModel);
          // convert lambda params into an initial set of InputParameterExpression objects; 
          foreach (var prm in queryInfo.Lambda.Parameters) {
            var inpParam = new ExternalValueExpression(prm);
            builderContext.ExternalValues.Add(inpParam);
          }
          //Analyze/transform query expression
          var expressions = ExpressionChain.Build(queryInfo.Lambda.Body);
          var selectExpr = BuildSelectExpression(expressions, builderContext);
          // Analyze external values (parameters?), create DbParameters
          var dbParams = CreateSqlParameters(queryInfo, builderContext.ExternalValues);
          // If there's at least one parameter that must be converted to literal (ex: value list), we cannot cache the query
          bool canCache = !builderContext.ExternalValues.Any(v => v.SqlUse == ExternalValueSqlUse.Literal);
          //Build SQL, compile object materializer
          var sqlStatement = SqlBuilder.BuildSelectSql(selectExpr);
          var sql = sqlStatement.ToString();
          var rowObjectCreator = CompileRowCreator(builderContext);
          var dbQuery = new SqlQuery(sql, dbParams, rowObjectCreator, selectExpr.ExecuteMethodName, canCache);
          return dbQuery;
        }

        // analyzes external values and creates db parameters
        private List<SqlQueryParameter> CreateSqlParameters(EntityQueryRuntimeInfo queryInfo, IList<ExternalValueExpression> externalValues) {
          var sqlProvider = _dbModel.LinqSqlProvider;
          var paramValues = queryInfo.ParamValues.ToArray(); 
          var sqParams = new List<SqlQueryParameter>();
          var count = 0;
          foreach (var extValue in externalValues) {
            if (extValue.FreeUseCount == 0)
              continue; // it already has SqlUse = NotUsed
            sqlProvider.CheckQueryParameter(extValue);
            switch (extValue.SqlUse) {
              case ExternalValueSqlUse.NotUsed:
                continue; //next value
              case ExternalValueSqlUse.Literal:
                // We cannot use this value as a parameter (ex: list/array of values with Contains() method); 
                // in this case we transform it into literal value - it will be embedded into SQL directly. 
                // The query becomes non-cacheable (because of this embedded literal)
                var vreader = BuildParameterValueReader(queryInfo.Lambda.Parameters, extValue.SourceExpression);
                var readFunc = vreader.Compile(); 
                extValue.LiteralValue = readFunc(paramValues);
                break;
              case ExternalValueSqlUse.Parameter:
                var prmName = "P" + count++;
                var valueReader = BuildParameterValueReader(queryInfo.Lambda.Parameters, extValue.SourceExpression);
                var prm = extValue.SqlParameter = new SqlQueryParameter(prmName, extValue.SourceExpression.Type, valueReader);
                sqParams.Add(prm);
                break;
            }//switch
          }//foreach extValue
          return sqParams; 
        }

        private Expression<Func<object[], object>> BuildParameterValueReader(IList<ParameterExpression> queryParams, Expression valueSourceExpression) {
          // One trouble - for Binary object, we need to convert them to byte[]
          if (valueSourceExpression.Type == typeof(System.Data.Linq.Binary)) {
            var methToArray = typeof(System.Data.Linq.Binary).GetMethod("ToArray");
            valueSourceExpression = Expression.Call(valueSourceExpression, methToArray);
          }
          // .SourceExpression has a shape like (@p1, @p2... @pn) => (SomeTypedValue). 
          // If we compile it, we can execute the compiled delegate using DynamicInvoke - not so good. 
          // We would like to use strongly typed call to Func<object[], object> - so we need to build lambda of this shape and compile it.
          // We wrap the valueSource lambda into another lambda that accepts object array as parameter.
          var arrayParam = Expression.Parameter(typeof(object[]), "Prms");
          var valueReadLambda = Expression.Lambda(valueSourceExpression, queryParams);
          var wrapperArgs = new List<Expression>();
          for (int i = 0; i < queryParams.Count; i++) {
            var qpType = queryParams[i].Type;
            var arg = Expression.Convert(Expression.ArrayAccess(arrayParam, Expression.Constant(i)), qpType);
            wrapperArgs.Add(arg);
          }
          var callCompute = Expression.Invoke(valueReadLambda, wrapperArgs);
          //Convert result to object
          var convResult = Expression.Convert(callCompute, typeof(object));
          // Build lambda:  (object[]) -> object
          var lambda = (Expression<Func<object[], object>>) Expression.Lambda(convResult, arrayParam);
          return lambda;
        }



        /// <summary>
        /// Builds and chains the provided Expressions
        /// </summary>
        /// <param name="expressionChain"></param>
        /// <param name="builderContext"></param>
        protected virtual SelectExpression BuildSelectExpression(ExpressionChain expressionChain, BuilderContext builderContext) {
          var previousExpression = ExpressionDispatcher.CreateTableExpression(expressionChain[0], builderContext);
          previousExpression = BuildSelectExpression(expressionChain, previousExpression, builderContext);
          BuildOffsetsAndLimits(builderContext);
          // then prepare Parts for SQL translation
          CheckTablesAlias(builderContext);
          CheckColumnsAlias(builderContext);
          PrepareSqlOperands(builderContext);
          // now, we optimize anything we can
          OptimizeQuery(builderContext);
          // in the very end, we keep the SELECT clause
          return builderContext.CurrentSelect;
        }

        /// <summary>
        /// Finds all registered tables or columns with the given name.
        /// We exclude parameter because they won't be prefixed/suffixed the same way (well, that's a guess, I hope it's a good one)
        /// </summary>
        /// <param name="name"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        protected virtual IList<Expression> FindExpressionsByName(string name, BuilderContext builderContext)
        {
            var expressions = new List<Expression>();
            expressions.AddRange((from t in builderContext.EnumerateAllTables() where t.Alias == name select (Expression)t).Distinct());
            expressions.AddRange(from c in builderContext.EnumerateScopeColumns() where c.Alias == name select (Expression)c);
            return expressions;
        }

        protected virtual string MakeName(string aliasBase, int index, string anonymousBase, BuilderContext builderContext)
        {
            if (string.IsNullOrEmpty(aliasBase))
                aliasBase = anonymousBase;
            return string.Format("{0}{1}", aliasBase, index);
        }

        protected virtual string MakeTableName(string aliasBase, int index, BuilderContext builderContext)
        {
            return MakeName(aliasBase, index, "t", builderContext);
        }

        protected virtual string MakeParameterName(string aliasBase, int index, BuilderContext builderContext)
        {
            return MakeName(aliasBase, index, "p", builderContext);
        }

        /// <summary>
        /// Give all non-aliased tables a name
        /// </summary>
        /// <param name="builderContext"></param>
        protected virtual void CheckTablesAlias(BuilderContext builderContext)
        {
            var tables = builderContext.EnumerateAllTables().Distinct().ToList();
            // just to be nice: if we have only one table involved, there's no need to alias it
            if (tables.Count == 1)
            {
                tables[0].Alias = null;
            }
            else
            {
                foreach (var tableExpression in tables)
                {
                    // if no alias, or duplicate alias
                    if (string.IsNullOrEmpty(tableExpression.Alias) ||
                        FindExpressionsByName(tableExpression.Alias, builderContext).Count > 1)
                    {
                        int anonymousIndex = 0;
                        var aliasBase = tableExpression.Alias;
                        // we try to assign one until we have a unique alias
                        do
                        {
                            tableExpression.Alias = MakeTableName(aliasBase, ++anonymousIndex, builderContext);
                        } while (FindExpressionsByName(tableExpression.Alias, builderContext).Count != 1);
                    }
                }
            }
        }

        //RI: added this to fix some SQL errors when multiple columns in the output have the same name
        protected virtual void CheckColumnsAlias(BuilderContext builderContext) {
          foreach(var select in builderContext.SelectExpressions) { 
            var outColumns = select.Operands.OfType<ColumnExpression>().ToList();
            CheckColumnsAlias(outColumns);
            foreach (var g in select.Group) {
              CheckColumnsAlias(g.Columns); 
            }
          }//for i
        } //method

        protected virtual void CheckColumnsAlias(IList<ColumnExpression> columns) {
            var dupColNames = columns.Where(c => string.IsNullOrEmpty(c.Alias)).GroupBy(c => c.Name).Where(g => g.Count() > 1).Select(g => g.Key).ToList();
            if (dupColNames.Count == 0) return; 
            var index = 0;
            foreach (var col in columns) {
              if (dupColNames.Contains(col.Name))
                col.Alias = "_c_" + (index++);
            }
        } //method


        /// <summary>
        /// Builds the ExpressionQuery main Expression, given a Table (or projection) expression
        /// </summary>
        /// <param name="expressions"></param>
        /// <param name="tableExpression"></param>
        /// <param name="builderContext"></param>
        /// <returns></returns>
        protected Expression BuildSelectExpression(ExpressionChain expressions, Expression tableExpression, BuilderContext builderContext)
        {
            tableExpression = ExpressionDispatcher.Analyze(expressions, tableExpression, builderContext);
            ExpressionDispatcher.BuildSelect(tableExpression, builderContext);
            return tableExpression;
        }

        /// <summary>
        /// This is a hint for SQL generations
        /// </summary>
        /// <param name="builderContext"></param>
        protected virtual void BuildOffsetsAndLimits(BuilderContext builderContext)
        {
            foreach (var selectExpression in builderContext.SelectExpressions)
            {
              //RI: because of some problems with DISTINCT+TOP combination, we always force using Skip/Take combination (avoiding TOP)
              bool hasAny = selectExpression.Offset != null || selectExpression.Limit != null;
              if (!hasAny) continue;
              if (selectExpression.Offset == null)
                selectExpression.Offset = Expression.Constant(0);
              //RI: old code; check for offse!=null is redundant now
              if (selectExpression.Offset != null && selectExpression.Limit != null)
              {
                  selectExpression.OffsetAndLimit = Expression.Add(selectExpression.Offset, selectExpression.Limit);
              }
              //Check that we have OrderBy - it is required if we have offset/limit
              if (selectExpression.OrderBy.Count == 0) {
                var col1 = selectExpression.Operands.FirstOrDefault() as ColumnExpression;
                if (col1 == null)
                  col1 = selectExpression.Columns[0];
                selectExpression.OrderBy.Add(new OrderByExpression(false, col1));
              }
            }
        }

        /// <summary>
        /// Builds the delegate to create a row
        /// </summary>
        /// <param name="builderContext"></param>
        protected Func<IDataRecord, EntitySession, object> CompileRowCreator(BuilderContext builderContext) {
          var reader = builderContext.CurrentSelect.Reader;
          // RI: reader is lambda returning typed object; to easier manage the execution,
          // so we don't have to convert it to generic version, convert the result to 'object' 
          // Now, we might have already a convert inside body to strongly typed object
          // (Table row reader adds a call to AttachEntity which returns object). 
          // In this case just remove this last conversion
          var body = reader.Body;
          var convExpr = body as UnaryExpression;
          if (body.NodeType == ExpressionType.Convert && convExpr.Operand.Type == typeof(object)) {
            reader = Expression.Lambda(convExpr.Operand, reader.Parameters);
          } else {
            var newBody = Expression.Convert(reader.Body, typeof(object));
            reader = Expression.Lambda(newBody, reader.Parameters); 
          }

          reader = (LambdaExpression)SpecialExpressionTranslator.Translate(reader);
          reader = (LambdaExpression)ExpressionOptimizer.Optimize(reader, builderContext);
          var creator = (Func<IDataRecord, EntitySession, object>) reader.Compile();
          return creator; 
        }

        /// <summary>
        /// Prepares SELECT operands to help SQL transaltion
        /// </summary>
        /// <param name="builderContext"></param>
        protected virtual void PrepareSqlOperands(BuilderContext builderContext)
        {
            ProcessExpressions(PrequelAnalyzer.Analyze, true, builderContext);
        }

        /// <summary>
        /// Processes all expressions in query, with the option to process only SQL targetting expressions
        /// This method is generic, it receives a delegate which does the real processing
        /// </summary>
        /// <param name="processor"></param>
        /// <param name="processOnlySqlParts"></param>
        /// <param name="builderContext"></param>
        protected virtual void ProcessExpressions(Func<Expression, BuilderContext, Expression> processor,
                                                  bool processOnlySqlParts, BuilderContext builderContext)
        {
            for (int scopeExpressionIndex = 0; scopeExpressionIndex < builderContext.SelectExpressions.Count; scopeExpressionIndex++)
            {
                // no need to process the select itself here, all ScopeExpressions that are operands are processed as operands
                // and the main ScopeExpression (the SELECT) is processed below
                var scopeExpression = builderContext.SelectExpressions[scopeExpressionIndex];

                // where clauses
                List<int> whereToRemove = new List<int>(); // List of where clausole evaluating TRUE (could be ignored and so removed)
                bool falseWhere = false; // true when the full where evaluate to FALSE
                for (int whereIndex = 0; whereIndex < scopeExpression.Where.Count; whereIndex++)
                {
                    Expression whereClausole = processor(scopeExpression.Where[whereIndex], builderContext);
                    ConstantExpression constantWhereClausole = whereClausole as ConstantExpression;
                    if (constantWhereClausole != null)
                    {
                        if (constantWhereClausole.Value.Equals(false))
                        {
                            falseWhere = true;
                            break;
                        }
                        else if (constantWhereClausole.Value.Equals(true))
                        {
                            whereToRemove.Add(whereIndex);
                            continue;
                        }
                    }
                    scopeExpression.Where[whereIndex] = whereClausole;
                }
                if (scopeExpression.Where.Count > 0)
                {
                    if (falseWhere)
                    {
                        scopeExpression.Where.Clear();
                        scopeExpression.Where.Add(Expression.Equal(Expression.Constant(true), Expression.Constant(false)));
                    }
                    else
                        foreach (int whereIndex in whereToRemove)
                            scopeExpression.Where.RemoveAt(whereIndex);
                }

                // limit clauses
                if (scopeExpression.Offset != null)
                    scopeExpression.Offset = processor(scopeExpression.Offset, builderContext);
                if (scopeExpression.Limit != null)
                    scopeExpression.Limit = processor(scopeExpression.Limit, builderContext);
                if (scopeExpression.OffsetAndLimit != null)
                    scopeExpression.OffsetAndLimit = processor(scopeExpression.OffsetAndLimit, builderContext);

                builderContext.SelectExpressions[scopeExpressionIndex] = scopeExpression;
            }
            // now process the main SELECT
            if (processOnlySqlParts)
            {
                // if we process only the SQL Parts, these are the operands
                var newOperands = new List<Expression>();
                var oldOperands = builderContext.CurrentSelect.Operands;
                foreach (var operand in oldOperands)
                    newOperands.Add(processor(operand, builderContext));
                builderContext.CurrentSelect = builderContext.CurrentSelect.ChangeOperands(newOperands, oldOperands);
            }
            else
            {
                // the output parameters and result builder
                builderContext.CurrentSelect = (SelectExpression)processor(builderContext.CurrentSelect, builderContext);
            }
        }

        /// <summary>
        /// Optimizes the query by optimizing subexpressions, and preparsing constant expressions
        /// </summary>
        /// <param name="builderContext"></param>
        protected virtual void OptimizeQuery(BuilderContext builderContext)
        {
            ProcessExpressions(ExpressionOptimizer.Optimize, false, builderContext);
        }


    }//class
}
