﻿
using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;

using Vita.Common;
using Vita.Data.Linq.Translation.Expressions;

namespace Vita.Data.Linq.Translation.SqlGen {

    internal class SqlBuilder 
    {
        public ExpressionQualifier ExpressionQualifier { get; set; }
        private Vita.Data.Model.DbModel _dbModel;
        private Vita.Data.Driver.LinqSqlProvider _sqlProvider;

        public SqlBuilder(Vita.Data.Model.DbModel dbModel) {
          _dbModel = dbModel;
          _sqlProvider = _dbModel.LinqSqlProvider; 
          ExpressionQualifier = new ExpressionQualifier(); // ObjectFactory.Get<IExpressionQualifier>();
        }

        /// <summary>
        /// Returns a list of sorted tables, given a select expression.
        /// The tables are sorted by dependency: independent tables first, dependent tables next
        /// </summary>
        /// <param name="selectExpression"></param>
        /// <returns></returns>
        protected IList<TableExpression> GetSortedTables(SelectExpression selectExpression)
        {
            var tables = new List<TableExpression>();
            foreach (var table in selectExpression.Tables)
            {
                // the rules are:
                // a table climbs up to 0 until we find the table it depends on
                // we keep the index and insert on it
                // we place joining tables under joined tables
                int tableIndex;
                for (tableIndex = tables.Count; tableIndex > 0; tableIndex--)
                {
                    // above us, the joined table? Stop now
                    if (tables[tableIndex - 1] == table.JoinedTable)
                        break;
                    // if the current table is joining and we have a non-joining table above, we stop here too
                    if (table.JoinExpression != null && tables[tableIndex - 1].JoinExpression == null)
                        break;
                }
                tables.Insert(tableIndex, table);
            }
            return tables;
        }

        /// <summary>
        /// Main SQL builder
        /// </summary>
        /// <param name="selectExpression"></param>
        /// <param name="queryContext"></param>
        /// <returns></returns>
        public SqlStatement BuildSelectSql(SelectExpression selectExpression)
        {
            selectExpression = _sqlProvider.PreviewSelect(selectExpression);

            // A scope usually has:
            // - a SELECT: the operation creating a CLR object with data coming from SQL tier
            // - a FROM: list of tables
            // - a WHERE: list of conditions
            // - a GROUP BY: grouping by selected columns
            // - a ORDER BY: sort
            var select = BuildSelect(selectExpression);
            if (select.ToString() == string.Empty)
            {
                SubSelectExpression subselect = null;
                if (selectExpression.Tables.Count == 1)
                    subselect = selectExpression.Tables[0] as SubSelectExpression;
                if(subselect != null)
                    return _sqlProvider.GetParenthesis(BuildSelectSql(subselect.Select));
            }

            // TODO: the following might be wrong (at least this might be the wrong place to do this
            if (select.ToString() == string.Empty)
                select = new SqlStatement("SELECT " + _sqlProvider.GetLiteral(null) + " AS " + _sqlProvider.GetSafeName("Empty"));

            var tables = GetSortedTables(selectExpression);
            var from = BuildFrom(tables);
            var join = BuildJoin(tables);
            var where = BuildWhere(tables, selectExpression.Where);
            var groupBy = BuildGroupBy(selectExpression.Group);
            var having = BuildHaving(selectExpression.Where);
            var orderBy = BuildOrderBy(selectExpression.OrderBy);
            select = Join(select, from, join, where, groupBy, having, orderBy);
            select = BuildLimit(selectExpression, select);

            if (selectExpression.NextSelectExpression != null)
            {
                var nextLiteralSelect = BuildSelectSql(selectExpression.NextSelectExpression);
                select = _sqlProvider.GetLiteral(
                    selectExpression.NextSelectExpressionOperator,
                    select, nextLiteralSelect);
            }

            return select;
        }

        public SqlStatement Join(params SqlStatement[] clauses)
        {
            return SqlStatement.Join(_sqlProvider.NewLine,
                               (from clause in clauses where clause.ToString() != string.Empty select clause).ToList());
        }

        /// <summary>
        /// The simple part: converts an expression to SQL
        /// This is not used for FROM clause
        /// </summary>
        /// <param name="expression"></param>
        /// <param name="queryContext"></param>
        /// <returns></returns>
        protected virtual SqlStatement BuildExpression(Expression expression)
        {

            if (expression is ConstantExpression)
              return _sqlProvider.GetLiteral(((ConstantExpression)expression).Value);

            var currentPrecedence = ExpressionQualifier.GetPrecedence(expression);
            // first convert operands
            var operands = expression.GetOperands();
            var literalOperands = new List<SqlStatement>();
            foreach (var operand in operands)
            {
                var operandPrecedence = ExpressionQualifier.GetPrecedence(operand);
                var literalOperand = BuildExpression(operand);
                if (operandPrecedence > currentPrecedence)
                    literalOperand = _sqlProvider.GetParenthesis(literalOperand);
                literalOperands.Add(literalOperand);
            }

            // then converts expression
            if (expression is SpecialExpression) {
              var specExpr = (SpecialExpression)expression;
              //RI: Special case for multiple "*" operands
              if (specExpr.SpecialNodeType == SpecialExpressionType.Count && literalOperands.Count > 0) {
                literalOperands.Clear();
                literalOperands.Add("*");              
              }
              return _sqlProvider.GetLiteral((specExpr).SpecialNodeType, literalOperands);
            }
            if (expression is TableExpression)
            {
                var tableExpression = (TableExpression)expression;
                if (tableExpression.Alias != null) // if we have an alias, use it
                {
                    return _sqlProvider.GetColumn(_sqlProvider.GetTableAlias(tableExpression.Alias),
                                                 _sqlProvider.GetColumns());
                }
                return _sqlProvider.GetColumns();
            }

            //RI: We might have NewExpression here! Query: (from b in books select new {b.Title}).Count();
            // in this case the NewExpression is 'hidden' inside subquery and it is not visible to CutOutOperands
            // We just return list of arguments (columns) of New expression
            if (expression is NewExpression) 
              return new SqlStatement(literalOperands); 
            //RI: adding this
            if (expression is MetaTableExpression) {
              var metaTable = (MetaTableExpression)expression;
              return _sqlProvider.GetColumns(); 
            }

            if (expression is ColumnExpression)
            {
                var columnExpression = (ColumnExpression)expression;
                if (columnExpression.Table.Alias != null)
                {
                    return _sqlProvider.GetColumn(_sqlProvider.GetTableAlias(columnExpression.Table.Alias),
                                                 columnExpression.Name);
                }
                //RI: changed this to keep output type
                var sqlPart = new SqlLiteralPart(_sqlProvider.GetColumn(columnExpression.Name), expression.Type);
                return new SqlStatement(sqlPart);
                //return _sqlProvider.GetColumn(columnExpression.Name);
            }

            if (expression is ExternalValueExpression)  {
                var extValue = (ExternalValueExpression)expression;
                
                switch(extValue.SqlUse) {
                  case ExternalValueSqlUse.Literal:
                    var sql = _sqlProvider.GetLiteral(extValue.LiteralValue);
                    return sql; 
                  case ExternalValueSqlUse.Parameter:
                    var prmName = extValue.SqlParameter.Name;
                    return new SqlStatement(new SqlParameterPart(_sqlProvider.GetParameterName(prmName), prmName));
                  default:
                    // we should never get here
                    Util.Throw("LINQ engine error: encountered InputParameterExpression with invalid Usage type: {0}, Expression: {1}",
                      extValue.SqlUse, extValue.SourceExpression);
                    return null; //never happens 
                } 
            }
            if (expression is SelectExpression)
                return BuildSelectSql((SelectExpression)expression);
            if (expression is GroupExpression)
                return BuildExpression(((GroupExpression)expression).GroupedExpression);

            StartIndexOffsetExpression indexExpression = expression as StartIndexOffsetExpression;
            if (indexExpression!=null)
            {
                if (indexExpression.StartsAtOne)
                {
                    literalOperands.Add(BuildExpression(Expression.Constant(1)));
                    return _sqlProvider.GetLiteral(ExpressionType.Add, literalOperands);
                }
                else
                    return literalOperands.First();
            }
            if (expression.NodeType == ExpressionType.Convert || expression.NodeType == ExpressionType.ConvertChecked)
            {
                var unaryExpression = (UnaryExpression)expression;
                var firstOperand = literalOperands.First();
                if (IsConversionRequired(unaryExpression))
                    return _sqlProvider.GetLiteralConvert(firstOperand, unaryExpression.Type);
                return firstOperand;
            }
            return _sqlProvider.GetLiteral(expression.NodeType, literalOperands);
        }

        /// <summary>
        /// Determines if a SQL conversion is required
        /// </summary>
        /// <param name="expression"></param>
        /// <returns></returns>
        private bool IsConversionRequired(UnaryExpression expression)
        {
            // obvious (and probably never happens), conversion to the same type
            if (expression.Type == expression.Operand.Type)
                return false;
            if (IsEnumAndInt(expression.Type, expression.Operand.Type))
              return false; 
            //RI: special case - do not convert bit column expression to int - they are already ints
            if (expression.Type == typeof(int) && expression.Operand is ColumnExpression && expression.Operand.Type == typeof(bool))
              return false; 
            // second, nullable to non-nullable for the same type
            if (expression.Type.IsNullableValueType() && !expression.Operand.Type.IsNullableValueType())
            {
                if (expression.Type.GetUnderlyingType() == expression.Operand.Type)
                    return false;
            }
            // third, non-nullable to nullable
            if (!expression.Type.IsNullableValueType() && expression.Operand.Type.IsNullableValueType())
            {
                if (expression.Type == expression.Operand.Type.GetUnderlyingType())
                    return false;
            }
            // found no excuse not to convert? then convert
            return true;
        }

        private bool IsEnumAndInt(Type x, Type y) {
          return x.IsEnum && Enum.GetUnderlyingType(x) == y || y.IsEnum && Enum.GetUnderlyingType(y) == x;
        }

        protected virtual bool MustDeclareAsJoin(IList<TableExpression> tables, TableExpression table)
        {
            // the first table can not be declared as join
            if (table == tables[0])
                return false;
            // we must declare as join, whatever the join is,
            // if some of the registered tables are registered as complex join
            if (tables.Any(t => t.JoinType != TableJoinType.Inner))
                return table.JoinExpression != null;
            return false;
        }

        protected virtual SqlStatement BuildFrom(IList<TableExpression> tables)
        {
          var sqlProvider = this._dbModel.LinqSqlProvider;
            var fromClauses = new List<SqlStatement>();
            foreach (var tableExpression in tables)
            {
                if (!MustDeclareAsJoin(tables, tableExpression))
                {
                    if (tableExpression.Alias != null)
                    {
                        string tableAlias;

                        // All subqueries has an alias in FROM
                        SubSelectExpression subquery = tableExpression as SubSelectExpression;
                        if (subquery == null)
                            tableAlias = _sqlProvider.GetTableAsAlias(tableExpression.Name, tableExpression.Alias);
                        else
                        {
                            var subqueryStatements = new SqlStatement(BuildSelectSql(subquery.Select));
                            tableAlias = _sqlProvider.GetSubQueryAsAlias(subqueryStatements.ToString(), tableExpression.Alias);
                        }

                        if ((tableExpression.JoinType & TableJoinType.LeftOuter) != 0)
                            tableAlias = "/* LEFT OUTER */ " + tableAlias;
                        if ((tableExpression.JoinType & TableJoinType.RightOuter) != 0)
                            tableAlias = "/* RIGHT OUTER */ " + tableAlias;
                        fromClauses.Add(tableAlias);
                    }
                    else
                    {
                        fromClauses.Add(_sqlProvider.GetTable(tableExpression.Name));
                    }
                }
            }
            return _sqlProvider.GetFromClause(fromClauses.ToArray());
        }

        /// <summary>
        /// Builds join clauses
        /// </summary>
        /// <param name="tables"></param>
        /// <param name="queryContext"></param>
        /// <returns></returns>
        protected virtual SqlStatement BuildJoin(IList<TableExpression> tables)
        {
            var sqlProvider = _dbModel.LinqSqlProvider;
            var joinClauses = new List<SqlStatement>();
            foreach (var tableExpression in tables)
            {
                // this is the pending declaration of direct tables
                if (MustDeclareAsJoin(tables, tableExpression))
                {
                    // get constitutive Parts
                    var joinExpression = BuildExpression(tableExpression.JoinExpression);
                    var tableAlias = _sqlProvider.GetTableAsAlias(tableExpression.Name, tableExpression.Alias);
                    SqlStatement joinClause;
                    switch (tableExpression.JoinType)
                    {
                        case TableJoinType.Inner:
                            joinClause = _sqlProvider.GetInnerJoinClause(tableAlias, joinExpression);
                            break;
                        case TableJoinType.LeftOuter:
                            joinClause = _sqlProvider.GetLeftOuterJoinClause(tableAlias, joinExpression);
                            break;
                        case TableJoinType.RightOuter:
                            joinClause = _sqlProvider.GetRightOuterJoinClause(tableAlias, joinExpression);
                            break;
                        case TableJoinType.FullOuter:
                            throw new NotImplementedException();
                        default:
                            throw new ArgumentOutOfRangeException();
                    }
                    joinClauses.Add(joinClause);
                }
            }
            return _sqlProvider.GetJoinClauses(joinClauses.ToArray());
        }

        protected virtual bool IsHavingClause(Expression expression)
        {
            bool isHaving = false;
            expression.Recurse(delegate(Expression e)
                                   {
                                       if (e is GroupExpression)
                                           isHaving = true;
                                       return e;
                                   });
            return isHaving;
        }

        protected virtual SqlStatement BuildWhere(IList<TableExpression> tables, IList<Expression> wheres)
        {
            var sqlProvider = _dbModel.LinqSqlProvider;
            var whereClauses = new List<SqlStatement>();
            foreach (var tableExpression in tables)
            {
                if (!MustDeclareAsJoin(tables, tableExpression) && tableExpression.JoinExpression != null)
                    whereClauses.Add(BuildExpression(tableExpression.JoinExpression));
            }
            foreach (var whereExpression in wheres)
            {
                if (!IsHavingClause(whereExpression))
                    whereClauses.Add(BuildExpression(whereExpression));
            }
            return _sqlProvider.GetWhereClause(whereClauses.ToArray());
        }

        protected virtual SqlStatement BuildHaving(IList<Expression> wheres)
        {
            var havingClauses = new List<SqlStatement>();
            foreach (var whereExpression in wheres)
            {
                if (IsHavingClause(whereExpression))
                    havingClauses.Add(BuildExpression(whereExpression));
            }
            return _sqlProvider.GetHavingClause(havingClauses.ToArray());
        }

        protected virtual SqlStatement GetGroupByClause(ColumnExpression columnExpression)
        {
            if (columnExpression.Table.Alias != null)
            {
                return _sqlProvider.GetColumn(_sqlProvider.GetTableAlias(columnExpression.Table.Alias),
                                             columnExpression.Name);
            }
            return _sqlProvider.GetColumn(columnExpression.Name);
        }

        protected virtual SqlStatement BuildGroupBy(IList<GroupExpression> groupByExpressions)
        {
            var groupByClauses = new List<SqlStatement>();
            foreach (var groupByExpression in groupByExpressions)
            {
                if (groupByExpression.IsDistinct) continue; //RI: added this
                foreach (var operand in groupByExpression.Columns)
                {
                    var columnOperand = operand as ColumnExpression;
                    if (columnOperand == null)
                        Util.Throw("S0201: Groupby argument must be a ColumnExpression");
                    groupByClauses.Add(GetGroupByClause(columnOperand));
                }
            }
            return _sqlProvider.GetGroupByClause(groupByClauses.ToArray());
        }

        protected virtual SqlStatement BuildOrderBy(IList<OrderByExpression> orderByExpressions)
        {
            var orderByClauses = new List<SqlStatement>();
            foreach (var clause in orderByExpressions)
            {
                orderByClauses.Add(_sqlProvider.GetOrderByColumn(BuildExpression(clause.ColumnExpression),
                                                                clause.Descending));
            }
            return _sqlProvider.GetOrderByClause(orderByClauses.ToArray());
        }

        protected virtual SqlStatement BuildSelect(Expression select)
        {
            var selectClauses = new List<SqlStatement>();
            var ops = select.GetOperands().ToList(); 
            foreach (var outputExpr in ops)
            {
                var expressionString = BuildExpression(outputExpr);
                if (outputExpr is SelectExpression)
                    expressionString = _sqlProvider.GetParenthesis(expressionString);
                else if (outputExpr is ColumnExpression) {
                  //RI: added this to add alias
                  var colExpr = (ColumnExpression) outputExpr;
                  if (!string.IsNullOrWhiteSpace(colExpr.Alias))
                    expressionString += " " + colExpr.Alias;
                } 
                selectClauses.Add(expressionString);
            }
            SelectExpression selectExp = select as SelectExpression;
            if (selectExp != null)
            {
                if (selectExp.Group.Count == 1 && selectExp.Group[0].IsDistinct)
                {
                    //RI: changed code
                  var g = selectExp.Group[0];
                  var outCols = new List<SqlStatement>();
                  foreach (var col in g.Columns) {
                    var sqlCol = col.Table.Alias == null ? 
                        _sqlProvider.GetColumn(col.Name) : 
                        _sqlProvider.GetColumn(_sqlProvider.GetTableAlias(col.Table.Alias), col.Name);
                    if (!string.IsNullOrEmpty(col.Alias))
                      sqlCol += " " + col.Alias; 
                    outCols.Add(sqlCol);  
                  }
                  return _sqlProvider.GetSelectDistinctClause(outCols.ToArray());
                  /*
                  // this is a select DISTINCT expression
                    // TODO: better handle selected columns on DISTINCT: I suspect this will not work in some cases
                    if (selectClauses.Count == 0)
                    {
                        selectClauses.Add(_sqlProvider.GetColumns());
                    }
                    return _sqlProvider.GetSelectDistinctClause(selectClauses.ToArray());
                   */ 
                } 
            }
            return _sqlProvider.GetSelectClause(selectClauses.ToArray());
        }

        protected virtual SqlStatement BuildLimit(SelectExpression select, SqlStatement literalSelect)
        {
            if (select.Limit != null)
            {
                var literalLimit = BuildExpression(select.Limit);
                if (select.Offset != null)
                {
                    var literalOffset = BuildExpression(select.Offset);
                    var literalOffsetAndLimit = BuildExpression(select.OffsetAndLimit);
                    return _sqlProvider.GetLiteralLimit(literalSelect, literalLimit,
                                                                                       literalOffset,
                                                                                       literalOffsetAndLimit);
                }
                return _sqlProvider.GetLiteralLimit(literalSelect, literalLimit);
            }
            return literalSelect;
        }
    }
}