﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data;
using System.Reflection;

using Vita.Common;
using Vita.Entities.Model;
using Vita.Logging;
using Vita.Data.Model;
using Vita.Data.Driver;
using Vita.Activation;


namespace Vita.Data.Management {

  /// <summary>Loads DB model from database. </summary>
  public partial class DbModelLoader {
    DbSettings _settings;
    DbModel _model; 
    DbDriver _driver;
    bool _supportsSchemas;
    DbSchemaManager _schemaManager;
    DbTypeMapper _typeMapper; 
    Log _log;
    StringSet _schemas = new StringSet(); 

    public DbModel LoadModel(DbSettings settings, Log log) {
      _settings = settings;
      _model = new DbModel(_settings);
      _log = log;
      _driver =  settings.Driver;
      _typeMapper = _driver.CreateTypeMapper(settings);
      _schemas.UnionWith(settings.Schemas);
      //note: if _schemas is empty, it means - load all.
      _supportsSchemas = _driver.Supports(DbCapabilities.Schemas);
      try {
        _schemaManager = _driver.CreateSchemaManager(_settings.SchemaManagementConnectionString);
        LoadSchemas();
        LoadTables();
        LoadTableColumns();
        LoadTableKeys();
        LoadRefConstraints();
        LoadTableIndexes();
        LoadProcedures();
        _schemaManager.OnModelLoaded(_model); 
      } finally {
        if (_schemaManager != null)
          _schemaManager.Close(); 
      }
      return _model; 
    }//method

    private void LogError(string message, params object[] args) {
      if (_log == null) //if no log then throw
        Util.Throw(message, args); 
      _log.Error(message, args);
    }

    private bool IncludeSchema(string schema) {
      if (!_supportsSchemas || _schemas.Count == 0)
        return true; 
      return _schemas.Contains(schema);
    }

    //TODO: fix all this; make full load of routines, with text and parameters - to compare
    private void LoadProcedures() {
      var supportsRoutines = _driver.Supports(DbCapabilities.StoredProcedures);
      if (!supportsRoutines) return;
      var data = _schemaManager.GetProcedures(null);
      foreach (DataRow row in data.Rows) {
        var schema = row.GetAsString("ROUTINE_SCHEMA");
        if (!IncludeSchema(schema)) continue; 
        var name = row.GetAsString("ROUTINE_NAME");
        var procText = row.GetAsString("ROUTINE_DEFINITION");
        var dbCommand = new DbCommandInfo(_model, schema, name, procText);
        dbCommand.StoredProcBody = row.GetAsString("ROUTINE_BODY");
        //detect if it is auto-generated
        var descrTag = _schemaManager.GetProcedureTag(row);
        if (string.IsNullOrEmpty(descrTag))
          dbCommand.Flags |= DbCommandFlags.NotGenerated;
        else
          dbCommand.DescriptiveTag = descrTag; 
        //if will be added to model.Commands automatically
      }//foreach
    }//method

    //Loads schemas into dbModel.Setup.Schemas only if it was originally empty
    private void LoadSchemas() {
      _model.Schemas.Clear(); //important - clear schemas that were copied from Setup
      if (!_supportsSchemas)
        return; 
      var data = _schemaManager.GetSchemas();
      foreach (DataRow row in data.Rows) {
        var schema = row.GetAsString("SCHEMA_NAME");
        _model.Schemas.Add(schema);
      }
    }

    private void LoadTables() {
      var data = _schemaManager.GetTables();
      foreach (DataRow row in data.Rows) {
        var schema = row.GetAsString("TABLE_SCHEMA");
        if (!IncludeSchema(schema)) continue; 
        var tableName = row.GetAsString("TABLE_NAME");
        var tblInfo = new DbTableInfo(_model, schema, tableName, null);
      }
    }

    private void LoadTableColumns() {
      DbTableInfo currTblInfo = null;
      var data = _schemaManager.GetColumns();
      foreach (DataRow row in data.Rows) {
        var schema = row.GetAsString("TABLE_SCHEMA");
        if (!IncludeSchema(schema)) continue;
        var tableName = row.GetAsString("TABLE_NAME");
        if(currTblInfo == null || currTblInfo.SchemaName != schema || currTblInfo.TableName != tableName) {
          currTblInfo = _model.GetTable(schema, tableName);
        }
        if(currTblInfo == null) continue; //it is a view
        var colName = row.GetAsString("COLUMN_NAME");
        var dataTypeString = row.GetAsString("DATA_TYPE");
        var size = row.GetAsLong("CHARACTER_MAXIMUM_LENGTH");
        var defaultExpr = row.GetAsString("COLUMN_DEFAULT");
        var typeInfo = _typeMapper.GetDbTypeInfo(row);
        if (typeInfo == null) {
          typeInfo = new DbTypeInfo(DbType.String, "nvarchar"); //fake type info for column in error
          LogError(
            "Failed to find TypeInfo for SQL data type [{0}]. Table {1}, column {2}.", dataTypeString, tableName, colName);
          var col = new DbColumnInfo(null, currTblInfo, colName, typeInfo);
          col.Flags |= DbColumnFlags.Error;
          continue;
        }
        var column = new DbColumnInfo(null, currTblInfo, colName, typeInfo);
        column.DefaultExpression = row.GetAsString("COLUMN_DEFAULT");
        // Let schema manager add any provider-specific info
        _schemaManager.OnColumnLoaded(_model, column, row);
      }//foreach
    }

    private void LoadTableKeys() {
      var data = _schemaManager.GetTableConstraints();
      foreach (DataRow row in data.Rows) {
        var schema = row.GetAsString("TABLE_SCHEMA");
        if (!IncludeSchema(schema)) continue;
        var tableName = row.GetAsString("TABLE_NAME");
        var table = _model.GetTable(schema, tableName);
        var constrName = row.GetAsString("CONSTRAINT_NAME");
        var constrTypeStr = row.GetAsString("CONSTRAINT_TYPE");
        if(table == null) continue;
        KeyType keyType;
        switch(constrTypeStr.Trim()) {
          case "PRIMARY KEY": keyType = KeyType.PrimaryKey;       break;
          case "FOREIGN KEY": keyType = KeyType.ForeignKey;       break;
          default: continue; //skip this constraint
        }
        var dbKey = new DbKeyInfo(constrName, table, keyType);
        if (keyType == KeyType.PrimaryKey)
          table.PrimaryKey = dbKey;
      }
      // Read key columns
      data = _schemaManager.GetConstraintColumnUsage();
      foreach (DataRow row in data.Rows) {
        var schema = row.GetAsString("TABLE_SCHEMA");
        if (!IncludeSchema(schema)) continue;
        var tableName = row.GetAsString("TABLE_NAME");
        var table = _model.GetTable(schema, tableName);
        if (table == null)  continue; 
        var colName = row.GetAsString("COLUMN_NAME");
        var keyName = row.GetAsString("CONSTRAINT_NAME");
        var key = table.Keys.FirstOrDefault(k => k.Name == keyName); 
        if(key == null) continue;
        var col = key.Table.Columns.FirstOrDefault(c => c.ColumnName == colName);
        key.Columns.Add(col);
      }
      //Mark PK columns as no update
      foreach (var t in _model.Tables)
        if (t.PrimaryKey != null)
          foreach (var c in t.PrimaryKey.Columns)
            c.Flags |= DbColumnFlags.NoUpdate; 
    }

    private void LoadRefConstraints() {
      var data = _schemaManager.GetReferentialConstraintsExt();
      foreach (DataRow row in data.Rows) {
        //Load names for Foreign key and Unique key
        var fkSchema = row.GetAsString("CONSTRAINT_SCHEMA");
        var ucSchema = row.GetAsString("UNIQUE_CONSTRAINT_SCHEMA");
        if (!IncludeSchema(ucSchema) || !IncludeSchema(fkSchema)) 
          continue;
        var fkTableName = row.GetAsString("C_TABLE");
        var fkName = row.GetAsString("CONSTRAINT_NAME");
        var fromKey = FindKey(fkSchema, fkTableName, fkName);

        var ucTableName = row.GetAsString("U_TABLE");
        var ucName = row.GetAsString("UNIQUE_CONSTRAINT_NAME");
        var toKey = FindKey(ucSchema, ucTableName, ucName);
        if (toKey == null) {
          //catch special case - foreign key to non-key set of fields
          if (fromKey != null && fromKey.KeyType.IsSet(KeyType.ForeignKey))
            fromKey.NotSupported = true; 
          continue; //it is not ref constraint
        }
        bool cascadeDelete = row.GetAsString("DELETE_RULE") == "CASCADE";
        var refConstraint = new DbRefConstraintInfo(_model, fromKey, toKey, cascadeDelete);
        fromKey.Table.RefConstraints.Add(refConstraint);
      }
    }

    private DbKeyInfo FindKey(string schema, string tableName, string keyName) {
      var table = _model.GetTable(schema, tableName);
      var key = table.Keys.FirstOrDefault(k => k.Name == keyName);
      return key; 
    }

    private void LoadTableIndexes() {
      var data = _schemaManager.GetIndexes();
      foreach (DataRow row in data.Rows) {
        var schema = row.GetAsString("TABLE_SCHEMA");
        if (!IncludeSchema(schema)) continue;
        var tableName = row.GetAsString("TABLE_NAME");
        var table = _model.GetTable(schema, tableName);
        if (table == null)
          continue;
        var indexName = row.GetAsString("INDEX_NAME");
        //indexName might be null for SQL Server (type_desc="HEAP") - just ignore these
        if (string.IsNullOrWhiteSpace(indexName)) 
          continue; 
        //primary keys are added through table constraints, so skip them here - except mark them as clustered
        var isPk = IsTrueOrNonZero(row, "primary_key");
        bool isClustered = _driver.Supports(DbCapabilities.ClusteredIndexes) && (row["clustered"].ToString() == "CLUSTERED");             
        if (isPk && isClustered) 
            table.PrimaryKey.KeyType = KeyType.ClusteredPrimaryKey; 
        if (isPk)
            continue; //PKs are added through constraints
        var isUnique = IsTrueOrNonZero(row, "unique");
        // Find existing, or create a new one
        var key = table.Keys.FindByName(indexName);
        //If key not exists yet, create it
        if (key == null)
          key = new DbKeyInfo(indexName, table, KeyType.Index);
        else
          key.KeyType |= KeyType.Index; 
        if (isClustered)
          key.KeyType |= KeyType.ClusteredIndex;
         if (isUnique)
          key.KeyType |= KeyType.UniqueIndex;
      }//while
      LoadIndexColumns(); 
    }

    private void LoadIndexColumns() {
      var data = _schemaManager.GetIndexColumns();
      foreach (DataRow row in data.Rows) {
        var schema = row.GetAsString("TABLE_SCHEMA");
        if (!IncludeSchema(schema)) continue;
        var tableName = row.GetAsString("TABLE_NAME");
        var table = _model.GetTable(schema, tableName);
        if (table == null) continue;
        var index_name = row.GetAsString("INDEX_NAME");
        var colName = row.GetAsString("COLUMN_NAME");
        var key = table.Keys.FindByName(index_name);
        if (key == null) {
          LogError("Loading index columns: failed to find owner index/key {0} for column {1}, table {2}.", index_name, colName, tableName);
          continue;
        }
        var dbCol = table.Columns.FindByName(colName);
        if (dbCol == null) continue;
        if (!key.Columns.Contains(dbCol))
          key.Columns.Add(dbCol);
        var desc = row.GetAsInt("IS_DESCENDING");
        bool isDesc = desc == 1;
        if (isDesc)
          key.DescendingColumns.Add(dbCol);
      }//foreach row
    }//method

    private static bool IsTrueOrNonZero(DataRow row, string columnName) {
      var value = row[columnName];
      if (value == null || value == DBNull.Value) return false;
      var t = value.GetType();
      if (t == typeof(bool))
        return (bool)value;
      if (t.IsInt()) {
        var lv = (long)value;
        return lv != 0;
      }
      return false;
    }

  }//class

}
