﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data;

using Vita.Common; 
using Vita.Entities.Model;
using Vita.Data.Model;
using Vita.Logging;
using Vita.Data.Driver;

namespace Vita.Data.Management {

  public class DbSchemaUpdater {
    Database _database;
    Log _log;
    DbSettings _settings; 
    DbSchemaManager _schemaManager;

    public DbSchemaUpdater(Database database, Log log) {
      _database = database; 
      _log = log;
      _settings = database.Settings;
      _schemaManager = _settings.Driver.CreateSchemaManager(_settings.SchemaManagementConnectionString);
    }

    public DbSchemaUpdateInfo GetSchemaUpdates() {
      //Check if commands compiled in DbModel
      var loader = new DbModelLoader();
      var oldModel = loader.LoadModel(_settings, _log);
      //Compare two models and get changes
      var modelComparer = new DbModelComparer();
      var updateInfo = modelComparer.GetSchemaChanges(_database.DbModel, oldModel, _log, _settings.SchemaUpdateOptions);
      return updateInfo;
    }

    public void ApplySchemaUpdates(DbSchemaUpdateInfo updates) {
      if (updates.Actions.Count == 0)
        return;
      try {
        ApplySchemaUpdatesImpl(updates);
      } finally {
        _schemaManager.Close();
      }
    }//method

    private void ApplySchemaUpdatesImpl(DbSchemaUpdateInfo updates) {
      var actionGroups = updates.Actions.GroupBy(a => a.ActionType).OrderBy(g => (int)g.Key);
      foreach (var ag in actionGroups) {
        var actionType = ag.Key; 
        ExecuteCustomUpdateActions(actionType);
        var actions = ag.ToList();
        foreach (var action in actions) {
          var target = action.ObjectInfo; 
          switch (action.ActionType) {
            case DbSchemaChangeActionType.SchemaDrop:   _schemaManager.DropSchema((string) target);    break;
            case DbSchemaChangeActionType.SchemaAdd:    _schemaManager.CreateSchema((string)target);   break; 
            case DbSchemaChangeActionType.TableDrop:    _schemaManager.DropTable((DbTableInfo) target);    break;
            case DbSchemaChangeActionType.TableAdd: _schemaManager.CreateTable((DbTableInfo)target); break;
            case DbSchemaChangeActionType.TableRename:
              var t = (DbTableInfo)target; 
              _schemaManager.RenameTable(t, t.Peer); 
              break;
            case DbSchemaChangeActionType.ColumnDrop: _schemaManager.DropColumn((DbColumnInfo)target); break;
            //New columns are always added as nullable; we change them to nullable later.
            case DbSchemaChangeActionType.ColumnAdd: _schemaManager.AddColumn((DbColumnInfo)target); break;
            case DbSchemaChangeActionType.ColumnRename: 
              var col =(DbColumnInfo)action.ObjectInfo;
              _schemaManager.RenameColumn(col, col.Peer); 
              break;
            case DbSchemaChangeActionType.ColumnInit:
            case DbSchemaChangeActionType.ColumnAddedInit: ResetNullValues((DbColumnInfo)target); break;
            case DbSchemaChangeActionType.ColumnModify: _schemaManager.ModifyColumn((DbColumnInfo)target); break;
            case DbSchemaChangeActionType.KeyDrop:      _schemaManager.DropDbKey((DbKeyInfo)target);  break;
            case DbSchemaChangeActionType.KeyAdd:
            case DbSchemaChangeActionType.PrimaryKeyAdd:
            _schemaManager.CreateDbKey((DbKeyInfo)target); 
            break;
            case DbSchemaChangeActionType.RefConstraintDrop:     _schemaManager.DropReferentialConstraint((DbRefConstraintInfo) target);   break;
            case DbSchemaChangeActionType.RefConstraintAdd: _schemaManager.CreateReferentialConstraint((DbRefConstraintInfo)target); break;
            case DbSchemaChangeActionType.RoutineDrop:           _schemaManager.DropRoutine((string) target);   break;
            case DbSchemaChangeActionType.RoutineAdd:    _schemaManager.AddRoutine((DbCommandInfo) target); break;
            // Columns are added as nullable; now change to non-nullable if it is non-nullable
            case DbSchemaChangeActionType.ColumnAddComplete:    CompleteAddingColumn((DbColumnInfo)target);        break; 
          }//switch ChangeType
        }//foreach action
      } //foreach actionGroup
    }//method

    private void ResetNullValues(DbColumnInfo column) {
      if (!HasNullValues(column))  
        return;
      if (string.IsNullOrWhiteSpace(column.TypeInfo.InitExpression)) { 
        _log.Error("Table {0}, column {1}: cannot initialize values before setting the column to nullable. " + 
              "The initial value is not provided by the driver.",
          column.Table.FullName, column.ColumnName);
      return; 
    }
      _schemaManager.InitNullColumnValues(column); 
    }

    private void CompleteAddingColumn(DbColumnInfo column) {
      if (column.Flags.IsSet(DbColumnFlags.Nullable)) {
        //Check if there are any rows that are null
        if (!HasNullValues(column)) 
          return; 
        // there are rows with null values. Check if we can update them to default value
        if (column.Member.ForeignKeyOwner != null) { //check if it is foreign key
          _log.Error("Table {0}, column {1}: cannot set to nullable. The column is a foreign key, and data contains rows with values not set. Use custom update actions to update the column.",
            column.Table.FullName, column.ColumnName);
          return; 
        }
      }
      //otherwise modify it
      _schemaManager.ModifyColumn(column);
    }

    private bool HasNullValues(DbColumnInfo column) {
      var sql = string.Format("Select Count(*) AS CNT FROM {0} WHERE {1} IS NULL;", column.Table.FullName, column.ColumnName);
      var dt = _schemaManager.ExecuteSelect(sql);
      int count = dt.Rows[0].GetAsInt("CNT");
      return (count > 0);
    }

    private void ExecuteCustomUpdateActions(DbSchemaChangeActionType actionType) {
      if (_settings.CustomUpdateActions.Count == 0) return;
      foreach (var upd in _settings.CustomUpdateActions) {
        if (upd.RunBefore == actionType)
          _schemaManager.ExecuteNonQuery(upd.Sql);

      }
    }

  }//class
}
