﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data;
using System.Reflection;

using Vita.Common;
using Vita.Entities;
using Vita.Entities.Model;
using Vita.Entities.Model.Construction;
using Vita.Entities.Runtime;
using Vita.Data.Driver;
using Vita.Logging;
using Vita.Activation;

namespace Vita.Data.Model.Construction {

  /// <summary>Builds Db model from entity model.</summary>
  public partial class DbModelBuilder {
    EntityModelInfo _entityModelInfo;
    DbModel _dbModel; 
    DbDriver _driver;
    DbTypeMapper _typeMapper;
    DbSettings _settings;
    int _tableKeyIndex; //is used to generate unique index names
    Log _log;

    public static DbModel Build(EntityModel model, DbSettings settings, Log log) {
      var dbmBuilder = new DbModelBuilder(model.ModelInfo, settings, log);
      var dbModel = dbmBuilder.Build();
      return dbModel;
    }

    private DbModelBuilder(EntityModelInfo entityModelInfo, DbSettings settings, Log log) {
      _log = log;
      _entityModelInfo = entityModelInfo;
      _settings = settings;
      _driver = _settings.Driver;
      _typeMapper = _driver.CreateTypeMapper(_settings); 
    }

    public DbModel Build() {
      _dbModel = new DbModel(_settings, _entityModelInfo.Model); 
      if (!CheckExplicitDbTypes())
        return _dbModel;
      BuildTables();
      CreateTableKeys();
      SetupOrderBy();
      //ref constraints are created in a separate loop, after creating PKs
      BuildRefConstraints();
      CheckObjectNames();
      BuildCrudCommands();
      _driver.OnDbModelConstructed(_dbModel);
      return _dbModel;

    }//method

    private void LogError(string message, params object[] args) {
      _log.Error(message, args);
    }

    private bool IsActive(EntityArea area) {
      return _dbModel.Schemas.Contains(area.SchemaName); 
    }

    private void BuildCrudCommands() {
      bool useSPs = _driver.Supports(DbCapabilities.StoredProcedures) && _settings.Options.IsSet(DbOptions.UseStoredProcs);
      var dbSqlBuilder = _driver.CreateSqlBuilder(_dbModel);
      foreach (var entityCommand in _entityModelInfo.GetCrudCommands()) {
        // Compiling only commands in the matching area
        if (!_dbModel.Schemas.Contains(entityCommand.Area.SchemaName)) continue;
        // We try to match commands by CommandID; it sometimes happens that two entity commands are identical and have the same underlying db command
        var cmd = _dbModel.GetCommand(entityCommand.DescriptiveTag);
        if (cmd != null) { //if it exists already, register it with other entity command
          _dbModel.RegisterDbObject(entityCommand, cmd); 
          continue;
        }
        cmd = dbSqlBuilder.BuildCommand(entityCommand);
        if (useSPs)
          _driver.ConvertToStoredProc(cmd);           
      }
    }

    //Create tables and regular "value" columns
    private void BuildTables() {
      foreach (var entityInfo in _entityModelInfo.Entities) {
        if (!IsActive(entityInfo.Area))
          continue; 
        var tableName = ConstructDefaultTableName(entityInfo);
        var table = new DbTableInfo(_dbModel, entityInfo.Area.SchemaName, tableName, entityInfo);
        //create Value columns 
        foreach (var member in entityInfo.Members)
          if (member.Kind == MemberKind.Column)
            CreateDbColumn(table, member);
      }//foreach entityInfo
    }

    private bool CheckExplicitDbTypes() {
      var result = true; 
      foreach (var ent in _entityModelInfo.Entities) {
        foreach (var member in ent.Members) {
          if (member.ExplicitDbType != null)
            result &= CheckExplicitDbType(member);
        }
      }
      return result; 
    }
    private bool CheckExplicitDbType(EntityMemberInfo member) {
      if (member.ExplicitDbType == null)
        return true;
      //assign default sizes
      switch (member.ExplicitDbType) {
        case CustomDbTypes.SmallMoney:
          if (member.Scale == 0) member.Scale = 10;
          if (member.Precision == 0) member.Precision = 4;
          break;
        case DbType.Currency:
          if (member.Scale == 0) member.Scale = 18;
          if (member.Precision == 0) member.Precision = 4;
          break;
        case CustomDbTypes.NText:
        case CustomDbTypes.Text: 
          member.Flags |= EntityMemberFlags.IsMemo;
          break; 
      }

      var dataType = member.DataType;
      if (dataType.IsNullableValueType())
        dataType = dataType.GetGenericArguments()[0];
      // lambda to construct error message on demand
      Func<string, string> getError = (string validTypes) =>
        Util.SafeFormat("Property {0}: DbType value {1} conflicts with property type {2}. Property type must be {3}.",
        member, CustomDbTypes.AsString(member.ExplicitDbType.Value), dataType.Name, validTypes);
      switch (member.ExplicitDbType) {
        //Standard (.NET-defined) DbType values
        case DbType.AnsiStringFixedLength:
        case DbType.StringFixedLength:
          if (dataType == typeof(string) || dataType == typeof(char)) return true;
          _log.Error(getError("string or char"));
          return false;

        case DbType.Xml:
        case DbType.AnsiString:
        case DbType.String:
        case CustomDbTypes.NText:
        case CustomDbTypes.Text: ;
          if (dataType == typeof(string)) return true;
          _log.Error(getError("string"));
          return false;

        case DbType.Boolean:
          if (dataType == typeof(bool)) return true;
          _log.Error(getError("bool"));
          return false;

        case DbType.Date:
        case DbType.Time:
        case DbType.DateTime:
        case DbType.DateTime2:
          if (dataType == typeof(DateTime)) return true;
          _log.Error(getError("DateTime"));
          return false;

        case DbType.DateTimeOffset:
          if (dataType == typeof(DateTimeOffset)) return true;
          _log.Error(getError("DateTimeOffset"));
          return false;

        case DbType.Guid:
          if (dataType == typeof(Guid)) return true;
          _log.Error(getError("Guid"));
          return false;

        case DbType.Double:
        case DbType.Single:
        case DbType.Decimal:
        case DbType.Currency:
        case DbType.VarNumeric:
        case CustomDbTypes.SmallMoney:
          if (dataType == typeof(double) || dataType == typeof(Single) || dataType == typeof(Decimal)) return true;
          _log.Error(getError("double, float(Single) or decimal"));
          return false;

        case DbType.Byte:
        case DbType.Int16:
        case DbType.Int32:
        case DbType.Int64:
        case DbType.SByte:
        case DbType.UInt16:
        case DbType.UInt32:
        case DbType.UInt64:
          if (dataType.IsInt()) return true;
          _log.Error(getError("int type"));
          return false;


        case DbType.Object:
          if (dataType == typeof(object)) return true;
          _log.Error(getError("Object"));
          return false;

        case DbType.Binary:
        case CustomDbTypes.VarBinary:
        case CustomDbTypes.Geography:
        case CustomDbTypes.Geometry:
        case CustomDbTypes.HierarchyId:
        case CustomDbTypes.Image:
        case CustomDbTypes.RowVersion:
          if (dataType == typeof(byte[])) return true;
          _log.Error(getError("byte[]"));
          return false;

      }//switch
      return true;
    }//method



    private DbColumnInfo CreateDbColumn(DbTableInfo table, EntityMemberInfo member) {
      bool isError = false; 
      var colName = member.ColumnName;
      string colDefault = member.ColumnDefault; //comes from attribute
      if (colDefault != null && member.DataType == typeof(string) && !colDefault.StartsWith("'"))
        colDefault = colDefault.Quote();

      //Figure out dbType
      var lookupType = member.DataType;
      if (lookupType.IsEnum)
        lookupType = typeof(Int32);
      var dbTypeInfo = _typeMapper.GetDbTypeInfo(member);
      if (dbTypeInfo == null) {
        isError = true;
        LogError("Driver failed to match db type for data type {0}, member {1}.{2}", member.DataType, member.Entity.FullName, member.MemberName);
        dbTypeInfo = new DbTypeInfo(DbType.String, "(error)"); //create "fake" type info, so we can continue
      }
      var dbColumn = new DbColumnInfo(member, table, colName, dbTypeInfo);
      if (!string.IsNullOrEmpty(colDefault))
        dbColumn.DefaultExpression = colDefault;
      if (member.AutoValueType == AutoType.Identity)
        dbColumn.Flags |= DbColumnFlags.Identity | DbColumnFlags.NoUpdate | DbColumnFlags.NoInsert;
      if (member.Flags.IsSet(EntityMemberFlags.Secret))
        dbColumn.Flags |= DbColumnFlags.NoUpdate; //updated only thru custom update method
      if (isError)
        dbColumn.Flags |= DbColumnFlags.Error;
      if (member.Flags.IsSet(EntityMemberFlags.IsMemo) && _driver.Supports(DbCapabilities.ForceNullableMemo)) //case for SQL CE
        dbColumn.Flags |= DbColumnFlags.Nullable;
      return dbColumn;
    }//method

    private void CreateTableKeys() {
      var createIndexesOnForeignKeys = _driver.Supports(DbCapabilities.NoIndexOnForeignKeys) 
                                       && _settings.Options.IsSet(DbOptions.AutoIndexForeignKeys);
      var supportsClustedIndex = _driver.Supports(DbCapabilities.ClusteredIndexes); 

      foreach (var table in _dbModel.Tables) {
        var entity = table.Entity;
        _tableKeyIndex = 0; //counter used to create unique index names
        foreach (var entityKey in entity.Keys) {
          var cols = new List<DbColumnInfo>(); 
          //Find columns and add them to the key
          foreach (var member in entityKey.ExpandedMembers) {
            var col = table.Columns.First(c => c.Member == member); 
            cols.Add(col);
          }//foreach member
          //Create DBKey
          var keyType = entityKey.KeyType;
          if (!supportsClustedIndex)
            keyType &= ~KeyType.Clustered;
          var keyName = ConstructDbKeyName(table, keyType, cols, entityKey.Name);
          var dbKey = new DbKeyInfo(keyName, table, keyType, entityKey);
          dbKey.Columns.AddRange(cols); 
          //Assign PK if it is PK
          if (keyType.IsSet(KeyType.PrimaryKey)) 
            table.PrimaryKey = dbKey;
          if (keyType.IsSet(KeyType.Clustered))
            foreach (var col in dbKey.Columns)
              col.Flags |= DbColumnFlags.ClusteredIndex;
          //copy descending members
          foreach (var col in dbKey.Columns)
            if (entityKey.IsDescending(col.Member))
              dbKey.DescendingColumns.Add(col); 
        }//foreach key
        // Check primary key and clustered index
        foreach (var col in table.PrimaryKey.Columns)
          col.Flags |= DbColumnFlags.NoUpdate | DbColumnFlags.PrimaryKey;
        // Check that clustered index exists; if not, change PK to be clustered - unless there's flag saying that entity has NO clustered index 
        if (_dbModel.Driver.Supports(DbCapabilities.ClusteredIndexes)) {
          var ci = table.Keys.FirstOrDefault(c => c.KeyType.IsSet(KeyType.ClusteredIndex));
          if (ci == null && table.PrimaryKey != null && !table.Entity.Flags.IsSet(EntityFlags.NoClusteredIndex))
            table.PrimaryKey.KeyType = KeyType.ClusteredPrimaryKey;
        }
        //Create indexes on foreign keys
        if (createIndexesOnForeignKeys) {
          //copy FKeys into a separate list, to avoid trouble when adding indexes to collection that we are iterating
          var foreignKeys = table.Keys.Where(k => k.KeyType == KeyType.ForeignKey).ToList();
          foreach (var key in foreignKeys) {
            //Check if there is already index with the same starting columns
            var matchingIndex = FindMatchingIndex(key);
            if (matchingIndex != null)
              continue;
            // construct index name, strip "IX_" prefix and replace it with "IX_FK_"
            var indexName = "IX_FK_" + ConstructDbKeyName(table, KeyType.Index, key.Columns).Substring(3);
            var dbIndex = new DbKeyInfo(indexName, table, KeyType.Index);
            dbIndex.Columns.AddRange(key.Columns);
          }
        }//if createIndexes

      }//foreach table
    }

    private DbKeyInfo FindMatchingIndex(DbKeyInfo key) {
      var table = key.Table;
      foreach (var index in key.Table.Keys) {
        //We are interested only in indexes or primary keys (or clustered primary keys)
        if (!index.KeyType.IsSet(KeyType.Index | KeyType.PrimaryKey)) 
          continue;  
        if (index.Columns.Count < key.Columns.Count) //it is not a match if it has fewer columns
          continue; 
        bool match = true;
        for (int i = 0; i < key.Columns.Count; i++)
          match &= (key.Columns[i] == index.Columns[i]);
        if (match)
          return index;
      }//foreach
      return null;   
    }

    private void BuildRefConstraints() {
      foreach (var table in _dbModel.Tables) {
        var entity = table.Entity;
        foreach (var member in entity.Members) {
          if (member.Kind != MemberKind.EntityRef)
            continue;
          var refInfo = member.ReferenceInfo;
          if (!IsActive(refInfo.ToKey.Entity.Area)) 
            continue; //target entity/table is not in this db model; so no foreign key link 
          var fromDbKey = _dbModel.LookupDbObject<DbKeyInfo>(refInfo.FromKey);
          var toDbKey = _dbModel.LookupDbObject<DbKeyInfo>(refInfo.ToKey);
          var cascadeDelete = member.Flags.IsSet(EntityMemberFlags.CascadeDelete);
          var dbRC = new DbRefConstraintInfo(_dbModel, fromDbKey, toDbKey, cascadeDelete, refInfo);
          table.RefConstraints.Add(dbRC);
        }
      }//foreach entityInfo
    }

    private void SetupOrderBy() {
      foreach (var table in _dbModel.Tables) {
        var orderBy = table.Entity.DefaultOrderBy;
        if (orderBy == null) continue; 
          table.DefaultOrderBy = ConstructOrderBy(table);
      }//foreach entity
    }//method

    private IList<DbOrderByEntryInfo> ConstructOrderBy(DbTableInfo table) {
      var orderByList = new List<DbOrderByEntryInfo>(); 
      var ent = table.Entity;
      foreach (var entry in ent.DefaultOrderBy) {
        var col = table.Columns.FirstOrDefault(c=>c.Member == entry.Member); 
        if (col == null)
          LogError("Failed to build ORDER BY list for table {0}: cannot find column for member {1}.", table.TableName, entry.Member.MemberName);
        var dbEntry = new DbOrderByEntryInfo() {Column = col, Descending = entry.Descending};
        orderByList.Add(dbEntry);
      }
      if (orderByList.Count == 0)
        return null; 
      return orderByList; 
    }

    //default, we run thru naming policy later
    private string ConstructDefaultTableName(EntityInfo entity) {
      var tableName = entity.TableName; 
      if (string.IsNullOrEmpty(tableName)) {
        tableName = entity.EntityType.Name;
        if (tableName.StartsWith("I"))
          tableName = tableName.Substring(1);
      }
      return tableName; 
    }

    private void CheckObjectNames() {
      foreach (var table in _dbModel.Tables) {
        if (_settings.Options.IsSet(DbOptions.PluralizeTableNames))
          table.TableName = Util.Pluralize(table.TableName); 
        table.TableName = _settings.NamingPolicy.CheckName(DbNamedObjectType.Table, table.TableName, table);
        foreach(var col in table.Columns)
          col.ColumnName = _settings.NamingPolicy.CheckName(DbNamedObjectType.Column, col.ColumnName, col);
        foreach (var key in table.Keys)
          key.Name = _settings.NamingPolicy.CheckName(DbNamedObjectType.Key, key.Name, key);
      }
    }//method


    private string ConstructDbKeyName(DbTableInfo table, KeyType keyType, IList<DbColumnInfo> columns, string entityKeyName = null) {
      var tableName = table.TableName;
      var joinedColNames = columns.GetNames(removeUnderscores: true); 
      string keyName = string.Empty; 
      switch(keyType) {
        case KeyType.ClusteredPrimaryKey:
        case KeyType.PrimaryKey: 
          keyName = "PK_" + tableName; 
          break;
        case KeyType.ClusteredIndex:
          keyName =  "IXC_" + tableName + "_" + joinedColNames;
          break; 
        case KeyType.UniqueIndex:
          keyName = "IXU_" + tableName + "_" + joinedColNames;
          break;
        case KeyType.Index:
          keyName = "IX_" + tableName + "_" + joinedColNames;
          break;
        case KeyType.ForeignKey:
          keyName = entityKeyName; //entity key name is already well-formed
          break;
      }//switch
      //check that it does not exist; if it does, append index to it.
      var existing = table.Keys.FirstOrDefault(c => c.Name == keyName);
      if (existing != null)
        keyName = keyName + (_tableKeyIndex++); 
      return keyName; 
    }

  }//class
}//namespace
