﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data;

using Vita.Common;
using Vita.Entities.Runtime;
using Vita.Entities.Model;
using Vita.Data.Driver;
using Vita.Data.Management;
using Vita.Data.Model.Construction;
using Vita.Entities;
using Vita.Activation;

namespace Vita.Data.Model {

  //Physical model
  public class DbModel {
    public EntityModel EntityModel;
    public DbSettings Settings;
    public readonly object LockObject = new object();

    internal bool CustomCommandsCompiled { get; private set; }

    // Contains all DbObjects (tables, keys, commands) by name. Additionally commands are registered by CommandId
    Dictionary<string, DbModelObjectBase> _objectsTable = new Dictionary<string, DbModelObjectBase>(StringComparer.InvariantCultureIgnoreCase);  
    IList<DbTableInfo> _tables = new List<DbTableInfo>();
    IList<DbCommandInfo> _commands = new List<DbCommandInfo>();


    public IEnumerable<DbTableInfo> Tables {
      get {return _tables;}
    }
    public IEnumerable<DbCommandInfo> Commands {
      get { return _commands; }
    }
    public DbDriver Driver { get { return Settings.Driver; } }
    public readonly Vita.Data.Driver.LinqSqlProvider LinqSqlProvider;
    //Note: might be different from Settings.Schemas. For DbModels loaded from database, Settings.Schemas contains schemas to check
    // DbModel.Schemas contains a list of schemas actually found in the database. 

    public StringSet Schemas = new StringSet();


    //Table of all Db objects; accessed by entity-model object as key 
    private Dictionary<HashedObject, DbModelObjectBase> _allObjects = new Dictionary<HashedObject, DbModelObjectBase>();

    /// <summary>Constructs DbModel from EntityModel.</summary>
    /// <param name="settings"></param>
    /// <param name="entityModel"></param>
    internal DbModel(DbSettings settings, EntityModel entityModel) {
      Util.Check(entityModel != null, "entityModel parameter may not be null.");
      Settings = settings;
      EntityModel = entityModel; //EntityModel is null for dbModel created from database
      LinqSqlProvider = this.Driver.GetLinqSqlProvider();
      Schemas.UnionWith(Settings.Schemas);
    }

    /// <summary>Constructs an empty DbModel. This constructor is used for models loaded from database. </summary>
    /// <param name="settings"></param>
    internal DbModel(DbSettings settings) {
      Settings = settings;
    }


    #region Construction-time methods
    public void AddTable(DbTableInfo table) {
      Util.Check(!_objectsTable.ContainsKey(table.FullName), "Duplicate table in DbModel: {0}", table.FullName);
      _tables.Add(table);
      _objectsTable[table.FullName] = table; 
    }

    public void AddCommand(DbCommandInfo command) {
      Util.Check(!_objectsTable.ContainsKey(command.FullCommandName), "Duplicate proc in DbModel: {0}", command.FullCommandName);
      if (command == null) return; 
      _commands.Add(command);
      _objectsTable[command.FullCommandName] = command;
      var entCommand = command.EntityCommand;
      if (entCommand != null)
        _objectsTable[entCommand.DescriptiveTag] = command;
    }

    public void RegisterDbObject(HashedObject entityModelObject, DbModelObjectBase dbObject) {
      if (entityModelObject == null) return; 
      _allObjects[entityModelObject] = dbObject; 
    }//method
    #endregion

    public T LookupDbObject<T>(HashedObject key, bool throwNotFound = false) where T : DbModelObjectBase {
      DbModelObjectBase result;
      _allObjects.TryGetValue(key, out result);
      if(result == null && throwNotFound)
        Util.Throw("Failed to lookup DbObject for entity model object {0}", key);
      return (T)result;
    }
    public DbModelObjectBase GetDbObject(string fullName) {
      DbModelObjectBase obj;
      _objectsTable.TryGetValue(fullName, out obj);
      return obj;
    }
    public DbCommandInfo GetCommand(string fullNameOrId) {
      return GetDbObject(fullNameOrId) as DbCommandInfo;
    }
    public DbTableInfo GetTable(string fullName) {
      return GetDbObject(fullName) as DbTableInfo;
    }
    public DbTableInfo GetTable(string schemaName, string tableName) {
      var fullName = Settings.Driver.GetFullName(schemaName, tableName);
      return GetTable(fullName);
    }
    
    public DbTableInfo GetTable(Type entityType, bool throwIfNotFound = true) {
      var entInfo = this.EntityModel.ModelInfo.GetEntityInfo(entityType, throwIfNotFound);
      if (entInfo == null)
        return null; 
      var tableInfo = this.LookupDbObject<DbTableInfo>(entInfo, throwIfNotFound);
      return tableInfo;
    }

    public DbKeyInfo FindKey(string name) {
      foreach (var t in this.Tables)
        foreach (var key in t.Keys)
          if (string.Compare(key.Name, name, true) == 0) return key;
      return null;
    }

    public bool ContainsSchema(string schema) {
      var supportsSchemas = Settings.Driver.Supports(DbCapabilities.Schemas);
      if (!supportsSchemas && string.IsNullOrEmpty(schema)) 
        return true; // we consider NULL schema is OK 
      return Schemas.Contains(schema);
    }

    internal void CompileCustomCommands(string connectionString) {
      var driver = Settings.Driver;
      this.CustomCommandsCompiled = true;
      var customCommands = EntityModel.ModelInfo.GetCustomCommands();
      if (customCommands.Count == 0)
        return;
      /* //Temporarily disabled
            bool useSPs = driver.Supports(DbCapabilities.StoredProcedures) && Settings.Options.IsSet(DbOptions.UseStoredProcs);
            var compiler = driver.CreateCustomCommandCompiler(this, connectionString);
            foreach (var entityCommand in customCommands) {
              // Compiling only commands in the matching area
              if (!Schemas.Contains(entityCommand.Area.SchemaName)) continue;
              var cmd = compiler.CompileCommand(entityCommand);
              if (useSPs)
                driver.ConvertToStoredProc(cmd);
            }
       */ 
    }



  }//class

}
