﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data;
using System.Threading;

using Vita.Common; 
using Vita.Entities.Model; 
using Vita.Entities;
using Vita.Data.Management;

namespace Vita.Data.Model {

  public enum DbExecutionType {
    Query,
    NonQuery,
    Scalar,
  }

  //Change type for db objects - columns, tables, etc
  public enum DbObjectChange {
    None,
    New,
    Drop,
    Modify,
  }


  // Base for all DB model objects. Constructor makes automatic registration of the object in DbModel's dictionary. 
  // This dictionary is used to lookup DbObject by its entity model counterpart.
  public abstract class DbModelObjectBase {
    public DbModel DbModel;

    public DbModelObjectBase(DbModel dbModel, HashedObject entityModelObject) {
      DbModel = dbModel;
      if (DbModel != null)
        DbModel.RegisterDbObject(entityModelObject, this); 
    }
  }

  public class DbTableInfo : DbModelObjectBase {
    public string SchemaName; 
    public string TableName;
    public readonly string FullName;
    public IList<DbColumnInfo> Columns = new List<DbColumnInfo>();
    public IList<DbKeyInfo> Keys = new List<DbKeyInfo>();
    public IList<DbRefConstraintInfo> RefConstraints = new List<DbRefConstraintInfo>();
    public IList<DbOrderByEntryInfo> DefaultOrderBy; 

    public DbKeyInfo PrimaryKey;
    //Props for schema update process
    public DbObjectChange ChangeType;
    public EntityInfo Entity;
    
    public object Tag;
    //Used in analyzing changes. In old model, points to the object in new model, and vice versa
    public DbTableInfo Peer;
    
    public DbTableInfo(DbModel model, string schemaName, string tableName, EntityInfo entity) : base(model, entity) {
      SchemaName = schemaName; 
      TableName = tableName;
      Entity = entity; //might be null
      FullName = model.Driver.GetFullName(schemaName, tableName);
      model.AddTable(this); 
    }

    public override string ToString() {
      return TableName;
    }
    public override int GetHashCode() {
      return FullName.GetHashCode(); 
    }
  }//class
  


  public delegate object DbValueConverter(object value);

  public class DbTypeInfo {
    public Type ClrType; 
    public DbType DbType;
    public string SqlTypeSpec;
    public int Size;
    public byte Precision;
    public byte Scale;
    public bool IsNullable;
    public DbValueConverter ReadConverter; // Db -> entity
    public DbValueConverter WriteConverter; // entity -> Db
    public string InitExpression; //value to initialize nullable column when it switches to non-nullable; this string will be put directly into update SQL

    public DbTypeInfo(DbType dbType, string sqlTypeSpecTemplate, bool isNullable = false, long size = 0, 
              string initExpression = null,
              byte precision = 0, byte scale = 0, 
              DbValueConverter readConverter = null, DbValueConverter writeConverter = null, 
              Type clrType = null) { 
      DbType = dbType; 
      //Util.Check(DbType != null, "DbTypeInfo: invalid value of dbType parameter: CustomDbTypes.Unspecified.");  
      ClrType = clrType ?? dbType.GetDefaultClrType(); 
      if (sqlTypeSpecTemplate.Contains('{'))
        SqlTypeSpec = string.Format(sqlTypeSpecTemplate, size, precision, scale);
      else
        SqlTypeSpec = sqlTypeSpecTemplate;
      IsNullable = isNullable;
      Size = size > int.MaxValue ? -1 : (int)size;
      InitExpression = initExpression ?? GuessSqlInitExpression(ClrType); 
      Precision = precision;
      Scale = scale;
      ReadConverter = readConverter ?? _noconv;
      WriteConverter = writeConverter ?? _noconv;
    }

    private static DbValueConverter _noconv = x => x;

    public virtual string GuessSqlInitExpression(Type type) {
      if (type == typeof(string)) return "''";
      if (type.IsInt()) return "0";
      if (type == typeof(float) || type == typeof(double) || type == typeof(decimal)) return "0";
      if (type == typeof(Guid)) return "'" + Guid.Empty + "'";
      if (type == typeof(DateTime)) return "'1900-01-01'";
      if (type == typeof(bool)) return "0";
      return null;
    }

    public override string ToString() {
      return SqlTypeSpec; 
    }

  }//class


  public class DbOrderByEntryInfo {
    public DbColumnInfo Column;
    public bool Descending;
    public override string ToString() {
      var dq = "\"";
      return dq + Column.ColumnName + dq + (Descending ? " DESC" : string.Empty);
    }
  }
  

  [Flags]
  public enum DbColumnFlags {
    None = 0,
    Nullable = 1,
    PrimaryKey = 1 << 1,
    ClusteredIndex = 1 << 2,

    Identity = 1 << 4,
    NoUpdate = 1 << 5,  // columns with auto-values
    NoInsert = 1 << 6,

    Error = 1 << 16, // column info was loaded with error from database
  }
  
  public class DbColumnInfo : DbModelObjectBase {

    public readonly DbTableInfo Table;
    public EntityMemberInfo Member;
    
    public string ColumnName;
    public DbColumnFlags Flags;

    public DbTypeInfo TypeInfo;
    public string DefaultExpression;

    //Schema update
    //Used in analyzing changes. In old model, points to the object in new model, and vice versa
    public DbColumnInfo Peer;
    public DbObjectChange ChangeType;

    public DbColumnInfo(EntityMemberInfo member, DbTableInfo table, string columnName, DbTypeInfo typeInfo)
                          : base(table.DbModel, member) {
      Member = member;
      Table = table;
      ColumnName = columnName;
      TypeInfo = typeInfo;
      if (TypeInfo.IsNullable)
        Flags |= DbColumnFlags.Nullable;
      if (TypeInfo.DbType == CustomDbTypes.RowVersion)
        Flags |= DbColumnFlags.NoUpdate | DbColumnFlags.NoInsert;
      // TODO: come up with some more advanced algorithm of column ordering (PK, ClusteredIndex columns, Natural key columns, etc)
      //Make sure PK columns appear first
      if (Member != null && Member.Flags.IsSet(EntityMemberFlags.PrimaryKey))
        Table.Columns.Insert(0, this);
      else 
        Table.Columns.Add(this);
    }

    public override string ToString() {
      return Table.TableName + "." + ColumnName + ":" + TypeInfo;
    }

  }//class

  //DbKeyInfo is a generalized concept for all database objects based on group of columns: 
  // indexes, primary or foreign keys. Note that we use the term "foreign key" in a narrow sense, 
  // to refer ONLY to the set of columns in child table; the "relation" between foreign key and primary/unique
  // key in another table is presented by DbRefConstraintInfo class.
  public class DbKeyInfo : DbModelObjectBase {
    public string Name;
    public DbTableInfo Table;
    public KeyType KeyType;
    public List<DbColumnInfo> Columns = new List<DbColumnInfo>();
    public HashSet<DbColumnInfo> DescendingColumns = new HashSet<DbColumnInfo>();
    public EntityKeyInfo EntityKey;
    public List<DbOrderByEntryInfo> OrderBy;
    //used only when loading schema from database and finding a foreign key with target not PK, but a random set of fields
    public bool NotSupported; 

    //Used in analyzing changes. In old model, points to the object in new model, and vice versa
    public DbKeyInfo Peer;
    public DbObjectChange ChangeType; 

    public override string ToString() {
      return Name + "(" + KeyType + ")";
    }
    public DbKeyInfo(string name, DbTableInfo table, KeyType type, EntityKeyInfo entityKey = null) : base(table.DbModel, entityKey) {
      Name = name;
      Table = table;
      KeyType = type; 
      EntityKey = entityKey;
      table.Keys.Add(this);
    }

  }//class

  public class DbRefConstraintInfo : DbModelObjectBase {
    public DbKeyInfo FromKey;
    public DbKeyInfo ToKey;
    public EntityReferenceInfo OwnerReference; 
    //TODO: implement handling update rule for relations
    public bool CascadeDelete;

    //Used in analyzing changes. In old model, points to the object in new model, and vice versa
    public DbObjectChange ChangeType; 
    public DbRefConstraintInfo Peer;

    public DbRefConstraintInfo(DbModel dbModel, DbKeyInfo fromKey, DbKeyInfo toKey, bool cascadeDelete, EntityReferenceInfo ownerRef = null)  : base(dbModel, ownerRef) {
      FromKey = fromKey;
      ToKey = toKey;
      OwnerReference = ownerRef;
      CascadeDelete = cascadeDelete;
    }

    public override string ToString() {
      return FromKey.Name;
    }
  }

  public class DbParamInfo {
    public EntityCommandParameter Owner; 
    public string Name;
    public DbTypeInfo TypeInfo;
    public ParameterDirection Direction = ParameterDirection.Input;
    public bool IsNullable;
    public DbColumnInfo SourceColumn; //might be null
    // for parameters of commands created from LINQ queries; in fact, parameter is a literal constant 
    public object DefaultValue;
    //Index of value in argument array of the command call; 
    // For custom commands based on custom LINQ queries might be -1 (it is artificial parameter created by LINQ for some constant);
    // in this case the DefaultValue is used
    public int ArgIndex; 

    public DbParamInfo(EntityCommandParameter owner, string name, DbColumnInfo sourceColumn, int argIndex) 
                        : this(owner, name, argIndex) {
      SourceColumn = sourceColumn;
      Direction = owner.Direction;
      TypeInfo = sourceColumn.TypeInfo;
      IsNullable = sourceColumn.Flags.IsSet(DbColumnFlags.Nullable);
    }

    public DbParamInfo(EntityCommandParameter owner, string name, DbTypeInfo typeInfo, int argIndex) : this(owner, name, argIndex) {
      TypeInfo = typeInfo;
      Direction = ParameterDirection.Input;
    }

    private DbParamInfo(EntityCommandParameter owner, string name, int argIndex) {
      Owner = owner;
      Name = name;
      ArgIndex = argIndex;
      DefaultValue = owner.DefaultValue; 
    }

    public override string ToString() {
      return this.Name + "/" + TypeInfo; 
    }
  }//class

  [Flags]
  public enum DbCommandFlags {
    None = 0,
    WithPaging = 1,
    NotGenerated = 1 << 1, // set for commands from procs found in the database that are not marked as generated by VITA

  }

  public class DbCommandInfo : DbModelObjectBase {
    public string CommandName { get; private set; }
    public string Description;
    public DbTableInfo Table;
    public string Sql;
    public EntityCommandKind Kind;
    public DbKeyInfo SelectionKey; //optional, set for Select-by-key commands, either primary or foreign key
    public DbExecutionType ExecutionType;
    public List<DbParamInfo> Parameters = new List<DbParamInfo>();
    public List<DbParamInfo> OutputParameters = new List<DbParamInfo>();
    public DbParamInfo ReturnValueParameter;
    public DbColumnInfo[] QueryColumnMappings; //for select command
    public DbCommandFlags Flags;
    public string DescriptiveTag; 

    //stored proc 
    public string Schema; //stored proc schema
    public string StoredProcText;
    public string StoredProcBody;
    public string FullCommandName;

    //for partial selects/updates - fields to be selected/updated
    //public FieldMask FieldMask;
    public DbCommandInfo Peer;
    public EntityCommand EntityCommand;

    public DbCommandInfo(EntityCommand entityCommand, string commandName, DbTableInfo table, DbExecutionType executionType, string sql)
      : base(table.DbModel, entityCommand) {
      EntityCommand = entityCommand;
      CommandName = commandName;
      Table = table;
      ExecutionType = executionType;
      Sql = sql;
      DescriptiveTag = EntityCommand.DescriptiveTag;
      Description = EntityCommand.Description;
      //derived entities
      Schema = entityCommand.Area.SchemaName;
      FullCommandName = Table.DbModel.Driver.GetFullName(Schema, CommandName);
      Kind = entityCommand.Kind;
      if (Kind == EntityCommandKind.SelectAllPaged)
        Flags |= DbCommandFlags.WithPaging;
      var dbModel = table.DbModel; 
      dbModel.AddCommand(this);
    }

    //Creates a stub for commands loaded from the database
    public DbCommandInfo(DbModel model, string schemaName, string commandName, string text) : base(model, null) {
      Schema = schemaName;
      CommandName = commandName;
      StoredProcText = text;
      FullCommandName = model.Driver.GetFullName(Schema, CommandName);
      model.AddCommand(this);
    }


    public DbParamInfo AddParameter(EntityCommandParameter owner, string name, DbColumnInfo column, int argIndex) {
      return AddParameter(new DbParamInfo(owner, name, column, argIndex));
    }

    public DbParamInfo AddParameter(EntityCommandParameter owner, string name, DbTypeInfo typeInfo, int argIndex) {
      return AddParameter(new DbParamInfo(owner, name, typeInfo, argIndex));
    }

    private DbParamInfo AddParameter(DbParamInfo parameter) {
      Parameters.Add(parameter);
      switch (parameter.Direction) {
        case ParameterDirection.Output:
        case ParameterDirection.InputOutput:
          OutputParameters.Add(parameter);
          break;
        case ParameterDirection.ReturnValue:
          ReturnValueParameter = parameter;
          break;
      }
      return parameter;
    }

    public override string ToString() {
      return FullCommandName;
    }

    public bool StoredProcDefined() {
      return !string.IsNullOrEmpty(StoredProcText);
    }

  }//class

  public static class DbModelExtensions {
    public static bool IsSet(this DbColumnFlags flags, DbColumnFlags flag) {
      return (flags & flag) != 0;
    }

    public static bool IsSet(this DbCommandFlags flags, DbCommandFlags flag) {
      return (flags & flag) != 0;
    }
    public static DbColumnInfo GetColumnByMemberName(this DbTableInfo table, string memberName) {
      return table.Columns.FirstOrDefault(c => c.Member.MemberName == memberName);
    }
  }




}
