﻿using System;
using System.Collections.Generic;
using System.Data;
using System.Linq;
using System.Text;

using Vita.Common;
using Vita.Entities;
using Vita.Entities.Model;
using Vita.Data.Driver;

namespace Vita.Data.Model {

  public static class DbModelHelper {

    public static Type GetDefaultClrType(this DbType dbType) {
      switch (dbType) {
        case DbType.AnsiString:
        case DbType.AnsiStringFixedLength:
        case DbType.String:
        case DbType.StringFixedLength:
          return typeof(string);
        case DbType.Xml: return typeof(string);

        case DbType.Binary: 
          return typeof(byte[]);
        case DbType.Boolean: return typeof(bool);
        case DbType.Guid: return typeof(Guid);

        case DbType.Currency:
        case DbType.Decimal:
        case DbType.VarNumeric:
          return typeof(Decimal);

        case DbType.Date:
        case DbType.DateTime:
        case DbType.DateTime2:
        case DbType.Time:
        case CustomDbTypes.SmallDateTime:
          return typeof(DateTime);

        case DbType.DateTimeOffset:
          return typeof(DateTimeOffset);

        case DbType.Byte: return typeof(Byte);
        case DbType.UInt16: return typeof(UInt16);
        case DbType.UInt32: return typeof(UInt32);
        case DbType.UInt64: return typeof(UInt64);
        case DbType.SByte: return typeof(SByte);
        case DbType.Int16: return typeof(Int16);
        case DbType.Int32: return typeof(Int32);
        case DbType.Int64: return typeof(Int64);

        case DbType.Single: return typeof(Single);
        case DbType.Double: return typeof(Double);

        //custom db types
        case CustomDbTypes.VarBinary:
        case CustomDbTypes.Image: 
          return typeof(byte[]);
        case CustomDbTypes.Text:
        case CustomDbTypes.NText: return typeof(string);
        case CustomDbTypes.RowVersion: return typeof(byte[]);
        //Not sure what's the best type for the following
        case CustomDbTypes.HierarchyId: 
        case CustomDbTypes.Geography: 
        case CustomDbTypes.Geometry: 
          return typeof(byte[]);
        case CustomDbTypes.SmallMoney: return typeof(Decimal);

        case DbType.Object: return typeof(object);
        default:
          return typeof(object);
      }//switch
    }

    public static bool IsInt(this DbType dbType) {
      switch (dbType) {
        case DbType.Byte:
        case DbType.Int16:
        case DbType.Int32:
        case DbType.Int64:
        case DbType.UInt16:
        case DbType.UInt32:
        case DbType.UInt64:
          return true;
        default:
          return false;
      }
    }

    public static string EnumToSqlString(object value, Type type) {
      return ((int)value).ToString();
    }
    public static string StringToSqlString(object value, Type type) {
      if (value == null || value == DBNull.Value) return "NULL";
      var sv = (string)value;
      if (sv.Contains("'")) //if contains quotes, then double them
        sv = sv.Replace("'", "''");
      return "'" + sv + "'";
    }

    public static string GetSqlNameListWithOrderSpec(this DbKeyInfo key) {
      if (key.DescendingColumns.Count == 0)
        return GetSqlNameList(key.Columns); 
      var colSpecs = new string[key.Columns.Count];
      for (int i = 0; i < key.Columns.Count; i++) {
        var col = key.Columns[i];
        var isDesc = key.DescendingColumns.Contains(col);
        var desc = isDesc ? " DESC" : string.Empty;
        colSpecs[i] = '"' + col.ColumnName + '"' + desc;
      }
      return string.Join(", ", colSpecs);
    }

    public static string GetSqlNameList(this IEnumerable<DbColumnInfo> columns, string tablePrefix = null) {
      if (string.IsNullOrEmpty(tablePrefix))
        return "\"" + columns.GetNames("\", \"", false) + "\"";
      var delim = "\", " + tablePrefix + ".\"";
      return tablePrefix + ".\"" + columns.GetNames(delim, false) + "\"";
    }

    public static string GetNames(this IEnumerable<DbColumnInfo> columns, string delimiter = "", bool removeUnderscores = false) {
      var names = columns.Select(c => c.ColumnName);
      var result = string.Join(delimiter, names);
      if (removeUnderscores)
        result = result.Replace("_", string.Empty);
      return result;
    }
    public static IEnumerable<DbColumnInfo> GetSelectable(this IEnumerable<DbColumnInfo> columns) {
      return columns.Where(col => !col.Member.Flags.IsSet(EntityMemberFlags.Secret));
    }
    public static DbColumnInfo FindByName(this IEnumerable<DbColumnInfo> columns, string name) {
      return columns.FirstOrDefault(c => string.Compare(c.ColumnName, name, true) == 0);
    }
    public static DbKeyInfo FindByName(this IEnumerable<DbKeyInfo> keys, string name) {
      return keys.FirstOrDefault(key => string.Compare(key.Name, name, ignoreCase: true) == 0);
    }

  }//class

}
