﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Vita.Common;
using Vita.Common.Graphs;
using Vita.Entities;
using Vita.Entities.Model;
using Vita.Entities.Runtime;

namespace Vita.Data {

  public class EntityUpdateBatch {
    public readonly Guid Id; //unique batch Id
    public readonly EntitySession Session;
    public readonly DateTime UpdateTime;
    //in case of error this field contains the exception
    public Exception Exception;
    public Dictionary<string, object> CustomData = new Dictionary<string, object>(); 

    // Note about isolated deletes and inserts.
    // Isolated deletes and inserts are for entities that are NOT referenced by any other records; deletes are performed first, 
    // and inserts are done last. This is a way to mitigate the conflict when some entity is deleted and "re-inserted" 
    // - inserted new entity with the same primary key or with the same value of some unique key.
    // The typical case for this is link records (in many-to-many relationships) - when user removes/adds link, or simply 
    // the order is changed by inserting/deleting linked records
    public readonly List<EntityRecord> IsolatedDeletes = new List<EntityRecord>();
    public readonly List<EntityRecord> IsolatedInserts = new List<EntityRecord>();
    public readonly List<EntityRecord> Inserts = new List<EntityRecord>();
    public readonly List<EntityRecord> Updates = new List<EntityRecord>();
    public readonly List<EntityRecord> Deletes = new List<EntityRecord>();
    public readonly HashSet<EntityInfo> EntityInfos = new HashSet<EntityInfo>();
    public readonly List<EntityRecord> AllRecords = new List<EntityRecord>(); //keeps all records that are affected by this batch

    public bool IsPrepared { get; private set; } //true means that records are split
    public bool IsSequenced { get; private set; }

    public EntityUpdateBatch(EntitySession session, DateTime updateTime) {
      Session = (EntitySession) session;
      UpdateTime = updateTime;
      Id = Vita.Common.Externals.NewSequentialGuid();
    }

    public object GetCustomData(string key) {
      object data;
      CustomData.TryGetValue(key, out data);
      return data; 
    }

    public void CheckPrepared() {
      if (!IsPrepared)
        Prepare(); 
    }

    //Private methods
    public void Prepare() {
      IsolatedDeletes.Clear();
      Inserts.Clear();
      Updates.Clear();
      Deletes.Clear();
      EntityInfos.Clear(); 
      //Split all records into 3 sets: inserts, updates, deletes
      foreach (var rec in Session.RecordsChanged) {
        EntityInfos.Add(rec.EntityInfo);
        AllRecords.Add(rec); 
        rec.TopologicalIndex = 0;
        switch (rec.Status) {
          case EntityStatus.New:
            if (rec.EntityInfo.Flags.IsSet(EntityFlags.IsReferenced))
              Inserts.Add(rec);
            else
              IsolatedInserts.Add(rec); 
            break;
          case EntityStatus.Modified: Updates.Add(rec); break;
          case EntityStatus.Deleting:
            if (rec.EntityInfo.Flags.IsSet(EntityFlags.IsReferenced))
              Deletes.Add(rec);
            else
              IsolatedDeletes.Add(rec);
            break;
        }//switch
      }//foreach cmd
      IsPrepared = true;
    }

    public void SequenceRecords() {
      CheckPrepared(); 
      // The sequence:
      //  1. Inserts in the reverse topological order
      //  2. Updates, no order
      //  3. Deletes, in direct topological order
      SequenceRecords(Inserts);
      Inserts.Reverse();
      SequenceRecords(Deletes);
      IsSequenced = true; 
    }//method


    private void SequenceRecords(List<EntityRecord> records) {
      if (records.Count < 2) return; 
      //Check if there are non-trivial subgroups; non-trivial means that sorting by table's index is not enough,
      // - entities are either self-refencing or there are entities with the same Topological index 
      // (which means there is a circular reference between entity types)
      // if yes, we need to sequence records inside groups
      var needRecordSequencing = records.Any(NeedsRecordLevelSequencing);
      if (needRecordSequencing) {
        var groups = records.GroupBy(r => r.EntityInfo.TopologicalIndex);
        foreach (var group in groups) {
          if (group.Count() < 2) continue;
          if (!group.Any(NeedsRecordLevelSequencing)) continue;
          SequenceRecordsGroup(group);
        }
      }//if needRecordSequencing
      //Finally sort all
      records.Sort(CompareTopologicalIndexes);
    }//method

    private bool NeedsRecordLevelSequencing(EntityRecord record) {
      var entInfo = record.EntityInfo;
      return entInfo.Flags.IsSet(EntityFlags.TopologicalGroupNonTrivial | EntityFlags.SelfReferencing); 
    }

    private void SequenceRecordsGroup(IEnumerable<EntityRecord> records) {
      var graph = new Graph();
      foreach (var rec in records) {
        var ent = rec.EntityInfo;
        foreach (var refMember in ent.RefMembers) {
          var targetEnt = rec.GetValue(refMember); 
          //If reference is not modified or not set, then nothing to do
          if (targetEnt == null)
            continue;
          var targetRec = EntityHelper.GetRecord(targetEnt);
          // we are interested only in case when both records are inserted, or both are deleted.
          if (targetRec.Status != rec.Status)
            continue;
          // finally, the target record's table must be in the same SCC group - i.e. have the same SccIndex
          if (targetRec.EntityInfo.TopologicalIndex != rec.EntityInfo.TopologicalIndex)
            continue;
          // We have potential conflict; add vertexes and link for the conflict
          var thisV = graph.FindOrAdd(rec);
          var targetV = graph.FindOrAdd(targetRec);
          thisV.AddLink(targetV);
        }
      }//foreach cmd
      //Check if any conflicts found
      if (graph.Vertexes.Count == 0) return;
      //Build SCC graph
      graph.BuildScc();
      // Once SCC is built, we have SCC indexes in Vertexes; use them to assign Record's TopologicalIndex
      bool hasNonTrivialGroups = false;
      foreach (var v in graph.Vertexes) {
        ((EntityRecord)v.Tag).TopologicalIndex = v.SccIndex;
        hasNonTrivialGroups |= v.NonTrivialGroup;
      }
      //if there are non-trivial groups, it means we have circular references in the set.
      if (hasNonTrivialGroups) {
          var descr = string.Join(",", records.Select(r=> r.EntityInfo.Name + "/" + r.PrimaryKey));
          Util.Throw("Detected circular references between entities in an update set. Cannot commit group update. Entities: [{0}].", descr); 
      }
    }

    public static int CompareTopologicalIndexes(EntityRecord x, EntityRecord y) {
      var entIndexCompare = x.EntityInfo.TopologicalIndex.CompareTo(y.EntityInfo.TopologicalIndex);
      if (entIndexCompare != 0) return entIndexCompare;
      return x.TopologicalIndex.CompareTo(y.TopologicalIndex);
    }


  }// class


}
