﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Microsoft.CSharp;
using System.CodeDom.Compiler;

using Vita.Entities.Model;
using Vita.Common;
using System.Data;
using Vita.Entities;
using Vita.Data.Model;
using Vita.Data;
using Vita.Logging;

namespace Vita.SourceCodeGen {

  public class EntityCodeGenerator {
    public string FileHeader = @"
/* 
  This file was automatically generated by VITA reverse-engineering tool from the database.
  Generated date/time: {0}
*/
using System;
using System.Linq;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data; 

using Vita.Activation;
using Vita.Entities;
using Vita.Data;  // used only in console app
";
    public Vita.Logging.Log Log;
    EntityModel _model;
    EntityCodeGeneratorConfig _config;
    StringBuilder _output;
    IDictionary<string, AutoValueInfo> _autoValues;

    public EntityCodeGenerator(EntityModel model, EntityCodeGeneratorConfig config, Log log = null) {
      _model = model;
      _config = config;
      Log = log ?? new Logging.Log();
      //Build autovalues lookup
      _autoValues = new Dictionary<string, AutoValueInfo>(StringComparer.InvariantCultureIgnoreCase); 
      config.AutoValues.Each(av => av.Names.Each(n => _autoValues[n] = av));
    }

    public string GenerateSource() {
      //select entities by schemas
      // create output string builder and write source
      _output = new StringBuilder();
      WriteSource(); 
      return _output.ToString(); 
    }//method

    private void WriteSource() {
      var header = Util.SafeFormat(FileHeader, DateTime.Now.ToString("s"));
      _output.AppendLine(header);
      _output.AppendFormat("namespace {0} {{\r\n", _config.Namespace);
      foreach (var module in _model.Modules) {
        if (module.Entities.Count == 0)
          continue; 
        _output.AppendLine("// Module : " + module.Name + " --------------------------------");
        foreach (var entType in module.Entities) {
          var ent = _model.ModelInfo.GetEntityInfo(entType); 
          WriteEntity(ent);
        }
        _output.AppendLine(); 
      }
      WriteModules();
      WriteEntityModel();
      if (_config.Options.IsSet(CodeGenOptions.GenerateConsoleAppCode))
        WriteConsoleAppCode(); 
      _output.AppendLine("}"); //close namespace
      if (!Log.IsEmpty) {
        var messages = Log.GetAllAsText(); 
        _output.Append("/* Messages: \r\n" + messages + "\r\n*/");
      }
    }

    private void WriteEntity(EntityInfo entity) {
      _output.AppendLine();
      WriteEntityAttributes(entity); 
      // Entity header
      _output.AppendFormat("  public interface {0} {{\r\n", entity.EntityType.Name);
      //members
      foreach (var member in entity.Members) {
        if (member.Flags.IsSet(EntityMemberFlags.ForeignKey))
          continue;
        if (member.Kind == MemberKind.EntityList && !_config.Options.IsSet(CodeGenOptions.AddOneToManyLists))
          continue; 
        _output.AppendLine();
        WriteMemberAttributes(member);
        // we do not create setters for auto types and entity lists
        bool getterOnly = member.Kind == MemberKind.EntityList || member.AutoValueType != AutoType.None; 
        var accessors = getterOnly ? "{ get; }" : "{ get; set; }"; // for ent lists we have getter only
        _output.AppendFormat("    {0} {1} {2}\r\n", member.DataType.GetCsName(), member.MemberName, accessors);
      }
      _output.AppendLine("  }");
    }

    private void WriteEntityAttributes(EntityInfo entity) {
      // printout keys and indexes
      var attrList = new StringList();
      attrList.Add(Util.SafeFormat("Entity(TableName = \"{0}\")", entity.TableName));
      foreach (var key in entity.Keys) {
        if (key.KeyType.IsSet(KeyType.ForeignKey))
          continue;
        //Skip single-column indexes - we set [Index] attribute directly on property; also skip erroneous keys with no members (if table has no PK, VITA creates PK with no columns)
        if (IsPropertyKey(key) || key.Members.Count == 0) 
          continue;
        var attrName = GetKeyAttribute(key.KeyType);
        var keyCols = GetKeyColumnList(key); 
        attrList.Add(Util.SafeFormat("{0}(\"{1}\")", attrName, keyCols));
      }
      // If PK is not found in database, system creates an empty fake PK
      if (entity.PrimaryKey.Members.Count == 0)
        Log.Warning("  Warning: Table {0} has no primary key.", entity.TableName);
      if (entity.Flags.IsSet(EntityFlags.NoClusteredIndex)) {
        attrList.Add("NoClusteredIndex");
        Log.Warning("  Warning: Table {0} has no clustered index.", entity.TableName);
      }
      _output.AppendFormat("  [{0}]\r\n", string.Join(", ", attrList));
    }

    private string GetKeyColumnList(EntityKeyInfo key) {
      var specs = new string[key.Members.Count];
      for (int i = 0; i < key.Members.Count; i++) {
        var member = key.Members[i];
        var memberSpec = member.ColumnName ?? member.MemberName;
        if (key.IsDescending(member))
          memberSpec += ":DESC";
        specs[i] = memberSpec;
      }
      return string.Join(",", specs);
      
    }

    private void WriteMemberAttributes(EntityMemberInfo member) {
      var autoValue = GetAutoValue(member);
      if (autoValue != null) {
        member.Flags |= EntityMemberFlags.AutoValue;
        member.AutoValueType = autoValue.AutoType;
      }
      if (member.Flags.IsSet(EntityMemberFlags.PrimaryKey) && _config.Options.IsSet(CodeGenOptions.AutoOnGuidPrimaryKeys) && member.DataType == typeof(Guid))
        member.AutoValueType = AutoType.NewGuid;

      var dataType = member.DataType;
      var isMemo = member.Flags.IsSet(EntityMemberFlags.IsMemo); 
      var needsSize = dataType == typeof(string) || dataType == typeof(byte[]);
      var skipSize = isMemo || dataType == typeof(string) && _config.Options.IsSet(CodeGenOptions.SkipSizeOnDefaultStringLength) && member.Size == _config.DefaultStringLength;
      if (skipSize)
        needsSize = false; 
      var isNVT = dataType.IsNullableValueType();
      var needsNullable = member.Flags.IsSet(EntityMemberFlags.Nullable) && !isNVT;
      var attrList = new StringList();
      switch (member.Kind) {
        case MemberKind.Column:
          var colArgs = new StringList();
          colArgs.Add('"' + member.ColumnName + '"');
          if (member.ExplicitDbType != null) 
            colArgs.Add("DbType = " + CustomDbTypes.AsString(member.ExplicitDbType.Value));
          if (member.Scale != 0) 
            colArgs.Add("Scale = " + member.Scale);
          if (member.Precision != 0) 
            colArgs.Add("Precision = " + member.Precision); 
          if (needsSize) 
            colArgs.Add("Size = " + member.Size);
          var colAttr = Util.SafeFormat("Column({0})", string.Join(", ", colArgs)); 
          attrList.Add(colAttr);
          break; 
        case MemberKind.EntityRef:
          var fkMembers = member.ReferenceInfo.FromKey.ExpandedMembers;
          var strCols = string.Join(",", fkMembers.Select(m => m.ColumnName));
          attrList.Add(Util.SafeFormat("EntityRef(KeyColumns = \"{0}\")", strCols));
          if (member.Flags.IsSet(EntityMemberFlags.CascadeDelete))
            attrList.Add("CascadeDelete");
          break; 
        case MemberKind.EntityList:
          // check if we have two or more lists of the same type - then we need to add [ManyToOne] attr to set explicitly the property name of back reference
          var count = member.Entity.Members.Where(m => m.DataType == member.DataType).Count();
          if (count > 1)
            attrList.Add(Util.SafeFormat("OneToMany(\"{0}\")", member.ChildListInfo.ParentRefMember.MemberName));
          break; 
      }
      // if (needsSize)  attrList.Add(Util.SafeFormat("Size({0})", member.Size));
      if (isMemo)
        attrList.Add("Memo");
      // Keys and Indexes. If there is single-column index on the column, put it on property directly
      var memberKeys = member.Entity.Keys.Where(key => IsPropertyKey(key) && key.Members[0] == member).ToList();
      foreach (var key in memberKeys) {
        var keyAttr = GetKeyAttribute(key.KeyType); 
        attrList.Add(keyAttr);
      }
      //Misc attributes
      if (needsNullable)
        attrList.Add("Nullable");
      if ((dataType == typeof(DateTime) || dataType == typeof(DateTime?)) && _config.Options.IsSet(CodeGenOptions.UtcDates))
        attrList.Add("Utc");

      //Auto type
      switch (member.AutoValueType) {
        case AutoType.None: break;
        case AutoType.NewGuid: attrList.Add("Auto"); break;
        case AutoType.Identity: attrList.Add("Identity"); break;
        case AutoType.RowVersion: attrList.Add("RowVersion"); break;
        default:
          attrList.Add(Util.SafeFormat("Auto(AutoType.{0})", autoValue.AutoType));
          break;
      }
      // Produce final list as string
      if (attrList.Count > 0)
        _output.AppendFormat("    [{0}]\r\n", string.Join(", ", attrList));
    }

    private void WriteModules() {
      _output.AppendLine();
      _output.AppendLine("  // -------------------  Entity modules --------------------------------------");
      foreach (var module in _model.Modules) {
        if (module.Entities.Count == 0)
          continue;
        var entModuleName = module.Name;
        _output.AppendFormat("  public class {0} : EntityModule {{\r\n", entModuleName);
        _output.AppendFormat("    public {0} (EntityArea area) : base(area, \"{0}\") {{\r\n", entModuleName);
        _output.Append("      RegisterEntities(");
        //build list of types
        var entTypeList = module.Entities.Select(t => "typeof(" + t.Name + ")").ToList();
        string indent = new string(' ', 12);
        int currLen = 200; //to force start from new line
        // write in lines with len < 120
        for (var i = 0; i < entTypeList.Count; i++) {
          var entType = entTypeList[i];
          if (currLen > 120) {
            _output.AppendLine();
            _output.Append(indent);
            currLen = indent.Length;
          }
          _output.Append(entType);
          if (i < entTypeList.Count - 1)
            _output.Append(", ");
          currLen += entType.Length + 2;
        } //for i
        // close the RegisterEntities call
        _output.AppendLine(");");
        //close constructor and class
        _output.AppendLine("    }"); //method
        _output.AppendLine("  }"); //class
        _output.AppendLine();
      }//foreach area
    }

    private void WriteEntityModel() {
      _output.AppendLine("  // -------------------  Entity model --------------------------------------");
      _output.AppendFormat("  public class {0} : EntityModel {{\r\n", _config.ModelClassName);
      _output.AppendFormat("    public {0} () {{\r\n", _config.ModelClassName);

      //Write creation of areas and modules
      var moduleFieldDecls = new StringList();
      foreach (var module in _model.Modules) {
        if (module.Entities.Count == 0)
          continue; 
        var suffix = module.Area.SchemaName.FirstCap();
        var areaVar = "area" + suffix;
        var moduleField = "Module" + suffix;
        _output.AppendFormat("      var {0} = this.AddArea(\"{1}\", \"{2}\");\r\n", areaVar, module.Area.Name, module.Area.SchemaName);
        _output.AppendFormat("      {0} = new {1}({2});\r\n", moduleField, module.Name, areaVar);
        var fieldDecl = Util.SafeFormat("    public readonly {0} {1};", module.Name, moduleField);
        moduleFieldDecls.Add(fieldDecl);
      }
      //close constructor and class
      _output.AppendLine("    }"); //constructor
      _output.AppendLine();
      //write Module fields
      _output.AppendLine(string.Join("\r\n", moduleFieldDecls));
      //close class
      _output.AppendLine("  }"); //class
      _output.AppendLine();
    }

    private void WriteConsoleAppCode() {
      const string template = @"
  
  // Sample Program class for a Console application. Add references to Vita.Core, Vita.Data, Vita.Data.MsSql assemblies
  // using Vita.Entities;
  // using Vita.Data;

  class Program {
    public static {{ModelClassName}} EntityModel;
    public static IEntityStore EntityStore; 

    static void Main(string[] args) {
      Console.WriteLine("" Sample application for VITA-generated model. "");
      Init();
      
      //Open session and run query
      var session = EntityStore.OpenSession();
      var query = from ent in session.EntitySet<{{EntityType}}>()  // just random entity
                  // where ?condition?
                  select ent;
      var entities = query.Take(5).ToList();

      Console.WriteLine(""Loaded "" + entities.Count + "" entities."");
      foreach(var ent in entities)
        Console.WriteLine(""  Entity: "" + ent.ToString()); // change to smth more meaningful 

      Console.WriteLine(""Press any key ..."");
      Console.ReadKey();
    }
    
    private static void Init() {
      EntityModel = new {{ModelClassName}}();
      EntityModel.DefaultModuleSettings.DefaultStringLength = {{DefaultStringLength}};
      var connString = @""{{ConnectionString}}"";
      var driver = new {{DriverType}}();
      var dbOptions = DbOptions.None;       // The app will use direct SQL statements.
      // var dbOptions = DbOptions.Default; // Change to this to use stored procedures; 
      var updateSchema = false;                       //   also set 'updateSchema'to true so system can generate stored procedures

      // Cache settings. If you do not add any types, cacheSettings will be ignored
      var cacheSettings = new Vita.Data.Caching.CacheSettings(
         fullyCachedTypes: new Type[] { },   // add fully cached types here
         sparselyCachedTypes: new Type[] { } // add sparsely cached types here
      );
      var logFile = ""appLog.log"";
      EntityStore = EntityModel.CreateEntityStore(driver, connString, dbOptions, updateSchema, logFile, logFile, cacheSettings);
    }
  }
";
      var sampleEntType = _model.Modules.First(m => m.Entities.Count > 0).Entities.First();
      //string.Format() pukes on this template, so have to do it through replace
      var source = template.Replace("{{ModelClassName}}", _config.ModelClassName)
                           .Replace("{{DefaultStringLength}}", _config.DefaultStringLength.ToString())
                           .Replace("{{ConnectionString}}", _config.ConnectionString)
                           .Replace("{{EntityType}}", sampleEntType.Name)
                           .Replace("{{DriverType}}", _config.Driver.GetType().FullName);
      _output.AppendLine(source); 
    }

    private AutoValueInfo GetAutoValue(EntityMemberInfo member) {
      AutoValueInfo result;
      if (_autoValues.TryGetValue(member.MemberName, out result)) return result;
      if (_autoValues.TryGetValue(member.ColumnName, out result)) return result;
      return null; 
    }

    private bool IsPropertyKey(EntityKeyInfo key) {
      if (key.Members.Count != 1)  return false;
      // Only Index or PK - but not ForeignKey; FK is hidden key, without attribute
      bool isIndexOrPk = key.KeyType.IsSet(KeyType.Index | KeyType.PrimaryKey);
      if (!isIndexOrPk)
        return false; 
      //if it is a key on fk column that is NOT explicitly exposed as property
      if (key.Members.Any(m => m.Flags.IsSet(EntityMemberFlags.ForeignKey)))
        return false; 
      return true; 
    }

    private bool CanBeIdentity(Type type) {
      return type == typeof(int) || type == typeof(uint) || type == typeof(long) || type == typeof(ulong);
    }

    private string GetKeyAttribute(KeyType keyType) {
      if (keyType.IsSet(KeyType.PrimaryKey))
        return "PrimaryKey";
      if (keyType.IsSet(KeyType.Clustered)) 
        return keyType.IsSet(KeyType.Unique) ? "UniqueClusteredIndex" : "ClusteredIndex";
      if (keyType.IsSet(KeyType.Unique))
        return "Unique";
      if (keyType.IsSet(KeyType.Index))
        return "Index";
      Util.Throw("Invalid key type for attribute: " + keyType);
      return null; 
    }
  
  } //class

}//ns
