﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Xml;

using Vita.Common;
using Vita.Entities;
using Vita.Entities.Xml;
using Vita.Data;
using Vita.Data.Driver;

namespace Vita.SourceCodeGen {

  [Flags]
  public enum CodeGenOptions {
    SkipSizeOnDefaultStringLength = 1,
    AutoOnGuidPrimaryKeys = 1 << 1,
    ChangeEntityNamesToSingular = 1 << 2, // Change plural to singular: table 'Orders' -> entity 'IOrder'

    UtcDates = 1 << 4, // add Utc to all date times
    AddOneToManyLists = 1 << 5,
    GenerateConsoleAppCode = 1 << 6,

    Default = SkipSizeOnDefaultStringLength | AutoOnGuidPrimaryKeys | AddOneToManyLists | GenerateConsoleAppCode, 
  }

  public class AutoValueInfo {
    public AutoType AutoType;
    public bool IsHidden; //Currently not used, will be in the future
    public HashSet<string> Names = new HashSet<string>(StringComparer.InvariantCultureIgnoreCase);
    public AutoValueInfo(AutoType autoType, bool isHidden, params string[] names) {
      AutoType = autoType;
      IsHidden = isHidden;
      Names.UnionWith(names);
    }
    public override string ToString() {
      return string.Join(",", Names) + "/" + AutoType + (IsHidden ? "/Hidden" : string.Empty);
    }
  }

  public class EntityCodeGeneratorConfig {
    public DbDriver Driver;
    public string DriverCode; //used when loading from configuration
    public string ConnectionString;
    public string Namespace; //base namespace
    public string ModelClassName;
    public string OutputPath;
    public CodeGenOptions Options;
    public int DefaultStringLength;
    public List<string> Schemas = new List<string>();
    public List<AutoValueInfo> AutoValues = new List<AutoValueInfo>();

    private EntityCodeGeneratorConfig() {}

    public EntityCodeGeneratorConfig(DbDriver driver, string connectionString, 
                                     string ns, string modelClassName, string outputPath,
                                     CodeGenOptions options = CodeGenOptions.Default, int defaultStringLength = 50,
                                     IEnumerable<string> schemas = null,
                                     bool testCompile = true, bool compareSchemas = true) {
      Driver = driver;
      ConnectionString = connectionString; 
      Namespace = ns;
      ModelClassName = modelClassName;
      OutputPath = outputPath;
      Options = options;
      DefaultStringLength = defaultStringLength;
      if (schemas != null)
        Schemas.AddRange(schemas); 
    }

    public static EntityCodeGeneratorConfig FromXml(XmlDocument xmlDoc) {
      var cfg = new EntityCodeGeneratorConfig();
      cfg.DriverCode = xmlDoc.GetValue( "//DriverCode");
      cfg.ConnectionString = xmlDoc.GetValue( "//ConnectionString");
      cfg.Namespace = xmlDoc.GetValue( "//Namespace");
      cfg.ModelClassName = xmlDoc.GetValue( "//ModelClassName");
      cfg.OutputPath = xmlDoc.GetValue( "//OutputPath");
      var schemas = xmlDoc.GetValue( "//Schemas");
      if (!string.IsNullOrWhiteSpace(schemas))
        cfg.Schemas.AddRange(schemas.Split(','));
      //Options
      var strOptions = xmlDoc.GetValue( "//Options");
      if (string.IsNullOrWhiteSpace(strOptions))
        cfg.Options = CodeGenOptions.Default;
      else {
        try {
          cfg.Options = (CodeGenOptions) Enum.Parse(typeof(CodeGenOptions), strOptions, ignoreCase: true);
        } catch (Exception ex) {
          Util.Throw("Failed to parse Options value '{0}' : {1}", strOptions, ex.Message);
        }          
      }//else
      var strlen = xmlDoc.GetValue( "//DefaultStringLength");
      if (!string.IsNullOrWhiteSpace(strlen))
        cfg.DefaultStringLength = int.Parse(strlen);
      //AutoValues
      var autoValues = xmlDoc.SelectSingleNode("//AutoValues");
      if (autoValues != null) {
        foreach (XmlElement child in autoValues.ChildNodes) {
          var sType = child.GetAttribute("autoType");
          AutoType autoType; 
          Util.Check(Enum.TryParse(sType, out autoType), "Invalid autoType attribute value: {0}", sType);
          var sHidden = child.GetAttribute("hidden");
          var names = child.InnerText;
          cfg.AutoValues.Add(new AutoValueInfo(autoType, sHidden == "true", names));
        }
      }//if autoValues
      //done
      return cfg; 
    }

  }


}
