﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Reflection;
using System.Data;

using Vita.Common;
using Vita.Entities;
using Vita.Entities.Model;
using Vita.Entities.Model.Construction;
using Vita.Data;
using Vita.Data.Model;
using Vita.Data.Management;
using Vita.Logging;

namespace Vita.SourceCodeGen {
  /// <summary>Reconstructs the entity model from the raw DbModel loaded from the database. 
  /// The constructed entity model may be used for generating the entity interfaces as c# source file. </summary>
  public class EntityModelReverseGenerator {
    EntityCodeGeneratorConfig _config;
    EntityModel _model; 
    EntityModelInfo _modelInfo;
    DbModel _dbModel;
    Log _log; 

    static int _callCount; // artificial counter to generate unique namespaces for multiple calls
    string _namespace;

    public EntityModel Generate(EntityCodeGeneratorConfig config) {
      _config = config;
      _log = new Log();
      var dbStt = new DbSettings(config.Driver, DbOptions.Default, _config.ConnectionString, DbSchemaUpdateOptions.VerifyAll, schemas: config.Schemas);
      var modelLoader = new DbModelLoader();
      _dbModel = modelLoader.LoadModel(dbStt, _log);
      Util.Check(_dbModel.Tables.Count() > 0, "No tables found in the database. Code generation aborted.");
      // Prepare type generator
      _callCount++;
      _namespace = "_dummy_" + _callCount; // artificial unique namespace for dummy interface types
      // Construct model setup and model
      _model = new EntityModel();
      GenerateModulesAndAreas(); 
      _modelInfo = _model.ModelInfo = new EntityModelInfo(_model);
      //generate entities and members
      GenerateEntities();
      SetupPrimaryKeys();
      GenerateReferenceMembers();
      CreateIndexes();
      SetupKeyMembers();
      SetupDescendingKeyMembers();
      return _model; 
    }

    private void GenerateModulesAndAreas() {
      foreach (var sch in _dbModel.Schemas) {
        if (_config.Schemas.Count > 0 && !_config.Schemas.Contains(sch))
          continue; 
        var area = _model.AddArea(sch, sch);
        var module = new EntityModule(area, "EntityModule" + sch.FirstCap());
      }
    }


    private void GenerateEntities() {
      foreach (var table in _dbModel.Tables) {
        var module = GetModule(table.SchemaName);
        var entName = DbNameToCsName(table.TableName);
        if (_config.Options.IsSet(CodeGenOptions.ChangeEntityNamesToSingular))
          entName = Util.Unpluralize(entName); 
        entName = "I" + entName;
        var entType = CreateDummyEntityType(entName); //dummy type, just to have unique type instance
        module.Entities.Add(entType); // register type in module
        var ent = new EntityInfo(module, entType);
        ent.TableName = table.TableName;
        table.Entity = ent; 
        _modelInfo.RegisterEntity(ent); 
        // generate entity members
        foreach (var col in table.Columns) {
          var nullable = col.Flags.IsSet(DbColumnFlags.Nullable);
          var dataType = col.TypeInfo.ClrType;
          if (dataType.IsValueType && nullable)
            dataType = ReflectionHelper.GetNullable(dataType);
          var memberName = CheckMemberName(DbNameToCsName(col.ColumnName), ent);
          var member = col.Member = new EntityMemberInfo(ent, MemberKind.Column, memberName, dataType);
          member.ColumnName = col.ColumnName;
          // added to ent.Members automatically in constructor
          if (nullable)
            member.Flags |= EntityMemberFlags.Nullable; // in case it is not set (for strings)
          if (col.Flags.IsSet(DbColumnFlags.Identity))
            member.AutoValueType = AutoType.Identity;
          member.Size = col.TypeInfo.Size;
          member.Scale = col.TypeInfo.Scale;
          member.Precision = col.TypeInfo.Precision;
          //special Db types
          var dbType = col.TypeInfo.DbType; 
          if (CustomDbTypes.IsCustom(dbType))
            member.ExplicitDbType = dbType;
          else {
            switch (dbType) {
              case DbType.Currency:
              case DbType.StringFixedLength:
              case DbType.AnsiStringFixedLength:
              case DbType.AnsiString:
              case DbType.Date:
              case DbType.DateTime:
              case DbType.Time:
              case DbType.Binary:
              case DbType.Xml:
              case DbType.VarNumeric:
              case DbType.Object:
                member.ExplicitDbType = dbType;
                break;
              case CustomDbTypes.RowVersion:
                member.ExplicitDbType = CustomDbTypes.RowVersion;
                member.AutoValueType = AutoType.RowVersion;
                member.DataType = typeof(byte[]);
                member.Size = 8;
                member.Flags |= EntityMemberFlags.AutoValue;
                break;
            }//switch
          }//else
        }
      }//foreach table
    }//method

    private void SetupPrimaryKeys() {
      foreach (var table in _dbModel.Tables) {
        var ent = table.Entity;
        ent.PrimaryKey = new EntityKeyInfo("PK_" + ent.Name, KeyType.PrimaryKey, ent);
        if (table.PrimaryKey == null) {
          _log.Warning("Table {0} has no primary key.", table.TableName);
          table.PrimaryKey = new DbKeyInfo("PK_" + table.TableName + "_EMPTY", table, KeyType.PrimaryKey);
          continue; // next table
        }
        foreach (var col in table.PrimaryKey.Columns) {
          col.Member.Flags |= EntityMemberFlags.PrimaryKey;
          ent.PrimaryKey.ExpandedMembers.Add(col.Member);
        }
      }
    }

    private void CreateIndexes() {
      var fkAutoIndexed = !_dbModel.Driver.Supports(Data.Driver.DbCapabilities.NoIndexOnForeignKeys);
      foreach (var table in _dbModel.Tables) {
        bool hasClusteredIndex = false; 
        foreach (var key in table.Keys) {
          if (key.KeyType.IsSet(KeyType.Clustered))
            hasClusteredIndex = true;
          if (!key.KeyType.IsSet(KeyType.Index)) 
            continue;
          if (fkAutoIndexed && key.KeyType.IsSet(KeyType.ForeignKey)) {
            // NOTE: commented out - do not put explicit index on FK in entity model; it is added only on DbModel!
            //if (key.EntityKey != null)  key.EntityKey.KeyType |= KeyType.Index;
            continue;
          }
          var entKey = key.EntityKey = new EntityKeyInfo(key.Name, key.KeyType, table.Entity); //added automatically to entity.Keys
          foreach (var dbCol in key.Columns)
            entKey.ExpandedMembers.Add(dbCol.Member); 
        }
        if (!hasClusteredIndex && !table.PrimaryKey.KeyType.IsSet(KeyType.Clustered))
          table.Entity.Flags |= EntityFlags.NoClusteredIndex;
      }//foreach table
    }//method

    // When we initially created entity keys, we filled out key.ExplandedMembers list - these include real FK members (ex: IBook.Publisher_Id). 
    // We now need to build key.Members lists that include refMembers (ex: IBook.Publisher) instead of FK members in key.ExpandedMembers
    // c# entity definitions use entity reference members in keys and indexes, ex: [Index("Publisher"] on IBook interface.
    private void SetupKeyMembers() {
      foreach (var ent in _modelInfo.Entities) {
        foreach (var key in ent.Keys) {
          foreach (var member in key.ExpandedMembers) {
            if (member.Flags.IsSet(EntityMemberFlags.ForeignKey)) {
              // It is foreign key member; replace it with refMember 
              var refMember = member.ForeignKeyOwner;
              var allFkMembers = member.ForeignKeyOwner.ReferenceInfo.FromKey.ExpandedMembers;
              bool canUseRefMember = (key.ExpandedMembers.ContainsAll(allFkMembers));
              if (canUseRefMember) {
                // add it if it is not there yet. Note we might have multiple FK members for a single ref member (if FK is composite)
                if (!key.Members.Contains(refMember)) key.Members.Add(refMember);
              } else //add member itself
                key.Members.Add(member); 
            } else //if is FK
              // it is regular member, not FK. Add it to Members list
              key.Members.Add(member); 
            
          }//foreach member
        }//foreach key
      }//foreach ent            
    }//method

    // Copies Descending flags (if any are true) for DbKeys to EntityKeys
    private void SetupDescendingKeyMembers() {
      foreach (var table in _dbModel.Tables) 
        foreach(var key in table.Keys) 
          foreach (var descCol in key.DescendingColumns) {
            var member = descCol.Member;
            if (member.ForeignKeyOwner != null)
              member = member.ForeignKeyOwner;
            key.EntityKey.DescendingMembers.Add(member); 
          }
    }


    // Creates ref members and foreign keys
    private void GenerateReferenceMembers() {
      bool addListMembers = _config.Options.IsSet(CodeGenOptions.AddOneToManyLists);
      foreach (var table in _dbModel.Tables) {
        var ent = table.Entity; 
        foreach(var constr in table.RefConstraints) {
          //create ref member
          var targetEnt = constr.ToKey.Table.Entity;
          string memberName;
          if (constr.FromKey.Columns.Count == 1) {
            var fkColName = constr.FromKey.Columns[0].ColumnName;
            memberName = fkColName;
            // cut-off PK name
            var targetColName = constr.ToKey.Columns[0].ColumnName;
            if (memberName.EndsWith(targetColName, StringComparison.InvariantCultureIgnoreCase)) {
              memberName = memberName.Substring(0, memberName.Length - targetColName.Length);
              if (string.IsNullOrEmpty(memberName))
                memberName = targetEnt.Name;
            }
            if (memberName == fkColName) //that is a problem - we cannot have dup names; change to target entity name
              memberName = targetEnt.Name;
          } else 
            memberName = targetEnt.Name; //without starting "I"
          memberName = DbNameToCsName(memberName);
          memberName = CheckMemberName(memberName, ent, trySuffix: "Ref"); 
          var refMember = new EntityMemberInfo(ent, MemberKind.EntityRef, memberName, targetEnt.EntityType);
          if (constr.CascadeDelete)
            refMember.Flags |= EntityMemberFlags.CascadeDelete;
          // create foreign key
          var dbFk = constr.FromKey;
          var entFk = dbFk.EntityKey = new EntityKeyInfo(dbFk.Name, KeyType.ForeignKey, ent, refMember); // added to ent.Keys automatically
          foreach(var fkCol in dbFk.Columns) {
            entFk.ExpandedMembers.Add(fkCol.Member); 
            fkCol.Member.Flags |= EntityMemberFlags.ForeignKey;
            fkCol.Member.ForeignKeyOwner = refMember;
            if (fkCol.Member.Flags.IsSet(EntityMemberFlags.Nullable))
              refMember.Flags |= EntityMemberFlags.Nullable; 
          }
          refMember.ReferenceInfo = new EntityReferenceInfo(refMember, entFk, targetEnt.PrimaryKey);
          if (addListMembers) {
            //create List member on target entity
            var listName = Util.Pluralize(ent.EntityType.Name.Substring(1)); //remove 'I' prefix and pluralize
            listName = CheckMemberName(listName, targetEnt);
            var listType = typeof(IList<>).MakeGenericType(ent.EntityType);
            var listMember = new EntityMemberInfo(targetEnt, MemberKind.EntityList, listName, listType);
            var listInfo = new ChildEntityListInfo();
            listInfo.RelationType = EntityRelationType.ManyToOne;
            listInfo.TargetEntity = ent;
            listInfo.ParentRefMember = refMember;
            listMember.ChildListInfo = listInfo; 
          }
        }
      }//foreach table
    }

    private string CheckMemberName(string baseName, EntityInfo entity, string trySuffix = null) {
      var result = baseName; 
      if (entity.GetMember(result) == null)
        return result;
      // For entity references we might have occasional match of baseName with the FK column name. To avoid adding numbers, we try to add "Ref" at the end.
      if (!string.IsNullOrEmpty(trySuffix)) {
        result = baseName + trySuffix;
        if (entity.GetMember(result) == null)
          return result; 
      }
      // try adding number at the end.
      for (int i = 1; i < 10; i++) {
        result = baseName + i;
        if (entity.GetMember(result) == null)
          return result; 
      }
      Util.Throw("Failed to generate property name for entity {0}, base name {1}", entity.Name, baseName);
      return null; 
    }

    private Type CreateDummyEntityType(string entityTypeName) {
      var fullName = _namespace + "." + entityTypeName;
      var typeAttrs = TypeAttributes.Interface | TypeAttributes.Public | TypeAttributes.Abstract;
      var typeBuilder = EntityClassesAssembly.Instance.ModuleBuilder.DefineType(fullName, typeAttrs);
      return typeBuilder.CreateType();
    }

    //We have one-module per area, so we find module by associated area's schema 
    private EntityModule GetModule(string schema) {
      var module = _model.Modules.FirstOrDefault(m => m.Area.SchemaName.Equals(schema, StringComparison.InvariantCultureIgnoreCase));
      Util.Check(module != null, "EntityModule for schema {0} not found.", schema);
      return module; 
    }

    // removes spaces and underscores, changes to sentence case
    private static string DbNameToCsName(string name) {
      if (name == null) return null;
      if (name.IndexOf(' ') >= 0)
        name = name.Replace(" ", "_"); 
      if (name.IndexOf('_') == -1)
        return name.FirstCap();
      //Convert to first cap all segments separated by underscore
      var parts = name.Split('_', ' ');
      var result = String.Join(string.Empty, parts.Select(p => p.FirstCap()));
      return result;
    }



  }//class
}//ns
