﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Linq.Expressions;
using System.Diagnostics;

using Vita.Common;
using Vita.Entities.Runtime;
using Vita.Entities.Model;
using Vita.Entities.Linq;

namespace Vita.Data.MsSql {
  // Used for custom commands generation
  // Builds Query.DbExpression from original query expression. DbExpression will be passed to Linq2Sql engine to generate SQL statement. 
  // 1. Replace abstract entity sets with ITable instances.
  // 2. Replaces Parameter objects with default constant values
  // 3. Replaces locally evaluatable expressions with their values.
  // 4. Adds 'shaping union' to expression. This is workaround to prevent LINQ optimization that removes columns (autotype members) that are 
  //    assigned from parameters. The shaping union adds extra union to the query that prevents this from happening
  public class MsSqlCommandQueryRewriter : ExpressionVisitor {
    EntityQuery _query; 
    RuntimeEntityModel _model;
    IDataContext _dataContext; 

    Dictionary<Expression, object> _constantsLookup = new Dictionary<Expression,object>();

    public MsSqlCommandQueryRewriter(RuntimeEntityModel model, IDataContext dataContext) {
      _model = model;
      _dataContext = dataContext; 
    }

    public Expression Rewrite(EntityQuery query, bool addShapingUnion, Type targetEntityType) {
      _query = query; 
      //Build dict of expr->value for constant expressions, to replace during rewrite
      _constantsLookup.Clear();
      var evaluatables = query.QueryInfo.LocallyEvaluatable;
      var evalValues = query.QueryInfo.LocallyEvaluatedValues;
      for (int i = 0; i < evaluatables.Count; i++) {
        var evExpr = evaluatables[i];
        var constValue = evalValues[i + 1]; // +1 because arg#0 is reserved for data context
        _constantsLookup[evExpr] = constValue;
      }
      //Add shaping union if necessary
      var expr = query.Expression;
      if (addShapingUnion)
        expr = AddShapingUnion(expr, targetEntityType);
      //Rewrite the expression, replacing parameters with constants and abstract EntitySets with ITable references
      var result = Visit(expr);
      return result; 
    }

    public override Expression Visit(Expression node) {
      if (node == null)
        return node; 
      object value; 
      //Replace locally evaluatable expressions with their values
      if (_constantsLookup.TryGetValue(node, out value))
        return Expression.Constant(value); 
      return base.Visit(node); 
    }

    //Check for EntitySet
    protected override Expression VisitConstant(ConstantExpression node) {
      var entSet = node.Value as IEntitySet;
      if (entSet != null) {
        var table = _dataContext.GetTable(entSet.ElementType);
        return Expression.Constant(table);
      }
      return base.VisitConstant(node);
    }
    // Need to reconstruct Binary nodes to adjust binary op, since we replace arguments
    protected override Expression VisitBinary(BinaryExpression node) {
      var left = Visit(node.Left);
      var right = Visit(node.Right);
      var newNode = Expression.MakeBinary(node.NodeType, left, right);
      return newNode;
    }

    #region Adding shaping union
    /* 
     For update, delete, inserts based on a custom query returning sequence of autotypes 
     we need to do an extra adjustment. 
     Linq2Sql has one annoying optimization. If auto-type's property value 
     is assigned directly from a parameter or a constant (which is automatically converted into parameter)
     then Linq eliminates this member from the SQL query.  
     For example, let's say we create a custom update command to set price of all old books to $1 for quick sale.
     The command is based on the following query:
         from b in books 
             where b.PublishedOn < 2000/1/1
             select new (b.Id, Price = 1 }
     
     If we try to get SQL for this query, the Price column would not be in result set. Linq2Sql reasons that 
     it does not make sense to send "1" constant to database and add extra column to SQL results with this 
     fixed value. So instead, it assigns this value to auto-type property on the client, when it forms the results.
     This optimization is a trouble for us - the query is intended for use as subquery inside custom update, insert 
     statements, so we need it fully formed. 
     The only way I found to force Linq to create full resultset is to add an extra, shaping query
     with no rows returned, and Union it with original query. We form the shaping subquery from the fields of original 
     entity:
      
       (from b in books select new { Id = b.Id, Price = b.Price }).Take(0)
       .Union(
           -- original query --
           from b in books 
             where b.PublishedOn < 2000/1/1
             select new (b.Id, Price = 1 }
       )
*/

    private Expression AddShapingUnion(Expression expression, Type targetEntityType) {
      var outAnonType = expression.Type.GetGenericArguments()[0];
      //Some preparation work. Get Queryable methods
      var methods = typeof(Queryable).GetMethods();
      //Find Union overload with 2 parameters
      var unionMethodGen = methods.First(m => m.Name == "Union" && m.GetParameters().Length == 2);
      // There are 2 Queryable.Select overloads. Find the one with selector (second) parameter of type: 'Expression<Func<TSource, TResult>>'
      var selectMethodGen = methods.First(m => m.Name == "Select" && IsTwoArgFuncExpression(m.GetParameters()[1].ParameterType));
      var takeMethodGen = methods.First(m => m.Name == "Take");
      // Construct non-generic method references
      var takeMethod = takeMethodGen.MakeGenericMethod(outAnonType);
      var unionMethod = unionMethodGen.MakeGenericMethod(outAnonType);
      var selectMethod = selectMethodGen.MakeGenericMethod(targetEntityType, outAnonType);

      var anonTypeProps = outAnonType.GetProperties();
      var anonTypeConstrArgTypes = anonTypeProps.Select(p => p.PropertyType).ToArray();
      var anonTypeConstr = outAnonType.GetConstructor(anonTypeConstrArgTypes);
      var targetEntProps = targetEntityType.GetProperties();

      // Create shaping subquery. Start with lambda parameter
      var selectLambdaPrm = Expression.Parameter(targetEntityType, "ent");
      //Build expressions for New (constructor) expression
      var constrArgs = new Expression[anonTypeConstrArgTypes.Length];
      for (int i = 0; i < constrArgs.Length; i++) {
        var propName = anonTypeProps[i].Name;
        var getter = targetEntProps.First(p => p.Name == propName).GetGetMethod();
        constrArgs[i] = Expression.Property(selectLambdaPrm, getter);
      }
      var callNewAnonType = Expression.New(anonTypeConstr, constrArgs);
      //Create lambda for select argument
      var selectLambda = Expression.Lambda(callNewAnonType, selectLambdaPrm);
      //Create entity set and empty select query on it
      var targetEntitySet = CreateEntitySet(targetEntityType);
      var targetEntitySetConst = Expression.Constant(targetEntitySet);
      var selectQuery = Expression.Call(null, selectMethod, targetEntitySetConst, selectLambda);
      // Append .Take(0)
      var selectQueryWithTake0 = Expression.Call(takeMethod, selectQuery, Expression.Constant(0));
      //Create union of empty shaping select and original expresssion
      var union = Expression.Call(null, unionMethod, selectQueryWithTake0, expression);
      return union;
    }

    //Returns true if type is: Expression<Func<TSource, TResult>>
    private bool IsTwoArgFuncExpression(Type type) {
      if (!type.IsGenericType) return false;
      var funcType = type.GetGenericArguments()[0];
      var funcTypeArgCount = funcType.GetGenericArguments().Length;
      return funcTypeArgCount == 2;
    }

    private IQueryable CreateEntitySet(Type entityType) {
      var setType = typeof(EntitySet<>).MakeGenericType(entityType);
      var entSet = Activator.CreateInstance(setType) as IQueryable;
      return entSet;
    }
    #endregion

  }//class

}
