﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data;
using System.Data.Common;
using System.Linq.Expressions;

using Vita.Common;
using Vita.Entities;
using Vita.Entities.Model;
using Vita.Entities.Runtime;
using Vita.Logging;
using Vita.Data.Model;
using System.Data.Linq;

namespace Vita.Data.MsSql {

  // Construction of custom commands based on LINQ queries
  public partial class MsSqlCustomCommandCompiler {
    protected DbModel DbModel;
    protected Log Log;
    protected DbTypeMapper TypeMapper;
    protected IDataContext DataContext;

    // Error Messages
    const string _errorLinqQueryParamCountMismatch =
      "Command compiler, command {0} - parameter count mismatch with DB command/query." +
      " Query param count:  {1}, compiled Db command count: {2}. Use 'Param<T>()' methods to explicitly add LINQ query parameters." +
      " Note: might be internal VITA bug or constraint - try to reformulate the query.";
    const string _errorMemberNotFound = "Member {0} is not found in target entity {1}; custom query/proc: {2}.";
    const string _primaryKeyMissingMessage =
      "Error compiling command {0}: all primary key columns must be included in the output of the query. ";
    const string _errorCompoundKeysNotSupportedInInserts = "Compound foreign keys not supported in custom INSERT statements; proc/query: {0}";
    const string _shapingUnionComment = "  -- NOTE: a union with artificial empty query is added - it is a workaround for some unwanted LINQ optimization.\r\n";

    public MsSqlCustomCommandCompiler(DbModel dbModel, IDataContext dataContext) { 
      DbModel = dbModel;
      DataContext = dataContext;
      TypeMapper = DbModel.Driver.CreateTypeMapper(DbModel.Settings); 
      Log = DbModel.EntityModel.Log;
    }

    public virtual DbCommandInfo CompileCommand(EntityCommand command) { 
      switch (command.Kind) {
        case EntityCommandKind.CustomSelect:
          return CompileCustomSelect(command);
        case EntityCommandKind.CustomInsert:
          return CompileCustomInsert(command);
        case EntityCommandKind.CustomUpdate:
          return CompileCustomUpdate(command);
        case EntityCommandKind.CustomDelete:
          return CompileCustomDelete(command);
        default:
          Util.Throw ("DbCustomCommandCompiler: Invalid command kind {0}, command {1}.", command.Kind, command.CommandName);
          return null; 
      }
    }

    public virtual DbCommandInfo CompileCustomSelect(EntityCommand command) {
      var baseQueryDbCommand = CompileBaseQuery(command);
      var table = DbModel.LookupDbObject<DbTableInfo>(command.TargetEntityInfo);
      var cmdInfo = CreateCustomDbCommandInfo(command, command.CommandName, table, DbExecutionType.Query, baseQueryDbCommand.CommandText);
      //for selects we build column mappings first time we execute the query
      return cmdInfo;
    }

    // Update command 
    /* Sample query:
    update prod.Product
      set 
        Quantity = V.Quantity
    from prod.Product T INNER JOIN 
     (select Id, quantity + 1 as Quantity
          from prod.Product) V
     on V.Id = T.Id;
    */
    const string _updateTemplate = @"
UPDATE {0}
  SET 
    {1}
FROM {0} T INNER JOIN (
  {2} 
  ) V 
  ON {3};
";

    public virtual DbCommandInfo CompileCustomUpdate(EntityCommand command) {
      var baseQueryDbCommand = CompileBaseQuery(command, true);
      var entInfo = command.TargetEntityInfo;
      var table = DbModel.LookupDbObject<DbTableInfo>(entInfo);
      var queryOutputType = command.CustomBaseQuery.ElementType;
      var outputProps = queryOutputType.GetProperties(); //it might be auto-object, with fewer propertis
      // build SET and JOIN clauses
      var setParts = new StringList();
      var joinParts = new StringList();
      foreach (var prop in outputProps) {
        var member = entInfo.GetMember(prop.Name);
        if (member == null) {
          Log.Error(_errorMemberNotFound, prop.Name, entInfo.EntityType, command);
          continue;
        }
        var col = DbModel.LookupDbObject<DbColumnInfo>(member);
        if (col == null) continue;
        var colName = col.ColumnName;
        if (member.Flags.IsSet(EntityMemberFlags.PrimaryKey))
          joinParts.Add(string.Format("V.[{0}] = T.[{0}]", colName));
        else
          setParts.Add(string.Format("[{0}] = V.[{0}]", col.ColumnName));
      }
      //Check that all PK columns are present among properties
      if (joinParts.Count != entInfo.PrimaryKey.ExpandedMembers.Count) {
        Log.Error(_primaryKeyMissingMessage, command);
        return null;
      }
      var joinClause = string.Join(" AND ", joinParts);
      var setClause = string.Join(", ", setParts);
      //Build SQL and construct commandINfo
      var sql = string.Format(_updateTemplate, table.FullName, setClause, baseQueryDbCommand.CommandText, joinClause);
      var cmdInfo = CreateCustomDbCommandInfo(command, command.CommandName, table, DbExecutionType.NonQuery, sql);
      return cmdInfo;
    }

    /* DELETE example from MSDN:
    DELETE FROM Sales.SalesPersonQuotaHistory 
      FROM Sales.SalesPersonQuotaHistory AS spqh
           INNER JOIN Sales.SalesPerson AS sp
           ON spqh.BusinessEntityID = sp.BusinessEntityID
           WHERE sp.SalesYTD > 2500000.00;
    */

    const string _deleteTemplate = @"
DELETE FROM {0}
FROM {0} T INNER JOIN (
   {1}
   ) V ON {2};
";

    public virtual DbCommandInfo CompileCustomDelete(EntityCommand command) {
      var baseQueryDbCommand = CompileBaseQuery(command);
      var entInfo = command.TargetEntityInfo;
      var table = DbModel.LookupDbObject<DbTableInfo>(entInfo);
      var queryOutputType = command.CustomBaseQuery.ElementType;
      var outputProps = queryOutputType.GetProperties(); //return type is not command.Entity, it is just IDs
      // build JOIN ON condition clause
      var joinParts = new StringList();
      foreach (var prop in outputProps) {
        var member = entInfo.GetMember(prop.Name);
        if (member == null) {
          Log.Error(_errorMemberNotFound, prop.Name, entInfo.EntityType, command);
          continue;
        }
        var col = DbModel.LookupDbObject<DbColumnInfo>(member);
        if (col == null) continue;
        if (member.Flags.IsSet(EntityMemberFlags.PrimaryKey))
          joinParts.Add(string.Format("V.[{0}] = T.[{0}]", col.ColumnName));
      }
      //Check that all PK columns are present among properties
      if (joinParts.Count != entInfo.PrimaryKey.ExpandedMembers.Count)
        Log.Error(_primaryKeyMissingMessage, command);
      var joinClause = string.Join(" AND ", joinParts);
      //Build SQL and construct commandINfo
      var sql = string.Format(_deleteTemplate, table.FullName, baseQueryDbCommand.CommandText, joinClause);
      var cmdInfo = CreateCustomDbCommandInfo(command, command.CommandName, table, DbExecutionType.NonQuery, sql);
      return cmdInfo;
    }

    //   ---------------- INSERT statement ----------------------------------------------------------------------
    const string _insertTemplate = @"
INSERT INTO {0} ({1})
SELECT {1} FROM (
     {2}
   ); 
";

    public virtual DbCommandInfo CompileCustomInsert(EntityCommand command) {
      var baseQueryDbCommand = CompileBaseQuery(command, true);
      var entInfo = command.TargetEntityInfo;
      var table = DbModel.LookupDbObject<DbTableInfo>(entInfo);
      //Get column names list
      var queryOutputType = command.CustomBaseQuery.ElementType;
      var outputProps = queryOutputType.GetProperties();
      var colNames = new StringList();
      foreach (var prop in outputProps) {
        var member = entInfo.GetMember(prop.Name);
        if (member == null) {
          Log.Error(_errorMemberNotFound, prop.Name, entInfo.EntityType, command);
          continue;
        }
        var col = DbModel.LookupDbObject<DbColumnInfo>(member);
        if (col == null && member.Kind == MemberKind.EntityRef) {
          var fkMembers = member.ReferenceInfo.FromKey.ExpandedMembers;
          if (fkMembers.Count > 1)
            Util.Throw(_errorCompoundKeysNotSupportedInInserts, command);
          col = DbModel.LookupDbObject<DbColumnInfo>(fkMembers[0]);
        }
        colNames.Add(col.ColumnName);
      }
      var strColumns = string.Join(", ", colNames);
      //Build SQL and construct commandInfo
      var sql = string.Format(_insertTemplate, table.FullName, baseQueryDbCommand.CommandText, strColumns);
      var cmdInfo = CreateCustomDbCommandInfo(command, command.CommandName, table, DbExecutionType.NonQuery, sql);
      return cmdInfo;
    }

    protected DbCommand CompileBaseQuery(EntityCommand command, bool addShapingUnion = false) {
      //Rewrite the expression
      var rewriter = new MsSqlCommandQueryRewriter(DbModel.EntityModel.RuntimeModel, this.DataContext);
      var dbExpr = rewriter.Rewrite(command.CustomBaseQuery, addShapingUnion, command.TargetEntityType);
      var realQuery = this.DataContext.Provider.CreateQuery(dbExpr);
      //Compile query into SQL    
      var dtx = (DataContext) this.DataContext.RealDataContext;
      var dbCommand = dtx.GetCommand(realQuery);
      //Check we got Linq parameters right
      if (command.ExtendedParameters.Count != dbCommand.Parameters.Count) {
        Log.Error(_errorLinqQueryParamCountMismatch, command.CommandName, command.ExtendedParameters.Count, dbCommand.Parameters.Count);
        return dbCommand;
      }
      //Add comment about shaping union
      if (addShapingUnion)
        dbCommand.CommandText = _shapingUnionComment + dbCommand.CommandText;
      return dbCommand;
    }


    public virtual DbCommandInfo CreateCustomDbCommandInfo(EntityCommand entityCommand, string name, DbTableInfo mainTable,
                                                               DbExecutionType executionType, string sql) {
      var prmPrefix = GetParameterPrefix();
      var cmdInfo = new DbCommandInfo(entityCommand, name, mainTable, executionType, sql);
      //Create parameters from entity command parameters 
      foreach (var linqParam in entityCommand.ExtendedParameters) {
        var paramName = prmPrefix + linqParam.Name;
        var typeInfo = this.TypeMapper.GetDbTypeInfo(linqParam.DataType, size: linqParam.Size, explicitDbType: linqParam.ExplicitDbType, precision: linqParam.Precision, scale: linqParam.Scale);
        var argIndex = entityCommand.Parameters.IndexOf(linqParam); //might be -1 if it is not command parameter
        cmdInfo.AddParameter(linqParam, paramName, typeInfo, argIndex);
      }//foreach entParam
      return cmdInfo;
    }

    private string GetParameterPrefix() {
      var settings = this.DbModel.Settings;
      var driver = settings.Driver;
      var prefix = settings.Options.IsSet(DbOptions.UseStoredProcs) ? driver.StoredProcParameterPrefix : driver.DynamicSqlParameterPrefix;
      return prefix;
    }
    

  }//class
}
