﻿using System;
using System.Data;
using System.Data.Linq.Mapping;
using System.Data.SqlClient;
using System.Diagnostics;

using Vita.Common;
using Vita.Entities;
using Vita.Entities.Model;
using Vita.Data.Model;
using Vita.Data.Driver;

namespace Vita.Data.MsSql {

  public enum MsSqlVersion {
    V2008,
    V2012
  }

  public class MsSqlDriver : DbDriver {
    public const DbOptions DefaultMsSqlDbOptions = DbOptions.Default | DbOptions.AutoIndexForeignKeys;
    public const DbCapabilities DefaultCapabilities =
      DbCapabilities.Schemas | DbCapabilities.StoredProcedures | DbCapabilities.OutputParameters | DbCapabilities.DefaultParameterValues
        | DbCapabilities.ReferentialConstraints | DbCapabilities.ClusteredIndexes
        | DbCapabilities.Paging | DbCapabilities.BatchedUpdates | DbCapabilities.OrderedColumnsInIndexes | DbCapabilities.NoIndexOnForeignKeys
        | DbCapabilities.CustomAll;

    public readonly MsSqlVersion ServerVersion;

    public MsSqlDriver() : this(MsSqlVersion.V2008) {}

    public MsSqlDriver(MsSqlVersion serverVersion) {
      ServerVersion = serverVersion; 
      base.Capabilities = DefaultCapabilities;
      base.DynamicSqlParameterPrefix = base.StoredProcParameterPrefix = "@";
      base.CommandCallLogFormat = "EXEC {0} {1};";
    }

    // overrides
    public override IDbConnection CreateConnection(string connectionString) {
      return new SqlConnection(connectionString);
    }

    public override IDbCommand CreateDbCommand() {
      return new SqlCommand();
    }
    public override DbSchemaManager CreateSchemaManager(string connectionString) {
      return new MsSqlSchemaManager(this, connectionString);
    }

    public override DbSqlBuilder CreateSqlBuilder(DbModel dbModel) {
      return new MsSqlBuilder(dbModel); 
    }

    public override DbTypeMapper CreateTypeMapper(DbSettings settings) {
      return new MsSqlTypeMapper(settings); 
    }
    public override Vita.Data.Driver.LinqSqlProvider GetLinqSqlProvider() {
      return new MsSqlLinqSqlProvider();
    }

    public override void CheckDatabaseException(Exception exception, IDbCommand command = null) {
      var sqlEx = exception as SqlException;
      if (sqlEx == null) return;
      switch (sqlEx.Number) {
        case 2601: //unique index violation
          //sqlEx.Data contains 6 values which are mostly useless (info about provider)
          // we need to parse index name - stupidly the ex does not provide it as a separate value. 
          var indexName = ExtractIndexNameFromDuplicateErrorMessage(sqlEx.Message); 
          throw new UniqueIndexViolationException(sqlEx.Message, indexName, sqlEx); 
        case 1205: //Transaction lock, process killed
          //TODO: implement command repeat on transaction lock
          break; 
        case 547: // FK constraint violation on delete
          //TODO: implement FK violation on delete - throw some special exception
          break; 
        case 50000: //this number is for messages that are NOT registered in SQL server Messages table
          //custom error raised by CRUD stored proc. Message contains 4 segments: 'Code/op/table/pk', for ex: 'VITA:RowVersionConflict/Update/Employee/123'
          var msg = exception.Message;
          if (msg.StartsWith(ExceptionStrings.CodeConcurrentUpdateConflict)) {
            var segments = msg.Split('/');
            var op = segments[1];
            var tableName = segments[2];
            var pk = segments[3];
            throw new ConcurrentUpdateException(ExceptionStrings.MessageConcurrentUpdateConflict, tableName, op, pk, exception);
      }//if

          break; 
      }
      base.CheckDatabaseException(exception, command); 
    }

    private string ExtractIndexNameFromDuplicateErrorMessage(string message) {
      try {
        // Sample error message (it is two-line string): 
        // Cannot insert duplicate key row in object 'books.Publisher' with unique index 'IXU_Publisher_Name'.
        // The statement has been terminated.
        var uniqueIndexTag = "unique index";
        var uiPos = message.IndexOf(uniqueIndexTag);
        var q1Pos = message.IndexOf("'", uiPos);
        var q2Pos = message.IndexOf("'", q1Pos + 1);
        var indexName = message.Substring(q1Pos + 1, q2Pos - q1Pos - 1);
        return indexName;
      } catch (Exception ex) {
        return "(failed to identify index: " + ex.Message + ")"; //
      }
    }

    public override string GetFullName(string schema, string name) {
      return string.Format("[{0}].[{1}]", schema, name);
    }

    protected override IDbDataParameter CreateDbParameter(IDbCommand command, DbParamInfo prmInfo) {
      var typeInfo = prmInfo.TypeInfo;
      SqlParameter prm = null;
      switch (typeInfo.DbType) {
        case DbType.Currency:
          prm = new SqlParameter(prmInfo.Name, SqlDbType.Money);
          break;
        case CustomDbTypes.VarBinary:
          prm = new SqlParameter(prmInfo.Name, SqlDbType.VarBinary, typeInfo.Size);
          break;
        case CustomDbTypes.RowVersion:
          prm = new SqlParameter(prmInfo.Name, SqlDbType.Timestamp, 8);
          break; 
        case CustomDbTypes.NText: 
          prm = new SqlParameter(prmInfo.Name, SqlDbType.NText, -1);
          break;
        case CustomDbTypes.Text:
          prm = new SqlParameter(prmInfo.Name, SqlDbType.Text, -1);
          break;
        case CustomDbTypes.Image:
          prm = new SqlParameter(prmInfo.Name, SqlDbType.Image, -1);
          break;
        case DbType.Object:
          prm = new SqlParameter(prmInfo.Name, SqlDbType.Variant);
          break;
        case CustomDbTypes.HierarchyId:
          prm = new SqlParameter(prmInfo.Name, SqlDbType.VarBinary, 892);
          break;
        case CustomDbTypes.Geography:
          prm = new SqlParameter(prmInfo.Name, SqlDbType.Udt, -1);
          prm.UdtTypeName = "geography";
          break;
        case CustomDbTypes.Geometry:
          prm = new SqlParameter(prmInfo.Name, SqlDbType.Udt, -1);
          prm.UdtTypeName = "geometry";
          break;
        case CustomDbTypes.SmallMoney:
          prm = new SqlParameter(prmInfo.Name, SqlDbType.SmallMoney);
          break;
        case CustomDbTypes.SmallDateTime:
          prm = new SqlParameter(prmInfo.Name, SqlDbType.SmallDateTime);
          break;
        default:
          return base.CreateDbParameter(command, prmInfo);
      }
      prm.Direction = prmInfo.Direction;
      if (prmInfo.SourceColumn != null)
        prm.SourceColumn = prmInfo.SourceColumn.ColumnName;
      prm.Direction = prmInfo.Direction;
      if (typeInfo.Precision > 0) prm.Precision = typeInfo.Precision;
      if (typeInfo.Scale > 0) prm.Scale = typeInfo.Scale;
      if (typeInfo.Size > 0) prm.Size = typeInfo.Size;
      prm.Value = prmInfo.DefaultValue;
      return prm;
    }

    protected override IDbDataParameter CloneDbParameter(IDbCommand command, IDbDataParameter parameter) {
      var sqlPrm = (SqlParameter) parameter;
      var result = new SqlParameter(parameter.ParameterName, sqlPrm.SqlDbType, sqlPrm.Size, sqlPrm.Direction, 
                    sqlPrm.IsNullable, sqlPrm.Precision, sqlPrm.Scale, sqlPrm.SourceColumn, sqlPrm.SourceVersion, sqlPrm.Value);
      return result; 
    }

    public override bool ConvertToStoredProc(DbCommandInfo command) {
      var table = command.Table;
      string ProcBodyTemplate =
@"BEGIN
  SET NOCOUNT ON;
  {0}
END
";
      string CreateProcTemplate = @"{0}
CREATE PROCEDURE {1} 
{2} 
  AS 
{3}
";
      command.Schema = table.SchemaName;
      //Build command that creates stored proc and execute it
      var listParams = new StringList();
      foreach (var prm in command.Parameters) {
        var strNull = prm.IsNullable ? " = NULL" : string.Empty; 
        var strOut = (prm.Direction & ParameterDirection.Output) != 0 ? " OUTPUT" : string.Empty;
        listParams.Add("    " + prm.Name + " " + prm.TypeInfo.SqlTypeSpec + strNull + strOut);
      }
      var strParams = string.Join(",\r\n", listParams);
      var header = BuildProcHeader(command);
      command.StoredProcBody = string.Format(ProcBodyTemplate, command.Sql);
      command.StoredProcText =
        string.Format(CreateProcTemplate, header, command.FullCommandName, strParams, command.StoredProcBody);
      return true;
    }

    private string BuildProcHeader(DbCommandInfo command) {
      var entCommand = command.EntityCommand;
      //table/category/operation; for ex: Product/CRUD/Update
      const string ProcHeaderLineTemplate = @"
{0}={1}
-- Description: {2}
";
      return string.Format(ProcHeaderLineTemplate, DbDriver.GeneratedProcedureTag, command.EntityCommand.DescriptiveTag, entCommand.Description);
    }


  }//class
}
