﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data;

using Vita.Common;
using Vita.Data;
using Vita.Data.Model;
using Vita.Data.Driver;

namespace Vita.Data.MsSql {

  public partial class MsSqlSchemaManager : DbSchemaManager {
    public MsSqlSchemaManager(DbDriver driver, string connectionString) : base(driver, connectionString) { }

    //Columns:CATALOG_NAME	SCHEMA_NAME	SCHEMA_OWNER	DEFAULT_CHARACTER_SET_CATALOG	DEFAULT_CHARACTER_SET_SCHEMA 
    //	DEFAULT_CHARACTER_SET_NAME  
    public override DataTable GetSchemas() {
      //simple way to find non-system schemas - in MS SQL Server
      //var sql = "SELECT * FROM INFORMATION_SCHEMA.SCHEMATA WHERE  SCHEMA_NAME <> SCHEMA_OWNER;";
      var sql = "SELECT * FROM INFORMATION_SCHEMA.SCHEMATA;";
      return ExecuteSelect(sql);
    }

    //Columns: TABLE_CATALOG	TABLE_SCHEMA	TABLE_NAME	TABLE_TYPE  (table_type: "BASE TABLE") 
    public override DataTable GetTables(string schema = null) {
      var sql = "SELECT * FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_TYPE='BASE TABLE' ";
      if (!string.IsNullOrEmpty(schema))
        sql += string.Format(" AND TABLE_SCHEMA='{0}'", schema);
      return ExecuteSelect(sql);
    }

    //SQL Server INFORMATION_SCHEMA.ROUTINES view defines ROUTINE_DEFINITION column (proc body) as nvarchar(4000), 
    // so it is truncated for long SPs. We use a workaround to get full text
    public override DataTable GetProcedures(string schema = null) {
      const string getRoutinesTemplate = @"
SELECT R.ROUTINE_SCHEMA, R.ROUTINE_NAME, 
       OBJECT_DEFINITION(OBJECT_ID(R.ROUTINE_SCHEMA + '.' + R.Routine_Name)) as ROUTINE_DEFINITION,
       '' AS ROUTINE_BODY,
       r.ROUTINE_DEFINITION as RoutineDefTruncated
FROM INFORMATION_SCHEMA.ROUTINES R
WHERE R.ROUTINE_TYPE='PROCEDURE'
";
      var sql = getRoutinesTemplate;
      if (!string.IsNullOrEmpty(schema))
        sql += string.Format(" AND ROUTINE_SCHEMA='{0}'", schema);
      return ExecuteSelect(sql);
    }

    //INFORMATION_SCHEMA does not have a view for indexes, so we have to do it through MSSQL special objects
    public override DataTable GetIndexes(string schema = null) {
      //Slightly adjusting column names to match SQL CE Indexes view
      const string getIndexesTemplate = @"
SELECT ss.name AS table_schema, st.name AS table_name, si.name AS index_name, 
       si.is_primary_key as [primary_key], si.type_desc as [clustered], 
       si.is_unique as [unique], si.is_disabled as [disabled]
FROM   sys.indexes AS si INNER JOIN
       sys.tables AS st ON si.object_id = st.object_id INNER JOIN
       sys.schemas AS ss ON ss.schema_id = st.schema_id
WHERE (si.name IS NOT NULL) AND {0}
ORDER BY table_schema, table_name, index_name
";
      const string schemaCondTemplate = "(ss.[name] = '{0}')";
      var strCond = string.IsNullOrWhiteSpace(schema) ? "(1=1)" : string.Format(schemaCondTemplate, schema);
      var sql = string.Format(getIndexesTemplate, strCond); 
      return ExecuteSelect(sql);
    }

    public override DataTable GetIndexColumns(string schema = null) {
      const string getIndexColumnsTemplate = @"
SELECT ss.name AS table_schema, st.name AS table_name, si.name AS index_name, 
       sc.name AS column_name, sic.key_ordinal as column_ordinal_position, sic.is_descending_key as is_descending
FROM   sys.indexes AS si INNER JOIN
       sys.index_columns AS sic ON si.object_id = sic.object_id AND si.index_id = sic.index_id INNER JOIN
       sys.columns AS sc ON si.object_id = sc.object_id AND sc.column_id = sic.column_id INNER JOIN
       sys.tables AS st ON si.object_id = st.object_id INNER JOIN
       sys.schemas AS ss ON ss.schema_id = st.schema_id
{0}
ORDER BY table_schema, table_name, index_name, column_ordinal_position;
";
      const string whereClauseTemplate = "WHERE ss.[name] = '{0}' ";
      var strWhere = (schema == null) ? string.Empty : string.Format(whereClauseTemplate, schema);
      var sql = string.Format(getIndexColumnsTemplate, strWhere);
      return ExecuteSelect(sql);
    }

    //For MS SQL server, if a column has a default expression, this default is stored as default constraint, which must be dropped
    // before we drop the column. To drop the constraint, we need to know its name. 

    public override void DropColumn(DbColumnInfo column) {
      if (!string.IsNullOrEmpty(column.DefaultExpression))
        DropDefaultConstraint(column);
      base.DropColumn(column);
    }

    private void DropDefaultConstraint(DbColumnInfo column) {
      const string sqlFindConstraintName = @"
WITH Defaults AS (
  SELECT D.Name as ConstraintName,
       SCHEMA_NAME(D.schema_id) as TableSchema, 
       OBJECT_NAME(D.parent_object_id) AS TableName, 
       C.Name as ColumnName, 
       D.Definition as DefaultExpression
FROM sys.default_constraints D INNER JOIN sys.columns C 
  ON D.parent_column_id = C.column_id AND D.parent_object_id = c.object_id
)
SELECT * From Defaults   
WHERE TableSchema = '{0}' AND TableName = '{1}' AND ColumnName = '{2}';
";
      const string sqlDeleteConstraint = "ALTER TABLE {0}.{1} DROP CONSTRAINT {2};"; 
      var sql = string.Format(sqlFindConstraintName, column.Table.SchemaName, column.Table.TableName, column.ColumnName);
      var table = base.ExecuteSelect(sql);
      var row = table.Rows[0];
      var  cname = row.GetAsString("ConstraintName");
      //delete it
      sql = string.Format(sqlDeleteConstraint, column.Table.SchemaName, column.Table.TableName, cname);
      base.ExecuteNonQuery(sql);
    }//method

    //exec sys.sp_rename 'employees' , 'people'
    public override void RenameTable(DbTableInfo oldTable, DbTableInfo newTable) {
      ExecuteNonQuery("EXEC SYS.SP_RENAME '{0}.{1}' , '{2}'", oldTable.SchemaName, oldTable.TableName, newTable.TableName);
    }

    //MSSQL:   exec sys.sp_rename 'employees.employee_number' , 'employee_num', 'COLUMN'
    public override void RenameColumn(DbColumnInfo oldColumn, DbColumnInfo newColumn) {
      ExecuteNonQuery("EXEC SYS.SP_RENAME '{0}.{1}.{2}' , '{3}', 'COLUMN'",
          oldColumn.Table.SchemaName, oldColumn.Table.TableName, oldColumn.ColumnName, newColumn.ColumnName);
    }


    public override void DropSchemaObjects(string schema) {
      Util.Check(!string.IsNullOrEmpty(schema), "Schema parameter may not be empty. (MsSqlDriver.DropSchemaObjects)"); 
      //Delete procedures
      var procSchemaFilter = "ROUTINE_SCHEMA='" + schema + "'";
      var procs = this.GetProcedures(schema);
      foreach (DataRow row in procs.Rows) {
        var procName = row.GetAsString("ROUTINE_NAME");
        ExecuteNonQuery("DROP PROCEDURE [{0}].[{1}];", schema, procName);
      }
      //Delete foreign keys; the rest of the constraints will be deleted with tables
      var tableSchemaFilter = "TABLE_SCHEMA='" + schema + "'";
      var constraints = this.GetTableConstraints(schema);
      foreach (DataRow row in constraints.Rows)
        if (row.GetAsString("CONSTRAINT_TYPE") == "FOREIGN KEY") {
          var tableName = row.GetAsString("TABLE_NAME");
          var fkName = row.GetAsString("CONSTRAINT_NAME");
          ExecuteNonQuery("ALTER TABLE [{0}].[{1}] DROP CONSTRAINT [{2}]", schema, tableName, fkName);
        }
      //Delete tables
      var tables = this.GetTables(schema);
      foreach (DataRow row in tables.Rows) {
        var tableName = row.GetAsString("TABLE_NAME");
        ExecuteNonQuery("DROP TABLE [{0}].[{1}]", schema, tableName);
      }
      //Drop schema
      var schemas = this.GetSchemas();
      foreach (DataRow row in schemas.Rows) {
        if (row.GetAsString("SCHEMA_NAME") == schema) {
          ExecuteNonQuery("DROP SCHEMA [{0}] ;", schema);
          break;
        }
      }
    }

    public override void CreateSchema(string schema) {
      ExecuteNonQuery("CREATE SCHEMA [{0}] AUTHORIZATION dbo;", schema);
    }//method

    public override string GetProcedureTag(DataRow row) {
      var procText = row.GetAsString("ROUTINE_DEFINITION");
      if (procText == null) return null;
      var tagStart = procText.IndexOf(DbDriver.GeneratedProcedureTag);
      if (tagStart < 0)
        return null;
      var tagValueStart = procText.IndexOf("=", tagStart) + 1;
      var tagValueEnd = procText.IndexOfAny(new char[] { '\r', '\n' }, tagValueStart) - 1;
      if (tagValueEnd < 0)
        return null;
      var tag = procText.Substring(tagValueStart, tagValueEnd - tagValueStart).Trim();
      return tag;
    }

    public override void OnModelLoaded(DbModel model) {
      LoadIdentityColumnsInfo(model); 
    }

    private void LoadIdentityColumnsInfo(DbModel model) {
      var sql = @"
SELECT s.name AS table_schema, t.name AS table_name, c.name AS column_name 
  FROM sys.columns c 
  INNER JOIN sys.tables t ON t.object_id = c.object_id
  INNER JOIN sys.schemas s ON s.schema_id = t.schema_id 
  WHERE c.is_identity = 1
";
      var data = ExecuteSelect(sql);
      foreach (DataRow row in data.Rows) {
        var schema = row.GetAsString("TABLE_SCHEMA");
        if (!model.Schemas.Contains(schema))
          continue;
        var tableName = row.GetAsString("TABLE_NAME");
        var colName = row.GetAsString("COLUMN_NAME");
        var table = model.GetTable(schema, tableName);
        if (table == null) continue;
        var colInfo = table.Columns.FirstOrDefault(c => c.ColumnName == colName);
        if (colInfo != null)
          colInfo.Flags |= DbColumnFlags.Identity | DbColumnFlags.NoUpdate | DbColumnFlags.NoInsert;
      }
    }

  }//class
}
