﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data;

using Vita.Common;
using Vita.Entities;
using Vita.Entities.Model;
using Vita.Data;
using Vita.Data.Model;

namespace Vita.Data.MsSql {

  public class MsSqlTypeMapper : DbTypeMapper {
    public const int HierarchyIdSize = 892;

    public MsSqlTypeMapper(DbSettings settings) : base(settings) { }

    public override DbTypeInfo GetDbTypeInfo(Type dataType, int size = 0, DbType? explicitDbType = null, EntityMemberFlags flags = EntityMemberFlags.None, byte precision = 0, byte scale = 0) {
      var isNullable = flags.IsSet(EntityMemberFlags.Nullable);
      switch(explicitDbType) {
        case null: break; 
        case DbType.Xml: return new DbTypeInfo(explicitDbType.Value, "xml", isNullable, -1);
        case DbType.Currency:  return new DbTypeInfo(explicitDbType.Value, "money", isNullable, size, null, 19, 4);
        case DbType.Object:  return new DbTypeInfo(explicitDbType.Value, "sql_variant", isNullable);
        case DbType.Binary: return new DbTypeInfo(explicitDbType.Value, "binary({0})", isNullable, size);
        case CustomDbTypes.Image: return new DbTypeInfo(explicitDbType.Value, "image", isNullable, -1);
        case CustomDbTypes.Text: return new DbTypeInfo(explicitDbType.Value, "text", isNullable, -1);
        case CustomDbTypes.NText: return new DbTypeInfo(explicitDbType.Value, "ntext", isNullable, -1);
        case CustomDbTypes.HierarchyId: return new DbTypeInfo(explicitDbType.Value, "hierarchyid", isNullable, HierarchyIdSize);
        case CustomDbTypes.Geography: return new DbTypeInfo(explicitDbType.Value, "geography", isNullable);
        case CustomDbTypes.Geometry: return new DbTypeInfo(explicitDbType.Value, "geometry", isNullable);
        case CustomDbTypes.SmallMoney: return new DbTypeInfo(explicitDbType.Value, "smallmoney", isNullable, size, null, 10, 4);
        case CustomDbTypes.SmallDateTime: return new DbTypeInfo(explicitDbType.Value, "smalldatetime", isNullable);
        case CustomDbTypes.RowVersion: return new DbTypeInfo(explicitDbType.Value, "rowversion", false, 8);
      }
      if (dataType == typeof(byte[])) {
        if (size < 0)
          return new DbTypeInfo(CustomDbTypes.VarBinary, "varbinary(max)", isNullable, size);
        else
          return new DbTypeInfo(CustomDbTypes.VarBinary, "varbinary({0})", isNullable, size);
      }
      // Go thru type code, but first unwrap nullable value type
      if (dataType.IsNullableValueType())
        dataType = dataType.GetValueTypeFromNullable();
      //First special cases
      if (dataType == typeof(DateTimeOffset))
        return new DbTypeInfo(DbType.DateTimeOffset, "datetimeoffset", isNullable);
      //Use typecode
      var typeCode = Type.GetTypeCode(dataType);
      switch (typeCode) {
        //Note that tinyint in UNSIGNED byte!
        case TypeCode.Byte: //unsigned byte
          return new DbTypeInfo(DbType.Byte, "tinyint", isNullable);
        case TypeCode.SByte: //signed byte
          return new DbTypeInfo(DbType.Byte, "tinyint", isNullable, 0, null, 0, 0, x => (sbyte)(byte)x, x => (byte)(sbyte)x);
        case TypeCode.UInt16:
          return new DbTypeInfo(DbType.Int16, "smallint", isNullable, 0, null, 0, 0, x => (UInt16)(Int16)x, x => (Int16)(UInt16)x);
        case TypeCode.UInt32:
          return new DbTypeInfo(DbType.Int32, "int", isNullable, 0, null, 0, 0, x => (UInt32)(Int32)x, x => (Int32)(UInt32)x);
        case TypeCode.UInt64:
          return new DbTypeInfo(DbType.Int64, "bigint", isNullable, 0, null, 0, 0, x => (UInt64)(Int64)x, x => (Int64)(UInt64)x);
        case TypeCode.String:
          if (size < 0) {
            switch (explicitDbType) {
              case DbType.AnsiString:
                return new DbTypeInfo(DbType.AnsiString, "varchar(max)", isNullable, -1);
              default:
                return new DbTypeInfo(DbType.String, "nvarchar(max)", isNullable, -1);
            }
          } else
            return base.GetDbTypeInfo(dataType, size, explicitDbType, flags, precision, scale);
        case TypeCode.Decimal:
            return base.GetDbTypeInfo(dataType, size, explicitDbType, flags, precision, scale);
      }//switch
      return base.GetDbTypeInfo(dataType, size, explicitDbType, flags, precision, scale);
    }

    public override DbTypeInfo GetDbTypeInfo(DataRow schemaRow) {
      int charSize = schemaRow.GetAsInt("CHARACTER_MAXIMUM_LENGTH");
      int byteSize = schemaRow.GetAsInt("CHARACTER_OCTET_LENGTH");
      var dataTypeString = schemaRow.GetAsString("DATA_TYPE");
      var isNullable = (schemaRow.GetAsString("IS_NULLABLE") == "YES");
      string typeSpec = string.Empty; 
      dataTypeString = dataTypeString.ToLowerInvariant();
      switch (dataTypeString) {
        case "tinyint":
          return new DbTypeInfo(DbType.Byte, dataTypeString, isNullable);
        case "datetime2":
          return new DbTypeInfo(DbType.DateTime2, dataTypeString, isNullable);
        case "datetimeoffset":
          return new DbTypeInfo(DbType.DateTimeOffset, dataTypeString, isNullable);
        case "money":
          var prec = schemaRow.GetAsByte("NUMERIC_PRECISION");
          var scale = schemaRow.GetAsByte("NUMERIC_SCALE");
          return new DbTypeInfo(DbType.Currency, dataTypeString, isNullable, precision: prec, scale: scale);
        case "nvarchar":
          if (charSize == -1) {
            typeSpec = "nvarchar(max)";
          } else 
            typeSpec = string.Format("nvarchar({0})", charSize); 
          return new DbTypeInfo(DbType.String, typeSpec, isNullable, charSize);
        case "timestamp":
        case "rowversion":
          return new DbTypeInfo(CustomDbTypes.RowVersion, "rowversion", false, 8);
        case "image":
          return new DbTypeInfo(CustomDbTypes.Image, "image", isNullable, -1);
        case "sql_variant":
          return new DbTypeInfo(DbType.Object, "sql_variant", isNullable);
        case "text":
          return new DbTypeInfo(CustomDbTypes.Text, "text", isNullable, -1);
        case "ntext":
          return new DbTypeInfo(CustomDbTypes.NText, "ntext", isNullable, -1);
        case "hierarchyid":
          return new DbTypeInfo(CustomDbTypes.HierarchyId, "hierarchyid", isNullable, HierarchyIdSize, clrType: typeof(byte[]));
        case "geography":
          return new DbTypeInfo(CustomDbTypes.Geography, "geography", isNullable);
        case "geometry":
          return new DbTypeInfo(CustomDbTypes.Geometry, "geometry", isNullable);
        case "smalldatetime":
          return new DbTypeInfo(CustomDbTypes.SmallDateTime, "smalldatetime", isNullable);
        case "smallmoney":
          var prec2 = schemaRow.GetAsByte("NUMERIC_PRECISION");
          var scale2 = schemaRow.GetAsByte("NUMERIC_SCALE");
          return new DbTypeInfo(CustomDbTypes.SmallMoney, "smallmoney", isNullable, precision: prec2, scale: scale2);

      }//switch
      return base.GetDbTypeInfo(schemaRow);
    }

  }//class
}

