﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data;

using Vita.Common;
using Vita.Entities;
using Vita.Entities.Model;
using Vita.Data.Model;
using Vita.Data.Driver;

namespace Vita.Data.MySql {
  
  public partial class MySqlBuilder : DbSqlBuilder {

    public MySqlBuilder(DbModel dbModel) : base(dbModel) { }

    protected override SqlLimitClause GetSqlLimitClause(string count) {
      return new SqlLimitClause() { LimitClause = "Limit " + count};
    }

    public override DbCommandInfo BuildSelectAllPagedCommand(EntityCommand entCommand) {
      const string SqlSelectAllPaged = @"
SELECT {0} FROM {1}
  {2}
  LIMIT {3}, {4};
";

      var table = DbModel.LookupDbObject<DbTableInfo>(entCommand.TargetEntityInfo, throwNotFound: true);
      //Build column list
      var strColumns = table.Columns.GetSelectable().GetSqlNameList(); 
      var strOrderBy = BuildOrderBy(table.DefaultOrderBy);
      var cmdName = Settings.NamingPolicy.ConstructDbCommandName(entCommand, table.TableName, "SelectAllPaged");
      var cmdInfo = CreateDbCommandInfo(entCommand, cmdName, table, DbExecutionType.Query, string.Empty);
      var skipPrm = cmdInfo.Parameters[0].Name;
      var takePrm = cmdInfo.Parameters[1].Name;
      cmdInfo.Sql = string.Format(SqlSelectAllPaged, strColumns, table.FullName, strOrderBy, skipPrm, takePrm);
      cmdInfo.Flags |= DbCommandFlags.WithPaging; 
      return cmdInfo;
    }

    // 
    public override DbCommandInfo BuildSqlInsertCommand(EntityCommand entityCommand) {
      const string SqlInsertTemplate = @"
INSERT INTO {0} 
  ({1}) 
  VALUES
    ({2}); 
{3}
";
      const string SqlGetIdentityTemplate = "\r\nSET {0} = LAST_INSERT_ID();";
      var table = DbModel.LookupDbObject<DbTableInfo>(entityCommand.TargetEntityInfo, throwNotFound: true);
      var idClause = string.Empty;
      var listColumns = new List<DbColumnInfo>();
      var listValues = new StringList();
      var cmdName = Settings.NamingPolicy.ConstructDbCommandName(entityCommand, table.TableName, "Insert");
      var dbCmdInfo = CreateDbCommandInfo(entityCommand, cmdName, table, DbExecutionType.NonQuery, null);
      foreach (var prm in dbCmdInfo.Parameters) {
        var col = prm.SourceColumn;
        if (col.Flags.IsSet(DbColumnFlags.Identity))
          idClause = string.Format(SqlGetIdentityTemplate, prm.Name);
        if (!col.Flags.IsSet(DbColumnFlags.NoInsert)) {
          listColumns.Add(col);
          listValues.Add(prm.Name);
        }
      }
      //build SQL
      var strColumns = listColumns.GetSqlNameList();
      var strValues = string.Join(", ", listValues);
      dbCmdInfo.Sql = string.Format(SqlInsertTemplate, table.FullName, strColumns, strValues, idClause);
      return dbCmdInfo;
    }

    public override DbCommandInfo BuildSqlUpdateCommand(EntityCommand entityCommand) {
      var hasRowVersion = entityCommand.TargetEntityInfo.Flags.IsSet(EntityFlags.HasRowVersion);
      if (hasRowVersion)
        return BuildSqlUpdateCommandWithConcurrencyCheck(entityCommand);
      else
        return base.BuildSqlUpdateCommand(entityCommand); 
    }

    public override DbCommandInfo BuildSqlDeleteCommand(EntityCommand entityCommand) {
      var hasRowVersion = entityCommand.TargetEntityInfo.Flags.IsSet(EntityFlags.HasRowVersion);
      if (hasRowVersion)
        return BuildSqlDeleteCommandWithConcurrencyCheck(entityCommand);
      else
        return base.BuildSqlDeleteCommand(entityCommand); 
    }

    private DbCommandInfo BuildSqlUpdateCommandWithConcurrencyCheck(EntityCommand entityCommand) {
      const string SqlUpdateTemplate = @"
UPDATE {0}
  SET {1} 
  WHERE {2}; {3}{4}
";

      var table = DbModel.LookupDbObject<DbTableInfo>(entityCommand.TargetEntityInfo, throwNotFound: true);
      if (entityCommand == null)
        return null;
      var cmdName = this.Settings.NamingPolicy.ConstructDbCommandName(entityCommand, table.TableName, "Update");
      var cmdInfo = CreateDbCommandInfo(entityCommand, cmdName, table, DbExecutionType.NonQuery, null);
      var updateParams = cmdInfo.Parameters.Where(p => !p.SourceColumn.Flags.IsSet(DbColumnFlags.PrimaryKey | DbColumnFlags.NoUpdate));
      // Some tables (like many-to-many link entities) might have no columns to update
      if (!updateParams.Any())
        return null;
      //Build Where expression
      var pkParams = cmdInfo.Parameters.Where(p => p.SourceColumn.Flags.IsSet(DbColumnFlags.PrimaryKey)).ToList();
      var whereParams = pkParams.ToList(); //make a copy
      var rvParam = cmdInfo.Parameters.First(p => p.TypeInfo.DbType == CustomDbTypes.RowVersion);
      whereParams.Add(rvParam);       
      var whereExpr = BuildWhereClause(whereParams);
      //Build Update clause
      var updList = new StringList();
      foreach (var prm in updateParams)
        updList.Add("[" + prm.SourceColumn.ColumnName + "] = " + prm.Name);
      var strUpdates = string.Join(", ", updList);
      var strCheckRowCount = BuildRowCountCheckStatement(table, pkParams, entityCommand.Kind);
      var getRowVersion = string.Format("\r\nSET {0} = @@DBTS;", rvParam.Name);
      //Build SQL
      cmdInfo.Sql = string.Format(SqlUpdateTemplate, table.FullName, strUpdates, whereExpr, strCheckRowCount, getRowVersion);
      return cmdInfo;
    }

    //Note: we do not check RowVersion on delete (no need for this) - just checking that @@RowCount is not zero.
    private DbCommandInfo BuildSqlDeleteCommandWithConcurrencyCheck(EntityCommand entityCommand) {
      const string SqlDeleteTemplate = @"
DELETE FROM {0} 
  WHERE {1}; {2}"; 
      var table = DbModel.LookupDbObject<DbTableInfo>(entityCommand.TargetEntityInfo, throwNotFound: true);
      //Load by primary key
      var cmdName = this.Settings.NamingPolicy.ConstructDbCommandName(entityCommand, table.TableName, "Delete");
      var cmdInfo = CreateDbCommandInfo(entityCommand, cmdName, table, DbExecutionType.NonQuery, null);
      var strWhere = BuildWhereClause(cmdInfo.Parameters);
      var strCheckRowCount = BuildRowCountCheckStatement(table, cmdInfo.Parameters, entityCommand.Kind);
      cmdInfo.Sql = string.Format(SqlDeleteTemplate, table.FullName, strWhere, strCheckRowCount);
      return cmdInfo;
    }

    //creates a TSQL statement that raises error with custom message like 'VITA:Concurrency/Update/books.Author/123' (123 is primary key value)
    private string BuildRowCountCheckStatement(DbTableInfo table, List<DbParamInfo> pkParams, EntityCommandKind commandKind) {
      const string sqlCheckRowCount = @"
IF @@RowCount = 0
BEGIN
  DECLARE @msg NVARCHAR(200) = {0};
  RAISERROR(@msg, 11, 111);
END
";
      //Build message expression
      var opType = commandKind == EntityCommandKind.Update ? "Update" : "Delete";
      var msg = "'" + ExceptionStrings.CodeConcurrentUpdateConflict + "/" + opType + "/" + table.FullName + "/' + ";
      var pkExprs = pkParams.Select(p => string.Format("CAST({0} AS NVARCHAR(50))", p.Name));
      var strPks = string.Join(" + ';' + ", pkExprs);
      msg += strPks;
      var result = string.Format(sqlCheckRowCount, msg);
      return result; 

    }

  }//class
}
