﻿using System;
using System.Data;
using System.Diagnostics;
using System.Linq;

using Vita.Common;
using Vita.Entities;
using Vita.Entities.Model;
using Vita.Data;
using Vita.Data.Model;
using Vita.Data.Driver;
using Vita.Data.Model.Construction;
using System.Collections.Generic;
using MySql.Data.MySqlClient;

namespace Vita.Data.MySql {
  /* Note:
   * 1. Add the following options to sql-mode variable in my.ini file: 
   *      ANSI_QUOTES
   * 2. MySql supports ordered columns in indexes, but there's no way to get this information when loading index columns from the database. So we suppress ordering; 
   *     we also set all column direction to ASC after construction DbModel
   * 3. Connection string must have ' Old Guids=true' to properly handle Guids
   
   */

  public class MySqlDriver : DbDriver {
    public const DbCapabilities DefaultCapabilities = DbCapabilities.Schemas | DbCapabilities.StoredProcedures | DbCapabilities.OutputParameters
        | DbCapabilities.ReferentialConstraints | DbCapabilities.Paging | DbCapabilities.BatchedUpdates; // | DbCapabilities.OrderedColumnsInIndexes;
    public const DbOptions DefaultMySqlDbOptions = DbOptions.Default | DbOptions.AutoIndexForeignKeys;

    public MySqlDriver()  {
      Capabilities = DefaultCapabilities;
      base.StoredProcParameterPrefix = "prm";
      base.DynamicSqlParameterPrefix = "@";
      base.CommandCallLogFormat = "CALL {0} ({1});";
    }

    // overrides
    public override IDbConnection CreateConnection(string connectionString) {
      return new MySqlConnection(connectionString);
    }
    public override IDbCommand CreateDbCommand() {
      return new MySqlCommand();
    }

    public override DbSchemaManager CreateSchemaManager(string connectionString) {
      return new MySqlSchemaManager(this, connectionString);
    }
    public override DbSqlBuilder CreateSqlBuilder(DbModel dbModel) {
      return new MySqlBuilder(dbModel); 
    }
    public override DbTypeMapper CreateTypeMapper(DbSettings settings) {
      return new MySqlTypeMapper(settings); 
    }

    public override Vita.Data.Driver.LinqSqlProvider GetLinqSqlProvider() {
      return new MySqlLinqSqlProvider();
    }

    public override string GetFullName(string schema, string name) {
      return string.Format("\"{0}\".\"{1}\"", schema, name);
    }

    public override void CheckDatabaseException(Exception exception, IDbCommand command = null) {
      var sqlEx = exception as MySqlException;
      if (sqlEx == null) return;
      switch (sqlEx.Number) {
        case 1062:
          var indexName = GetIndexName(sqlEx.Message); 
          throw new UniqueIndexViolationException(sqlEx.Message, indexName, sqlEx); 
        
        default: 
          break; 
      } //switch
      base.CheckDatabaseException(exception, command); 
    }

    private string GetIndexName(string message) {
      try {
        // Sample message: Duplicate entry 'MS Books' for key 'IXU_Publisher_Name'
        var constrNameTag = "for key '";
        var tagPos = message.IndexOf(constrNameTag);
        var start = tagPos + constrNameTag.Length;
        var end = message.IndexOf("'", start);
        var indexName = message.Substring(start, end - start).Trim();
        return indexName;
      } catch (Exception ex) {
        return "(failed to identify index in message: " + ex.Message + ")"; //
      }
    }

    protected override IDbDataParameter CreateDbParameter(IDbCommand command, DbParamInfo prmInfo) {
      var typeInfo = prmInfo.TypeInfo;
      MySqlParameter prm = null;
      switch (typeInfo.DbType) {
        case DbType.Object:
        case CustomDbTypes.HierarchyId:
        case CustomDbTypes.Geography:
          prm = null; //to avoid compiler error
          Util.Throw("Not supported custom DbType {0}, parameter {1}.", typeInfo.DbType, prmInfo.Name);
          break;
        case DbType.Currency:
          prm = new MySqlParameter(prmInfo.Name, MySqlDbType.Decimal);
          break;
        case CustomDbTypes.VarBinary:
          prm = new MySqlParameter(prmInfo.Name, MySqlDbType.VarBinary, typeInfo.Size);
          break;
        case CustomDbTypes.RowVersion:
          prm = new MySqlParameter(prmInfo.Name, MySqlDbType.Int64, 8);
          break; 
        case CustomDbTypes.NText:
        case CustomDbTypes.Text:
          prm = new MySqlParameter(prmInfo.Name, MySqlDbType.Text, -1);
          break;
        case CustomDbTypes.Image:
          prm = new MySqlParameter(prmInfo.Name, MySqlDbType.Blob, -1);
          break;

        case CustomDbTypes.Geometry:
          prm = new MySqlParameter(prmInfo.Name, MySqlDbType.Geometry, -1);
          break;
        case CustomDbTypes.SmallMoney:
          prm = new MySqlParameter(prmInfo.Name, MySqlDbType.Decimal);
          break;
        case CustomDbTypes.SmallDateTime:
          prm = new MySqlParameter(prmInfo.Name, MySqlDbType.DateTime);
          break;
        default:
          return base.CreateDbParameter(command, prmInfo);
      }
      prm.Direction = prmInfo.Direction;
      if (prmInfo.SourceColumn != null)
        prm.SourceColumn = prmInfo.SourceColumn.ColumnName;
      prm.Direction = prmInfo.Direction;
      if (typeInfo.Precision > 0) prm.Precision = typeInfo.Precision;
      if (typeInfo.Scale > 0) prm.Scale = typeInfo.Scale;
      if (typeInfo.Size > 0) prm.Size = typeInfo.Size;
      prm.Value = prmInfo.DefaultValue;
      return prm;
    }

    protected override IDbDataParameter CloneDbParameter(IDbCommand command, IDbDataParameter parameter) {
      var sqlPrm = (MySqlParameter) parameter;
      var result = new MySqlParameter(parameter.ParameterName, sqlPrm.MySqlDbType, sqlPrm.Size, sqlPrm.Direction, 
                    sqlPrm.IsNullable, sqlPrm.Precision, sqlPrm.Scale, sqlPrm.SourceColumn, sqlPrm.SourceVersion, sqlPrm.Value);
      return result; 
    }

    //Note: as of May/13, MySql does not preserve comments preceding SP header; they are going to patch this.
    // For now, we use COMMENTS clause to save the Descriptive tag
    public override bool ConvertToStoredProc(DbCommandInfo command) {
      var table = command.Table;
      string ProcBodyTemplate =
@"BEGIN
  -- Description: {0}
  {1}
END";
      string CreateProcTemplate = 
@"CREATE PROCEDURE {0}(
  {1}) 
  SQL SECURITY INVOKER
  COMMENT '{2}' 
{3}";
      command.Schema = table.SchemaName;
      //Build command that creates stored proc and execute it
      var listParams = new StringList();
      foreach (var prm in command.Parameters) {
        var strOut = (prm.Direction & ParameterDirection.Output) != 0 ? "OUT " : string.Empty;
        var prmSpec = string.Format("    {0}{1} {2}", strOut, prm.Name, prm.TypeInfo.SqlTypeSpec);
        listParams.Add(prmSpec);
      }
      var strParams = string.Join(",\r\n", listParams);
      var tag = DbDriver.GeneratedProcedureTag + "=" + command.DescriptiveTag;
      command.StoredProcBody = string.Format(ProcBodyTemplate, command.Description, command.Sql);
      command.StoredProcText = string.Format(CreateProcTemplate, command.FullCommandName, strParams, tag, command.StoredProcBody);
      return true;
    }

    // MySql stores only proc body, without header (proc name with parameters), so we compare bodies only
    public override bool StoredProceduresMatch(DbCommandInfo x, DbCommandInfo y) {
      if (string.IsNullOrEmpty(x.StoredProcBody) || string.IsNullOrEmpty(y.StoredProcBody))
        return false; 
      if (x.StoredProcBody == y.StoredProcBody)
        return true;
      if (x.StoredProcBody.Trim() == y.StoredProcBody.Trim())
        return true;
      return false;
    }


    public override void OnDbModelConstructed(DbModel dbModel) {
      foreach (var table in dbModel.Tables) {
        // Names of PK constraints in MySql is 'PRIMARY', cannot be changed
        // PK is also clustered index
        if (table.PrimaryKey != null) {
          table.PrimaryKey.Name = "PRIMARY";
          table.PrimaryKey.KeyType |= KeyType.Clustered;
        }
        foreach (var key in table.Keys) {
          // All foreign keys have a supporting index; if there's no matching index already, it is created automatically.
          if (key.KeyType.IsSet(KeyType.ForeignKey)) {
            var supportingIndex = FindSupportingIndex(table.Keys, key);
            if (supportingIndex == null) //if no supporting index, then mark this key as an index
              key.KeyType |= KeyType.Index;
          }
          //Drop descending columns - see notes at the top of the file
          if (key.DescendingColumns != null && key.DescendingColumns.Count > 0)
            key.DescendingColumns.Clear(); 
        }
        // auto_increment (identity) columns must be associated with key (PK or index). For auto-inc columns that are NOT PKs we create artificial index
        var autoIncCols = table.Columns.Where(c => c.Flags.IsSet(DbColumnFlags.Identity));
        foreach (var col in autoIncCols) {
          if (col.Flags.IsSet(DbColumnFlags.PrimaryKey))
            continue;
          var ind = new DbKeyInfo(col.ColumnName, col.Table, KeyType.Index); //added automatically to table.Keys list
          ind.Columns.Add(col);
        }
      
      }     
      // Remove double-quotes from StoredProcedure names - MySqlDriver does not like it
      foreach (var cmd in dbModel.Commands) {
        cmd.FullCommandName = cmd.FullCommandName.Replace("\"", "");
      }
      base.OnDbModelConstructed(dbModel);
    }

    private DbKeyInfo FindSupportingIndex(IList<DbKeyInfo> allKeys, DbKeyInfo foreignKey) {
      foreach (var key in allKeys) {
        if (!key.KeyType.IsSet(KeyType.Index | KeyType.PrimaryKey) || key.Columns.Count < foreignKey.Columns.Count) 
          continue;
        //Check columns are the same
        bool match = true; 
        for (int i = 0; i < foreignKey.Columns.Count; i++) {
          if (foreignKey.Columns[i] != key.Columns[i])
            match = false; 
        }
        if (match)
          return key; 
      }
      return null; 
    }

  }//class
}
