﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data;

using Vita.Common;
using Vita.Entities.Model;
using Vita.Data;
using Vita.Data.Model;
using Vita.Data.Driver;

namespace Vita.Data.MySql {

  public partial class MySqlSchemaManager : DbSchemaManager {
    public MySqlSchemaManager(DbDriver driver, string connectionString) : base(driver, connectionString) { }

    //Columns:CATALOG_NAME	SCHEMA_NAME	SCHEMA_OWNER	DEFAULT_CHARACTER_SET_CATALOG	DEFAULT_CHARACTER_SET_SCHEMA 
    //	DEFAULT_CHARACTER_SET_NAME  
    public override DataTable GetSchemas() {
      //simple way to find non-system schemas - in MS SQL Server
      //var sql = "SELECT * FROM INFORMATION_SCHEMA.SCHEMATA WHERE  SCHEMA_NAME <> SCHEMA_OWNER;";
      var sql = "SELECT * FROM INFORMATION_SCHEMA.SCHEMATA;";
      return ExecuteSelect(sql);
    }

    //Columns: TABLE_CATALOG	TABLE_SCHEMA	TABLE_NAME	TABLE_TYPE  (table_type: "BASE TABLE") 
    public override DataTable GetTables(string schema = null) {
      var sql = "SELECT * FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_TYPE='BASE TABLE' ";
      if (!string.IsNullOrEmpty(schema))
        sql += string.Format(" AND TABLE_SCHEMA='{0}'", schema);
      return ExecuteSelect(sql);
    }

    // ROUTINE_COMMENT is extra column (extension to standard); we use it to store proc tag.
    public override DataTable GetProcedures(string schema = null) {
      const string getRoutinesTemplate = @"
SELECT R.ROUTINE_SCHEMA, R.ROUTINE_NAME, R.ROUTINE_DEFINITION AS ROUTINE_BODY, '' AS ROUTINE_DEFINITION, R.ROUTINE_COMMENT
FROM INFORMATION_SCHEMA.ROUTINES R
WHERE R.ROUTINE_TYPE='PROCEDURE'
";
      var sql = getRoutinesTemplate;
      if (!string.IsNullOrEmpty(schema))
        sql += string.Format(" AND ROUTINE_SCHEMA='{0}'", schema);
      return ExecuteSelect(sql);
    }

    //INFORMATION_SCHEMA does not have a view for indexes, so we have to do it through MSSQL special objects
    public override DataTable GetIndexes(string schema = null) {
      const string getIndexesTemplate = @"
SELECT DISTINCT TABLE_SCHEMA, TABLE_NAME, INDEX_NAME, (0=1) AS PRIMARY_KEY, '' AS CLUSTERED, 
  (-NON_UNIQUE + 1) AS ""UNIQUE"", 0 AS DISABLED  
  from information_schema.STATISTICS
where
  INDEX_NAME != 'Primary' AND {0}
ORDER BY TABLE_SCHEMA, TABLE_NAME, INDEX_NAME
";
      const string schemaCondTemplate = "(TABLE_SCHEMA = '{0}')";
      var strCond = string.IsNullOrWhiteSpace(schema) ? "(1=1)" : string.Format(schemaCondTemplate, schema);
      var sql = string.Format(getIndexesTemplate, strCond); 
      return ExecuteSelect(sql);
    }

    public override DataTable GetIndexColumns(string schema = null) {
      const string getIndexColumnsTemplate = @"
SELECT TABLE_SCHEMA, TABLE_NAME, INDEX_NAME, COLUMN_NAME, 
  SEQ_IN_INDEX AS COLUMN_ORDINAL_POSITION, 0 AS IS_DESCENDING
  FROM information_schema.STATISTICS
where
  INDEX_NAME != 'Primary' AND {0}
ORDER BY TABLE_SCHEMA, TABLE_NAME, INDEX_NAME, COLUMN_ORDINAL_POSITION
";
      const string schemaCondTemplate = "(TABLE_SCHEMA = '{0}')";
      var strCond = string.IsNullOrWhiteSpace(schema) ? "(1=1)" : string.Format(schemaCondTemplate, schema);
      var sql = string.Format(getIndexColumnsTemplate, strCond);
      return ExecuteSelect(sql);
    }

    public override void RenameTable(DbTableInfo oldTable, DbTableInfo newTable) {
      ExecuteNonQuery(@"ALTER TABLE {0} RENAME TO {1};", oldTable.FullName, newTable.FullName);
    }

    public override void RenameColumn(DbColumnInfo oldColumn, DbColumnInfo newColumn) {
      // ALTER TABLE  "upd"."parententity" CHANGE StringProp2_OldName StringProp2_NewName VARCHAR(40) [NOT NULL];
      var colSpec = GetColumnSpec(newColumn, DbColumnSpecOptions.None);
      ExecuteNonQuery("ALTER TABLE {0}  CHANGE \"{1}\" {2};",
          oldColumn.Table.FullName, oldColumn.ColumnName, colSpec);
    }

    public override void DropSchemaObjects(string schema) {
      Util.Check(!string.IsNullOrEmpty(schema), "Schema parameter may not be empty. (MsSqlDriver.DropSchemaObjects)"); 
      //Delete procedures
      var procSchemaFilter = "ROUTINE_SCHEMA='" + schema + "'";
      var procs = this.GetProcedures(schema);
      foreach (DataRow row in procs.Rows) {
        var procName = row.GetAsString("ROUTINE_NAME");
        ExecuteNonQuery("DROP PROCEDURE {0};", GetFullName(schema, procName));
      }
      //Delete foreign keys; the rest of the constraints will be deleted with tables
      var tableSchemaFilter = "TABLE_SCHEMA='" + schema + "'";
      var constraints = this.GetTableConstraints(schema);
      foreach (DataRow row in constraints.Rows)
        if (row.GetAsString("CONSTRAINT_TYPE") == "FOREIGN KEY") {
          var tableName = row.GetAsString("TABLE_NAME");
          var fkName = row.GetAsString("CONSTRAINT_NAME");
          ExecuteNonQuery("ALTER TABLE {0} DROP FOREIGN KEY {1};", GetFullName(schema, tableName),  fkName);
        }
      //Delete tables
      var tables = this.GetTables(schema);
      foreach (DataRow row in tables.Rows) {
        var tableName = row.GetAsString("TABLE_NAME");
        ExecuteNonQuery("DROP TABLE {0}", GetFullName(schema, tableName));
      }
      //Drop schema
      var schemas = this.GetSchemas();
      foreach (DataRow row in schemas.Rows) {
        if (row.GetAsString("SCHEMA_NAME") == schema) {
          ExecuteNonQuery("DROP SCHEMA \"{0}\";", schema);
          break;
        }
      }
    }

    private string GetFullName(string schema, string name) {
      if (string.IsNullOrEmpty(schema))
        return string.Format("\"{0}\"", name);
      return string.Format("\"{0}\".\"{1}\"", schema, name);
    }

    private string GetKeyTypeAsString(DbKeyInfo key) {
      if (key.KeyType.IsSet(KeyType.ForeignKey))
        return "FOREIGN KEY";
      if (key.KeyType.IsSet(KeyType.PrimaryKey))
        return "PRIMARY KEY";
      if (key.KeyType.IsSet(KeyType.Index))
        return "INDEX";
      return "CONSTRAINT";
    }

    public override void CreateSchema(string schema) {
      ExecuteNonQuery("CREATE SCHEMA IF NOT EXISTS \"{0}\";", schema);
    }//method

    protected override string GetColumnSpec(DbColumnInfo column, DbColumnSpecOptions options) {
      var typeStr = column.TypeInfo.SqlTypeSpec;
      var nullable = options.IsSet(DbColumnSpecOptions.ForceNull) || column.Flags.IsSet(DbColumnFlags.Nullable);
      var nullStr = nullable ? "NULL" : "NOT NULL";
      var strAutoInc = string.Empty;
      if (options.IsSet(DbColumnSpecOptions.NewColumn) && column.Flags.IsSet(DbColumnFlags.Identity)) {
        // MySql requires that auto-incr column is supported by a key - either a primary key, or an index
        var strKeyType = column.Flags.IsSet(DbColumnFlags.PrimaryKey) ? "PRIMARY KEY" : "KEY";
        strAutoInc = string.Format("AUTO_INCREMENT, {0}(\"{1}\")", strKeyType, column.ColumnName);
      }
      string defaultStr = null;
      //Default constraint can be set only on new columns
      if (!string.IsNullOrWhiteSpace(column.DefaultExpression) && options.IsSet(DbColumnSpecOptions.NewColumn))
        defaultStr = "DEFAULT " + column.DefaultExpression;
      var spec = string.Format(@" ""{0}"" {1} {2} {3}", column.ColumnName, typeStr, nullStr, strAutoInc);
      return spec;
    }

    //ALTER TABLE hr.employees MODIFY COLUMN employee_name varchar(100) Null;
    public override void ModifyColumn(DbColumnInfo column) {
      var colSpec = GetColumnSpec(column, DbColumnSpecOptions.None);
      var tbl = column.Table;
      ExecuteNonQuery(@"ALTER TABLE ""{0}"".""{1}"" MODIFY COLUMN {2}", tbl.SchemaName, tbl.TableName, colSpec);
    }
    public override void CreatePrimaryKey(DbKeyInfo key) {
      // PK for Identity (Auto-increment) columns is created when table/ID columns is created
      if (key.Columns[0].Flags.IsSet(DbColumnFlags.Identity))
        return; 
      var fullTableRef = key.Table.FullName;
      var pkFields = key.Columns.GetSqlNameList();
      // PK name is always 'PRIMARY'
      ExecuteNonQuery("ALTER TABLE {0} ADD CONSTRAINT PRIMARY KEY ({1});", fullTableRef, pkFields);
    }

    public override void DropDbKey(DbKeyInfo dbKey) {
      string template = string.Empty;
      if (dbKey.KeyType.IsSet(KeyType.ForeignKey))
        template += "ALTER TABLE {0} DROP FOREIGN KEY {1};\r\n";
      if (dbKey.KeyType.IsSet(KeyType.PrimaryKey)) {
        if (!dbKey.Columns[0].Flags.IsSet(DbColumnFlags.Identity))
          //do not drop PKs with identity - it will fail
          template += "ALTER TABLE {0} DROP PRIMARY KEY;\r\n";
      }
      if (dbKey.KeyType.IsSet(KeyType.Index))
        template += "ALTER TABLE {0} DROP INDEX {1};\r\n";
      if (!string.IsNullOrEmpty(template))
        ExecuteNonQuery(template, dbKey.Table.FullName, dbKey.Name);
    }

    public override void DropTableConstraint(DbKeyInfo key) {
      //do nothing here - we drop FK together with corresponding index, we drop keys
    }

    // In MySql all PKs are named 'PRIMARY'; because of this the SQL in DbSchemaManager for constraints does NOT work. 
    // But in MySql we can use even a simpler query, without joins - default ref_constraints view contains the tables we need 
    public override DataTable GetReferentialConstraintsExt(string schema = null) {
      var sql = @"
  SELECT rc.*, rc.TABLE_NAME as C_TABLE, rc.REFERENCED_TABLE_NAME AS U_TABLE 
    FROM INFORMATION_SCHEMA.REFERENTIAL_CONSTRAINTS rc;
";
      if (!string.IsNullOrEmpty(schema))
        sql += string.Format(" WHERE tc1.CONSTRAINT_SCHEMA='{0}'", schema);
      return ExecuteSelect(sql);
    }

    public override string GetProcedureTag(DataRow row) {
      var tag = row.GetAsString("ROUTINE_COMMENT");
      if (string.IsNullOrEmpty(tag))
        return null;
      var eqPos = tag.IndexOf("=");
      if (eqPos < 0)
        return null;
      var value = tag.Substring(eqPos + 1);
      return value; 
    }

    public override void OnModelLoaded(DbModel model) {
      foreach (var table in model.Tables) {
        if (table.PrimaryKey != null) //all primary keys are clustered keys by default
          table.PrimaryKey.KeyType |= KeyType.Clustered;
      }
      LoadIdentityColumnsInfo(model); 
      base.OnModelLoaded(model);
    }

    private void LoadIdentityColumnsInfo(DbModel model) {
      var sql = @"
SELECT table_schema, table_name, column_name
FROM INFORMATION_SCHEMA.Columns 
  WHERE EXTRA = 'auto_increment'
";
      var data = ExecuteSelect(sql);
      foreach (DataRow row in data.Rows) {
        var schema = row.GetAsString("TABLE_SCHEMA");
        if (!model.Schemas.Contains(schema))
          continue;
        var tableName = row.GetAsString("TABLE_NAME");
        var colName = row.GetAsString("COLUMN_NAME");
        var table = model.GetTable(schema, tableName);
        if (table == null) continue;
        var colInfo = table.Columns.FirstOrDefault(c => c.ColumnName == colName); 
        if (colInfo != null)
          colInfo.Flags |= DbColumnFlags.Identity | DbColumnFlags.NoUpdate | DbColumnFlags.NoInsert;
      }
    }

  }//class
}
