﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data;

using Vita.Common;
using Vita.Entities;
using Vita.Entities.Model;
using Vita.Data;
using Vita.Data.Model;

namespace Vita.Data.MySql {

  public class MySqlTypeMapper : DbTypeMapper {

    public MySqlTypeMapper(DbSettings settings) : base(settings) { }

    public override DbTypeInfo GetDbTypeInfo(Type dataType, int size = 0, DbType? explicitDbType = null, EntityMemberFlags flags = EntityMemberFlags.None, byte precision = 0, byte scale = 0) {
      var isNullable = flags.IsSet(EntityMemberFlags.Nullable); 
      switch(explicitDbType) {
        case null: break; 
        case DbType.Xml: return new DbTypeInfo(explicitDbType.Value, "longtext", isNullable, -1);
        case DbType.Binary: return new DbTypeInfo(explicitDbType.Value, "longblob", isNullable, size);
        case DbType.Currency: return new DbTypeInfo(DbType.Decimal, "decimal(19,4)", isNullable, size, null, 19, 4);
        case CustomDbTypes.Image: return new DbTypeInfo(explicitDbType.Value, "longblob", isNullable, -1);
        case CustomDbTypes.Text:
        case CustomDbTypes.NText: return new DbTypeInfo(explicitDbType.Value, "longtext", isNullable, -1);
        case CustomDbTypes.Geometry: return new DbTypeInfo(explicitDbType.Value, "geometry", isNullable);
        default:
          //we may end up calling base method
          break; 
      }
      if (dataType == typeof(byte[])) {
        if (size < 0)
          return new DbTypeInfo(CustomDbTypes.VarBinary, "longblob", isNullable, -1);
        else
          return new DbTypeInfo(CustomDbTypes.VarBinary, "blob({0})", isNullable, size);
      }
      // Go thru type code, but first unwrap nullable value type
      if (dataType.IsNullableValueType())
        dataType = dataType.GetValueTypeFromNullable();
      //First special cases
      if (dataType == typeof(Guid))
        return new DbTypeInfo(DbType.Binary, "binary(16)", isNullable, 16, readConverter: BinaryToGuid, writeConverter: GuidToBinary); 
      //Use typecode
      var typeCode = Type.GetTypeCode(dataType);
      switch (typeCode) {
        //Note that tinyint in SIGNED byte!
        case TypeCode.Byte: //unsigned byte
          return new DbTypeInfo(DbType.Int16, "smallint", isNullable, readConverter: x => (byte)(Int16)x, writeConverter: x => (Int16)(byte)x);
        case TypeCode.SByte: //signed byte
          return new DbTypeInfo(DbType.SByte, "tinyint", isNullable); 
        case TypeCode.UInt16:
          return new DbTypeInfo(DbType.Int16, "smallint", isNullable, 0, null, 0, 0, x => (UInt16)(Int16)x, x => (Int16)(UInt16)x);
        case TypeCode.UInt32:
          return new DbTypeInfo(DbType.Int32, "int", isNullable, 0, null, 0, 0, x => (UInt32)(Int32)x, x => (Int32)(UInt32)x);
        case TypeCode.UInt64:
          return new DbTypeInfo(DbType.Int64, "bigint", isNullable, 0, null, 0, 0, x => (UInt64)(Int64)x, x => (Int64)(UInt64)x);
        case TypeCode.Single:
          return new DbTypeInfo(DbType.Single, "float", isNullable);
        case TypeCode.Double:
          return new DbTypeInfo(DbType.Double, "double", isNullable);
        case TypeCode.DateTime: // there's no datetime2 in MYSQL
          return new DbTypeInfo(DbType.DateTime, "datetime", isNullable);
        case TypeCode.Char:
          switch (explicitDbType) {
            case DbType.AnsiStringFixedLength:
              return new DbTypeInfo(DbType.AnsiStringFixedLength, "char(1)", isNullable, 1, null, 0, 0, s => ((string)s)[0], c => c.ToString());
            case DbType.StringFixedLength:
            default:
              return new DbTypeInfo(DbType.StringFixedLength, "char(1)", isNullable, 1, null, 0, 0, s => ((string)s)[0], c => c.ToString());
          }

        case TypeCode.String:
          if (size < 0) 
              return new DbTypeInfo(DbType.String, "longtext", isNullable, -1);
          switch (explicitDbType) {
            case DbType.AnsiStringFixedLength: return new DbTypeInfo(DbType.AnsiStringFixedLength, "char({0})", isNullable, size);
            case DbType.StringFixedLength: return new DbTypeInfo(DbType.StringFixedLength, "char({0})", isNullable, size);
            case DbType.AnsiString: return new DbTypeInfo(DbType.AnsiString, "varchar({0})", isNullable, size);
            default:
              return new DbTypeInfo(DbType.String, "varchar({0})", isNullable, size);
          }
        case TypeCode.Boolean:
          return new DbTypeInfo(DbType.Boolean, "bit", isNullable, 
            readConverter: x => (ulong)x == 1, 
            writeConverter: x => x);
        
        
        case TypeCode.Decimal:
            return base.GetDbTypeInfo(dataType, size, explicitDbType, flags, precision, scale);
      }//switch
      return base.GetDbTypeInfo(dataType, size, explicitDbType, flags, precision, scale);
    }

    private static object BinaryToGuid(object value) {
      if (value == DBNull.Value) return DBNull.Value;
      // It might happen that driver performs conversion to Guid automatically for binary(16) columns
      if (value is Guid)
        return (Guid)value; 
      var bytes = (byte[])value;
      var guid = new Guid(bytes);
      return guid; 
    }
    private static object GuidToBinary(object value) {
      if (value == null || value == DBNull.Value) 
        return DBNull.Value;
      var guid = (Guid)value;
      var bytes = guid.ToByteArray(); 
      return bytes;
    }

    public override DbTypeInfo GetDbTypeInfo(DataRow schemaRow) {
      var charSize = schemaRow.GetAsLong("CHARACTER_MAXIMUM_LENGTH");
      var byteSize = schemaRow.GetAsLong("CHARACTER_OCTET_LENGTH");
      var dataTypeString = schemaRow.GetAsString("DATA_TYPE");
      var isNullable = (schemaRow.GetAsString("IS_NULLABLE") == "YES");
      string typeSpec = string.Empty; 
      dataTypeString = dataTypeString.ToLowerInvariant();
      switch (dataTypeString) {
        case "set":
          return new DbTypeInfo(DbType.String, dataTypeString, isNullable, 255);
        case "enum":
          return new DbTypeInfo(DbType.String, dataTypeString, isNullable, 255);

        case "mediumint":
          return new DbTypeInfo(DbType.Int32, dataTypeString, isNullable);
        case "real":
        case "float":
          return new DbTypeInfo(DbType.Single, dataTypeString, isNullable);
        case "double":
          return new DbTypeInfo(DbType.Double, dataTypeString, isNullable);
        case "timestamp":
          return new DbTypeInfo(DbType.DateTime, dataTypeString, isNullable, 8);
        case "year":
          return new DbTypeInfo(DbType.Int16, dataTypeString, isNullable, 8);
        case "tinyblob":
          return new DbTypeInfo(DbType.Binary, dataTypeString, isNullable, byteSize);
        case "blob":
        case "longblob":
        case "mediumlob":
          return new DbTypeInfo(CustomDbTypes.VarBinary, dataTypeString, isNullable, byteSize);

        case "nvarchar":
        case "varchar":
          if (charSize == -1) {
            typeSpec = "nvarchar(max)";
          } else 
            typeSpec = string.Format("nvarchar({0})", charSize); 
          return new DbTypeInfo(DbType.String, typeSpec, isNullable, charSize);
        case "time":
          return new DbTypeInfo(DbType.Time, dataTypeString, isNullable, 8);
        case "tinytext":
          return new DbTypeInfo(DbType.String, dataTypeString, isNullable, 255);
        case "text":
          return new DbTypeInfo(DbType.String, dataTypeString, isNullable, -1);
        case "mediumtext":
        case "longtext":
          return new DbTypeInfo(CustomDbTypes.NText, dataTypeString, isNullable, -1);
        case "geometry":
          return new DbTypeInfo(CustomDbTypes.Geometry, "geometry", isNullable);
      }//switch
      return base.GetDbTypeInfo(schemaRow);
    }

  }//class
}

