﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data;
using System.Data.Linq.Mapping;
using System.Data.SqlServerCe;

using Vita.Common;
using Vita.Entities;
using Vita.Entities.Model;
using Vita.Data.Model;
using Vita.Data.Driver;

namespace Vita.Data.SqlCe {

  public class SqlCeDbDriver : DbDriver {
    public const DbOptions DefaultDbOptions = DbOptions.Default | DbOptions.AutoIndexForeignKeys;
    public const DbCapabilities DefaultCapabilities = DbCapabilities.ReferentialConstraints | DbCapabilities.Paging | DbCapabilities.BatchedUpdates
                                                    | DbCapabilities.OrderedColumnsInIndexes | DbCapabilities.NoIndexOnForeignKeys 
                                                    | DbCapabilities.ForceNullableMemo | DbCapabilities.NoMemoInDistinct
                                                    | DbCapabilities.CustomSelect;

    // constructor and initialization
    public SqlCeDbDriver(){
      this.Capabilities = DefaultCapabilities;
      this.DynamicSqlParameterPrefix = "@";
      base.CommandCallLogFormat = "EXEC {0} {1};";
    }

    public override IDbConnection CreateConnection(string connectionString) {
      return new SqlCeConnection(connectionString);
    }
    public override IDbCommand CreateDbCommand() {
      return new SqlCeCommand();
    }

    public override DbSchemaManager CreateSchemaManager(string connectionString) {
      return new SqlCeSchemaManager(this, connectionString);
    }
    public override DbTypeMapper CreateTypeMapper(DbSettings settings) {
      return new SqlCeTypeMapper(settings); 
    }
    public override DbSqlBuilder CreateSqlBuilder(DbModel dbModel) {
      return new SqlCeSqlBuilder(dbModel);
    }

    public override Vita.Data.Driver.LinqSqlProvider GetLinqSqlProvider() {
      return new SqlCeLinqSqlProvider();
    }

    public override void CheckDatabaseException(Exception exception, IDbCommand command = null) {
      var sqlEx = exception as SqlCeException;
      if (sqlEx == null) return;
      switch (sqlEx.NativeError) {
        case 25016:  //unique index violation
          var indexName = GetIndexName(sqlEx.Message); 
          throw new UniqueIndexViolationException(sqlEx.Message, indexName, sqlEx); 
      }
    }

    private string GetIndexName(string message) {
      try {
        // Sample message: A duplicate value cannot be inserted into a unique index. [ Table name = Publisher,Constraint name = IXU_Publisher_Name ]
        var constrNameTag = "Constraint name =";
        var tagPos = message.IndexOf(constrNameTag);
        var start = tagPos + constrNameTag.Length; 
        var end = message.IndexOf("]", start);
        var indexName = message.Substring(start, end - start).Trim();
        return indexName;
      } catch (Exception ex) {
        return "(failed to identify index in message: " + ex.Message + ")"; //
      }
    }
    //SQL CE has no schemas
    public override string GetFullName(string schema, string name) {
      return "[" + name + "]";
    }

    // We need to change SqlDbType for ntext parameters - otherwise it will fail
    public override IDbCommand CreateDbCommand(IDbConnection connection, DbCommandInfo commandInfo) {
      var command = base.CreateDbCommand(connection, commandInfo);
      for(int i=0; i< commandInfo.Parameters.Count; i++){
        var prmInfo = commandInfo.Parameters[i];
        if (prmInfo.TypeInfo.SqlTypeSpec == "ntext") {
          var prm = (SqlCeParameter) command.Parameters[i];
          prm.SqlDbType = SqlDbType.NText;
        }
      }
      return command; 
    }

    protected override IDbDataParameter CreateDbParameter(IDbCommand command, DbParamInfo prmInfo) {
      var typeInfo = prmInfo.TypeInfo;
      SqlCeParameter prm = null;
      switch (typeInfo.DbType) {
        case CustomDbTypes.VarBinary:
          prm = new SqlCeParameter(prmInfo.Name, SqlDbType.VarBinary, typeInfo.Size);
          break;
        case DbType.Currency:
          prm = new SqlCeParameter(prmInfo.Name, SqlDbType.Money);
          break;
        case CustomDbTypes.RowVersion:
          prm = new SqlCeParameter(prmInfo.Name, SqlDbType.Timestamp, 8);
          break;
        case CustomDbTypes.NText:
          prm = new SqlCeParameter(prmInfo.Name, SqlDbType.NText, -1);
          break;
        case CustomDbTypes.Text:
          prm = new SqlCeParameter(prmInfo.Name, SqlDbType.Text, -1);
          break;
        case CustomDbTypes.Image:
          prm = new SqlCeParameter(prmInfo.Name, SqlDbType.Image, -1);
          break;
        case CustomDbTypes.SmallMoney:
          prm = new SqlCeParameter(prmInfo.Name, SqlDbType.SmallMoney);
          break;
        case CustomDbTypes.SmallDateTime:
          prm = new SqlCeParameter(prmInfo.Name, SqlDbType.SmallDateTime);
          break;
        case DbType.Object:
          // should never happen
          Util.Throw("Object/Sql_variant Db type is not supported. Command/parameter: {0}/{1}.", command.CommandText, prmInfo.Name);
          break;
        default:
          return base.CreateDbParameter(command, prmInfo);
      }
      prm.Direction = prmInfo.Direction;
      if (prmInfo.SourceColumn != null)
        prm.SourceColumn = prmInfo.SourceColumn.ColumnName;
      prm.Direction = prmInfo.Direction;
      if (typeInfo.Precision > 0) prm.Precision = typeInfo.Precision;
      if (typeInfo.Scale > 0) prm.Scale = typeInfo.Scale;
      if (typeInfo.Size > 0) prm.Size = typeInfo.Size;
      prm.Value = prmInfo.DefaultValue;
      return prm;
    }

    protected override IDbDataParameter CloneDbParameter(IDbCommand command, IDbDataParameter parameter) {
      var sqlPrm = (SqlCeParameter)parameter;
      var result = new SqlCeParameter(parameter.ParameterName, sqlPrm.SqlDbType, sqlPrm.Size, sqlPrm.Direction,
                    sqlPrm.IsNullable, sqlPrm.Precision, sqlPrm.Scale, sqlPrm.SourceColumn, sqlPrm.SourceVersion, sqlPrm.Value);
      return result;
    }

    public override void OnDbModelConstructed(DbModel dbModel) {
      // NText columns must be nullable
      foreach (var table in dbModel.Tables) {
        foreach (var col in table.Columns) {
          if (col.IsMemo())
            col.Flags |= DbColumnFlags.Nullable;
        }
      }
      base.OnDbModelConstructed(dbModel);
    }
  }//class

}
