﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data;

using Vita.Common;
using Vita.Data.Model;
using Vita.Data.Driver;

namespace Vita.Data.SqlCe {

  public class SqlCeSchemaManager : DbSchemaManager {
    public SqlCeSchemaManager(DbDriver driver, string connectionString) : base(driver, connectionString) {  }

    public override DataTable GetSchemas() {
      return EmptyTable;
    }
    public override DataTable GetDatabases() {
      return EmptyTable;
    }
    public override DataTable GetProcedures(string schema = null) {
      return EmptyTable;
    }

    //This is an exception - standard INFORMATION_SCHEMA does not provide INDEXES view. SQL CE does provide it - nice! - but there's a problem.
    // Indexes view is in fact a list of index columns; so we have to use DISTINCT
    // NOTE schema is ignored - SQL CE does not support schemas
    public override DataTable GetIndexes(string schema = null) {
      var sql = @"
SELECT DISTINCT table_schema, table_name, index_name, primary_key, [unique], [clustered], 0 as [disabled]
FROM INFORMATION_SCHEMA.Indexes
ORDER BY table_name,  index_name
";  
      return ExecuteSelect(sql);
    }
    public override DataTable GetIndexColumns(string schema = null) {
      var sql = @"
SELECT table_schema, table_name, index_name, column_name, ordinal_position as column_ordinal_position, (collation - 1) as is_descending
FROM INFORMATION_SCHEMA.Indexes
ORDER BY table_name, index_name, ordinal_position
";
      return ExecuteSelect(sql);
    }


    public override void CreateSchema(string schema) {
      //SQL CE does not have schemas
    }
    public override void DropDatabase(string name) {
    }
    public override void DropSchema(string schema) {
    }

    //Columns: TABLE_CATALOG	TABLE_SCHEMA	TABLE_NAME	TABLE_TYPE  (table_type: "TABLE") 
    public override DataTable GetTables(string schema = null) {
      var sql = "SELECT * FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_TYPE='TABLE' ";
      return ExecuteSelect(sql);
    }

    //Columns: CONSTRAINT_CATALOG, CONSTRAINT_SCHEMA, CONSTRAINT_NAME, UNIQUE_CONSTRAINT_CATALOG, UNIQUE_CONSTRAINT_SCHEMA, 
    //  UNIQUE_CONSTRAINT_NAME, MATCH_OPTION, UPDATE_RULE, DELETE_RULE, C_TABLE, U_TABLE
    // Extended version with joins to bring Table names for both constraints
    // Overriding default to get rid of Schema comparisons
    public override DataTable GetReferentialConstraintsExt(string schema = null) {
      var sql = @"
        SELECT rc.*, tc1.TABLE_NAME as C_TABLE, tc2.TABLE_NAME AS U_TABLE 
        FROM INFORMATION_SCHEMA.REFERENTIAL_CONSTRAINTS rc
          INNER JOIN INFORMATION_SCHEMA.TABLE_CONSTRAINTS tc1 
            ON (tc1.CONSTRAINT_NAME = rc.CONSTRAINT_NAME)
          INNER JOIN INFORMATION_SCHEMA.TABLE_CONSTRAINTS tc2 
            ON (tc2.CONSTRAINT_NAME = rc.UNIQUE_CONSTRAINT_NAME)";
      return ExecuteSelect(sql);
    }

    //SqlCe does not allow NULL/NOT NULL spec for 'ntext' columns
    protected override string GetColumnSpec(DbColumnInfo column, DbColumnSpecOptions options) {
 //     if (column.IsMemo())
 //       options |= DbColumnSpecOptions.SuppressNull;
      return base.GetColumnSpec(column, options);
    }

    public override void ModifyColumn(DbColumnInfo column) {
      //SqlCe does not allow any modification of 'ntext'/memo columns
      if (column.IsMemo()) 
        return; 
      base.ModifyColumn(column);
    }
    public override void DropIndex(DbKeyInfo key) {
      ExecuteNonQuery("DROP INDEX {0}.\"{1}\";", key.Table.FullName, key.Name);
    }

    public override void RenameTable(DbTableInfo oldTable, DbTableInfo newTable) {
      ExecuteNonQuery("sp_rename '{0}', '{1}';", oldTable.TableName, newTable.TableName);
    }

    public override void DropSchemaObjects(string schema) {
      var constraints = this.GetTableConstraints(null);
      foreach (DataRow row in constraints.Rows)
        if (row.GetAsString("CONSTRAINT_TYPE") == "FOREIGN KEY") {
          var tableRef = row.GetAsString("TABLE_NAME");
          var fkName = row.GetAsString("CONSTRAINT_NAME");
          ExecuteNonQuery("ALTER TABLE {0} DROP CONSTRAINT \"{1}\"", tableRef, fkName);
        }
      //Delete tables
      var tables = this.GetTables(null);
      foreach (DataRow row in tables.Rows) {
        var tableName = row.GetAsString("TABLE_NAME");
        ExecuteNonQuery("DROP TABLE [{0}]", tableName);
      }
    }

    public override void OnColumnLoaded(DbModel model, DbColumnInfo column, DataRow columnRow) {
      //check identity flag; SQL CE provides an extra column in information_schema.columns view for identity flag
      var identityFlag = columnRow.GetAsInt("autoinc_increment");
      if (identityFlag == 1)
        column.Flags |= DbColumnFlags.Identity | DbColumnFlags.NoUpdate | DbColumnFlags.NoInsert;
    }

    public override void OnModelLoaded(DbModel model) {
      base.OnModelLoaded(model); 
    }
  }//class

}//namespace
