﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data;

using Vita.Common;
using Vita.Entities;
using Vita.Entities.Model;
using Vita.Data.Model;

namespace Vita.Data.SqlCe {
  public class SqlCeTypeMapper : DbTypeMapper {

    public SqlCeTypeMapper(DbSettings settings) : base(settings) { }

    public override DbTypeInfo GetDbTypeInfo(Type dataType, int size = 0, DbType? explicitDbType = null, EntityMemberFlags flags = EntityMemberFlags.None, byte precision = 0, byte scale = 0) {
      var isNullable = flags.IsSet(EntityMemberFlags.Nullable);
      if (dataType == typeof(byte[])) {
        switch (explicitDbType) {
          case CustomDbTypes.NText: return new DbTypeInfo(CustomDbTypes.Image, "image", isNullable, Int16.MaxValue);
          case CustomDbTypes.VarBinary: return new DbTypeInfo(CustomDbTypes.VarBinary, "varbinary({0})", isNullable, Int16.MaxValue);
          //We treat rowversion in SQL CE as just binary field. SQL CE does support rowversion, but using it actually does not seem to make much sense - SQL CE is local, singlue-user database
          // maybe in the future?
          case CustomDbTypes.RowVersion: return new DbTypeInfo(DbType.Binary, "binary(8)", false, 8);
          default:
            return new DbTypeInfo(DbType.Binary, "varbinary({0})", isNullable, size);
        }
      }
      // Go thru type code, but first unwrap nullable value type
      if (dataType.IsNullableValueType())
        dataType = dataType.GetValueTypeFromNullable();
      //First special cases
      if (dataType == typeof(DateTimeOffset))
        return GetDatetimeOffsetTypeInfo(isNullable);
      // Go thru type code
      var typeCode = Type.GetTypeCode(dataType);
      switch (typeCode) {
        case TypeCode.DateTime:
          DbValueConverter readTime = null;
          if (flags.IsSet(EntityMemberFlags.Utc))
            readTime = ReadUtcTime;
          return new DbTypeInfo(DbType.DateTime, "datetime", isNullable, readConverter: readTime);

        //Note that tinyint in UNSIGNED byte!
        case TypeCode.Byte: //signed byte
          return new DbTypeInfo(DbType.Byte, "tinyint", isNullable);
        case TypeCode.SByte: //signed byte
          return new DbTypeInfo(DbType.Byte, "tinyint", isNullable, 1, null, 0, 0, x => (sbyte)(byte)x, x => (byte)(sbyte)x);
        case TypeCode.UInt16:
          return new DbTypeInfo(DbType.UInt16, "smallint", isNullable, 2, null, 0, 0, x => (UInt16)(Int16)x, x => (Int16)(UInt16)x);
        case TypeCode.UInt32:
          return new DbTypeInfo(DbType.UInt32, "int", isNullable, 4, null, 0, 0, x => (UInt32)(Int32)x, x => (Int32)(UInt32)x);
        case TypeCode.UInt64:
          return new DbTypeInfo(DbType.UInt64, "bigint", isNullable, 4, null, 0, 0, x => (UInt64)(Int64)x, x => (Int64)(UInt64)x);
        case TypeCode.String:
          if (size < 0) {
            switch (explicitDbType) {
              case CustomDbTypes.NText:
                return new DbTypeInfo(CustomDbTypes.NText, "ntext", isNullable, -1);
              case CustomDbTypes.Text:
                return new DbTypeInfo(CustomDbTypes.Text, "text", isNullable, -1);
              default:
                return new DbTypeInfo(DbType.String, "ntext", isNullable, -1);
            }
          } else
            return base.GetDbTypeInfo(dataType, size, explicitDbType, flags, precision, scale);

        case TypeCode.Decimal:
          if (precision <= 0) precision = 18;
          if (scale <= 0) scale = 4;
          switch (explicitDbType) {
            case DbType.Currency: return new DbTypeInfo(DbType.Currency, "money", isNullable, size, null, precision, scale);
            default:
              return new DbTypeInfo(DbType.Decimal, "numeric({1},{2})", isNullable, size, null, precision, scale);
          }
        default:
          return base.GetDbTypeInfo(dataType, size, explicitDbType, flags, precision, scale);
      }//switch
    }

    // In SQL CE datetimeoffset is mapped to nvarchar(34)
    private DbTypeInfo GetDatetimeOffsetTypeInfo(bool isNullable) {
      return new DbTypeInfo(DbType.String, "nvarchar({0})", isNullable, 34,
        readConverter: SqlCeUtility.StringToDateTimeOffset, writeConverter: SqlCeUtility.DateTimeOffsetToString);
    }

    public override DbTypeInfo GetDbTypeInfo(DataRow schemaRow) {
      var dataTypeString = schemaRow.GetAsString("DATA_TYPE");
      var isNullable = (schemaRow.GetAsString("IS_NULLABLE") == "YES");
      string typeSpec = string.Empty;
      dataTypeString = dataTypeString.ToLowerInvariant();
      switch (dataTypeString) {
        case "tinyint":
          return new DbTypeInfo(DbType.Byte, dataTypeString, isNullable);
        case "datetime2":
          return new DbTypeInfo(DbType.DateTime2, dataTypeString, isNullable);
        case "datetimeoffset":
          return GetDatetimeOffsetTypeInfo(isNullable);
        case "money":
          return new DbTypeInfo(DbType.Currency, dataTypeString, isNullable);
        case "nvarchar":
          var size = schemaRow.GetAsInt("CHARACTER_MAXIMUM_LENGTH");
          if (size == -1) {
            typeSpec = "nvarchar(max)";
          } else
            typeSpec = string.Format("nvarchar({0})", size);
          return new DbTypeInfo(DbType.String, typeSpec, isNullable, size);
      }//switch
      return base.GetDbTypeInfo(schemaRow);
    }

  }//class
}
