﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Globalization;
using System.Threading;
using System.Data;

using Vita.Common;
using Vita.Data.Model;

namespace Vita.Data.SqlCe {

  public static class SqlCeUtility {
    internal static bool IsMemo(this DbColumnInfo column) {
      return column.TypeInfo.DbType == DbType.String && column.TypeInfo.Size < 0;
    }


    
    // In SQL CE database, DateTimeOffset is represented by nvarchar(34). To allow for proper sorting, we store UTC datetime (not local)
    // followed by offset. However, if we try to use a DateTimeOffset constant in a LINQ query, it will be converted to string (nvachar)
    // using DateTimeOffset.ToString() method. There are two problems with this method.
    // First, the result is a local date-time (!) followed by offset. So to be compatible with database, we need to "shift" the value 
    // to London, so that local time is UTC time, so it will appear in ToString() representation.
    // Second, the DateTimeOffset.ToString uses ShortDatePattern to format the date, which results in 'mm/dd/yyyy' in US locale.
    // What we need is 'yyyy/mm/dd'. The only way to do this is to change short datetime format on current culture on current thread.
    private static CultureInfo _adjustedCulture;
    public static void AdjustDefaultDateTimeFormatForSqlCe() {
      var currThread = Thread.CurrentThread;
      var currCulture = currThread.CurrentCulture;
      if (currCulture == _adjustedCulture) return;
      if (_adjustedCulture == null) {
        _adjustedCulture = currCulture.Clone() as CultureInfo;
        var fmtInfo = _adjustedCulture.DateTimeFormat;
        fmtInfo.ShortDatePattern = "yyyy-MM-dd";
        fmtInfo.LongTimePattern = "HH:mm:ss.fffffff";
      } 
      currThread.CurrentCulture = _adjustedCulture;
    }

    //Use this method in LINQ queries when using constant DateTimeOffset value, when your target database is SQL CE
    public static DateTimeOffset AdjustDateTimeOffsetForLinq(DateTimeOffset value) {
      AdjustDefaultDateTimeFormatForSqlCe(); 
      return value.Subtract(value.Offset);
    }

    // The following two methods are used as read/write converters for SQL parameters
    //Note: default ToString() representation of DatetimeOffset is local time (!) followed by offset.
    // To make values comparable in database, we use UtcTime in string representation, so we do our custom parsing
    public static object StringToDateTimeOffset(object value) {
      if (value == null || value == DBNull.Value) return DBNull.Value;
      var svalue = (string)value;
      if (string.IsNullOrWhiteSpace(svalue)) return DBNull.Value; 
      var datePart = svalue.Substring(0, 27);
      var offsPart = svalue.Substring(28, 6);
      var utcDate = DateTime.ParseExact(datePart, "yyyy-MM-dd HH:mm:ss.fffffff", null);
      var offs = TimeSpan.Parse(offsPart);
      var localDate = utcDate.Add(offs);
      var result = new DateTimeOffset(localDate, offs); //expects local time
      return result;
    }
    public static object DateTimeOffsetToString(object value) {
      //Format: '"YYYY-MM-DD hh:mm:ss.nnnnnnn [+/-] hh:mm"'
      if (value == null || value == DBNull.Value) 
        return null;
      DateTimeOffset dtOffset;
      if (value is DateTimeOffset?) {
        var temp = (DateTimeOffset?)value;
        if (temp == null)
          return null;
        dtOffset = temp.Value;
      } else if (value is DateTimeOffset)
        dtOffset = (DateTimeOffset)value;
      else {
        Util.Throw("DateTimeOffset converterter: invalid property value. Expected DateTimeOffset, found: {0} (type: {1})", value, value.GetType().Name);
        return null; 
      }
      //Format the value
      var dtPart = dtOffset.UtcDateTime.ToString("yyyy-MM-dd HH:mm:ss.fffffff");
      var offsPart = dtOffset.Offset.ToString(@"hh\:mm");
      var sign = (dtOffset.Offset.Hours >= 0) ? "+" : "-";
      var result = dtPart + " " + sign + offsPart;
      return result;
    }




  }//class
}
