﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Vita.Entities.Runtime;
using Vita.Common;
using Vita.Entities.Model;

namespace Vita.Modules.ChangeTracking {

  public interface IChangeTrackingLogSerializer {
    string Serialize(IList<EntityChangeInfo> changes);
    IList<EntityChangeInfo> Deserialize(string text, EntityModelInfo model);
  }

  //Serializers/deserializes changes into a plain multiline text separated with NewLine. 
  /// <summary> Serializers/deserializes changes into a plain multiline text.</summary>
  /// <remarks>
  /// Each line is an entity change in the format: {EntityFullName}|{PrimaryKeyValue}|{ChangeType}
  /// For example: 'books.Book|0f235feb-aff9-11e1-8f87-f0def1783701|New'
  /// </remarks>
  public class ChangeLogSimpleSerializer : IChangeTrackingLogSerializer {

    public string Serialize(IList<EntityChangeInfo> changes) {
      var result = string.Join(Environment.NewLine, changes.Select(
          change => Util.SafeFormat("{0}|{1}|{2}", 
            change.EntityName, change.KeyValues, change.ChangeType)));
      return result; 
    }

    public IList<EntityChangeInfo> Deserialize(string text, EntityModelInfo model) {
      var result = new List<EntityChangeInfo>(); 
      var delimArray = new string[] { "\r\n" };
      var arrChanges = text.Split(delimArray, StringSplitOptions.RemoveEmptyEntries);
      foreach (var changeLine in arrChanges) {
        var temp = changeLine.Split('|');
        var entName = temp[0];
        var strPk = temp[1];
        var changeType = ParseChangeType(temp[2]);
        var entInfo = model.GetEntityInfo(entName, true);
        var pk = EntityKey.KeyFromString(entInfo.PrimaryKey, strPk);
        var trackChange = new EntityChangeInfo(pk, changeType);
        result.Add(trackChange);
      }
      return result; 
    }

    private static EntityChangeType ParseChangeType(string typeString) {
      EntityChangeType result;
      Util.Check(Enum.TryParse<EntityChangeType>(typeString, out result),
        "TrackingChangeLog parsing error: invalid status value {0}.", typeString);
      return result;
    }
  
  }//class
}
