﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Vita.Entities;
using Vita.Entities.Runtime;
using Vita.Common;
using Vita.Entities.Model;
using Vita.Services;

namespace Vita.Modules.ChangeTracking {

  public interface IChangeTrackingService {
    void RegisterLog(ChangeTrackingLog log);
    DateTime GetFirstEntryDateTime(IEntitySession session, ChangeTrackingLog log);
    List<EntityChangeInfo> GetChanges(IEntitySession session, ChangeTrackingLog log, Guid partition, DateTime fromDateTime, DateTime? tillDateTime = null);
  }


  public class ChangeTrackingModule : EntityModule, IChangeTrackingService {
    public readonly List<ChangeTrackingLog> ChangeLogs = new List<ChangeTrackingLog>();
    public IChangeTrackingLogSerializer ChangeLogSerializer;

    //private IEntityCommand _getChangeEntries, _getFirstEntry;
    private ITimeService _timeService; 

    public ChangeTrackingModule(EntityArea area, string moduleName, IChangeTrackingLogSerializer serializer = null) : base(area, moduleName) {
      ChangeLogSerializer = serializer ?? new ChangeLogSimpleSerializer();
      RegisterEntities(typeof(IChangeTrackEntry));
      Model.RegisterService(typeof(IChangeTrackingService), this);
/*
      // compile commands
      var entries = EntitySet<IChangeTrackEntry>();
      var qryGetChanges = from cl in entries
                          where cl.LogName == StringParam(50) && cl.Partition == Param<Guid>() &&
                          cl.CreatedOn > Param<DateTime>() && cl.CreatedOn <= Param<DateTime>()
                          orderby cl.CreatedOn descending
                          select cl;
      _getChangeEntries = CreateSelect<IChangeTrackEntry>("ChangeLogSelectByDatePartition", "Selects change tracking records.", qryGetChanges);
      var qryGetFirstEntry = (from cl in entries
                              where cl.LogName == StringParam(50)
                              orderby cl.CreatedOn
                              select cl).Take(1);
      _getFirstEntry = CreateSelect<IChangeTrackEntry>("ChangeLogGetFirstEntry", "Finds the first entry in the change log identified by log name.", qryGetFirstEntry);
 */ 
    }

    public override void Activate(AppActivationContext context) {
      base.Activate(context);
      _timeService = EntityStore.GetService<ITimeService>();
      var entStore = (EntityStore)this.EntityStore;
      entStore.Events.SavingChanges += EntitySession_Saving;
      foreach (var log in ChangeLogs)
        log.Init(this); 
    }

    void EntitySession_Saving(EntitySession session, EventArgs args) {
      foreach (var log in ChangeLogs)
        log.OnEntitySessionSaving(session); 
    }

    #region Custom commands
    internal IList<IChangeTrackEntry> GetChangeLogEntries(IEntitySession session, ChangeTrackingLog log, Guid partition,
                DateTime fromDateTime, DateTime tillDateTime) {
      Util.Check(log != null, "Parameter 'log' may not be null.");
      var entries = session.EntitySet<IChangeTrackEntry>();
      var qryGetChanges = from cl in entries
                          where cl.LogName == log.Name && cl.Partition == partition &&
                          cl.CreatedOn > fromDateTime && cl.CreatedOn <= tillDateTime
                          orderby cl.CreatedOn descending
                          select cl;
      return qryGetChanges.ToList();
      //return session.ExecuteQuery<IChangeTrackEntry>(_getChangeEntries, log.Name, partition, fromDateTime, tillDateTime);
    }

    internal IChangeTrackEntry GetFirstEntry(IEntitySession session, ChangeTrackingLog log) {
      Util.Check(log != null, "Parameter 'log' may not be null.");
      var entries = session.EntitySet<IChangeTrackEntry>();
      var qryGetFirstEntry = (from cl in entries
                              where cl.LogName == log.Name
                              orderby cl.CreatedOn
                              select cl);
      return qryGetFirstEntry.FirstOrDefault();
      
      // return session.ExecuteQuery<IChangeTrackEntry>(_getFirstEntry, log.Name).FirstOrDefault();
    }

    #endregion


    #region IChangeTrackingService Members
    public void RegisterLog(ChangeTrackingLog log) {
      ChangeLogs.Add(log);
      //If we are registering log after 
      if (this.EntityStore != null)
        log.Init(this);
    }
    
    public DateTime GetFirstEntryDateTime(IEntitySession session, ChangeTrackingLog log) {
      var entry = GetFirstEntry(session, log);
      if (entry != null)
        return entry.CreatedOn;
      //If no entries, create anchor entry
      // Open another session and save it independently
      var tempSession = this.EntityStore.OpenSession();
      var newEntry = log.CreateLogEntry(tempSession, Guid.Empty, Guid.Empty, string.Empty, ChangleLogEntryType.Base);
      tempSession.SaveChanges(); 
      return newEntry.CreatedOn;
    }

    public List<EntityChangeInfo> GetChanges(IEntitySession session, ChangeTrackingLog log, Guid partition, DateTime fromDateTime, DateTime? tillDateTime = null) {
      var now = _timeService.Now;
      DateTime tillDateTimeValue;
      if (tillDateTime != null)
        tillDateTimeValue = tillDateTime.Value;
      else {
        tillDateTimeValue = now;
      }
      // Get raw changes, analyse them - remove/combine entries for the same entity
      var rawChanges = GetRawChanges(session, log, partition, fromDateTime, tillDateTimeValue);
      var changeTable = new Dictionary<string, EntityChangeInfo>(); //temp table
      //Note that changes are returned in CreatedOn/desc order. So the first ones in the list are the latest.
      foreach (var change in rawChanges) {
        var fullKey = change.GetFullKey();
        EntityChangeInfo laterChange;
        if (!changeTable.TryGetValue(fullKey, out laterChange)) {
          //if not seen yet, just add it
          changeTable.Add(fullKey, change);
          continue;
        }
        //we have a "later" change for this change. Let's analyse later status
        switch (laterChange.ChangeType) {
          case EntityChangeType.Delete:
            //ignore this change, record will be deleted anyway
            break;
          case EntityChangeType.Update:
            //check this change status - if it is New, then replace the later (modified)  with this new
            if (change.ChangeType == EntityChangeType.New)
              changeTable[fullKey] = change;
            break;
        }//switch
      }//foreach change

      //Build the result changeset
      var allChanges = changeTable.Values.ToList();
      return allChanges;
    }


    private static EntityChangeType ParseStatus(string status, Guid changeEntryId) {
      EntityChangeType result;
      Util.Check(Enum.TryParse<EntityChangeType>(status, out result),
        "TrackingChangeLog parsing error: invalid status value {0} in entry ID={1}.", status, changeEntryId);
      return result;
    }

    private List<EntityChangeInfo> GetRawChanges(IEntitySession session, ChangeTrackingLog log, Guid partition, DateTime fromDateTime, DateTime tillDateTime) {
      var entStore = (EntityStore)this.EntityStore; 
      var modelInfo = entStore.Model.ModelInfo;
      var result = new List<EntityChangeInfo>();
      var logEntries = GetChangeLogEntries(session, log, partition, fromDateTime, tillDateTime);
      foreach (var logEntry in logEntries) {
        var changesText = logEntry.Changes;
        if (string.IsNullOrEmpty(changesText)) continue;
        var changes = this.ChangeLogSerializer.Deserialize(changesText, modelInfo);
        foreach (var change in changes)
          result.Add(change);
      }
      return result;
    } //method


    #endregion

  }//class


}
