﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Vita.Entities.Model;
using Vita.Entities;
using Vita.Common;
using Vita.Entities.Runtime;
using System.Linq.Expressions;

namespace Vita.Modules.ChangeTracking {

  public abstract class DataGrouper {
    public abstract Type[] GetTypes(); 
    public abstract Funclet GetGrouperFunclet(Type entityType);
  }

  /// <summary>Defines a set of rules that classify entity instances into groups. </summary>
  /// <typeparam name="TGroup">The data type for group ID.</typeparam>
  /// <remarks>
  /// <para>Data grouper defines classification of data into separate sets belonging together according to some criteria.
  /// Once the grouper is constructed, it can be used to compute the group that a record belongs to. 
  /// Note that data groupers define sets of entity instances (records), not entity types/tables.</para>
  /// <para>For example, an application can define Department grouper, and then use it for different entities (employees, equipment, documents, etc)
  /// to determine the department it belongs to. 
  /// This grouper then can be used in Authorization module to restrict user access to records related to his/her department.
  /// Given a current user, and an entity (a document), we can use the DepartmentGrouper to retrieve document's ownership department. 
  /// Then authorization code would match it to user's department and allow or deny the access based on the match result.</para>
  /// <para>VITA authorization services rely on data grouper definitions provided by application to match users with data available to them. 
  /// The other use of data groups is in data sync service to classify the data changes in the central database into separate sets 
  /// that belong to different clients, to allow client machines to sync only the data subsets that are relevant to them.
  /// </para>
  /// <para>You setup the data grouper by providing a test method for each entity type: a function that returns a group (usually group ID) for a 
  /// given entity:</para>
  /// <code>
  ///   var depGroup = new DataGrouper&lt;Guid&gt;("DepartmentGroup");
  ///   depGroup.AddEntity&lt;IEmployee&gt;(empl => empl.Department.Id); // returns a department ID for a given employee.
  ///   depGroup.AddEntity&lt;IPurchaseOrder&gt;(order => order.Department.Id); // returns a department ID for a given purchase order document.
  ///   depGroup.AddEntity&lt;IUser&gt;(user => user.Employee.Department.Id); //returns a department ID for a given user.
  /// </code>
  /// </remarks>
  public class DataGrouper<TGroup> : DataGrouper where TGroup: IEquatable<TGroup> {

    public readonly string Name;
    private Dictionary<Type, Funclet> _funcs = new Dictionary<Type, Funclet>();  

    public DataGrouper(string name) {  
      Name = name; 
    }

    #region DataGrouper overrides
    public override Type[] GetTypes() {
      return _funcs.Keys.ToArray();
    }

    public override Funclet GetGrouperFunclet(Type entityType) {
      Funclet funcObject;
      _funcs.TryGetValue(entityType, out funcObject);
      return funcObject;
    }

    #endregion

    public void AddEntity<TEntity>(Expression<Func<TEntity, TGroup>> expr = null) where TEntity : class {
      var entType = typeof(TEntity);
      Util.Check(!_funcs.ContainsKey(entType), "Duplicate entries in DataGrouper {0} for entity {1}.", Name, entType);
      var func = new Funclet<TEntity, TGroup>(expr);
      _funcs.Add(entType, func);
    }

    public void Add(Type type) {
      var func = new Funclet<object, TGroup>( x => default(TGroup)); 
      _funcs.Add(type, func);
    }


    public TGroup GetGroup(EntityRecord record) {
      Util.Check(record != null, "Record argument is null. Grouper: {0}", Name);
      var funcObject = GetGrouperFunclet(record.EntityInfo.EntityType);
      if (funcObject == null) 
        return default(TGroup); 
      try {
        var result = funcObject.Evaluate(record.EntityInstance);
        return (TGroup)result;
      } catch (Exception ex) {
        Util.Rethrow(ex, @"Error evaluating group: {0}. Grouper: {1}, Record: {2}, FuncObject: {3}", ex.Message, Name, record, funcObject);
        throw; //never happens
      }
    }

    public TGroup GetGroup<TEntity>(TEntity entity) {
      var funcObject = GetGrouperFunclet(typeof(TEntity));
      if (funcObject == null) 
        return default(TGroup);
      try {
        var result = funcObject.Evaluate(entity);
        return (TGroup)result;
      } catch (Exception ex) {
        Util.Rethrow(ex, @"Error evaluating group: {0}. Grouper: {1}, Entity: {2}, FuncObject: {3}", ex.Message, Name, entity, funcObject);
        throw; //never happens
      }
    }

    public bool IsRegisteredType(Type entityType) {
      return _funcs.ContainsKey(entityType); 
    }

    public Dictionary<TGroup, IList<EntityRecord>> GroupRecords(IList<EntityRecord> records) {
      var result = new Dictionary<TGroup, IList<EntityRecord>>();
      for (int i = 0; i < records.Count; i++) {
        var rec = records[i];
        var groupId = GetGroup(rec);
        //Add record to appropriate group
        IList<EntityRecord> list;
        if (!result.TryGetValue(groupId, out list)) {
          list = new List<EntityRecord>();
          result[groupId] = list;
        }
        list.Add(rec);
      }
      return result;
    }//method

  }//class
}//ns
