﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Vita.Entities;

namespace Vita.Modules.ChangeTracking {

  // Base entry is an entry with empty 'Changes' column. 
  // It is used to mark the moment when change log started, when no changes are written yet. 
  // Data sync service relies on at least one entry to mark the moment when first snapshot was 
  // created. 
  // When entities do not change, there are not entries in change log. So by default, if application requests a sync set, 
  // a full snapshot should be returned. This snapshot will be returned again and again on repeated requests for sync set.
  // To avoid this, we create a Base changelog entry automatically when there are no entries
  public enum ChangleLogEntryType {
    Base = 0, //no changes, just marks the start of change log
    Changes = 1,
  }
  // ChangeTrackingLog table is used for recording all ongoing updates to a set of entities, optionally segregated by data group id
  // When list of changes is requested by the remote agent, system retrieves records in this table, filtering by log name and partition.
  // Filtering by partition is used when remote party is interested only in subset of changes (belonging to particular customer). 
  // In this case system can quickly find all changes in records related to the customer/client by filtering on partition. 
  [Entity, OrderBy("CreatedOn:DESC"), ClusteredIndex("CreatedOn,Id"), IsSystem, Dynamic, Index("LogName,Partition")]
  // Dynamic means that records must be discarded from session if transaction aborts or commits
  public interface IChangeTrackEntry {
    [Auto, PrimaryKey]
    Guid Id { get; set; }

    [Auto(AutoType.CreatedOn), Utc]
    DateTime CreatedOn { get; set; }

    [Size(50)]
    string LogName { get; set; }
    
    Guid TransactionId { get; set; }

    ChangleLogEntryType EntryType { get; set; }

    [OldNames("DataGroupId")]
    Guid Partition { get; set; }

    [Memo, Nullable]
    string Changes { get; set; }
  }

}
