﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Xml;
using System.Xml.Serialization;

using Vita.Common;
using Vita.Entities.Runtime;
using Vita.Entities.Model;
using Vita.Modules.ChangeTracking;

namespace Vita.Modules.DataSync {

  [Serializable]
  public class DataSyncResult {
    public Guid SyncId;
    public SyncStatus Status;
    public int RecordCount; 
    public string Details;

    //public DataSyncResult() { } // needed for deserialization
    public DataSyncResult(Guid syncId, SyncStatus status, int recordCount, string details = null) {
      SyncId = syncId;
      Status = status;
      RecordCount = recordCount;
      Details = details;
    }
  }

  [Serializable]
  public class EntitySyncInfo {
    public string EntityName;
    public EntityChangeType ChangeType;
    public string KeyValues;
    public EntityRecord Record; // null for changeType delete

    public EntitySyncInfo() {}

    public EntitySyncInfo(string entityName, EntityChangeType changeType, string keyValues, EntityRecord record = null) {
      EntityName = entityName;
      ChangeType = changeType;
      KeyValues = keyValues;
      Record = record; 
    }
    // Used for full snapshots - so all records are treated as NEW
    public static EntitySyncInfo CreateForFullSync(EntityRecord record) {
      return new EntitySyncInfo(record.EntityInfo.FullName, EntityChangeType.New, record.PrimaryKey.ValuesToString(), record); 
    }
    public override string ToString() {
      return EntityName + "/" + KeyValues + "/" + ChangeType;
    }
  }

  [Serializable]
  public class EntitySyncGroup {
    public string EntityName;
    public bool IsFullSnapshot;
    public List<EntitySyncInfo> SyncChanges = new List<EntitySyncInfo>();

    public EntitySyncGroup() { }
    public EntitySyncGroup(string entityName, bool isFullSnapshot) {
      EntityName = entityName;
      IsFullSnapshot = isFullSnapshot;
    }
    public override string ToString() {
      var str = EntityName + "/" + SyncChanges.Count + " records;";
      if (IsFullSnapshot)
        str += " (Full snapshot)";
      return str;         
    }
  }

  public class EntitySnapshotSet {
    public string EntityName;
    public List<EntityRecord> Records = new List<EntityRecord>();

    public EntitySnapshotSet(string entityName) {
      EntityName = entityName;
    }
    public override string ToString() {
      return EntityName + "/Count=" + Records.Count;
    }
  }


  /// <summary>
  /// Encapsulates all changes for entities in a certain date range.
  /// </summary>
  [Serializable]
  public class DataSyncSet {
    public Guid Id; // == sync log entry ID
    public string Source;
    public DateTimeOffset Since;
    public DateTimeOffset Until;
    public List<EntitySyncGroup> SyncGroups = new List<EntitySyncGroup>();

    public DataSyncSet() { } //needed for deserialization

    public DataSyncSet(DateTimeOffset since, DateTimeOffset until) : this() {
      Id = Guid.NewGuid(); 
      Since = since;
      Until = until;
    }
    public int GetRecordCount() {
      return SyncGroups.Sum(g => g.SyncChanges.Count);
    }
    public override string ToString() {
      return Id + "/" + Since + ".." + Until + "/GroupCount=" + SyncGroups.Count; 
    }
    /*
    #region IXmlSerializable Members

    public System.Xml.Schema.XmlSchema GetSchema() {
      return null;
    }

    public void ReadXml(XmlReader reader) {
      reader.SkipWs();
      reader.ReadStartElement(); //read root 
      reader.SkipWs();
      this.Id = reader.ReadElementContentAsGuid();
      reader.SkipWs();
      this.MachineId = reader.ReadElementContentAsGuid();
      reader.SkipWs();
      this.Since = reader.ReadElementContentAsDateTime().CheckKind(DateTimeKind.Utc);
      reader.SkipWs();
      this.Until = reader.ReadElementContentAsDateTime().CheckKind(DateTimeKind.Utc);
      reader.SkipWs();
      this.IsFullSnapshot = reader.ReadElementContentAsBoolean();
      reader.SkipWs();

      var model = EntityModel.Current;
      Util.Check(model != null, "EntityModel.Current was not set before deserialization - cannot deserialize records."); 
      //Read inserts
      this.InsertsUpdates = reader.ReadRecordList(model, "InsertsUpdates");
      //Read deletes
      this.Deletes = reader.ReadRecordKeyList(model, "Deletes");
      //read root end
      reader.ReadEndElement(); 
    }

    public void WriteXml(XmlWriter writer) {
      //Writer has already root element written - that's the way Xml serializer works
      writer.WriteElementString("Id", Id.ToString());
      writer.WriteElementString("MachineId", MachineId.ToString());
      writer.WriteElementString("FromDateTime", this.Since.ToString("s"));
      writer.WriteElementString("TillDateTime", Until.ToString("s"));
      writer.WriteElementString("IsFullSnapshot", (this.IsFullSnapshot ? "true" : "false"));
      writer.WriteRecordList("InsertsUpdates", this.InsertsUpdates);
      writer.WriteRecordKeyList("Deletes", this.Deletes);
    }
    #endregion
    */
  } //class

}
