﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using System.Text;

using Vita.Common;
using Vita.Entities;
using Vita.Entities.Runtime;
using Vita.Entities.Model;
using Vita.Data;

namespace Vita.Modules.DataSync {

  public abstract class DataSnapshotLoader {
    public abstract List<EntitySnapshotSet> LoadSnapshots(IEntitySession session, object partitionKey);
  }//class


  //Loads all records from all entities of specified types; ignores partition
  public class FullSnapshotLoader : DataSnapshotLoader {
    public HashSet<Type> EntityTypes = new HashSet<Type>(); 

    public FullSnapshotLoader(IEnumerable<Type> types) {
      EntityTypes.UnionWith(types); 
    }
    public FullSnapshotLoader(params Type[] types) {
      EntityTypes.UnionWith(types);
    }
    public override List<EntitySnapshotSet> LoadSnapshots(IEntitySession session, object partitionKey) {
      var result = new List<EntitySnapshotSet>(); 
      var entSession = session as EntitySession; 
      var model = entSession.EntityStore.Model;
      //var paging = new Paging(skip: 0, take: 100 * 1000);
      foreach (var type in EntityTypes) {
        var entInfo = model.ModelInfo.GetEntityInfo(type, true);
        var snapshot = new EntitySnapshotSet(entInfo.FullName);
        var recs = entSession.GetRecords(entInfo);
        snapshot.Records.AddRange(recs);
        result.Add(snapshot); 
      }
      return result; 
    }
  } //class

  /// <summary> Loads partitioned set of entities - snapshot for a specific partition value. </summary>
  /// <typeparam name="TPartition">Partitioning key type.</typeparam>
  public class PartitionedSnapshotLoader<TPartition> : DataSnapshotLoader {
    public event EventHandler<SnapshotLoadEventArgs<TPartition>> LoadingShapsot;

    public override List<EntitySnapshotSet> LoadSnapshots(IEntitySession session, object partition) {
      var args = new SnapshotLoadEventArgs<TPartition>(session, (TPartition) partition);
      if (LoadingShapsot != null)
        LoadingShapsot(this, args);
      return args.Snapshots;
    }
  }//class

  public class SnapshotLoadEventArgs<TPartition> : EventArgs {
    public readonly IEntitySession Session;
    public readonly TPartition Partition; 
    public readonly List<EntitySnapshotSet> Snapshots = new List<EntitySnapshotSet>();
    private EntityModelInfo _modelInfo; 

    public SnapshotLoadEventArgs(IEntitySession session, TPartition partition) {
      Session = session;
      Partition = partition;
      var entSession = session as EntitySession;
      _modelInfo = entSession.EntityStore.Model.ModelInfo; 
    }

    public void Add<TEntity>(IList<TEntity> entities) {
      var entInfo = _modelInfo.GetEntityInfo(typeof(TEntity), true);
      var snapshot = new EntitySnapshotSet(entInfo.FullName); 
      foreach(var ent in entities)
        snapshot.Records.Add(EntityHelper.GetRecord(ent));
      Snapshots.Add(snapshot); 
    }

    public void Load<TEntity>(Expression<Func<TEntity, bool>> filter) where TEntity: class {
      this.Add(Session.EntitySet<TEntity>().Where(filter).ToList());

    }
  }//class
}
