﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Vita.Common;
using Vita.Entities.Runtime;
using Vita.Entities.Model;
using Vita.Modules.ChangeTracking;
using Vita.Entities;

namespace Vita.Modules.DataSync {

  public abstract class DataSyncChannel {
    public string Name;
    public readonly Dictionary<Type, ClearValueRule> ClearValueRules = new Dictionary<Type, ClearValueRule>();

    internal DataSyncChannel(string name) {
      Name = name;
    }

    public void AddClearValueRule<TEntity>(params string[] propertyNames) {
      AddClearValueRule(typeof(TEntity), propertyNames);
    }
    public override string ToString() {
      return Name;
    }
    public void AddClearValueRule(Type entityType, string[] propertyNames) {
      Util.Check(!ClearValueRules.ContainsKey(entityType), "Duplicate reset value rule for entity type {0}, change log {1}", entityType.Name, this.Name);
      ClearValueRules[entityType] = new ClearValueRule(entityType, propertyNames);
    }

    public void ApplyClearValueRules(EntityRecord record) {
      if (!_rulesInitialized)
        InitRules(record.Session.EntityStore);
      ClearValueRule rule;
      if (!ClearValueRules.TryGetValue(record.EntityInfo.EntityType, out rule)) return;
      //We have a rule to reset smth in this record
      foreach (var member in rule.Members) {
        //record.SetValue(member, null); -- this goes thru auth check, and will fail - auth treats it as an attempt to modify the record
        member.SetValueRef(record, member, member.DeniedValue);
      }
    }

    private bool _rulesInitialized; 
    private void InitRules(EntityStore entityStore) {
      foreach (var rule in ClearValueRules.Values)
        rule.Init(this.Name, entityStore);
      _rulesInitialized = true; 
    }//method


  }//class

  [Flags]
  public enum ImportSyncActions {
    None = 0,

    Insert = 1,
    Update = 1 << 1,
    Delete = 1 << 2,

    All = Insert | Update | Delete,
  }

  public class ImportSyncChannel : DataSyncChannel {
    public readonly HashSet<Type> ImportTypes = new HashSet<Type>();
    public readonly ImportSyncActions Actions;
    // Snapshot loader is used for importing full snapshot when Delete action is allowed. 
    // In this case, the system must preload all current records covered by import, insert/update all records from the import pack, 
    // and then DELETE all existing records that are not in the import pack. To do the delete, it must be able to load all records before import 
    // using snapshot loader.
    public readonly DataSnapshotLoader SnapshotLoader; 

    public ImportSyncChannel(string name, IEnumerable<Type> importTypes = null, 
                             ImportSyncActions actions = ImportSyncActions.All, DataSnapshotLoader snapshotLoader = null) : base(name) {
      Actions = actions;
      SnapshotLoader = snapshotLoader;
      if (importTypes != null)
        ImportTypes.UnionWith(importTypes);
    }
  }

  public class ExportSyncChannel : DataSyncChannel {
    public readonly DataSnapshotLoader SnapshotLoader;
    public readonly ChangeTrackingLog ChangeLog;
    public readonly HashSet<Type> ExportTypes = new HashSet<Type>();
    public ExportSyncChannel(string name, DataSnapshotLoader snapshotLoader, ChangeTrackingLog changeLog, IEnumerable<Type> exportTypes)
      : base(name) {
      SnapshotLoader = snapshotLoader;
      ChangeLog = changeLog;
      if (exportTypes != null)
        ExportTypes.UnionWith(exportTypes); 
    }

  }//class

}//ns
