﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Vita.Entities;
using Vita.Entities.Runtime;
using Vita.Common;
using Vita.Modules.ChangeTracking;
using Vita.Services;

namespace Vita.Modules.DataSync {

  public interface IDataSyncService {
    DateTime? GetSyncedUntil(DataSyncChannel channel);
    DataSyncSet GetSyncSet(IEntitySession session, ExportSyncChannel channel, Guid partition, DateTime since, DateTime? until);
    void ApplySyncSet(IEntitySession session, ImportSyncChannel channel, DataSyncSet syncSet);
    void CreateSyncLogEntry(IEntitySession session, string channelName, Guid syncSetId, SyncType syncType, SyncStatus status, string remoteParty, Guid partition,
              DateTime since, DateTime until, int recordCount);
    void UpdateSyncLogEntry(IEntitySession session, Guid syncSetId, SyncStatus status, string messages);
  }

  public class DataSyncModule : EntityModule, IDataSyncService {
    private IChangeTrackingService _changeTracking;
    private ITimeService _timeService; 
    //private IEntityCommand _getLastSync;

    public DataSyncModule(EntityArea area, string moduleName) : base(area, moduleName) {
      RegisterEntities(typeof(IDataSyncLog));
      Model.RegisterService(typeof(IDataSyncService), this); 
/*
      // compile commands
      var syncLog = EntitySet<IDataSyncLog>();
      var getLastQuery = (from sl in syncLog
                         where sl.ChannelName == StringParam(50) && sl.Status == Const(SyncStatus.Success)
                         orderby sl.CreatedOn descending
                         select sl).Take(1);
      _getLastSync = CreateSelect<IDataSyncLog>("DataSyncLogGetLatest", "Gets last successful sync record.", getLastQuery);
 */ 
    }

    public override void Activate(AppActivationContext context ) {
      base.Activate(context);
      _timeService = this.EntityStore.GetService<ITimeService>(); 
      _changeTracking = this.EntityStore.GetService<IChangeTrackingService>();
      if (_changeTracking == null)
        context.Log.Error("DataSyncModule: ChangeTracking service not registered.");
    }

    #region IDataSyncService Members
    public DateTime? GetSyncedUntil(DataSyncChannel channel) {
      var session = EntityStore.OpenSession();
      var syncLog = session.EntitySet<IDataSyncLog>();
      var getLastQuery = (from sl in syncLog
                          where sl.ChannelName == channel.Name && sl.Status == SyncStatus.Success
                          orderby sl.CreatedOn descending
                          select sl).Take(1);
      var lastImport = getLastQuery.FirstOrDefault();
      //var lastImport = session.ExecuteQuery<IDataSyncLog>(_getLastSync, channel.Name).FirstOrDefault();
      if (lastImport == null)
        return null; 
      return lastImport.ChangesUntil;
    }

    public void CreateSyncLogEntry(IEntitySession session, string channelName, Guid syncSetId, SyncType syncType, SyncStatus status, string remoteParty, Guid partition,
              DateTime sinceUtc, DateTime untilUtc, int recordCount) {
      var syncLog = session.NewDataSyncLog(channelName, syncSetId, syncType, status, remoteParty, partition, sinceUtc, untilUtc, recordCount);
    }

    public void UpdateSyncLogEntry(IEntitySession session, Guid syncSetId, SyncStatus status, string messages) {
      var syncLog = session.GetEntity<IDataSyncLog>(syncSetId);
      Util.Check(syncLog != null, "DataSyncLog record not found, Id={0}.", syncSetId);
      syncLog.Status = status;
      syncLog.Messages = messages;
    }

    // Returns export sync set for a specified export sync channel and date range. 
    // Checks if there is a change log entries prior to 'since' date, including Anchor change log entries. 
    // If yes, then client has been synced at least once by this starting date -
    // so method returns incremental sync set. If there are no changelog entires prior to 'since' date-time 
    // then method returns a full sync set for a current date. 
    public DataSyncSet GetSyncSet(IEntitySession session, ExportSyncChannel channel, Guid partition, DateTime since, DateTime? until) {
      //Figure out time interval
      var sinceUtc = since.ToUniversalTime(); 
      var untilUtcNotNull = until == null ? _timeService.Now.ToUniversalTime() : until.Value;
      untilUtcNotNull = untilUtcNotNull.ToUniversalTime(); 
      //check if we need full snapshot
      var firstEntryDate = _changeTracking.GetFirstEntryDateTime(session, channel.ChangeLog);
      // TODO: fix rounding to milliseconds
      // Trying to fix changeLog problem - to deliver incremental changes when appropriate
      //firstEntryDate = RoundToMilliseconds(firstEntryDate); 
      var needFullSnapshot = sinceUtc < firstEntryDate;
      //Create sync set
      DataSyncSet syncSet; 
      if (needFullSnapshot) {
        //we need a full snapshot; load it using channel's SnapshotLoader
        var snapshots = channel.SnapshotLoader.LoadSnapshots(session, partition);
        syncSet = CreateSnapshotSyncSet(sinceUtc, untilUtcNotNull, snapshots, channel.ExportTypes);
      } else {
        //we need incremental changes log
        var changes = _changeTracking.GetChanges(session, channel.ChangeLog, partition, sinceUtc, untilUtcNotNull);
        syncSet = CreateIncrementalSyncSet(session, sinceUtc, untilUtcNotNull, changes, channel.ExportTypes);
      }
      //Filter sync set
      //Apply clear value rules to all records
      if (channel.ClearValueRules.Count > 0)
        ApplyClearValueRules(channel, syncSet); 
      return syncSet; 
    }

    // Imports/applies sync set to the local database
    public void ApplySyncSet(IEntitySession session, ImportSyncChannel channel, DataSyncSet syncSet) {
      var entSession = session as EntitySession;
      var modelInfo = entSession.EntityStore.Model.ModelInfo;
      if (channel.ClearValueRules.Count > 0)
        ApplyClearValueRules(channel, syncSet);
      foreach (var syncGroup in syncSet.SyncGroups) {
        //Check if entity type is allowed to be imported.
        var entInfo = modelInfo.GetEntityInfo(syncGroup.EntityName, true);
        if (!channel.ImportTypes.Contains(entInfo.EntityType)) continue; 
        if (syncGroup.IsFullSnapshot)
          ApplyFullSnapshot(session, syncGroup, channel.Actions.IsSet(ImportSyncActions.Delete));
        else
          ApplyIncrementalChanges(session, syncGroup);
      }//foreach syncGroup
    }

    private static void ApplyFullSnapshot(IEntitySession session, EntitySyncGroup group, bool deleteNotImported) {
      //load all entities
      var entSession = session as EntitySession;
      var modelInfo = entSession.EntityStore.Model.ModelInfo;
      var entInfo = modelInfo.GetEntityInfo(group.EntityName, true);
      var allKeys = new StringSet();
      if (deleteNotImported) {
        //Load all records and save all pk values in hashset. As we go thru sync group records, 
        // we remove keys from this hashset. Those keys that are left in hashset after processing - these records should be deleted.
        var recs = entSession.GetRecords(entInfo);
        allKeys.UnionWith(recs.Select(r => r.PrimaryKey.ValuesToString()));
      }
      foreach (var syncInfo in group.SyncChanges) {
        var pkValues = syncInfo.KeyValues;
        if (deleteNotImported)
          allKeys.Remove(pkValues);
        var pk = EntityKey.KeyFromString(entInfo.PrimaryKey, pkValues); 
        var existingRec = entSession.GetRecord(pk);
        if (existingRec == null)
          AttachAsNew(entSession, syncInfo.Record);
        else
          UpdateExistingRecord(existingRec, syncInfo.Record);
      } //foreach syncInfo
      if (deleteNotImported) {
        //If there are keys remaining in allKeys, delete these records
        foreach (var pkValues in allKeys) {
          var pk = EntityKey.KeyFromString(entInfo.PrimaryKey, pkValues);
          var rec = entSession.GetRecord(pk);
          entSession.DeleteRecord(rec);
        }
      }
    }

    private static void ApplyIncrementalChanges(IEntitySession session, EntitySyncGroup group) {
      var entSession = session as EntitySession;
      var modelInfo = entSession.EntityStore.Model.ModelInfo;
      var entInfo = modelInfo.GetEntityInfo(group.EntityName); 
      foreach (var syncInfo in group.SyncChanges) {
        var pk = EntityKey.KeyFromString(entInfo.PrimaryKey, syncInfo.KeyValues);
        var existingRec = entSession.GetRecord(pk);
        if (syncInfo.ChangeType == EntityChangeType.Delete) {
          if (existingRec != null)
            entSession.DeleteRecord(existingRec);
        } else {
          if (existingRec == null) 
            AttachAsNew(entSession, syncInfo.Record); 
          else
            UpdateExistingRecord(existingRec, syncInfo.Record); 
        }//else (if ChangeType...)
      }//foreach syncInfo
    } //method

    private static void UpdateExistingRecord(EntityRecord existing, EntityRecord fromRecord) {
      Array.Copy(fromRecord.ValuesOriginal, existing.ValuesModified, existing.ValuesModified.Length);
      existing.Status = EntityStatus.Modified;
      existing.SuppressAutoValues = true;
    }

    private static void AttachAsNew(EntitySession session, EntityRecord record) {
      session.Attach(record);
      record.Status = EntityStatus.New;
      record.SuppressAutoValues = true; 
    }

    private static void ApplyClearValueRules(DataSyncChannel channel, DataSyncSet syncSet) {
      if (channel.ClearValueRules.Count == 0)
        return; 
      var allSyncInfos = syncSet.SyncGroups.SelectMany(g=>g.SyncChanges).ToList(); 
      foreach(var syncInfo in allSyncInfos) {
        if (syncInfo.Record != null && syncInfo.ChangeType != EntityChangeType.Delete)
          channel.ApplyClearValueRules(syncInfo.Record); 
      }
    }

    //Snapshot
    private DataSyncSet CreateSnapshotSyncSet(DateTime since, DateTime until, List<EntitySnapshotSet> snapshots, HashSet<Type> filterTypes) {
      var entStore = (EntityStore)this.EntityStore;
      var modelInfo = entStore.Model.ModelInfo;
      var syncSet = new DataSyncSet(since, until);
      foreach (var sn in snapshots) {
        if (sn.Records.Count == 0)
          continue;
        var entInfo = modelInfo.GetEntityInfo(sn.EntityName);
        if (!filterTypes.Contains(entInfo.EntityType))
          continue; 
        var syncGroup = new EntitySyncGroup(sn.EntityName, isFullSnapshot: true); 
        foreach(var rec in sn.Records) 
          syncGroup.SyncChanges.Add(EntitySyncInfo.CreateForFullSync(rec));
        syncSet.SyncGroups.Add(syncGroup);
      }
      return syncSet;
    }

    private DataSyncSet CreateIncrementalSyncSet(IEntitySession session, DateTime since, DateTime until,
                List<EntityChangeInfo> changes, HashSet<Type> filterTypes) {
      var entSession = session as EntitySession; 
      var modelInfo = entSession.EntityStore.Model.ModelInfo;
      var syncSet = new DataSyncSet(since, until);
      //Group changes by entity type
      var changeGroups = changes.GroupBy(ch => ch.EntityName);
      foreach (var changeGroup in changeGroups) {
        var entName = changeGroup.Key;
        var entInfo = modelInfo.GetEntityInfo(entName);
        if (!filterTypes.Contains(entInfo.EntityType)) continue; 
        var syncGroup = new EntitySyncGroup(entName, isFullSnapshot: false);
        syncSet.SyncGroups.Add(syncGroup);
        foreach (var changeInfo in changeGroup) {
          EntityRecord record = null;
          if (changeInfo.ChangeType != EntityChangeType.Delete) {
            var pk = EntityKey.KeyFromString(entInfo.PrimaryKey, changeInfo.KeyValues);
            record = entSession.GetRecord(pk);
            if (record == null) //record might not exist anymore
              continue; 
          }
          var syncInfo = new EntitySyncInfo(changeInfo.EntityName, changeInfo.ChangeType, changeInfo.KeyValues, record);
          syncGroup.SyncChanges.Add(syncInfo);
        }//foreach changeInfo
      }
      return syncSet;
    }
    #endregion

    private static DateTime RoundToMilliseconds(DateTime date) {
      var result = new DateTime(date.Year, date.Month, date.Day, date.Hour, date.Minute, date.Second, date.Millisecond, date.Kind);
      return result; 
    }
  }//class
}
