﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Xml;

using Vita.Common; 

namespace Vita.Modules.ErrorLogging {

  //Helper class for creating HTML pages to view errors in a browser
  public class ErrorHtmlPageBuilder : HtmlPageBuilder {

    const string PrevNextStyle = " .prevnext { font-family: arial; color:gray; font-size: 10px; text-align:left; }";
    string _listPageUrlTemplate;
    string _detailsPageUrlTemplate;

    public ErrorHtmlPageBuilder(string listPageUrlTemplate = "errors?skip={0}&take={1}",
                                string detailsPageUrlTemplate = "error/{0}", 
                                string style = null) : base(style) {
      _listPageUrlTemplate = listPageUrlTemplate;
      _detailsPageUrlTemplate = detailsPageUrlTemplate;
      base.Style += PrevNextStyle; 
    }

    public void CreateErrorDetailsPage(IErrorInfo error) {
      if (error == null) {
        BeginPage("Not found", false);
        WriteRaw("Error info not found.");
        EndPage();
        return; 
      }
      //Normal page
      BeginPage("Error Details");
      WriteNameValue("Message:   ", error.Message);
      WriteNameValue("DATE/TIME: ", error.CreatedOn.ToString("g"));
      WriteNameValue("MACHINE:   ", error.MachineName);
      WriteNameValue("USER:      ", error.UserName);
      WriteNameValue("DETAILS:   ", string.Empty);
      WriteMultilineText(error.Details);
      WriteNameValue("LOG: ", string.Empty);
      WriteMultilineText(error.Log);
      WriteNameValue("Error ID:  ", error.Id.ToString().ToUpperInvariant());
      //Link to the first page of full error list.
      WriteRaw("<hr/>");
      var listPageUrl = Util.SafeFormat(_listPageUrlTemplate, 0, 50);
      WriteHomeLink("All Errors", listPageUrl);
      EndPage();
    }

    public void CreateErrorListPage(IList<IErrorInfo> errors, int skip, int take) {
      BeginPage("Errors");
      WriteTableHeader("DateTime", "Error Message", "Machine/User", "Id");
      foreach (var err in errors) {
        WriteStartElement("tr");
        WriteElement("td", err.CreatedOn.ToString("s"));
        WriteStartElement("td");
        var url = Util.SafeFormat(_detailsPageUrlTemplate, err.Id);
        WriteAnchor(err.Message, url);
        WriteEndElement(); //td
        WriteElement("td", err.MachineName + " / " + err.UserName);
        WriteElement("td", err.Id.ToString());
        WriteEndElement(); //tr
      }
      WriteEndElement(); //table
      WritePrevNext(errors.Count, skip, take);
      EndPage();
    }

    private void WritePrevNext(int errorCount, int skip, int take) {
      var hasPrev = skip > 0;
      var hasNext = errorCount >= take;
      WriteStartElement("div");
      WriteAttribute("class", "prevnext");
      if (hasPrev) {
        var newSkip = skip - take;
        if (newSkip < 0) newSkip = 0;
        var url = Util.SafeFormat(_listPageUrlTemplate, newSkip, take);
        WriteAnchor("Prev", url);
      } else
        WriteString("Prev");
      WriteString(" . . . ");
      if (hasNext)
        WriteAnchor("Next", Util.SafeFormat(_listPageUrlTemplate, skip + take, take));
      else
        WriteString("Next");
      WriteEndElement(); 
    }

    private void WriteNameValue(string name, string value, bool lineBreak = false) {
      WriteStartElement("p");
      WriteElement("b", name);
      if (lineBreak)
        WriteRaw("<br/>");
      WriteString(value); //writes with escaping
      WriteEndElement(); // /p
    }
  }//class

}//ns
