﻿using System;
using System.ComponentModel; 
using System.Collections.Generic;
using System.Diagnostics; 
using System.Linq;
using System.Text;
using System.IO;
using System.Xml;

using Vita.Common;
using Vita.Entities;
using Vita.Entities.Runtime;
using Vita.Entities.Model;
using Vita.Data;
using Vita.Services;

namespace Vita.Modules.ErrorLogging {

  //Entities
  [Description("Error information record.")]
  [Entity, OrderBy("CreatedOn:DESC"), Paged, ClusteredIndex("CreatedOn,Id"), IsSystem]
  public interface IErrorInfo {
    [PrimaryKey, Auto]
    Guid Id { get; set; }
    [Auto(AutoType.CreatedOn), Utc]
    DateTime CreatedOn { get; set; }
    /*
        DateTimeOffset? LocalTimeOffset { get; set; }
    */
    [Auto(AutoType.CreatedOn)]
    DateTime LocalTime { get; set; }

    [Size(50), Nullable]
    string MachineName { get; set; }
    [Size(50), Nullable]
    string UserName { get; set; }
    [Size(50), Nullable]
    string AppName { get; set; }
    [Size(250)]
    string Message { get; set; }
    [Size(100)]
    string ExceptionType { get; set; }
    [Memo, Nullable]
    string Details { get; set; }
    [Memo, Nullable]
    string Log { get; set; }
  }

  public class ErrorLogModule : EntityModule, IErrorLogService {
    public IErrorLogService FallbackService;
    private ITimeService _timeService; 

    public ErrorLogModule(EntityArea area)   : base(area, "ErrorLog", "Error Logging Module.") {
      RegisterEntities(typeof(IErrorInfo));
      Model.RegisterService(typeof(IErrorLogService), this);
    }

    public override void  Activate(AppActivationContext context) {
 	    base.Activate(context);
      _timeService = EntityStore.GetService<ITimeService>(); 
    }

    #region IErrorLogService Members

    public bool LogError(Exception exception, string userLog, string user = null, string appName = null, string machineName = null) {
      if (exception == null)
        return true; 
      try {
        machineName = machineName ?? Environment.MachineName;
        var session = this.EntityStore.OpenSession();
        var errInfo = session.NewEntity<IErrorInfo>();
        errInfo.AppName = appName;
        errInfo.ExceptionType = exception.GetType().Name;
        //Some messages might be really long; check length to fit into the field
        errInfo.Message = Util.CheckLength(exception.Message, 250);
        errInfo.Log = userLog;
        errInfo.UserName = user;
        errInfo.MachineName = machineName;
        errInfo.Details = exception.ToLogString(); //writes exc.ToString() and exc.Data collection, along with all inner exception details
        session.SaveChanges();
        return true;
      } catch (Exception logEx) {
        LogToFallback(logEx, exception.ToString(), userLog, user, appName);
        return false;
      }
    }
    public bool LogError(string message, string details, string userLog, string user = null, string appName = null, string machineName = null) {
      try {
        var session = this.EntityStore.OpenSession();
        var errInfo = session.NewEntity<IErrorInfo>();
        errInfo.AppName = appName;
        errInfo.ExceptionType = "Unknown";
        //Some messages might be really long; check length to fit into the field
        errInfo.Message = Util.CheckLength(message, 250);
        errInfo.Log = userLog;
        errInfo.UserName = user;
        errInfo.MachineName = machineName;
        errInfo.Details = details;
        session.SaveChanges();
        return true;
      } catch (Exception logEx) {
        LogToFallback(logEx, details, userLog, user, appName);
        return false;
      }
    }
    #endregion

    internal bool LogToFallback(Exception logException, string originalError, string userLog, string user, string appName) {
      //Trace
      var msg = "Fatal error when trying to log error: \r\n" + logException.ToString();
      var delimLine = (new string('=', 80)) + Environment.NewLine;
      Trace.WriteLine(delimLine);
      Trace.WriteLine(msg);
      Trace.WriteLine("\r\n     Original exception: \r\n" + originalError);
      Trace.WriteLine(delimLine);
      if (FallbackService != null) {
        try {
          FallbackService.LogError(logException, userLog, user, appName);
          FallbackService.LogError("Original error (see Details)", originalError, userLog, user, appName);
          return true;
        } catch { }
      }
      //We are here if Fallback is null or failed for any reason
      // Let's try EventLog
      try {
        var srcName = "VITA_ERROR_LOG";
        if (!EventLog.SourceExists(srcName))
          EventLog.CreateEventSource(srcName, srcName);
        EventLog.WriteEntry(srcName, msg);
      } catch { }
      return false;
    }


  }// ErrorLogModule class


}//ns
