﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Security.Cryptography;
using System.IO;

using Vita.Entities;
using Vita.Common;
using Vita.Entities.Runtime;
using System.Diagnostics;

namespace Vita.Modules.Login {
  using Binary = System.Data.Linq.Binary;

  public class LoginModule : EntityModule, ILoginService {
    public const int HashSize = 128;
    public readonly int IterationCount;
    //A flag controlling if user name should be stored in plain text. Should be set to false if application or web site 
    // use sensitive info as user name - for ex, some websites use emails as user names. 
    // The system hashes the username anyway and uses this hash to locate the login record. 
    public readonly bool SavePlainTextUserNames; 
    // The Salt is not a real secret when using RFC2898. One common requirement is that it should be different
    // for each user. The PasswordSalt property is a prefix for salt common for all users.
    // The actual salt is calculated as: 
    //   salt = Encoding.Unicode.GetBytes(PasswordSalt + "_" + userName);
    // Use some random GUID as password salt: "dbbc76f0-b6ff-4f0d-9fc1-e8ea5a9489ab"
    private string _passwordSalt;
    private IEntityCommand _updatePasswordHashCommand; //_loginCommand 


    public LoginModule(EntityArea area, string name, string passwordSalt, bool savePlainTextUserNames = true, int iterationCount = 1000)
            : base(area, name, "VITA Login Module") {
      Util.Check(!string.IsNullOrWhiteSpace(passwordSalt), "Password salt may not be empty.");
      Util.Check(iterationCount > 100 && iterationCount < 100000, "IterationCount should be between 100 and 100 000.");
      ValidateStableHash();
      _passwordSalt = passwordSalt;
      SavePlainTextUserNames = savePlainTextUserNames;
      IterationCount = iterationCount; 
      RegisterEntities(typeof(ILogin));
      area.Model.RegisterService(typeof(ILoginService), this);
      CreateCommands();
    }

    //.NET framework's GetHashCode() is not garanteed to be stable between .NET versions. 
    // If we want to keep hashes in database, we need a stable hash implementation - it is provided by Util.StableHash() method.
    // However, somebody (guess who?) might tweak this method between versions (trying to improve it foolishly), and this will break all existing 
    // records in databases that store this hash value. To detect this, we keep a test case (input and hash-value) and verify that 
    // Util.StableHash still works the same. 
    private static void ValidateStableHash() {
      const string _stableHashTestInput = "abcdefg";
      const int _stableHashTestValue = 340755172; 

      var currentHashValue = Util.StableHash(_stableHashTestInput);
      Util.Check(currentHashValue == _stableHashTestValue,
        "LoginModule: Util.StableHash verification failed - method implementation changed, existing data is broken.");
    }

    //TODO: remove UserName from search criteria
    private void CreateCommands() {
/* CustomSelects are temporarily disabled
      var logins = EntitySet<ILogin>();
      // IMPORTANT! We need to provide parameters size (HashSize) in LINQ expressions; otherwise, we'll not have match in WHERE clause
      var qryLogin = from lg in logins
                  where 
                  (lg.UserName == StringParam(50) || lg.UserNameHash == Param<int>()) && lg.PasswordHash == Param<byte[]>(HashSize) &&
                        lg.DomainId == Param<Guid>()
                  select lg;
      _loginCommand = CreateSelect<ILogin>("LoginFindByUserHashDomain", "Finds login record by user name, password hash and domain ID.", qryLogin);
 */ 
      // update password hash
      _updatePasswordHashCommand = CreatePartialUpdate<ILogin>("LoginUpdatePasswordHash", "Updates password hash.", lg => new { lg.PasswordHash });
    }

    #region ILoginService Members

    public ILogin NewLogin(IEntitySession session, string userName, string password,
            Guid? domainId, DateTime? expires = null, LoginFlags flags = LoginFlags.None) {
      var login = session.NewEntity<ILogin>();
      if (SavePlainTextUserNames)
        login.UserName = userName;
      login.UserNameHash = Util.StableHash(userName); 
      login.PasswordHash = GetPasswordHash(userName, password);
      login.DomainId = (domainId == null) ? Guid.Empty : domainId.Value;
      login.Expires = expires;
      login.Flags = flags;
      return login;
    }

    public ILogin Login(IEntitySession session, string userName, string password, Guid? domainId = null) {
      var hash = GetPasswordHash(userName, password);
      var bhash = new Binary(hash); 
      var userNameHash = Util.StableHash(userName);
      Guid domainIdValue = (domainId == null) ? Guid.Empty : domainId.Value;
      var logins = session.EntitySet<ILogin>();
      // TODO: we do not compare usernames, only UserNameHash values; UserName might be null if we don't save them
      var qryLogin = from lg in logins
                     where  (lg.UserNameHash == userNameHash) && lg.PasswordHash == bhash &&     
                            lg.DomainId == domainIdValue
                     select lg;
      var login = qryLogin.FirstOrDefault();
      return login;
    }

    public void UpdateCredentials(IEntitySession session, ILogin login, string userName, string password) {
      Util.Check(login != null, "Login not found.");
      var hash = GetPasswordHash(login.UserName, password);
      session.ExecuteOnSave(_updatePasswordHashCommand, ExecuteOptions.None,  login.Id, hash);
    }

    #endregion

    //Implements RFC2898-based hash generations
    private byte[] GetPasswordHash(string userName, string password) {
      var saltBytes = Encoding.Unicode.GetBytes(_passwordSalt + "_" + userName);
      return GetRfc2898Hash(password, saltBytes, HashSize, IterationCount);
    }

    public static byte[] GetRfc2898Hash(string value, byte[] saltBytes, int hashSize, int iterations) {
      Rfc2898DeriveBytes k1 = new Rfc2898DeriveBytes(value, saltBytes, iterations);
      var result = k1.GetBytes(hashSize);
      return result;
    }

  }//module

}//ns
