﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Vita.Entities;
using Vita.Entities.Runtime; 

namespace Vita.Samples.BooksModel {
  //Helper methods to create entities
  public static class BookExtensions {

    public static IBook NewBook(this IEntitySession session, BookEdition editions, BookCategory category, string title, string description,
         IPublisher publisher, DateTime publishedOn, decimal price) {
      var book = session.NewEntity<IBook>();
      book.Editions = editions;
      book.Category = category; 
      book.Title = title;
      book.Description = description;
      book.Publisher = publisher;
      book.PublishedOn = publishedOn;
      book.Price = price;
      return book;
    }

    public static IPublisher NewPublisher(this IEntitySession session, string name) {
      var pub = session.NewEntity<IPublisher>();
      pub.Name = name;
      return pub;
    }

    public static IAuthor NewAuthor(this IEntitySession session, string firstName, string lastName, string bio = null) {
      var auth = session.NewEntity<IAuthor>();
      auth.FirstName = firstName;
      auth.LastName = lastName;
      auth.Bio = bio;
      return auth;
    }

    public static IUser NewUser(this IEntitySession session, string userName, UserType type, bool isActive = true) {
      var user = session.NewEntity<IUser>();
      user.UserName = userName;
      user.DisplayName = userName;
      user.Type = type;
      user.IsActive = isActive; 
      return user; 
    }
    public static ICoupon NewCoupon(this IEntitySession session, string promoCode, double discountPerc, DateTime expires) {
      var coupon = session.NewEntity<ICoupon>();
      coupon.PromoCode = promoCode;
      coupon.DiscountPerc = discountPerc;
      coupon.ExpiresOn = expires;
      return coupon;
    }
    public static IBookOrder NewOrder(this IEntitySession session, IUser user) {
      var order = session.NewEntity<IBookOrder>();
      order.User = user;
      order.Status = OrderStatus.Open;
      return order;
    }
    public static IBookOrderLine Add(this IBookOrder order, IBook book, int quantity) {
      var session = EntityHelper.GetSession(order);
      var line = session.NewEntity<IBookOrderLine>();
      line.Order = order; 
      line.Book = book;
      line.Quantity = quantity; 
      line.Price = book.Price;
      order.Lines.Add(line); 
      return line; 
    }
    public static void CompleteOrder(this IBookOrder order, string promoCode = null) {
      order.Total = order.Lines.Sum(line => line.Price * line.Quantity);
      if (!string.IsNullOrWhiteSpace(promoCode)) {
        var session = EntityHelper.GetSession(order); 
        var coupon = LookupCoupon(session, promoCode);
        if (coupon != null && coupon.ExpiresOn >= DateTime.Now && coupon.AppliedOn == null) {
          coupon.AppliedOn = DateTime.Now;
          order.Total = (decimal) (((double)order.Total) * ((100 - coupon.DiscountPerc) / 100.0));
        }
      }
      order.Status = OrderStatus.Completed; 
    }

    public static IBookReview NewReview(this IEntitySession session, IUser user, IBook book, int rating, string caption, string text) {
      var review = session.NewEntity<IBookReview>();
      review.User = user;
      review.Book = book;
      review.Rating = rating; 
      review.Caption = caption;
      review.Review = text;
      return review;
    }

    public static ICoupon LookupCoupon(this IEntitySession session, string code) {
      var query = from c in session.EntitySet<ICoupon>()
                   where c.PromoCode == code
                   select c;
      var coupon = query.FirstOrDefault();
      return coupon; 
    }

    public static bool IsSet(this BookEdition flags, BookEdition flag) {
      return (flags & flag) != 0;
    }
    public static bool IsSet(this UserType flags, UserType flag) {
      return (flags & flag) != 0;
    }
  }//class
}
