﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Linq.Expressions;

using Vita.Entities;
using Vita.Entities.Model;
using Vita.Entities.Authorization;

namespace Vita.Samples.BooksModel {

  /*  ROLES:
   * WebVisitor - can browse books, publishers, authors, reviews, 
   *              create/edit reviews, buy books (create book orders), optionall using coupon codes, 
   *              view/edit his orders
   * Author -     can browse all books, reviews; can update his own Bio; cannot edit books, except modifying 
   *               Abstract  and Descriptions of his books. Author is allowed to update Bio, book Abstract and Description
   *               ONLY within explicitly started activity AuthorEdit.
   * BookEditor - can create books, authors, but not publishers; cannot see any book orders or coupons
   * StoreManager - can create publishers, coupons, can adjust orders; can moderate (delete any) reviews 
   */ 

  public static class BooksAuthorization {
    // A key for UserContext.Values dictionary to store the ID of order under udjustment
    public const string AdjustedOrderIdKey = "AdjustedOrderId";

    public static Role WebVisitor;
    public static Role Author;
    public static Role BookEditor;
    public static Role StoreManager;

    public static DynamicActivityGrant AuthorEditGrant;
    public static DynamicActivityGrant ManagerAdjustOrderGrant;

    public const string UserIdKey = "UserId";

    public static IList<Role> GetRoles(this IUser user) {
      if (!_initialized)
        Init(); 

      var roles = new List<Role>(); 
      var type = user.Type;
      if (type.IsSet(UserType.WebVisitor)) 
        roles.Add(WebVisitor);
      if (type.IsSet(UserType.Author)) 
        roles.Add(Author);
      if (type.IsSet(UserType.BookEditor)) 
        roles.Add(BookEditor);
      if (type.IsSet(UserType.StoreAdmin)) 
        roles.Add(StoreManager);
      return roles;
    }

    private static bool _initialized; 
    public static void Init() {
      if (_initialized) return;
      _initialized = true;

      // Data filters
      // user ID reader reads current user ID from user context.
      var userIdReader = new ContextValueReader<Guid>(UserIdKey); 
      // Data filter detects if a given entity "belongs" to a data group identified 
      // by a value in UserContext. The following data filter matches entities against UserId in UserContext
      // The filter definition uses the AddEntity overload that expects a function that returns the group Id (userId) 
      // for an entity (func: entity -> groupId). The returned Id for an entity will be matched against the value
      // returned by userIdReader.
      // The other AddEntity overload used for author data filter expects a method that returns true/false: 
      //   func: (entity, userId) -> bool
      var userDataFilter = new DataFilter<Guid>("UserData", userIdReader);
      userDataFilter.AddEntity<IBookOrder>(bo => bo.User.Id);
      userDataFilter.AddEntity<IBookOrderLine>(ol => ol.Order.User.Id);
      userDataFilter.AddEntity<IBookReview>(r => r.User.Id);
      userDataFilter.AddEntity<IUser>(u => u.Id);
      // Author data filter. 
      var authorDataFilter = new DataFilter<Guid>("AuthorData", userIdReader);
      // Simple case. IAuthor record is matched by filter if IAuthor.User.Id == Current-user-Id
      // We use static helper function that takes into account that author.User might be null
      authorDataFilter.AddEntity<IAuthor>(a => GetAuthorUserId(a));
      // More complex case. The argument of AddEntity method is two-parameter function returning bool. 
      //  We provide an expression that checks if userId matches Id of any author of a book.
      authorDataFilter.AddEntity<IBook>((b, userId) => (b.Authors.Any(a => GetAuthorUserId(a) == userId)));
      // BookOrder filter for adjusting order by StoreManager
      var adjustmentOrderIdReader = new ContextValueReader<Guid>(AdjustedOrderIdKey);
      var adjustOrderFilter = new DataFilter<Guid>("OrderToAdjust", adjustmentOrderIdReader);
      adjustOrderFilter.AddEntity<IBookOrder>(bo => bo.Id);
      adjustOrderFilter.AddEntity<IBookOrderLine>(line => line.Order.Id);

      //Entity resources
      var books = new EntityGroupResource("Books", typeof(IBook), typeof(IBookAuthor), typeof(IAuthor));
      var publishers = new EntityGroupResource("Publishers", typeof(IPublisher));
      var orders = new EntityGroupResource("Orders", typeof(IBookOrder), typeof(IBookOrderLine));
      var reviews = new EntityGroupResource("BookReviews", typeof(IBookReview));
      var users = new EntityGroupResource("Users", typeof(IUser));
      var authorEditData = new EntityGroupResource("AuthorEditData");
      authorEditData.Add(typeof(IAuthor), "Bio");
      authorEditData.Add(typeof(IBook), "Description,Abstract");
      // authorEditData.Add<IBook>(b => b.Description, b => b.Abstract); -- alternative way
      var coupons = new EntityGroupResource("Coupons", typeof(ICoupon));
      var couponAppliedOn = new EntityGroupResource("CouponAppliedOn");
      couponAppliedOn.Add<ICoupon>(c => c.AppliedOn);

      //Permissions
      var browseBooks = new EntityGroupPermission("BrowseBooks", Actions.Read, books, publishers);
      var createOrders = new EntityGroupPermission("CreateOrders", Actions.CRUD, orders);
      var browseReviews = new EntityGroupPermission("BrowseReviews", Actions.Read, reviews);
      var writeReviews = new EntityGroupPermission("WriteReviews", Actions.CRUD, reviews);
      var editBooks = new EntityGroupPermission("EditBooks", Actions.CRUD, books);
      var deleteReviews = new EntityGroupPermission("DeleteReviews", Actions.Read | Actions.Delete, reviews);
      var editPublishers = new EntityGroupPermission("EditPublishers", Actions.CRUD, publishers);
      var editByAuthor = new EntityGroupPermission("EditByAuthor", Actions.Update, authorEditData);
      var editPersInfo = new EntityGroupPermission("EditPersInfo", Actions.CRUD, users);
      var adjustOrder = new EntityGroupPermission("AdjustOrder", Actions.Read | Actions.Update | Actions.Delete, orders);

      // We grant Peek permission here for Coupons. The app code can find a coupon by code, but user cannot see it 
      // (no READ permission) or see any coupons. User receives coupon in email, and uses it when buying a book. 
      // System looks up the coupon code and applies the discount. But user cannot never see any coupons in the system.
      var lookupCoupon = new EntityGroupPermission("LookupCoupon", Actions.Peek, coupons);
      //permission to update ICoupon.AppliedOn property - we set it when user uses coupon
      var useCoupon = new EntityGroupPermission("UseCoupon", Actions.UpdateStrict, couponAppliedOn);
      var manageCoupons = new EntityGroupPermission("ManageCoupons", Actions.CRUD, coupons);

      //Activities
      var browsing = new Activity("Browsing", browseBooks, browseReviews);
      var shopping = new Activity("Shopping", createOrders, lookupCoupon, useCoupon);
      var reviewing = new Activity("Reviewing", writeReviews);
      var editingPersInfo = new Activity("EditingPersInfo", editPersInfo);
      var bookEditing = new Activity("BookEditing", editBooks);
      var editingByAuthor = new Activity("EditingByAuthor", editByAuthor);
      var managingStore = new Activity("ManagingStore", editPublishers, manageCoupons);
      var moderatingReviews = new Activity("ModeratingReviews", deleteReviews);
      var adjustingOrders = new Activity("AdjustingOrders", adjustOrder);

      //Roles
      //Browse all books; view any reviews;  view/edit orders only for current user; edit reviews only created by current user 
      WebVisitor = new Role("WebVisitor", browsing);
      WebVisitor.Grant(userDataFilter, shopping, reviewing, editingPersInfo);
      BookEditor = new Role("BookEditor", browsing, bookEditing);
      Author = new Role("Author"); 
      Author.ChildRoles.Add(WebVisitor);
      //We save the grant in static field, to explicitly enable the activity at runtime for limited scope  
      AuthorEditGrant = Author.GrantDynamic(authorDataFilter, editingByAuthor);
      StoreManager = new Role("StoreManager", managingStore, moderatingReviews);
      // Store manager is able to adjust orders, but only in the context of dynamic (explicitly started) activity
      // When adjusting activity starts, it saves order Id in user context under AdjustedOrderIdKey. 
      // All records being edited are then verified against this order Id.
      // This garantees that during adjustment editing we modify only data for the order that we started adjustment for.
      ManagerAdjustOrderGrant = StoreManager.GrantDynamic(adjustOrderFilter, adjustingOrders, AdjustedOrderIdKey);
    }

    private static Guid GetAuthorUserId(IAuthor author) {
      var user = author.User;
      return user == null ? Guid.Empty : user.Id; 
    }
  }//class

}
