﻿using System;
using System.Collections.Generic;
using System.ComponentModel;

using Vita.Entities; 

namespace Vita.Samples.BooksModel {

  //Entities for Books module

  [Flags, Description("Available editions.")]
  public enum BookEdition {
    [Description("Paperback edition.")]
    Paperback = 0x01,
    [Description("Hardcover edition.")]
    Hardcover = 0x02,
    [Description("Electronic download.")]
    EBook = 0x04,
  }

  [Description("Book category.")]
  public enum BookCategory {
    [Description("Books on programming.")]
    Programming,
    [Description("Fiction.")]
    Fiction,
    [Description("Kids books.")]
    Kids,
  }

  public enum OrderStatus {
    Open, //in progress, being edited by the user
    Completed, 
    Canceled,
  }

  [Flags, Description("User types.")]
  public enum UserType {
    WebVisitor = 1,
    Author = 1 << 1,
    BookEditor = 1 << 2,
    StoreAdmin = 1 << 3,
  }

  // Note that Book and Author have Paged attribute - meaning they are paged in database (in stored procedure)
  // Publisher does not have this attribute, we expect it to be a small table. We can still read/query it with 
  // skip/take parameters, but paging is performed on the client: all records are loaded and then code extracts the span of requested records. 
  [Entity, Paged, OrderBy("Title"), Validate(typeof(BooksModule), "ValidateBook")]
  [Description("Represents a book.")]
  public interface IBook {

    [PrimaryKey, Auto]
    Guid Id { get; set; }

    [Description("Book publisher.")] // We add some decriptions to book properties - they will appear on Web service help page.
    IPublisher Publisher { get; set; }
    
    [Description("Book title.")]
    string Title { get; set; } //uses default length
    
    [Size(250), Nullable, Description("Book description.")]
    string Description { get; set; }

    [Description("Publication date.")]
    DateTime PublishedOn { get; set; }

    [Memo, Nullable, Description("Book abstract.")]
    string Abstract { get; set; }
    
    [Description("Book category.")]
    BookCategory Category { get; set; }

    [Description("Available editions.")]
    BookEdition Editions { get; set; }

    [Description("Book price.")]
    Decimal Price { get; set; }

    [ManyToMany(typeof(IBookAuthor)), Description("Book authors.")]
    IList<IAuthor> Authors { get; }
  }

  [Entity, OrderBy("Name"), SmallTable, Description("Represents a publisher.")]
  public interface IPublisher {
    [PrimaryKey, Auto]
    Guid Id { get; set; }
    string Name { get; set; }
    [OrderBy("PublishedOn:DESC")]
    IList<IBook> Books { get; }
  }

  [Entity, Paged, OrderBy("LastName,FirstName"), Description("Represents an author.")]
  [Display("{LastName}, {FirstName}")]
  public interface IAuthor {
    [PrimaryKey, Auto]
    Guid Id { get; set; }
    [Size(30)]

    string FirstName { get; set; }
    [Size(30)]
    string LastName { get; set; }
    [Memo, Nullable]
    string Bio { get; set; }
    [ManyToMany(typeof(IBookAuthor))]
    IList<IBook> Books { get; }
    [Nullable]
    IUser User { get; set; } //author might be a user

    [Computed(typeof(BooksModule), "GetFullName"), DependsOn("FirstName,LastName")] //DependsOn is optional, used for auto PropertyChanged firing
    string FullName { get; }
  }

  // Note: we set CascadeDelete on reference to Book, but not Author. 
  // We can always delete a book, and the link to the author will be deleted automatically.
  // But we cannot delete an Author if there are any books in the system by this author. 
  [Entity, PrimaryKey("Book,Author")]
  public interface IBookAuthor {
    [CascadeDelete]
    IBook Book { get; set; }
    IAuthor Author { get; set; }
  }

  [Entity, OrderBy("UserName")]
  public interface IUser {
    [PrimaryKey, Auto]
    Guid Id { get; set; }
    [Size(30), Unique]
    string UserName { get; set; }
    [Size(30)]
    string DisplayName { get; set; }
    UserType Type { get; set; } //might be combination of several type flags
    bool IsActive { get; set; }
  }

  [Entity, OrderBy("CreatedOn"), ClusteredIndex("User,CreatedOn,Id")]
  public interface IBookOrder {
    [PrimaryKey, Auto]
    Guid Id { get; set; }

    [Auto(AutoType.CreatedOn), Utc]
    DateTime CreatedOn { get; }

    [NoUpdate]
    IUser User { get; set; }

    decimal Total { get; set; }

    OrderStatus Status { get; set; }

    [PersistOrderIn("LineNumber")]
    IList<IBookOrderLine> Lines { get; }

    [Computed(typeof(BooksModule), "GetOrderSummary"), DependsOn("User,Total,CreatedOn")] //DependsOn is optional, used for auto PropertyChanged firing
    string Summary { get; }
  }

  [Entity, OrderBy("LineNumber"), ClusteredIndex("Order,Id")]
  public interface IBookOrderLine {
    [PrimaryKey, Auto]
    Guid Id { get; set; }
    [NoUpdate, CascadeDelete] //Delete all lines when order is deleted
    IBookOrder Order { get; set; }
    int LineNumber { get; set; } //automatically maintained line order - see IBookOrder.Lines property
    int Quantity { get; set; }
    IBook Book { get; set; }
    Decimal Price { get; set; } //captures price at the time the order was created
  }

  [Entity, OrderBy("CreatedOn"), ClusteredIndex("Book,CreatedOn,Id")]
  public interface IBookReview {
    [PrimaryKey, Auto]
    Guid Id { get; set; }

    [Auto(AutoType.CreatedOn), Utc]
    DateTime CreatedOn { get; set; }

    [NoUpdate]
    IBook Book { get; set; }
    
    [NoUpdate]
    IUser User { get; set; }

    //1 .. 5
    int Rating { get; set; }
    
    [Size(100)]
    string Caption { get; set; }
    [Memo]
    string Review { get; set; }
  }

  [Entity, OrderBy("CreatedOn"), ClusteredIndex("CreatedOn,Id")]
  public interface ICoupon {
    [PrimaryKey, Auto]
    Guid Id { get; set; }

    [Auto(AutoType.CreatedOn), Utc]
    DateTime CreatedOn { get; set; }

    DateTime ExpiresOn { get; set; }

    [Size(20), Index]
    string PromoCode { get; set; }

    double DiscountPerc { get; set; }

    DateTime? AppliedOn { get; set; }
  }
  
}
