﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text; 

using Vita.Entities; 

namespace Vita.Samples.BooksModel {
  //We could register Book entities directly at program startup. But for better code modularity, we create 
  // a data module that handles all book-related functionality - including entity registration, stored procedures, etc.
  // We also define a small static extension class to add handy entity-creation methods.
  public class BooksModule : EntityModule {
    //IEntityCommand _getBooksByAuthor, _getPublisherLatestBook, _booksChangePrice, _deleteOldCheapBooks, _bookDiscountOldFiction;

    public BooksModule(EntityArea area) : base(area, "Books", "Books module") {
      RegisterEntities(typeof(IBook), typeof(IPublisher), typeof(IAuthor), typeof(IBookAuthor), typeof(IBookReview),
                       typeof(IUser), typeof(IBookOrder), typeof(IBookOrderLine),  typeof(ICoupon)); 
      //Register companion types that describe keys and indexes on entities
      RegisterCompanionTypes(
          typeof(IBookKeys), typeof(IAuthorKeys), typeof(IPublisherKeys));
      //Compile commands
      CreateCustomCommands();
    }

    //Define and compile stored procedures for custom Select, Update, Delete operations using LINQ syntax.
    // This method will be called during model activation: by firing the CompilingCommands event 
    // the model suggests to custom modules to create stored procedures they will use.
    public void CreateCustomCommands() {
/*
      var books = EntitySet<IBook>();
      var pubs = EntitySet<IPublisher>();
      var authors = EntitySet<IAuthor>();
      var bookAuthors = EntitySet<IBookAuthor>();

      //Let's use 3-table join and find all books of an author identified by name
      var queryGetByAuthor = from bk in books
                             join ba in bookAuthors on bk equals ba.Book
                             join au in authors on ba.Author equals au
                             where au.LastName == Param<string>(30) //parameter size
                             select bk;
      _getBooksByAuthor = CreateSelect<IBook>("_BooksGetByAuthor", "Gets books by author.", queryGetByAuthor);

      // Using Take(1) - this is special case for VITA linq engine
      var queryGetPubLatestBook = (from bk in books
                                  where bk.Publisher.Name == StringParam(50)
                                  orderby bk.PublishedOn descending
                                  select bk).Take(1);
      _getPublisherLatestBook = CreateSelect<IBook>("_BookGetLatestByPublisher", "Gets latest book by publisher", queryGetPubLatestBook);

      //Update query must return auto-type containing two kinds of fields: primary key to identify the record, 
      // and fields to be updated in the record identified by PK
      var prmPrice = Param<double>();
      var changePriceQuery = from bk in books
                             where bk.Price > prmPrice // bk.Category == Param<BookCategory>()
                             select new { Id = bk.Id, Price = bk.Price * Param<double>() };
      _booksChangePrice = CreateUpdate<IBook>("_BooksChangePrice", "Changes all prices on books in a given category using a multiplier.", changePriceQuery);

      //Delete query must return ID's of books to be deleted
      var delQuery = from b in books
                     where b.Price < Const(2.0) && b.PublishedOn < Param<DateTime>()  
                     select new { b.Id };
      _deleteOldCheapBooks = CreateDelete<IBook>("_BookDeleteOldCheap",
                          "Deletes all books published before a given date and price less than $2.", delQuery);

      // Notice the use of 'Param<T>.Value' function when specifying value of anonymous type property.
      // If we used 'Param<double>' then property type would be 'EntityCommandParameter', not double. So we use 'Value' accessor to make it double.
      // Also, just for testing purposes, we use a literal constant BookCategory.Fiction directly in the query, 
      // without wrapping it in Const<>() call. LINQ will create an extra parameter for this value, but we do NOT need to provide
      // it when calling the method - VITA will fill it automatically
      var queryOldFictionBooks =
              from b in books
              where b.PublishedOn < Param<DateTime>() && (b.Category == BookCategory.Fiction)
              select new { b.Id, Price = Param<double>().Value };
      _bookDiscountOldFiction = CreateUpdate<IBook>("_BookDiscountOldFiction", "Sets discount price on fiction books published before certain date.", 
                                                    queryOldFictionBooks);
 */ 
    }

    // Static method computing FullName computed property for an Author
    public static string GetFullName(IAuthor author) {
      return author.FirstName + " " + author.LastName;
    }
    // Static method computing FullName computed property for an Author
    public static string GetOrderSummary(IBookOrder order) {
      return order.CreatedOn.ToString("s") + " " + order.User.DisplayName + ", Total: " + order.Total.ToString("###.##");
    }

    // Static method validating Book entity
    public static void ValidateBook(EntityValidator validator, IBook book) {
      validator.Check(book.Price > 0.01m, book.Price, "Book price must be greater than 1 cent.", null, "PriceNegative", "Price", book);
    }

    //Methods executing stored procedures 
/*
    public IList<IBook> GetBooksByAuthor(IEntitySession session, string lastName) {
      return session.ExecuteQuery<IBook>(_getBooksByAuthor, lastName);
    }

    public IBook GetLatestBookByPublisher(IEntitySession session, string pubName) {
      return session.ExecuteQuery<IBook>(_getPublisherLatestBook, pubName).FirstOrDefault();
    }

    public void ChangePrice(IEntitySession session, BookCategory category, double changePerc) {
      var factor = (100.0 + changePerc) / 100.0d;
      session.ExecuteNonQuery(_booksChangePrice, category, factor);
    }

    public void DeleteOldCheapBooks(IEntitySession session, DateTime date) {
      session.ExecuteNonQuery(_deleteOldCheapBooks, date);
    }

    public void DiscountOldFictionBooks(IEntitySession session, DateTime publishedBefore, double newPrice) {
      session.ExecuteNonQuery(_bookDiscountOldFiction, publishedBefore, newPrice);
    }
*/

  }//BooksModule 

}
