﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Vita.Activation;
using Vita.Data;

namespace Vita.Tutorial.Part1 {
  // Demo-ed features:
  //   Defining simple entities (Book and Publisher); basic CRUD operations, loading entities by ID, using LINQ-2-SQL for
  //   selecting entities from database; lasy loading of entity references. 

  public class Part1_Sample {
    public static void RunSample() {
      Console.Clear(); 
      Console.WriteLine("Running Sample.");
      Setup.SchemaManager.DropSchemaObjects("part1");

      Console.WriteLine("Opening an entity store...");
      var model = new Part1EntityModel();
      var entityStore = model.CreateEntityStore(new Vita.Data.MsSql.MsSqlDriver(), Setup.ConnectionString);
      Console.WriteLine("Success. Updated database schema, opened entity store.");

      //Create a publisher and 2 books
      Console.WriteLine("Creating sample entities: a publisher and 2 books...");
      var session = entityStore.OpenSession();
      var msPub = session.NewEntity<IPublisher>();
      msPub.Id = Guid.NewGuid(); 
      msPub.Name = "MS Publishing";
      //Create a book
      var bk = session.NewEntity<IBook>();
      bk.Id = Guid.NewGuid(); 
      bk.Author = "John Sharp";
      bk.Title = "c# Programming";
      bk.PublishedOn = DateTime.Today;
      bk.Publisher = msPub;
      bk.Price = 25.0;
      //Create another book
      bk = session.NewEntity<IBook>();
      bk.Id = Guid.NewGuid();
      bk.Author = "Jack Pound";
      bk.Title = "VB Programming";
      bk.PublishedOn = DateTime.Today;
      bk.Publisher = msPub;
      bk.Price = 20.0;
      session.SaveChanges(); //Submit to database
      Console.WriteLine("Success. Created a publisher and 2 books.");
      Console.WriteLine();
      //let's remember some IDs, we'll use it later
      var msPubId = msPub.Id; 
      var bk2Id = bk.Id; 

      Console.WriteLine("Loading list of created books ...");
      session = entityStore.OpenSession();
      var books = session.GetEntities<IBook>(take: 10);
      foreach (var b in books)
        Console.WriteLine("  Book: " + b.Title);
      Console.WriteLine("Loading a book by ID = " + bk2Id);
      var bk2 = session.GetEntity<IBook>(bk2Id);
      Console.WriteLine("  Success - loaded book '" + bk2.Title + "' by " + bk2.Author);
      Console.WriteLine();

      Console.WriteLine("Testing lazy loading - reading Publisher property of the book...");
      Console.WriteLine("  Book publisher: " + bk2.Publisher.Name);
      Console.WriteLine();

      Console.WriteLine("Using direct LINQ - finding books by author.");
      Console.WriteLine("  Printing books by John Sharp");
      var qBooks = from b in session.EntitySet<IBook>()
              where b.Author == "John Sharp"
              select b; 
      foreach(var b in qBooks)
        Console.WriteLine("  Book: '" + b.Title + "' by " + b.Author);
      Console.WriteLine();

      Console.WriteLine("Updating entities - changing publishing date to a year ago.");
      var allBooks = session.GetEntities<IBook>(take: 10);
      //Note: initially publishing date was Today
      foreach (var b in allBooks) 
        b.PublishedOn = DateTime.Today.AddYears(-1);
      session.SaveChanges();
      Console.WriteLine("  Verifying... ");
      session = entityStore.OpenSession(); //Let's open another session to make sure there are no entities cached in session
      var yesterday = DateTime.Today.AddDays(-1);
      var updatedBooks = from b in session.EntitySet<IBook>()
                 where b.PublishedOn < yesterday
                 select b;
      var updatedCount = updatedBooks.ToList().Count; 
      if (updatedCount != 2)
        throw new Exception("Books update failed!");
      Console.WriteLine(" Verified: #books with updated publish date: " + updatedCount);
      Console.WriteLine("Demo completed. Browse the database to see the objects created by the demo.");
      Console.WriteLine();
    
    }
  }

}
