﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Vita.Activation;
using Vita.Data;

namespace Vita.Tutorial.Part2 {

  //Demoed features:
  // 1. Using enumerations  
  // 2. Using extension methods for creating entities
  // 3. Attributes: Auto, Size, Nullable, Memo, OrderBy, Paged 
  // 4. Many-to-one and Many-to-many relations; Entity lists in properties; 
  // 5. Querying with paging; 

  
  public static class Part2_Sample {

    public static void RunSample() {
      Console.Clear();
      Console.WriteLine("Running Sample.");
      Setup.SchemaManager.DropSchemaObjects("part2");

      Console.WriteLine("Opening an entity store...");
      var model = new Part2EntityModel();
      var entityStore = model.CreateEntityStore(new Vita.Data.MsSql.MsSqlDriver(), Setup.ConnectionString);
      Console.WriteLine("Success. Updated database schema, opened entity store.");
      Console.WriteLine("Creating sample entities: publisher, authors, books.");
      var session = entityStore.OpenSession();
      //publisher
      var msPub = session.NewPublisher("MS Publishing");
      //authors
      var john = session.NewAuthor("John", "Sharp");
      var jack = session.NewAuthor("Jack", "Pound");
      //Books
      var csBook = session.NewBook(BookEdition.Hardcover, BookCategory.Programming, "c# Programming", "Expert c# programming", msPub, DateTime.Today.AddYears(-1), 25); 
      csBook.Authors.Add(john);
      var vbBook = session.NewBook(BookEdition.EBook | BookEdition.Paperback, BookCategory.Programming, "VB Programming", "Expert VB programming", msPub, 
         DateTime.Today, 15); 
      vbBook.Authors.Add(jack);
      vbBook.Authors.Add(john);

      session.SaveChanges(); //Submit to database
      Console.WriteLine("Success. Created a publisher and 2 books.");
      Console.WriteLine();
      //let's remember some IDs, we'll use them later
      var msPubId = msPub.Id;
      var vbBookId = vbBook.Id;
      session = entityStore.OpenSession(); //open new session, to make sure we read fresh entities from database

      Console.WriteLine("Many-to-one relations: loading a publisher, listing its books.");
      var pub = session.GetEntity<IPublisher>(msPubId);
      Console.WriteLine("  Loaded publisher: " + pub.Name);
      if (pub.Books.Count != 2) // we can access publisher.Books property directly - it is loaded when we first read it
        throw new Exception("Unexpected # of books for a publisher.");
      foreach(var bk in pub.Books)
        Console.WriteLine("  Book: " + bk.Title);
      Console.WriteLine();

      Console.WriteLine("Many-to-many relations: loading a book, listing its authors.");
      Console.WriteLine("                        using a FullName computed property.  ");
      var book = session.GetEntity<IBook>(vbBookId);
      Console.WriteLine("  Loaded a book: " + book.Title);
      if (book.Authors.Count != 2) // we can access book.Authors property directly - it is loaded when we first read it
        throw new Exception("Unexpected # of authors for VB book.");
      foreach(var author in book.Authors)
        Console.WriteLine("  Author: " + author.FullName);
      Console.WriteLine();

      Console.WriteLine("Paging - reading paged book list.");
      var allBooks = session.GetEntities<IBook>(take: 10);
      Console.WriteLine("  Loaded all books, count = " + allBooks.Count);
      // Books are ordered by Published date, desc. CS book is published a year ago, VB book is published today - so VB book is first in the list. 
      // If we query with skip=1, we should get CS book only.
      Console.WriteLine("  Loading all books with skip = 1 ...");
      var pagedBooks = session.GetEntities<IBook>(skip: 1, take: 1);
      if (pagedBooks.Count != 1) 
        throw new Exception("Unexpected # of paged books.");
      var title = pagedBooks[0].Title;
      Console.WriteLine("  Loaded book: " + title);
      if (title != "c# Programming")
        throw new Exception("Unexpected book in paged book list.");
      Console.WriteLine();

      Console.WriteLine("Demo completed. ");
      Console.WriteLine();
    }



  }//class
}//ns
