﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Vita.Activation;
using Vita.Entities;
using Vita.Data;

namespace Vita.Tutorial.Part4 {

  //Demoed features:
  // Computed members (FullName for IAuthor)
  // Entity validation; custom validation methods
  // Non-fatal exceptions:  ValidationException, UniqueIndexViolationException

  public static class Part4_Sample {

    public static void RunSample() {
      Console.Clear();
      Console.WriteLine("Running Sample.");
      Setup.SchemaManager.DropSchemaObjects("part4");

      Console.WriteLine("Opening an entity store...");
      var model = new Part4EntityModel();
      var entityStore = model.CreateEntityStore(new Vita.Data.MsSql.MsSqlDriver(), Setup.ConnectionString);
      Console.WriteLine("Success. Updated database schema, opened entity store.");
      Console.WriteLine("Creating sample entities: publisher, authors, books.");

      var session = entityStore.OpenSession();
      //publishers
      var msPub = session.NewPublisher("MS Publishing");
      //authors
      var john = session.NewAuthor("John", "Sharp");
      var jack = session.NewAuthor("Jack", "Pound");
      //Books
      var csBook = session.NewBook(BookEdition.Hardcover, BookCategory.Programming, "c# Programming", "Expert c# programming", msPub, DateTime.Today.AddYears(-1), 25); 
      csBook.Authors.Add(john);
      var vbBook = session.NewBook(BookEdition.EBook | BookEdition.Paperback, BookCategory.Programming, "VB Programming", "Expert VB programming", msPub, 
         DateTime.Today, 15); 
      vbBook.Authors.Add(jack);
      vbBook.Authors.Add(john);

      session.SaveChanges(); //Submit to database
      Console.WriteLine("Success. Created a publisher and 2 books.");
      Console.WriteLine();

      Console.WriteLine("Validation - submitting invalid entities, expecting 4 validation errors.");
      var invalidAuthor = session.NewAuthor(null, "VeryLoooooooooooooooooooooooooooongLaaaaaaaaaaaaaaaaaaaaaaaaaaaaaastNaaaaaaaaaaaaaaaaame");
      var invalidBook = session.NewBook(BookEdition.EBook, BookCategory.Fiction, "Not valid book", "Some invalid book", null, DateTime.Now, -5.0);
      // We expect 4 errors: Author's first name should not be null; Author's last name is too long; Publisher cannot be null;
      // Price must be > 1 cent. The first 3 errors are found by built-in validation; the last error, price check, is added
      // by custom validation method.
      try {
        session.SaveChanges();
      } catch (ValidationException vex) {
        var errCount = vex.Errors.Count;
        if(errCount != 4) 
          throw new Exception("Wrong number of validation errors; expected 4, received " + errCount);
        foreach (var err in vex.Errors)
          Console.WriteLine(" Error: " + err.Message);   // + "  (" + err.TargetEntity + "/" + err.TargetKey + ")"); // add these to see entity type and Id
      }
      Console.WriteLine();

      Console.WriteLine("Non-fatal exceptions: submitting duplicate publisher.");
      session = entityStore.OpenSession(); 
      var msPub2 = session.NewPublisher("MS Publishing");
      try {
        session.SaveChanges();
      } catch (UniqueIndexViolationException ex) {
        if (ex.KeyName == "PubName")
          Console.WriteLine("Duplicate publisher.");
        else //should not happen
          Console.WriteLine("Unrecognized unique index violation: " + ex.Message);
      }
      Console.WriteLine();

      Console.WriteLine("Demo completed. ");
      Console.WriteLine();
    }



  }//class
}//ns
