﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;

namespace Vita.Samples.BooksSample.UnitTests {

#if USE_NUNIT
  using NUnit.Framework;
  using TestClass = NUnit.Framework.TestFixtureAttribute;
  using TestMethod = NUnit.Framework.TestAttribute;
  using TestInitialize = NUnit.Framework.SetUpAttribute;
  using TestCategory = NUnit.Framework.CategoryAttribute;
#else
  using Microsoft.VisualStudio.TestTools.UnitTesting;
#endif

  using Vita.Common;
  using Vita.Entities;
  using Vita.Entities.Authorization;
  using Vita.Samples.BooksModel;
  using Vita.Entities.Runtime;


  [TestClass]
  public class AuthorizationTests {
    delegate void TestAction(); 

    [TestInitialize]
    public void TestInit() {
      SetupHelper.Init();
      SetupHelper.CreateSampleData();
    }

    [TestMethod, TestCategory("Vita_Books")]
    public void TestAuthorization() {
      // First let's open non-secure session, load Users and lookup ID's of some existing objects - we will use them in tests
      var entityStore = SetupHelper.EntityStore;
      var nonSecureSession = entityStore.OpenSession();
      // Preparation - find specific entities (books, users, etc)
      var allUsers = nonSecureSession.GetEntities<IUser>(take: 10);
      var allBooks = nonSecureSession.GetEntities<IBook>(take: 10);
      var allAuthors = nonSecureSession.GetEntities<IAuthor>(take: 10);
      var allPubs = nonSecureSession.GetEntities<IPublisher>(take: 10);
      var allOrders = nonSecureSession.GetEntities<IBookOrder>(take: 10);
      var dora = allUsers.First(u => u.UserName == "Dora");
      var diego = allUsers.First(u => u.UserName == "Diego");
      var lindaTheEditor = allUsers.First(u => u.UserName == "linda");
      var charlieTheManager = allUsers.First(u => u.UserName == "charlie");
      var johnTheAuthor = allUsers.First(u => u.UserName == "john");
      var jackTheAuthor = allUsers.First(u => u.UserName == "jack");
      // Get IDs of some entities
      var csBookId = allBooks.First(b=>b.Title.StartsWith("c#")).Id;
      var vbBookId = allBooks.First(b => b.Title.StartsWith("VB")).Id;
      var kidBookId = allBooks.First(b => b.Title.StartsWith("Three")).Id; //3 little piggies
      var authorJohnId = allAuthors.First(a => a.FirstName == "John").Id;
      var authorJackId = allAuthors.First(a => a.FirstName == "Jack").Id;
      var msPubId = allPubs.First(p => p.Name.StartsWith("MS")).Id;
      var kidPubId = allPubs.First(p => p.Name.StartsWith("Kid")).Id;
      var oldDiegoOrderId = allOrders.First(ord => ord.User == diego).Id;

      nonSecureSession = null; // to never use it after this

      #region WebVisitor Role ========================================================================================= 

      EnsurePasses("Web user can update his/her User record", () => {
        var secureSession = OpenSecureSession(dora);
        var doraRec = secureSession.GetEntity<IUser>(dora.Id);
        doraRec.DisplayName += " the Explorer"; //Dora the Explorer
        secureSession.SaveChanges();
      });

      EnsureFails(dora, "Web user cannot read other user's User record", () => {
        var secureSession = OpenSecureSession(dora);
        var diegoRec = secureSession.GetEntity<IUser>(diego.Id); //it should fail here
      });

      EnsurePasses("Web user can browse books, authors, publishers, reviews.", () => {
        var secureSession = OpenSecureSession(dora);
        secureSession.RequireReadAccess = ReadAccessLevel.Read; //Requires Read permission
        var csBook = secureSession.GetEntity<IBook>(csBookId);
        var csTitle = csBook.Title; //try read Title
        var authorJohn = secureSession.GetEntity<IAuthor>(authorJohnId);
        var johnFullName = authorJohn.FullName; //read IAuthor properties
        var johnBio = authorJohn.Bio;
        var reviews = secureSession.GetEntities<IBookReview>(take: 10);
        string reviewText;
        foreach (var rv in reviews)
          reviewText = rv.Review; //can access any review
      });

      var doraReview1Id = Guid.Empty;
      var doraReview2Id = Guid.Empty;
      EnsurePasses("Web user can write reviews", () => {
        var secureSession = OpenSecureSession(dora);
        var csBook = secureSession.GetEntity<IBook>(csBookId);
        var vbBook = secureSession.GetEntity<IBook>(vbBookId);
        var review1 = secureSession.NewReview(dora, csBook, 5, "Excellent!", "Recommend to everyone!");
        var review2 = secureSession.NewReview(dora, vbBook, 4, "Very good.", "Must read for VB programmers");
        secureSession.SaveChanges();
        doraReview1Id = review1.Id; //save it in a local var, wil use it in the next test
        doraReview2Id = review2.Id; 
      });

      EnsureFails(diego, "Web user cannot update other users' reviews.", () => {
        var secureSession = OpenSecureSession(diego); 
        var doraReview = secureSession.GetEntity<IBookReview>(doraReview1Id);
        doraReview.Review = " Not so good... "; // blows up
      });

      EnsureFails(diego, "Web user cannot create a review linked to other user.", () => {
        var secureSession = OpenSecureSession(diego); 
        var csBook = secureSession.GetEntity<IBook>(csBookId);
        var doraReview = secureSession.NewReview(dora, csBook, 3, "Dora likes it", "I know Dora likes it. Diego");
        secureSession.SaveChanges(); //should blow up here
      });

      EnsureFails(diego, "Web user cannot delete other users' reviews.", () => {
        var secureSession = OpenSecureSession(diego); 
        var doraReview = secureSession.GetEntity<IBookReview>(doraReview1Id);
        secureSession.DeleteEntity(doraReview);
      });

      EnsurePasses("Web user can update, delete his/her reviews.", () => {
        //Update
        var secureSession = OpenSecureSession(dora); //back to Dora
        var doraReview = secureSession.GetEntity<IBookReview>(doraReview1Id);
        doraReview.Review += " (update: some more info)";
        secureSession.SaveChanges();
        //Delete
        secureSession = OpenSecureSession(dora);
        doraReview = secureSession.GetEntity<IBookReview>(doraReview1Id);
        secureSession.DeleteEntity(doraReview); // we test that we can call this method
        // secureSession.SaveChanges(); do not actually delete
      });


      Guid newDoraOrderId = Guid.Empty;
      EnsurePasses("Web user can buy books, use coupons", () => {
        var secureSession = OpenSecureSession(dora);
        var doraOrder = secureSession.NewOrder(dora);
        var csBook = secureSession.GetEntity<IBook>(csBookId);
        doraOrder.Add(csBook, 1);
        doraOrder.CompleteOrder("C2"); // C2 is coupon code; this will lookup coupon entity
        secureSession.SaveChanges();
        newDoraOrderId = doraOrder.Id; //save it in a local var, will use it later
      });

      // WebUser can Peek coupons (order-creation code looks up coupon by code), but cannot Read (see) them.
      EnsureFails(dora, "Web user cannot READ coupon entities.", () => {
        var secureSession = OpenSecureSession(dora);
        secureSession.RequireReadAccess = ReadAccessLevel.Read; // Means that property reads are for showing stuff to the User 
        var coupon = secureSession.GetEntities<ICoupon>(take: 10).First(); //That will work, because WebUser has Peek permission
        var promoCode = coupon.PromoCode;  //should blow up
      });

      EnsureFails(diego, "Web user cannot see other users' orders", () => {
        var secureSession = OpenSecureSession(diego); 
        var doraOrder = secureSession.GetEntity<IBookOrder>(newDoraOrderId); //should blow up
      });

      EnsureFails(diego, "Web user cannot delete other users' orders", () => {
        var secureSession = OpenSecureSession(diego); 
        //TODO: review returning RecordStubs later; might need better authorization enforcement
        // Comment: we do a trick here - we load a record Stub - an empty record with PrimaryKey initialized.  
        // System does not check access rights for stubs initially, to avoid forcing unnecessary record load. 
        // But it will check permissions as soon as we try to access record values, or delete the record
        var doraOrder = secureSession.GetEntity<IBookOrder>(newDoraOrderId, LoadFlags.Stub); //does not fail yet
        secureSession.DeleteEntity(doraOrder); // blows up
      });

      EnsureFails(dora, "Web user cannot update books", () => {
        var secureSession = OpenSecureSession(dora);
        var csBook = secureSession.GetEntity<IBook>(csBookId);
        csBook.Price += 1; //should fail here
      });

      EnsureFails(dora, "Web user cannot delete books", () => {
        var secureSession = OpenSecureSession(dora);
        var csBook = secureSession.GetEntity<IBook>(csBookId);
        secureSession.DeleteEntity(csBook); //blows here
      });

      EnsureFails(dora, "Web user cannot update Authors", () => {
        var secureSession = OpenSecureSession(dora);
        var authorJohn = secureSession.GetEntity<IAuthor>(authorJohnId);
        authorJohn.Bio = "Spam spam spam"; // blows up
      });

      EnsureFails(dora, "Web user cannot update Publishers", () => {
        var secureSession = OpenSecureSession(dora);
        var msPub = secureSession.GetEntity<IPublisher>(msPubId);
        msPub.Name = "Microsoft Publishing"; //fails here
      });
      #endregion

      #region Author Role =============================================================
      EnsurePasses("Author can edit his Bio and his book description in scope of AuthorEditGrant.", () => {
        var secureSession = OpenSecureSession(johnTheAuthor);
        using (BooksAuthorization.AuthorEditGrant.Execute(secureSession.UserContext)) {
          var john = secureSession.GetEntity<IAuthor>(authorJohnId);
          john.Bio = "Some new Bio";
          var csBook = secureSession.GetEntity<IBook>(csBookId);
          csBook.Description = "New description";
          secureSession.SaveChanges();
        }
      });
      EnsureFails(johnTheAuthor, "Author cannot edit his LastName in scope of AuthorEditGrant.", () => {
        var secureSession = OpenSecureSession(johnTheAuthor);
        using (BooksAuthorization.AuthorEditGrant.Execute(secureSession.UserContext)) {
          var john = secureSession.GetEntity<IAuthor>(authorJohnId);
          john.LastName = "Sharpp";
        }
      });
      EnsureFails(johnTheAuthor, "Author cannot edit his book Title even in scope of AuthorEditGrant.", () => {
        var secureSession = OpenSecureSession(johnTheAuthor);
        using (BooksAuthorization.AuthorEditGrant.Execute(secureSession.UserContext)) {
          var csBook = secureSession.GetEntity<IBook>(csBookId);
          csBook.Title = "c# programming for experts";
        }
      });
      EnsureFails(johnTheAuthor, "Author cannot change his book Price even in scope of AuthorEditGrant.", () => {
        var secureSession = OpenSecureSession(johnTheAuthor);
        using (BooksAuthorization.AuthorEditGrant.Execute(secureSession.UserContext)) {
          var csBook = secureSession.GetEntity<IBook>(csBookId);
          csBook.Price += 10;
        }
      });
      EnsureFails(johnTheAuthor, "Author cannot edit his Bio outside AuthorEditGrant.", () => {
        var secureSession = OpenSecureSession(johnTheAuthor);
        var john = secureSession.GetEntity<IAuthor>(authorJohnId);
        john.Bio = "Some another Bio";
      });
      EnsureFails(johnTheAuthor, "Author cannot edit his book description outside AuthorEditGrant.", () => {
        var secureSession = OpenSecureSession(johnTheAuthor);
        var csBook = secureSession.GetEntity<IBook>(csBookId);
        csBook.Description = "New description";
      });
      EnsureFails(johnTheAuthor, "Author cannot edit other author's Bio even in scope of AuthorEditGrant.", () => {
        var secureSession = OpenSecureSession(johnTheAuthor);
        using (BooksAuthorization.AuthorEditGrant.Execute(secureSession.UserContext)) {
          var jack = secureSession.GetEntity<IAuthor>(authorJackId);
          jack.Bio = "Corrected Bio";
        }
      });
      EnsureFails(johnTheAuthor, "Author cannot edit a random book Description even in scope of AuthorEditGrant.", () => {
        var secureSession = OpenSecureSession(johnTheAuthor);
        using (BooksAuthorization.AuthorEditGrant.Execute(secureSession.UserContext)) {
          var kidBook = secureSession.GetEntity<IBook>(kidBookId);
          kidBook.Description = "New description";
        }
      });
      EnsureFails(johnTheAuthor, "Author cannot access any orders.", () => {
        var secureSession = OpenSecureSession(johnTheAuthor);
        var doraOrder = secureSession.GetEntity<IBookOrder>(newDoraOrderId);
      });
      EnsurePasses("Author can read Publishers.", () => {
        var secureSession = OpenSecureSession(johnTheAuthor);
        var pub = secureSession.GetEntity<IPublisher>(msPubId);
        var pubName = pub.Name;
        pub = secureSession.GetEntity<IPublisher>(kidPubId);
        pubName = pub.Name;
      });
      EnsureFails(johnTheAuthor, "Author cannot update publishers.", () => {
        var secureSession = OpenSecureSession(johnTheAuthor);
        var pub = secureSession.GetEntity<IPublisher>(msPubId);
        pub.Name = "Microsoft Publishing";
      });
      EnsureFails(johnTheAuthor, "Author cannot create publishers.", () => {
        var secureSession = OpenSecureSession(johnTheAuthor);
        var pub = secureSession.NewEntity<IPublisher>();
      });
      #endregion

      #region BookEditor Role =============================================================
      EnsurePasses("Book editor can create/update/delete authors, books.", () => {
        var secureSession = OpenSecureSession(lindaTheEditor);
        var kidPub = secureSession.GetEntity<IPublisher>(kidPubId);
        var mtwain = secureSession.NewAuthor("Mark", "Twain");
        var mtwainBook = secureSession.NewBook(BookEdition.Paperback, BookCategory.Fiction,
            "Tom Soyer", "Tom Soyer adventures.", kidPub, DateTime.Now.AddYears(-1).Date, 10);
        mtwainBook.Authors.Add(mtwain);
        secureSession.SaveChanges(); 
        // now let's delete it; note that IBookAuthor link is deleted automatically through CascadeDelete
        // attribute on IBookAuthor.Book property. 
        secureSession.DeleteEntity(mtwainBook);
        secureSession.DeleteEntity(mtwain);
        secureSession.SaveChanges();
      });
      EnsureFails(lindaTheEditor, "Book editor cannot create publishers.", () => {
        var secureSession = OpenSecureSession(lindaTheEditor);
        var pub = secureSession.NewEntity<IPublisher>();
      });
      EnsureFails(lindaTheEditor, "Book editor cannot update publishers.", () => {
        var secureSession = OpenSecureSession(lindaTheEditor);
        var msPub = secureSession.GetEntity<IPublisher>(msPubId);
        msPub.Name = "Microsoft Publishing";
      });
      EnsureFails(lindaTheEditor, "Book editor cannot delete reviews.", () => {
        var secureSession = OpenSecureSession(lindaTheEditor);
        var doraReview = secureSession.GetEntity<IBookReview>(doraReview2Id);
        secureSession.DeleteEntity(doraReview);
      });

      #endregion

      #region StoreManager Role =============================================================
      EnsurePasses("StoreManager can create/delete publishers.", () => {
        var secureSession = OpenSecureSession(charlieTheManager);
        var pub = secureSession.NewPublisher("SciFi Publishing");
        secureSession.SaveChanges();
        secureSession.DeleteEntity(pub);
        secureSession.SaveChanges();
      });
      EnsurePasses("StoreManager can delete reviews.", () => {
        var secureSession = OpenSecureSession(charlieTheManager);
        var doraReview = secureSession.GetEntity<IBookReview>(doraReview2Id);
        secureSession.DeleteEntity(doraReview);
        secureSession.SaveChanges(); 
      });
      EnsurePasses("StoreManager can create/delete coupons.", () => {
        var secureSession = OpenSecureSession(charlieTheManager);
        var coupon = secureSession.NewCoupon("S01", 10, DateTime.Now.AddMonths(3));
        secureSession.SaveChanges();
        secureSession.DeleteEntity(coupon);
        secureSession.SaveChanges();
      });
      EnsurePasses("StoreManager can adjust orders within ManagerAdjustOrderGrant.", () => {
        var secureSession = OpenSecureSession(charlieTheManager);
        using (BooksAuthorization.ManagerAdjustOrderGrant.Execute(secureSession.UserContext, null, newDoraOrderId)) {
          var doraOrder = secureSession.GetEntity<IBookOrder>(newDoraOrderId);
          // Verify that we can modify IBookOrder and IBookOrderLine entities
          doraOrder.Lines[0].Price -= 2; // make a discount on first line
          doraOrder.Total -= 2; // change total
          secureSession.SaveChanges();
        }
      });
      EnsureFails(charlieTheManager, "StoreManager can adjust ONLY the order for which the ManagerAdjustOrderGrant started.", () => {
        var secureSession = OpenSecureSession(charlieTheManager);
        //We activate the grant for doraOrderId, but will try to modify diego's order inside - this should fail
        using (BooksAuthorization.ManagerAdjustOrderGrant.Execute(secureSession.UserContext, null, newDoraOrderId)) {
          var diegoOrder = secureSession.GetEntity<IBookOrder>(oldDiegoOrderId);
          diegoOrder.Total += -2; // should blow up
        }
      });
      #endregion

      #region Retrieving access rights without actually accessing the data
      // For retrieving rights at table level, use ISecureSession.IsActionAllowed<TEntity>(action) method. 
      // To retrieve rights for particular entity and property, use EntityHelper.GetEntityAccess helper method to get the descriptor, 
      // and then use the descriptor to get the detailed info.
      {
        //Check that book editor can edit books, but cannot edit publishers
        var secureSession = OpenSecureSession(lindaTheEditor);
        Assert.IsTrue(secureSession.IsActionAllowed<IBook>(Actions.CRUD), 
                        "Invalid access descriptor: editor can create, update, delete books.");
        Assert.IsFalse(secureSession.IsActionAllowed<IPublisher>(Actions.UpdateStrict), "Invalid access descriptor: editor cannot update publishers.");
        // Check that author can edit book Abstract, but not Title (within AuthorEditGrant)
        secureSession = OpenSecureSession(johnTheAuthor);
        using (BooksAuthorization.AuthorEditGrant.Execute(secureSession.UserContext)) {
          var csBook = secureSession.GetEntity<IBook>(csBookId);
          var csBookAccess = EntityHelper.GetEntityAccess(csBook);
          Assert.IsTrue(csBookAccess.CanUpdate<IBook>(bk => bk.Abstract), "Failed access check: author can update book abstract.");
          // the same with different overload
          Assert.IsTrue(csBookAccess.CanUpdate("Abstract"), "Failed access check: author can update book abstract.");
          // should not be able to update title
          Assert.IsFalse(csBookAccess.CanUpdate<IBook>(bk => bk.Title), "Failed access check: author cannot update the title.");
        }
      }
      #endregion

    }//method



    // Helpers and Utilities

    private void EnsurePasses(string testStep, TestAction action) {
      try {
        action.Invoke();
      } catch (AuthorizationException authExc) {
        Debug.WriteLine("Test step: " + testStep);
        //Report user's rights to entity from AuthorizationException
        Debug.WriteLine("AuthorizationException Summary:\r\n" + authExc.Summary);
        throw;
      } catch (Exception) {
        Debug.WriteLine("Test step: " + testStep);
        throw; 
      }
    }

    //User is used for reporting permissons to Debug output, in case anticipated exception is not thrown
    private void EnsureFails(IUser user, string testStep, TestAction action) {
      try {
        action.Invoke();
        //if invoke did not fail, throw. 
        throw new Exception("Anticipated exception was not thrown. Test: " + testStep);
      } catch (AuthorizationException) {
        return; // it is expected, so suppress it
      } 
    }

    // Note: Authority lookup might be further optimized by saving authority.Key for the user and then looking up by this key: authService.TryGetAuthority(authKey) 
    private ISecureSession OpenSecureSession(IUser user) {
      var authService = SetupHelper.BooksModel.Authorization; 
      var roles = BooksAuthorization.GetRoles(user);
      var authority = authService.GetAuthority(roles); 
      var userContext = new UserContext(user.UserName, authority, user.Id, user);
      userContext.Values[BooksAuthorization.UserIdKey] = user.Id;
      return authService.OpenSecureSession(userContext);
    }
  
  }//class
}
