﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;
using System.Data.SqlClient;

using Vita.Data;

namespace Vita.Samples.BooksSample.UnitTests {

#if USE_NUNIT
  using NUnit.Framework;
  using TestClass = NUnit.Framework.TestFixtureAttribute;
  using TestMethod = NUnit.Framework.TestAttribute;
  using TestInitialize = NUnit.Framework.SetUpAttribute;
  using TestCategory = NUnit.Framework.CategoryAttribute;
#else
  using Microsoft.VisualStudio.TestTools.UnitTesting;
#endif

#if MSSQL
  [TestClass]
#endif
  public class BatchPerfTest {
    [TestInitialize]
    public void TestInit() {
      SetupHelper.Init();
    }
    [TestCleanup]
    public void TearDown() {
      SetupHelper.TearDown();
    }

#if MSSQL
    [TestMethod, TestCategory("Vita_Books")]
#endif
    // A simple run to estimate batch perf improvement - this feature is coming, so this is estimate of expected benefit. 
    // We execute the same statements creating sample data as Books\SampleDataGenerator.cs does. 
    // executing updates one-by-one takes about 20 ms for 2 transactions. Running one big text batch with SP calls takes 2-3 ms (not consistently, but best cases show this). 
    public void TestCompareBatchPerformance() {
      //If we are running without stored procs, this test can't be run
      var connString = SetupHelper.GetConnectionString();
      var conn = new SqlConnection(connString);
      var cmd = conn.CreateCommand();
      //delete all
      conn.Open();
      cmd.CommandText = SqlDeleteAll;
      cmd.ExecuteNonQuery();
      conn.Close();
      //test batch
      cmd.CommandText = SqlCreateBooksAuthorsOrders;
      conn.Open();
      var sw = new Stopwatch();
      sw.Start();
      var start = sw.ElapsedMilliseconds;
      cmd.ExecuteNonQuery();
      var end = sw.ElapsedMilliseconds;
      conn.Close();
      Trace.WriteLine("Batch time: " + (end - start)); //This will appear in test results report
    }

    const string SqlDeleteAll = @"
delete from books.BookOrderLine;
delete from books.BookOrder;
delete from books.Coupon;
delete from books.BookReview;
delete from books.Book
delete from books.Publisher;
delete from books.Author;
delete from books.[User];
";

    // This SQL was directly copied from VITA's app Log after test run. A nice thing about Log is that it produces directly re-executable SQL.
    const string SqlCreateBooksAuthorsOrders = @"
BEGIN TRANSACTION;
Exec [books].[PublisherInsert] 'b463a6b2-3f25-11e2-b998-f0def1783701', 'Kids Books'; -- @9 ms , 12/05/2012 13:49:55 
Exec [books].[PublisherInsert] 'b463a6b1-3f25-11e2-b998-f0def1783701', 'MS Books'; -- @4 ms , 12/05/2012 13:49:55 
Exec [books].[UserInsert] 'b463a6c1-3f25-11e2-b998-f0def1783701', 'jack', 'jack', 2/*Author*/; -- @4 ms , 12/05/2012 13:49:55 
Exec [books].[UserInsert] 'b463a6c0-3f25-11e2-b998-f0def1783701', 'john', 'john', 2/*Author*/; -- @4 ms , 12/05/2012 13:49:55 
Exec [books].[UserInsert] 'b463a6bf-3f25-11e2-b998-f0def1783701', 'charlie', 'charlie', 8/*StoreAdmin*/; -- @3 ms , 12/05/2012 13:49:55 
Exec [books].[UserInsert] 'b463a6be-3f25-11e2-b998-f0def1783701', 'linda', 'linda', 4/*BookEditor*/; -- @3 ms , 12/05/2012 13:49:55 
Exec [books].[UserInsert] 'b463a6bd-3f25-11e2-b998-f0def1783701', 'Diego', 'Diego', 1/*WebVisitor*/; -- @3 ms , 12/05/2012 13:49:55 
Exec [books].[UserInsert] 'b463a6bc-3f25-11e2-b998-f0def1783701', 'Dora', 'Dora', 1/*WebVisitor*/; -- @3 ms , 12/05/2012 13:49:55 
Exec [books].[AuthorInsert] 'b463a6b4-3f25-11e2-b998-f0def1783701', 'Jack', 'Pound', NULL, 'b463a6c1-3f25-11e2-b998-f0def1783701'; -- @4 ms , 12/05/2012 13:49:55 
Exec [books].[AuthorInsert] 'b463a6b3-3f25-11e2-b998-f0def1783701', 'John', 'Sharp', '  Lorem ipsum dolor ...  ...pteur sint occaecat ', 'b463a6c0-3f25-11e2-b998-f0def1783701'; -- @3 ms , 12/05/2012 13:49:55 
Exec [books].[BookInsert] 'b463a6b8-3f25-11e2-b998-f0def1783701', 'Windows programming', 'Introduction to Windows Programming', [2001-12-05T00:00:00], NULL, 0/*Programming*/, 2/*Hardcover*/, 30, 'b463a6b1-3f25-11e2-b998-f0def1783701'; -- @5 ms , 12/05/2012 13:49:55 
Exec [books].[BookInsert] 'b463a6b7-3f25-11e2-b998-f0def1783701', 'Three little pigs', 'Folk tale', [2011-12-05T00:00:00], NULL, 2/*Kids*/, 2/*Hardcover*/, 10, 'b463a6b2-3f25-11e2-b998-f0def1783701'; -- @4 ms , 12/05/2012 13:49:55 
Exec [books].[BookInsert] 'b463a6b6-3f25-11e2-b998-f0def1783701', 'VB Programming', 'Expert programming in VB', [2011-12-05T00:00:00], NULL, 0/*Programming*/, 3/*Paperback, Hardcover*/, 25, 'b463a6b1-3f25-11e2-b998-f0def1783701'; -- @3 ms , 12/05/2012 13:49:55 
Exec [books].[BookInsert] 'b463a6b5-3f25-11e2-b998-f0def1783701', 'c# Programming', 'Expert programming in c#', [2011-12-05T00:00:00], 'Expert guide to prog...  ...rts.  Covers c# 4.0.', 0/*Programming*/, 5/*Paperback, EBook*/, 20, 'b463a6b1-3f25-11e2-b998-f0def1783701'; -- @3 ms , 12/05/2012 13:49:55 
Exec [books].[CouponInsert] 'b463a6b9-3f25-11e2-b998-f0def1783701', [2012-12-05T21:49:55], [2013-01-05T13:49:55], 'C1', 10, NULL; -- @5 ms , 12/05/2012 13:49:55 
Exec [books].[CouponInsert] 'b463a6ba-3f25-11e2-b998-f0def1783701', [2012-12-05T21:49:55], [2013-01-05T13:49:55], 'C2', 10, NULL; -- @3 ms , 12/05/2012 13:49:55 
Exec [books].[CouponInsert] 'b463a6bb-3f25-11e2-b998-f0def1783701', [2012-12-05T21:49:55], [2013-01-05T13:49:55], 'C3', 10, NULL; -- @3 ms , 12/05/2012 13:49:55 
Exec [books].[BookAuthorInsert] 'b463a6b5-3f25-11e2-b998-f0def1783701', 'b463a6b3-3f25-11e2-b998-f0def1783701'; -- @4 ms , 12/05/2012 13:49:55 
Exec [books].[BookAuthorInsert] 'b463a6b5-3f25-11e2-b998-f0def1783701', 'b463a6b4-3f25-11e2-b998-f0def1783701'; -- @4 ms , 12/05/2012 13:49:55 
Exec [books].[BookAuthorInsert] 'b463a6b6-3f25-11e2-b998-f0def1783701', 'b463a6b4-3f25-11e2-b998-f0def1783701'; -- @3 ms , 12/05/2012 13:49:55 
Exec [books].[BookAuthorInsert] 'b463a6b8-3f25-11e2-b998-f0def1783701', 'b463a6b3-3f25-11e2-b998-f0def1783701'; -- @3 ms , 12/05/2012 13:49:55 
COMMIT TRANSACTION; --  time: 14ms 

BEGIN TRANSACTION;
Exec [books].[BookOrderInsert] 'b463a6c6-3f25-11e2-b998-f0def1783701', [2012-12-05T21:49:55], 45, 1/*Completed*/, 'b463a6bd-3f25-11e2-b998-f0def1783701'; -- @5 ms , 12/05/2012 13:49:55 
Exec [books].[BookOrderInsert] 'b463a6c3-3f25-11e2-b998-f0def1783701', [2012-12-05T21:49:55], 36, 1/*Completed*/, 'b463a6bc-3f25-11e2-b998-f0def1783701'; -- @4 ms , 12/05/2012 13:49:55 
Exec [books].[CouponUpdate] 'b463a6b9-3f25-11e2-b998-f0def1783701', [2013-01-05T13:49:55], 'C1', 10, [2012-12-05T13:49:55]; -- @8 ms , 12/05/2012 13:49:55 
Exec [books].[BookOrderLineInsert] 'b463a6c4-3f25-11e2-b998-f0def1783701', 1, 1, 20, 'b463a6c3-3f25-11e2-b998-f0def1783701', 'b463a6b5-3f25-11e2-b998-f0def1783701'; -- @5 ms , 12/05/2012 13:49:55 
Exec [books].[BookOrderLineInsert] 'b463a6c5-3f25-11e2-b998-f0def1783701', 2, 2, 10, 'b463a6c3-3f25-11e2-b998-f0def1783701', 'b463a6b7-3f25-11e2-b998-f0def1783701'; -- @5 ms , 12/05/2012 13:49:55 
Exec [books].[BookOrderLineInsert] 'b463a6c7-3f25-11e2-b998-f0def1783701', 1, 1, 25, 'b463a6c6-3f25-11e2-b998-f0def1783701', 'b463a6b6-3f25-11e2-b998-f0def1783701'; -- @3 ms , 12/05/2012 13:49:55 
Exec [books].[BookOrderLineInsert] 'b463a6c8-3f25-11e2-b998-f0def1783701', 2, 1, 20, 'b463a6c6-3f25-11e2-b998-f0def1783701', 'b463a6b5-3f25-11e2-b998-f0def1783701'; -- @4 ms , 12/05/2012 13:49:55 
Exec [books].[BookReviewInsert] 'b463a6c9-3f25-11e2-b998-f0def1783701', [2012-12-05T21:49:55], 5, 'Very interesting book!', 'Liked it very much!', 'b463a6b5-3f25-11e2-b998-f0def1783701', 'b463a6bc-3f25-11e2-b998-f0def1783701'; -- @5 ms , 12/05/2012 13:49:55 
Exec [books].[BookReviewInsert] 'b463a6ca-3f25-11e2-b998-f0def1783701', [2012-12-05T21:49:55], 1, 'Not worth it.', 'Did not like it at all.', 'b463a6b6-3f25-11e2-b998-f0def1783701', 'b463a6bd-3f25-11e2-b998-f0def1783701'; -- @4 ms , 12/05/2012 13:49:55 
COMMIT TRANSACTION; --  time: 7ms 
";
  
  }//class
}//ns
