﻿using System;
using System.ComponentModel;
using System.Collections.Generic;
using System.Linq;
using System.Diagnostics;
using System.Data.SqlClient;
using System.Threading;

using Vita.Entities;
using Vita.Entities.Runtime;
using Vita.Modules.ErrorLogging;

using Vita.Samples.BooksModel;

namespace Vita.Samples.BooksSample.UnitTests {

#if USE_NUNIT
  using NUnit.Framework;
  using TestClass = NUnit.Framework.TestFixtureAttribute;
  using TestMethod = NUnit.Framework.TestAttribute;
  using TestInitialize = NUnit.Framework.SetUpAttribute;
  using TestCategory = NUnit.Framework.CategoryAttribute;
#else
  using Microsoft.VisualStudio.TestTools.UnitTesting;
  using Vita.Data.Caching;
  using Vita.Services;
#endif

  [TestClass]
  public class LinqTests {

    [TestInitialize]
    public void TestInit() {
#if USE_OLD_DATA
      SetupHelper.Init(false);
#else
      SetupHelper.Init();
      SetupHelper.CreateSampleData();
#endif
    }

    [TestCleanup]
    public void TearDown() {
      SetupHelper.TearDown(); 
    }

    #region Helper objects for LINQ test
    public const string PublisherNameConst = "c# const string"; // to use in tests, to check how Linq handles references to constants
    // field and prop to test local evaluation in LINQ queries
    private string _someClassField;
    private string SomeClassProp { get { return "boo!"; } }
    public static string SomeStaticClassField;

    //Helper class to use as output in queries - testing LINQ engine with custom (non-anon) output types
    class BookInfo {
      public string Title;
      public string Publisher;
      public decimal Price;

      public BookInfo() { }
      public BookInfo(string title, string publisher, decimal price) {
        Title = title;
        Publisher = publisher;
        Price = price;
      }
    }

    //An artificial container for query parameters - to check how LINQ engine handles parameters in the form qp.Field or qp.Prop
    class QueryParamsContainer {
      public string Field;
      public string Prop { get; set; }
    }
    #endregion

    [TestMethod, TestCategory("Vita_Books")]
    public void TestLinq() {
      var entityStore = SetupHelper.EntityStore;
      //Init
      var session = entityStore.OpenSession();
      var books = session.EntitySet<IBook>();
      var authors = session.EntitySet<IAuthor>();
      var users = session.EntitySet<IUser>();
      var pubs = session.EntitySet<IPublisher>();

      var csTitle = "c# Programming";
      var vbTitle = "VB Programming";

      // book by title
      var qBooksByTitle = from b in books
                          where b.Title == csTitle
                          select b;
      var lstBooksByTitle = qBooksByTitle.ToList();
      Assert.IsTrue(lstBooksByTitle.Count > 0, "Books by title failed.");

      // Join through nullable reference - this does not work with cached lists!!!
      if (!SetupHelper.CacheEnabled) {
        var qAuthorsJack = from a in authors
                           where a.User.UserName == "jack"
                           select a;
        var lstAuthorsJack = qAuthorsJack.ToList();
        Assert.AreEqual(1, lstAuthorsJack.Count, "Failed to find author by user name");
      }

      // Books by publisher's name; skip, take test
      var msbooks = from b in books
                    where b.Publisher.Name == "MS Books"
                    orderby b.Title
                    select b;
      var msBookList = msbooks.ToList();
      Assert.IsTrue(3 == msBookList.Count, "Invalid number of MS books.");
      var msbook0 = msBookList[0]; //c# book should be the first book when sorted by title
      Assert.IsTrue(msbook0.Title == "c# Programming", "Invalid title of c# book.");
      //Same with skip, take
      var qSkipTake = msbooks.Skip(1).Take(1);
      var lstSkipTake = qSkipTake.ToList();
      Assert.IsTrue(lstSkipTake.Count == 1, "Invalid # of books in skip/take.");
      
      // Test bitwise op in SQL. Finding books by editions - we use bit-wise operation on book.Editions property - which is a flagset enum
      session = entityStore.OpenSession();
      books = session.EntitySet<IBook>();
      var eBooks = from b in books
                   where (b.Editions & BookEdition.EBook) != 0
                   select b;
      var eBooksList = eBooks.ToList();
      Assert.IsTrue(eBooksList.Count == 1, "Invalid number of e-books");

      // Using == for entity objects in queries
      var progCat = BookCategory.Programming;
      var msPub = session.EntitySet<IPublisher>().Single(p => p.Name == "MS Books");
      var msbooks2 = from b in books
                     where b.Publisher == msPub && b.Category == progCat
                     orderby b.Title
                     select b;
      msBookList = msbooks2.ToList();
      Assert.IsTrue(msBookList.Count > 0, "Query with entity object == comparison failed.");

      //BookAuthors link table
      var bookAuthors = session.EntitySet<IBookAuthor>();
      var aJack = session.EntitySet<IAuthor>().Where(a => a.FirstName == "Jack").First();
      var qBA = from ba in bookAuthors
                where ba.Author == aJack && ba.Book.Title == "c# Programming"
                select ba;
      var lstBAs = qBA.ToList();
      Assert.IsTrue(lstBAs.Count > 0, "Failed to find book-author record by author and book title.");

      // LINQ with FirstOrDefault() - these are executed slightly differently from queries that return result sets
      session = entityStore.OpenSession();
      var msp = (from p in session.EntitySet<IPublisher>()
                 where p.Name == "MS Books"
                 select p).FirstOrDefault();
      Assert.IsTrue(msp != null, "Query with FirstOrDefault failed.");
      Assert.AreEqual("MS Books", msp.Name, "Query with FirstOrDefault failed, invalid name.");
      //Check that entity is attached to session
      var mspRec = EntityHelper.GetRecord(msp);
      Assert.IsTrue(mspRec.Session != null, "FirstOrDefault query failed: the result record is not attached to session.");
      Assert.IsTrue(msp.Books.Count > 0, "Failed to read publisher's books: entity is not attached to session.");
      
      // Query that returns a derived entity, different from the original 'EntitySet' entity
      session = entityStore.OpenSession();
      //Return publishers of books in Kids category - we have one kids book, so expect one pub record
      var kidPubs = from b in session.EntitySet<IBook>()
                    where b.Category == BookCategory.Kids
                    select b.Publisher;
      var listPubs = kidPubs.ToList();
      Assert.AreEqual(1, listPubs.Count, "Unexpected pub count in special LINQ query.");
      // The same but with FirstOrDefault
      var firstKidPub = (from b in session.EntitySet<IBook>()
                         where b.Category == BookCategory.Kids
                         select b.Publisher).FirstOrDefault();
      Assert.IsNotNull(firstKidPub, "First kid Publisher is null in special LINQ query.");
      
      // Test FirstOrDefault with null result - special case for EntityCache.CloneEntity method
      session = entityStore.OpenSession();
      var none = session.EntitySet<IBook>().Where(b => b.Title == "123").FirstOrDefault();
      Assert.IsNull(none, "FirstOrDefault with null result failed.");

      // Query returning anonymous type
      // Return pairs book-title, price
      session = entityStore.OpenSession();
      var someId = Guid.Empty;
      var qryBooksAnon = from b in session.EntitySet<IBook>()
                         where b.Category == BookCategory.Programming
                         select new { Id = b.Id.ToString(), Title = b.Title, Price = b.Price }; //testing how ToString() works with Guids - only in select output!
      var listBooksAnon = qryBooksAnon.ToList();
      Assert.IsTrue(listBooksAnon.Count > 0, "Failed to retrieve anon type.");
      Assert.IsTrue(!string.IsNullOrEmpty(listBooksAnon[0].Id), "ToString query translation failed");

      //Same but with parameter in output 
#if !SQLCE 
      //SQL CE does not allow this (query like: SELECT Title, Price, @P3)
      var someKey = "someKey"; //some value to include into results - testing how such field initialized from parameter is handled
      var qryBooksAnon2 = from b in session.EntitySet<IBook>()
                        where b.Category == BookCategory.Programming
                        select new { b.Title, b.Price, Key = someKey };
      var listBooksAnon2 = qryBooksAnon2.ToList();
      Assert.IsTrue(listBooksAnon2.Count > 0, "Failed to retrieve anon type.");
#endif

      // Return pairs of (book, publisher)
      session = entityStore.OpenSession();
      var bpPairs = from b in session.EntitySet<IBook>()
                        where b.Category == BookCategory.Programming
                        select new { Book = b, Publisher = b.Publisher };
      var bpPairsList = bpPairs.ToList();
      Assert.IsTrue(bpPairsList.Count > 0, "Failed to retrieve anon type.");

      // Return anon type with pair of entities, one of them is nullable
      session = entityStore.OpenSession();
      var qAuthorUser = from a in authors
               select new { Author = a, User = a.User };
      var listAuthorUser = qAuthorUser.ToList();
      Assert.IsTrue(listAuthorUser.Count > 0, "Author-User query failed.");
      // Author Jim Hacker is not a user, so its User prop should be null
      var jimInfo = listAuthorUser.First(au => au.Author.LastName == "Hacker");
      Assert.IsTrue(jimInfo.User == null, "User prop is not null for Author Jim Hacker");

      // query with custom type in output (not anon type)
      var qBkInfos = from b in books
                     where b.Price > 1
                     select new BookInfo() { Title = b.Title, Publisher = b.Publisher.Name, Price = b.Price };
      var lstBkInfos = qBkInfos.ToList();
      Assert.IsTrue(lstBkInfos.Count > 0, "BookInfo query failed.");
      // Same with non-default constructor
      qBkInfos = from b in books
                 where b.Price > 1
                 select new BookInfo(b.Title, b.Publisher.Name, b.Price);
      lstBkInfos = qBkInfos.ToList();
      Assert.IsTrue(lstBkInfos.Count > 0, "BookInfo query failed.");

      // Some odd query returning list of constants
      books = session.EntitySet<IBook>();
      var numberQuery = from b in books
                        where b.Publisher.Name == "MS Books"
                        select 1;
      var numbersFromQuery = numberQuery.ToList();
      Assert.IsTrue(numbersFromQuery.Count > 0, "Number query did not return rows");

      // Aggregate methods
      session = entityStore.OpenSession();
      books = session.EntitySet<IBook>();
      var maxPrice = books.Max(b => b.Price);
      Assert.IsTrue(maxPrice > 0, "Max price is invalid.");
      // Queryable.Average method is special in some way - it is not a generic on result type, but has a number of overloads
      // for specific numeric types. This test checks how fwk handles this
      var avgPrice = books.Average(b => b.Price);
      Assert.IsTrue(avgPrice > 0, "Average price is invalid.");

      // Test 'locally evaluatable' pieces in queries
      session = entityStore.OpenSession();
      pubs = session.EntitySet<IPublisher>();
      var msBooksName = "local var value";
      _someClassField = "field value";
      var pq = from p in pubs
               where p.Name == "Literal'string" || p.Name == PublisherNameConst //these 2 should turn into consts in query, not parameters
                     || p.Name == msBooksName
                     || p.Name == _someClassField
                     || p.Name != "blah"
                     || p.Name == SomeClassProp
               select p;
      var pqList = pq.ToList();
      Assert.IsTrue(pqList.Count > 0, "Query with local variables failed.");

      //Using extra container object
      var qparams = new QueryParamsContainer() { Field = csTitle, Prop = vbTitle };
      var qBooksWithParamObj = from b in books
                               where b.Title == qparams.Field || b.Title == qparams.Prop
                               select b;
      var lstBooksWithParamObj = qBooksWithParamObj.ToList();
      Assert.AreEqual(2, lstBooksWithParamObj.Count, "Query with param object failed");

      // Test using predicates comparing with null. It is special case - must be translated to 'IS NULL' in SQL
      var bq = from b in books
               where b.Abstract == null
               select b;
      var booksWithoutAbstract = bq.ToList();
      Assert.AreEqual(3, booksWithoutAbstract.Count, "Query with predicate checking for null failed.");
      // testing entity reference null check (foreign key != null)
      var qAuthUsers = from a in authors
                        where a.User != null
                        select a;
      var lstAuthUsers = qAuthUsers.ToList();
      Assert.IsTrue(lstAuthUsers.Count > 0, "Query with check for non null failed.");

      // Test for MySql handling of Guid IDs
      // Must use an entity that is not in full entity cache: like IBookOrder, not IBook.
      session = entityStore.OpenSession();
      var someOrder = session.GetEntities<IBookOrder>(take: 10).First();
      var someOrderId = someOrder.Id;
      var qGetOrderById = from bo in session.EntitySet<IBookOrder>()
                     where bo.Id == someOrderId
                     select bo;
      var listOrders = qGetOrderById.ToList();
      Assert.AreEqual(1, listOrders.Count, "Failed to get order by Id.");
      Assert.AreEqual(someOrderId, listOrders[0].Id, "Failed to get order by Id.");
    }

    [TestMethod, TestCategory("Vita_Books")]
    public void TestLinqWithEntityCache() {
      if (!SetupHelper.CacheEnabled)
        return;
      SetupHelper.ReloadCache(); //make sure cache is reloaded and fresh
      // Some queries (ex: involving many-to-many relations) are not supported by Linq2Sql, 
      // but run OK against full entity cache. 
      // Entity cache is returning clones of entities in cache, and these clones must be attached to the caller session.
      // We test that results of queries are attaced to current session.
      var entityStore = SetupHelper.EntityStore;

      // Many-to-many relations is NOT supported currently, only queries in cache can handle this. It is on TODO list to support m2m. 
      // We check CloneEntities interceptor injected at the top level, for the final results of the query
      var session = entityStore.OpenSession();
      var qMultiAuthorBooks = from b in session.EntitySet<IBook>()
                              where b.Authors.Count > 1
                              orderby b.Title descending
                              select b;
      var multiAuthorBooks = qMultiAuthorBooks.ToList();
      // NOTE: do not run this test standalone with USE_OLD_DATA - test starts with cache invalidated, and query is sent to database where it fails
      Assert.IsTrue(multiAuthorBooks.Count > 0, "No multi-author books found.");
      Assert.IsTrue(IsAttachedTo(multiAuthorBooks[0], session), "Entities are not attached to current session!");

      // Query returning anon object with entity (Publisher) inside
      // CloneEntity call is injected into an argument of anon object constructor
      session = entityStore.OpenSession();
      var bkInfos = from b in session.EntitySet<IBook>()
                    select new { Title = b.Title, Publisher = b.Publisher };
      var lstBkInfos = bkInfos.ToList();
      Assert.IsTrue(lstBkInfos.Count > 0, "Bk info query failed.");
      Assert.IsTrue(IsAttachedTo(lstBkInfos[0].Publisher, session), "BookInfo.Publisher entity is not attached to current session.");

      // Query returning list of lists in many-to-many; not supported in Linq2Sql, only in entity cache
      // CloneEntities call is injected into lambda parameter of 'Select' method.
      session = entityStore.OpenSession();
      var authQuery = from b in session.EntitySet<IBook>()
                      where b.Category == BookCategory.Programming
                      select b.Authors;
      var authListOfLists = authQuery.ToList();
      Assert.IsTrue(authListOfLists.Count > 0, "Failed to retrieve list-of-lists.");
      Assert.IsTrue(IsAttachedTo(authListOfLists[0][0], session), "Entities in list-of-lists queries are not attached to current session!");

      // checking perf, do it a few more times and look at the log; 
      // The first run of LINQ query in cache - the compilation takes most of the time; the query is compiled and saved in cache,
      // so it should be much faster next run(s)
      var temp1 = qMultiAuthorBooks.ToList();
      temp1 = qMultiAuthorBooks.ToList();
      temp1 = qMultiAuthorBooks.ToList();
      var temp2 = bkInfos.ToList();
      temp2 = bkInfos.ToList();
      temp2 = bkInfos.ToList();
      // yes, it is faster - all queries run with 0ms reported.
    }

    private bool IsAttachedTo(object entity, IEntitySession session) {
      var recSession = EntityHelper.GetSession(entity);
      return recSession == session; 
    }

    [TestMethod, TestCategory("Vita_Books")]
    public void TestLinqGroupByAndJoins() {
      var entityStore = SetupHelper.EntityStore;
      var session = entityStore.OpenSession();
      var books = session.EntitySet<IBook>();
      var pubs = session.EntitySet<IPublisher>();
      var authors = session.EntitySet<IAuthor>();
      var users = session.EntitySet<IUser>();

      // Join of sub queries using multiple 'from' + 'where' (without 'join' keyword); 
      // Note that joins using 'join' keyword do not support subqueries
      // also checking use of parameter inside sub-query
      string nonName = "NoName";
      var subUsers = users.Where(u => u.IsActive && u.UserName != "noname");
      var subAuthors = authors.Where(a => a.FirstName != nonName);
      var qSubQueries = from u in subUsers
                        from a in subAuthors
                        where a.User == u 
                        select new { A = a, U = u };
      var listSubQueries = qSubQueries.ToList(); 
      Assert.IsTrue(listSubQueries.Count > 0, "Sub-queries join failed.");
      
      // Group-By
      //Note: not supported in output: ' Books = g.ToList() }' - this is what is known '(n+1)-select' style query
      // usually implemented thru running (n+1) selects (one for main records + n queries for children)
      // No plans to support this, for several fundamental reasons (not because we can't, but because we shouldn't)
      var booksByCat = from b in books
                       group b by b.Category into g
                       orderby g.Key
                       select new { Category = g.Key, BookCount = g.Count(), MaxPrice = g.Max(b => b.Price) };  
      var lstBooksByCat = booksByCat.ToList();
      Assert.IsTrue(lstBooksByCat.Count > 0, "GroupBy test returned 0 groups.");

      // simple join thru property. Note: currently fails if Author.User == null, on TODO list
      var qjp = from a in authors
                select new { User = a.User, Author = a };
      var lstQjp = qjp.ToList();
      Assert.IsTrue(lstQjp.Count > 0, "Failed to retrieve users with authors with users.");

      // Cross join
      // RI: join condition on Entities (ent equals ent) is handled ok, automatically changed to comparing keys, just like in WHERE
      var booksWithPubs = from b in books
                          join p in pubs on b.Publisher equals p
                          select new { Book = b, Publisher = p };
      var lstBooksWithPubs = booksWithPubs.ToList();
      Assert.IsTrue(lstBooksWithPubs.Count > 0, "Books-Join-Publishers test returned 0 results.");

      // Outer join. See here: http://msdn.microsoft.com/en-us/library/vstudio/bb397895.aspx
      var qUsersWithAuthors = from u in users
                              join a in authors on u.Id equals a.User.Id into tempAuthors // equals does not work with ent references for now, have to use PKs
                              from a2 in tempAuthors.DefaultIfEmpty()
                              select new { User = u.UserName, Author = a2.LastName };
                              // select new { User = u, Author = a2 }; //for now does not work with 'nullable' entities
      var lstUsersWithAuthors = qUsersWithAuthors.ToList();
      Assert.IsTrue(lstUsersWithAuthors.Count > 0, "Failed to retrieve users with authors.");

      // Similar join, but inner
      var q1 = from u in users
               from a in authors
               where a.User == u 
               select new { A = a, U = u };
      var q1List = q1.ToList();
      Assert.IsTrue(q1List.Count > 0, "Cross join with where failed.");

    }


    [TestMethod, TestCategory("Vita_Books")]
    public void TestLinqArrayContains() {
      var entityStore = SetupHelper.EntityStore;
      var session = entityStore.OpenSession();
      var bookOrders = session.EntitySet<IBookOrder>();
      //Note: it is important to use table that is not fully cached, so we use IBookOrder entity

      // Test constList.Contains()
      var userTypes = new UserType[] { UserType.WebVisitor, UserType.Author};
      var qOrders = from bo in bookOrders
              where userTypes.Contains(bo.User.Type)
              select bo;
      var orders = qOrders.ToList();
      Assert.IsTrue(orders.Count > 0, "No orders by type found.");

      // Orders for books from MS pub
      var someOrders = bookOrders.Take(2).ToList();
      var orderIdsArray = someOrders.Select(o => o.Id).ToArray(); 
      var qOrders2 = from bo in bookOrders
                        where orderIdsArray.Contains(bo.Id)
                        select bo;
      orders = qOrders.ToList();
      Assert.AreEqual(orderIdsArray.Length, orders.Count, "Test Array.Contains failed: order counts do not match.");
      
      // Try again with a single Id
      var arrOneId = new Guid[] { orderIdsArray[0] };
      qOrders = from bo in bookOrders
                    where arrOneId.Contains(bo.Id)
                    select bo;
      orders = qOrders.ToList();
      Assert.AreEqual(1, orders.Count, "Test Array.Contains with one Id failed: order counts do not match.");
      // Again with empty list
      var arrEmpty = new Guid[] {};
      qOrders = from bo in bookOrders
                where arrEmpty.Contains(bo.Id)
                select bo;
      orders = qOrders.ToList();
      Assert.AreEqual(0, orders.Count, "Test Array.Contains with empty array failed: order counts do not match.");
      // Again with list, not array
      var orderIdsList = orderIdsArray.ToList();
      qOrders = from bo in bookOrders
                where orderIdsList.Contains(bo.Id)
                select bo;
      orders = qOrders.ToList();
      Assert.AreEqual(orderIdsList.Count, orders.Count, "Test constList.Contains, repeated query failed: order counts do not match.");
    }

    [TestMethod, TestCategory("Vita_Books")]
    public void TestLinqCompiledQueryCache() {
      var entityStore = SetupHelper.EntityStore;
      var session = entityStore.OpenSession();
      var books = session.EntitySet<IBook>();
      var authors = session.EntitySet<IAuthor>();

      //Testing how expr like 'where someEnt.OtherEnt == null' are handled
      var qa = from a in authors
               where a.User == null
               select a;
      var aList = qa.ToList(); // just checking SQL is correct and it runs


      //test compiled query caching
      var kidBooksQ = from b in books
                      where b.Publisher.Name == "Kids Books"
                      orderby b.Title
                      select b;
      var entQuery = kidBooksQ as EntityQuery;
      var kidBooksList = kidBooksQ.ToList();
      //Just for debugging, do it 3 times
      kidBooksList = kidBooksQ.ToList();
      kidBooksList = kidBooksQ.ToList();
      Assert.IsTrue(kidBooksList.Count > 0, "Compiled query cache test failed: kid books not found.");
      // var l2 = kidBooksQ.ToList(); // for checking in debug mode - second time the query must be cached (compiled query cache)

      // Query with disabled compiled query cache. Disable cache for 'search' queries, to avoid polluting cache
      // with many custom searches that users enter in search form. 
      // Note about the test query - we must make sure it is of unique shape, otherwise system will pickup previously compiled query.
      var noCacheBooksQ = session.EntitySet<IBook>().DisableQueryCache();
      var kidBooksQDesc = from b in noCacheBooksQ
                          where b.Publisher.Name == "Kids Books"
                          orderby b.Title descending
                          select b;
      var kidBooksListDesc = kidBooksQ.ToList();
      Assert.IsTrue(kidBooksListDesc.Count > 0, "Query with disabled query cache failed: multiple-author books not found.");
    }

    [TestMethod, TestCategory("Vita_Books")]
    public void TestLinqBoolBitFields() {
      // We test that LINQ engine correctly handles bit fields
      var entityStore = SetupHelper.EntityStore;
      var session = entityStore.OpenSession();
      var users = session.EntitySet<IUser>();

      // Expression over bit field: 'u.IsActive==true' should be replaced with 'u.IsActive = 1'; we also check ! operator and bool parameter
      bool boolParam = true; 
      var q2 = from u in users
               where u.IsActive && u.IsActive == true && u.IsActive == boolParam || !u.IsActive
               select u; // new { U = u.UserName, A = u.IsActive };
      var lstUsers2 = q2.ToList();
      Assert.IsTrue(lstUsers2.Count > 0, "Bit field expr test failed");

      // bool/bit field used in Where expressions directly - should be replaced with 'u.IsActive = 1'
      var q1 = from u in users
               where u.IsActive // & u.IsActive == true
               select u; // new { U = u.UserName, A = u.IsActive };
      var lstUsers1 = q1.ToList();
      Assert.IsTrue(lstUsers1.Count > 0, "No active users found.");

      // Using bit field in anon type initializer
      var q3 = from u in users
               //where u.IsActive
               select new { U = u.UserName, A = u.IsActive };
      var lstUsers3 = q3.ToList();
      Assert.IsTrue(lstUsers3.Count > 0, "Bit field - use in anon object failed.");

    }

    [TestMethod, TestCategory("Vita_Books")]
    public void TestLinqSpecialQueries() {
      //Some seemingly simple queries that caused problems for LINQ engine initially
      var entityStore = SetupHelper.EntityStore;
      var session = entityStore.OpenSession();
      var books = session.EntitySet<IBook>();
      var pubs = session.EntitySet<IPublisher>();

      // Join/Distinct/Count combinations. Encountered error: The column 'Id' was specified multiple times for 'X'. 
#if SQLCE
      //SQLCE has some specifics - it does not allow Text(memo) columns in output of DISTINCT clause; 
      // so doing easier query without book.Abstract column
      var qJoin = from b in books
                  join p in pubs on b.Publisher equals p
                  where p.Name == "MS Books"
                  select new { b.Id, b.Title, PubId = p.Id, p.Name }; //new { B = b, P = p }; //
#else
      var qJoin = from b in books
                  join p in pubs on b.Publisher equals p
                  where p.Name == "MS Books"
                  select new { B = b, P = p }; //
#endif

      var joinDistinct = qJoin.Distinct().ToList();
      Assert.IsTrue(joinDistinct.Count > 0, "JoinDistinct query failed.");
      var joinCount = qJoin.Count();
      Assert.IsTrue(joinCount > 0, "JoinCount query failed.");
      var joinDistinctCount = qJoin.Distinct().Count();
      Assert.IsTrue(joinDistinctCount > 0, "JoinDistinctCount query failed.");
      var joinDistinctTake = qJoin.Distinct().Take(3).ToList();
      Assert.IsTrue(joinDistinctTake.Count > 0, "JoinDistinctTake query failed.");
      var joinDistinctSkipTake = qJoin.Distinct().Skip(1).Take(3).ToList();
      Assert.IsTrue(joinDistinctSkipTake.Count > 0, "JoinDistinctSkipTake query failed.");

      // New operator hidden inside Count(). Added special handling of New in SqlBuilder
      var qJoin2 = from b in books
                   select new { Title = b.Title };
      var list = qJoin2.ToList();
      var joinDistinctCount2 = qJoin2.Count();
      Assert.IsTrue(joinDistinctCount2 > 0, "Count query failed.");

      //Take/skip
      var qbase = books.Where(b => b.Publisher.Name == "MS Books");
      // with order by
      var qTakeSkip = qbase.OrderBy(b => b.Title).Skip(1).Take(2);
      var lTakeSkip = qTakeSkip.ToList();
      Assert.AreEqual(2, lTakeSkip.Count, "TakeSkip query failed.");
      // without Order by - should automatically add 'order by Id'
      qTakeSkip = qbase.Skip(1).Take(2);
      lTakeSkip = qTakeSkip.ToList();
      Assert.AreEqual(2, lTakeSkip.Count, "TakeSkip query failed.");

      //using min/max dates - parameter should use DbType.DateTime2, not DateTime to avoid overflow
      var qByDates = books.Where(b => b.PublishedOn > DateTime.MinValue && b.PublishedOn < DateTime.MaxValue);
      var lByDates = qByDates.ToList();
      Assert.IsTrue(lByDates.Count > 0, "MinDate query failed");

      // SQLCE and maybe others: when using OFFSET and LIMIT clauses, ORDER BY must be specified
      // If not, LINQ engine must add default (by primary key)
      var qSkipTake = books.Skip(1).Take(1);
      var lSkipTake = qSkipTake.ToList();
      Assert.AreEqual(1, lSkipTake.Count, "Skip/take without order by query failed");

      // Simplest query - not so simple, had to fix some code to make it work
      var allBooks = books.ToList();
      Assert.IsTrue(allBooks.Count > 0, "books.ToList() failed.");

    }

    [TestMethod, TestCategory("Vita_Books")]
    public void TestLinqUnionExcept() {
      //Some seemingly simple queries that caused problems for LINQ engine initially
      var entityStore = SetupHelper.EntityStore;
      var session = entityStore.OpenSession();
      var books = session.EntitySet<IBook>();
      var pubs = session.EntitySet<IPublisher>();

      //simple union
#if !SQLCE
      // SQL CE fails on this query; it requires UNION ALL if you include Text/Memo columns
      var qUnion = books.Where(b => b.Title == "c# Programming").Union(books.Where(b => b.Title == "VB Programming"));
      var lstUnion = qUnion.ToList();
      Assert.IsTrue(lstUnion.Count == 2, "Union query failed.");
#endif
      // Union of anon types
      var q1 = books.Where(b => b.Title == "c# Programming").Select(b => new { Id = b.Id, Title = b.Title });
      var q2 = books.Where(b => b.Title == "VB Programming").Select(b => new { Id = b.Id, Title = b.Title });
      var q12 = q1.Union(q2);
      var list12 = q12.ToList();
      Assert.IsTrue(list12.Count == 2, "Union query2 failed.");

      // Except
      // SQL CE, MySql do not support Except
#if !SQLCE && !MYSQL
      if (!SetupHelper.CacheEnabled) {
        // for cached entities, the comparisons in filters are performed for CLR AUTO objects (.NET-produced comparison), and they do not work
        // 'correctly'. Smth we have to live with - no way of fixing auto objects comparison
        var qe1 = books.Where(b => b.Title.Contains("Programming")).Select(b => new { Id = b.Id, Title = b.Title });
        var qe2 = books.Where(b => b.Title.Contains("c#")).Select(b => new { Id = b.Id, Title = b.Title });
        var qe12 = qe1.Except(qe2);
        var listEx12 = qe12.ToList(); // we should have VB progr, Windows progr, but not c# progr
        Assert.IsTrue(listEx12.Count == 2, "Except query2 failed.");
      }
#endif
    }

    #region LinqTestBed
    // Not real tests - just random very specific expressions for debugging latest found bugs - should all work OK now
    [TestMethod, TestCategory("Vita_Books")]
    public void Test_LinqTestBed() {
      var entityStore = SetupHelper.EntityStore;
      //Init
      var session = entityStore.OpenSession();
      var authors = session.EntitySet<IAuthor>();

      var qa = from a in authors
               select new { Author = a, User = a.User  }; 
      var alist = qa.ToList();
      Assert.IsTrue(alist.Count > 0, "Author/User query failed.");
      // User might be null, but Author is always there
      var auth0 = alist[0].Author;
      var authRec0 = EntityHelper.GetRecord(auth0);
      Assert.AreEqual(session, authRec0.Session, "Sessions do not match");

    }
    #endregion


  }//class
}
