﻿using System;
using System.ComponentModel;
using System.Collections.Generic;
using System.Linq;
using System.Diagnostics;
using System.Data.SqlClient;
using System.Threading;

using Vita.Entities;
using Vita.Entities.Runtime;
using Vita.Modules.ErrorLogging;
using Vita.Data.Caching;

using Vita.Samples.BooksModel;

namespace Vita.Samples.BooksSample.UnitTests {

#if USE_NUNIT
  using NUnit.Framework;
  using TestClass = NUnit.Framework.TestFixtureAttribute;
  using TestMethod = NUnit.Framework.TestAttribute;
  using TestInitialize = NUnit.Framework.SetUpAttribute;
  using TestCategory = NUnit.Framework.CategoryAttribute;
#else
  using Microsoft.VisualStudio.TestTools.UnitTesting;
#endif

  [TestClass]
  public class MiscTests {

    [TestInitialize]
    public void TestInit() {
      SetupHelper.Init();
      SetupHelper.CreateSampleData();
    }
    [TestCleanup]
    public void TearDown() {
      SetupHelper.TearDown(); 
    }

    [TestMethod, TestCategory("Vita_Books")]
    public void TestCRUD() {
      var entityStore = SetupHelper.EntityStore;
      var userCtx = new UserContext("Test", null);
      //Load all books, get IDs
      var session = entityStore.OpenSession(userCtx);
      var allBooks = session.GetEntities<IBook>(take: 10); //we have to provide paging (IBook is not marked by SmallTable attribute)
      Assert.IsTrue(4 == allBooks.Count, "Invalid # of books");
      // alt way:
      var bkCount = session.EntitySet<IBook>().Count();
      Assert.IsTrue(4 == bkCount, "Invalid # of books from EntitSet().Count().");
      // Save some IDs in local variables 
      var csBook = allBooks.Single(b => b.Title.StartsWith("c#"));
      var vbBook = allBooks.Single(b => b.Title.StartsWith("VB"));
      var windowsBook = allBooks.Single(b => b.Title.StartsWith("Windows"));
      var kidBook = allBooks.Single(b => b.Title.StartsWith("Three"));
      var csBookId = csBook.Id;
      var vbBookId = vbBook.Id;
      var windowsBookId = windowsBook.Id;
      var kidBookId = kidBook.Id;
      var msPubId = csBook.Publisher.Id;
      var kidPubId = kidBook.Publisher.Id;

      //Loading entities by primary key --------------------------------------------------------
      // Verify that parent objects referenced by the same FK value result in the same object
      session = entityStore.OpenSession(userCtx);
      // load book by ID, check the title
      csBook = session.GetEntity<IBook>(csBookId);
      Assert.IsTrue(csBook != null, "Failed to find the book by Id.");
      Assert.IsTrue(csBook.Title == "c# Programming", "Loading book by Id: wrong book loaded, title mismatch.");
      vbBook = session.GetEntity<IBook>(vbBookId);
      Assert.IsTrue(vbBook != null, "Failed to find the book by Id.");
      // Done - loaded and verified 2 programming books.
      //Compare publishers as objects
      object csPubObj = csBook.Publisher;
      object vbPubObj = vbBook.Publisher;
      Assert.IsTrue(csPubObj != null, "Publisher is null!");
      Assert.IsTrue(csPubObj == vbPubObj, "c# and vb books publishers is not the same Publisher instance.");

      // Default sorting. Books are sorted by title by default, so proper order is: Cs book; Three little pigs; Vb book; Windows book
      session = entityStore.OpenSession(userCtx);
      allBooks = session.GetEntities<IBook>(take: 10);
      Assert.IsTrue(allBooks.Count >= 4, "Invalid book count");
      //let's re-sort and compare order
      var sortedBooks = allBooks.OrderBy(b => b.Title).ToList();
      for (int i = 0; i < allBooks.Count; i++)
        Assert.AreEqual(allBooks[i], sortedBooks[i], "Books are not sorted by title");

      //Paging thru books - load with skip=1, take 1
      session = entityStore.OpenSession(userCtx);
      var pagedBooks = session.GetEntities<IBook>(skip: 1, take: 1);
      Assert.IsTrue(1 == pagedBooks.Count, "Invalid # of paged list of books");
      var bk0 = pagedBooks[0];
      Assert.IsTrue(bk0.Title == "Three little pigs", "Invalid book title.");
      Assert.IsTrue(bk0.Publisher.Name == "Kids Books", "Invalid publisher.");

      //Updating, deleting entities ---------------------------------------------------------------
      //Create 2 authors
      session = entityStore.OpenSession(userCtx);
      var j1 = session.NewAuthor("John", "One");
      var j2 = session.NewAuthor("Jack", "Two");
      var j1Id = j1.Id;
      var j2Id = j2.Id;
      session.SaveChanges();
      session = entityStore.OpenSession(userCtx);
      j1 = session.GetEntity<IAuthor>(j1Id);
      j2 = session.GetEntity<IAuthor>(j2Id);
      Assert.IsNotNull(j1, "j1 is not saved/reloaded.");
      Assert.IsNotNull(j2, "j2 is not saved/reloaded.");
      //Modify
      j1.LastName += " Jr";
      j2.FirstName = "Jacky";
      session.SaveChanges();
      //reload and check that names changed
      session = entityStore.OpenSession(userCtx);
      j1 = session.GetEntity<IAuthor>(j1Id);
      j2 = session.GetEntity<IAuthor>(j2Id);
      Assert.AreEqual("John One Jr", j1.FullName, "Invalid modified name of j1");
      Assert.AreEqual("Jacky Two", j2.FullName, "Invalid modified name of j2");
      //Delete
      session = entityStore.OpenSession(userCtx);
      j1 = session.GetEntity<IAuthor>(j1Id);
      session.DeleteEntity(j1);           // delete  entity
      //other way of deleting, by creating stub - to avoid reloading
      var j2Stub = session.GetEntity<IAuthor>(j2Id, LoadFlags.Stub);  
      session.DeleteEntity<IAuthor>(j2Stub);
      session.SaveChanges(); 
      //Check that entities are deleted
      session = entityStore.OpenSession(userCtx);
      j1 = session.GetEntity<IAuthor>(j1Id);
      j2 = session.GetEntity<IAuthor>(j2Id);
      Assert.IsNull(j1, "j1 is not deleted.");
      Assert.IsNull(j2, "j2 is not deleted.");

      // Entity lists, one-to-many -----------------------------------------------------------------------
      // For a publisher, get the books (Publisher.Books)
      session = entityStore.OpenSession(userCtx);
      var msPub = session.GetEntity<IPublisher>(msPubId);
      Assert.IsTrue(3 == msPub.Books.Count, "Invalid # of books from MS Books");
      foreach (var bk in msPub.Books)
        Assert.IsTrue(!string.IsNullOrWhiteSpace(bk.Title), "Book title is empty.");  

      //Entity lists, many-to-many -----------------------------------------------------------------------
      // Loading a book (about c#) and enumerating its Authors property.
      session = entityStore.OpenSession(userCtx);
      csBook = session.GetEntity<IBook>(csBookId);
      Assert.IsTrue(2 == csBook.Authors.Count, "Invalid authors count for c# book");
      foreach (var a in csBook.Authors)
        Assert.IsTrue(!string.IsNullOrWhiteSpace(a.FullName), "Author full name is empty."); 
      //remove one author, save and check if it is removed
      var a1 = csBook.Authors[0];
      csBook.Authors.Remove(a1);
      session.SaveChanges();
      session = entityStore.OpenSession(userCtx);
      csBook = session.GetEntity<IBook>(csBookId);
      var csAuthors = csBook.Authors;
      Assert.IsTrue(1 == csAuthors.Count, "Author is not removed for c# book");
      //Add author back, we gonna need him
      a1 = session.GetEntity<IAuthor>(a1.Id);
      csBook.Authors.Add(a1);
      session.SaveChanges(); 

      // Special tests ========================================================================
      // Testing remove/add with many-to-many lists
      session = entityStore.OpenSession(userCtx);
      csBook = session.GetEntity<IBook>(csBookId);
      var jack = (from a in csBook.Authors
                 where a.FirstName == "Jack"
                 select a).FirstOrDefault();
      csBook.Authors.Remove(jack);
      csBook.Authors.Add(jack);
      session.SaveChanges(); //we just verify that it works


      //test EnsureLoaded - just check we can call it and it does not blow up
      session = entityStore.OpenSession(userCtx); 
      csBook = session.GetEntity<IBook>(csBookId); 
      EntityHelper.EnsureLoaded(csBook, b => b.Authors);
      EntityHelper.EnsureLoaded(csBook, b => new {b.Authors, b.Publisher});

    }//method

    [TestMethod, TestCategory("Vita_Books")]
    public void TestWithComputedColumn() {
      var entityStore = SetupHelper.EntityStore;
      var session = entityStore.OpenSession();
      
      // Testing loading entity with computed column; I've encountered trouble with computed columns in new linq engine, so verifying the fix here.
      // Important - we should use entity that is NOT in full cache, like IBookOrder (it has computed field Summary).
      var orders = session.EntitySet<IBookOrder>();
      var qOrders = from ord in orders
                    where ord.Status == OrderStatus.Completed
                    select ord;
      var lstOrders = qOrders.ToList();
      Assert.IsTrue(lstOrders.Count > 0, "Failed to retrieve orders.");
      var summ0 = lstOrders[0].Summary;
      Assert.IsFalse(string.IsNullOrWhiteSpace(summ0), "Failed to read order summary");
    }

    [TestMethod, TestCategory("Vita_Books")]
    public void TestUniqueKey() {
      var entityStore = SetupHelper.EntityStore;
      var session = entityStore.OpenSession();
      // Unique key test ----------------------------------------------------------------------------------------
      var msPub2 = session.NewPublisher("MS Books");
      try {
        session.LogMessage("\r\n-- TEST: expecting duplicate key exception ....");
        session.SaveChanges();
        Assert.IsTrue(false, "Expected UniqueIndexViolationException was not thrown.");
      } catch (UniqueIndexViolationException ex) {
        Assert.IsTrue(ex.KeyName == "PublisherName", "Unexpected key name in Unique index violation.");
        Assert.IsTrue(ex.MemberNames == "Name", "Unexpected member name in Unique index violation.");
        SetupHelper.ErrorLog.LogError(ex, session.GetUserLogContents());
      } catch(Exception) {
        throw; //just a stop for a breakpoint
      }
    }


    [TestMethod, TestCategory("Vita_Books")]
    public void TestPersistOrderInAttr() {
      var entityStore = SetupHelper.EntityStore;
      var session = entityStore.OpenSession();
      var doraOrder = session.EntitySet<IBookOrder>().First(o => o.User.UserName == "Dora");
      var lines = doraOrder.Lines;
      var item0 = lines[0];
      var item1 = lines[1];
      //Change line ordering; 
      lines[0] = item1;
      lines[1] = item0;
      session.SaveChanges();
      // IBookOrderLine has [PersistOrderIn("LineNumber")] attribute, so changed order in the list
      // should be reflected in LineNumber property (after save); and if we read lines back, they should come in new order
      Assert.AreEqual(lines[0].LineNumber, 1, "Ordering not applied.");
      Assert.AreEqual(lines[1].LineNumber, 2, "Ordering not applied.");
      //Open fresh session, read lines back and check order
      session = entityStore.OpenSession();
      doraOrder = session.GetEntity<IBookOrder>(doraOrder.Id); //get fresh copy
      Assert.AreEqual(doraOrder.Lines[0].Id, item1.Id, "Explicit ordering failed!");
      Assert.AreEqual(doraOrder.Lines[1].Id, item0.Id, "Explicit ordering failed!");
    }

    [TestMethod, TestCategory("Vita_Books")]
    public void TestValidation() {
      var entityStore = SetupHelper.EntityStore;
      var session = entityStore.OpenSession();
      // Entity validation: trying to save entities with errors -----------------------------------------
      var invalidAuthor = session.NewAuthor(null, "VeryLoooooooooooooooooooooooooooongLastName");
      var invalidBook = session.NewBook(BookEdition.EBook, BookCategory.Fiction, "Not valid book", "Some invalid book", null, DateTime.Now, -5.0m);
      //We expect 4 errors: Author's first name should not be null; Author's last name is too long; Publisher cannot be null;
      // Price must be > 1 cent. The first 3 errors are found by built-in validation; the last error, price check, is added
      // by custom validation method.
      try {
        session.SaveChanges();
        Assert.IsTrue(false, "Validation exception is not thrown.");
      } catch (ValidationException vex) {
        Assert.AreEqual(4, vex.Errors.Count, "Wrong # of validation errors.");
        foreach (var err in vex.Errors)
          Assert.IsTrue(!string.IsNullOrWhiteSpace(err.Message), "Validation message is empty");
      }
    }

    [TestMethod, TestCategory("Vita_Books")]
    public void TestCanDelete() {
      var entityStore = SetupHelper.EntityStore;
      var session = entityStore.OpenSession();
      // Test CanDelete. There are no orders for Windows book, so we may be able to delete it
      var winBook = session.EntitySet<IBook>().Single(b => b.Title.StartsWith("Windows"));
      var winBookId = winBook.Id;
      Type[] blockingEntities;
      // We can delete a book even if there are links for it in BookAuthor link table (BookAuthor.Book has CascadeDelete attr)
      // so the result here should be true
      var canDelete = session.CanDeleteEntity(winBook, out blockingEntities);
      Assert.IsTrue(canDelete, "CanDelete WinBook failed.");
      session.DeleteEntity(winBook);
      session.SaveChanges();
      //check it is deleted; we also check here that entity cache works correctly and does not keep object.
      winBook = session.GetEntity<IBook>(winBookId);
      Assert.IsNull(winBook, "Winbook not deleted.");

      //Now check Publisher: mspub has associated books, so it should not be allowed to be deleted.
      var msPub = session.EntitySet<IPublisher>().Single(p => p.Name.StartsWith("MS"));
      canDelete = session.CanDeleteEntity(msPub, out blockingEntities);
      Assert.IsFalse(canDelete, "Error: MS publisher should not be allowed to be deleted.");
      Assert.IsTrue(blockingEntities != null && blockingEntities.Length > 0, "Blocking entities empty.");
      //Still try to delete
      session.LogMessage("\r\n-- TEST: expecting integrity violation exception ...."); //to make clear in Log that exc is expected
      Debug.WriteLine("\r\n-- TEST: expecting integrity violation exception ....");
      try {
        session.DeleteEntity(msPub);
        session.SaveChanges();
      } catch (Exception ex) {
        SetupHelper.ErrorLog.LogError(ex, session.GetUserLogContents());
        Debug.WriteLine(ex.ToString());
      }
    }



    [TestMethod, TestCategory("Vita_Books")]
    public void TestNotifyPropertyChanged() {
      var entityStore = SetupHelper.EntityStore;
      var session = entityStore.OpenSession();
      // INotifyPropertyChanged and DependsOn test ----------------------------------------------------------------
      // FullName has DependsOn attribute with FirstName, LastName as targets. 
      // So when we change the FirstName, we get two events fired -
      // first for FirstName, then for FullName
      var auth1 = session.GetEntities<IAuthor>(take: 1)[0];
      var iNpc = auth1 as INotifyPropertyChanged;
      var raisedEvents = new Stack<string>();
      iNpc.PropertyChanged += (ent, args) => { raisedEvents.Push(args.PropertyName); };
      auth1.FirstName = "NewName";
      //We should have 2 properties in the stack
      Assert.AreEqual(2, raisedEvents.Count, "Invalid number of properties in events stack.");
      var prop1 = raisedEvents.Pop();
      Assert.AreEqual("FullName", prop1, "Expected property changed on FullName");
      var prop2 = raisedEvents.Pop();
      Assert.AreEqual("FirstName", prop2, "Expected property changed on FirstName");
    }


    [TestMethod, TestCategory("Vita_Books")]
    public void TestSparceEntityCache() {
      if (!SetupHelper.CacheEnabled)
        return;
      var ctx = new UserContext("UnitTest", null); 
      var entityStore = SetupHelper.EntityStore;
      // Books catalog (books, authors, publishers) is kept in FULL entity cache - entire tables are held in memory
      // Book orders, order lines are cached in sparse cache - only single records are kept for limited time (30 seconds).
      
      // 1. EntityCache 
      var session = entityStore.OpenSession(ctx);
      var bk = session.EntitySet<IBook>().First(); //this will start reloading cache
      Thread.Sleep(500); // let cache reload
      session = entityStore.OpenSession(ctx); //open another session
      bk = session.EntitySet<IBook>().First(); //now it should come from cache
      var bkRec = EntityHelper.GetRecord(bk);
      Assert.IsNotNull(bkRec.CachedRecord, "Entity cache test failed: book is not cached.");

      // Sparse cache
      // Remember that records in sparse cache expire in 30 seconds, 
      // so if you step through in debugger and stop for too long, the test might fail.
      session = entityStore.OpenSession(ctx);
      var orderId = session.EntitySet<IBookOrder>().First().Id;
      // Open session and get order by Id - this will place order entity into sparse cache
      session = entityStore.OpenSession(ctx);
      var order1 = session.GetEntity<IBookOrder>(orderId);

      // Open new session and read order again - it should come from sparse cache
      session = entityStore.OpenSession(ctx);
      var order2 = session.GetEntity<IBookOrder>(orderId);
      var orderRec = EntityHelper.GetRecord(order2);
      Assert.IsNotNull(orderRec.CachedRecord, "Sparse cache test failed: order record is not cached in sparse cache.");

      // Testing deleting from sparse entity cache
      session = entityStore.OpenSession(ctx);
      var billy = session.NewUser("billy", UserType.WebVisitor);
      var billyId = billy.Id; 
      session.SaveChanges();

      // billy should be cached in sparse cache, get it and check it's coming from cache
      // note that if we update cache asynchronously, then immediate GET of the object would not bring cached record - 
      // the cache is updated on background thread
      session = entityStore.OpenSession(ctx);
      bool cacheUpdatedImmediately = !SetupHelper.CacheSettings.Options.IsSet(CacheOptions.AsyncUpdate);
      billy = session.GetEntity<IUser>(billyId);
      // If we setup cache to update synchronously, then immediate read should deliver record from cache
      if (cacheUpdatedImmediately) {
        var billyRec = EntityHelper.GetRecord(billy);
        Assert.IsNotNull(billyRec.CachedRecord, "Billy rec is not delivered from cache.");
      }
      //Now try to delete billy - make sure it does not stay in cache after being deleted in the database
      session.DeleteEntity(billy);
      session.SaveChanges(); 

      // read it back, make sure it is expelled from sparse cache
      session = entityStore.OpenSession(ctx);
      billy = session.GetEntity<IUser>(billyId);
      Assert.IsNull(billy, "Billy is not deleted.");
    }

  }//class
}
