﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;
using System.Configuration;
using System.Threading;
using System.Reflection;
using System.IO;

using Vita.Common;
using Vita.Entities;
using Vita.Entities.Authorization;
using Vita.Entities.Model;
using Vita.Services;
using Vita.Activation;
using Vita.Data;
using Vita.Data.Caching;
using Vita.Data.Driver;
using Vita.Data.Management;
using Vita.Data.Model;
using Vita.Modules.ErrorLogging;
using Vita.Samples.BooksModel;

namespace Vita.Samples.BooksSample.UnitTests {

  public static class SetupHelper {
    public static bool _initialized;
    static DbDriver _driver;
    private static DbOptions _dbOptions = DbOptions.Default;
    private static string _connectionStringKey;

    public static BooksEntityModel BooksModel;
    public static IEntityStore EntityStore;
    public static bool CacheEnabled;
    public static CacheSettings CacheSettings;
    public static IErrorLogService ErrorLog; 


    //static constructor
    static SetupHelper() {
#if SQLCE
      _dbOptions = Vita.Data.SqlCe.SqlCeDbDriver.DefaultDbOptions;
      _driver = new Vita.Data.SqlCe.SqlCeDbDriver();
     _connectionStringKey = "SqlCeConnectionString";
#elif MYSQL
      // if compiler fails here, add reference to Vita.Data.MySql assembly
      _driver = new Vita.Data.MySql.MySqlDriver();
      _connectionStringKey = "MySqlConnectionString";
      _dbOptions |= DbOptions.IgnoreTableNamesCase;
#elif MSSQL08
      _connectionStringKey = "MsSql2008ConnectionString";
      _driver = new Vita.Data.MsSql.MsSqlDriver(Data.MsSql.MsSqlVersion.V2008);
      _dbOptions |= DbOptions.AutoIndexForeignKeys;
#elif MSSQL12
      _connectionStringKey = "MsSql2012ConnectionString";
      _driver = new Vita.Data.MsSql.MsSqlDriver(Data.MsSql.MsSqlVersion.V2012);
      _dbOptions |= DbOptions.AutoIndexForeignKeys;
#else
      ERROR_DEFINE_MISSING();
      //  Error: No conditional symbol defined for driver type. Define one of the following symbols in project properties:
      //     MSSQL08, MSSQL12, SQLCE, MYSQL
#endif
    }

    public static void Init(bool dropSchema = true) {
      if (_initialized)
        return;
      try {
        _initialized = true;
        if (dropSchema)
          DropSchemaObjects("books");
        //Setup model, initialize Books module, create database model, update schema -------------------------------------------------
        BooksModel = new BooksEntityModel();
        CacheEnabled = ConfigurationManager.AppSettings["enableCache"] == "true";
        CacheSettings = null; 
        if (CacheEnabled) {
          CacheSettings = new CacheSettings(
              new Type[] { typeof(IBook), typeof(IPublisher), typeof(IAuthor), typeof(IBookAuthor) }, //fully cached types
              new Type[] { typeof(IBookOrder), typeof(IBookOrderLine), typeof(IUser), },             // sparsely cached types
              fullSetCacheExpiration: TimeSpan.FromSeconds(300), sparseCacheExpiration: TimeSpan.FromSeconds(60),
            // CacheOptions.Default: cache updates synchronously, so that updating session waits for cache write lock and completing cache updates
            // CacheOptions.AsyncUpdate: cache updates asynchronously, so that updating thread does not wait, but returns right after completing db operation;
            //     in this case, the session.UserContext.RequireCacheVersion is set to higher value than current version, so if session immediately reads some cached value,
            //     the read should be correct: if cache has not been updated yet, system detects the version mismatch and directs the query to the database.
              options: CacheOptions.AsyncUpdate | CacheOptions.UseSimpleInvalidator // CacheOptions.None  
              );
        }
        var connString = GetConnectionString(); 
        var logPath = GetLogFilePath();
        var updateOptions = DbSchemaUpdateOptions.VerifyAll | DbSchemaUpdateOptions.UpdateSchema;
        var schemas = new string[] { "books", "log" };
        var dbSettings = new DbSettings(_driver, _dbOptions, connString, updateOptions, schemas: schemas);
        EntityStore = BooksModel.CreateEntityStore(dbSettings, CacheSettings, logFilePath: logPath);
        //retrieve error log
        ErrorLog = EntityStore.GetService<IErrorLogService>(); 
      } catch (ValidationException vex) {
        Debug.WriteLine("Validation errors: \r\n" + vex.ToString());
        throw; 
      } catch (ActivationException sx) {
        //Unit test framework shows only ex message, not details; let's write specifics into debug output - it will be shown in test failure report
        Debug.WriteLine("EntityStore init encountered errors: ");
        Debug.WriteLine(sx.Log);
        throw;
      }
    }

    public static void TearDown() {
      Thread.Sleep(200); //delay to let log flush to disk
    }

    public static DbDriver GetDriver() {
      return _driver; 
    }

    public static void DropSchemaObjects(string schema) {
      var schemaMgr = _driver.CreateSchemaManager(GetConnectionString());
      schemaMgr.DropSchemaObjects(schema);
    }

    public static void ReloadCache() {
      if (!CacheEnabled)
        return;
      var cacheControl = EntityStore.GetService<ICacheControlService>();
      cacheControl.Invalidate(reload: true, waitForReloadComplete: true);
    }

    public static void CreateSampleData() {
      var cacheService = EntityStore.GetService<ICacheControlService>();
      if (cacheService != null)
        cacheService.CacheEnabled = false;
      var db = Database.GetDatabase(EntityStore);
      var entStore = (Vita.Entities.Runtime.EntityStore)EntityStore;
      var entitiesToClear = entStore.Model.GetAllEntityTypes().ToList();
      // We create sample data multiple times, so later test wipes out data from previous test. 
      // So do not wipe out ErrorInfo, just for demo purposes, so the info stays after all tests are run
      entitiesToClear.Remove(typeof(IErrorInfo));
      ActivationHelper.DeleteAllData(db, entitiesToClear);
      SampleDataGenerator.CreateSampleData(EntityStore);
      if (cacheService != null)
        cacheService.CacheEnabled = true;
      if (CacheEnabled)
        ReloadCache();
    }

    public static string GetConnectionString() {
      var connString = ConfigurationManager.AppSettings[_connectionStringKey];
      if (connString.Contains("{bin}")) { //For SQL CE, fix the path
        var asmPath = Assembly.GetExecutingAssembly().Location;
        var binFolder = Path.GetDirectoryName(asmPath);
        connString = connString.Replace("{bin}", binFolder);
      }
      return connString;
    }
    public static string GetLogFilePath() {
      return ConfigurationManager.AppSettings["LogFilePath"];
    }

  }//class
}
