﻿using System;
using System.ComponentModel;
using System.Linq;
using System.Data;
using System.Diagnostics;
using System.Globalization;
using System.Threading;

using Vita.Entities;
using Vita.Entities.Runtime;
using Vita.Data;

namespace Vita.UnitTests.DataTypes {
  using Binary = System.Data.Linq.Binary;

#if USE_NUNIT
  using NUnit.Framework;
  using TestClass = NUnit.Framework.TestFixtureAttribute;
  using TestMethod = NUnit.Framework.TestAttribute;
  using TestInitialize = NUnit.Framework.SetUpAttribute;
  using TestCategory = NUnit.Framework.CategoryAttribute;
#else
  using Microsoft.VisualStudio.TestTools.UnitTesting;
  using Vita.Data.Model;
#endif

  [TestClass] 
  public class DataTypesTests {

      [Entity]
      public interface IDataTypesEntity {

        // Guid
        [PrimaryKey]
        Guid Id { get; set; }
        
        // strings
        [Size(20)]
        string StringProp { get; set; }
        [Memo, Nullable]
        string MemoProp { get; set; }

        //bool 
        bool BoolProp { get; set; }

        // integer types
        byte ByteProp { get; set; }
        Int16 Int16Prop { get; set; }
        Int32 Int32Prop { get; set; }
        Int64 Int64Prop { get; set; }

        // floating point types
        double DoubleProp { get; set; }
        double? DoublePropNull { get; set; }
        float SingleProp { get; set; } //equiv System.Single
        // decimal/ money
        decimal DecProp { get; set; }
        [Currency]
        decimal MoneyProp { get; set; }

        // date-time
        DateTime DateTime2Prop { get; set; }

        // binary 
        [Size(64)]
        byte[] ByteArrayProp { get; set; }
        [Size(128), Nullable]
        System.Data.Linq.Binary BinaryProp { get; set; }

        //These types are not supported by SQL Server directly - but framework provides support for them and converts automatically when needed.
        sbyte SByteProp { get; set; }
        UInt16 UInt16Prop { get; set; }
        UInt32 UInt32Prop { get; set; }
        UInt64 UInt64Prop { get; set; }

        char CharProp { get; set; }
      }

      [Entity]
      //special, exotic types
      public interface ISpecialDataTypesEntity {
        // Guid
        [PrimaryKey]
        Guid Id { get; set; }

        DateTimeOffset DateTimeOffsetProp { get; set; }

        [Column(DbType = DbType.AnsiStringFixedLength, Size = 10)]
        string CharNProp { get; set; }

        [Column(DbType = DbType.StringFixedLength, Size = 10)]
        string NCharNProp { get; set; }

        [Column(DbType = DbType.AnsiString, Size = 10)]
        string VarCharProp { get; set; }

        [Column(DbType = DbType.Date)]
        DateTime DateProp { get; set; }

        [Column(DbType = DbType.DateTime)] //not DateTime2 which is default
        DateTime DateTimeProp { get; set; }

        [Column(DbType = DbType.Time)] 
        DateTime TimeProp { get; set; }

        [Column(DbType = CustomDbTypes.SmallDateTime)]
        DateTime SmallDateTimeProp { get; set; }

        [RowVersion]
        byte[] TimeStampProp { get; set; }

        [Column(DbType = CustomDbTypes.Geography), Nullable]
        byte[] GeographyProp { get; set; }

        [Column(DbType = CustomDbTypes.Geometry), Nullable]
        byte[] GeometryProp { get; set; }

        [Column(DbType = CustomDbTypes.HierarchyId), Nullable]
        byte[] HierarchyIdProp { get; set; }

        [Column(DbType = CustomDbTypes.Image)]
        byte[] ImageProp { get; set; }

        [Column(DbType = CustomDbTypes.NText)]
        string NTextProp { get; set; }

        [Column(DbType = CustomDbTypes.Text)]
        string TextProp { get; set; }

        [Column(DbType = DbType.Xml)]
        string XmlProp { get; set; }

        [Column(DbType = CustomDbTypes.SmallMoney)]
        Decimal SmallMoneyProp { get; set; }

        [Column(DbType = DbType.Binary, Size = 100)]
        byte[] BinaryProp { get; set; }

        [Column(DbType = DbType.Object)] //SQL_Variant in SQL server
        object SqlVariantProp { get; set; }
      }

      // We skip defining custom entity module and use base EntityModule class
      public class DataTypesTestEntityModel : EntityModel {
        public DataTypesTestEntityModel() {
          var area = AddArea("Types", "types");
          var mainModule = new EntityModule(area, "MainModule");
          mainModule.RegisterEntities(typeof(IDataTypesEntity));
          if (SetupHelper.ServerType == ServerType.MsSql)
            mainModule.RegisterEntities(typeof(ISpecialDataTypesEntity));
        }
      }//class

      DataTypesTestEntityModel _model;
      IEntityStore _entityStore;

      [TestInitialize]
      public void Init() {
        _model = new DataTypesTestEntityModel();
        _entityStore = SetupHelper.CreateTestEntityStore(_model);
      }


    [TestMethod, TestCategory("Vita_Base")]
    public void TestDataTypes() {

      var session = _entityStore.OpenSession();
      var ent = session.NewEntity<IDataTypesEntity>();
      var id = ent.Id = Guid.NewGuid();
      ent.StringProp = "abcd";
      ent.MemoProp = "0123456789 0123456789";
      ent.ByteProp = 255;
      ent.BoolProp = true;
      ent.Int16Prop = -234;
      ent.Int32Prop = -345;
      ent.Int64Prop = -456;
      ent.CharProp = 'X';

      ent.DoubleProp = 3.456;
      ent.DoublePropNull = 1.2345;
      ent.SingleProp = 4.567f;
      ent.DecProp = 2.34m;
      ent.MoneyProp = 3.45m;
      ent.DateTime2Prop = DateTime.Now;
      ent.ByteArrayProp = new byte[] { 1, 2, 3 };
      ent.BinaryProp = new System.Data.Linq.Binary(new byte[] { 4, 5, 6 });

      ent.SByteProp = 12;
      ent.UInt16Prop = 456;
      ent.UInt32Prop = 567;
      ent.UInt64Prop = 678;

      session.SaveChanges();

      var session2 = _entityStore.OpenSession();
      var ent2 = session2.GetEntity<IDataTypesEntity>(id);
      Assert.IsNotNull(ent2, "Failed to load saved record");

      Assert.AreEqual(ent.StringProp, ent2.StringProp.Trim());
      Assert.AreEqual(ent.MemoProp, ent2.MemoProp); 
      Assert.AreEqual(ent.BoolProp, ent2.BoolProp);
      Assert.AreEqual(ent.CharProp, ent2.CharProp);
      Assert.AreEqual(ent.ByteProp, ent2.ByteProp);
      Assert.AreEqual(ent.Int16Prop, ent2.Int16Prop);
      Assert.AreEqual(ent.Int32Prop, ent2.Int32Prop);
      Assert.AreEqual(ent.Int64Prop, ent2.Int64Prop);

      Assert.AreEqual(ent.DoubleProp, ent2.DoubleProp);
      Assert.AreEqual(ent.DoublePropNull, ent2.DoublePropNull);
      Assert.AreEqual(ent.SingleProp, ent2.SingleProp);
      Assert.AreEqual(ent.DecProp, ent2.DecProp);
      Assert.AreEqual(ent.MoneyProp, ent2.MoneyProp);

      Assert.IsTrue(Equal(ent.DateTime2Prop, ent2.DateTime2Prop));
      Assert.IsTrue(Equal(ent.ByteArrayProp, ent2.ByteArrayProp));
      Assert.IsTrue(Equal(ent.BinaryProp.ToArray(), ent2.BinaryProp.ToArray()));

      Assert.AreEqual(ent.SByteProp, ent2.SByteProp);
      Assert.AreEqual(ent.UInt16Prop, ent2.UInt16Prop);
      Assert.AreEqual(ent.UInt32Prop, ent2.UInt32Prop);
      Assert.AreEqual(ent.UInt64Prop, ent2.UInt64Prop);

      //try to set nullable double to Null
      ent.DoublePropNull = null;
      session.SaveChanges();
      session = _entityStore.OpenSession(); 
      ent = session.GetEntity<IDataTypesEntity>(id);
      Assert.AreEqual(null, ent.DoublePropNull, "Null expected in Nullable property");

/*
      // LINQ
      var dtHourAgo = DateTimeOffset.Now.AddHours(-1);

#if SQLCE
      //SQL CE needs some adjustment, read more in method comments
      dtHourAgo = Vita.Data.SqlCe.SqlCeUtility.AdjustDateTimeOffsetForLinq(dtHourAgo);
#endif
      var qry = from e in session.EntitySet<IDataTypesEntity>()
                where e.DateTimeOffsetProp > dtHourAgo
                select e;
      var qryList = qry.ToList();
      Assert.IsTrue(qryList.Count > 0, "Failed to query by DateTimeOffset");
*/
    }//method

#if MSSQL
    [TestMethod, TestCategory("Vita_Base")]
#endif
    public void TestSpecialDataTypes() {
      // Verify data types
      var db = Database.GetDatabase(_entityStore); 
      var schemaMgr = db.DbModel.Driver.CreateSchemaManager(db.ConnectionString);
      var dtColumns = schemaMgr.GetColumns("types");
      CheckDataType(dtColumns, "CharNProp", "char", 10);
      CheckDataType(dtColumns, "NCharNProp", "nchar", 10);
      CheckDataType(dtColumns, "VarCharProp", "varchar", 10);
      CheckDataType(dtColumns, "DateProp", "date");
      CheckDataType(dtColumns, "DateTimeProp", "datetime");
      CheckDataType(dtColumns, "TimeProp", "time");
      CheckDataType(dtColumns, "SmallDateTimeProp", "smalldatetime");
      CheckDataType(dtColumns, "TimeStampProp", "timestamp");
      CheckDataType(dtColumns, "GeographyProp", "geography");
      CheckDataType(dtColumns, "GeometryProp", "geometry");
      CheckDataType(dtColumns, "HierarchyIdProp", "hierarchyid");
      CheckDataType(dtColumns, "ImageProp", "image");
      CheckDataType(dtColumns, "NTextProp", "ntext");
      CheckDataType(dtColumns, "TextProp", "text");
      CheckDataType(dtColumns, "XmlProp", "xml");
      CheckDataType(dtColumns, "SmallMoneyProp", "smallmoney");
      CheckDataType(dtColumns, "BinaryProp", "binary", 100);
      CheckDataType(dtColumns, "SqlVariantProp", "sql_variant");

      var session = _entityStore.OpenSession();
      var ent = session.NewEntity<ISpecialDataTypesEntity>();
      ent.Id = Guid.NewGuid();
      ent.CharNProp = "12345678";
      ent.NCharNProp = "234567";
      ent.VarCharProp = "3456789";
      ent.DateProp = DateTime.Now.Date;
      ent.DateTimeProp = DateTime.Now;
      ent.TimeProp = DateTime.Now;
      ent.SmallDateTimeProp = DateTime.Now;
      ent.DateTimeOffsetProp = DateTimeOffset.Now;

      // ent.TimeStampProp is set automatically by database
/* 
      // Have no idea how to properly assign these properties; assigning random data does not work, so made columns nullable and skip assignment 
      ent.GeographyProp = new byte[] { 1, 2, 3, 4 };
      ent.GeometryProp = new byte[] { 5, 6, 7, 8 };
      ent.HierarchyIdProp = new byte[] { 11, 12, 13, 14 };
 */ 
      ent.ImageProp = new byte[] { 21, 22, 23, 24 };
      ent.NTextProp = "abcd";
      ent.TextProp = "defg";
      ent.XmlProp = "<foo/>";
      ent.SmallMoneyProp = 1.23m;
      ent.BinaryProp = new byte[] { 31, 32 };
      ent.SqlVariantProp = 1234;

      session.SaveChanges(); 
    }//test method

    private void CheckDataType(DataTable dtColumns, string columnName, string dataType, int size = 0) {
      //find row
      dtColumns.DefaultView.RowFilter = "column_name = '" + columnName + "'";
      var colRow = dtColumns.DefaultView[0];
      Assert.AreEqual(dataType, (string)colRow["Data_Type"], "Data type does not match, column: " + columnName);
      if (size > 0)
        Assert.AreEqual(size, (int)colRow["character_maximum_length"], "Size does not match, column: " + columnName);
    }
    //Helper methods to compare 
    private bool Equal(DateTime x, DateTime y) {
      int precMs = 10;
#if MYSQL
      precMs = 1000;
#endif
      return y > x.AddMilliseconds(-precMs) && y < x.AddMilliseconds(precMs); 
    }
    private bool Equal(DateTimeOffset x, DateTimeOffset y) {
      return y > x.AddMilliseconds(-10) && y < x.AddMilliseconds(10);
    }

    private bool Equal(byte[] x, byte[] y) {
      if (x.Length != y.Length)
        return false;
      for (int i = 0; i < x.Length; i++)
        if (x[i] != y[i])
          return false;
      return true; 
    }

  }

}
