﻿using System;
using System.Text;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Diagnostics;

using Vita.Entities;
using Vita.Entities.Runtime;

namespace Vita.UnitTests.EntityRefs {
#if USE_NUNIT
  using NUnit.Framework;
  using TestClass = NUnit.Framework.TestFixtureAttribute;
  using TestMethod = NUnit.Framework.TestAttribute;
  using TestInitialize = NUnit.Framework.SetUpAttribute;
  using TestCategory = NUnit.Framework.CategoryAttribute;
#else
  using Microsoft.VisualStudio.TestTools.UnitTesting;
  using System.ComponentModel;
  using Vita.Common;
  using Vita.Services;
#endif

  [Entity, OrderBy("Model")]
  public interface ICar {
    [PrimaryKey, Auto]
    Guid Id { get; set; }

    [Size(30)]
    //we test that these attributes will be passed to the object's property
    [System.ComponentModel.Description("Car model description.")]
    [Browsable(true), DisplayName("Car model"), System.ComponentModel.CategoryAttribute("Miscellaneous")]
    string Model { get; set; }

    IPerson Owner { get; set; }
    [Nullable]
    IPerson Driver { get; set; }
  }

  [Entity]
  public interface IPerson {
    [PrimaryKey, Auto]
    Guid Id { get; set; }

    [Size(30)]
    string Name { get; set; }

    [OneToMany("Owner")]
    IList<ICar> Cars { get; set; }

  }

  // We skip defining custom entity module and use base EntityModule class
  public class MiscTestsEntityModel : EntityModel {
    public MiscTestsEntityModel() {
      var area = AddArea("refs", "refs");
      var mainModule = new EntityModule(area, "MainModule");
      mainModule.RegisterEntities(typeof(ICar), typeof(IPerson));
    }
  }//class


  [TestClass]
  public class MiscTests {
    MiscTestsEntityModel _model;
    IEntityStore _entityStore; 

    [TestInitialize]
    public void Init() {
      _model = new MiscTestsEntityModel();
      _entityStore = SetupHelper.CreateTestEntityStore(_model);
    }

    [TestMethod, TestCategory("Vita_Base")]
    public void TestEntityReferences() {
      SetupHelper.DeleteAll(_entityStore, typeof(ICar), typeof(IPerson));
      var session = _entityStore.OpenSession();

      // Quick test for fix of a bug - entityRef entities are not null (initialized with stubs) on new entities
      var carx = session.NewEntity<ICar>();
      var carxOwner = carx.Owner;
      Assert.IsNull(carxOwner, "EntityRef on new entity is not null!");

      session = _entityStore.OpenSession();
      var john = session.NewEntity<IPerson>();
      john.Name = "John S";
      var johnId = john.Id;

      var jane = session.NewEntity<IPerson>();
      jane.Name = "Jane M";
      var janeId = jane.Id; 

      var car1 = session.NewEntity<ICar>();
      car1.Model = "Beatle";
      car1.Owner = john;
      car1.Driver = john;
      var car1Id = car1.Id; 
      var car2 = session.NewEntity<ICar>();
      car2.Model = "Explorer";
      car2.Owner = john;
      car2.Driver = john;
      var car2Id = car2.Id;
      session.SaveChanges();

      //Start new session, load all and check that IDs and relationships are correct
      session = _entityStore.OpenSession();
      john = session.GetEntity<IPerson>(johnId);
      Assert.IsNotNull(john, "John is not found.");
      Assert.AreEqual(2, john.Cars.Count, "Invalid # of cars with John.");
      car1 = john.Cars[0];
      car2 = john.Cars[1];
      Assert.IsTrue(car1.Owner == john, "Car1 owner expected to be John");
      Assert.IsTrue(car2.Owner == john, "Car2 owner expected to be John");
      Assert.IsTrue(car1.Driver == john, "Car1 driver expected to be John");
      Assert.IsTrue(car2.Driver == john, "Car2 driver expected to be John");
      //Transfer car2 to jane
      jane = session.GetEntity<IPerson>(janeId);
      car2.Owner = jane; 
      car2.Driver = null;
      session.SaveChanges();

      session = _entityStore.OpenSession();
      john = session.GetEntity<IPerson>(johnId);
      Assert.AreEqual(1, john.Cars.Count, "Invalid # of John's cars after transfer.");
      jane = session.GetEntity<IPerson>(janeId);
      Assert.AreEqual(1, jane.Cars.Count, "Invalid # of Jane's cars after transfer.");
      car2 = jane.Cars[0];
      Assert.AreEqual(jane, car2.Owner, "Invalid car2 owner after transfer.");
      Assert.IsNull(car2.Driver, "Car2 driver expected to be NULL");

      //Check that entity reference is returned initially as record stub
      session = _entityStore.OpenSession();
      car1 = session.GetEntity<ICar>(car1Id);
      var car1driver = car1.Driver;
      var rec = EntityHelper.GetRecord(car1driver);
      Assert.AreEqual(EntityStatus.Stub, rec.Status, "Entity reference is not loaded as stub.");
    }//method

    // The framework automatically copies some component model attribute (lie Description, DisplayName, Browsable)
    // for entity interfaces to generated classes. We test here this copying process.
    [TestMethod, TestCategory("Vita_Base")]
    public void TestStandardAttributes() {
      SetupHelper.DeleteAll(_entityStore, typeof(ICar), typeof(IPerson));
      var session = _entityStore.OpenSession();
      var car = session.NewEntity<ICar>(); 
      var modelProp = car.GetType().GetProperty("Model");
      var attrs = modelProp.GetCustomAttributes(true).ToList();
      Assert.IsTrue(attrs.Count > 0, "No standard attributes found on Car.Model property in the generated class.");
      Assert.IsNotNull(attrs.FirstOrDefault(a => a is System.ComponentModel.DisplayNameAttribute), "DisplayNameAttribute not found in the generated class.");
      Assert.IsNotNull(attrs.FirstOrDefault(a => a is System.ComponentModel.DescriptionAttribute), "DescriptionAttribute not found in the generated class.");
      Assert.IsNotNull(attrs.FirstOrDefault(a => a is System.ComponentModel.BrowsableAttribute), "BrowsableAttribute not found in the generated class.");
      Assert.IsNotNull(attrs.FirstOrDefault(a => a is System.ComponentModel.CategoryAttribute), "CategoryAttribute not found in the generated class.");
    }

// Disabled, enable if need to check this
    [TestMethod, TestCategory("Vita_Base")]
    public void TestMultipleReadWrites() {
      SetupHelper.DeleteAll(_entityStore, typeof(ICar), typeof(IPerson));
      //We test here that multiple sessions/selects don't bring down the system - that connections are opened/closed appropriately 
      // and connection pool is not exausted.
      var session = _entityStore.OpenSession();
      var timeService = _entityStore.GetService<ITimeService>(); 

      var john = session.NewEntity<IPerson>();
      john.Name = "John M";
      var johnId = john.Id;
      session.SaveChanges(); 

      int Count = 200;
      //GC.Collect();
      //Thread.Sleep(100);

      var start = timeService.ElapsedMilliseconds;
      //writes
      for (int i = 0; i < Count; i++) {
        session = _entityStore.OpenSession();
        john = session.GetEntity<IPerson>(johnId);
        john.Name = "John_" + i;
        session.SaveChanges();
      }

      //SQL CE: normal execution time is around 450 ms; but sometimes SQL CE works slower, 10 times slower.
      // the test runs at 6500 ms instead of 450ms
      // TODO: investigate SQL CE slow behavior
      int Timeout = 5000; 
#if SQLCE
      Timeout = 10000; 
#endif
      var ticks = timeService.ElapsedMilliseconds - start; // Expected: MS SQL - 1600ms, SQL CE - 450 ms
      Assert.IsTrue(ticks < Timeout, "Too much time for multiple reaq/writes test, ms: " + ticks); 
      System.Diagnostics.Debug.WriteLine("\r\nTestMultipleReadWrites: ticks = " + ticks + "\r\n");
    }

  }
}
