﻿using System;
using System.Text;
using System.Collections.Generic;
using System.Linq;
using System.Threading;

using Vita.Common;
using Vita.Data.Caching;
using Vita.Entities.Runtime;


namespace Vita.UnitTests {
#if USE_NUNIT
  using NUnit.Framework;
  using TestClass = NUnit.Framework.TestFixtureAttribute;
  using TestMethod = NUnit.Framework.TestAttribute;
  using TestInitialize = NUnit.Framework.SetUpAttribute;
  using TestCategory = NUnit.Framework.CategoryAttribute;
#else
  using Microsoft.VisualStudio.TestTools.UnitTesting;
  using System.Threading.Tasks;
  using System.Diagnostics;
  using Vita.Services;
#endif

  [TestClass]
  public class ObjectCacheTest {

    // Note: this test might not work in step-by-step debugger, it depends on timing of background thread activation and millisecond-scale timing
    [TestMethod, TestCategory("Vita_Base")]
    public void TestObjectCache() {
      var objCache = new ObjectCache<string, string>(10, TimeSpan.FromMilliseconds(100), TimeSpan.FromMilliseconds(50));
      // Setup cleanup event listener
      bool cleanupHappened = false;
      objCache.Cleanup += (src, args) => { cleanupHappened = true; };
      // Test 1 - test time-based expiration and auto-cleanup of expired values -------------------------------------------------------
      //fill out cache
      for (int i = 0; i < 5; i++)
        objCache.Add("k" + i, "V" + i);
      //read cached value
      var v1 = objCache.Lookup("k1");
      Assert.AreEqual("V1", v1, "Cache lookup for k1 failed.");
      // shift time, wait for cleanup
      Thread.Sleep(200);
      //wait for cleanup; since time is +200 ms, it should expell all records
      cleanupHappened = false;
      while (!cleanupHappened)
        Thread.Sleep(1);
      // now the cache should be empty  
      var count = objCache.Count;
      Assert.AreEqual(0, count, "Cache is not empty after expiration");

      // Test 2 - test auto-cleanup on cache capacity overflow -------------------------------------------------------------------------
      objCache.Clear();
      // fill up
      for (int i = 0; i < 10; i++) //fill-up
        objCache.Add("k" + i, "V" + i);
      // shift time
      Thread.Sleep(20);
      objCache.RefreshTime(); // forces cache to read current tick count
      // add 10 more - the cache will become over capacity
      for (int i = 10; i < 20; i++)
        objCache.Add("k" + i, "V" + i); //overflow will request immediate cleanup
      var oldCount = objCache.Count; // we should have overflow now, cache over capacity
      //wait until cleanup runs
      cleanupHappened = false; 
      while (!cleanupHappened)
        Thread.Sleep(1);
      // the cleanup should have expelled all 'older' values, to bring count under capacilty
      // We should have only new values, but not all - with cache cleanup on overflow, cache is reduced to 80% perc of max size
      var newCount = objCache.Count;
      Assert.IsTrue(newCount >= 7, "Cache after cleanup is less than 80% of max size.");

      // check all old values are gone
      for (int i = 0; i < 10; i++) {
        var v = objCache.Lookup("k" + i);
        Assert.IsNull(v, "Cache did not expell old value V" + i);
      }
    }


  }
}
