﻿using System;
using System.ComponentModel;
using System.Linq;
using System.Diagnostics;

using Vita.Entities;
using Vita.Entities.Runtime;

namespace Vita.UnitTests.DataTypes {
#if USE_NUNIT
  using NUnit.Framework;
  using TestClass = NUnit.Framework.TestFixtureAttribute;
  using TestMethod = NUnit.Framework.TestAttribute;
  using TestInitialize = NUnit.Framework.SetUpAttribute;
  using TestCategory = NUnit.Framework.CategoryAttribute;
#else
  using Microsoft.VisualStudio.TestTools.UnitTesting;
#endif

  //RowVersion and Identity are not supported for SQL CE
#if !SQLCE
  [TestClass]
#endif
  public class RowVersionTests {

    [Entity]
    public interface IVersionedProduct {

      // Guid
      [PrimaryKey, Identity]
      int Id { get; }
      string Name { get; set; }
      [Size(100)]
      string Description { get; set; }
      double Price { get; set; }

      [RowVersion]
      byte[] RowVersion { get; set; }
    }

    // We skip defining custom entity module and use base EntityModule class
    public class RowVersionTestsEntityModel : EntityModel {
      public RowVersionTestsEntityModel() {
        var area = AddArea("rv", "rv");
        var mainModule = new EntityModule(area, "MainModule");
        mainModule.RegisterEntities(typeof(IVersionedProduct));
      }
    }//class

    [TestInitialize]
    public void Setup() {
    }

#if MSSQL
    [TestMethod]
#endif
    public void TestRowVersion() {
      SetupHelper.DropSchemaObjects("rv");
      var model = new RowVersionTestsEntityModel();
      var entStore = SetupHelper.CreateTestEntityStore(model);

      try {
        var session = entStore.OpenSession();
        var prod = session.NewEntity<IVersionedProduct>();
        prod.Name = "Coffee Maker 2000";
        prod.Description = "Automatic coffee maker.";
        prod.Price = 22.99;
        session.SaveChanges();
        var prodId = prod.Id;

        var rv1 = prod.RowVersion;
        prod.Price = 19.99;
        session.SaveChanges();
        var rv2 = prod.RowVersion;
        Assert.IsFalse(rv2.EqualValues(rv1), "Row version did not increase.");

        // Test LINQ query against table with Row version column - make LINQ can read rowversion column
        session = entStore.OpenSession();
        var q = from vp in session.EntitySet<IVersionedProduct>()
                where vp.Price > 5
                select vp;
        var pList = q.ToList();
        Assert.IsTrue(pList.Count > 0, "Failed to run LINQ query.");
        var rv3 = pList[0].RowVersion;
        Assert.IsTrue(rv3.EqualValues(rv2), "Row version from LINQ query mismatch.");

        // Concurrent update through different sessions
        // Original session and product entity
        session = entStore.OpenSession();
        prod = session.GetEntity<IVersionedProduct>(prodId);
        // other, concurrent session, updates the same entity
        var session2 = entStore.OpenSession();
        var prodCopy = session2.GetEntity<IVersionedProduct>(prodId);
        prodCopy.Price = 18.99;
        session2.SaveChanges();
        // Original session tries to update now stale entity - should get ConcurrentUpdateException
        try {
          prod.Price = 17.99;
          session.SaveChanges();
          Assert.IsTrue(false, "Expected concurrent update exception.");
        } catch (ConcurrentUpdateException cux) {
          Assert.AreEqual("[rv].[VersionedProduct]", cux.TableName, "TableName mismatch in concurrent update exception");
          Debug.WriteLine("Concurrent update conflict - as expected.");
        }


      } catch (ValidationException ve) {
        Debug.WriteLine("Validation errors: " + string.Join("\r\n", ve.GetMessages()));
        throw;
      } catch (Exception ex) {
        Debug.WriteLine("Exception: " + ex.ToString());
        throw;
      }
    }
  
  }//class
}//namespace
