﻿using System;
using System.Linq;
using System.Diagnostics;

using Vita.Entities;
using Vita.Entities.Model;
using Vita.Data;
using Vita.Data.Model;
using Vita.Common;

namespace Vita.UnitTests.SchemaUpdate {
#if USE_NUNIT
  using NUnit.Framework;
  using TestClass = NUnit.Framework.TestFixtureAttribute;
  using TestMethod = NUnit.Framework.TestAttribute;
  using TestInitialize = NUnit.Framework.SetUpAttribute;
  using TestCategory = NUnit.Framework.CategoryAttribute;
#else
  using Microsoft.VisualStudio.TestTools.UnitTesting;
  using Vita.Data.Model;
  using Vita.Data.Management;
#endif

  [TestClass] 
  public class SchemaUpdateTests {


    //Original model consisting of 2 tables
    public class OriginalModule : EntityModule {
      [Entity, Index("IntProp,StringProp")]
      public interface IParentEntity {
        [PrimaryKey, Auto, ClusteredIndex]
        Guid Id { get; set; }
        [Size(40)]
        string StringProp { get; set; }
        int IntProp { get; set; }
        [Column(Default="123")]
        int IntPropWithDefault { get; set; }
        double DoubleProp { get; set; }
        System.Single SingleProp { get; set; }
        [Size(20)]
        string StringProp2_OldName { get; set; }
        IEntityToDelete EntityToDeleteRef { get; set; }
      }

      [Entity]
      public interface IChildEntity {
        [PrimaryKey, Auto]
        Guid Id { get; set; }
        IParentEntity Parent { get; set; }
        IParentEntity ParentRefToDelete { get; set; }
        [Size(20)]
        string Name { get; set; }
      }

      [Entity]
      public interface IEntityToDelete {
        [PrimaryKey, Auto]
        Guid Id { get; set; }
        string Name { get; set; }
      }

      // An entity with some reserved words used as names, to test that it is handled properly
      [Entity]
      public interface ITable {
        [PrimaryKey, Auto]
        Guid Id { get; set; }
        DateTime Begin { get; set; }
        DateTime End { get; set; }
        string User { get; set; }
        string Insert { get; set; }
      }

      public OriginalModule(EntityArea area, string name) : base(area, name) {
        base.RegisterEntities(typeof(IParentEntity), typeof(IChildEntity), typeof(IEntityToDelete), typeof(ITable));
      }
    }//class

    //Changed model
    public class ChangedModule : EntityModule {
      [Entity, Index("IntProp:desc,StringProp:asc")]
      public interface IParentEntity {
        [PrimaryKey, Auto]
        Guid Id { get; }
        [Size(50)] //change size; clustered index changes as well, but we specify it now using Companion type.
        string StringProp { get; set; }
        int IntProp { get; set; }
        double DoubleProp { get; set; }
        [Size(20), OldNames("StringProp2_OldName")]
        string StringProp2_NewName { get; set; }

        // Add non-nullable properties
        int IntPropAdded { get; set; }
        DateTime DateTimePropAdded { get; set; }
      }

      [Entity, OldNames("IChildEntity")]
      public interface IChildEntityRenamed {
        [PrimaryKey]
        Guid Id { get; set; }
        IParentEntity Parent { get; set; }
        [Size(20)]
        string Name { get; set; }
      }

      [Entity]
      public interface ITable {
        [PrimaryKey, Auto]
        Guid Id { get; set; }
        DateTime Begin { get; set; }
        DateTime End { get; set; }
        string Insert { get; set; }
      }

      [ClusteredIndex("StringProp,Id")]
      public interface IParentEntityKeys : IParentEntity {
      }

      public ChangedModule(EntityArea area, string name) : base(area, name) {
        base.RegisterEntities(typeof(IParentEntity), typeof(IChildEntityRenamed), typeof(ITable));
        base.RegisterCompanionTypes(typeof(IParentEntityKeys));

      }
    }//class

    [TestMethod, TestCategory("Vita_Base")]
    public void TestSchemaUpdate() {
      try {
        TestSchemaUpdateImpl(); 
      } catch (Exception ex) {
        // write to debug stream - it will appear in test report
        Debug.WriteLine(ex.ToString());
        throw;
        
      }
    }

    private void TestSchemaUpdateImpl() {
      var schema = "upd";
      SetupHelper.DropSchemaObjects(schema); 

      //version 1 of model/schema
      var entModel = new EntityModel();
      var area = entModel.AddArea("SchemaUpdates", schema);
      var module = new OriginalModule(area, "model");


      var entStore = SetupHelper.CreateTestEntityStore(entModel); //updates schema
      //SetupHelper.DropSchemaObjects(entStore, schema);
      var db = Database.GetDatabase(entStore);
      var supportsClustIndex = db.DbModel.Driver.Supports(Vita.Data.Driver.DbCapabilities.ClusteredIndexes);

      // Load DbModel and verify it is correct
      var dbModel = SetupHelper.LoadDbModel(schema);
      Assert.AreEqual(4, dbModel.Tables.Count(), "Expected 4 tables.");
      var parTable = dbModel.GetTable(schema, "ParentEntity");
      Assert.AreEqual(8, parTable.Columns.Count, "Invalid number of columns in parent table.");
      var keyCount = CountKeys(parTable);
      //Keys: PK, FK to IEntityToDelete, index on FK to IEntityToDelete, index on IntProp,StringProp, (ClusteredIndex?) 
      Assert.AreEqual(5, keyCount, "Invalid # of keys in parent table.");

      if (supportsClustIndex) {
        var parCI = parTable.Keys.First(k => k.KeyType.IsSet(KeyType.ClusteredIndex));
        Assert.AreEqual(1, parCI.Columns.Count, "Invalid numbe of fields in clustered index.");
      }

      //child entity
      var childTable = dbModel.GetTable(schema, "ChildEntity");
      Assert.AreEqual(4, childTable.Columns.Count, "Invalid number of columns in child table.");
      // some servers like MySql create FK and Index on it under the same name. When loading, such a key should be marked with 2 flags
      // so let's count these separately; should be 5: PK + 2 foreign keys + 2 indexes on FK
      keyCount = CountKeys(childTable);
      Assert.AreEqual(5, keyCount,  "Invalid # of keys in child table.");

      //Create a few records
      var session = entStore.OpenSession();
      var parent = session.NewEntity<OriginalModule.IParentEntity>();
      parent.IntProp = 4;
      parent.StringProp = "Some string";
      parent.SingleProp = 4.56f;
      parent.DoubleProp = 5.67;
      parent.IntPropWithDefault = 567;
      parent.StringProp2_OldName = "Old string";
      var child = session.NewEntity<OriginalModule.IChildEntity>();
      child.Parent = parent;
      child.ParentRefToDelete = parent; 
      child.Name = "Child name";
      var entToDelete = session.NewEntity<OriginalModule.IEntityToDelete>();
      entToDelete.Name = "some-name";
      parent.EntityToDeleteRef = entToDelete; 
      session.SaveChanges(); 

      //Now change to version 2 ================================================================
      entModel = new EntityModel();
      area = entModel.AddArea("SchemaUpdates", schema);
      var module2 = new ChangedModule(area, "model");
      entStore = SetupHelper.CreateTestEntityStore(entModel, dropUnknownTables: true); 
      db = Database.GetDatabase(entStore); 

      //At this point the schema should have been updated; let's check it      
      // Load DbModel and verify it is correct
      dbModel = SetupHelper.LoadDbModel(schema);
      Assert.AreEqual(3, dbModel.Tables.Count(), "Expected 3 tables after update.");
      parTable = dbModel.GetTable(schema, "ParentEntity");
      Assert.AreEqual(7, parTable.Columns.Count, "Invalid number of columns in parent table after schema update.");
      Assert.AreEqual(3, parTable.Keys.Count, //PK, Clustered index, index on IntProp,StringProp
         "Invalid # of keys in parent table after update.");
      if (supportsClustIndex) {
        //SQL CE does not support clustered indexes
        var parCI = parTable.Keys.First(k => k.KeyType.IsSet(KeyType.ClusteredIndex));
        Assert.AreEqual(2, parCI.Columns.Count, "Invalid number of fields in clustered index."); //
      }
      //child entity
      childTable = dbModel.GetTable(schema, "ChildEntityRenamed");
      Assert.AreEqual(3, childTable.Columns.Count, "Invalid number of columns in child table after update.");
      keyCount = CountKeys(childTable); // = 3:  Clustered PK, FK to parent, index on FK
      Assert.AreEqual(3, keyCount, "Invalid # of keys in child table after update.");

      // Now create model again, compare it and make sure no schema updates
      entStore = SetupHelper.CreateTestEntityStore(entModel, updateSchema: false);
      db = Database.GetDatabase(entStore);
      var log = new Logging.Log();
      var dbUpdater = new DbSchemaUpdater(db,  log);
      var changes = dbUpdater.GetSchemaUpdates();
      if (changes.Actions.Count > 0) {
        var strUpdates = string.Join(Environment.NewLine, changes.Actions);
        Debug.WriteLine("Detected updates when no schema changes should be present:");
        Debug.WriteLine(strUpdates);
        Assert.IsTrue(false, "Schema changes count should be zero.");
      }


    }//method

    // For some providers keys might be combined; so to properly account for this, we count keys this way - separately by each type.
    private int CountKeys(DbTableInfo table) {
      return CountKeys(table, KeyType.PrimaryKey) + CountKeys(table, KeyType.ForeignKey) + CountKeys(table, KeyType.Index);
    }

    private int CountKeys(DbTableInfo table, KeyType keyType) {
      return table.Keys.Count(k => k.KeyType.IsSet(keyType));
    }

  }//class


}
