﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Configuration;
using System.Diagnostics;

using Vita.Common;
using Vita.Entities;
using Vita.Entities.Model;
using Vita.Entities.Runtime;
using Vita.Data;
using Vita.Data.Driver;
using Vita.Data.Management;
using Vita.Data.Model;
using Vita.Activation;

namespace Vita.UnitTests {

  public enum ServerType {
    MsSql,
    SqlCe,
    MySql,
    Oracle,
  }

  public static class SetupHelper {
    private static DbDriver _driver;
    private static DbOptions _dbOptions = DbOptions.Default;
    private static string _connectionStringKey;
    public static ServerType ServerType;

    //static constructor
    static SetupHelper() {
#if SQLCE
    ServerType = ServerType.SqlCe;
      _dbOptions = Vita.Data.SqlCe.SqlCeDbDriver.DefaultDbOptions;
      _driver = new Vita.Data.SqlCe.SqlCeDbDriver();
     _connectionStringKey = "SqlCeConnectionString";
#elif MYSQL
    ServerType = ServerType.MySql;
      _driver = new Vita.Data.MySql.MySqlDriver();
      _connectionStringKey = "MySqlConnectionString";
      _dbOptions |= DbOptions.IgnoreTableNamesCase;
#elif MSSQL08
      ServerType = ServerType.MsSql;
      _connectionStringKey = "MsSql2008ConnectionString";
      _driver = new Vita.Data.MsSql.MsSqlDriver(Data.MsSql.MsSqlVersion.V2008);
      _dbOptions |= DbOptions.AutoIndexForeignKeys;
#elif MSSQL12
      ServerType = ServerType.MsSql;
      _connectionStringKey = "MsSql2012ConnectionString";
      _driver = new Vita.Data.MsSql.MsSqlDriver(Data.MsSql.MsSqlVersion.V2012);
      _dbOptions |= DbOptions.AutoIndexForeignKeys;
#else
      ERROR_DEFINE_MISSING(); 
      //  Error: No conditional symbol defined for driver type. Define one of the following symbols in project properties:
      //     MSSQL08, MSSQL12, SQLCE, MYSQL
#endif
    }

    public const string LogFilePathKey = "LogFilePath";

    public static IEntityStore CreateTestEntityStore(EntityModel model, bool updateSchema = true, bool dropUnknownTables = false) {
      try {
        CheckCanConnect();
        var logPath = GetLogFilePath();
        var connString = GetConnectionString();
        var updateOptions = DbSchemaUpdateOptions.VerifyAll;
        if (updateSchema)
          updateOptions |= DbSchemaUpdateOptions.UpdateSchema;
        if (dropUnknownTables)
          updateOptions |= DbSchemaUpdateOptions.DropUnknownTables;
        var schemas = model.Areas.Select(a => a.SchemaName);
        var dbSettings = new DbSettings(_driver, _dbOptions, connString, updateOptions, schemas: schemas);
        var entStore = model.CreateEntityStore(dbSettings, logFilePath: logPath); 
        return entStore;
      } catch (ActivationException sx) {
        //Unit test framework shows only ex message, not details; let's write specifics into debug output - it will be shown in test failure report
        Debug.WriteLine("EntityStore init exception: ");
        Debug.WriteLine(sx.Log);
        throw;
      }
    }

    private static void CheckCanConnect() {
      var conn = _driver.CreateConnection(GetConnectionString());
      conn.Open();
      conn.Close();
    }

    public static void DeleteAll(IEntityStore store, params Type[] entitiesToDelete) {
      var entStore = (EntityStore)store;
      var db = Database.GetDatabase(entStore); 
      ActivationHelper.DeleteAllData(db, entitiesToDelete); 
    }

    public static void DropSchemaObjects(string schema) {
      var connString = GetConnectionString();
      var schemaMgr = _driver.CreateSchemaManager(connString);
      schemaMgr.DropSchemaObjects(schema);
    }
      

    public static DbModel LoadDbModel(string schema) {
      var loader = new Vita.Data.Management.DbModelLoader();
      var log = new Vita.Logging.Log();
      var connString = GetConnectionString();
      var dbSettings = new DbSettings(_driver, _dbOptions, connString, schemas: new string[] {schema});
      var dbModel = loader.LoadModel(dbSettings, log);
      if (log.HasErrors)
        Util.Throw("Model loading errors: \r\n" + log.GetAllAsText());
      return dbModel;
    }


    public static string GetConnectionString() {
      var connString = ConfigurationManager.AppSettings[_connectionStringKey];
      if (connString.Contains("{bin}")) {
        var asmPath = Assembly.GetExecutingAssembly().Location;
        var binFolder = Path.GetDirectoryName(asmPath);
        connString = connString.Replace("{bin}", binFolder);
      }
      return connString;
    }
    public static string GetLogFilePath() {
      return ConfigurationManager.AppSettings[LogFilePathKey];
    }

  }//class

}
