﻿using System;
using System.CodeDom.Compiler;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Xml;
using Vita.Common;
using Vita.Data;
using Vita.Data.Model;
using Vita.Entities;
using Vita.SourceCodeGen;

namespace Vita.Tools.EntityGenerator {
  

  class Program {
    const string ErrorLogFile = "entitygen-error.log";
    static bool _noinput;

    static int Main(string[] args) {
      Console.Title = "VITA Entity Generator";
      Console.WindowWidth = 120;
      Console.WriteLine("VITA Application Framework. Entity Generator Tool.");
      Console.WriteLine("  http://vita.codeplex.com");
      Console.WriteLine(); 
      Console.WriteLine("This tool generates VITA entities (interfaces) from the database.");
      Console.WriteLine();
      var showHelp = args.FirstOrDefault(arg => arg.StartsWith("/?") || arg.StartsWith("/h")) != null;
      if (showHelp) {
        ShowHelp();
        return 0; 
      }
      _noinput = args.FirstOrDefault(arg => arg.StartsWith("/noinput")) != null; //no wait for input
      //get config file name
      var configFile = GetConfig(args);
      if (string.IsNullOrEmpty(configFile)) {
        ShowHelp();
        LogError("Invalid arguments - missing config file parameter.");
        WaitPressKey(); 
        return -1; //return error
      }
      //generate code
      try {
        GenerateCode(configFile);
      } catch (Exception ex) {
        var err = ex.ToString();
        Console.ForegroundColor = ConsoleColor.Red;
        Console.WriteLine(); 
        Console.WriteLine("Exception: ");
        Console.WriteLine(err);
        Console.ResetColor();
        LogError(err + "\r\n");
        WaitPressKey();
        return -1;
      }
      WaitPressKey(); 
      return 0; //success
    }//main

    private static void GenerateCode(string configFileName) {
      Console.WriteLine("Loading config file: " + configFileName);
      Util.Check(System.IO.File.Exists(configFileName), "File not found: {0}", configFileName); 
      var log = new Logging.Log();
      //Load file
      var xmlConfig = new XmlDocument();
      xmlConfig.Load(configFileName);
      var cfg = EntityCodeGeneratorConfig.FromXml(xmlConfig);  
      //setup driver. Note: code generation is not supported for SQL CE
      switch(cfg.DriverCode) {
        case "MSSQL": cfg.Driver = new Vita.Data.MsSql.MsSqlDriver(); break;
        case "MYSQL": cfg.Driver = new Vita.Data.MySql.MySqlDriver(); break;
        default: Util.Throw("Unknown DriverCode '{0}' in config file; must be MSSQL.", cfg.DriverCode); break;
      }
      // check connection
      Console.Write("Testing database connection...");
      var conn = cfg.Driver.CreateConnection(cfg.ConnectionString); 
      try {
        conn.Open();
        conn.Close(); 
        Console.WriteLine(" Done.");       
      } catch(Exception ex) {
        Console.WriteLine(" Failed: " + ex.Message);
        throw;       
      }
      // load, generate model and generate source
      Console.Write("Generating entity definitions...");
      var entModelGen = new EntityModelReverseGenerator();
      var entModel = entModelGen.Generate(cfg);
      var codeGen = new EntityCodeGenerator(entModel, cfg, log);
      var source = codeGen.GenerateSource();
      System.IO.File.WriteAllText(cfg.OutputPath, source);
      Console.WriteLine(" Done.");
      // check log
      if (log.HasErrors)
        throw new Exception("Errors encountered: " + log.GetAllAsText());
      if (!log.IsEmpty) {
        Console.WriteLine();
        Console.WriteLine("Warnings: ----------------------------------");
        Console.WriteLine(log.GetAllAsText());
        Console.WriteLine("End warnings -------------------------------");        
      }
      //Everything is ok
      Console.WriteLine("The source code is saved to: " + cfg.OutputPath);
      // Do test compile 
      CompilerResults compilerResults = null;
      Console.Write("Verifying generated code - compiling...");
      compilerResults = SourceCodeGenUtil.CompileSources(cfg.Driver.GetType(), source);
      if (compilerResults.Errors.Count > 0) {
        Console.WriteLine(" Failed. Errors: ");
        foreach (CompilerError err in compilerResults.Errors) {
          var strErr = "Compile error: " + err.ErrorText + " Line: " + err.Line;
          Console.WriteLine(strErr);
        }
        Console.WriteLine(); 
        return;
      }
      Console.WriteLine(" Done.");
      //Compare schema
      Console.Write("Verifying generated entities. Running schema comparison ...");
      var fullModelName = cfg.Namespace + "." + cfg.ModelClassName;
      var modelType = compilerResults.CompiledAssembly.GetType(fullModelName);

      // important - do not use DbOptions.AutoIndexForeignKeys - which is recommended for MS SQL, but is not helpful here.
      // This will create a bunch of extra indexes on FKs in entities schema and result in extra differences with original schema.
      var nonProcActions = SourceCodeGenUtil.CompareDatabaseSchemas(compilerResults.CompiledAssembly, modelType, cfg,  
            cfg.Driver, DbOptions.Default, cfg.ConnectionString, log);
      string schemaMessage; 
      if (nonProcActions.Count == 0) {
        schemaMessage = "Schema verification completed, schemas are identical.";
      } else {
        schemaMessage = Util.SafeFormat("Schema verification: detected {0} differences (schema update actions).\r\n", nonProcActions.Count);
        schemaMessage += "Schema update actions:\r\n  ";
        schemaMessage += string.Join("\r\n  ", nonProcActions);

        schemaMessage += @"

Note: Non-empty update action list represents the delta between the original database schema and 
      the schema from the generated entities. Ideally this list should be empty. If it is not, 
      then probably some features in your database are not currently supported by VITA. 
      Or it can indicate an error in this entity code generator tool. 
      Please report the incident with some details to the developer on the project website. 
";
      }
      //Dump the message to the source file as comment:
      System.IO.File.AppendAllText(cfg.OutputPath, "\r\n/*\r\n" + schemaMessage + "\r\n*/");
      Console.WriteLine(); 
      Console.WriteLine(schemaMessage);
    }

    private static string GetConfig(string[] args) {
      const string prefix = "/c:";
      if (args.Length > 0) {
        var cfgArg = args.FirstOrDefault(arg => arg.StartsWith(prefix));
        if (cfgArg != null)
          return cfgArg.Substring(prefix.Length);  
      }
      return null;
    }

    private static void ShowHelp() {
      Console.WriteLine("Usage:");
      Console.WriteLine("  entitygen /c:configfile [/noinput] [/?]");
      Console.WriteLine("Switches:");
      Console.WriteLine("    /?       - show help.");
      Console.WriteLine("    /c:      - configuration file name/path.");
      Console.WriteLine("    /noinput - unattended batch mode execution.");
      WaitPressKey(); 
    }

    private static void WaitPressKey() {
      if (_noinput) return; //do not stop in batch mode
      Console.WriteLine("Press any key...");
      Console.ReadKey();
    }

    // If we are running in batch mode, try to log error
    private static void LogError(string message) {
      if (_noinput)
        try { System.IO.File.AppendAllText(ErrorLogFile, message + "\r\n"); } catch { }
    }

  }
}
